/*
 * zen/include/kara_defaults.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef KARA_DEFAULTS_H
#define KARA_DEFAULTS_H
				/* This file is not intended to be
				   included by users.  It is useful
				   only in kara files and should NOT
				   be used elsewhere */

#include <stdio.h>

#ifdef KARA_CUT_SET
# include "karadefs.h"
#endif

#define DAss        1
#define CAdd        2
#define AddC        4
#define Ass         8
#define Add        16
#define Sub        32
#define SAss       64
#define Mix       128
#define MSub      256
#define MK3       512
#define SK2      1024

#include "zbn.h"

#ifndef PURE_KARA
# define PURE_KARA 0
#endif

#ifndef KARA_DEBUG
# define KARA_DEBUG 0		/* 0-5 */
#endif

#ifndef KARAm_LIM
# if PURE_KARA
#  define KARAm_LIM 1		/* do not change this line */
# else
#  ifndef KARA_LIM_SET		/* if set, KARAm_LIM will be a static
                                   variable. This is used to find the best cutoff */
#   define KARAm_LIM 8		/* 8 seems a good default value on most machines */
#  endif
# endif
#endif

#ifndef KARAs_LIM
# if PURE_KARA
#  define KARAs_LIM 1		/* do not change this line */
# else
#  ifndef KARA_LIM_SET		/* if set, KARAs_LIM will be a static variable. This
				   is used to find the best cutoff */
#   define KARAs_LIM 8		/* 8 seems a good default value on most machines */
#  endif
# endif
#endif

#define ZBNMultiply2Digit(x,a,b)		\
if (1) {					\
  BigNumDigit _h, _l, _d;			\
						\
  _d = (b)[0];					\
  zbnmul(_h, _l, (a)[0], _d);			\
  _h += Add2Digit((x)[0], _l);			\
  (x)[1] += _h;					\
} else
#define ZBNSquare1Digit(x,a) ZBNMultiply2Digit(x,a,a)

#define IS_0(x) (!(x))
#define IS_1(x) IS_0(~(x))
#define Add2Digit(a,b) (((a)=(a)+(b))<(b))
#define Sub2Digit(a,b) (save = (a), (((a)=(save)-(b))>(save)))

#if ZBN_KARA_OPTS & CAdd
# define _CompAdd(buf,N) if(1) {ZBNComplement(buf,N); ZBNAddCarry(buf,N,(BigNumDigit)1);} else
#else
# define _CompAdd(buf,N) if(1) {BigNum pta,ptlim; \
	 ptlim = N + (pta = buf) ; \
	 do { *pta = ~(*pta); } while ((!(++(*pta++)))&&(pta < ptlim)); \
	 while(pta < ptlim) {pta++; pta[-1] = ~pta[-1];}} else
#endif

#if ZBN_KARA_OPTS & AddC
# define _AddCarry(buf,N) ZBNAddCarry(buf,N,(BigNumDigit)1)
#else
# define _AddCarry(buf,N) if(1) {BigNum pta,ptlim; \
	   ptlim = N + (pta=buf); \
	   while((!(++(*pta++)))&&(pta < ptlim)); } else
#endif

#if ZBN_KARA_OPTS & Ass
# define _Assign(a,N) ZBNAssign(a+N,a,N)
#else
# define _Assign(a,N) if(1) {BigNum pt1,pt2,ptlim; \
	     ptlim = pt1 = (pt2 = a) + (N); \
	     do { *(pt1++)=*(pt2++) ; } while (pt2 < ptlim); } else
#endif

#if ZBN_KARA_OPTS & Add
# define _Add(a,al,b,bl,cin) ZBNAdd(a,al,b,bl,cin)
#else
# define _Add(a,al,b,bl,cin) if(1) {BigNum pta=a,ptb=b; BigNumDigit save,c=cin; BigNumLength nl=bl; \
      do { if (c) { \
	if(!(~(save = *pta))) { \
		*(pta++)= *(ptb++); c=1; } \
	else { \
		c= *(ptb++); save += c; *(pta ++) = ++save; \
		c= (save < c ? 1 : 0 ); } } \
	    else { \
		c = (save = *(ptb++)) + *pta; *(pta++) = c; \
		c = (c < save ? 1 : 0 ); } } while (--nl); \
	if(c) while((pta < a+al)&&(!(++(*pta++)))) ; \
			      } else		    
#endif

#if ZBN_KARA_OPTS & Sub
# define _Sub(a,al,b,bl,sb) ZBNSubtract(a,al,b,bl,sb)
#else
# define _Sub(a,al,b,bl,sb) if(1) {BigNum pta=a,ptb=b; BigNumDigit invn,c=sb; BigNumLength nl=bl; \
	do { invn = ~(*(ptb++)); if(c) {BigNumDigit save; \
		if(!(~(save = *pta))) { \
			*(pta++)=invn; c=1; } \
		else { \
			save += invn; *(pta++) = ++save; \
			c= (save<invn ? 1 : 0); } } \
	       else { \
		c = invn + *pta; *(pta++) = c ; \
		c = (c<invn ? 1 : 0); } \
	   } while(--nl); \
	if(!c) while((pta < a+al)&&(!((*(pta++))--))); \
			      } else
#endif

#if ZBN_KARA_OPTS & Mix      
# define _Mix(x,N,l,l2,z) if(1) { \
 if(ZBNAdd(x+N,N,x+l,N,(BigNumDigit)0)) _AddCarry(x+l+N,z); \
 _Assign(x+N,N); \
 _Add(x+l,l2,x+l+N,z,(BigNumDigit)0); \
 _Add(x+N,l2+N,x,N,(BigNumDigit)0); } else			  		      
#else
# define _Mix(x,N,l,l2,z) if(1) {BigNum pta,ptb; BigNumDigit save,c; BigNumLength nl; \
      c=0; ptb=N + (pta = x + (nl = N)); \
      do { if (c) { \
	if(!(~(save = *pta))) { \
		*(pta++)= *(ptb++); c=1; } \
	else { \
		c= *ptb; save += c; *(pta ++) = *(ptb++) = ++save; \
		c= (save < c ? 1 : 0 ); } } \
	    else { \
		c = (save = *ptb) + *pta; *(pta++) = *(ptb++) = c; \
		c = (c < save ? 1 : 0 ); } } while (--nl); \
	if(c) while((ptb < x+l+N+z) && (!(++(*ptb++)))); \
	     _Add(x+l,l2,x+l+N,z,(BigNumDigit)0); \
		_Add(x+N,l2+N,x,N,(BigNumDigit)0); \
			      } else			      
#endif
#endif
