/*
 * zen/include/sys.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __SYS__
#define __SYS__

#include "available.h"

				/* This is used when compiling a C++ program
				   against zen headers file */
#if defined(__GNUG__) || defined(__cplusplus)
# define extern extern "C"
#endif

#ifdef ___
#  undef ___
#endif
#ifndef NOCANSI
# define ___(args) args
#else
# define ___(args) ()
#endif

#ifdef TRUE
# undef TRUE
#endif
#define TRUE 1
#ifdef true
# undef true
#endif
#define true 1

#ifdef FALSE
# undef FALSE
#endif
#define FALSE 0
#ifdef false
# undef false
#endif
#define false 0

/* The value of |zen_error.lib| is one of the following. */
typedef enum {
  ZEN_LIB_UNKNOWN= 0,
  ZEN_LIB_ZBN= 1,
  ZEN_LIB_ZBZ= 2,
  ZEN_LIB_ZE2= 3,
  ZEN_LIB_ZEPS= 4,
  ZEN_LIB_ZEXT= 5,
  ZEN_LIB_ZEG= 6,
  ZEN_LIB_ZED= 7,
  ZEN_LIB_ZEP= 8,
  ZEN_LIB_ZETAB= 9,
  ZEN_LIB_ZELOG= 10,
  ZEN_LIB_ZER= 11,
  ZEN_LIB_ZEF= 12,
  ZEN_LIB_ZEC= 13,
  ZEN_LIB_ZEM= 14
} arith;

/* \TEX{

   Procedures of the |ZEN| library raise errors by their outputs.
   There are 2 types of errors, system and mathematical errors.

   System errors occur when a function returns |ZENERR| or
   |ZENNULL|. In that case, the internal global variable |zen_error|
   the type of which is |zen_err| was set by this function with
   |ZENSetError()| (see the file |sys.h|). Here, the user only has to
   call the function |ZENError()| in its code to handle this error.

   Most functions of the library can be used in any rings (only few of
   them can only apply in finite fields) even if they are not defined
   everywhere.  Mathematical errors generally occures when a function
   returns |ZEN_NO_INVERSE|.  That means that somewhere in the code, a
   function tried to inverse an element $e$ modulo a non prime integer
   $n$ or a non irreducible polynomial $P(X)$ and that the gcd of $e$
   with $n$ or $P(X)$ is not 1. In that case, a factor of the
   corresponding modulo is put in |ZENRingFact(Rg)| where |Rg| is the
   current finite ring.

   The situation can be sligtly more complicated when building several
   not prime or not irreducible finite rings. For instance, when you are
   first working in $R_0=\Z/15\Z$ and then in $R_1=R_0[T]/(T^2+4T+3)$.
   On the first hand, a call to |ZENEltInverse()| with argument $2T+6$
   will return |ZEN_NO_INVERSE| and a call to |ZENRingFact(|$R_1$|)|
   will return $T+3$. On the other hand, a call to |ZENEltInverse()|
   with argument $3T+6$ will return |ZEN_NO_INVERSE| and a call to
   |ZENRingFact(|$R_1$|)| will return $3$. Here, the degree of
   |ZENRingFact(|$R_1$|)| is 0 and so, we can call
   |ZENRingFact(|$R_0$|)| to finally get $3$, a factor of $15$, the
   modulo of the first finite ring.
   } */

/*
   \TEX{
   The values of |ZENERR| and |ZENNULL| are defined as follows:
   }
   */
#define ZENWGINPUT -2		/* used only in BigNum by Karatsuba's
				   procedures
				   \CiteC   */
#define ZENERR -1
#define ZENNULL NULL				/* \EndCiteC */
#if ZEN_NO_TEST
# define ZERR ZENERR
# define ZNULL ZENNULL
#else
/* \TEX{ In fact, in case of error, the functions of \zen\ return one
   of the following }
   \CiteC */
# define ZERR _ZENERR()
# define ZNULL _ZENNULL() /* \EndCiteC */
#endif
/* \TEX{ This is intended to allow interactive debugging by giving the
   possibility of setting breakpoints in the functions |_ZENERR| and
   |_ZENNULL|.  Therefore this feature is disabled in optimized
   compilation. In that case, most of the tests are skipped, except
   those that check if an inverse was impossible to compute.

   In debugging mode, the flag |debugflag| can be set to non zero.
   This activate verbose error outputs. Each time one of the two functions |_ZENERR()| or |_ZENNULL| is called, an error message as printed by |ZENError()| is printed on the standard error output, according to the value of |zen_error|
   The real structure of |zen_error| is as follows.
   }
   \CiteC
*/
typedef struct zen_err{
  arith lib;
  int fct;
  int err;
} zen_err;
extern zen_err zen_error;
extern int debugflag; /* \EndCiteC */
/* \TEX{
   The internal functions of |ZEN| raised errors with the
   function |ZENSetError()|.
   }

   The value of |zen_error.err| is one of the following.
   */

#define ZEN_UNKNOWN      0

#define ZEN_ALLOC        1      /* "Malloc error" */
#define ZEN_SPRINTF      2      /* "Sprintf error" */
#define ZEN_SSCANF       3      /* "Sscanf error" */
#define ZEN_FPRINTF      4      /* "Fprintf error" */
#define ZEN_FSCANF       5      /* "Fscanf error" */
#define ZEN_SYNTAX       6      /* "Syntax error" */
#define ZEN_RGCLOSE      7      /* "Ring freed before its extensions" */
#define ZEN_STRCAT       8      /* "Strcat error" */
#define ZEN_NONMONIC     9      /* "Non invertible highest coefficient in the definition polynomial" */
#define ZEN_MAT_DIM     10      /* "Matrix dimensions error" */
#define ZEN_MAT_DIM0    11      /* "Uncorrect matrix dimension" */
#define ZEN_NOTPERM     12      /* "Permutation operation on non permutation matrix" */
#define ZEN_ISPERM      13      /* "Assigning a permutation matrix with non permutation" */
#define ZEN_MAT_TYPE    14      /* "Uncorrect type of matrix" */
#define ZEN_MAT_NOTINV  15      /* "Matrix should be invertible" */
#define ZEN_DIVBYZERO   16      /* "Division by 0" */
#define ZEN_TOO_LARGE   17      /* "Too big ZENRing for tabulating" */
#define ZEN_NOT_EXT     18      /* "This ZENRing is not an extension" */
#define ZEN_UNAVAILABLE 19      /* "This arithmetic is unavailable" */
#define ZEN_INPUT       20      /* "Wrong input" */
#define ZEN_ELT_NOTINV  21      /* "Wrong input" */
#define ZEN_ELT_NOTFF   22      /* "This ZENRing is not a finite field" */

/* The value of |zen_error.fct| is one of the following. */

#define ZEN_WEIGHT              1
#define ZEN_GCDPLAIN            2
#define ZEN_GCDBINARY           3
#define ZEN_EEAPLAIN            4
#define ZEN_EEALEHMER           5
#define ZEN_GCD                 6
#define ZEN_EEA                 7
#define ZEN_PRINTTOSTRING       8
#define ZEN_READFROMSTRING      9
#define ZEN_PUTTOSTRING         10              
#define ZEN_GETFROMSTRING       11
#define ZEN_PRINTTOFILE         12
#define ZEN_READFROMFILE        13      
#define ZEN_PUTTOFILE           14
#define ZEN_GETFROMFILE         15
#define ZEN_INIT                16
#define ZEN_CLOSE               17
#define ZEN_MODULO              18
#define ZEN_ZENELTISASQRCP      19
#define ZEN_ZENELTSQRTCP        20
#define ZEN_INVERSE             21
#define ZEN_ADDPRC              22
#define ZEN_EXPPRC              23
#define ZEN_MULTKARA            24
#define ZEN_SQUKARA             25
#define ZEN_ELTTOZ              26
#define ZEN_ZTOELT              27
#define ZEN_KARAMC              28
#define ZEN_KARASC              29
#define ZEN_KARAPRC             30
#define ZEN_MALLOC              31
#define ZEN_SQRT                32
#define ZEN_ELTMULTPRC          33
#define ZEF_FACTORIAL           34
#define ZEN_CONVERT             35
#define ZEN_EXP                 36
#define ZEN_ELTTRACPRC          37

#define ____ZEN_BASE            37 /* this is the number of basic errors */

#define ZEN_POLYCREATE          101
#define ZEN_POLYPRINTTOSTRING   102
#define ZEN_POLYREADFROMSTRING  103
#define ZEN_POLYPRINTTOFILE     104
#define ZEN_POLYREADFROMFILE    105
#define ZEN_POLYPUTTOSTRING     106
#define ZEN_POLYGETFROMSTRING   107     
#define ZEN_POLYPUTTOFILE       108
#define ZEN_POLYGETFROMFILE     109
#define ZEN_POLYCREATEFROMFILE  110
#define ZEN_POLYTOZ             111
#define ZEN_ZTOPOLY             112

#define ____ZEN_POLYN            12 /* this is the number of poly errors */

#define ZEN_MATADD              201
#define ZEN_MATNEGATE           202
#define ZEN_MATSUBTRACT         203
#define ZEN_MATASSIGN           204
#define ZEN_MAT_CREATE          205 
#define ZEN_MATCREPMS           206
#define ZEN_MATGETBLOC          207
#define ZEN_MATSETBLOC          208
#define ZEN_MATGETDIGIT         209
#define ZEN_MATSETDIGIT         211
#define ZEN_MATMULTIPLY         211
#define ZEN_MATWINOGRAD         212
#define ZEN_MATSETZERO          213
#define ZEN_MATSETONE           214
#define ZEN_MATSETRANDOM        215
#define ZEN_MATPERMUTEBNS       216
#define ZEN_MATPERMUTEDG        217
#define ZEN_MATMULLEFTPERM      218
#define ZEN_MATMULRIGHTPERM     219
#define ZEN_MATGAUSS            220
#define ZEN_MATGAUSSINV         221
#define ZEN_MATINVERSE          222
#define ZEN_MATKERNEL           223
#define ZEN_MATPRINTTOSTRING    224
#define ZEN_MATREADFROMSTRING   225
#define ZEN_MATPRINTTOFILE      226
#define ZEN_MATREADFROMFILE     227
#define ZEN_MATPUTTOSTRING      228
#define ZEN_MATGETFROMSTRING    229     
#define ZEN_MATPUTTOFILE        230
#define ZEN_MATGETFROMFILE      231
#define ZEN_MATMULBNPERM        232
#define ZEN_MATMULSTPERMDG      233
#define ZEN_MATMULOTPERMDG      234
#define ZEN_MATGAUSSTRIG        235
#define ZEN_MATMULPERM          236
#define ZEN_MATRANK             237
#define ZEN_MATSETSUBMAT        238
#define ZEN_MATGETSUBMAT        239
#define ZEN_MATCONVERT          240
#define ZEN_VECT2MAT            241
#define ZEN_MAT2VECT            242

#define ____ZEN_MAT              42 /* this is the number of mat errors */

#define ZEN_PERMCREATE          301
#define ZEN_PERMASSIGN          302
#define ZEN_PERMSETZERO         303
#define ZEN_PERMSETONE          304
#define ZEN_PERMSETRANDOM       305
#define ZEN_PERMGETBLOC         306
#define ZEN_PERMSETBLOC         307
#define ZEN_PERMGETDIGIT        308
#define ZEN_PERMSETDIGIT        309
#define ZEN_PERMPERMUTEBNS      310
#define ZEN_PERMISONE           311

#define ____ZEN_PERM             11 /* this is the number of permutation errors */

#define ZEN_SRCREATE            401
#define ZEN_SRPRINTTOSTRING     402
#define ZEN_SRREADFROMSTRING    403
#define ZEN_SRPRINTTOFILE       404
#define ZEN_SRREADFROMFILE      405
#define ZEN_SRPUTTOSTRING       406
#define ZEN_SRGETFROMSTRING     407     
#define ZEN_SRPUTTOFILE         408
#define ZEN_SRGETFROMFILE       409
#define ZEN_SRCREATEFROMFILE    410

#define ____ZEN_SR            10 /* this is the number of series errors */

#define ZEN_ECCREATE            501
#define ZEN_ECPTCREATE          502
#define ZEN_ECPTPRINTTOSTRING   503
#define ZEN_ECPTREADFROMSTRING  504
#define ZEN_ECPTPRINTTOFILE     505
#define ZEN_ECPTREADFROMFILE    506
#define ZEN_ECPTPUTTOSTRING     507
#define ZEN_ECPTGETFROMSTRING   508     
#define ZEN_ECPTPUTTOFILE       509
#define ZEN_ECPTGETFROMFILE     510
#define ZEN_ECPTCREATEFROMFILE  511
#define ZEN_ECPTMULTIPLY        512
#define ZEN_ECPTADD             513

#define ____ZEN_EC            12 /* this is the number of elliptic curve errors */


/* Handling errors. */
# define ZENErrLib zen_error.lib
# define ZENErrFct zen_error.fct
# define ZENErrErrno zen_error.err
#if 0
/*
   \TEX{
   \title{Setting the error condition flags}
   \input{The three flags describing the error exception:
   \begin{itemize}
   \item the arithmetic in which it occured,
   \item the function of |ZEN| in which it occured,
   \item the type of error.
   \end{itemize}
   }
   \side{The |zen_error| global structure is set.}
   \note{This is not a user purpose function. It is described here to
   inform users about the way the errors are handled by the library.}
   }
   */
void ZENSetError(arit,function,error)
     arith arit;
     int function,error; {}
#endif
# define ZENSetError(lb,ft,er) (ZENErrLib = lb, ZENErrFct = ft,ZENErrErrno  = er)

#ifndef ZEN_NO_TEST
# define ZENCheckError(X,ret) if(X) return(ret); else
# define ZENTest(X,er,ft,lb,ret) if(X) \
  {ZENSetError(lb,ft,er) ; return(ret); } else
#else
# define ZENCheckError(X,ret) (X)
# define ZENTest(X,err,fct,lib,ret) 
#endif

/* Procedure to handle errors
 ****************************/
extern void ZENError ___((void));
extern int _ZENERR ___((void));
extern void * _ZENNULL ___((void));

#if !ZEN_MALLOC_TEST
# define ZENMalloc(nl) ZENMallocPtr((size_t)(nl))
# define ZENRealloc(n,nl) ZENReallocPtr((void *)(n), (size_t)(nl))
# define ZENFree(n) ZENFreePtr((void *)(n))
/* Procedure to handle allocation
*********************************/
extern void * (*ZENMallocPtr) ___((size_t));
extern void * (*ZENReallocPtr) ___((void *, size_t));
extern void (*ZENFreePtr) ___((void *));

extern void ZENSetMemoryFunctions ___((
     void * (*) ___((size_t)),
     void * (*) ___((void *, size_t)),
     void   (*) ___((void *))));

#else  /* ZEN_MALLOC_TEST */
# define ZENMalloc(nl) ZENMallocPtr((size_t)(nl),__FILE__,__LINE__)
# define ZENRealloc(n,nl) ZENReallocPtr((void *)(n), (size_t)(nl),__FILE__,__LINE__)
# define ZENFree(n) ZENFreePtr((void *)(n),__FILE__,__LINE__)
/* Procedure to handle allocation
*********************************/
extern void * (*ZENMallocPtr) ___((size_t, char *, int));
extern void * (*ZENReallocPtr) ___((void *, size_t, char *, int));
extern void (*ZENFreePtr) ___((void *,char *, int));

extern void ZENSetMemoryFunctions ___((
     void * (*) ___((size_t, char *, int)),
     void * (*) ___((void *, size_t, char *, int)),
     void   (*) ___((void *, char *, int))));

#endif  /* ZEN_MALLOC_TEST */

/* Procedure to generate random values
 **************************************/
#define zsrandom(seed, seed_len) _zsrandom(seed, seed_len)
#define zrandom()                _zrandom()

extern void          (* _zsrandom) ___((unsigned char *, size_t));
extern unsigned char (* _zrandom)  ___((void));

extern unsigned char ZENNativeRand         ___((void));
extern void          ZENNativeSrand        ___((unsigned char *, size_t));
extern int           ZENSetRandomFunctions ___((int (*in) (void), void (*sr) (unsigned char *, size_t), unsigned char (*ra) (void)));

/* Procedure to get user time
 ****************************/
extern double runtime ___((void));
#endif
/*
   \TEX{
\input{error}   
   } 
   */

