/*
 * zen/include/zbnconfig.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsubsection{Assembler directives}
   } */

/* \TEX{
   Unlike the library |BigNum| or the library |GMP|, one important
   principle followed while designing |ZBN| is to minimize as much as
   possible code written in assembler. There are two important
   benefits for such an approach:
   \begin{itemize}
   %
   \item It is no more a problem to port the library to a new
   architecture. In particular, to optimize it since we only have to
   write only few assembler sentences.
   %
   \item Compilers are becoming more and more powerfull. They are able
   to optimize codes much more efficiently than us (unrolling loops,
   managing cache misses, \ldots).
   %
   \end{itemize}
   That's why, we limited our assembler directives to the few
   arithmetic assembler instructions which are not easily accessed in
   C. The coresponding macros are described below.
   } */

/* Linux or cygnus, 386, gcc
 ***************************/
#if (defined(linux386) || defined(cygwin386)) && defined(__GNUC__)
#if !defined(LONG_LONG) || (LONG_LONG == NO)
# define BN_DIGIT_SIZE 32

typedef unsigned long BigNumDigit;
# define CST_BNDIGIT(d) d##U

typedef long BigNumLength;
# define CST_BNLENGTH(d) d##L

#ifndef NO_ASM
# define zbnadd(c1, c0, a1, a0, b1, b0)		\
do {						\
  BigNumDigit _c1 = a1, _c0 = a0;		\
						\
  __asm__("addl %5,%1\n adcl %3,%0"		\
	  : "=g"  ((BigNumDigit)(_c1)),		\
	    "=&g" ((BigNumDigit)(_c0))		\
	  : "%0"  ((BigNumDigit)(a1)),		\
	    "ri"   ((BigNumDigit)(b1)),		\
	    "%1"  ((BigNumDigit)(a0)),		\
	    "ri"   ((BigNumDigit)(b0)));	\
  c1 = _c1;  c0 = _c0;				\
} while(0)

# define zbnsub(c1, c0, a1, a0, b1, b0)		\
do {						\
  BigNumDigit _c1 = c1, _c0 = c0;		\
						\
__asm__("subl %5,%1\nsbbl %3,%0"		\
	: "=g"  ((BigNumDigit)(_c1)),		\
	  "=&g" ((BigNumDigit)(_c0))		\
	: "0"   ((BigNumDigit)(a1)),		\
	  "ri"   ((BigNumDigit)(b1)),		\
	  "1"   ((BigNumDigit)(a0)),		\
	  "ri"   ((BigNumDigit)(b0)));		\
  c1 = _c1;  c0 = _c0;				\
} while(0)

# define zbnmul(c1, c0, a, b)			\
__asm__("mull %3"				\
	: "=a" ((BigNumDigit)(c0)),		\
	  "=d" ((BigNumDigit)(c1))		\
	: "%0" ((BigNumDigit)(a)),		\
	  "rm" ((BigNumDigit)(b)))

# define zbnmuladds(h, m, n, d, c)						\
do {										\
  BigNumDigit __dummy;								\
__asm__ ("mull %3\naddl %5, %%eax\nadcl $0,%1\naddl %%eax,%0\nadcl $0,%1"	\
	 : "=g"       ((BigNumDigit)(m)),					\
	   "=&d"      ((BigNumDigit)(h)),					\
	   "=a"       ((BigNumDigit)(__dummy))					\
	 : "rm"       ((BigNumDigit)(d)),					\
	   "2"        ((BigNumDigit)(n)),					\
	   "g"        ((BigNumDigit)(c)));					\
} while(0)

# define zbndiv(q, r, n1, n0, d)		\
__asm__("divl %4"				\
	: "=a" ((BigNumDigit)(q)),		\
	  "=d" ((BigNumDigit)(r))		\
	: "0"  ((BigNumDigit)(n0)),		\
	  "1"  ((BigNumDigit)(n1)),		\
	  "rm" ((BigNumDigit)(d)))
#endif

# ifdef  KARA_DEFAULTS_H
#  define ZBN_KARA_OPTS (DAss  | Add | Sub | Mix | MSub | MK3 | SK2)   
#  define SK2_NOSQ 1 
# endif
#endif
#endif

/* Solaris sparc computers
 ***************************/
#if defined(solarissparc)
#if !defined(LONG_LONG) || (LONG_LONG == NO)
# define BN_DIGIT_SIZE 32

typedef unsigned long BigNumDigit;
# define CST_BNDIGIT(d) d##U

typedef long BigNumLength;
# define CST_BNLENGTH(d) d##L

#ifndef NO_ASM
# if defined(__GNUC__)
#  define zbnadd(c1, c0, a1, a0, b1, b0)	\
__asm__("addcc %r2,%3,%0\naddx %r4,%5,%1"	\
	: "=r"  ((BigNumDigit)(c0)),		\
	  "=&r" ((BigNumDigit)(c1))		\
	: "%rJ" ((BigNumDigit)(a0)),		\
	  "rI"  ((BigNumDigit)(b0)),		\
	  "%rJ" ((BigNumDigit)(a1)),		\
	  "rI"  ((BigNumDigit)(b1)))

#  define zbnsub(c1, c0, a1, a0, b1, b0)	\
__asm__("subcc %r2,%3,%0\nsubx %r4,%5,%1"	\
	: "=r"  ((BigNumDigit)(c0)),		\
	  "=&r" ((BigNumDigit)(c1))		\
	: "%rJ" ((BigNumDigit)(a0)),		\
	  "rI"  ((BigNumDigit)(b0)),		\
	  "%rJ" ((BigNumDigit)(a1)),		\
	  "rI"  ((BigNumDigit)(b1)))
#  define zbnmul(c1, c0, a, b)			\
__asm__("umul %2,%3,%0\nrd %%y,%1"		\
	: "=r" ((BigNumDigit)(c0)),		\
	  "=r" ((BigNumDigit)(c1))		\
	: "r" ((BigNumDigit)(a)),		\
	  "r" ((BigNumDigit)(b)))
#  define zbndiv(q, r, n1, n0, d)		\
do {						\
  BigNumDigit _qp;				\
    __asm__ ("mov %1,%%y;udiv %2,%3,%0"		\
	     : "=r" ((BigNumDigit)(_qp))	\
	     : "r" ((BigNumDigit)(n1)),		\
	       "r" ((BigNumDigit)(n0)),		\
	       "r" ((BigNumDigit)(d)));		\
    (r) = (n0) - _qp * (d);			\
    (q) = _qp;					\
} while(0)

/* -O2 -mv8 */
# endif
#endif

# if defined(__SUNPRO_C)
#  define zbnmul(c1, c0, a, b) _zbnmul(&(c1), &(c0), a, b)
static void _zbnmul(p_c1, p_c0, a, b)
     BigNumDigit *p_c1, *p_c0, a, b;
{
  unsigned long long _cp;

  _cp = ((unsigned long long)(a))*(b);
  *p_c0 = _cp;
  *p_c1 = _cp >> BN_DIGIT_SIZE;
}
/* -fast */
# endif

# ifdef  KARA_DEFAULTS_H
#  define ZBN_KARA_OPTS (Add | Sub | SK2)
#  define SK2_NOSQ 1 
#endif

#endif
#endif

/* OSF1, alpha computers
 **************************/
#ifdef osfalpha
# define BN_DIGIT_SIZE 64

typedef unsigned long BigNumDigit;
# define CST_BNDIGIT(d) d##UL

typedef long BigNumLength;
# define CST_BNLENGTH(d) d##L

#ifndef NO_ASM
# ifdef __GNUC__
#  define zbnmul(c1, c0, a, b)			\
do {						\
  BigNumDigit _cp;				\
						\
  __asm__ ("umulh %r1,%2,%0"			\
	   : "=r" (_cp)				\
	   : "%rJ" ((BigNumDigit)(a)),		\
	     "rI"  ((BigNumDigit)(b)));		\
  (c0) = (a)*(b);				\
  (c1) = _cp;					\
} while(0)
/* -O3 */
# endif

# ifdef __DECC
#  include <c_asm.h>
#  define zbnmul(c1, c0, a, b)				\
do {							\
  BigNumDigit _cp;					\
							\
  _cp   = asm("umulh  %a0, %a1, %v0", (a), (b));	\
  (c0) = (a)*(b);					\
  (c1) = _cp;						\
} while(0)
/*  -arch host -fast -tune host -g1 -O4 -inline speed -Olimit 9000 */
# endif
#endif

# ifdef  KARA_DEFAULTS_H
/* very difficult to define, as the optimizer takes the final
   executable as input. Hence, squaring options seem to greatly
   decrease multiplication performances ! */
#  define ZBN_KARA_OPTS (Ass | Sub | MSub | MK3 ) 
#  define SK2_NOSQ 0
# endif

#endif

/* Aix, rs6000
 **************/
#if defined(aixrs6000) || defined(aixppc)
#if !defined(LONG_LONG) || (LONG_LONG == NO)
# define BN_DIGIT_SIZE 32

typedef unsigned long BigNumDigit;
# define CST_BNDIGIT(d) d##U

typedef long BigNumLength;
# define CST_BNLENGTH(d) d##L

#ifndef NO_ASM
# ifdef __GNUC__
/* Included from GMP-2.0.2 */
#define zbnadd(sh, sl, ah, al, bh, bl)					\
  do {									\
    if (__builtin_constant_p (bh) && (bh) == 0)				\
      __asm__ ("{a%I4|add%I4c} %1,%3,%4\n\t{aze|addze} %0,%2"		\
	     : "=r" ((BigNumDigit)(sh)),				\
	       "=&r" ((BigNumDigit)(sl))				\
	     : "%r" ((BigNumDigit)(ah)),				\
	       "%r" ((BigNumDigit)(al)),				\
	       "rI" ((BigNumDigit)(bl)));				\
    else if (__builtin_constant_p (bh) && (bh) ==~(BigNumDigit) 0)	\
      __asm__ ("{a%I4|add%I4c} %1,%3,%4\n\t{ame|addme} %0,%2"		\
	     : "=r" ((BigNumDigit)(sh)),				\
	       "=&r" ((BigNumDigit)(sl))				\
	     : "%r" ((BigNumDigit)(ah)),				\
	       "%r" ((BigNumDigit)(al)),				\
	       "rI" ((BigNumDigit)(bl)));				\
    else								\
      __asm__ ("{a%I5|add%I5c} %1,%4,%5\n\t{ae|adde} %0,%2,%3"		\
	     : "=r" ((BigNumDigit)(sh)),				\
	       "=&r" ((BigNumDigit)(sl))				\
	     : "%r" ((BigNumDigit)(ah)),				\
	       "r" ((BigNumDigit)(bh)),					\
	       "%r" ((BigNumDigit)(al)),				\
	       "rI" ((BigNumDigit)(bl)));				\
  } while (0)
#define zbnsub(sh, sl, ah, al, bh, bl)					\
  do {									\
    if (__builtin_constant_p (ah) && (ah) == 0)				\
      __asm__ ("{sf%I3|subf%I3c} %1,%4,%3\n\t{sfze|subfze} %0,%2"	\
	       : "=r" ((BigNumDigit)(sh)),				\
		 "=&r" ((BigNumDigit)(sl))				\
	       : "r" ((BigNumDigit)(bh)),				\
		 "rI" ((BigNumDigit)(al)),				\
		 "r" ((BigNumDigit)(bl)));				\
    else if (__builtin_constant_p (ah) && (ah) ==~(BigNumDigit) 0)	\
      __asm__ ("{sf%I3|subf%I3c} %1,%4,%3\n\t{sfme|subfme} %0,%2"	\
	       : "=r" ((BigNumDigit)(sh)),				\
		 "=&r" ((BigNumDigit)(sl))				\
	       : "r" ((BigNumDigit)(bh)),				\
		 "rI" ((BigNumDigit)(al)),				\
		 "r" ((BigNumDigit)(bl)));				\
    else if (__builtin_constant_p (bh) && (bh) == 0)			\
      __asm__ ("{sf%I3|subf%I3c} %1,%4,%3\n\t{ame|addme} %0,%2"		\
	       : "=r" ((BigNumDigit)(sh)),				\
		 "=&r" ((BigNumDigit)(sl))				\
	       : "r" ((BigNumDigit)(ah)),				\
		 "rI" ((BigNumDigit)(al)),				\
		 "r" ((BigNumDigit)(bl)));				\
    else if (__builtin_constant_p (bh) && (bh) ==~(BigNumDigit) 0)	\
      __asm__ ("{sf%I3|subf%I3c} %1,%4,%3\n\t{aze|addze} %0,%2"		\
	       : "=r" ((BigNumDigit)(sh)),				\
		 "=&r" ((BigNumDigit)(sl))				\
	       : "r" ((BigNumDigit)(ah)),				\
		 "rI" ((BigNumDigit)(al)),				\
		 "r" ((BigNumDigit)(bl)));				\
    else								\
      __asm__ ("{sf%I4|subf%I4c} %1,%5,%4\n\t{sfe|subfe} %0,%3,%2"	\
	       : "=r" ((BigNumDigit)(sh)),				\
		 "=&r" ((BigNumDigit)(sl))				\
	       : "r" ((BigNumDigit)(ah)),				\
		 "r" ((BigNumDigit)(bh)),				\
		 "rI" ((BigNumDigit)(al)),				\
		 "r" ((BigNumDigit)(bl)));				\
  } while (0)
#  if defined (_ARCH_PPC)
#define zbnmul(ph, pl, m0, m1)			\
  do {						\
    BigNumDigit __m0 = (m0), __m1 = (m1);	\
    __asm__ ("mulhwu %0,%1,%2"			\
	     : "=r" ((BigNumDigit) ph)		\
	     : "%r" (__m0),			\
	       "r" (__m1));			\
    (pl) = __m0 * __m1;				\
  } while (0)
#  else
#define zbnmul(xh, xl, m0, m1)			\
  do {						\
    BigNumDigit __m0 = (m0), __m1 = (m1);	\
    __asm__ ("mul %0,%2,%3"			\
	     : "=r" ((BigNumDigit)(xh)),	\
	       "=q" ((BigNumDigit)(xl))		\
	     : "r" (__m0),			\
	       "r" (__m1));			\
    (xh) += ((((int) __m0 >> 31) & __m1)	\
	     + (((int) __m1 >> 31) & __m0));	\
  } while (0)
#  endif
# endif
#endif

#endif
#endif

/* HPUX, PA-RISC1.
 ****************/
#if defined(hppa1)
#if !defined(LONG_LONG) || (LONG_LONG == NO)
# define BN_DIGIT_SIZE 32

typedef unsigned long BigNumDigit;
# define CST_BNDIGIT(d) d##U

typedef long BigNumLength;
# define CST_BNLENGTH(d) d##L

#ifndef NO_ASM
# if defined(__GNUC__)
#define zbnadd(c1, c0, a1, a0, b1, b0)		\
__asm__ ("add %4,%5,%1\naddc %2,%3,%0"		\
	 : "=r"  ((BigNumDigit)(c1)),		\
	   "=&r" ((BigNumDigit)(c0))		\
	 : "%rM" ((BigNumDigit)(a1)),		\
	   "rM"  ((BigNumDigit)(b1)),		\
	   "%rM" ((BigNumDigit)(a0)),		\
	   "rM"  ((BigNumDigit)(b0)))
     
#define zbnsub(c1, c0, a1, a0, b1, b0)		\
__asm__ ("sub %4,%5,%1\nsubb %2,%3,%0"		\
	 : "=r"  ((BigNumDigit)(c1)),		\
	   "=&r" ((BigNumDigit)(c0))		\
	 : "rM"  ((BigNumDigit)(a1)),		\
	   "rM"  ((BigNumDigit)(b1)),		\
	   "rM"  ((BigNumDigit)(a0)),		\
	   "rM"  ((BigNumDigit)(b0)))

#  if defined (_PA_RISC1_1)
#define zbnmul(c1, c0, a, b)			\
do {						\
  union {					\
    unsigned long long __c;			\
    struct {BigNumDigit __c1, __c0;} __c10;	\
  } __cu;					\
  __asm__ ("xmpyu %1,%2,%0"			\
	   : "=*f" (__cu.__c10)			\
	   : "*f"  ((BigNumDigit)(a)),		\
	     "*f"  ((BigNumDigit)(b)));		\
  (c1) = __cu.__c10.__c1;			\
  (c0) = __cu.__c10.__c0;			\
} while (0)
#  endif
# endif
#endif

#endif
#endif

/* HPUX, PA-RISC2.
 ****************/
#if defined(hppa2)
# define BN_DIGIT_SIZE 64

typedef unsigned long long BigNumDigit;
# define CST_BNDIGIT(d) d##ULL

typedef long long BigNumLength;
# define CST_BNLENGTH(d) d##LL
#endif

/* Default types 
 ****************/
#ifndef BN_DIGIT_SIZE
# if !defined(LONG_LONG) || (LONG_LONG == NO)
#  define BN_DIGIT_SIZE 32

typedef unsigned long BigNumDigit;
#  define CST_BNDIGIT(d) d

typedef long BigNumLength;
#  define CST_BNLENGTH(d) d
# else
#  define BN_DIGIT_SIZE 64

typedef unsigned long long BigNumDigit;
#  define CST_BNDIGIT(d) d##ULL

typedef long long BigNumLength;
#  define CST_BNLENGTH(d) d##LL
# endif
#endif

#ifndef BN_BYTE_SIZE
# define BN_BYTE_SIZE 8
#endif

/* Default macros
 ****************/
#ifndef zbnlow
# define zbnlow(t)       (((BigNumDigit)(t))&((CST_BNDIGIT(1)<<(BN_DIGIT_SIZE/2))-1))
#endif

#ifndef zbnhigh
# define zbnhigh(t)      (((BigNumDigit)(t))>>(BN_DIGIT_SIZE/2))
#endif

#ifndef zbnhighlow
# define zbnhighlow(h,l) ((((BigNumDigit)h)<<(BN_DIGIT_SIZE/2))|((BigNumDigit)l))
#endif

#if 0
/* \TEX{
   \title{Adding two digits}
   \input{Four digits |a0| and |b0|, |a1| and |b1| to add.}
   \side{$|c0| = |a0|+|b0| \bmod 2^{|SIZE_BLOC|}$ and $|c1| =
   |a1|+|b1|+((|a0|+|b0|)\div 2^{|SIZE_BLOC|}) \bmod 2^{|SIZE_BLOC|}$.}
   \note{This macro is sometimes an assembler directive.}
   } */
void zbnadd(c1, c0, a1, a0, b1, b0)
     BigNumDigit c1, c0;
     BigNumDigit a1, a0;
     BigNumDigit b1, b0;{}
#endif
#ifndef zbnadd
# define zbnadd(c1, c0, a1, a0, b1, b0)		\
if (1) {					\
  BigNumDigit _cp;				\
						\
  _cp = (a0)+(b0);				\
  (c1) = (a1)+(b1)+(_cp < (a0));	\
  (c0) = _cp;					\
} else
#endif

#if 0
/* \TEX{
   \title{Subtracting two digits}
   \input{Four digits |a0| and |b0|, |a1| and |b1| to subtract.}
   \side{$|c0| = |a0|-|b0| \bmod 2^{|SIZE_BLOC|} $ and $|c1| =
   |a1|-|b1|-((|a0|-|b0|\bmod 2^{|SIZE_BLOC|+1})\div 2^{|SIZE_BLOC|})
   \bmod 2^{|SIZE_BLOC|}$.}
   \note{This macro is sometimes an assembler directive.}
   } */
void zbnsub(c1, c0, a1, a0, b1, b0)
     BigNumDigit c1, c0;
     BigNumDigit a1, a0;
     BigNumDigit b1, b0;{}
#endif
#ifndef zbnsub
# define zbnsub(c1, c0, a1, a0, b1, b0)		\
if (1) {					\
  BigNumDigit _cp;				\
						\
  _cp = (a0)-(b0);				\
  (c1) = _cp > (a0) ? (a1)-(b1)-1 : (a1)-(b1);	\
  (c0) = _cp;					\
} else
#endif

#if 0
/* \TEX{
   \title{Multiplying digits}
   \input{Two digits |a| and |b|.}
   \side{$|c0| = |a0|\times|b0| \bmod 2^{|SIZE_BLOC|}$ and $|c1| =
   |a1|\times|b1|+((|a0|\times|b0|)\div 2^{|SIZE_BLOC|})
   \bmod 2^{|SIZE_BLOC|}$.}
   \note{This macro is sometimes an assembler directive.}
   } */
void zbnmul(c1, c0, a, b)
     BigNumDigit c1, c0;
     BigNumDigit a;
     BigNumDigit b;{}
#endif
#ifndef zbnmul
# define zbnmul(c1, c0, a, b)				\
if (1) {						\
  BigNumDigit _cp0, _cp1a, _cp1b, _cp2;			\
							\
  _cp0 = zbnlow(a)*zbnlow(b);				\
							\
  _cp1a =  zbnlow(a)*zbnhigh(b);			\
  _cp1a += zbnhigh(_cp0);				\
  _cp1b =  zbnhigh(a)*zbnlow(b);			\
  _cp1a += _cp1b;					\
							\
  _cp2 = zbnhigh(a)*zbnhigh(b);				\
  if (_cp1a < _cp1b)					\
    _cp2 += CST_BNDIGIT(1)<<(BN_DIGIT_SIZE/2);		\
							\
  (c0) = zbnhighlow(zbnlow(_cp1a),zbnlow(_cp0));	\
  (c1) = _cp2+zbnhigh(_cp1a);				\
} else
#endif

#if 0
/* \TEX{
   \title{Dividing 2  digits}
   \input{Two digits |n1| and |n0| to be divided by $|d|$. One must
   have $|d| \div 2^{|SIZE_BLOC|-1} = 1$.}
   \side{$|q| = (|n1|\times2^{|SIZE_BLOC|}+|n0|)\div |d| $ and $|r| =
   (|n1|\times2^{|SIZE_BLOC|}+|n0|)\bmod |d|$.}
   \note{This macro is sometimes an assembler directive.}
   } */
void zbndiv(q, r, n1, n0, d)
     BigNumDigit q;
     BigNumDigit r;
     BigNumDigit n1, n0;
     BigNumDigit d;{}
#endif
/* From Knuth, ``Semi numerical algorithms'', vol 2, pp 235--238. */
#ifndef zbndiv
#define zbndiv(q, r, n1, n0, d)			\
        zbndiv_normalized(q, r, n1, n0, d)
#define zbndiv_normalized(q, r, n1, n0, d)	\
if (1) {					\
  BigNumDigit _da, _db, _qa, _qb, _rp, _p;	\
						\
  _da = zbnhigh(d);				\
  _db = zbnlow(d);				\
						\
  _qa = (n1)/_da;				\
  _rp = (n1)%_da;				\
  _rp = zbnhighlow(_rp, zbnhigh(n0));		\
  _p = _qa*_db;					\
  if (_rp < _p) {				\
    _qa--;					\
    _rp += (d);					\
    if (_rp >= (d)) {				\
      if (_rp < _p) {				\
	_qa--;					\
	_rp += (d);				\
      }						\
    }						\
  }						\
  _rp -= _p;					\
						\
  _qb = _rp/_da;				\
  _rp = _rp%_da;				\
  _rp = zbnhighlow(_rp, zbnlow(n0));		\
  _p = _qb*_db;					\
  if (_rp < _p) {				\
    _qb--;					\
    _rp += (d);					\
    if (_rp >= (d)) {				\
      if (_rp < _p) {				\
	_qb--;					\
	_rp += (d);				\
      }						\
    }						\
  }						\
  _rp -= _p;					\
						\
  (q) = zbnhighlow(_qa, _qb);			\
  (r) = _rp;					\
} else
#endif

#if 0
/* \TEX{
   \title{Hamming weight of one digit}
   \input{A digit |d|.}
   \side{The Hamming weight of |d| is stored in |w|.}
   \note{This is a macro which is sometimes an assembler
   directive. When none assembler directive is available, the
   algorithm used depends on the flag |LOW_MEMORY|.}
   } */
void zbnwght(w, d)
     BigNumLength w;
     BigNumDigit d;{}
#endif
#ifndef zbnwght
# ifdef LOW_MEMORY
#define zbnwght(w, d)				\
if (1) {					\
  BigNumDigit _dp, _wp;				\
						\
  _dp = (d);					\
  _wp = 0;					\
  while(_dp) {					\
    _dp &= (_dp-1);				\
    _wp++;					\
  }						\
  (w) = _wp;					\
} else
# else
#define zbnwght(w, d)						\
if (1) {							\
  BigNumDigit _dp, _wp;						\
								\
  switch(sizeof(BigNumDigit)) {					\
  case sizeof(short):						\
    _wp = __ZEN_WEIGHT[(d)];					\
    break;							\
  case 2*sizeof(short):						\
    _wp = __ZEN_WEIGHT[(d) & 65535]+__ZEN_WEIGHT[((d)>>16)];	\
    break;							\
  case 4*sizeof(short):						\
    _wp = __ZEN_WEIGHT[(d) & 65535] +				\
         __ZEN_WEIGHT[((d) >> 16) & 65535] +			\
         __ZEN_WEIGHT[((d) >> 32) & 65535] +			\
         __ZEN_WEIGHT[((d) >> 48)];				\
    break;							\
  default:							\
    {								\
      BigNumLength i;						\
								\
      _dp = (d); _wp = 0;					\
      for(i = 0;						\
	  i < sizeof(BigNumDigit)/sizeof(short);		\
	  i++) {						\
	_wp += __ZEN_WEIGHT[_dp & 65535];			\
	_dp >>= 16;						\
      }								\
    break;							\
  }								\
  (w) = _wp;							\
} else
# endif
#endif

/* (h, m ) =  n * d  + (0, c) + (0, m) */
#ifndef zbnmuladds
# define zbnmuladds(h, m, n, d, c)		\
 do {						\
   zbnmul((h), (n), (n), (d));			\
   zbnadd((h), (n), (h), (n), CST_BNDIGIT(0), (c));	\
   zbnadd((h), (m), (h), (m), CST_BNDIGIT(0), (n));	\
 } while(0)
#endif

#ifdef  KARA_DEFAULTS_H
# ifndef ZBN_KARA_OPTS
/* Default values (seems to be one of the best compromise for the
   above architectures */
#  define ZBN_KARA_OPTS (Add | Sub | SK2)
#  define SK2_NOSQ 1
# endif
#endif
