/*
 * zen/include/ze2.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZE2__
#define __ZE2__
#include "zen.h"

#define opXor(a,b) (((a)||(b))&&(!((a)&&(b))))

/* \TEX{
   All the |_Ze2Fct| functions have corresponding macros |Ze2Fct|.
   The macros are always at least as fast as the function call and sometimes even faster.
   Therefore, it is best to use macros.
   }
   */
#define Ze2Elt2Int(a,Rg) ZENElt2Int(a,Rg)

/* On elements
 *************/

/* Allocation */
#define Ze2EltCreate(a, Rg)
#define Ze2EltFree(a, Rg)

/* Assigning */
#define Ze2EltAssign(a,b,Rg) (Ze2Elt2Int(a, Rg)) = (Ze2Elt2Int(b, Rg))
#define Ze2EltSetToZero(a,Rg) (Ze2Elt2Int(a, Rg)) = 0
#define Ze2EltSetToOne(a,Rg) (Ze2Elt2Int(a, Rg)) = 1
#define Ze2EltSetRandom(a,Rg) (Ze2Elt2Int(a, Rg)) = zbnrandom() & 1
#define Ze2EltSetNext(a,Rg) (Ze2Elt2Int(a, Rg)) = !(Ze2Elt2Int(a, Rg))

/* Test */
#define Ze2EltAreEqual(a,b,Rg) ((Ze2Elt2Int(a, Rg)) == (Ze2Elt2Int(b, Rg)))
#define Ze2EltIsZero(a,Rg) ((Ze2Elt2Int(a, Rg)) == 0)
#define Ze2EltIsOne(a,Rg) ((Ze2Elt2Int(a, Rg)) == 1)

/* Arithmetic */
#define Ze2EltAdd(r,a,Rg) (Ze2Elt2Int(r, Rg)) ^= (Ze2Elt2Int(a, Rg))
#define Ze2EltNegate(c,a,Rg) Ze2EltAssign(c,a,Rg) 
#define Ze2EltSubtract(c,a,Rg) Ze2EltAdd(c,a,Rg)
#define Ze2EltSquare(r,a,Rg) (Ze2Elt2Int(r, Rg)) = (Ze2Elt2Int(a, Rg))
#define Ze2EltMultiply(r,a,b,Rg) (Ze2Elt2Int(r, Rg)) = (Ze2Elt2Int(a, Rg)) & (Ze2Elt2Int(b, Rg))
#define Ze2EltInverse(r,a,Rg) \
((Ze2Elt2Int(a, Rg)) ? \
((Ze2Elt2Int(r, Rg)) = 1, ZEN_HAS_INVERSE) : \
((Ze2Elt2Int(r, Rg)) = 0, ZENSetError(ZEN_LIB_ZE2, ZEN_INVERSE, ZEN_DIVBYZERO), ZERR))
#define Ze2EltTrace(r,a,Rg) (Ze2Elt2Int(r, Rg)) = (Ze2Elt2Int(a, Rg))
#define Ze2EltExp(r,k,kl,a,Rg) Ze2EltAssign(r,a,Rg)

/* io */
#define Ze2EltReadFromString(a,s,base,Rg) _Ze2EltReadFromString(&(a),s,base,Rg) 
#define Ze2EltPrintToString(a,base,Rg) _Ze2EltPrintToString(a,base,Rg)
#define Ze2EltReadFromFile(a,file,base,Rg) _Ze2EltReadFromFile(a,file,base,Rg) 
#define Ze2EltPrintToFile(file,a,base,Rg) _Ze2EltPrintToFile(file,a,base,Rg)

#define Ze2EltGetFromString(a,s,Rg) _Ze2EltReadFromString(&(a),s,10,Rg)
#define Ze2EltPutToString(a,Rg) _Ze2EltPrintToString(a,10,Rg)
#define Ze2EltGetFromFile(a,file,Rg) _Ze2EltReadFromFile(&(a),file,10,Rg) 
#define Ze2EltPutToFile(file,a,Rg) _Ze2EltPrintToFile(file,a,10,Rg)

#define Ze2EltToZ(p,p_pl,a,R) (ZBNDigit(p, 0) = Ze2Elt2Int(a, Rg), *(p_pl) = 1)
#define Ze2ZToElt(a,p,pl,R) (Ze2Elt2Int(a, Rg) = ZBNDigit(p, 0))

/* On polynomials
 ****************/
#define Ze2PolyDeg(p,Rg) ZENPolyDeg(p,Rg)
#define Ze2PolyLgt(p,Rg) ZENPolyLgt(p,Rg)

#define Ze2PolyDigit(r,Rg) ZENPolyDigit(r,Rg)
#define Ze2PolyGetDigit(r,i,Rg) ZENPolyGetDigit(r,i,Rg)

#define Ze2PolyChar(r,Rg) ZENPolyChar(r,Rg)
#define Ze2PolyGetChar(r,i,Rg) ZENPolyGetChar(r,i,Rg)

/* Allocation */
#define Ze2PolyCreate(d, Rg) _Ze2PolyCreate(d, Rg) 
#define Ze2PolyCreateLarge(d, Rg) _Ze2PolyCreate(d, Rg) 
#define Ze2PolyFree(r, Rg) _Ze2PolyFree(r, Rg) 

/* Assigning */
#define Ze2PolyAssign(r,PX,Rg) _Ze2PolyAssign(r,PX,Rg)
#define Ze2PolySetToZero(r,Rg) _Ze2PolySetToZero(r,Rg)
#define Ze2PolySetToXi(r,i,Rg) _Ze2PolySetToXi(r,i,Rg)
#define Ze2PolySetRandom(r,i,Rg) _Ze2PolySetRandom(r,i,Rg)
#define Ze2PolySetNext(r,Rg) _Ze2PolySetNext(r,Rg)
#define Ze2PolySetCoeff(r,i,b,Rg) _Ze2PolySetCoeff(r,i,b,Rg)
#define Ze2PolyGetCoeffPtr(r,i,Rg) _Ze2PolyGetCoeffPtr(r,i,Rg)
#define Ze2PolyUpdateDegree(r,Rg) _Ze2PolyUpdateDegree(r,Rg)

/* Test */
#define Ze2PolyAreEqual(r,PX,Rg) _Ze2PolyAreEqual(r,PX,Rg)
#define Ze2PolyIsZero(r,Rg) (!Ze2PolyDeg(r,Rg) && !(Ze2PolyGetDigit(r,0,Rg) & 1))
#define Ze2PolyIsXi(r,i,Rg)  _Ze2PolyIsXi(r,i,Rg)

/* Arithmetic */
#define Ze2PolyAdd(r,PX,Rg) _Ze2PolyAdd(r,PX,Rg)
#define Ze2PolySubtract(r,PX,Rg) _Ze2PolyAdd(r,PX,Rg)
#define Ze2PolyNegate(r,PX,Rg) _Ze2PolyAssign(r,PX,Rg)

#define Ze2PolySquare(c,r,Rg) _Ze2PolySquare(c,r,Rg)
#define Ze2PolySquareLazy(c,r,Rg) _Ze2PolySquare(c,r,Rg)

#define Ze2PolyMultiply(c,r,p,Rg) _Ze2PolyMultiply(c,r,p,Rg)
#define Ze2PolyMultiplyLazy(c,r,p,Rg) _Ze2PolyMultiply(c,r,p,Rg)
#define Ze2PolyMultiplyKaratsuba(c,r,p,Rg) _Ze2PolyMultiplyKaratsuba(c,r,p,Rg)
#define Ze2PolyMultiplyKaratsubaLazy(c,r,p,Rg) _Ze2PolyMultiplyKaratsubaLazy(c,r,p,Rg)

#define Ze2PolyMultiplyScalar(c,r,e,Rg) _Ze2PolyMultiply(c,r,e,Rg)
#define Ze2PolyDot(r,p,q,Rg) _Ze2PolyDot(r,p,q,Rg)
#define Ze2PolyEval(f,p,e,Rg) _Ze2PolyEval(&(f),p,e,Rg)
#define Ze2PolyDivide(r,MX,PX,QX,Rg) _Ze2PolyDivide(r,MX,PX,QX,Rg)
#define Ze2PolyDivideLazy(r,MX,PX,QX,Rg) _Ze2PolyDivideLazy(r,MX,PX,QX,Rg)
#define Ze2PolyModuloLazy(MX,PX,QX,k,Rg) _Ze2PolyModuloLazy(MX,PX,QX,k,Rg)

#define Ze2PolyMultiplyPrecompute() _Ze2PolyMultiplyPrecompute()
#define Ze2PolyDividePrecompute() _Ze2PolyDividePrecompute()

/* io */
#define Ze2PolyReadFromString(r,s,base,Rg) ZENPolyReadFromString(r,s,base,Rg) 
#define Ze2PolyPrintToString(r,base,Rg) ZENPolyPrintToString(r,base,Rg)
#define Ze2PolyReadFromFile(r,file,base,Rg) ZENPolyReadFromFile(r,file,base,Rg) 
#define Ze2PolyPrintToFile(file,r,base,Rg) ZENPolyPrintToFile(file,r,base,Rg)

#define Ze2PolyGetFromString(r,s,Rg) ZENPolyReadFromString(r,s,10,Rg)
#define Ze2PolyPutToString(r,Rg) ZENPolyPrintToString(r,10,Rg)
#define Ze2PolyGetFromFile(r,file,Rg) ZENPolyReadFromFile(r,file,10,Rg) 
#define Ze2PolyPutToFile(file,r,Rg) ZENPolyPrintToFile(file,r,10,Rg)

/* On matrices
 ****************/
#define Ze2MatNbPtr(M,R) ZENMatNbPtr(M,R)
#define Ze2MatSize(M,R) ZENMatSize(M,R)
#define Ze2MatNbBn(M,R) ZENMatNbBn(M,R)
#define Ze2MatNbBnd(M,R) ZENMatNbBnd(M,R)
#define Ze2MatNbDigit(M,R) ZENMatNbDigit(M,R)
#define Ze2MatNbBloc(M,R) ZENMatNbBloc(M,R)
#define Ze2MatType(M,R) ZENMatType(M,R)
#define Ze2MatFirst(M,R) ZENMatFirst(M,R)

#define Ze2MatIsRowType(M,R) ZENMatIsRowType(M,R)
#define Ze2MatIsPermutation(M,R) ZENMatIsPermutation(M,R)
#define Ze2MatAreSameType(A,B,R) ZENMatAreSameType(A,B,R)

#define Ze2MatNbRow(M,R) ZENMatNbRow(M,R)
#define Ze2MatNbCol(M,R) ZENMatNbCol(M,R)


/* Allocation */
#define Ze2MatCreate(l,c,R) ZENMatCreate(l,c,R)
#define Ze2MatRowCreate(l,c,R) ZENMatRowCreate(l,c,R)
#define Ze2MatColCreate(l,c,R) ZENMatColCreate(l,c,R)
#define Ze2PermutationCreate(n,t,R) ZENPermutationCreate(n,t,R)
#define Ze2PermutationRowCreate(n,R) ZENPermutationRowCreate(n,R)
#define Ze2PermutationColCreate(n,R) ZENPermutationColCreate(n,R)
#define Ze2MatFree(M,R) _ZedsMatFree(M,R)
#define Ze2MatCopy(M,R) ZENMatCopy(M,R)

/* Assignation */
#define Ze2MatFastPermute(M,l1,l2,R) ZENMatFastPermute(M,l1,l2,R)
#define Ze2MatGetCoeff(Z,M,l,c,R) ZENMatGetCoeff(Z,M,l,c,R)  
#define Ze2MatSetCoeff(M,l,c,Z,R) ZENMatSetCoeff(M,l,c,Z,R)

#define Ze2MatPermuteRow(M,l1,l2,R) ZENMatPermuteRow(M,l1,l2,R)
#define Ze2MatPermuteCol(M,l1,l2,R) ZENMatPermuteCol(M,l1,l2,R)


/* Arithmetic */
#define Ze2MatMultiply(X,A,B,R) ZENMatMultiply(X,A,B,R)
#define Ze2MatSquare(A,B,R) ZENMatSquare(A,B,R)

/* io */
#define Ze2MatReadFromString(r,s,base,Rg) ZENMatReadFromString(r,s,base,Rg)
#define Ze2MatPrintToString(r,base,Rg) ZENMatPrintToString(r,base,Rg)
#define Ze2MatReadFromFile(r,file,base,Rg) ZENMatReadFromFile(r,file,base,Rg)
#define Ze2MatPrintToFile(file,r,base,Rg) ZENMatPrintToFile(file,r,base,Rg)

/* Following the local Ze2 functions */

#define Ze2MatLnCoeffFast(M,l,dc,mc,R) ((Ze2MatBloc(M,l,dc,R)>>(mc))&((BigNumDigit)1))
#define Ze2MatLnCoeffTest(M,l,dc,Emc,R) (Ze2MatBloc(M,l,dc,R)&(Emc))
#define Ze2MatLnCoeff(M,l,c,R) Ze2MatLnCoeffFast(M,l,divSizeBloc(c),modSizeBloc(c),R)
#define Ze2MatLnChangeCoeffFast(M,l,dc,Emc,R) Ze2MatBloc(M,l,dc,R)^=Emc
#define Ze2MatLnChangeCoeff(M,l,c,R) Ze2MatLnChangeCoeffFast(M,l,divSizeBloc(c),BlocExp2(modSizeBloc(c)),R)

#define Ze2MatColCoeff(M,l,c,R) Ze2MatLnCoeff(M,c,l,R)
#define Ze2MatColChangeCoeff(M,l,c,R) Ze2MatLnChangeCoeff(M,c,l,R)
#define Ze2MatColCoeffFast(M,dl,ml,c,R) Ze2MatLnCoeffFast(M,c,dl,ml,R)
#define Ze2MatColCoeffTest(M,dl,Eml,c,R) Ze2MatLnCoeffTest(M,c,dl,Eml,R)
#define Ze2MatColChangeCoeffFast(M,dl,Eml,c,R) Ze2MatLnChangeCoeffFast(M,c,dl,Eml,R)


/* Allocation */
#define Ze2MatCreateParams(ts,l,c,p_pms,R) _Ze2MatCreateParams(ts,l,c,p_pms,R)
     
/* Assigning */
#define Ze2MatAssign(A,B,R) _Ze2MatAssign(A,B,R)
#define Ze2MatSameTypeAssign(A,B,R) ZBNAssign(Ze2MatFirst(A,R),Ze2MatFirst(B,R),Ze2MatNbBnd(A,R))
#define Ze2MatOppositeTypeAssign(A,B,R) _Ze2MatOppositeTypeAssign(A,B,R)
#define Ze2MatSetToZero(A,R) ZBNSetToZero(Ze2MatFirst(A,R),Ze2MatNbBnd(A,R))
#define Ze2MatSetToOne(M,R) _Ze2MatSetToOne(M,R)
#define Ze2MatSetRandom(M,R) _Ze2MatSetRandom(M,R)

#define Ze2MatSetBloc(M,l,c,Z,R) _Ze2MatSetBloc(M,l,c,Z,R)
#define Ze2MatGetBloc(Z,M,l,c,R) _Ze2MatGetBloc(Z,M,l,c,R)
#define Ze2MatSetDigit(M,l,c,Z,R) Ze2MatSetBloc(M,c,l,Z,R)
#define Ze2MatGetDigit(Z,M,l,c,R) Ze2MatGetBloc(Z,M,c,l,R)
 
#define Ze2MatPermuteBn(M,l1,l2,R) _Ze2MatPermuteBn(M,l1,l2,R)
#define Ze2MatPermuteDigit(M,l1,l2,R) _Ze2MatPermuteDigit(M,l1,l2,R)

/* Test */
#define Ze2MatSameTypeAreEqual(A,B,R) (!(ZBNCompare(Ze2MatFirst(A,R),Ze2MatNbBnd(A,R),Ze2MatFirst(B,R),Ze2MatNbBnd(B,R))))
#define Ze2MatIsZero(A,R) ZBNIsZero(Ze2MatFirst(A,R),Ze2MatNbBnd(A,R))
#define Ze2MatIsOne(M,R) _Ze2MatIsOne(M,R)

/* Arithmetic */
#define Ze2MatAdd(A,B,R) _Ze2MatAdd(A,B,R)
#define Ze2MatOppositeTypeAdd(A,B,R) _Ze2MatOppositeTypeAdd(A,B,R)
#define Ze2MatSameTypeAdd(A,B,R) _Ze2MatSameTypeAdd(A,B,R)
#define Ze2MatAddScalar(A,b,R) if(Ze2EltIsOne(b,R))_Ze2MatAddScalarOne(A,R); else
#define Ze2MatNegate(A,B,R) Ze2MatAssign(A,B,R)
#define Ze2MatSubtract(A,B,R) Ze2MatAdd(A,B,R)
#define Ze2MatMultiplyPlain(X,A,B,R) _Ze2MatMultiplyPlain(X,A,B,R)
#define Ze2MatMultiplyScalar(A,b,R) if(Ze2EltIsZero(b,R)) Ze2MatSetToZero(A,R); else
#define Ze2MatMultiplyRRR(X,A,B,R) _Ze2MatMultiplyRRR(X,A,B,R)
#define Ze2MatMultiplyRCR(X,A,B,R) _Ze2MatMultiplyRCR(X,A,B,R)
#define Ze2MatMultiplyRRC(X,A,B,R) _Ze2MatMultiplyRRC(X,A,B,R)
#define Ze2MatMultiplyRCC(X,A,B,R) _Ze2MatMultiplyRCC(X,A,B,R)
#define Ze2MatMultiplyCCC(X,A,B,R) Ze2MatMultiplyRRR(X,B,A,R)	
#define Ze2MatMultiplyCCR(X,A,B,R) Ze2MatMultiplyRCR(X,B,A,R)
#define Ze2MatMultiplyCRC(X,A,B,R) Ze2MatMultiplyRRC(X,B,A,R)
#define Ze2MatMultiplyCRR(X,A,B,R) Ze2MatMultiplyRCC(X,B,A,R)

#define Ze2MatMultiplyLeftPermutationRRR(X,A,B,R) _Ze2MatMultiplyLeftPermutationRRR(X,A,B,R)
#define Ze2MatMultiplyLeftPermutationRCR(X,A,B,R) _Ze2MatMultiplyLeftPermutationRCR(X,A,B,R)
#define Ze2MatMultiplyLeftPermutationCRC(X,A,B,R) __Ze2MatMultiplyLeftPermutationXRC(Ze2MatNbDigit(X,R),X,A,B,R)
#define Ze2MatMultiplyLeftPermutationCCC(X,A,B,R) __Ze2MatMultiplyLeftPermutationXCC(Ze2MatNbDigit(X,R),X,A,B,R)
#define Ze2MatMultiplyLeftPermutationRRC(X,A,B,R) __Ze2MatMultiplyLeftPermutationXRC(Ze2MatNbBn(X,R),X,A,B,R)
#define Ze2MatMultiplyLeftPermutationCRR(X,A,B,R) Ze2MatMultiplyLeftPermutationRRC(X,A,B,R) 
#define Ze2MatMultiplyLeftPermutationRCC(X,A,B,R) __Ze2MatMultiplyLeftPermutationXCC(Ze2MatNbBn(X,R),X,A,B,R)
#define Ze2MatMultiplyLeftPermutationCCR(X,A,B,R) Ze2MatMultiplyLeftPermutationRCC(X,A,B,R) 
#define Ze2MatMultiplyRightPermutationCCC(X,A,B,R) Ze2MatMultiplyLeftPermutationRRR(X,B,A,R)
#define Ze2MatMultiplyRightPermutationRRC(X,A,B,R) Ze2MatMultiplyLeftPermutationCRC(X,B,A,R)
#define Ze2MatMultiplyRightPermutationRCC(X,A,B,R) Ze2MatMultiplyLeftPermutationCRR(X,B,A,R)
#define Ze2MatMultiplyRightPermutationCRC(X,A,B,R) Ze2MatMultiplyLeftPermutationRRC(X,B,A,R)
#define Ze2MatMultiplyRightPermutationCCR(X,A,B,R) Ze2MatMultiplyLeftPermutationRCR(X,B,A,R)
#define Ze2MatMultiplyRightPermutationRRR(X,A,B,R) Ze2MatMultiplyLeftPermutationCCC(X,B,A,R)
#define Ze2MatMultiplyRightPermutationRCR(X,A,B,R) Ze2MatMultiplyLeftPermutationCCR(X,B,A,R)
#define Ze2MatMultiplyRightPermutationCRR(X,A,B,R) Ze2MatMultiplyLeftPermutationRCC(X,B,A,R)
#define Ze2MatGaussPlain(D,S,P,p_rk,R) ZENMatGaussPlain(D,S,P,p_rk,R)
#define Ze2MatGaussRank(D,p_rk,R) ZENMatGaussRank(D,p_rk,R)
#define Ze2MatGaussLazy(D,P,p_rk,R) ZENMatGaussLazy(D,P,p_rk,R)
#define Ze2MatGaussInverse(D,S,R) ZENMatGaussInverse(D,S,R)
#define Ze2MatInverse(I,M,R) _Ze2MatInverse(I,M,R)
#define Ze2MatKernel(K,M,R) _Ze2MatKernel(K,M,R)
#define Ze2MatGetSubMat(S,M,l,c,R) _Ze2MatGetSubMat(S,M,l,c,R) 
#define Ze2MatSetSubMat(M,l,c,R) _Ze2MatSetSubMat(M,l,c,S,R)

/* io */
#define Ze2MatGetFromString(r,s,Rg) _Ze2MatGetFromString(r,s,Rg)
#define Ze2MatPutToString(r,Rg) _Ze2MatPutToString(r,Rg)
#define Ze2MatGetFromFile(r,file,Rg) _Ze2MatGetFromFile(r,file,Rg)
#define Ze2MatPutToFile(file,r,Rg) _Ze2MatPutToFile(file,r,Rg)

/* Procedure on rings
 ***********************/
extern int _Ze2Init ___((ZENRing));
extern int _Ze2Close ___((ZENRing));

/* Procedure on elements
 ***********************/
/* Allocation */
extern int _Ze2EltCreate ___((ZENElt *, ZENRing)); /* Create elements */
extern void _Ze2EltFree ___((ZENElt, ZENRing)); /* Free elements */

/* Assigning */
extern void _Ze2EltAssign ___((ZENElt *, ZENElt, ZENRing)); /* Assign an element with another */
extern void _Ze2EltSetToZero ___((ZENElt *, ZENRing)); /* Set an element to zero */
extern void _Ze2EltSetToOne ___((ZENElt *, ZENRing)); /* Set an element to one */
extern void _Ze2EltSetRandom ___((ZENElt *, ZENRing)); /* Set randomly an element */
extern void _Ze2EltSetNext ___((ZENElt *, ZENRing)); /* Set an element to the next */

/* Test */
extern int _Ze2EltAreEqual ___((ZENElt, ZENElt, ZENRing)); /* Is an element equal to another */
extern int _Ze2EltIsZero ___((ZENElt, ZENRing)); /* Is an element equal to zero */
extern int _Ze2EltIsOne ___((ZENElt, ZENRing)); /* Is an element equal to one */

/* Arithmetic */
extern void _Ze2EltAdd ___((ZENElt *, ZENElt, ZENRing));	/* Addition of elements */
extern void _Ze2mEltSquare ___((ZENElt *, ZENElt, ZENRing));
extern void _Ze2EltSquare ___((ZENElt *, ZENElt, ZENRing)); /* Squaring of elements */
extern void _Ze2mEltMultiply ___((ZENElt *, ZENElt, ZENElt, ZENRing));
extern void _Ze2EltMultiply ___((ZENElt *, ZENElt, ZENElt, ZENRing)); /* Multiplication of elements */
extern int _Ze2EltInverse ___((ZENElt *, ZENElt, ZENRing)); /* Inverse of elements */
extern void _Ze2EltTrace ___((ZENElt *, ZENElt, ZENRing));	/* Trace of elements */

/* io */
extern int _Ze2EltReadFromString ___((ZENElt *, char *, int, ZENRing)); /* Read a polynomial from a string */
extern char* _Ze2EltPrintToString ___((ZENElt, int, ZENRing)); /* Print a polynomial to a string */
extern int _Ze2EltReadFromFile ___((ZENElt *, FILE *, int, ZENRing));	/* Read a polynomial from a file */
extern int _Ze2EltPrintToFile ___((FILE *, ZENElt, int, ZENRing)); /* Print a polynomial to a file */
extern int _Ze2EltGetFromString ___((ZENElt *, char *, ZENRing)); /* Get a polynomial from a string */
extern char* _Ze2EltPutToString ___((ZENElt, ZENRing)); /* Put a polynomial to a string */
extern int _Ze2EltGetFromFile ___((ZENElt *, FILE *, ZENRing)); /* Get a polynomial from a file */
extern int _Ze2EltPutToFile ___((FILE *, ZENElt, ZENRing)); /* Put a polynomial to a file */

extern void _Ze2EltToZ ___((BigNum,BigNumLength*,ZENElt,ZENRing)); /* Evaluate in $p$ an element */
extern void _Ze2ZToElt ___((ZENElt *,BigNum,BigNumLength,ZENRing)); /* Get the element from an evaluation in $p$ */
extern int _Ze2EltConvert ___((ZENElt *,ZENRing,ZENElt,ZENRing)); /* Convert elements between ZENRing */

/* Procedures on polynomials
 ***************************/
/* Allocation */
extern ZENPoly _Ze2PolyCreate ___((int, ZENRing)); /* Create polynomials */
extern void _Ze2PolyFree ___((ZENPoly, ZENRing)); /* Free polynomials */

/* Assigning */
extern void _Ze2PolyAssign ___((ZENPoly, ZENPoly, ZENRing)); /* Assign a polynomial with another */
extern void _Ze2PolySetToZero ___((ZENPoly, ZENRing)); /* Set a polynomial to zero */
extern void _Ze2PolySetToXi ___((ZENPoly, int, ZENRing)); /* Set a polynomial to $X^i$ */
extern void _Ze2PolySetRandom ___((ZENPoly, int, ZENRing)); /* Set randomly a polynomial */
extern void _Ze2PolySetCoeff ___((ZENPoly, int, ZENElt, ZENRing)); /* Set a coefficient of a polynomial */
extern void _Ze2PolyGetCoeff ___((ZENElt *, ZENPoly, int, ZENRing)); /* Get a coefficient of a polynomial */
extern ZENElt _Ze2PolyGetCoeffPtr ___((ZENPoly, int, ZENRing)); /* Get a pointer on a coefficient */
extern void _Ze2PolyUpdateDegree ___((ZENPoly, ZENRing)); /* Update the degree of a polynomial */

/* Test */
extern int _Ze2PolyAreEqual ___((ZENPoly, ZENPoly, ZENRing)); /* Is a polynomial equal to another */
extern int _Ze2PolyIsZero ___((ZENPoly, ZENRing)); /* Is a polynomial equal to zero */
extern int _Ze2PolyIsXi ___((ZENPoly, int, ZENRing)); /* Is a polynomial equal to $X^i$ */

/* Arithmetic */
extern void _Ze2PolyAdd ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _Ze2PolySubtract ___((ZENPoly, ZENPoly, ZENRing)); /* Substraction of polynomials */
extern int _Ze2PolySquare ___((ZENPoly, ZENPoly, ZENRing)); /* Squaring of polynomials */
extern int _Ze2PolyMultiply ___((ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Multiplication of polynomials */
extern int _Ze2PolyMultiplyKaratsuba ___((ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Multiplication of polynomials */
extern int _Ze2PolyMultiplyKaratsubaLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Multiplication of polynomials */
extern void _Ze2PolyMultiplyScalar ___((ZENPoly, ZENPoly, ZENElt, ZENRing)); /* Multiplication of polynomials by a scalar*/
extern void _Ze2PolyDot  ___((ZENElt *, ZENPoly, ZENPoly, ZENRing)); /* Scalar product of 2 polynomials */
extern void _Ze2PolyEval  ___((ZENElt *, ZENPoly, ZENElt, ZENRing)); /* Evaluation of a polynomial */
extern int _Ze2PolyDivide ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Division of polynomials */
extern int _Ze2PolyDivideLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Lazy division of polynomials */
extern int _Ze2PolyModuloLazy ___((ZENPoly, ZENPoly, ZENPoly, int, ZENRing)); /* Lazy modulo of polynomials */
extern int _Ze2PolyDerive ___((ZENPoly, ZENPoly, ZENRing)); 

/* Procedures on series
 ***************************/
/* Allocation */
extern ZENSr _Ze2SrCreate ___((int, ZENRing)); /* Create series */
extern void _Ze2SrFree ___((ZENSr, ZENRing)); /* Free series */

/* Assigning */
extern void _Ze2SrAssign ___((ZENSr, ZENSr, ZENRing)); /* Assign a valuation with another */
extern void _Ze2SrSetToZero ___((ZENSr,ZENRing)); /* Set a valuation to zero */
extern void _Ze2SrSetRandom ___((ZENSr,ZENRing)); /* Set randomly a valuation */
extern void _Ze2SrSetCoeff ___((ZENSr, int, ZENElt, ZENRing)); /* Set a coefficient of a valuation */
extern void _Ze2SrGetCoeff ___((ZENElt, ZENSr, int, ZENRing)); /* Get a coefficient of a valuation */
extern ZENElt _Ze2SrGetCoeffPtr ___((ZENSr, int, ZENRing)); /* Get a pointer on a coefficient of a valuation */
extern void _Ze2SrUpdateValuation ___((ZENSr, ZENRing)); /* Update the valuation of a valuation */

/* Test */
extern int _Ze2SrAreEqual ___((ZENSr, ZENSr, ZENRing)); /* Is a valuation equal to another */
extern int _Ze2SrIsZero ___((ZENSr, ZENRing)); /* Is a valuation equal to zero */

/* Arithmetic */
extern void _Ze2SrAdd ___((ZENSr, ZENSr, ZENRing)); /* Addition of series */
extern void _Ze2SrNegate ___((ZENSr, ZENSr, ZENRing)); /* Negation of series */
extern void _Ze2SrSubtract ___((ZENSr, ZENSr, ZENRing)); /* Substraction of series */
extern int _Ze2SrSquareChar2 ___((ZENSr, ZENSr, ZENRing)); /* Squaring of series in characteristic 2*/
extern int _Ze2SrSquareCharp ___((ZENSr, ZENSr, ZENRing)); /* Squaring of series in characteristic p*/
extern int _Ze2SrMultiply ___((ZENSr, ZENSr, ZENSr, ZENRing)); /* Multiplication of series */
extern void _Ze2SrMultiplyScalar ___((ZENSr, ZENElt, ZENSr, ZENRing)); /* Multiplication of series by a scalar */
extern int _Ze2SrDivide ___((ZENSr, ZENSr, ZENRing)); /* Division of series */

/* Procedures on matrices
 ***************************/
/* Allocation */
extern int _Ze2MatCreateParams ___((ZENMatTypeStruct,
			   Dim,Dim,
			   ZENMat *,
			   ZENRing));		/* Creates a matrix parameters
						   structure */
/* Assignment */
extern int _Ze2MatAssign ___((ZENMat,ZENMat,ZENRing));
extern int _Ze2Permutation2Matrix ___((ZENMat,ZENMat,ZENRing));
extern int _Ze2MatSetToZero ___((ZENMat,ZENRing));
extern int _Ze2MatSetToOne ___((ZENMat,ZENRing));
extern int _Ze2MatSetRandom ___((ZENMat,ZENRing));
extern int _Ze2MatGetBloc ___((ZENElt *,ZENMat,Dim,Dim,ZENRing));
extern int _Ze2MatGetSubMat ___((ZENMat,ZENMat,Dim,Dim,ZENRing));
extern int _Ze2MatSetSubMat ___((ZENMat,Dim,Dim,ZENMat,ZENRing));
extern ZENElt _Ze2MatGetBlocPtr ___((ZENMat,Dim,Dim,ZENRing));
extern int _Ze2MatSetBloc ___((ZENMat,Dim,Dim,ZENElt,ZENRing));
extern int _Ze2MatPermuteBn ___((ZENMat,Dim,Dim,ZENRing));
extern int _Ze2MatPermuteDigit ___((ZENMat,Dim,Dim,ZENRing));

/*Tests */
extern int _Ze2MatIsZero ___((ZENMat,ZENRing));
extern int _Ze2MatIsOne ___((ZENMat,ZENRing));
extern int _Ze2MatAreEqual ___((ZENMat,ZENMat,ZENRing));

/*Arithmetic */
extern void _Ze2MatOppositeTypeAdd ___((ZENMat,ZENMat,ZENRing));
extern void _Ze2MatSameTypeAdd ___((ZENMat,ZENMat,ZENRing));
extern int _Ze2MatAdd ___((ZENMat,ZENMat,ZENRing));
extern int _Ze2MatAddScalar ___((ZENMat,ZENElt,ZENRing));
extern void _Ze2MatAddScalarOne ___((ZENMat,ZENRing));
extern int _Ze2MatNegate ___((ZENMat,ZENMat,ZENRing));
extern int _Ze2MatSubtract ___((ZENMat,ZENMat,ZENRing));
extern int _Ze2MatMultiplyPlain ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern int _Ze2MatMultiplyScalar ___((ZENMat,ZENElt,ZENRing));
extern void _Ze2MatMultiplyRRR ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern void _Ze2MatMultiplyRCR ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern void _Ze2MatMultiplyRRC ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern void _Ze2MatMultiplyRCC ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern void _Ze2MatMultiplyLeftPermutationRRR ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern void _Ze2MatMultiplyLeftPermutationRCR ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern void __Ze2MatMultiplyLeftPermutationXRC ___((int,ZENMat,ZENMat,ZENMat,ZENRing));
extern void __Ze2MatMultiplyLeftPermutationXCC ___((int,ZENMat,ZENMat,ZENMat,ZENRing));
extern int _Ze2MatMultiplyRightPermutation ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern int _Ze2MatMultiplyLeftPermutation ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern int _Ze2MatGauss ___((ZENMat,
			     ZENMat,
			     ZENMat,
			     Dim *,
			     ZENElt *,
			     int (*) ___((ZENMat,int,int,ZENRing)),
			     ZENRing));
extern int _Ze2MatInverse ___((ZENMat,ZENMat,ZENRing));
extern int _Ze2MatKernel ___((ZENMat *,ZENMat,ZENRing));

/* io */
extern int _Ze2MatReadFromString ___((ZENMat *,char *,int,ZENRing)); /* Read a matrix from a string */
extern char* _Ze2MatPrintToString ___((ZENMat,int,ZENRing)); /* Print a matrix to a string */
extern int _Ze2MatReadFromFile ___((ZENMat *,FILE *,int,ZENRing));	/* Read a matrix from a file */
extern int _Ze2MatPrintToFile ___((FILE *,ZENMat,int,ZENRing)); /* Print a matrix to a file */
extern int _Ze2MatGetFromString ___((ZENMat *,char *,ZENRing)); /* Get a matrix from a string */
extern char* _Ze2MatPutToString ___((ZENMat,ZENRing)); /* Put a matrix to a string */
extern int _Ze2MatGetFromFile ___((ZENMat *,FILE *,ZENRing)); /* Get a matrix from a file */
extern int _Ze2MatPutToFile ___((FILE *,ZENMat,ZENRing)); /* Put a matrix to a file */

extern int _Ze2MatConvert ___((ZENMat, ZENRing, ZENMat, ZENRing));

/*   
   \TEX{
   \Section{Ring initialization}
   \input{ze2init}

   \Section{Element operations}
   \input{ze2elt_alloc}
   \input{ze2elt_assign}
   \input{ze2elt_test}
   \input{ze2elt_add}
   \input{ze2elt_multiply}
   \input{ze2elt_inverse}
   \input{ze2elt_trace}
   \input{ze2elt_io}

   \Section{Polynomials operations}
   \input{ze2poly_alloc}
   \input{ze2poly_assign}
   \input{ze2poly_test}
   \input{ze2poly_add}
   \input{ze2poly_multiply}
   \input{ze2poly_derivative}

   \Section{Matrices operations}
   \input{ze2mat_create}
   \input{ze2mat_assign}
   \input{ze2permutation_assign}
   \input{ze2mat_permute}
   \input{ze2mat_test}
   \input{ze2mat_add}
   \input{ze2mat_multiply}
   \input{ze2mat_gauss}
   \input{ze2mat_inverse}
   \input{ze2mat_io}
   }
*/

#endif
