/*
 * zen/include/zec.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZEC__
#define __ZEC__

#include "zen.h"
#define CHINESE_VERBOSE 0

/* \TEX{
   In zec.h
   }
   */

#define ZecElt2Elts(a, Rg) ZENElt2Elts(a, Rg)
#define ZecElt2Elt(a, n, Rg) ZecElt2Elts(a, Rg)[n]

#define ZecRingDef(Rg) ZENRingDef(Rg)
#define ZecRingOrigin(R) ZENRingOrigin(R) 
#define ZecRingCln(R) ZENRingCln(R)

#define ZecRingPms(R) ZENRingPms(R).zec
#define ZecRingNbS(R) ZecRingPms(R).nbsubrings
#define ZecRingModinvs(R) ZecRingPms(R).modinvs
#define ZecRingSubrings(R) ZecRingPms(R).subrings
#define ZecRingSubring(R,n) ZecRingSubrings(R)[n]
#define ZecRingModinv(R,n) ZecRingModinvs(R)[n]

/* On elements
 *************/
/* Allocation */
#define ZecEltCreate(a, Rg) _ZecEltCreate(a, Rg) 
#define ZecEltFree(a, Rg) _ZecEltFree(a, Rg)

/* Assigning */
#define ZecEltAssign(a,b,Rg) _ZecEltAssign(a,b,Rg) 
#define ZecEltSetToZero(a,Rg) _ZecEltSetToZero(a,Rg) 
#define ZecEltSetToOne(a,Rg) _ZecEltSetToOne(a,Rg) 
#define ZecEltSetRandom(a,Rg) _ZecEltSetRandom(a,Rg) 
#define ZecEltSetNext(a,Rg) _ZecEltSetNext(a,Rg) 
#define ZecCloneToElt(a,b,Rg) _ZecCloneToElt(&(a),b,Rg)
#define ZecEltToClone(a,b,Rg) _ZecEltToClone(&(a),b,Rg)

/* Test */
#define ZecEltAreEqual(a,b,Rg) _ZecEltAreEqual(a,b,Rg) 
#define ZecEltIsZero(a,Rg) _ZecEltIsZero(a,Rg) 
#define ZecEltIsOne(a,Rg) _ZecEltIsOne(a,Rg) 

/* Arithmetic */
#define ZecEltAdd(r,a,Rg) _ZecEltAdd(r,a,Rg) 
#define ZecEltNegate(r,a,Rg) _ZecEltNegate(r,a,Rg) 
#define ZecEltSubtract(r,a,Rg) _ZecEltSubtract(r,a,Rg) 
#define ZecEltSquare(r,a,Rg) _ZecEltSquare(r,a,Rg) 
#define ZecEltMultiply(r,a,b,Rg) _ZecEltMultiply(r,a,b,Rg) 

#define ZecEltInverse(r,a,Rg) _ZecEltInverse(r,a,Rg) 

#define ZecEltTrace(r,a,Rg) _ZecEltTrace(r,a,Rg) 

/* io */
#define ZecEltReadFromString(a,s,base,Rg) _ZecEltReadFromString(a,s,base,Rg) 
#define ZecEltPrintToString(a,base,Rg) _ZecEltPrintToString(a,base,Rg) 
#define ZecEltReadFromFile(a,file,base,Rg) _ZecEltReadFromFile(a,file,base,Rg) 
#define ZecEltPrintToFile(file,a,base,Rg) _ZecEltPrintToFile(file,a,base,Rg) 

#define ZecEltGetFromString(a,s,Rg) _ZecEltGetFromString(a,s,Rg) 
#define ZecEltPutToString(a,Rg) _ZecEltPutToString(a,Rg) 
#define ZecEltGetFromFile(a,file,Rg) _ZecEltGetFromFile(a,file,Rg) 
#define ZecEltPutToFile(file,a,Rg) _ZecEltPutToFile(file,a,Rg) 

#define ZecEltToZ(p,p_pl,a,Rg) _ZecEltToZ(p,p_pl,a,Rg) 
#define ZecZToElt(a,p,p_pl,Rg) _ZecZToElt(a,p,p_pl,Rg) 


/* Procedure on rings
 ***********************/
extern int _ZecInit ___((ZENRing));
extern int _ZecClose ___((ZENRing));

extern int _ZecPrcInitEltExp ___((ZENRing));
extern int _ZecPrcCloseEltExp ___((ZENRing));

extern int _ZecPrcInitPolyRootsCanonical ___((ZENRing));
extern int _ZecPrcClosePolyRootsCanonical ___((ZENRing));

/* Procedure on elements
 ***********************/

/* Allocation */
extern int _ZecEltCreate ___((ZENElt *, ZENRing)); 
extern void _ZecEltFree ___((ZENElt, ZENRing));

/* Assigning */
extern void _ZecEltAssign ___((ZENElt *, ZENElt, ZENRing));
extern void _ZecEltSetToZero ___((ZENElt *, ZENRing));
extern void _ZecEltSetToOne ___((ZENElt *, ZENRing));
extern void _ZecEltSetRandom ___((ZENElt *, ZENRing));
extern void _ZecEltSetNext ___((ZENElt *, ZENRing));

extern void _ZecEltToClone ___((ZENElt *, ZENElt, ZENRing));
extern void _ZecCloneToElt ___((ZENElt *, ZENElt, ZENRing));

/* Test */
extern int _ZecEltAreEqual ___((ZENElt, ZENElt, ZENRing));
extern int _ZecEltIsZero ___((ZENElt, ZENRing));
extern int _ZecEltIsOne ___((ZENElt, ZENRing));

/* Arithmetic */
extern void _ZecEltAdd ___((ZENElt *, ZENElt, ZENRing));
extern void _ZecEltNegate ___((ZENElt *, ZENElt, ZENRing));
extern void _ZecEltSubtract ___((ZENElt *, ZENElt, ZENRing));
extern void _ZecEltSquare ___((ZENElt *, ZENElt, ZENRing));
extern void _ZecEltMultiply ___((ZENElt *, ZENElt, ZENElt, ZENRing));
extern int _ZecEltInverse ___((ZENElt *, ZENElt, ZENRing));
extern int _ZecEltExp ___((ZENElt *, BigNum, BigNumLength, ZENElt, ZENRing));
extern void _ZecEltTrace ___((ZENElt *, ZENElt, ZENRing));

/* io */
extern void _ZecEltToZ ___((BigNum, BigNumLength *, ZENElt, ZENRing));
extern void _ZecZToElt ___((ZENElt *, BigNum, BigNumLength, ZENRing));


/* 
   \TEX{
   \Section{Ring initialization}
   \input{zecinit}

   \Section{Element operations}
   \input{zecelt_alloc}
   \input{zecelt_assign}
   \input{zecelt_test}
   \input{zecelt_add}
   \input{zecelt_multiply}
   \input{zecelt_exp}
   \input{zecelt_inverse}
   \input{zecelt_trace}
   \input{zecelt_io}

   }
   */
#endif

