/*
 * zen/include/zed.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZED__
#define __ZED__

/* \TEX{
   \Section{The default functions of the library}
   }
   */

#include "zen.h"

#define ZedPolyAllocStruct(Rg) ZENPolyAllocStruct(Rg)
#define ZedPolyCof(p,Rg) ZENPolyCof(p,Rg)
#define ZedPolyGetCof(p,i,Rg) ZENPolyGetCof(p,i,Rg)
#define ZedPolyLgt(PX, Rg) ZENPolyLgt(PX, Rg)

#define ZedMatBloc(M,l,c,R) ZENMatBloc(M,l,c,R)

#define ZedSrAllocStruct(Rg) ZENSrAllocStruct(Rg)
#define ZedSrCof(PX, Rg) ZENSrCof(PX, Rg)
#define ZedSrGetCof(PX, i, Rg) ZENSrGetCof(PX, i, Rg)

/* Procedures on polynomials
 ***************************/
/* Allocation */
extern ZENPoly _ZedsPolyCreate ___((int, ZENRing)); /* Create polynomials */
extern void _ZedsPolyFree ___((ZENPoly, ZENRing)); /* Free polynomials */
extern ZENPoly _ZedPolyCreate ___((int, ZENRing)); /* Create polynomials */
extern void _ZedPolyFree ___((ZENPoly, ZENRing)); /* Free polynomials */

/* Assigning */
extern void _ZedPolySetCoeff ___((ZENPoly,int,ZENElt,ZENRing)); /* Set a coefficient of a polynomial */
extern ZENElt _ZedPolyGetCoeffPtr ___((ZENPoly,int,ZENRing)); /* Get a pointer on a coefficient of a polynomial */

/* Procedures on matrices
 ***************************/
/* Allocation */
extern int _ZedMatCreateParams ___((ZENMatTypeStruct,Dim,Dim,ZENMat *,ZENRing));		
extern int __ZedMatCreate ___((ZENMat,ZENRing));
extern void _ZedMatFree ___((ZENMat,ZENRing));
extern int _ZedsMatCreateParams ___((ZENMatTypeStruct,Dim,Dim,ZENMat *,ZENRing));		
extern int __ZedsMatCreate ___((ZENMat,ZENRing));
extern void _ZedsMatFree ___((ZENMat,ZENRing));

extern int _ZedMatGetBloc ___((ZENElt *,ZENMat,Dim,Dim,ZENRing));
extern ZENElt _ZedMatGetBlocPtr ___((ZENMat,Dim,Dim,ZENRing));
extern int _ZedMatSetBloc ___((ZENMat,Dim,Dim,ZENElt,ZENRing));
extern int _ZedMatGetSubMat ___((ZENMat, ZENMat, Dim, Dim, ZENRing));
extern int _ZedMatSetSubMat ___((ZENMat,Dim,Dim,ZENMat,ZENRing));


/* Procedures on series
 ***************************/
/* Allocation */
extern ZENSr _ZedSrCreate ___((int, ZENRing)); /* Create series */
extern void _ZedSrFree ___((ZENSr, ZENRing)); /* Free series */

/* Assigning */
extern void _ZedSrSetCoeff ___((ZENSr,int,ZENElt,ZENRing)); /* Set a coefficient of a valuation */
extern ZENElt _ZedSrGetCoeffPtr ___((ZENSr,int,ZENRing)); /* Get a pointer on a coefficient of a valuation */
#endif
/* 
   \TEX{
   \Section{Procedures on polynomials}
   \input{zedpoly_alloc}
   
   \Section{Procedures on matrices}
   \input{zedmat_create}
   \input{zedmat_assign}

   \Section{Procedures on series}
   \input{zedsr_alloc}

   \Section{Precomputations}
   \input{zedprc}
   }
   */

