/*
 * zen/include/zeg.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZEG__
#define __ZEG__

#include "zen.h"

#define LN2PREZENMULT 8		/* Used in precomputations in GF(2)
				   value of 8 indicates tabulation on
				   8 bits. */


#define ZENMatMultiplyPermRRR(X,A,B,R) _ZENMatMultiplyPermRRR(X,A,B,R)
#define ZENMatMultiplyPermCCC(X,A,B,R) _ZENMatMultiplyPermRRR(X,B,A,R)
#define ZENMatMultiplyPermRRC(X,A,B,R) _ZENMatMultiplyPermRRR(X,A,B,R)
#define ZENMatMultiplyPermCRC(X,A,B,R) _ZENMatMultiplyPermRRR(X,B,A,R)
#define ZENMatMultiplyPermRCC(X,A,B,R) _ZENMatMultiplyPermRCC(X,A,B,R)
#define ZENMatMultiplyPermCRR(X,A,B,R) _ZENMatMultiplyPermRCC(X,B,A,R)
#define ZENMatMultiplyPermCCR(X,A,B,R) _ZENMatMultiplyPermCCR(X,A,B,R)
#define ZENMatMultiplyPermRCR(X,A,B,R) _ZENMatMultiplyPermCCR(X,B,A,R)


#define ZENEcPtMultPlain(R, k, kl, P, E) _ZENEcPtMultPlain(R, k, kl, P, E)
#define ZENEcPtMultSigned(R, k, kl, P, E) _ZENEcPtMultSigned(R, k, kl, P, E)
#define ZENEcPtMultWindowSigned(R, m, d, k, kl, P, E) _ZENEcPtMultWindowSigned(R, m, d, k, kl, P, E)

/* Flags to Add several points at the same time
***********************************************/
#define MEc_No 0
#define MEc_Add 1
#define MEc_Double 2


/* Procedure on rings
 ***********************/
extern ZENRing _ZENBaseRingCreate ___((BigNum,BigNumLength));
extern ZENRing _ZENExtRingCreate ___((ZENPoly,ZENRing));
extern ZENRing _ZENChineseRingCreate ___((int,ZENRing *));
extern int _ZENRingClose ___((ZENRing));
extern ZENRing _ZENRingCopy ___((ZENRing));
extern ZENRing _ZENRingClone ___((ZENRing,ZENCln));
extern int _ZENRingFullClose ___((ZENRing));

extern int _ZENRingPrintToFile ___((FILE *, ZENRing));
extern int _ZENRingReadFromFile ___((ZENRing *,FILE *));
extern int _ZENRingAddPrc ___((ZENRing, ZENPrc));
extern int _ZENRingRmPrc ___((ZENRing, ZENPrc));

extern int _ZENPrcInitEltExp ___((ZENRing));
extern int _ZENPrcCloseEltExp ___((ZENRing));

extern int __ZENPrcInitEltExp ___((
     int (*) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing)),
     int (*) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing)),
     int (*) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing)),
     int (*) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing)),
     int (*) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing)),
     ZENRing));
extern int __ZENPrcCloseEltExp ___((int (*) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing)), ZENRing));

/* Procedures on elements
 ************************/
/* Assigning */
extern int _ZENEltFromBigNum ___((ZENElt*,BigNum,BigNumLength,ZENRing));
extern int _ZENEltCloneConvert ___((ZENElt *,ZENRing,ZENElt,ZENRing));
/* test */
extern int _ZENEltIsASquareChar2 ___((ZENElt, ZENRing)); /* Is a square  in a finite field whith chracteristic 2 */
extern int _ZENEltIsASquareCharp ___((ZENElt, ZENRing)); /* Is a square  in a finite field with odd chrarcteristic */

/* Arithmetic */
extern int _ZENEltSquareRootChar2 ___((ZENElt *, ZENElt, ZENRing)); /* Square root  in a finite field whith chracteristic 2 */
extern int _ZENEltSquareRootCharp ___((ZENElt *, ZENElt, ZENRing)); /* Square root in a finite field with odd chrarcteristic */


extern int _ZENEltExpPlain ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZENEltExp ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZENEltExpMod ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZENEltExpMod4 ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZENEltExpMod5 ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZENEltExpMod6 ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */

extern int _ZENEltAbsoluteTrace ___((ZENElt *, ZENElt, ZENRing));


/* IO */
extern int _ZENCloneEltReadFromString ___((ZENElt *, char *, int, ZENRing));
extern char* _ZENCloneEltPrintToString ___((ZENElt, int, ZENRing));
extern int _ZENCloneEltReadFromFile ___((ZENElt *, FILE *, int, ZENRing));
extern int _ZENCloneEltPrintToFile ___((FILE *, ZENElt, int, ZENRing));
extern int _ZENCloneEltGetFromString ___((ZENElt *, char *, ZENRing));
extern char* _ZENCloneEltPutToString ___((ZENElt, ZENRing));
extern int _ZENCloneEltGetFromFile ___((ZENElt *, FILE *, ZENRing));
extern int _ZENCloneEltPutToFile ___((FILE *, ZENElt, ZENRing));

/* Procedures on polynomials
 ***************************/
/* Allocation */
extern ZENPoly _ZENPolyCopy ___((ZENPoly,ZENRing)); /* Create polynomials */

/* Assigning */
extern void _ZENPolyGetCoeff ___((ZENElt *, ZENPoly, int, ZENRing)); /* Get a coefficient of a polynomial */
extern void _ZENPolyAssign ___((ZENPoly, ZENPoly, ZENRing)); /* Assign a polynomial with another */
extern void _ZENPolySetToZero ___((ZENPoly, ZENRing)); /* Set a polynomial to zero */
extern void _ZENPolySetToXi ___((ZENPoly, int, ZENRing)); /* Set a polynomial to $X^i$ */
extern void _ZENPolySetRandom ___((ZENPoly, int, ZENRing)); /* Set randomly a polynomial */
extern void _ZENPolyUpdateDegree ___((ZENPoly, ZENRing)); /* Update the degree of a polynomial */

/* Test */
extern int _ZENPolyAreEqual ___((ZENPoly, ZENPoly, ZENRing)); /* Is a polynomial equal to another */
extern int _ZENPolyIsZero ___((ZENPoly, ZENRing)); /* Is a polynomial equal to zero */
extern int _ZENPolyIsXi ___((ZENPoly, int, ZENRing)); /* Is a polynomial equal to $X^i$ */

/* Arithmetic */
extern void _ZENPolyAdd ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZENPolyNegate ___((ZENPoly, ZENPoly, ZENRing)); /* Negation of polynomials */
extern void _ZENPolySubtract ___((ZENPoly, ZENPoly, ZENRing)); /* Substraction of polynomials */
extern int _ZENPolySquareChar2 ___((ZENPoly, ZENPoly, ZENRing)); /* Squaring of polynomials in characteristic 2*/
extern int _ZENPolySquareCharp ___((ZENPoly, ZENPoly, ZENRing)); /* Squaring of polynomials in characteristic p*/
extern int _ZENPolyMultiply ___((ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Multiplication of polynomials */
extern void _ZENPolyMultiplyScalar ___((ZENPoly, ZENPoly, ZENElt, ZENRing)); /* Multiplication of polynomials */
extern void _ZENPolyDot ___((ZENElt *, ZENPoly, ZENPoly, ZENRing));  /* Scalar product of 2 polynomials */
extern void _ZENPolyEval  ___((ZENElt *, ZENPoly, ZENElt, ZENRing)); /* Evaluation of a polynomial */
extern int _ZENPolyMakeMonic ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZENPolyDivide ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Division of polynomials */
extern int _ZENPolyDivideLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Lazy division of polynomials */
extern int _ZENPolyModuloLazy ___((ZENPoly, ZENPoly, ZENPoly, int, ZENRing)); /* Lazy modulo of polynomials */
extern int _ZENPolyDerive ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZENPolyRootsDegree1 ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZENPolyRootsCanonical ___((ZENPoly, ZENElt, ZENRing));
extern int _ZENPolyRootsDegree2Char2 ___((ZENPoly, ZENPoly, ZENRing)); /* Roots of polynomials of degree 2 in finite fields of characteristic 2 */
extern int _ZENPolyRootsDegree2Charp ___((ZENPoly, ZENPoly, ZENRing)); /* Roots of polynomials of degree 2 in finite fields of odd characteristic */
extern int _ZENPolyBinaryReciprocal ___((ZENPoly,ZENPoly,ZENRing));
extern int _ZENPolyReciprocal ___((ZENPoly,ZENPoly,ZENRing));

extern int _ZENPolyGcd ___((ZENPoly,ZENPoly,ZENPoly,ZENRing));
extern int _ZENPolyExtGcd ___((ZENPoly,ZENPoly,ZENPoly,ZENRing));
extern int _ZENPolyResultant ___((ZENElt *, ZENPoly, ZENPoly, ZENRing));

/* io */
extern int _ZENPolyReadFromString ___((ZENPoly *, char *, int, int, ZENRing)); /* Read a polynomial from a string */
extern char* _ZENPolyPrintToString ___((ZENPoly, int, int, ZENRing)); /* Print a polynomial to a string */
extern int _ZENPolyReadFromFile ___((ZENPoly *, FILE *, int, int, ZENRing));        /* Read a polynomial from a file */
extern int _ZENPolyPrintToFile ___((FILE *, ZENPoly, int, int, ZENRing)); /* Print a polynomial to a file */
extern int _ZENPolyGetFromString ___((ZENPoly *, char *, ZENRing)); /* Get a polynomial from a string */
extern char* _ZENPolyPutToString ___((ZENPoly, ZENRing)); /* Put a polynomial to a string */
extern int _ZENPolyGetFromFile ___((ZENPoly *, FILE *, ZENRing)); /* Get a polynomial from a file */
extern int _ZENPolyPutToFile ___((FILE *, ZENPoly, ZENRing)); /* Put a polynomial to a file */

extern ZENPoly _ZENZToPoly ___((BigNum,BigNumLength,ZENRing));
extern BigNum _ZENPolyToZ ___((BigNumLength*,ZENPoly,ZENRing));

/* Procedures on matrices
 ***************************/
/* Allocation */
extern ZENMat _ZENMatCreate ___((Dim,Dim,ZENRing));
extern ZENMat _ZENMatRowCreate ___((Dim,Dim,ZENRing));
extern ZENMat _ZENMatColCreate ___((Dim,Dim,ZENRing));
extern ZENMat _ZENPermutationCreate ___((Dim,ZENMatTypeStruct,ZENRing));
extern void _ZENPermutationFree ___((ZENMat,ZENRing));

extern ZENMat _ZENMatCopy ___((ZENMat,ZENRing));

/* Assignment */
extern int __ZENMatAssign ___((ZENMat,ZENMat,ZENRing));
extern int __ZENMatSetToOne ___((ZENMat,ZENRing));
extern int __ZENMatIsOne ___((ZENMat,ZENRing));
extern int __ZENMatSetRandom ___((ZENMat,ZENRing));
extern ZENElt __ZENMatGetBlocPtr ___((ZENMat,Dim,Dim,ZENRing));
extern int __ZENMatGetBloc ___((ZENElt *,ZENMat,Dim,Dim,ZENRing));
extern int __ZENMatSetBloc ___((ZENMat,Dim,Dim,ZENElt,ZENRing));
extern int __ZENMatGetDigit ___((ZENElt *,ZENMat,Dim,Dim,ZENRing));
extern ZENElt __ZENMatGetDigitPtr ___((ZENMat,Dim,Dim,ZENRing));
extern int __ZENMatSetDigit ___((ZENMat,Dim,Dim,ZENElt,ZENRing));
extern int __ZENMatPermuteBn ___((ZENMat,Dim,Dim,ZENRing));
extern int __ZENMatPermuteDigit ___((ZENMat,Dim,Dim,ZENRing));

extern int _ZENPermutationAssign ___((ZENMat,ZENMat,ZENRing));
extern int _ZENPermutationSetToOne ___((ZENMat,ZENRing));
extern int _ZENPermutationIsOne ___((ZENMat,ZENRing));
extern int _ZENPermutationSetRandom ___((ZENMat,ZENRing));
extern int _ZENPermutationGetBloc ___((ZENElt *,ZENMat,Dim,Dim,ZENRing));
extern int _ZENPermutationSetBloc ___((ZENMat,Dim,Dim,ZENElt,ZENRing));
extern int _ZENPermutationGetDigit ___((ZENElt *,ZENMat,Dim,Dim,ZENRing));
extern int _ZENPermutationSetDigit ___((ZENMat,Dim,Dim,ZENElt,ZENRing));
extern int _ZENPermutationPermuteBn ___((ZENMat,Dim,Dim,ZENRing));

extern int _ZENMatAssign ___((ZENMat,ZENMat,ZENRing));
extern int _ZENPermutation2Matrix ___((ZENMat,ZENMat,ZENRing));
extern int _ZENMatSetToZero ___((ZENMat,ZENRing));
extern int _ZENMatSetToOne ___((ZENMat,ZENRing));
extern int _ZENMatSetRandom ___((ZENMat,ZENRing));
extern int _ZENMatPermuteBn ___((ZENMat,Dim,Dim,ZENRing));
extern int _ZENMatPermuteDigit ___((ZENMat,Dim,Dim,ZENRing));
extern int _ZENMatFastPermuteBn ___((ZENMat,Dim,Dim,ZENRing));

/*Tests */
extern int _ZENMatIsZero ___((ZENMat,ZENRing));
extern int _ZENMatIsOne ___((ZENMat,ZENRing));
extern int _ZENMatAreEqual ___((ZENMat,ZENMat,ZENRing));

/*Arithmetic */
extern int _ZENMatMultiply ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern void _ZENMatTranspose ___((ZENMat,ZENRing));
extern int _ZENMatAdd ___((ZENMat,ZENMat,ZENRing));
extern int _ZENMatAddScalar ___((ZENMat,ZENElt,ZENRing));
extern int _ZENMatNegate ___((ZENMat,ZENMat,ZENRing));
extern int _ZENMatSubtract ___((ZENMat,ZENMat,ZENRing));
extern int _ZENMatMultiplyPlain ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern int _ZENMatMultiplyScalar ___((ZENMat,ZENElt,ZENRing));
extern int _ZENMatWinograd ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern int _ZENMatMultiplyRightPermutation ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern int _ZENMatMultiplyLeftPermutation ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern int _ZENMatMultiplyPermutations ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern void _ZENMatMultiplyPermRRR ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern void _ZENMatMultiplyPermRCC ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern void _ZENMatMultiplyPermCCR ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern int _ZENMatRank ___((ZENMat,Dim *,ZENRing));
extern int _ZENMatGauss ___((ZENMat, ZENMat, ZENMat, Dim *, ZENElt *, int (*) ___((ZENMat,int,int,ZENRing)), ZENRing));
extern int _ZENMatGaussPlain ___((ZENMat,ZENMat,ZENMat,Dim *,ZENRing));
extern int _ZENMatGaussInverse ___((ZENMat,ZENMat,ZENRing));
extern int _ZENMatGaussLazy ___((ZENMat,ZENMat,Dim *,ZENRing));
extern int _ZENMatInverse ___((ZENMat,ZENMat,ZENRing));
extern int __ZENMatInverse ___((ZENMat,ZENMat,ZENRing));
extern int _ZENMatKernel ___((ZENMat *,ZENMat,ZENRing));
extern int _ZENMatDet ___((ZENMat, ZENElt *, ZENRing));

/* io */
extern int _ZENMatReadFromString ___((ZENMat *,char *,int,ZENRing)); /* Read a matrix from a string */
extern char* _ZENMatPrintToString ___((ZENMat,int,ZENRing)); /* Print a matrix to a string */
extern int _ZENMatReadFromFile ___((ZENMat *,FILE *,int,ZENRing));      /* Read a matrix from a file */
extern int _ZENMatPrintToFile ___((FILE *,ZENMat,int,ZENRing)); /* Print a matrix to a file */
extern int _ZENMatGetFromString ___((ZENMat *,char *,ZENRing)); /* Get a matrix from a string */
extern char* _ZENMatPutToString ___((ZENMat,ZENRing)); /* Put a matrix to a string */
extern int _ZENMatGetFromFile ___((ZENMat *,FILE *,ZENRing)); /* Get a matrix from a file */
extern int _ZENMatPutToFile ___((FILE *,ZENMat,ZENRing)); /* Put a matrix to a file */

/* matrix conversion */
extern int _ZENMatConvert ___((ZENMat, ZENRing, ZENMat, ZENRing));
extern int _ZENCloneVect2Mat ___((ZENMat, ZENMat, ZENRing));
extern int _ZENCloneMat2Vect ___((ZENMat, ZENMat, ZENRing));
 
/* Procedures on series
 ***************************/
/* Allocation */
extern ZENSr _ZENSrCopy ___((ZENSr,ZENRing));

/* Assigning */
extern void _ZENSrGetCoeff ___((ZENElt *, ZENSr, int, ZENRing)); /* Get a coefficient of a valuation */
extern void _ZENSrAssign ___((ZENSr, ZENSr, ZENRing)); /* Assign a valuation with another */
extern void _ZENSrSetToZero ___((ZENSr,ZENRing)); /* Set a valuation to zero */
extern void _ZENSrSetRandom ___((ZENSr,int,int,ZENRing)); /* Set randomly a valuation */
extern void _ZENSrUpdateValuation ___((ZENSr, ZENRing)); /* Update the valuation of a valuation */

/* Test */
extern int _ZENSrAreEqual ___((ZENSr, ZENSr, ZENRing)); /* Is a valuation equal to another */
extern int _ZENSrIsZero ___((ZENSr, ZENRing)); /* Is a valuation equal to zero */

/* Arithmetic */
extern void _ZENSrAdd ___((ZENSr, ZENSr, ZENRing)); /* Addition of series */
extern void _ZENSrNegate ___((ZENSr, ZENSr, ZENRing)); /* Negation of series */
extern void _ZENSrSubtract ___((ZENSr, ZENSr, ZENRing)); /* Substraction of series */
extern int _ZENSrSquareChar2 ___((ZENSr, ZENSr, ZENRing)); /* Squaring of series in characteristic 2*/
extern int _ZENSrSquareCharp ___((ZENSr, ZENSr, ZENRing)); /* Squaring of series in characteristic p*/
extern int _ZENSrMultiply ___((ZENSr, ZENSr, ZENSr, ZENRing)); /* Multiplication of series */
extern void _ZENSrMultiplyScalar ___((ZENSr, ZENSr, ZENElt, ZENRing)); /* Multiplication of series by a scalar */
extern int _ZENSrDivide ___((ZENSr, ZENSr, ZENSr, ZENRing)); /* Division of series */

/* io */
extern int _ZENSrReadFromString ___((ZENSr *,char *,int,ZENRing)); /* Read a polynomial from a string */
extern char* _ZENSrPrintToString ___((ZENSr,int,ZENRing)); /* Print a polynomial to a string */
extern int _ZENSrReadFromFile ___((ZENSr *,FILE *,int,ZENRing));        /* Read a polynomial from a file */
extern int _ZENSrPrintToFile ___((FILE *,ZENSr,int,ZENRing)); /* Print a polynomial to a file */
extern int _ZENSrGetFromString ___((ZENSr *,char *,ZENRing)); /* Get a polynomial from a string */
extern char* _ZENSrPutToString ___((ZENSr,ZENRing)); /* Put a polynomial to a string */
extern int _ZENSrGetFromFile ___((ZENSr *,FILE *,ZENRing)); /* Get a polynomial from a file */
extern int _ZENSrPutToFile ___((FILE *,ZENSr,ZENRing)); /* Put a polynomial to a file */

/* Procedures on elliptic curves
 ******************************/

/* Allocation */
extern ZENEc _ZENEcCreate ___((ZENRing)); /* Create an elliptic curve */

extern int _ZENEcInitialize ___((ZENEc, ZENElt, ZENElt, ZENElt, ZENElt, ZENElt)); /* Initialization of an elliptic curve */
extern int _ZENEcInitializeChar2 ___((ZENEc)); /* Charateristic 2 */
extern int _ZENEcInitializeChar3 ___((ZENEc)); /* Characteristic 3 */
extern int _ZENEcInitializeCharp ___((ZENEc)); /* Charateristic greater than 2 and 3 */

extern void _ZENEcFree ___((ZENEc)); /* Free an elliptic curve */

/* Allocation */
extern ZENEcPt _ZENEcPtCreate ___((ZENEc)); /* Create a point on an elliptic curve */
extern void _ZENEcPtFree ___((ZENEcPt, ZENEc)); /* Free a point an elliptic curve */

/* Assigning */
extern void _ZENEcPtSetToZero ___((ZENEcPt, ZENEc)); /* Set a point on an elliptic curve to zero */
extern int _ZENEcPtSetNext ___((ZENEcPt, ZENEcPt, ZENEc)); /* Enumerate points on an elliptic curve */
extern int _ZENEcPtSetNextChar23p ___((ZENEcPt, ZENEcPt, ZENEc));

extern void _ZENEcPtAssign ___((ZENEcPt, ZENEcPt, ZENEc)); /* Assigning points on an elliptic curve */

extern void _ZENEcPtGetX ___((ZENElt *, ZENEcPt, ZENEc)); /* Get the abscissa of a point on an elliptic curve */
extern void _ZENEcPtGetXChar23p ___((ZENElt *, ZENEcPt, ZENEc));

extern void _ZENEcPtGetY ___((ZENElt *, ZENEcPt, ZENEc)); /* Get the ordinate of a point on an elliptic curve */
extern void _ZENEcPtGetYChar23p ___((ZENElt *, ZENEcPt, ZENEc));

extern void _ZENEcPtSetXY ___((ZENEcPt, ZENElt, ZENElt, ZENEc)); /* Set the abscissa and ordinate of a point on an elliptic curve */
extern void _ZENEcPtSetXYChar23p ___((ZENEcPt, ZENElt, ZENElt, ZENEc));
  
extern  void _ZENEcPtGeneral2Canonical ___((ZENEcPt, ZENEcPt, ZENEc)); /* Compute a general point from a canonical */
extern  void _ZENEcPtCanonical2General ___((ZENEcPt, ZENEcPt, ZENEc)); /* Compute a canonical point from a general */

/* Test */
extern int _ZENEcPtAreEqual ___((ZENEcPt, ZENEcPt, ZENEc)); /* Are two points equal */
extern int _ZENEcPtIsZero ___((ZENEcPt, ZENEc)); /* Is a point equal to zero */

extern int _ZENEcPtIsOnEc ___((ZENEcPt, ZENEc)); /* Is a point on an elliptic curve */
extern int _ZENEcPtIsOnEcChar23p ___((ZENEcPt, ZENEc));

/* Arithmetic */
extern int _ZENEcPtAdd ___((ZENEcPt, ZENEcPt, ZENEcPt, ZENEc));	/* Add 2 points on an elliptic curve */

extern void _ZENEcPtPreAdd ___((ZENEcPt, ZENEcPt, ZENEc));

extern void _ZENEcPtPostAddAllChar ___((ZENEcPt, ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtPostAddChar2J0 ___((ZENEcPt, ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtPostAddChar2J1 ___((ZENEcPt, ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtPostAddChar3J0 ___((ZENEcPt, ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtPostAddChar3J1 ___((ZENEcPt, ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtPostAddCharp ___((ZENEcPt, ZENEcPt, ZENEcPt, ZENEc));

extern int _ZENEcPtDouble ___((ZENEcPt, ZENEcPt, ZENEc)); /* Doubling one point on an elliptic curve */

extern void _ZENEcPtPreDoubleAllChar ___((ZENEcPt, ZENEc));
extern void _ZENEcPtPreDoubleChar2J0 ___((ZENEcPt, ZENEc));
extern void _ZENEcPtPreDoubleChar2J1 ___((ZENEcPt, ZENEc));
extern void _ZENEcPtPreDoubleChar3J0 ___((ZENEcPt, ZENEc));
extern void _ZENEcPtPreDoubleChar3J1 ___((ZENEcPt, ZENEc));
extern void _ZENEcPtPreDoubleCharp ___((ZENEcPt, ZENEc));

extern void _ZENEcPtPostDoubleAllChar ___((ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtPostDoubleChar2J0 ___((ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtPostDoubleChar2J1 ___((ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtPostDoubleChar3J0 ___((ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtPostDoubleChar3J1 ___((ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtPostDoubleCharp ___((ZENEcPt, ZENEcPt, ZENEc));

extern void _ZENEcPtNegateAllChar ___((ZENEcPt, ZENEcPt, ZENEc)); /* Negation of a point on an elliptic curve */
extern void _ZENEcPtNegateChar2J0 ___((ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtNegateChar2J1 ___((ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtNegateChar3 ___((ZENEcPt, ZENEcPt, ZENEc));
extern void _ZENEcPtNegateCharp ___((ZENEcPt, ZENEcPt, ZENEc));

extern int _ZENEcPtSubtract ___((ZENEcPt, ZENEcPt, ZENEcPt, ZENEc)); /* Subtraction of points on an elliptic curve */

extern int _ZENEcPtMult ___((ZENEcPt, BigNum, BigNumLength, ZENEcPt, ZENEc)); /* Multiplication of a point by an integer */

extern int _ZENMEcPtAdd  ___((ZENEcPt *, ZENEcPt *, ZENEcPt *, ZENEc *, char *, int, ZENRing));
extern int _ZENMEcPtMult ___((ZENEcPt *, BigNum *, BigNumLength *, ZENEcPt *, ZENEc *, int, ZENRing));



/* io */
extern int _ZENEcPtReadFromString ___((ZENEcPt,char *,int,ZENEc)); /* Read a point from a string */
extern char* _ZENEcPtPrintToString ___((ZENEcPt,int,ZENEc)); /* Print a point to a string */
extern int _ZENEcPtReadFromFile ___((ZENEcPt,FILE *,int,ZENEc)); /* Read a point from a file */
extern int _ZENEcPtPrintToFile ___((FILE *,ZENEcPt,int,ZENEc));	/* Print a point to a file */
extern int _ZENEcPtGetFromString ___((ZENEcPt,char *,ZENEc)); /* Get a point from a string */
extern char* _ZENEcPtPutToString ___((ZENEcPt,ZENEc)); /* Put a point to a string */
extern int _ZENEcPtGetFromFile ___((ZENEcPt,FILE *,ZENEc));	/* Get a point from a file */
extern int _ZENEcPtPutToFile ___((FILE *,ZENEcPt,ZENEc)); /* Put a point to a file */

/* Debugging functions
 **********************/

extern int ZBP ___((BigNum, BigNumLength));
extern int ZBR ___((BigNum *, BigNumLength *));
extern int ZEP ___((ZENElt, ZENRing));
extern int ZFP ___((ZENElt, ZENRing));
extern int ZER ___((ZENElt, ZENRing));
extern int ZPP ___((ZENPoly, ZENRing));
extern int ZPR ___((ZENPoly *, ZENRing));
extern int ZMP ___((ZENMat, ZENRing));
extern int ZMR ___((ZENMat *, ZENRing));
extern int ZSP ___((ZENSr, ZENRing));
extern int ZSR ___((ZENSr *, ZENRing));

#endif
/* 
   \TEX{
   \Section{Procedures on elements}
   \input{zegelt_assign}
   \input{zegelt_exp}
   \input{zegelt_sqrt}

   \Section{Procedures on polynomials}
   \input{zegpoly_assign}
   \input{zegpoly_test}
   \input{zegpoly_add}
   \input{zegpoly_multiply}
   \input{zegpoly_derivative}
   \input{zegpoly_gcd.tex}
   \input{zegpoly_roots}
   \input{zegpoly_io}


   \Section{Procedures on matrices}
   \input{zegmat_matrix}
   \input{zegmat_create}
   \input{zegmat_assign}
   \input{zegmat_convert}
   \input{zegmat_permute}
   \input{zegmat_test}
   \input{zegmat_add}
   \input{zegmat_multiply}
   \input{zegmat_gauss}
   \input{zegmat_inverse}
   \input{zegmat_io}

   \Section{Procedures on permutations}
   \input{zegpermutation_create}
   \input{zegpermutation_assign}
   \input{zegpermutation_ops}
   
   \Section{Procedures on series}
   \input{zegsr_assign}
   \input{zegsr_test}
   \input{zegsr_add}
   \input{zegsr_multiply}
   \input{zegsr_io}

   \Section{Procedures on elliptic curves}
   \input{zegec_alloc}
   \input{zegecpt_alloc}
   \input{zegecpt_assign}
   \input{zegecpt_test}
   \input{zegecpt_add}
   \input{zegecpt_add_char2}
   \input{zegecpt_add_char3}
   \input{zegecpt_add_charp}
   \input{zegecpt_multiply}
   \input{zegec_io}

   \Section{Miscellanous}
   \input{zegdebug}
   \input{zegclone_io}
   }
   */

