/*
 * zen/include/zelog.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZELOG__
#define __ZELOG__
#include "zen.h"

/* \TEX{
   In zelog.h
   }
   */

#define ZelogElt2Int(a, Rg) ZENElt2Int(a, Rg)
#define ZelogRingPms(R) ZENRingPms(R).zelog

#define ZelogRingDef(Rg) ZENRingDef(Rg)
#define ZelogRingOrigin(R)  ZENRingOrigin(R) 
#define ZelogRingCln(R)  ZENRingCln(R) 

#define ZelogGenerator(R) ZelogRingPms(R).Gen
#define ZelogMinusOne(R) ZelogRingPms(R).MinusOne
#define ZelogHbuf(R) ZelogRingPms(R).Hbuf

/* The tables */
#define ZelogAddTablePtr(R)  (ZelogRingPms(R).add)
#define ZelogAddTable(R,i)  ((*ZelogAddTablePtr(R))[i])

#define ZelogLogTablePtr(R)  (ZelogRingPms(R).log)
#define ZelogLogTable(R,i)  ((*ZelogLogTablePtr(R))[i])

#define ZelogTrTablePtr(R)  (ZelogRingPms(R).tr)
#define ZelogTrTable(R,i)  ((*ZelogTrTablePtr(R))[i])

#define ZelogConvTablePtr(R)  (ZelogRingPms(R).conv)
#define ZelogConvTable(R,i)  ((ZelogConvTablePtr(R))[i])

/* On elements
 *************/
/* Allocation */
#define ZelogEltCreate(a, Rg) 0
#define ZelogEltFree(a, Rg)

/* Assigning */
#define ZelogAnyEltAssign(a,b,Rg)  ((a) = (b))
#define ZelogAnyEltSetToZero(a,Rg) ((a) = 0)
#define ZelogAnyEltSetToOne(a,Rg) ((a) = 1)
#define ZelogAnyEltSetRandom(a,Rg) ((a) = zbnrandom()% ZBNDigit(ZENRingQ(Rg), 0))
#define ZelogAnyEltSetNext(a,Rg) ((a) = (++(a) == ZBNDigit(ZENRingQ(Rg), 0) ? 0 : a))

#define ZelogEltAssign(a,b,Rg) ZelogAnyEltAssign(ZelogElt2Int(a, Rg),ZelogElt2Int(b, Rg),Rg)
#define ZelogEltSetToZero(a,Rg) ZelogAnyEltSetToZero(ZelogElt2Int(a, Rg),Rg)
#define ZelogEltSetToOne(a,Rg) ZelogAnyEltSetToOne(ZelogElt2Int(a, Rg),Rg)
#define ZelogEltSetRandom(a,Rg) ZelogAnyEltSetRandom(ZelogElt2Int(a, Rg),Rg)
#define ZelogEltSetNext(a,Rg) ZelogAnyEltSetNext(ZelogElt2Int(a, Rg),Rg)

/* Test */
#define ZelogAnyEltAreEqual(a,b,Rg) ((a) == (b))
#define ZelogAnyEltIsZero(a,Rg) ((a) == 0)
#define ZelogAnyEltIsOne(a,Rg) ((a) == 1)

#define ZelogEltAreEqual(a,b,Rg) ZelogAnyEltAreEqual(ZelogElt2Int(a, Rg),ZelogElt2Int(b, Rg),Rg)
#define ZelogEltIsZero(a,Rg) ZelogAnyEltIsZero(ZelogElt2Int(a, Rg),Rg)
#define ZelogEltIsOne(a,Rg) ZelogAnyEltIsOne(ZelogElt2Int(a, Rg),Rg)

/* Arithmetic */

#define ZelogAnyEltAdd(r,a,Rg) if (1) {					\
if (!r) {						\
  r = a;				\
}									\
else if (a) {						\
  BigNumDigit f;                \
  if (a > r) {			\
    if ((ZelogElt2Int(ZelogHbuf(Rg),Rg)=ZelogAddTable(Rg,a-r+1))){ \
      f =r+ZelogElt2Int(ZelogHbuf(Rg),Rg) -1;						\
    }									\
    else {							        \
      f = 0;						\
    }									\
  }									\
  else {								\
    if ((ZelogElt2Int(ZelogHbuf(Rg),Rg)=ZelogAddTable(Rg,r-a+1))){ \
      f = a+ZelogElt2Int(ZelogHbuf(Rg),Rg)-1;			\
    }									\
    else {								\
      f = 0;						\
    }									\
  }									\
  if (f >= ZBNDigit(ZENRingQ(Rg),0)) {			\
    f -= ZBNDigit(ZENRingQ(Rg),0)-1;			\
  }									\
  r = f; \
}} else

#define ZelogAnyEltTrace(r,a,Rg) r = ZelogTrTable(Rg,a) 

#define ZelogAnyEltMultiply(r,a,b,Rg) if(1) {\
BigNumDigit f;\
f= \
(!a || !b ? 0 : \
 a+b-1); \
 r=f>=ZBNDigit(ZENRingQ(Rg),0) ? \
 f-ZBNDigit(ZENRingQ(Rg),0)+1: \
 f;}

#define ZelogAnyEltInverse(r,a,Rg) (\
a > 1 ? \
(r=ZBNDigit(ZENRingQ(Rg),0)-a+1,ZEN_HAS_INVERSE): \
(!a ? \
(ZENSetError(ZEN_LIB_ZELOG, ZEN_INVERSE, ZEN_DIVBYZERO), ZERR):\
(r = 1, ZEN_HAS_INVERSE)))
#define ZelogAnyEltNegate(r,a,Rg) ZelogAnyEltMultiply(r,ZelogElt2Int(ZelogMinusOne(Rg), Rg),a,Rg)
#define ZelogAnyEltSubtract(r,a,Rg) if(1) {\
  BigNumDigit t; \
  ZelogAnyEltNegate(t, a, Rg);\
  ZelogAnyEltAdd(r, t, Rg);\
} else
#define ZelogAnyEltSquare(r,a,Rg) ZelogAnyEltMultiply(r,a,a,Rg)

#define ZelogEltNegate(r,a,Rg) ZelogEltMultiply(r,ZelogMinusOne(Rg),a,Rg)
#define ZelogEltAdd(r,a,Rg) ZelogAnyEltAdd(ZelogElt2Int(r, Rg),ZelogElt2Int(a, Rg),Rg)
#define ZelogEltSubtract(r,a,Rg) ZelogAnyEltSubtract(ZelogElt2Int(r, Rg),ZelogElt2Int(a, Rg),Rg)
#define ZelogEltMultiply(r,a,b,Rg) ZelogAnyEltMultiply(ZelogElt2Int(r, Rg),ZelogElt2Int(a, Rg),ZelogElt2Int(b, Rg),Rg)
#define ZelogEltSquare(r,a,Rg) ZelogAnyEltSquare(ZelogElt2Int(r, Rg),ZelogElt2Int(a, Rg),Rg)
#define ZelogEltInverse(r,a,Rg) ZelogAnyEltInverse(ZelogElt2Int(r, Rg),ZelogElt2Int(a, Rg),Rg)
#define ZelogEltTrace(r,a,Rg) ZelogAnyEltTrace(ZelogElt2Int(r, Rg),ZelogElt2Int(a, Rg),Rg) 

/* io */
#define ZelogEltReadFromString(a,s,base,Rg) _ZelogEltReadFromString(&(a),s,base,Rg) 
#define ZelogEltPrintToString(a,base,Rg) _ZelogEltPrintToString(a,base,Rg) 
#define ZelogEltReadFromFile(a,file,base,Rg) _ZelogEltReadFromFile(&(a),file,base,Rg)
#define ZelogEltPrintToFile(file,a,base,Rg) _ZelogEltPrintToFile(file,a,base,Rg) 

#define ZelogEltGetFromString(a,s,Rg) _ZelogEltGetFromString(&(a),s,Rg) 
#define ZelogEltPutToString(a,Rg) ZENEltPutToString(ZelogEltElt(Rg,a),ZelogRing(Rg)) 
#define ZelogEltGetFromFile(a,file,Rg) _ZelogEltGetFromFile(&(a),file,Rg)
#define ZelogEltPutToFile(file,a,Rg) ZENEltPutToFile(file,ZelogEltElt(Rg,a),ZelogRing(Rg)) 

#define ZelogEltToZ(p,p_pl,a,Rg) _ZelogEltToZ(p,p_pl,a,Rg) 

#define ZelogAnyZToElt(a,p,p_pl,Rg) ((a) = ZelogLogTable(Rg,ZBNDigit(p, 0)))
#define ZelogZToElt(a,p,p_pl,Rg) ZelogAnyZToElt(ZelogElt2Int(a, Rg),p,p_pl,Rg)

/* On polynomials
 ****************/
#define ZelogPolyDeg(p,Rg) ZENPolyDeg(p,Rg)
#define ZelogPolyLgt(p,Rg) ZENPolyLgt(p,Rg)

#define ZelogPolyCof(r,Rg) ZENPolyShort(r,Rg)
#define ZelogPolyGetCof(r,i,Rg) ZENPolyGetShort(r,i,Rg)

/* Allocation */
#define ZelogPolyCreate(d, Rg) _ZelogPolyCreate(d, Rg) 
#define ZelogPolyCreateLarge(d, Rg) _ZelogPolyCreate(d, Rg) 
#define ZelogPolyFree(r, Rg) _ZelogPolyFree(r, Rg) 
#define ZelogPolyCopy(PX,Rg) ZENPolyCopy(PX,Rg)

/* Assigning */
#define ZelogPolyAssign(r,PX,Rg) _ZelogPolyAssign(r,PX,Rg)
#define ZelogPolySetToZero(r,Rg) _ZelogPolySetToZero(r,Rg)
#define ZelogPolySetToXi(r,i,Rg) _ZelogPolySetToXi(r,i,Rg)
#define ZelogPolySetRandom(r,d,Rg) _ZelogPolySetRandom(r,d,Rg)
#define ZelogPolySetCoeff(r,i,b,Rg) ZelogAnyEltAssign(ZelogPolyGetCof(r, i, Rg), ZelogElt2Int(b, Rg), Rg)
#define ZelogPolyGetCoeff(c, r,i,Rg) ZelogAnyEltAssign(ZelogElt2Int(c, Rg), ZelogPolyGetCof(r, i, Rg), Rg)
#define ZelogPolyGetCoeffPtr(r,i,Rg) _ZelogPolyGetCoeffPtr(r,i,Rg)

#define ZelogPolyUpdateDegree(r,Rg) _ZelogPolyUpdateDegree(r,Rg)

/* Test */
#define ZelogPolyAreEqual(r,PX,Rg) _ZelogPolyAreEqual(r,PX,Rg)
#define ZelogPolyIsZero(r,Rg) (!(r)->deg && !ZelogPolyGetCof(r, 0, Rg))
#define ZelogPolyIsXi(p,i,Rg) _ZelogPolyIsXi(p.i,Rg)

/* Arithmetic */
#define ZelogPolyAdd(r,PX,Rg) _ZelogPolyAdd(r,PX,Rg)
#define ZelogPolySubtract(r,PX,Rg) _ZelogPolySubtract(r,PX,Rg)
#define ZelogPolyNegate(r,PX,Rg) _ZelogPolyNegate(r,PX,Rg)

#define ZelogPolySquareChar2(c,r,Rg) _ZelogPolySquareChar2(c,r,Rg)
#define ZelogPolySquareCharp(c,r,Rg) _ZelogPolySquareCharp(c,r,Rg)
#define ZelogPolySquareKaratsuba(c,r,Rg) _ZelogPolySquareKaratsuba(c,r,Rg)
#define ZelogPolySquareLazy(c,r,Rg) _ZelogPolySquareKaratsuba(c,r,Rg)
#define ZelogPolySquareKaratsubaLazy(c,p,Rg) _ZelogPolySquareKaratsubaLazy(c,p,Rg)

#define ZelogPolyMultiply(c,r,PX,Rg) _ZelogPolyMultiply(c,r,PX,Rg)
#define ZelogPolyMultiplyLazy(c,r,PX,Rg) _ZelogPolyMultiplyLazy(c,r,PX,Rg)
#define ZelogPolyMultiplyPlain(c,r,PX,Rg) _ZelogPolyMultiplyPlain(c,r,PX,Rg)
#define ZelogPolyMultiplyKaratsuba(c,r,PX,Rg) _ZelogPolyMultiplyKaratsuba(c,r,PX,Rg)
#define ZelogPolyMultiplyKaratsubaLazy(c,r,PX,Rg) _ZelogPolyMultiplyKaratsubaLazy(c,r,PX,Rg)

#define ZelogPolyMultiplyScalar(r,p,e,Rg) _ZelogPolyMultiplyScalar(r,p,e,Rg)
#define ZelogPolyDot(r,p,q,R) _ZelogPolyDot(r,p,q,R)

#define ZelogPolyDivide(r,MX,PX,QX,Rg) _ZelogPolyDivide(r,MX,PX,QX,Rg)
#define ZelogPolyDivideLazy(r,MX,PX,QX,Rg) _ZelogPolyDivideLazy(r,MX,PX,QX,Rg)
#define ZelogPolyModuloLazy(MX,PX,QX,k,Rg) _ZelogPolyModuloLazy(MX,PX,QX,k,Rg)

/* io */
#define ZelogPolyReadFromString(r,s,base,Rg) ZENPolyReadFromString(r,s,base,Rg) 
#define ZelogPolyPrintToString(r,base,Rg) ZENPolyPrintToString(r,base,Rg)
#define ZelogPolyReadFromFile(r,file,base,Rg) ZENPolyReadFromFile(r,file,base,Rg) 
#define ZelogPolyPrintToFile(file,r,base,Rg) ZENPolyPrintToFile(file,r,base,Rg)

#define ZelogPolyGetFromString(r,s,Rg) ZENPolyReadFromString(r,s,10,Rg)
#define ZelogPolyPutToString(r,Rg) ZENPolyPrintToString(r,10,Rg)
#define ZelogPolyGetFromFile(r,file,Rg) ZENPolyReadFromFile(r,file,10,Rg) 
#define ZelogPolyPutToFile(file,r,Rg) ZENPolyPrintToFile(file,r,10,Rg)

/* On matrices
 ****************/
#define ZelogMatNbPtr(M,R) ZENMatNbPtr(M,R)
#define ZelogMatSize(M,R) ZENMatSize(M,R)
#define ZelogMatNbBn(M,R) ZENMatNbBn(M,R)
#define ZelogMatNbBnd(M,R) ZENMatNbBnd(M,R)
#define ZelogMatNbDigit(M,R) ZENMatNbDigit(M,R)
#define ZelogMatNbBloc(M,R) ZENMatNbBloc(M,R)
#define ZelogMatType(M,R) ZENMatType(M,R)
#define ZelogMatFirst(M,R) ZENMatFirst(M,R)

#define ZelogMatIsRowType(M,R) ZENMatIsRowType(M,R)
#define ZelogMatIsPermutation(M,R) ZENMatIsPermutation(M,R)
#define ZelogMatAreSameType(A,B,R) ZENMatAreSameType(A,B,R)

#define ZelogMatNbRow(M,R) ZENMatNbRow(M,R)
#define ZelogMatNbCol(M,R) ZENMatNbCol(M,R)



/* Allocation */
#define ZelogMatCreate(l,c,R) ZENMatCreate(l,c,R)
#define ZelogMatRowCreate(l,c,R) ZENMatRowCreate(l,c,R)
#define ZelogMatColCreate(l,c,R) ZENMatColCreate(l,c,R)
#define ZelogPermutationCreate(n,t,R) ZENPermutationCreate(n,t,R)
#define ZelogPermutationRowCreate(n,R) ZENPermutationRowCreate(n,R)
#define ZelogPermutationColCreate(n,R) ZENPermutationColCreate(n,R)
#define ZelogMatFree(M,R) ZENMatFree(M,R)
#define ZelogMatCopy(M,R) ZENMatCopy(M,R)

/* Assignation */
#define ZelogMatFastPermute(M,l1,l2,R) ZENMatFastPermute(M,l1,l2,R)
#define ZelogMatGetCoeffPtr(M,l,c,R) ZENMatGetCoeffPtr(M,l,c,R)  
#define ZelogMatGetCoeff(Z,M,l,c,R) ZENMatGetCoeff(Z,M,l,c,R)  
#define ZelogMatSetCoeff(M,l,c,Z,R) ZENMatSetCoeff(M,l,c,Z,R)

#define ZelogMatPermuteRow(M,l1,l2,R) ZENMatPermuteRow(M,l1,l2,R)
#define ZelogMatPermuteCol(M,l1,l2,R) ZENMatPermuteCol(M,l1,l2,R)


/* Arithmetic */
#define ZelogMatMultiply(X,A,B,R) ZENMatMultiply(X,A,B,R)
#define ZelogMatSquare(A,B,R) ZENMatSquare(A,B,R)

/* io */
#define ZelogMatReadFromString(r,s,base,Rg) ZENMatReadFromString(r,s,base,Rg)
#define ZelogMatPrintToString(r,base,Rg) ZENMatPrintToString(r,base,Rg)
#define ZelogMatReadFromFile(r,file,base,Rg) ZENMatReadFromFile(r,file,base,Rg)
#define ZelogMatPrintToFile(file,r,base,Rg) ZENMatPrintToFile(file,r,base,Rg)

/* Following the local Zelog functions */

/* Allocation */
#define ZelogMatCreateParams(ts,l,c,p_pms,R) _ZelogMatCreateParams(ts,l,c,p_pms,R)

/* Assigning */
#define ZelogMatAssign(A,B,R) _ZelogMatAssign(A,B,R)
#define ZelogMatSameTypeAssign(A,B,R) ZBNAssign(ZENMatFirst(A,R),ZENMatFirst(B,R),ZENMatNbBnd(A,R))
#define ZelogMatSetToZero(A,R) ZBNSetToZero(ZENMatFirst(A,R),ZENMatNbBnd(A,R))
#define ZelogMatSetToOne(M,R) _ZelogMatSetToOne(M,R)
#define ZelogMatSetRandom(M,R) _ZelogMatSetRandom(M,R)

#define ZelogMatSetBloc(M,l,c,Z,R) _ZelogMatSetBloc(M,l,c,Z,R)
#define ZelogMatGetBloc(Z,M,l,c,R) _ZelogMatGetBloc(Z,M,l,c,R)
#define ZelogMatSetDigit(M,l,c,Z,R) ZelogMatSetBloc(M,c,l,Z,R)
#define ZelogMatGetDigit(Z,M,l,c,R) ZelogMatGetBloc(Z,M,c,l,R)
 
#define ZelogMatPermuteBn(M,l1,l2,R) _ZelogMatPermuteBn(M,l1,l2,R)
#define ZelogMatPermuteDigit(M,l1,l2,R) _ZelogMatPermuteDigit(M,l1,l2,R)

/* Test */
#define ZelogMatSameTypeAreEqual(A,B,R) (!(ZBNCompare(ZENMatFirst(A,R),ZENMatNbBnd(A,R),ZENMatFirst(B,R),ZENMatNbBnd(B,R))))
#define ZelogMatIsZero(A,R) ZBNIsZero(ZENMatFirst(A,R),ZENMatNbBnd(A,R))
#define ZelogMatIsOne(M,R) _ZelogMatIsOne(M,R)

/* Arithmetic */
#define ZelogMatAdd(A,B,R) _ZelogMatAdd(A,B,R)
#define ZelogMatSameTypeAdd(A,B,R) _ZelogMatSameTypeAdd(A,B,R)
#define ZelogMatAddScalar(A,b,R) _ZelogMatAddScalar(A,b,R)
#define ZelogMatNegate(A,B,R) _ZelogMatNegate(A,B,R)
#define ZelogMatSubtract(A,B,R) _ZelogMatSubtract(A,B,R)
#define ZelogMatMultiplyPlain(X,A,B,R) _ZelogMatMultiplyPlain(X,A,B,R)
#define ZelogMatMultiplyScalar(A,b,R) _ZelogMatMultiplyScalar(A,b,R)
#define ZelogMatGaussPlain(D,S,P,p_rk,R) ZENMatGaussPlain(D,S,P,p_rk,R)
#define ZelogMatGaussRank(D,p_rk,R) ZENMatGaussRank(D,p_rk,R)
#define ZelogMatGaussLazy(D,P,p_rk,R) ZENMatGaussLazy(D,P,p_rk,R)
#define ZelogMatGaussInverse(D,S,R) ZENMatGaussInverse(D,S,R)
#define ZelogMatInverse(I,M,R) _ZelogMatInverse(I,M,R)
#define ZelogMatKernel(K,M,R) _ZelogMatKernel(K,M,R)

/* io */
#define ZelogMatGetFromString(r,s,Rg) ZENMatGetFromString(r,s,Rg)
#define ZelogMatPutToString(r,Rg) ZENMatPutToString(r,Rg)
#define ZelogMatGetFromFile(r,file,Rg) ZENMatGetFromFile(r,file,Rg)
#define ZelogMatPutToFile(file,r,Rg) ZENMatPutToFile(file,r,Rg)

 
/* Procedure on rings
 ***********************/
extern int _ZelogInit ___((ZENRing));
extern int _ZelogClose ___((ZENRing));

extern int _ZelogPrcInitPolyRootsCanonical ___((ZENRing));
extern int _ZelogPrcClosePolyRootsCanonical ___((ZENRing));

/* Procedure on elements
 ***********************/

/* Allocation */
extern int _ZelogEltCreate ___((ZENElt *, ZENRing)); 
extern void _ZelogEltFree ___((ZENElt, ZENRing));

/* Assigning */
extern void _ZelogEltAssign ___((ZENElt *, ZENElt, ZENRing));
extern void _ZelogEltSetToZero ___((ZENElt *, ZENRing));
extern void _ZelogEltSetToOne ___((ZENElt *, ZENRing));
extern void _ZelogEltSetRandom ___((ZENElt *, ZENRing));
extern void _ZelogEltSetNext ___((ZENElt *, ZENRing));

extern void _ZelogEltToClone ___((ZENElt *, ZENElt, ZENRing));
extern void _ZelogCloneToElt ___((ZENElt *, ZENElt, ZENRing));

/* Test */
extern int _ZelogEltAreEqual ___((ZENElt, ZENElt, ZENRing));
extern int _ZelogEltIsZero ___((ZENElt, ZENRing));
extern int _ZelogEltIsOne ___((ZENElt, ZENRing));

/* Arithmetic */
extern void _ZelogEltAdd ___((ZENElt *, ZENElt, ZENRing));
extern void _ZelogEltNegate ___((ZENElt *, ZENElt, ZENRing));
extern void _ZelogEltSubtract ___((ZENElt *, ZENElt, ZENRing));
extern void _ZelogEltSquare ___((ZENElt *, ZENElt, ZENRing));
extern void _ZelogEltMultiply ___((ZENElt *, ZENElt, ZENElt, ZENRing));
extern int _ZelogEltInverse ___((ZENElt *, ZENElt, ZENRing));
extern void _ZelogEltTrace ___((ZENElt *, ZENElt, ZENRing));

/* io */
extern void _ZelogEltToZ ___((BigNum,BigNumLength*,ZENElt,ZENRing));
extern void _ZelogZToElt ___((ZENElt *,BigNum,BigNumLength,ZENRing));

/* Procedures on polynomials
 ***************************/
/* Allocation */
extern ZENPoly _ZelogPolyCreate ___((int, ZENRing)); /* Create polynomials */
extern void _ZelogPolyFree ___((ZENPoly, ZENRing)); /* Free polynomials */

/* Assigning */
extern void _ZelogPolyAssign ___((ZENPoly, ZENPoly, ZENRing)); /* Assign a polynomial with another */
extern void _ZelogPolySetToZero ___((ZENPoly, ZENRing)); /* Set a polynomial to zero */
extern void _ZelogPolySetToXi ___((ZENPoly, int, ZENRing)); /* Set a polynomial to $X^i$ */
extern void _ZelogPolySetRandom ___((ZENPoly, int, ZENRing)); /* Set randomly a polynomial */
extern void _ZelogPolySetCoeff ___((ZENPoly, int, ZENElt, ZENRing)); /* Set a coefficient of a polynomial */
extern void _ZelogPolyGetCoeff ___((ZENElt *, ZENPoly, int, ZENRing)); /* Get a coefficient of a polynomial */
extern ZENElt _ZelogPolyGetCoeffPtr ___((ZENPoly, int, ZENRing)); /* Get a pointer on a coefficient of a polynomial */
extern void _ZelogPolyUpdateDegree ___((ZENPoly, ZENRing)); /* Update the degree of a polynomial */

/* Test */
extern int _ZelogPolyAreEqual ___((ZENPoly, ZENPoly, ZENRing)); /* Is a polynomial equal to another */
extern int _ZelogPolyIsZero ___((ZENPoly, ZENRing)); /* Is a polynomial equal to zero */
extern int _ZelogPolyIsXi ___((ZENPoly, int, ZENRing)); /* Is a polynomial equal to $X^i$ */

/* Arithmetic */
extern void _ZelogPolyAdd ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZelogPolyNegate ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZelogPolySubtract ___((ZENPoly, ZENPoly, ZENRing)); /* Substraction of polynomials */

extern int _ZelogPolySquareChar2 ___((ZENPoly, ZENPoly, ZENRing)); /* Squaring of polynomials */
extern int _ZelogPolySquareCharp ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZelogPolySquareKaratsuba ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZelogPolySquareKaratsubaLazy ___((ZENPoly, ZENPoly, ZENRing));

extern int _ZelogPolyMultiply ___((ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Multiplication of polynomials */
extern int _ZelogPolyMultiplyLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZelogPolyMultiplyPlain ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZelogPolyMultiplyKaratsuba ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZelogPolyMultiplyKaratsubaLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

extern void _ZelogPolyMultiplyScalar ___((ZENPoly, ZENPoly, ZENElt, ZENRing)); /* Multiplication of polynomials */
extern void _ZelogPolyDot  ___((ZENElt *, ZENPoly, ZENPoly, ZENRing)); /* Scalar product of 2 polynomials */
extern int _ZelogPolyDivide ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Division of polynomials */
extern int _ZelogPolyDivideLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Lazy division of polynomials */
extern int _ZelogPolyModuloLazy ___((ZENPoly, ZENPoly, ZENPoly, int, ZENRing)); /* Lazy modulo of polynomials */

/* Procedures on matrices
 ***************************/
/* Allocation */
extern int _ZelogMatCreateParams ___((ZENMatTypeStruct,Dim,Dim,ZENMat *,ZENRing));

/* Assignment */
extern int _ZelogMatAssign ___((ZENMat,ZENMat,ZENRing));
extern int _ZelogMatSetToZero ___((ZENMat,ZENRing));
extern int _ZelogMatSetToOne ___((ZENMat,ZENRing));
extern int _ZelogMatSetRandom ___((ZENMat,ZENRing));
extern int _ZelogMatGetBloc ___((ZENElt *,ZENMat,Dim,Dim,ZENRing));
extern int _ZelogMatGetSubMat ___((ZENMat,ZENMat,Dim,Dim,ZENRing));
extern int _ZelogMatSetSubMat ___((ZENMat,Dim,Dim,ZENMat,ZENRing));
extern ZENElt _ZelogMatGetBlocPtr ___((ZENMat,Dim,Dim,ZENRing));
extern int _ZelogMatSetBloc ___((ZENMat,Dim,Dim,ZENElt,ZENRing));
extern int _ZelogMatPermuteBn ___((ZENMat,Dim,Dim,ZENRing));
extern int _ZelogMatPermuteDigit ___((ZENMat,Dim,Dim,ZENRing));

/*Tests */
extern int _ZelogMatIsZero ___((ZENMat,ZENRing));
extern int _ZelogMatIsOne ___((ZENMat,ZENRing));
extern int _ZelogMatAreEqual ___((ZENMat,ZENMat,ZENRing));

/*Arithmetic */
extern int _ZelogMatAdd ___((ZENMat,ZENMat,ZENRing));
extern int _ZelogMatAddScalar ___((ZENMat,ZENElt,ZENRing));
extern int _ZelogMatNegate ___((ZENMat,ZENMat,ZENRing));
extern int _ZelogMatSubtract ___((ZENMat,ZENMat,ZENRing));
extern int _ZelogMatMultiplyPlain ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern int _ZelogMatMultiplyScalar ___((ZENMat,ZENElt,ZENRing));
extern int _ZelogMatGauss ___((ZENMat,
			       ZENMat,
			       ZENMat,
			       Dim *,
			       ZENElt *,
			       int (*) ___((ZENMat,int,int,ZENRing)),
			       ZENRing));
extern int _ZelogMatInverse ___((ZENMat,ZENMat,ZENRing));
extern int _ZelogMatKernel ___((ZENMat *,ZENMat,ZENRing));

/* matrix conversion */
extern int _Zelog2mVect2Mat ___((ZENMat, ZENMat, ZENRing));
extern int _Zelog2mMat2Vect ___((ZENMat, ZENMat, ZENRing));

/* 
   \TEX{
   \Section{Ring initialization}
   \input{zeloginit}

   \Section{Element operations}
   \input{zelogelt_alloc.tex}
   \input{zelogelt_assign.tex}
   \input{zelogelt_test.tex}
   \input{zelogelt_add.tex}
   \input{zelogelt_multiply.tex}
   \input{zelogelt_inverse.tex}
   \input{zelogelt_trace.tex}
   \input{zelogelt_io.tex}

   \Section{Polynomial operations}
   \input{zelogpoly_alloc.tex}
   \input{zelogpoly_assign.tex}
   \input{zelogpoly_test.tex}
   \input{zelogpoly_add.tex}
   \input{zelogpoly_multiply.tex}

   \Section{Matricial operations}
   \input{zelogmat_create}
   \input{zelogmat_assign}
   \input{zelogmat_permute}
   \input{zelogmat_test}
   \input{zelogmat_add}
   \input{zelogmat_multiply}
   \input{zelogmat_gauss}
   \input{zelogmat_inverse}
   \input{zelogmat_convert}
   }
   */

#endif
