/*
 * zen/include/zem.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZEM__
#define __ZEM__

#include "zep.h"
#define MONTGOMERY_VERBOSE 0

#define ZemRingOne(R) ZemRingPms(R).One
#define ZemRingBn(R) ZemRingPms(R).bn
#define ZemRingBn2(R) ZemRingPms(R).bn2
#define ZemRingBn3(R) ZemRingPms(R).bn3
#define ZemRingBnl(R) ZemRingPms(R).bnl
#define ZemRingV(R) ZemRingPms(R).V
#define ZemRingVl(R) ZemRingPms(R).Vl
#define ZemRingMask(R) ZemRingPmsZemp(R).mask
#define ZemRingNbuf(R) ZemRingPmsZemp(R).Nbuf
#define ZemRingMbuf(R) ZemRingPmsZemp(R).Mbuf
#define ZemRingMbufl(R) ZemRingPmsZemp(R).Mbufl

/* \TEX{
   All the |_ZemFct| functions have corresponding macros |ZemFct|.
   The macros are always at least as fast as the function call and sometimes even faster.
   Therefore, it is best to use macros.
   }
   */
/* On elements
 *************/
#define ZemElt2Bn(a,Rg) ZepElt2Bn(a,Rg)

/* Allocation */
#define ZemEltCreate(a, Rg) ZepEltCreate(a, Rg) 
#define ZemEltFree(a, Rg) ZepEltFree(a, Rg) 

/* Assigning */
#define ZemEltAssign(a,b,Rg) ZepEltAssign(a,b,Rg) 
#define ZemEltSetToZero(a,Rg) ZepEltSetToZero(a,Rg) 
#define ZemEltSetToOne(a,Rg) ZemEltAssign(a,ZemRingOne(Rg),Rg) 
#define ZemEltSetRandom(a,Rg) ZepEltSetRandom(a,Rg) 
#define ZemEltSetNext(a,Rg) ZemEltAdd(a,ZemRingOne(Rg),Rg) 

#define ZemEltToClone(a,b,R) ZemZToElt(a,ZENElt2Bn(b,ZENRingOrigin(R)),ZENRingSizeElt(ZENRingOrigin(R)),R)
#define ZemCloneToElt(a,b,R) _ZemCloneToElt(&(a),b,R)

/* Test */
#define ZemEltAreEqual(a,b,Rg) ZepEltAreEqual(a,b,Rg) 
#define ZemEltIsZero(a,Rg) ZepEltIsZero(a,Rg) 
#define ZemEltIsOne(a,Rg) ZemEltAreEqual(a,ZemRingOne(Rg),Rg) 

/* Arithmetic */
#define ZemEltAdd(r,a,Rg) ZepEltAdd(r,a,Rg) 
#define ZemEltNegate(c,a,Rg) ZepEltNegate(c,a,Rg) 
#define ZemEltSubtract(c,a,Rg) ZepEltSubtract(c,a,Rg) 
#define ZemEltSquare(r,a,Rg) ZENEltSquare(r,a,Rg)
#define ZemEltMultiply(r,a,b,Rg) ZENEltMultiply(r,a,b,Rg)
#define ZemEltRedMod(n,R) _ZemEltRedMod(n,R)
#define ZemEltRedModKaratsuba(n,buf,R) _ZemEltRedModKaratsuba(n,buf,R)
#define ZemEltInverse(r,a,Rg) _ZemEltInverse(&(r),a,Rg)
#define ZemEltTrace(r,a,Rg) ZepEltTrace(r,a,Rg) 
#define ZemEltExp(r,k,kl,a,Rg) ZENEltExp(r,k,kl,a,Rg) 

/* io */
#define ZemEltReadFromString(a,s,base,Rg) _ZENCloneEltReadFromString(a,s,base,Rg)
#define ZemEltPrintToString(a,base,Rg) _ZENCloneEltPrintToString(a,base,Rg)
#define ZemEltReadFromFile(a,file,base,Rg) _ZENCloneEltReadFromFile(a,file,base,Rg)
#define ZemEltPrintToFile(file,a,base,Rg) _ZENCloneEltPrintToFile(file,a,base,Rg)

#define ZemEltGetFromString(a,s,Rg) ZepEltGetFromString(a,s,Rg) 
#define ZemEltPutToString(a,Rg) ZepEltPutToString(a,Rg) 
#define ZemEltGetFromFile(a,file,Rg) ZepEltGetFromFile(a,file,Rg) 
#define ZemEltPutToFile(file,a,Rg) ZepEltPutToFile(file,a,Rg) 

#define ZemEltToZ(p,p_pl,a,R) _ZemEltToZ(p,p_pl,a,R)
#define ZemZToElt(a,p,pl,R) _ZemZToElt(&(a),p,pl,R)

/* On polynomials
 ****************/
#define ZemPolyDeg(p,Rg) ZepPolyDeg(p,Rg)
#define ZemPolyLgt(p,Rg) ZepPolyLgt(p,Rg)

#define ZemPolyCof(r,Rg) ZepPolyCof(r,Rg)
#define ZemPolyGetCof(r,i,Rg) ZepPolyGetCof(r,i,Rg)

/* Allocation */
#define ZemPolyCreate(d, Rg) ZepPolyCreate(d, Rg) 
#define ZemPolyCreateLarge(d, Rg) ZepPolyCreateLarge(d, Rg) 
#define ZemPolyFree(r, Rg) ZepPolyFree(r, Rg) 
#define ZemPolyCopy(PX,Rg) ZepPolyCopy(PX,Rg)

/* Assigning */
#define ZemPolyAssign(r,PX,Rg) ZepPolyAssign(r,PX,Rg)
#define ZemPolyAssignLarge(r,PX,Rg) ZepPolyAssignLarge(r,PX,Rg)
#define ZemPolyLazyToVeryLazy(r,PX,Rg) _ZemPolyLazyToVeryLazy(r,PX,Rg)
#define ZemPolySetToZero(r,Rg) ZepPolySetToZero(r,Rg)
#define ZemPolySetToZeroLarge(r,Rg) ZepPolySetToZeroLarge(r,Rg)
#define ZemPolySetToXi(r,i,Rg) _ZemPolySetToXi(r,i,Rg)
#define ZemPolySetRandom(r,d,Rg) ZepPolySetRandom(r,d,Rg)
#define ZemPolySetCoeff(r,i,b,Rg) ZemEltAssign(ZemPolyGetCof(r, i, Rg), b, Rg)
#define ZemPolyGetCoeff(c, r,i,Rg) ZemEltAssign(c, ZemPolyGetCof(r, i, Rg), Rg)
#define ZemPolyGetCoeffPtr(r,i,Rg) ZemPolyGetCof(r,i,Rg)

#define ZemPolyDigit(r,Rg) (r)->coeff.n
#define ZemPolyGetDigit(r,i,Rg) ZBNDigit(ZemPolyDigit(r, Rg), i)
#define ZemPolySetDigit(r,i,n,Rg) ZemPolyGetDigit(r,i,Rg) = n

#define ZemPolyUpdateDegree(r,Rg) ZepPolyUpdateDegree(r,Rg)

/* Test */
#define ZemPolyAreEqual(r,PX,Rg) ZepPolyAreEqual(r,PX,Rg)
#define ZemPolyIsZero(r,Rg) (!(r)->deg && ZemEltIsZero(ZemPolyGetCoeffPtr(r, 0, Rg), Rg))
#define ZemPolyIsXi(p,i,Rg) _ZemPolyIsXi(p.i,Rg)

/* Arithmetic */
#define ZemPolyAdd(r,PX,Rg) _ZemPolyAdd(r,PX,Rg)
#define ZemPolyAddLarge(r,PX,Rg) _ZemPolyAddLarge(r,PX,Rg)
#define ZemPolySubtract(r,PX,Rg) _ZemPolySubtract(r,PX,Rg)
#define ZemPolySubtractLLarge(r,PX,Rg) _ZemPolySubtractLarge(r,PX,Rg)
#define ZemPolyNegate(r,PX,Rg) _ZemPolyNegate(r,PX,Rg)
#define ZemPolyNegateLarge(r,PX,Rg) _ZemPolyNegateLarge(r,PX,Rg)

#define ZemPolySquarePlain(c,r,Rg) _ZemPolySquarePlain (c,r,Rg)
#define ZemPolySquarePlainLazy(c,r,Rg) _ZemPolySquarePlain (c,r,Rg)
#define ZemPolySquarePlainVeryLazy(c,r,Rg) _ZepPolySquarePlainVeryLazy(c,r,Rg)

#define ZemPolySquareKaratsuba(c,r,Rg) _ZemPolySquareKaratsuba(c,r,Rg)
#define ZemPolySquareKaratsubaLazy(c,p,Rg) _ZemPolySquareKaratsubaLazy(c,p,Rg)
#define ZemPolySquareKaratsubaVeryLazy(c,p,Rg) _ZepPolySquareKaratsubaVeryLazy(c,p,Rg)

#define ZemPolyMultiply(c,r,PX,Rg) _ZemPolyMultiply(c,r,PX,Rg)
#define ZemPolyMultiplyLazy(c,r,PX,Rg) _ZemPolyMultiplyLazy(c,r,PX,Rg)
#define ZemPolyMultiplyVeryLazy(c,r,PX,Rg) _ZepPolyMultiplyVeryLazy(c,r,PX,Rg)

#define ZemPolyMultiplyPlain(c,r,PX,Rg) _ZemPolyMultiplyPlain(c,r,PX,Rg)
#define ZemPolyMultiplyPlainLazy(c,r,PX,Rg) _ZemPolyMultiplyPlain(c,r,PX,Rg)
#define ZemPolyMultiplyPlainVeryLazy(c,r,PX,Rg) _ZepPolyMultiplyPlainVeryLazy(c,r,PX,Rg)

#define ZemPolyMultiplyKaratsuba(c,r,PX,Rg) _ZemPolyMultiplyKaratsuba(c,r,PX,Rg)
#define ZemPolyMultiplyKaratsubaLazy(c,r,PX,Rg) _ZemPolyMultiplyKaratsubaLazy(c,r,PX,Rg)
#define ZemPolyMultiplyKaratsubaVeryLazy(c,r,PX,Rg) _ZepPolyMultiplyKaratsubaVeryLazy(c,r,PX,Rg)

#define ZemPolyMultiplyScalar(r,p,e,Rg) _ZemPolyMultiplyScalar(r,p,e,Rg)
#define ZemPolyDot(r,p,q,R) _ZemPolyDot(&(r),p,q,R)

#define ZemPolyDivide(r,MX,PX,QX,Rg) _ZemPolyDivide(r,MX,PX,QX,Rg)
#define ZemPolyDivideLazy(r,MX,PX,QX,Rg) _ZemPolyDivideLazy(r,MX,PX,QX,Rg)
#define ZemPolyDivideVeryLazy(r,MX,PX,QX,Rg) _ZemPolyDivideVeryLazy(r,MX,PX,QX,Rg)
#define ZemPolyModuloVeryLazy(MX,PX,QX,k,Rg) _ZemPolyModuloVeryLazy(MX,PX,QX,k,Rg)

/* io */
#define ZemPolyReadFromString(r,s,base,Rg) ZepPolyReadFromString(r,s,base,Rg) 
#define ZemPolyPrintToString(r,base,Rg) ZepPolyPrintToString(r,base,Rg)
#define ZemPolyReadFromFile(r,file,base,Rg) ZepPolyReadFromFile(r,file,base,Rg) 
#define ZemPolyPrintToFile(file,r,base,Rg) ZepPolyPrintToFile(file,r,base,Rg)

#define ZemPolyGetFromString(r,s,Rg) ZepPolyReadFromString(r,s,10,Rg)
#define ZemPolyPutToString(r,Rg) ZepPolyPrintToString(r,10,Rg)
#define ZemPolyGetFromFile(r,file,Rg) ZepPolyReadFromFile(r,file,10,Rg) 
#define ZemPolyPutToFile(file,r,Rg) ZepPolyPrintToFile(file,r,10,Rg)


/* Procedure on Rings
 ***********************/
extern int _ZemInit ___((ZENRing));
extern int _ZemClose ___((ZENRing));

extern int _ZemPrcInitPolyRootsCanonical ___((ZENRing));
extern int _ZemPrcClosePolyRootsCanonical ___((ZENRing));

extern int _ZemPrcInitEltKara ___((ZENRing));
extern int _ZemPrcCloseEltKara ___((ZENRing));

extern int _ZemPrcInitEltExp ___((ZENRing));
extern int _ZemPrcCloseEltExp ___((ZENRing));

/* Procedure on elements
 ***********************/
/* Assigning */
extern void _ZemEltSetToOne ___((ZENElt *, ZENRing)); /* Set an element to one */
extern void _ZemEltSetNext ___((ZENElt *, ZENRing)); /* Set an element to the next */

extern void _ZemEltToClone ___((ZENElt *, ZENElt, ZENRing));
extern void _ZemCloneToElt ___((ZENElt *, ZENElt, ZENRing));

extern int _ZemEltConvert ___((ZENElt *, ZENRing, ZENElt, ZENRing));

/* Test */
extern int _ZemEltIsOne ___((ZENElt, ZENRing)); /* Is an element equal to one */

/* Arithmetic */
extern BigNumLength _ZemEltRedMod ___((BigNum, ZENRing));
extern void _ZemModSquare ___((BigNum,BigNum,BigNumLength,ZENRing));
extern void _ZemModMultiply ___((BigNum,BigNum,BigNumLength,BigNum,BigNumLength,ZENRing));
extern void _ZemModMultiplyKaratsubaBuffer ___((BigNum,BigNum,BigNumLength,BigNum,BigNumLength,BigNum,ZENRing));
extern void _ZemModSquareKaratsubaBuffer ___((BigNum,BigNum,BigNumLength,BigNum,ZENRing));

extern BigNumLength _ZemEltRedModKaratsuba ___((BigNum, BigNum, ZENRing));
extern void _ZemEltSquare ___((ZENElt *,ZENElt,ZENRing)); /* Squaring of elements */
extern void _ZemEltMultiply ___((ZENElt *,ZENElt,ZENElt,ZENRing)); /* Multiplication of elements */
extern void _ZemEltSquareKaratsuba ___((ZENElt *,ZENElt,ZENRing)); /* Squaring of elements */
extern void _ZemEltMultiplyKaratsuba ___((ZENElt *,ZENElt,ZENElt,ZENRing)); /* Multiplication of elements */
extern int _ZemEltInverse ___((ZENElt *, ZENElt, ZENRing)); /* Inverse of elements */
extern void _ZemEltTrace ___((ZENElt *, ZENElt, ZENRing));	/* Trace of elements */

extern int _ZemEltExpPlain ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZemEltExp ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZemEltExpMod ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZemEltExpMod4 ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZemEltExpMod5 ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZemEltExpMod6 ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */

/* io */
extern void _ZemEltToZ ___((BigNum,BigNumLength*,ZENElt,ZENRing)); /* Evaluate in $p$ an element */
extern void _ZemZToElt ___((ZENElt *,BigNum,BigNumLength,ZENRing));	/* Get the element from an evaluation in $p$ */


/* Procedures on polynomials
 ***************************/

/* Assigning */
extern void _ZemPolySetToXi ___((ZENPoly, int, ZENRing)); /* Set a polynomial to $X^i$ */

/* Testing */
extern int _ZemPolyIsXi ___((ZENPoly, int, ZENRing)); /* Is a polynomial equal to $X^i$ */
extern void _ZemPolyLazyToVeryLazy ___((ZENPoly, ZENPoly, ZENRing));

/* Arithmetic */
extern void _ZemPolyAdd ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZemPolyAddLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZemPolyNegate ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZemPolyNegateLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZemPolySubtract ___((ZENPoly, ZENPoly, ZENRing)); /* Substraction of polynomials */
extern void _ZemPolySubtractLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Substraction of polynomials */

extern int _ZemPolySquarePlain ___((ZENPoly, ZENPoly, ZENRing)); /* Squaring of polynomials */

extern int _ZemPolySquareKaratsuba ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZemPolySquareKaratsubaLazy ___((ZENPoly, ZENPoly, ZENRing));


extern int _ZemPolyMultiply ___((ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Multiplication of polynomials */
extern int _ZemPolyMultiplyLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZemPolyMultiplyPlain ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

extern int _ZemPolyMultiplyKaratsuba ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZemPolyMultiplyKaratsubaLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));


extern void _ZemPolyMultiplyScalar ___((ZENPoly, ZENPoly, ZENElt, ZENRing)); /* Multiplication of polynomials */
extern void _ZemPolyDot  ___((ZENElt *, ZENPoly, ZENPoly, ZENRing)); /* Scalar product of 2 polynomials */
extern int _ZemPolyDivide ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Division of polynomials */
extern int _ZemPolyDivideLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Lazy division of polynomials */
extern int _ZemPolyDivideVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Lazy division of polynomials */
extern int _ZemPolyModuloVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, int, ZENRing)); /* Lazy modulo of polynomials */

/*   
   \TEX{
   \Section{Ring initialization}
   \input{zeminit}
   \input{zemmodulo}

   \Section{Element operations}


   \input{zemelt_assign}
   \input{zemelt_test}
   \input{zemelt_multiply}
   \input{zemelt_exp}
   \input{zemelt_inverse}
   \input{zemelt_trace}
   \input{zemelt_io}

   }
*/

#endif
