/*
 * zen/include/zep.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZEP__
#define __ZEP__

#include "zen.h"

#define ZepRingP1(R) ZENRingP1(R)

#define ZepRingMask(R) ZepRingPmsZemp(R).mask
#define ZepRingIbuf(R) ZepRingPms(R).Ibuf
#define ZepRingNbuf(R) ZepRingPmsZemp(R).Nbuf
#define ZepRingMbuf(R) ZepRingPmsZemp(R).Mbuf
#define ZepRingMbufl(R) ZepRingPmsZemp(R).Mbufl

/* \TEX{
   All the |_ZepFct| functions have corresponding macros |ZepFct|.
   The macros are always at least as fast as the function call and sometimes even faster.
   Therefore, it is best to use macros.
   }
   */
/* On elements
 *************/
#define ZepElt2Bn(a,Rg) ZENElt2Bn(a,Rg)

/* Allocation */
#define ZepEltCreate(a, Rg) ((ZENElt2Bn(a,Rg)=ZBNC(ZENRingSizeElt(Rg)))==NULL ? ZERR : 0)
#define ZepEltCreateLarge(a, Rg) ((ZENElt2Bn(a,Rg)=ZBNC(2*ZENRingSizeElt(Rg)+2))==NULL ? ZERR : 0)
#define ZepEltFree(a, Rg) ZENFree(ZENElt2Bn(a,Rg))

/* Assigning */
#define ZepEltAssign(a,b,Rg) ZBNAssign(ZENElt2Bn(a,Rg),ZENElt2Bn(b,Rg),ZENRingSizeElt(Rg))
#define ZepEltSetToZero(a,Rg) ZBNSetToZero(ZENElt2Bn(a,Rg),ZENRingSizeElt(Rg))
#define ZepEltSetToOne(a,Rg) if(1) {ZBNSetToZero(ZENElt2Bn(a,Rg)+1,ZENRingSizeElt(Rg)-1); ZENElt2Bn(a,Rg)[0]=1;} else
#define ZepEltSetRandom(a,Rg) _ZepEltSetRandom(&(a),Rg)
#define ZepEltSetNext(a,Rg) _ZepEltSetNext(&(a),Rg)

/* Test */
#define ZepEltAreEqual(a,b,Rg) (!ZBNCompare(ZepElt2Bn(a, Rg),ZENRingSizeElt(Rg),ZepElt2Bn(b, Rg),ZENRingSizeElt(Rg)))
#define ZepEltIsZero(a,Rg) ZBNIsZero(ZepElt2Bn(a, Rg),ZENRingSizeElt(Rg))
#define ZepEltIsOne(a,Rg) ((ZBNNumDigits(ZepElt2Bn(a, Rg),ZENRingSizeElt(Rg)) == 1) && (ZepElt2Bn(a, Rg)[0] == 1))

/* Arithmetic */
#define ZepEltAdd(r,a,Rg) ZBNModAdd(ZENElt2Bn(r,Rg),ZENRingSizeElt(Rg),ZENElt2Bn(a,Rg),ZENRingSizeElt(Rg),ZENRingP(Rg),ZENRingPl(Rg))
#define ZepEltAddMisc(r,a,Rg) ZBNAdd(ZENElt2Bn(r,Rg),ZENRingSizeElt(Rg),ZENElt2Bn(a,Rg),ZENRingSizeElt(Rg),(BigNumDigit) 0)
#define ZepEltNegate(c,a,Rg) _ZepEltNegate(&(c),a,Rg)
#define ZepEltSubtract(c,a,Rg) _ZepEltSubtract(&(c),a,Rg)
#define ZepEltSquare(r,a,Rg)   ZENEltSquare(r,a,Rg)
#define ZepEltMultiply(r,a,b,Rg) ZENEltMultiply(r,a,b,Rg)
#define ZepEltInverse(r,a,Rg) _ZepEltInverse(&(r),a,Rg)
#define ZepEltTrace(r,a,Rg) ZepEltAssign(r,a,Rg)
#define ZepEltExp(r,k,kl,a,Rg) ZENEltExp(r,k,kl,a,Rg)

/* io */
#define ZepEltReadFromString(a,s,base,Rg) _ZepEltReadFromString(&(a),s,base,Rg) 
#define ZepEltPrintToString(a,base,Rg) _ZepEltPrintToString(a,base,Rg)
#define ZepEltReadFromFile(a,file,base,Rg) _ZepEltReadFromFile(a,file,base,Rg) 
#define ZepEltPrintToFile(file,a,base,Rg) _ZepEltPrintToFile(file,a,base,Rg)

#define ZepEltGetFromString(a,s,Rg) _ZepEltReadFromString(&(a),s,10,Rg)
#define ZepEltPutToString(a,Rg) _ZepEltPrintToString(a,10,Rg)
#define ZepEltGetFromFile(a,file,Rg) _ZepEltReadFromFile(&(a),file,10,Rg) 
#define ZepEltPutToFile(file,a,Rg) _ZepEltPrintToFile(file,a,10,Rg)

#define ZepEltToZ(p,p_pl,a,R) \
ZBNAssign(p, ZepElt2Bn(a, R), (*(p_pl) = ZBNNumDigits(ZepElt2Bn(a, R), ZENRingSizeElt(R))))

#define ZepZToElt(a,p,pl,R) \
     if (1) { \
	       ZBNAssign(ZepElt2Bn(a, R), p, pl); \
	       ZBNSetToZero(ZepElt2Bn(a, R)+(pl), ZENRingSizeElt(R)-pl); \
	    } else


/* On polynomials
 ****************/
#define ZepPolyDeg(p,Rg) ZENPolyDeg(p,Rg)
#define ZepPolyLgt(p,Rg) ZENPolyLgt(p,Rg)

#define ZepPolyCof(r,Rg) ZENPolyCof(r,Rg)
#define ZepPolyGetCof(r,i,Rg) ZENPolyGetCof(r,i,Rg)

/* Allocation */
#define ZepPolyCreate(d, Rg) _ZepPolyCreate(d, Rg) 
#define ZepPolyCreateLarge(d, Rg) _ZepPolyCreateLarge(d, Rg) 
#define ZepPolyFree(r, Rg) _ZepPolyFree(r, Rg) 
#define ZepPolyCopy(PX,Rg) ZENPolyCopy(PX,Rg)

/* Assigning */
#define ZepPolyAssign(r,PX,Rg) _ZepPolyAssign(r,PX,Rg)
#define ZepPolyAssignLarge(r,PX,Rg) _ZepPolyAssignLarge(r,PX,Rg)
#define ZepPolyLazyToVeryLazy(r,PX,Rg) _ZepPolyLazyToVeryLazy(r,PX,Rg)
#define ZepPolySetToZero(r,Rg) _ZepPolySetToZero(r,Rg)
#define ZepPolySetToZeroLarge(r,Rg) _ZepPolySetToZeroLarge(r,Rg)
#define ZepPolySetToXi(r,i,Rg) _ZepPolySetToXi(r,i,Rg)
#define ZepPolySetRandom(r,d,Rg) _ZepPolySetRandom(r,d,Rg)
#define ZepPolySetCoeff(r,i,b,Rg) ZepEltAssign(ZepPolyGetCof(r, i, Rg), b, Rg)
#define ZepPolyGetCoeff(c, r,i,Rg) ZepEltAssign(c, ZepPolyGetCof(r, i, Rg), Rg)
#define ZepPolyGetCoeffPtr(r,i,Rg) ZepPolyGetCof(r,i,Rg)

#define ZepPolyDigit(r,Rg) (r)->coeff.n
#define ZepPolyGetDigit(r,i,Rg) ZBNDigit(ZepPolyDigit(r, Rg), i)
#define ZepPolySetDigit(r,i,n,Rg) ZepPolyGetDigit(r,i,Rg) = n

#define ZepPolyUpdateDegree(r,Rg) _ZepPolyUpdateDegree(r,Rg)

/* Test */
#define ZepPolyAreEqual(r,PX,Rg) _ZepPolyAreEqual(r,PX,Rg)
#define ZepPolyIsZero(r,Rg) (!(r)->deg && ZepEltIsZero(ZepPolyGetCoeffPtr(r, 0, Rg), Rg))
#define ZepPolyIsXi(p,i,Rg) _ZepPolyIsXi(p.i,Rg)

/* Arithmetic */
#define ZepPolyAdd(r,PX,Rg) _ZepPolyAdd(r,PX,Rg)
#define ZepPolyAddLarge(r,PX,Rg) _ZepPolyAddLarge(r,PX,Rg)
#define ZepPolySubtract(r,PX,Rg) _ZepPolySubtract(r,PX,Rg)
#define ZepPolySubtractLLarge(r,PX,Rg) _ZepPolySubtractLarge(r,PX,Rg)
#define ZepPolyNegate(r,PX,Rg) _ZepPolyNegate(r,PX,Rg)
#define ZepPolyNegateLarge(r,PX,Rg) _ZepPolyNegateLarge(r,PX,Rg)


#define ZepPolySquarePlain(c,r,Rg) _ZepPolySquarePlain(c,r,Rg)
#define ZepPolySquarePlainLazy(c,r,Rg) _ZepPolySquarePlain(c,r,Rg)
#define ZepPolySquarePlainVeryLazy(c,r,Rg) _ZepPolySquarePlainVeryLazy(c,r,Rg)

#define ZepPolySquareKaratsuba(c,r,Rg) _ZepPolySquareKaratsuba(c,r,Rg)
#define ZepPolySquareKaratsubaLazy(c,p,Rg) _ZepPolySquareKaratsubaLazy(c,p,Rg)
#define ZepPolySquareKaratsubaVeryLazy(c,p,Rg) _ZepPolySquareKaratsubaVeryLazy(c,p,Rg)

#define ZepPolyMultiply(c,r,PX,Rg) _ZepPolyMultiply(c,r,PX,Rg)
#define ZepPolyMultiplyLazy(c,r,PX,Rg) _ZepPolyMultiplyLazy(c,r,PX,Rg)
#define ZepPolyMultiplyVeryLazy(c,r,PX,Rg) _ZepPolyMultiplyVeryLazy(c,r,PX,Rg)

#define ZepPolyMultiplyPlain(c,r,PX,Rg) _ZepPolyMultiplyPlain(c,r,PX,Rg)
#define ZepPolyMultiplyPlainLazy(c,r,PX,Rg) _ZepPolyMultiplyPlain(c,r,PX,Rg)
#define ZepPolyMultiplyPlainVeryLazy(c,r,PX,Rg) _ZepPolyMultiplyPlainVeryLazy(c,r,PX,Rg)

#define ZepPolyMultiplyKaratsuba(c,r,PX,Rg) _ZepPolyMultiplyKaratsuba(c,r,PX,Rg)
#define ZepPolyMultiplyKaratsubaLazy(c,r,PX,Rg) _ZepPolyMultiplyKaratsubaLazy(c,r,PX,Rg)
#define ZepPolyMultiplyKaratsubaVeryLazy(c,r,PX,Rg) _ZepPolyMultiplyKaratsubaVeryLazy(c,r,PX,Rg)

#define ZepPolyMultiplyScalar(r,p,e,Rg) _ZepPolyMultiplyScalar(r,p,e,Rg)
#define ZepPolyDot(r,p,q,R) _ZepPolyDot(r,p,q,R)

#define ZepPolyDivide(r,MX,PX,QX,Rg) _ZepPolyDivide(r,MX,PX,QX,Rg)
#define ZepPolyDivideLazy(r,MX,PX,QX,Rg) _ZepPolyDivideLazy(r,MX,PX,QX,Rg)
#define ZepPolyDivideVeryLazy(r,MX,PX,QX,Rg) _ZepPolyDivideVeryLazy(r,MX,PX,QX,Rg)
#define ZepPolyModuloVeryLazy(MX,PX,QX,k,Rg) _ZepPolyModuloVeryLazy(MX,PX,QX,k,Rg)

/* io */
#define ZepPolyReadFromString(r,s,base,Rg) ZENPolyReadFromString(r,s,base,Rg) 
#define ZepPolyPrintToString(r,base,Rg) ZENPolyPrintToString(r,base,Rg)
#define ZepPolyReadFromFile(r,file,base,Rg) ZENPolyReadFromFile(r,file,base,Rg) 
#define ZepPolyPrintToFile(file,r,base,Rg) ZENPolyPrintToFile(file,r,base,Rg)

#define ZepPolyGetFromString(r,s,Rg) ZENPolyReadFromString(r,s,10,Rg)
#define ZepPolyPutToString(r,Rg) ZENPolyPrintToString(r,10,Rg)
#define ZepPolyGetFromFile(r,file,Rg) ZENPolyReadFromFile(r,file,10,Rg) 
#define ZepPolyPutToFile(file,r,Rg) ZENPolyPrintToFile(file,r,10,Rg)


/* On matrices
 ****************/
#define ZepMatNbPtr(M,R) ZENMatNbPtr(M,R)
#define ZepMatSize(M,R) ZENMatSize(M,R)
#define ZepMatNbBn(M,R) ZENMatNbBn(M,R)
#define ZepMatNbBnd(M,R) ZENMatNbBnd(M,R)
#define ZepMatNbDigit(M,R) ZENMatNbDigit(M,R)
#define ZepMatNbBloc(M,R) ZENMatNbBloc(M,R)
#define ZepMatType(M,R) ZENMatType(M,R)
#define ZepMatFirst(M,R) ZENMatFirst(M,R)

#define ZepMatIsRowType(M,R) ZENMatIsRowType(M,R)
#define ZepMatIsPermutation(M,R) ZENMatIsPermutation(M,R)
#define ZepMatAreSameType(A,B,R) ZENMatAreSameType(A,B,R)

#define ZepMatNbRow(M,R) ZENMatNbRow(M,R)
#define ZepMatNbCol(M,R) ZENMatNbCol(M,R)

/* Allocation */
#define ZepMatCreate(l,c,R) ZENMatCreate(l,c,R)
#define ZepMatRowCreate(l,c,R) ZENMatRowCreate(l,c,R)
#define ZepMatColCreate(l,c,R) ZENMatColCreate(l,c,R)
#define ZepPermutationCreate(n,t,R) ZENPermutationCreate(n,t,R)
#define ZepPermutationRowCreate(n,R) ZENPermutationRowCreate(n,R)
#define ZepPermutationColCreate(n,R) ZENPermutationColCreate(n,R)
#define ZepMatFree(M,R) ZENMatFree(M,R)
#define ZepMatCopy(M,R) ZENMatCopy(M,R)

/* Assignation */
#define ZepMatSetRandom(M,R) ZENMatSetRandom(M,R)
#define ZepMatFastPermute(M,l1,l2,R) ZENMatFastPermute(M,l1,l2,R)
#define ZepMatGetCoeffPtr(M,l,c,R) ZENMatGetCoeffPtr(M,l,c,R)  
#define ZepMatGetCoeff(Z,M,l,c,R) ZENMatGetCoeff(Z,M,l,c,R)  
#define ZepMatSetCoeff(M,l,c,Z,R) ZENMatSetCoeff(M,l,c,Z,R)

#define ZepMatPermuteRow(M,l1,l2,R) ZENMatPermuteRow(M,l1,l2,R)
#define ZepMatPermuteCol(M,l1,l2,R) ZENMatPermuteCol(M,l1,l2,R)
#define ZepMatPermuteBn(M,l1,l2,R) ZENMatPermuteBn(M,l1,l2,R)
#define ZepMatPermuteDigit(M,l1,l2,R) ZENMatPermuteDigit(M,l1,l2,R)



/* Arithmetic */
#define ZepMatMultiply(X,A,B,R) ZENMatMultiply(X,A,B,R)
#define ZepMatSquare(A,B,R) ZENMatSquare(A,B,R)
#define ZepMatAdd(A,B,R) ZENMatAdd(A,B,R)
#define ZepMatSameTypeAdd(A,B,R) ZepMatAdd(A,B,R)
#define ZepMatAddScalar(A,b,R) ZENMatAddScalar(A,b,R)
#define ZepMatNegate(A,B,R) ZENMatNegate(A,B,R)
#define ZepMatSubtract(A,B,R) ZENMatSubtract(A,B,R)
#define ZepMatMultiplyPlain(X,A,B,R) ZENMatMultiplyPlain(X,A,B,R)
#define ZepMatMultiplyScalar(A,b,R) ZENMatMultiplyScalar(A,b,R)
#define ZepMatGaussPlain(D,S,P,p_rk,R) ZENMatGaussPlain(D,S,P,p_rk,R)
#define ZepMatGaussRank(D,p_rk,R) ZENMatGaussRank(D,p_rk,R)
#define ZepMatGaussLazy(D,P,p_rk,R) ZENMatGaussLazy(D,P,p_rk,R)
#define ZepMatGaussInverse(D,S,R) ZENMatGaussInverse(D,S,R)
#define ZepMatInverse(I,M,R) ZENMatInverse(I,M,R)
#define ZepMatKernel(K,M,R) ZENMatKernel(K,M,R)

/* io */
#define ZepMatReadFromString(r,s,base,Rg) ZENMatReadFromString(r,s,base,Rg)
#define ZepMatPrintToString(r,base,Rg) ZENMatPrintToString(r,base,Rg)
#define ZepMatReadFromFile(r,file,base,Rg) ZENMatReadFromFile(r,file,base,Rg)
#define ZepMatPrintToFile(file,r,base,Rg) ZENMatPrintToFile(file,r,base,Rg)

/* Following the local Zep functions */

/* Allocation */
#define ZepMatCreateParams(ts,l,c,p_pms,R) _ZepMatCreateParams(ts,l,c,p_pms,R)

/* Assigning */
#define ZepMatAssign(A,B,R) _ZepMatAssign(A,B,R)
#define ZepMatSameTypeAssign(A,B,R) ZBNAssign(ZENMatFirst(A,R),ZENMatFirst(B,R),ZENMatNbBnd(A,R))
#define ZepMatSetToZero(A,R) ZBNSetToZero(ZENMatFirst(A,R),ZENMatNbBnd(A,R))
#define ZepMatSetToOne(M,R) _ZepMatSetToOne(M,R)

#define ZepMatSetBloc(M,l,c,Z,R) _ZedMatSetBloc(M,l,c,Z,R)
#define ZepMatGetBloc(Z,M,l,c,R) _ZedMatGetBloc(Z,M,l,c,R)
#define ZepMatSetDigit(M,l,c,Z,R) ZepMatSetBloc(M,c,l,Z,R)
#define ZepMatGetDigit(Z,M,l,c,R) ZepMatGetBloc(Z,M,c,l,R)
 
/* Test */
#define ZepMatSameTypeAreEqual(A,B,R) (!(ZBNCompare(ZENMatFirst(A,R),ZENMatNbBnd(A,R),ZENMatFirst(B,R),ZENMatNbBnd(B,R))))
#define ZepMatIsZero(A,R) ZBNIsZero(ZENMatFirst(A,R),ZENMatNbBnd(A,R))
#define ZepMatIsOne(M,R) _ZepMatIsOne(M,R)

/* io */
#define ZepMatGetFromString(r,s,Rg) ZENMatGetFromString(r,s,Rg)
#define ZepMatPutToString(r,Rg) ZENMatPutToString(r,Rg)
#define ZepMatGetFromFile(r,file,Rg) ZENMatGetFromFile(r,file,Rg)
#define ZepMatPutToFile(file,r,Rg) ZENMatPutToFile(file,r,Rg)

/* Procedure on Rings
 ***********************/
extern int _ZepPrcInitPolyRootsCanonical ___((ZENRing));
extern int _ZepPrcClosePolyRootsCanonical ___((ZENRing));

extern int _ZepPrcInitEltKara ___((ZENRing));
extern int _ZepPrcCloseEltKara ___((ZENRing));

extern int _ZepPrcInitEltExp ___((ZENRing));
extern int _ZepPrcCloseEltExp ___((ZENRing));

extern int _ZepInit ___((ZENRing));
extern int _ZepClose ___((ZENRing));

/* Procedure on elements
 ***********************/
/* Allocation */
extern int _ZepEltCreate ___((ZENElt *, ZENRing)); /* Create elements */
extern int _ZepEltCreateLarge ___((ZENElt *, ZENRing));
extern void _ZepEltFree ___((ZENElt, ZENRing)); /* Free elements */

/* Assigning */
extern void _ZepEltAssign ___((ZENElt *, ZENElt, ZENRing)); /* Assign an element with another */
extern void _ZepEltSetToZero ___((ZENElt *, ZENRing)); /* Set an element to zero */
extern void _ZepEltSetToOne ___((ZENElt *, ZENRing)); /* Set an element to one */
extern void _ZepEltSetRandom ___((ZENElt *, ZENRing)); /* Set randomly an element */
extern void _ZepEltSetNext ___((ZENElt *, ZENRing)); /* Set an element to the next */
extern int _ZepEltConvert ___((ZENElt *, ZENRing, ZENElt, ZENRing));

/* Test */
extern int _ZepEltAreEqual ___((ZENElt, ZENElt, ZENRing)); /* Is an element equal to another */
extern int _ZepEltIsZero ___((ZENElt, ZENRing)); /* Is an element equal to zero */
extern int _ZepEltIsOne ___((ZENElt, ZENRing)); /* Is an element equal to one */

/* Arithmetic */
extern void _ZepEltAdd ___((ZENElt *, ZENElt, ZENRing));	/* Addition of elements */
extern void _ZepEltAddMisc ___((ZENElt *, ZENElt, ZENRing));	/* Addition of elements */
extern void _ZepEltNegate ___((ZENElt *, ZENElt, ZENRing));	/* Addition of elements */
extern void _ZepEltSubtract ___((ZENElt *, ZENElt, ZENRing));	/* Addition of elements */
extern void _ZepEltSquare ___((ZENElt *, ZENElt, ZENRing)); /* Squaring of elements */
extern void _ZepEltMultiply ___((ZENElt *, ZENElt, ZENElt, ZENRing)); /* Multiplication of elements */
extern void _ZepEltSquareKaratsuba ___((ZENElt *, ZENElt, ZENRing)); /* Squaring of elements */
extern void _ZepEltMultiplyKaratsuba ___((ZENElt *, ZENElt, ZENElt, ZENRing)); /* Multiplication of elements */
extern int _ZepEltInverse ___((ZENElt *, ZENElt, ZENRing)); /* Inverse of elements */
extern int _ZepEltInverseMisc ___((ZENElt *, ZENElt, ZENRing)); /* Inverse of elements */
extern void _ZepEltTrace ___((ZENElt *, ZENElt, ZENRing));	/* Trace of elements */

extern int _ZepEltExpPlain ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZepEltExp ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZepEltExpMod ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZepEltExpMod4 ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZepEltExpMod5 ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
extern int _ZepEltExpMod6 ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));	/* Exponentiation in a finite field */
/* io */
extern int _ZepEltReadFromString ___((ZENElt *, char *, int, ZENRing)); /* Read a polynomial from a string */
extern char* _ZepEltPrintToString ___((ZENElt, int, ZENRing)); /* Print a polynomial to a string */
extern int _ZepEltReadFromFile ___((ZENElt *, FILE *, int, ZENRing));	/* Read a polynomial from a file */
extern int _ZepEltPrintToFile ___((FILE *, ZENElt, int, ZENRing)); /* Print a polynomial to a file */
extern int _ZepEltGetFromString ___((ZENElt *, char *, ZENRing)); /* Get a polynomial from a string */
extern char* _ZepEltPutToString ___((ZENElt, ZENRing)); /* Put a polynomial to a string */
extern int _ZepEltGetFromFile ___((ZENElt *, FILE *, ZENRing)); /* Get a polynomial from a file */
extern int _ZepEltPutToFile ___((FILE *, ZENElt, ZENRing)); /* Put a polynomial to a file */

extern void _ZepEltToZ ___((BigNum,BigNumLength*,ZENElt,ZENRing)); /* Evaluate in $p$ an element */
extern void _ZepZToElt ___((ZENElt *,BigNum,BigNumLength,ZENRing));	/* Get the element from an evaluation in $p$ */

/* Procedures on polynomials
 ***************************/
/* Allocation */
extern ZENPoly _ZepPolyCreate ___((int, ZENRing)); /* Create polynomials */
extern ZENPoly _ZepPolyCreateLarge ___((int, ZENRing)); /* Create polynomials */
extern void _ZepPolyFree ___((ZENPoly, ZENRing)); /* Free polynomials */

/* Assigning */
extern void _ZepPolyAssign ___((ZENPoly, ZENPoly, ZENRing)); /* Assign a polynomial with another */
extern void _ZepPolyAssignLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Assign a polynomial with another */
extern void _ZepPolyLazyToVeryLazy ___((ZENPoly, ZENPoly, ZENRing));
extern void _ZepPolySetToZero ___((ZENPoly, ZENRing)); /* Set a polynomial to zero */
extern void _ZepPolySetToZeroLarge ___((ZENPoly, ZENRing)); /* Set a polynomial to zero */
extern void _ZepPolySetToXi ___((ZENPoly, int, ZENRing)); /* Set a polynomial to $X^i$ */
extern void _ZepPolySetRandom ___((ZENPoly, int, ZENRing)); /* Set randomly a polynomial */
extern void _ZepPolySetCoeff ___((ZENPoly, int, ZENElt, ZENRing)); /* Set a coefficient of a polynomial */
extern void _ZepPolyGetCoeff ___((ZENElt *, ZENPoly, int, ZENRing)); /* Get a coefficient of a polynomial */
extern ZENElt _ZepPolyGetCoeffPtr ___((ZENPoly, int, ZENRing)); /* Get a pointer on a coefficient of a polynomial */
extern void _ZepPolyUpdateDegree ___((ZENPoly, ZENRing)); /* Update the degree of a polynomial */

/* Test */
extern int _ZepPolyAreEqual ___((ZENPoly, ZENPoly, ZENRing)); /* Is a polynomial equal to another */
extern int _ZepPolyIsZero ___((ZENPoly, ZENRing)); /* Is a polynomial equal to zero */
extern int _ZepPolyIsXi ___((ZENPoly, int, ZENRing)); /* Is a polynomial equal to $X^i$ */

/* Arithmetic */
extern void _ZepPolyAdd ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZepPolyAddLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZepPolyNegate ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZepPolyNegateLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZepPolySubtract ___((ZENPoly, ZENPoly, ZENRing)); /* Substraction of polynomials */
extern void _ZepPolySubtractLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Substraction of polynomials */

extern int _ZepPolySquarePlain ___((ZENPoly, ZENPoly, ZENRing)); /* Squaring of polynomials */
extern int _ZepPolySquarePlainVeryLazy ___((ZENPoly, ZENPoly, ZENRing));

extern int _ZepPolySquareKaratsuba ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZepPolySquareKaratsubaLazy ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZepPolySquareKaratsubaVeryLazy ___((ZENPoly, ZENPoly, ZENRing));


extern int _ZepPolyMultiply ___((ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Multiplication of polynomials */
extern int _ZepPolyMultiplyLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZepPolyMultiplyVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

extern int _ZepPolyMultiplyPlain ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZepPolyMultiplyPlainVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

extern int _ZepPolyMultiplyKaratsuba ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZepPolyMultiplyKaratsubaLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZepPolyMultiplyKaratsubaVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

extern void _ZepPolyMultiplyScalar ___((ZENPoly, ZENPoly, ZENElt, ZENRing)); /* Multiplication of polynomials */
extern void _ZepPolyDot  ___((ZENElt *, ZENPoly, ZENPoly, ZENRing)); /* Scalar product of 2 polynomials */
extern int _ZepPolyDivide ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Division of polynomials */
extern int _ZepPolyDivideLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Lazy division of polynomials */
extern int _ZepPolyDivideVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Lazy division of polynomials */
extern int _ZepPolyModuloVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, int, ZENRing)); /* Lazy modulo of polynomials */


/* Procedures on matrices
 ***************************/
/* Allocation */
extern int _ZepMatCreateParams ___((ZENMatTypeStruct,Dim,Dim,ZENMat *,ZENRing));

/* Assignment */
extern int _ZepMatAssign ___((ZENMat,ZENMat,ZENRing));
extern int _ZepMatSetToZero ___((ZENMat,ZENRing));
extern int _ZepMatSetToOne ___((ZENMat,ZENRing));

/*Tests */
extern int _ZepMatIsZero ___((ZENMat,ZENRing));
extern int _ZepMatIsOne ___((ZENMat,ZENRing));
extern int _ZepMatAreEqual ___((ZENMat,ZENMat,ZENRing));

/*   
   \TEX{
   \Section{Ring initialization}
   \input{zepinit}

   \Section{Element operations}
   \input{zepelt_alloc}
   \input{zepelt_assign}
   \input{zepelt_test}
   \input{zepelt_add}
   \input{zepelt_multiply}
   \input{zepelt_exp}
   \input{zepelt_inverse}
   \input{zepelt_trace}
   \input{zepelt_io}

   \Section{Polynomials operations}
   \input{zeppoly_alloc}
   \input{zeppoly_assign}
   \input{zeppoly_test}
   \input{zeppoly_add}
   \input{zeppoly_multiply}
   }
*/

#endif
