/*
 * zen/include/zeps.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZEPS__
#define __ZEPS__
#include "zen.h"

#define ZepsRingPms(R) ZENRingPms(R).zeps
#define ZepsRingEbuf(Rg) ZepsRingPms(Rg).Ebuf
/* #define ZepsRingTab(Rg) ZepsRingPms(Rg).Tbuf */

#define ZepsMultTable0Ptr(R)  ZepsRingPms(R).tab0
#define ZepsMultTable0(R,i)  ((*ZepsMultTable0Ptr(R))[i])

#define ZepsMultTable1Ptr(R)  ZepsRingPms(R).tab1
#define ZepsMultTable1(R,i)  ((*ZepsMultTable1Ptr(R))[i])

#define ZepsMultTable2Ptr(R)  ZepsRingPms(R).tab2
#define ZepsMultTable2(R,i)  ((*ZepsMultTable2Ptr(R))[i])

#define ZepsMultTable3Ptr(R)  ZepsRingPms(R).tab3
#define ZepsMultTable3(R,i)  ((*ZepsMultTable3Ptr(R))[i])

/* \TEX{
   All the |_ZepsFct| functions have corresponding macros |ZepsFct|.
   The macros are always at least as fast as the function call and sometimes even faster.
   Therefore, it is best to use macros.
   }
   */

#define ZepsElt2Int(a, Rg) ZENElt2Int(a, Rg)

/* On elements
 *************/

/* Allocation */
#define ZepsEltCreate(a, Rg) 0
#define ZepsEltFree(a, Rg)

/* Assigning */
#define ZepsEltAssign(a,b,Rg)  (ZepsElt2Int(a, Rg)) = (ZepsElt2Int(b, Rg))
#define ZepsEltSetToZero(a,Rg) (ZepsElt2Int(a, Rg)) = 0
#define ZepsEltSetToOne(a,Rg) (ZepsElt2Int(a, Rg)) = 1
#define ZepsEltSetRandom(a,Rg) (ZepsElt2Int(a, Rg)) = zbnrandom()% ZBNDigit(ZENRingP(Rg), 0)
#define ZepsEltSetNext(a,Rg) ZepsElt2Int(a, Rg) = ((++ZepsElt2Int(a, Rg)) == ZBNDigit(ZENRingQ(Rg), 0) ? 0 : ZepsElt2Int(a, Rg))

/* Test */
#define ZepsEltAreEqual(a,b,Rg) ((ZepsElt2Int(a, Rg)) == (ZepsElt2Int(b, Rg)))
#define ZepsEltIsZero(a,Rg) ((ZepsElt2Int(a, Rg)) == 0)
#define ZepsEltIsOne(a,Rg) ((ZepsElt2Int(a, Rg)) == 1)

/* Arithmetic */
#define ZepsEltAdd(r,a,Rg) (\
(ZepsElt2Int(ZepsRingEbuf(Rg), Rg) = ZepsElt2Int(r, Rg)+ZepsElt2Int(a, Rg), \
ZepsElt2Int(r, Rg) = \
((ZepsElt2Int(ZepsRingEbuf(Rg), Rg) >= ZBNDigit(ZENRingP(Rg), 0)) || \
((ZepsElt2Int(ZepsRingEbuf(Rg), Rg) < (ZepsElt2Int(a, Rg))))) ?  \
ZepsElt2Int(ZepsRingEbuf(Rg), Rg)-ZBNDigit(ZENRingP(Rg), 0) : \
ZepsElt2Int(ZepsRingEbuf(Rg), Rg)) \
)
#define ZepsEltNegate(r,a,Rg) ZepsElt2Int(r, Rg) = (ZepsElt2Int(a, Rg) ? (ZBNDigit(ZENRingP(Rg), 0)-ZepsElt2Int(a, Rg)):0)
#define ZepsEltSubtract(r,a,Rg) ZepsElt2Int(r, Rg) = \
(ZepsElt2Int(r, Rg) >= ZepsElt2Int(a, Rg) ? \
ZepsElt2Int(r, Rg) - ZepsElt2Int(a, Rg) : \
(ZBNDigit(ZENRingP(Rg), 0)-ZepsElt2Int(a, Rg))+ZepsElt2Int(r, Rg))
#define ZepsEltSquare(r,a,Rg) ZENEltSquare(r,a,Rg)
#define ZepsEltMultiply(r,a,b,Rg) ZENEltMultiply(r,a,b,Rg)
#define ZepsEltInverse(r,a,Rg) _ZepsEltInverse(&(r),a,Rg)
#define ZepsEltTrace(r,a,Rg) ZepsEltAssign(r,a,Rg)
#define ZepsEltExp(r,k,kl,a,R) ZENEltExp(r,k,kl,a,R) 

/* io */
#define ZepsEltReadFromString(a,s,base,Rg) _ZepsEltReadFromString(&(a),s,base,Rg) 
#define ZepsEltPrintToString(a,base,Rg) ZBNPrintToString(&ZepsElt2Int(a, Rg), 1, base)
#define ZepsEltReadFromFile(a,file,base,Rg) _ZepsEltReadFromFile(&(a),file,base,Rg)
#define ZepsEltPrintToFile(file,a,base,Rg) ZBNPrintToFile(file, &ZepsElt2Int(a, Rg), 1, base)

#define ZepsEltGetFromString(a,s,Rg) _ZepsEltGetFromString(&(a),s,Rg) 
#define ZepsEltPutToString(a,Rg) _ZepsEltPutToString(a,Rg) 
#define ZepsEltGetFromFile(a,file,Rg) _ZepsEltGetFromFile(&(a),file,Rg)
#define ZepsEltPutToFile(file,a,Rg) _ZepsEltPutToFile(file,a,Rg)

#define ZepsEltToZ(p,p_pl,a,R) (ZBNDigit(p, 0) = ZepsElt2Int(a, Rg), *(p_pl) = 1)
#define ZepsZToElt(a,p,pl,R) (ZepsElt2Int(a, Rg) = ZBNDigit(p, 0))

/* On polynomials
 ****************/
#define ZepsPolyDeg(p,Rg) ZENPolyDeg(p,Rg)
#define ZepsPolyLgt(p,Rg) ZENPolyLgt(p,Rg)

#define ZepsPolyCof(r,Rg) ZENPolyCof(r,Rg)
#define ZepsPolyGetCof(r,i,Rg) ZENPolyGetCof(r,i,Rg)

/* Allocation */
#define ZepsPolyCreate(d, Rg) _ZedsPolyCreate(d, Rg) 
#define ZepsPolyCreateLarge(d, Rg) _ZedsPolyCreate(d, Rg) 
#define ZepsPolyFree(r, Rg) _ZedsPolyFree(r, Rg) 
#define ZepsPolyCopy(PX,Rg) ZENPolyCopy(PX,Rg)

/* Assigning */
#define ZepsPolyAssign(r,PX,Rg) _ZepsPolyAssign(r,PX,Rg)
#define ZepsPolySetToZero(r,Rg) _ZepsPolySetToZero(r,Rg)
#define ZepsPolySetToXi(r,i,Rg) _ZepsPolySetToXi(r,i,Rg)
#define ZepsPolySetRandom(r,d,Rg) _ZepsPolySetRandom(r,d,Rg)
#define ZepsPolySetCoeff(r,i,b,Rg) ZepsEltAssign(ZepsPolyGetCof(r, i, Rg), b, Rg)
#define ZepsPolyGetCoeff(c, r,i,Rg) ZepsEltAssign(c, ZepsPolyGetCof(r, i, Rg), Rg)
#define ZepsPolyGetCoeffPtr(r,i,Rg) ZepsPolyGetCof(r,i,Rg)

#define ZepsPolyDigit(r,Rg) (r)->coeff.n
#define ZepsPolyGetDigit(r,i,Rg) ZBNDigit(ZepsPolyDigit(r, Rg), i)
#define ZepsPolySetDigit(r,i,n,Rg) ZepsPolyGetDigit(r,i,Rg) = n

#define ZepsPolyUpdateDegree(r,Rg) _ZepsPolyUpdateDegree(r,Rg)

/* Test */
#define ZepsPolyAreEqual(r,PX,Rg) _ZepsPolyAreEqual(r,PX,Rg)
#define ZepsPolyIsZero(r,Rg) (!(r)->deg && ZepsEltIsZero(ZepsPolyGetCoeffPtr(r, 0, Rg), Rg))
#define ZepsPolyIsXi(p,i,Rg) _ZepsPolyIsXi(p.i,Rg)

/* Arithmetic */
#define ZepsPolyAdd(r,PX,Rg) _ZepsPolyAdd(r,PX,Rg)
#define ZepsPolySubtract(r,PX,Rg) _ZepsPolySubtract(r,PX,Rg)
#define ZepsPolyNegate(r,PX,Rg) _ZepsPolyNegate(r,PX,Rg)

#define ZepsPolySquarePlain(c,r,Rg) _ZepsPolySquarePlain(c,r,Rg)
#define ZepsPolySquareKaratsuba(c,r,Rg) _ZepsPolySquareKaratsuba(c,r,Rg)
#define ZepsPolySquareKaratsubaLazy(c,p,Rg) _ZepsPolySquareKaratsubaLazy(c,p,Rg)

#define ZepsPolyMultiply(c,r,PX,Rg) _ZepsPolyMultiply(c,r,PX,Rg)
#define ZepsPolyMultiplyLazy(c,r,PX,Rg) _ZepsPolyMultiplyLazy(c,r,PX,Rg)
#define ZepsPolyMultiplyPlain(c,r,PX,Rg) _ZepsPolyMultiplyPlain(c,r,PX,Rg)
#define ZepsPolyMultiplyKaratsuba(c,r,PX,Rg) _ZepsPolyMultiplyKaratsuba(c,r,PX,Rg)
#define ZepsPolyMultiplyKaratsubaLazy(c,r,PX,Rg) _ZepsPolyMultiplyKaratsubaLazy(c,r,PX,Rg)

#define ZepsPolyMultiplyScalar(r,p,e,Rg) _ZepsPolyMultiplyScalar(r,p,e,Rg)
#define ZepsPolyDot(r,p,q,R) _ZepsPolyDot(r,p,q,R)

#define ZepsPolyDivide(r,MX,PX,QX,Rg) _ZepsPolyDivide(r,MX,PX,QX,Rg)
#define ZepsPolyDivideLazy(r,MX,PX,QX,Rg) _ZepsPolyDivideLazy(r,MX,PX,QX,Rg)
#define ZepsPolyModuloLazy(MX,PX,QX,k,Rg) _ZepsPolyModuloLazy(MX,PX,QX,k,Rg)

/* io */
#define ZepsPolyReadFromString(r,s,base,Rg) ZENPolyReadFromString(r,s,base,Rg) 
#define ZepsPolyPrintToString(r,base,Rg) ZENPolyPrintToString(r,base,Rg)
#define ZepsPolyReadFromFile(r,file,base,Rg) ZENPolyReadFromFile(r,file,base,Rg) 
#define ZepsPolyPrintToFile(file,r,base,Rg) ZENPolyPrintToFile(file,r,base,Rg)

#define ZepsPolyGetFromString(r,s,Rg) ZENPolyReadFromString(r,s,10,Rg)
#define ZepsPolyPutToString(r,Rg) ZENPolyPrintToString(r,10,Rg)
#define ZepsPolyGetFromFile(r,file,Rg) ZENPolyReadFromFile(r,file,10,Rg) 
#define ZepsPolyPutToFile(file,r,Rg) ZENPolyPrintToFile(file,r,10,Rg)



/* Rings
 *******/
extern int _ZepsInit ___((ZENRing));
extern int _ZepsClose ___((ZENRing));

extern int _ZepsPrcInitEltMultiply ___((ZENRing));
extern int _ZepsPrcCloseEltMultiply ___((ZENRing));

/* Procedure on elements
 ***********************/
/* Allocation */
extern int _ZepsEltCreate ___((ZENElt *, struct zenring *)); /* Create elements */
extern void _ZepsEltFree ___((ZENElt, struct zenring *)); /* Free elements */

/* Assigning */
extern void _ZepsEltAssign ___((ZENElt *, ZENElt, struct zenring *)); /* Assign an element with another */
extern void _ZepsEltSetToZero ___((ZENElt *, struct zenring *)); /* Set an element to zero */
extern void _ZepsEltSetToOne ___((ZENElt *, struct zenring *)); /* Set an element to one */
extern void _ZepsEltSetRandom ___((ZENElt *, struct zenring *)); /* Set randomly an element */
extern void _ZepsEltSetNext ___((ZENElt *, struct zenring *)); /* Set an element to the next */

/* Test */
extern int _ZepsEltAreEqual ___((ZENElt, ZENElt, struct zenring *)); /* Is an element equal to another */
extern int _ZepsEltIsZero ___((ZENElt, struct zenring *)); /* Is an element equal to zero */
extern int _ZepsEltIsOne ___((ZENElt, struct zenring *)); /* Is an element equal to one */

/* Arithmetic */
extern void _ZepsEltAdd ___((ZENElt *, ZENElt, struct zenring *));	/* Addition of elements */
extern void _ZepsEltNegate ___((ZENElt *, ZENElt, struct zenring *));	/* Negation of elements */
extern void _ZepsEltSubtract ___((ZENElt *, ZENElt, struct zenring *));	/* Subtraction of elements */
extern void _ZepsEltSquareSmallp ___((ZENElt *, ZENElt, struct zenring *)); /* Squaring of elements for small modulos */
extern void _ZepsEltSquareLargep ___((ZENElt *, ZENElt, struct zenring *)); /* Squaring of elements for large modulos */
extern void _ZepsEltMultiplySmallp ___((ZENElt *, ZENElt, ZENElt, struct zenring *)); /* Multiplication of elements for small modulos */
extern void _ZepsEltMultiplyLargep ___((ZENElt *, ZENElt, ZENElt, struct zenring *)); /* Multiplication of elements for large modulos*/
extern int _ZepsEltInverse ___((ZENElt *, ZENElt, struct zenring *)); /* Inverse of elements */
extern void _ZepsEltTrace ___((ZENElt *, ZENElt, struct zenring *));	/* Trace of elements */

/* io */
extern int _ZepsEltReadFromString ___((ZENElt *, char *, int, struct zenring *)); /* Read a polynomial from a string */
extern char* _ZepsEltPrintToString ___((ZENElt, int, struct zenring *)); /* Print a polynomial to a string */
extern int _ZepsEltReadFromFile ___((ZENElt *, FILE *, int, struct zenring *));	/* Read a polynomial from a file */
extern int _ZepsEltPrintToFile ___((FILE *, ZENElt, int, struct zenring *)); /* Print a polynomial to a file */
extern int _ZepsEltGetFromString ___((ZENElt *, char *, struct zenring *)); /* Get a polynomial from a string */
extern char* _ZepsEltPutToString ___((ZENElt, struct zenring *)); /* Put a polynomial to a string */
extern int _ZepsEltGetFromFile ___((ZENElt *, FILE *, struct zenring *)); /* Get a polynomial from a file */
extern int _ZepsEltPutToFile ___((FILE *, ZENElt, struct zenring *)); /* Put a polynomial to a file */

extern void _ZepsEltToZ ___((BigNum,BigNumLength*,ZENElt,ZENRing)); /* Evaluate in $p$ an element */
extern void _ZepsZToElt ___((ZENElt *,BigNum,BigNumLength,ZENRing));	/* Get the element from an evaluation in $p$ */
extern int _ZepsEltConvert ___((ZENElt *, ZENRing, ZENElt, ZENRing));

extern int _ZepsEP ___((ZENElt, ZENRing));
extern int _ZepsER ___((ZENElt *, ZENRing));

/* Procedures on polynomials
 ***************************/
/* Allocation */
extern ZENPoly _ZepsPolyCreate ___((int, struct zenring *)); /* Create polynomials */
extern void _ZepsPolyFree ___((ZENPoly, struct zenring *)); /* Free polynomials */

/* Assigning */
extern void _ZepsPolyAssign ___((ZENPoly, ZENPoly, struct zenring *)); /* Assign a polynomial with another */
extern void _ZepsPolySetToZero ___((ZENPoly, struct zenring *)); /* Set a polynomial to zero */
extern void _ZepsPolySetToXi ___((ZENPoly, int, struct zenring *)); /* Set a polynomial to $X^i$ */
extern void _ZepsPolySetRandom ___((ZENPoly, int, struct zenring *)); /* Set randomly a polynomial */
extern void _ZepsPolySetCoeff ___((ZENPoly, int, ZENElt, struct zenring *)); /* Set a coefficient of a polynomial */
extern void _ZepsPolyGetCoeff ___((ZENElt *, ZENPoly, int, struct zenring *)); /* Get a coefficient of a polynomial */
extern ZENElt _ZepsPolyGetCoeffPtr ___((ZENPoly, int, struct zenring *)); /* Get a pointer on a coefficient of a polynomial */
extern void _ZepsPolyUpdateDegree ___((ZENPoly, struct zenring *)); /* Update the degree of a polynomial */

/* Test */
extern int _ZepsPolyAreEqual ___((ZENPoly, ZENPoly, struct zenring *)); /* Is a polynomial equal to another */
extern int _ZepsPolyIsZero ___((ZENPoly, struct zenring *)); /* Is a polynomial equal to zero */
extern int _ZepsPolyIsXi ___((ZENPoly, int, struct zenring *)); /* Is a polynomial equal to $X^i$ */

/* Arithmetic */
extern void _ZepsPolyAdd ___((ZENPoly, ZENPoly, struct zenring *)); /* Addition of polynomials */
extern void _ZepsPolyNegate ___((ZENPoly, ZENPoly, struct zenring *)); /* Addition of polynomials */
extern void _ZepsPolySubtract ___((ZENPoly, ZENPoly, struct zenring *)); /* Substraction of polynomials */

extern int _ZepsPolySquarePlain ___((ZENPoly, ZENPoly, ZENRing)); /* Squaring of polynomials */
extern int _ZepsPolySquareKaratsuba ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZepsPolySquareKaratsubaLazy ___((ZENPoly, ZENPoly, ZENRing));


extern int _ZepsPolyMultiply ___((ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Multiplication of polynomials */
extern int _ZepsPolyMultiplyLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));  /* Multiplication of polynomials */
extern int _ZepsPolyMultiplyPlain ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZepsPolyMultiplyKaratsuba ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZepsPolyMultiplyKaratsubaLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

extern void _ZepsPolyMultiplyScalar ___((ZENPoly, ZENPoly, ZENElt, struct zenring *)); /* Multiplication of polynomials */
extern void _ZepsPolyDot  ___((ZENElt *, ZENPoly, ZENPoly, struct zenring *)); /* Scalar product of 2 polynomials */
extern int _ZepsPolyDivide ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, struct zenring *)); /* Division of polynomials */
extern int _ZepsPolyDivideLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, struct zenring *)); /* Lazy division of polynomials */
extern int _ZepsPolyModuloLazy ___((ZENPoly, ZENPoly, ZENPoly, int, struct zenring *)); /* Lazy modulo of polynomials */

/* Procedures on series
 ***************************/
/* Allocation */
extern ZENSr _ZepsSrCreate ___((int, struct zenring *)); /* Create series */
extern void _ZepsSrFree ___((ZENSr, struct zenring *)); /* Free series */

/* Assigning */
extern void _ZepsSrAssign ___((ZENSr, ZENSr, struct zenring *)); /* Assign a valuation with another */
extern void _ZepsSrSetToZero ___((ZENSr,struct zenring *)); /* Set a valuation to zero */
extern void _ZepsSrSetRandom ___((ZENSr,struct zenring *)); /* Set randomly a valuation */
extern void _ZepsSrSetCoeff ___((ZENSr, int, ZENElt, struct zenring *)); /* Set a coefficient of a valuation */
extern void _ZepsSrGetCoeff ___((ZENElt, ZENSr, int, struct zenring *)); /* Get a coefficient of a valuation */
extern ZENElt _ZepsSrGetCoeffPtr ___((ZENSr, int, struct zenring *)); /* Get a pointer on a coefficient of a valuation */
extern void _ZepsSrUpdateValuation ___((ZENSr, struct zenring *)); /* Update the valuation of a valuation */

/* Test */
extern int _ZepsSrAreEqual ___((ZENSr, ZENSr, struct zenring *)); /* Is a valuation equal to another */
extern int _ZepsSrIsZero ___((ZENSr, struct zenring *)); /* Is a valuation equal to zero */

/* Arithmetic */
extern void _ZepsSrAdd ___((ZENSr, ZENSr, struct zenring *)); /* Addition of series */
extern void _ZepsSrNegate ___((ZENSr, ZENSr, struct zenring *)); /* Negation of series */
extern void _ZepsSrSubtract ___((ZENSr, ZENSr, struct zenring *)); /* Substraction of series */
extern int _ZepsSrSquareChar2 ___((ZENSr, ZENSr, struct zenring *)); /* Squaring of series in characteristic 2*/
extern int _ZepsSrSquareCharp ___((ZENSr, ZENSr, struct zenring *)); /* Squaring of series in characteristic p*/
extern int _ZepsSrMultiply ___((ZENSr, ZENSr, ZENSr, struct zenring *)); /* Multiplication of series */
extern void _ZepsSrMultiplyScalar ___((ZENSr, ZENElt, ZENSr, struct zenring *)); /* Multiplication of series by a scalar */
extern int _ZepsSrDivide ___((ZENSr, ZENSr, struct zenring *)); /* Division of series */

/* Procedures on matrix
 **********************/
/* Allocation */
extern ZENMat _ZepsMatCreate ___((ZENMatTypeStruct,Dim,Dim,ZENMat *,struct zenring *)); /* Create a matrix */
extern void _ZepsMatFree ___((struct zenmat *, struct zenring *)); /* Free a matrix  */

/* Assigning */
extern void _ZepsMatAssign ___((struct zenmat *, struct zenmat *, struct zenring *)); /* Assign a matrix with another */
extern void _ZepsMatSetToZero ___((ZENMat, struct zenring *)); /* Set a matrix to zero */
extern void _ZepsMatSetToOne ___((struct zenmat *, struct zenring *)); /* Set a matrix to identity */
extern void _ZepsMatSetRandom ___((struct zenmat *, struct zenring *)); /* Set randomly a matrix */
extern void _ZepsMatSetCoeff ___((struct zenmat *, Dim, Dim, ZENElt, struct zenring *)); /* Set a coefficient of a matrix */
extern ZENElt _ZepsMatGetCoeff ___((ZENElt *, struct zenmat *, Dim, Dim, struct zenring *)); /* Get a coefficient of a matrix */

/* Test */
extern int _ZepsMatAreEqual ___((ZENMat, ZENMat, struct zenring *)); /* Is a matrix equal to another */
extern int _ZepsMatIsZero ___((ZENMat, struct zenring *)); /* Is a matrix equal to zero */
extern int _ZepsMatIsOne ___((struct zenmat *, struct zenring *)); /* Is a matrix equal to one */

/* Arithmetic */
extern void _ZepsMatAdd ___((ZENMat, ZENMat, struct zenring *));	/* Addition of Matrix */
extern void _ZepsMatSubtract ___((ZENMat, ZENMat, struct zenring *)); /* Substraction of Matrix */
extern void _ZepsMatMultiply ___((ZENMat, ZENElt, struct zenring *)); /* Multiplication of Matrix */
extern int _ZepsMatInverse ___((ZENMat, ZENMat, struct zenring *)); /* Inverse of a Matrix */
extern int _ZepsMatKernel ___((ZENMat *, ZENMat, struct zenring *)); /* Kernel of a Matrix */

extern int _ZepmMat2Vect ___((ZENMat, ZENMat, ZENRing));
extern int _ZepmVect2Mat ___((ZENMat, ZENMat, ZENRing));

/* 
   \TEX{
   \Section{Ring initialization}
   \input{zepsinit}

   \Section{Element operations}
   \input{zepselt_alloc.tex}
   \input{zepselt_assign.tex}
   \input{zepselt_test.tex}
   \input{zepselt_add.tex}
   \input{zepselt_multiply.tex}
   \input{zepselt_inverse.tex}
   \input{zepselt_trace.tex}
   \input{zepselt_io.tex}

   \Section{Polynomial operations}
   \input{zepspoly_assign.tex}
   \input{zepspoly_test.tex}
   \input{zepspoly_add.tex}
   \input{zepspoly_multiply.tex}

   \Section{Matricial operations}
   \input{zepsmat_convert.tex}
   }
   */

#endif
