/*
 * zen/include/zetab.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */


#ifndef __ZETAB__
#define __ZETAB__

#include "zen.h"

/* \TEX{
   In zetab.h
   }
   */

#define ZetabElt2Int(a, Rg) ZENElt2Int(a, Rg)
#define ZetabRingPms(R) ZENRingPms(R).zetab

#define ZetabRingDef(Rg) ZENRingDef(Rg)
#define ZetabRingOrigin(R) ZENRingOrigin(R) 
#define ZetabRingCln(R) ZENRingCln(R) 

/* The tables */
#define ZetabAddTablePtr(R)  (ZetabRingPms(R).add)
#define ZetabAddTable(R,i,j)  ((*ZetabAddTablePtr(R))[i][j])

#define ZetabNegTablePtr(R)  (ZetabRingPms(R).neg)
#define ZetabNegTable(R,i)  ((*ZetabNegTablePtr(R))[i])

#define ZetabMultTablePtr(R)  (ZetabRingPms(R).mult)
#define ZetabMultTable(R,i,j)  ((*ZetabMultTablePtr(R))[i][j])

#define ZetabInvTablePtr(R)  (ZetabRingPms(R).inv)
#define ZetabInvTable(R,i)  ((*ZetabInvTablePtr(R))[i])

#define ZetabTrTablePtr(R)  (ZetabRingPms(R).tr)
#define ZetabTrTable(R,i)  ((*ZetabTrTablePtr(R))[i])

#define ZetabConvTablePtr(R)  (ZetabRingPms(R).conv)
#define ZetabConvTable(R,i)  ((ZetabConvTablePtr(R))[i])

/* On elements
 *************/
/* Allocation */
#define ZetabEltCreate(a, Rg) 0
#define ZetabEltFree(a, Rg)

/* Assigning */
#define ZetabAnyEltAssign(a,b,Rg)  ((a) = (b))
#define ZetabAnyEltSetToZero(a,Rg) ((a) = 0)
#define ZetabAnyEltSetToOne(a,Rg) ((a) = 1)
#define ZetabAnyEltSetRandom(a,Rg) ((a) = zbnrandom()% ZBNDigit(ZENRingQ(Rg), 0))
#define ZetabAnyEltSetNext(a,Rg) ((a) = ((++(a) == ZBNDigit(ZENRingQ(Rg), 0) ? 0 : (a))))

#define ZetabEltAssign(a,b,Rg) ZetabAnyEltAssign(ZetabElt2Int(a, Rg), ZetabElt2Int(b, Rg),Rg)
#define ZetabEltSetToZero(a,Rg) ZetabAnyEltSetToZero(ZetabElt2Int(a, Rg),Rg)
#define ZetabEltSetToOne(a,Rg) ZetabAnyEltSetToOne(ZetabElt2Int(a, Rg),Rg)
#define ZetabEltSetRandom(a,Rg) ZetabAnyEltSetRandom(ZetabElt2Int(a, Rg),Rg)
#define ZetabEltSetNext(a,Rg) ZetabAnyEltSetNext(ZetabElt2Int(a, Rg),Rg)

/* Test */
#define ZetabAnyEltAreEqual(a,b,Rg) ((a) == (b))
#define ZetabAnyEltIsZero(a,Rg) ((a) == 0)
#define ZetabAnyEltIsOne(a,Rg) ((a) == 1)

#define ZetabEltAreEqual(a,b,Rg) ZetabAnyEltAreEqual(ZetabElt2Int(a, Rg),ZetabElt2Int(b, Rg),Rg)
#define ZetabEltIsZero(a,Rg) ZetabAnyEltIsZero(ZetabElt2Int(a, Rg),Rg)
#define ZetabEltIsOne(a,Rg) ZetabAnyEltIsOne(ZetabElt2Int(a, Rg),Rg)

/* Arithmetic */
#define ZetabAnyEltAdd(r,a, Rg) ((r)=ZetabAddTable(Rg, (r), (a)))
#define ZetabAnyEltNegate(r,a, Rg) ((r)=ZetabNegTable(Rg, (a)))
#define ZetabAnyEltSubtract(r,a, Rg) ((r)=ZetabAddTable(Rg, (r), ZetabNegTable(Rg, (a))))
#define ZetabAnyEltSquare(r,a, Rg) ((r)=ZetabMultTable(Rg, (a),  (a)))
#define ZetabAnyEltMultiply(r,a,b, Rg) ((r)=ZetabMultTable(Rg, (a), (b)))

#define ZetabAnyEltInverse(r,a, Rg) (((r)=ZetabInvTable(Rg, (a)))?ZEN_HAS_INVERSE:__ZetabEltInverse(&(r),(BigNumDigit)(a), Rg))

#define ZetabAnyEltTrace(r,a, Rg) ((r)=ZetabTrTable(Rg,(a)))

#define ZetabEltAdd(r,a,Rg) ZetabAnyEltAdd(ZetabElt2Int(r, Rg),ZetabElt2Int(a, Rg),Rg)
#define ZetabEltNegate(r,a,Rg) ZetabAnyEltNegate(ZetabElt2Int(r, Rg),ZetabElt2Int(a, Rg),Rg)
#define ZetabEltSubtract(r,a,Rg) ZetabAnyEltSubtract(ZetabElt2Int(r, Rg),ZetabElt2Int(a, Rg),Rg)
#define ZetabEltSquare(r,a,Rg) ZetabAnyEltSquare(ZetabElt2Int(r, Rg),ZetabElt2Int(a, Rg),Rg)
#define ZetabEltMultiply(r,a,b,Rg) ZetabAnyEltMultiply(ZetabElt2Int(r, Rg),ZetabElt2Int(a, Rg),ZetabElt2Int(b, Rg),Rg)

#define ZetabEltInverse(r,a,Rg) ZetabAnyEltInverse(ZetabElt2Int(r, Rg),ZetabElt2Int(a, Rg),Rg)

#define ZetabEltTrace(r,a,Rg) ZetabAnyEltTrace(ZetabElt2Int(r, Rg),ZetabElt2Int(a, Rg),Rg)

/* io */
#define ZetabEltReadFromString(a,s,base,Rg) _ZetabEltReadFromString(&(a),s,base,Rg) 
#define ZetabEltPrintToString(a,base,Rg) _ZetabEltPrintToString(a,base,Rg) 
#define ZetabEltReadFromFile(a,file,base,Rg) _ZetabEltReadFromFile(&(a),file,base,Rg)
#define ZetabEltPrintToFile(file,a,base,Rg) _ZetabEltPrintToFile(file,a,base,Rg) 

#define ZetabEltGetFromString(a,s,Rg) _ZetabEltGetFromString(&(a),s,Rg) 
#define ZetabEltPutToString(a,Rg) ZENEltPutToString(ZetabEltElt(Rg,a),ZetabRing(Rg)) 
#define ZetabEltGetFromFile(a,file,Rg) _ZetabEltGetFromFile(&(a),file,Rg)
#define ZetabEltPutToFile(file,a,Rg) ZENEltPutToFile(file,ZetabEltElt(Rg,a),ZetabRing(Rg)) 

#define ZetabAnyEltToZ(p,p_pl,a, Rg) (ZBNDigit(p, 0)=(a),*(p_pl)=1)
#define ZetabAnyZToElt(a,p,p_pl, Rg) ((a)=ZBNDigit(p, 0))

#define ZetabEltToZ(p,p_pl,a,Rg) ZetabAnyEltToZ(p,p_pl,ZetabElt2Int(a, Rg),Rg)
#define ZetabZToElt(a,p,p_pl,Rg) ZetabAnyZToElt(ZetabElt2Int(a, Rg),p,p_pl,Rg)

/* On polynomials
 ****************/
#define ZetabPolyDeg(p,Rg) ZENPolyDeg(p,Rg)
#define ZetabPolyLgt(p,Rg) ZENPolyLgt(p,Rg)

#define ZetabPolyCof(r,Rg) ZENPolyChar(r,Rg)
#define ZetabPolyGetCof(r,i,Rg) ZENPolyGetChar(r,i,Rg)

/* Allocation */
#define ZetabPolyCreate(d, Rg) _ZetabPolyCreate(d, Rg) 
#define ZetabPolyCreateLarge(d, Rg) _ZetabPolyCreate(d, Rg) 
#define ZetabPolyFree(r, Rg) _ZetabPolyFree(r, Rg) 
#define ZetabPolyCopy(PX,Rg) ZENPolyCopy(PX,Rg)

/* Assigning */
#define ZetabPolyAssign(r,PX,Rg) _ZetabPolyAssign(r,PX,Rg)
#define ZetabPolySetToZero(r,Rg) _ZetabPolySetToZero(r,Rg)
#define ZetabPolySetToXi(r,i,Rg) _ZetabPolySetToXi(r,i,Rg)
#define ZetabPolySetRandom(r,d,Rg) _ZetabPolySetRandom(r,d,Rg)
#define ZetabPolySetCoeff(r,i,b,Rg) (ZetabPolyGetCof(r, i, Rg) = (unsigned char)ZetabElt2Int(b, Rg))
#define ZetabPolyGetCoeff(c, r,i,Rg) (ZetabElt2Int(c, Rg) = (BigNumDigit)ZetabPolyGetCof(r, i, Rg))
#define ZetabPolyGetCoeffPtr(r,i,Rg) _ZetabPolyGetCoeffPtr(r,i,Rg) 

#define ZetabPolyDigit(r,Rg) (r)->coeff.n
#define ZetabPolyGetDigit(r,i,Rg) ZBNDigit(ZetabPolyDigit(r, Rg), i)
#define ZetabPolySetDigit(r,i,n,Rg) ZetabPolyGetDigit(r,i,Rg) = n

#define ZetabPolyUpdateDegree(r,Rg) _ZetabPolyUpdateDegree(r,Rg)

/* Test */
#define ZetabPolyAreEqual(r,PX,Rg) _ZetabPolyAreEqual(r,PX,Rg)
#define ZetabPolyIsZero(r,Rg) (!(r)->deg && !(ZetabPolyGetCof(r, 0, Rg)))
#define ZetabPolyIsXi(p,i,Rg) _ZetabPolyIsXi(p.i,Rg)

/* Arithmetic */
#define ZetabPolyAdd(r,PX,Rg) _ZetabPolyAdd(r,PX,Rg)
#define ZetabPolySubtract(r,PX,Rg) _ZetabPolySubtract(r,PX,Rg)
#define ZetabPolyNegate(r,PX,Rg) _ZetabPolyNegate(r,PX,Rg)

#define ZetabPolySquareChar2(c,r,Rg) _ZetabPolySquareChar2(c,r,Rg)
#define ZetabPolySquareCharp(c,r,Rg) _ZetabPolySquareCharp(c,r,Rg)
#define ZetabPolySquareKaratsuba(c,r,Rg) _ZetabPolySquareKaratsuba(c,r,Rg)
#define ZetabPolySquareLazy(c,r,Rg) _ZetabPolySquareKaratsuba(c,r,Rg)
#define ZetabPolySquareKaratsubaLazy(c,p,Rg) _ZetabPolySquareKaratsubaLazy(c,p,Rg)

#define ZetabPolyMultiply(c,r,PX,Rg) _ZetabPolyMultiply(c,r,PX,Rg)
#define ZetabPolyMultiplyLazy(c,r,PX,Rg) _ZetabPolyMultiplyLazy(c,r,PX,Rg)
#define ZetabPolyMultiplyPlain(c,r,PX,Rg) _ZetabPolyMultiplyPlain(c,r,PX,Rg)
#define ZetabPolyMultiplyKaratsuba(c,r,PX,Rg) _ZetabPolyMultiplyKaratsuba(c,r,PX,Rg)
#define ZetabPolyMultiplyKaratsubaLazy(c,r,PX,Rg) _ZetabPolyMultiplyKaratsubaLazy(c,r,PX,Rg)

#define ZetabPolyDivide(r,MX,PX,QX,Rg) _ZetabPolyDivide(r,MX,PX,QX,Rg)
#define ZetabPolyDivideLazy(r,MX,PX,QX,Rg) _ZetabPolyDivideLazy(r,MX,PX,QX,Rg)
#define ZetabPolyModuloLazy(MX,PX,QX,k,Rg) _ZetabPolyModuloLazy(MX,PX,QX,k,Rg)

/* io */
#define ZetabPolyReadFromString(r,s,base,Rg) ZENPolyReadFromString(r,s,base,Rg) 
#define ZetabPolyPrintToString(r,base,Rg) ZENPolyPrintToString(r,base,Rg)
#define ZetabPolyReadFromFile(r,file,base,Rg) ZENPolyReadFromFile(r,file,base,Rg) 
#define ZetabPolyPrintToFile(file,r,base,Rg) ZENPolyPrintToFile(file,r,base,Rg)

#define ZetabPolyGetFromString(r,s,Rg) ZENPolyReadFromString(r,s,10,Rg)
#define ZetabPolyPutToString(r,Rg) ZENPolyPrintToString(r,10,Rg)
#define ZetabPolyGetFromFile(r,file,Rg) ZENPolyReadFromFile(r,file,10,Rg) 
#define ZetabPolyPutToFile(file,r,Rg) ZENPolyPrintToFile(file,r,10,Rg)

/* On matrices
 ****************/
#define ZetabMatNbPtr(M,R) ZENMatNbPtr(M,R)
#define ZetabMatSize(M,R) ZENMatSize(M,R)
#define ZetabMatNbBn(M,R) ZENMatNbBn(M,R)
#define ZetabMatNbBnd(M,R) ZENMatNbBnd(M,R)
#define ZetabMatNbDigit(M,R) ZENMatNbDigit(M,R)
#define ZetabMatNbBloc(M,R) ZENMatNbBloc(M,R)
#define ZetabMatType(M,R) ZENMatType(M,R)
#define ZetabMatFirst(M,R) ZENMatFirst(M,R)

#define ZetabMatIsRowType(M,R) ZENMatIsRowType(M,R)
#define ZetabMatIsPermutation(M,R) ZENMatIsPermutation(M,R)
#define ZetabMatAreSameType(A,B,R) ZENMatAreSameType(A,B,R)

#define ZetabMatNbRow(M,R) ZENMatNbRow(M,R)
#define ZetabMatNbCol(M,R) ZENMatNbCol(M,R)



/* Allocation */
#define ZetabMatCreate(l,c,R) ZENMatCreate(l,c,R)
#define ZetabMatRowCreate(l,c,R) ZENMatRowCreate(l,c,R)
#define ZetabMatColCreate(l,c,R) ZENMatColCreate(l,c,R)
#define ZetabPermutationCreate(n,t,R) ZENPermutationCreate(n,t,R)
#define ZetabPermutationRowCreate(n,R) ZENPermutationRowCreate(n,R)
#define ZetabPermutationColCreate(n,R) ZENPermutationColCreate(n,R)
#define ZetabMatFree(M,R) ZENMatFree(M,R)
#define ZetabMatCopy(M,R) ZENMatCopy(M,R)

/* Assignation */
#define ZetabMatFastPermute(M,l1,l2,R) ZENMatFastPermute(M,l1,l2,R)
#define ZetabMatGetCoeffPtr(M,l,c,R) ZENMatGetCoeffPtr(M,l,c,R)  
#define ZetabMatGetCoeff(Z,M,l,c,R) ZENMatGetCoeff(Z,M,l,c,R)  
#define ZetabMatSetCoeff(M,l,c,Z,R) ZENMatSetCoeff(M,l,c,Z,R)

#define ZetabMatPermuteRow(M,l1,l2,R) ZENMatPermuteRow(M,l1,l2,R)
#define ZetabMatPermuteCol(M,l1,l2,R) ZENMatPermuteCol(M,l1,l2,R)


/* Arithmetic */
#define ZetabMatMultiply(X,A,B,R) ZENMatMultiply(X,A,B,R)
#define ZetabMatSquare(A,B,R) ZENMatSquare(A,B,R)

/* io */
#define ZetabMatReadFromString(r,s,base,Rg) ZENMatReadFromString(r,s,base,Rg)
#define ZetabMatPrintToString(r,base,Rg) ZENMatPrintToString(r,base,Rg)
#define ZetabMatReadFromFile(r,file,base,Rg) ZENMatReadFromFile(r,file,base,Rg)
#define ZetabMatPrintToFile(file,r,base,Rg) ZENMatPrintToFile(file,r,base,Rg)

/* Following the local Zetab functions */

/* Allocation */
#define ZetabMatCreateParams(ts,l,c,p_pms,R) _ZetabMatCreateParams(ts,l,c,p_pms,R)

/* Assigning */
#define ZetabMatAssign(A,B,R) _ZetabMatAssign(A,B,R)
#define ZetabMatSameTypeAssign(A,B,R) ZBNAssign(ZENMatFirst(A,R),ZENMatFirst(B,R),ZENMatNbBnd(A,R))
#define ZetabMatSetToZero(A,R) ZBNSetToZero(ZENMatFirst(A,R),ZENMatNbBnd(A,R))
#define ZetabMatSetToOne(M,R) _ZetabMatSetToOne(M,R)
#define ZetabMatSetRandom(M,R) _ZetabMatSetRandom(M,R)

#define ZetabMatSetBloc(M,l,c,Z,R) _ZetabMatSetBloc(M,l,c,Z,R)
#define ZetabMatGetBloc(Z,M,l,c,R) _ZetabMatGetBloc(Z,M,l,c,R)
#define ZetabMatSetDigit(M,l,c,Z,R) ZetabMatSetBloc(M,c,l,Z,R)
#define ZetabMatGetDigit(Z,M,l,c,R) ZetabMatGetBloc(Z,M,c,l,R)
 
#define ZetabMatPermuteBn(M,l1,l2,R) _ZetabMatPermuteBn(M,l1,l2,R)
#define ZetabMatPermuteDigit(M,l1,l2,R) _ZetabMatPermuteDigit(M,l1,l2,R)

/* Test */
#define ZetabMatSameTypeAreEqual(A,B,R) (!(ZBNCompare(ZENMatFirst(A,R),ZENMatNbBnd(A,R),ZENMatFirst(B,R),ZENMatNbBnd(B,R))))
#define ZetabMatIsZero(A,R) ZBNIsZero(ZENMatFirst(A,R),ZENMatNbBnd(A,R))
#define ZetabMatIsOne(M,R) _ZetabMatIsOne(M,R)

/* Arithmetic */
#define ZetabMatAdd(A,B,R) _ZetabMatAdd(A,B,R)
#define ZetabMatSameTypeAdd(A,B,R) _ZetabMatSameTypeAdd(A,B,R)
#define ZetabMatAddScalar(A,b,R) _ZetabMatAddScalar(A,b,R)
#define ZetabMatNegate(A,B,R) _ZetabMatNegate(A,B,R)
#define ZetabMatSubtract(A,B,R) _ZetabMatSubtract(A,B,R)
#define ZetabMatMultiplyPlain(X,A,B,R) _ZetabMatMultiplyPlain(X,A,B,R)
#define ZetabMatMultiplyScalar(A,b,R) _ZetabMatMultiplyScalar(A,b,R)
#define ZetabMatGaussPlain(D,S,P,p_rk,R) ZENMatGaussPlain(D,S,P,p_rk,R)
#define ZetabMatGaussRank(D,p_rk,R) ZENMatGaussRank(D,p_rk,R)
#define ZetabMatGaussLazy(D,P,p_rk,R) ZENMatGaussLazy(D,P,p_rk,R)
#define ZetabMatGaussInverse(D,S,R) ZENMatGaussInverse(D,S,R)
#define ZetabMatInverse(I,M,R) _ZetabMatInverse(I,M,R)
#define ZetabMatKernel(K,M,R) _ZetabMatKernel(K,M,R)

/* io */
#define ZetabMatGetFromString(r,s,Rg) ZENMatGetFromString(r,s,Rg)
#define ZetabMatPutToString(r,Rg) ZENMatPutToString(r,Rg)
#define ZetabMatGetFromFile(r,file,Rg) ZENMatGetFromFile(r,file,Rg)
#define ZetabMatPutToFile(file,r,Rg) ZENMatPutToFile(file,r,Rg)

/* Procedure on rings
 ***********************/
extern int _ZetabInit ___((ZENRing));
extern int _ZetabClose ___((ZENRing));

extern int _ZetabPrcInitPolyRootsCanonical ___((ZENRing));
extern int _ZetabPrcClosePolyRootsCanonical ___((ZENRing));

/* Procedure on elements
 ***********************/

/* Allocation */
extern int _ZetabEltCreate ___((ZENElt *, ZENRing)); 
extern void _ZetabEltFree ___((ZENElt, ZENRing));

/* Assigning */
extern void _ZetabEltAssign ___((ZENElt *, ZENElt, ZENRing));
extern void _ZetabEltSetToZero ___((ZENElt *, ZENRing));
extern void _ZetabEltSetToOne ___((ZENElt *, ZENRing));
extern void _ZetabEltSetRandom ___((ZENElt *, ZENRing));
extern void _ZetabEltSetNext ___((ZENElt *, ZENRing));

extern void _ZetabEltToClone ___((ZENElt *, ZENElt, ZENRing));
extern void _ZetabCloneToElt ___((ZENElt *, ZENElt, ZENRing));

/* Test */
extern int _ZetabEltAreEqual ___((ZENElt, ZENElt, ZENRing));
extern int _ZetabEltIsZero ___((ZENElt, ZENRing));
extern int _ZetabEltIsOne ___((ZENElt, ZENRing));

/* Arithmetic */
extern void _ZetabEltAdd ___((ZENElt *, ZENElt, ZENRing));
extern void _ZetabEltNegate ___((ZENElt *, ZENElt, ZENRing));
extern void _ZetabEltSubtract ___((ZENElt *, ZENElt, ZENRing));
extern void _ZetabEltSquare ___((ZENElt *, ZENElt, ZENRing));
extern void _ZetabEltMultiply ___((ZENElt *, ZENElt, ZENElt, ZENRing));
extern int __ZetabEltInverse ___((BigNumDigit *, BigNumDigit, ZENRing));
extern int _ZetabEltInverse ___((ZENElt *, ZENElt, ZENRing));
extern void _ZetabEltTrace ___((ZENElt *, ZENElt, ZENRing));

/* io */
extern void _ZetabEltToZ ___((BigNum, BigNumLength *, ZENElt, ZENRing));
extern void _ZetabZToElt ___((ZENElt *, BigNum, BigNumLength, ZENRing));

/* Procedures on polynomials
 ***************************/
/* Allocation */
extern ZENPoly _ZetabPolyCreate ___((int, ZENRing)); /* Create polynomials */
extern void _ZetabPolyFree ___((ZENPoly, ZENRing)); /* Free polynomials */

/* Assigning */
extern void _ZetabPolyAssign ___((ZENPoly, ZENPoly, ZENRing)); /* Assign a polynomial with another */
extern void _ZetabPolySetToZero ___((ZENPoly, ZENRing)); /* Set a polynomial to zero */
extern void _ZetabPolySetToXi ___((ZENPoly, int, ZENRing)); /* Set a polynomial to $X^i$ */
extern void _ZetabPolySetRandom ___((ZENPoly, int, ZENRing)); /* Set randomly a polynomial */
extern void _ZetabPolySetCoeff ___((ZENPoly, int, ZENElt, ZENRing)); /* Set a coefficient of a polynomial */
extern void _ZetabPolyGetCoeff ___((ZENElt *, ZENPoly, int, ZENRing)); /* Get a coefficient of a polynomial */
extern ZENElt _ZetabPolyGetCoeffPtr ___((ZENPoly, int, ZENRing)); /* Get a pointer on a coefficient of a polynomial */
extern void _ZetabPolyUpdateDegree ___((ZENPoly, ZENRing)); /* Update the degree of a polynomial */

/* Test */
extern int _ZetabPolyAreEqual ___((ZENPoly, ZENPoly, ZENRing)); /* Is a polynomial equal to another */
extern int _ZetabPolyIsZero ___((ZENPoly, ZENRing)); /* Is a polynomial equal to zero */
extern int _ZetabPolyIsXi ___((ZENPoly, int, ZENRing)); /* Is a polynomial equal to $X^i$ */

/* Arithmetic */
extern void _ZetabPolyAdd ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZetabPolyNegate ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZetabPolySubtract ___((ZENPoly, ZENPoly, ZENRing)); /* Substraction of polynomials */

extern int _ZetabPolySquareChar2 ___((ZENPoly, ZENPoly, ZENRing)); /* Squaring of polynomials */
extern int _ZetabPolySquareCharp ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZetabPolySquareKaratsuba ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZetabPolySquareKaratsubaLazy ___((ZENPoly, ZENPoly, ZENRing));

extern int _ZetabPolyMultiply ___((ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Multiplication of polynomials */
extern int _ZetabPolyMultiplyLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZetabPolyMultiplyPlain ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZetabPolyMultiplyKaratsuba ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZetabPolyMultiplyKaratsubaLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

extern void _ZetabPolyMultiplyScalar ___((ZENPoly, ZENPoly, ZENElt, ZENRing)); /* Multiplication of polynomials */
extern void _ZetabPolyDot ___((ZENElt *, ZENPoly, ZENPoly, ZENRing)); /* Scalar product of 2 polynomials */
extern int _ZetabPolyDivide ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Division of polynomials */
extern int _ZetabPolyDivideLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Lazy division of polynomials */
extern int _ZetabPolyModuloLazy ___((ZENPoly, ZENPoly, ZENPoly, int, ZENRing)); /* Lazy modulo of polynomials */

/* Procedures on matrices
 ***************************/
/* Allocation */
extern int _ZetabMatCreateParams ___((ZENMatTypeStruct,Dim,Dim,ZENMat *,ZENRing));

/* Assignment */
extern int _ZetabMatAssign ___((ZENMat,ZENMat,ZENRing));
extern int _ZetabMatSetToZero ___((ZENMat,ZENRing));
extern int _ZetabMatSetToOne ___((ZENMat,ZENRing));
extern int _ZetabMatSetRandom ___((ZENMat,ZENRing));
extern int _ZetabMatGetBloc ___((ZENElt *,ZENMat,Dim,Dim,ZENRing));
extern int _ZetabMatGetSubMat ___((ZENMat,ZENMat,Dim,Dim,ZENRing));
extern int _ZetabMatSetSubMat ___((ZENMat,Dim,Dim,ZENMat,ZENRing));
extern ZENElt _ZetabMatGetBlocPtr ___((ZENMat,Dim,Dim,ZENRing));
extern int _ZetabMatSetBloc ___((ZENMat,Dim,Dim,ZENElt,ZENRing));
extern int _ZetabMatPermuteBn ___((ZENMat,Dim,Dim,ZENRing));
extern int _ZetabMatPermuteDigit ___((ZENMat,Dim,Dim,ZENRing));

/*Tests */
extern int _ZetabMatIsZero ___((ZENMat,ZENRing));
extern int _ZetabMatIsOne ___((ZENMat,ZENRing));
extern int _ZetabMatAreEqual ___((ZENMat,ZENMat,ZENRing));

/*Arithmetic */
extern int _ZetabMatAdd ___((ZENMat,ZENMat,ZENRing));
extern int _ZetabMatAddScalar ___((ZENMat,ZENElt,ZENRing));
extern int _ZetabMatNegate ___((ZENMat,ZENMat,ZENRing));
extern int _ZetabMatSubtract ___((ZENMat,ZENMat,ZENRing));
extern int _ZetabMatMultiplyPlain ___((ZENMat,ZENMat,ZENMat,ZENRing));
extern int _ZetabMatMultiplyScalar ___((ZENMat,ZENElt,ZENRing));
extern int _ZetabMatGauss ___((ZENMat,ZENMat,ZENMat,Dim *,ZENElt *,
			       int (*) ___((ZENMat,int,int,ZENRing)),
			       ZENRing));
extern int _ZetabMatInverse ___((ZENMat,ZENMat,ZENRing));
extern int _ZetabMatKernel ___((ZENMat *,ZENMat,ZENRing));

/* matrix conversion */
extern int _Zetab2mVect2Mat ___((ZENMat, ZENMat, ZENRing));
extern int _Zetab2mMat2Vect ___((ZENMat, ZENMat, ZENRing));

/* 
   \TEX{
   \Section{Ring initialization}
   \input{zetabinit}

   \Section{Element operations}
   \input{zetabelt_alloc}
   \input{zetabelt_assign}
   \input{zetabelt_test}
   \input{zetabelt_add}
   \input{zetabelt_multiply}
   \input{zetabelt_inverse}
   \input{zetabelt_trace}
   \input{zetabelt_io}

   \Section{Polynomial operations}
   \input{zetabpoly_alloc}
   \input{zetabpoly_assign}
   \input{zetabpoly_test}
   \input{zetabpoly_add}
   \input{zetabpoly_multiply}

   \Section{Matricial operations}
   \input{zetabmat_create}
   \input{zetabmat_assign}
   \input{zetabmat_permute}
   \input{zetabmat_test}
   \input{zetabmat_add}
   \input{zetabmat_multiply}
   \input{zetabmat_gauss}
   \input{zetabmat_inverse}
   \input{zetabmat_convert}
   }
   */
#endif

