/*
 * zen/include/zext.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZEXT__
#define __ZEXT__
#include "zen.h"



#define ZextRingRbuf(R) ZENRingRbuf(R)
#define ZextRingHbuf(R) ZENRingHbuf(R)
#define ZextRingIbuf(R) ZENRingIbuf(R)
#define ZextRingPbuf(R) ZENRingPbuf(R)
#define ZextRingTbuf(R) ZENRingTbuf(R)
#define ZextRingTmp1(R) ZENRingTmp1(R)
#define ZextRingTmp2(R) ZENRingTmp2(R)
#define ZextRingRec(R)  ZENRingRec(R)

/* \TEX{
   The basic principle of extension creation is to use the polyniomials functions of the current |ZENRing| to create a new |ZENRing| for which the |Elt| functions will be the polynomial ones of the underlevel |ZENRing|.
   
   }
   */

#define ZextElt2Pol(a, Rg) ZENElt2Pol(a, Rg)
#define ZextEltSetToX(a, Rg) (ZextEltSetToZero(a,Rg), ZENPolySetToXi(ZENElt2Pol(a, Rg),1,ZENRingDef(Rg)))

/* On elements
 *************/

/* Allocation */
#define ZextEltCreate(a, Rg) ((ZextElt2Pol(a, Rg) = ZENPolyCreate(ZENRingDeg(Rg)-1, ZENRingDef(Rg))) ? 0 : ZERR)
#define ZextEltCreateLarge(a, Rg) 							\
(											\
 (											\
  (ZENRingArith(ZENRingDef(Rg)) == ZEN_LIB_ZE2)						\
  ?											\
  (ZextElt2Pol(a, Rg) = ZENPolyCreateLarge(4*ZENRingDeg(Rg)+SIZE_BLOC, ZENRingDef(Rg)))	\
  :											\
  (ZextElt2Pol(a, Rg) = ZENPolyCreateLarge(4*ZENRingDeg(Rg)+1, ZENRingDef(Rg)))		\
  )											\
 ?											\
 0											\
 :											\
 ZERR											\
)
#define ZextEltFree(a, Rg) ZENPolyFree(ZextElt2Pol(a, Rg), ZENRingDef(Rg))

/* Assigning */
#define ZextEltAssign(a,b,Rg) ZENPolyAssign(ZextElt2Pol(a, Rg), ZextElt2Pol(b, Rg), ZENRingDef(Rg))
#define ZextEltSetToZero(a,Rg) ZENPolySetToZero(ZextElt2Pol(a, Rg), ZENRingDef(Rg))
#define ZextEltSetToOne(a,Rg) ZENPolySetToXi(ZextElt2Pol(a, Rg), 0, ZENRingDef(Rg))
#define ZextEltSetRandom(a,Rg) ZENPolySetRandom(ZextElt2Pol(a, Rg), ZENRingDeg(Rg)-1, ZENRingDef(Rg))
#define ZextEltSetNext(a,Rg) _ZextEltSetNext(&(a),Rg)

/* Test */
#define ZextEltAreEqual(a,b,Rg) ZENPolyAreEqual(ZextElt2Pol(a, Rg), ZextElt2Pol(b, Rg), ZENRingDef(Rg))
#define ZextEltIsZero(a,Rg) ZENPolyIsZero(ZextElt2Pol(a, Rg), ZENRingDef(Rg))
#define ZextEltIsOne(a,Rg) ZENPolyIsXi(ZextElt2Pol(a, Rg), 0, ZENRingDef(Rg))

/* Arithmetic */
#define ZextEltAdd(r,a,Rg) ZENPolyAdd(ZextElt2Pol(r, Rg), ZextElt2Pol(a, Rg), ZENRingDef(Rg))
#define ZextEltNegate(c,a,Rg) ZENPolyNegate(ZextElt2Pol(c, Rg), ZextElt2Pol(a, Rg), ZENRingDef(Rg))
#define ZextEltSubtract(c,a,Rg)  ZENPolySubtract(ZextElt2Pol(c, Rg), ZextElt2Pol(a, Rg), ZENRingDef(Rg))
#define ZextEltSquare(r,a,Rg) ZENEltSquare(r,a,Rg)
#define ZextEltMultiply(r,a,b,Rg) ZENEltMultiply(r,a,b,Rg)
#define ZextEltInverse(r,a,Rg) _ZextEltInverse(&(r),a,Rg)
#define ZextEltTrace(r,a,Rg) ZENEltTrace(r,a,Rg) 
#define ZextEltTracePrecomputation(Rg) _ZextEltTracePrecomputation(Rg)
#define ZextEltExp(r,k,kl,a,R) ZENEltExp(r,k,kl,a,R) 
#define ZextEltTraceOne(R) _ZextEltTraceOne(R)
#define ZextEltConvert(a,R,b,Rg) _ZextEltConvert(&(a),R,b,Rg)

/* io */
#define ZextEltReadFromString(a,s,base,Rg) _ZextEltReadFromString(&(a),s,base,Rg) 
#define ZextEltPrintToString(a,base,Rg) _ZENPolyPrintToString(ZextElt2Pol(a, Rg),ZENRingVar(Rg),base,ZENRingDef(Rg))
#define ZextEltReadFromFile(a,file,base,Rg) _ZextEltReadFromFile(&(a),file,base,Rg)
#define ZextEltPrintToFile(file,a,base,Rg) _ZENPolyPrintToFile(file,ZextElt2Pol(a, Rg),ZENRingVar(Rg),base,ZENRingDef(Rg))

#define ZextEltGetFromString(a,s,Rg) _ZextEltGetFromString(&(a),s,Rg) 
#define ZextEltPutToString(a,Rg) ZENPolyPutToString(ZextElt2Pol(a, Rg),ZENRingDef(Rg))
#define ZextEltGetFromFile(a,file,Rg) _ZextEltGetFromFile(&(a),file,Rg) 
#define ZextEltPutToFile(file,a,Rg) ZENPolyPutToFile(file,ZextElt2Pol(a, Rg),ZENRingDef(Rg))

#define ZextEltToZ(p,p_pl,a,R) _ZextEltToZ(p,p_pl,a,R)
#define ZextZToElt(a,p,pl,R) _ZextZToElt(&(a),p,pl,R) 

/* On polynomials
 ****************/
#define ZextPolyDeg(p,Rg) ZENPolyDeg(p,Rg)
#define ZextPolyLgt(p,Rg) ZENPolyLgt(p,Rg)

#define ZextPolyCof(r,Rg) ZENPolyCof(r,Rg)
#define ZextPolyGetCof(r,i,Rg) ZENPolyGetCof(r,i,Rg)

/* Allocation */
#define ZextPolyCreate(d, Rg) _ZextPolyCreate(d, Rg) 
#define ZextPolyCreateLarge(d, Rg) _ZextPolyCreateLarge(d, Rg) 
#define ZextPolyFree(r, Rg) _ZextPolyFree(r, Rg) 
#define ZextPolyCopy(PX,Rg) ZENPolyCopy(PX,Rg)

/* Assigning */
#define ZextPolyAssign(r,PX,Rg) _ZextPolyAssign(r,PX,Rg)
#define ZextPolyAssignLarge(r,PX,Rg) _ZextPolyAssignLarge(r,PX,Rg)
#define ZextPolyLazyToVeryLazy(r,PX,Rg) _ZextPolyLazyToVeryLazy(r,PX,Rg)
#define ZextPolySetToZero(r,Rg) _ZextPolySetToZero(r,Rg)
#define ZextPolySetToZeroLarge(r,Rg) _ZextPolySetToZeroLarge(r,Rg)
#define ZextPolySetToXi(r,i,Rg) _ZextPolySetToXi(r,i,Rg)
#define ZextPolySetRandom(r,d,Rg) _ZextPolySetRandom(r,d,Rg)
#define ZextPolySetCoeff(r,i,b,Rg) ZENEltAssign(ZextPolyGetCof(r, i, Rg), b, Rg)
#define ZextPolyGetCoeff(c, r,i,Rg) ZENEltAssign(c, ZextPolyGetCof(r, i, Rg), Rg)
#define ZextPolyGetCoeffPtr(r,i,Rg) ZextPolyGetCof(r,i,Rg)

#define ZextPolyDigit(r,Rg) (r)->coeff.n
#define ZextPolyGetDigit(r,i,Rg) ZBNDigit(ZextPolyDigit(r, Rg), i)
#define ZextPolySetDigit(r,i,n,Rg) ZextPolyGetDigit(r,i,Rg) = n

#define ZextPolyUpdateDegree(r,Rg) _ZextPolyUpdateDegree(r,Rg)

/* Test */
#define ZextPolyAreEqual(r,PX,Rg) _ZextPolyAreEqual(r,PX,Rg)
#define ZextPolyIsZero(r,Rg) (!(r)->deg && ZextEltIsZero(ZextPolyGetCof(r, 0, Rg), Rg))
#define ZextPolyIsXi(p,i,Rg) _ZextPolyIsXi(p.i,Rg)

/* Arithmetic */
#define ZextPolyAdd(r,PX,Rg) _ZextPolyAdd(r,PX,Rg)
#define ZextPolyAddLarge(r,PX,Rg) _ZextPolyAddLarge(r,PX,Rg)
#define ZextPolySubtract(r,PX,Rg) _ZextPolySubtract(r,PX,Rg)
#define ZextPolySubtractLarge(r,PX,Rg) _ZextPolySubtractLarge(r,PX,Rg)
#define ZextPolyNegate(r,PX,Rg) _ZextPolyNegate(r,PX,Rg)
#define ZextPolyNegateLarge(r,PX,Rg) _ZextPolyNegateLarge(r,PX,Rg)

#define ZextPolySquareChar2(c,r,Rg) _ZextPolySquareChar2(c,r,Rg)

#define ZextPolySquareCharpPlainVeryLazy(c,r,Rg) \
     _ZextPolySquareCharpPlainVeryLazy(c,r,Rg)
#define ZextPolySquareCharpPlainLazy(c,r,Rg) \
     _ZextPolySquareCharpPlainVeryLazy(c,r,Rg)
#define ZextPolySquareCharpPlain(c,r,Rg) \
     _ZextPolySquareCharpPlain(c,r,Rg)

#define ZextPolySquareKaratsuba(c,r,Rg) _ZextPolySquareKaratsuba(c,r,Rg)
#define ZextPolySquareKaratsubaLazy(c,p,Rg) _ZextPolySquareKaratsubaLazy(c,p,Rg)
#define ZextPolySquareKaratsubaVeryLazy(c,p,Rg) _ZextPolySquareKaratsubaVeryLazy(c,p,Rg)

#define ZextPolyMultiply(c,r,PX,Rg) _ZextPolyMultiply(c,r,PX,Rg)
#define ZextPolyMultiplyLazy(c,r,PX,Rg) _ZextPolyMultiplyLazy(c,r,PX,Rg)
#define ZextPolyMultiplyVeryLazy(c,r,PX,Rg) _ZextPolyMultiplyVeryLazy(c,r,PX,Rg)

#define ZextPolyMultiplyPlain(c,r,PX,Rg) _ZextPolyMultiplyPlain(c,r,PX,Rg)
#define ZextPolyMultiplyPlainLazy(c,r,PX,Rg) _ZextPolyMultiplyPlain(c,r,PX,Rg)
#define ZextPolyMultiplyPlainVeryLazy(c,r,PX,Rg) _ZextPolyMultiplyPlainVeryLazy(c,r,PX,Rg)

#define ZextPolyMultiplyKaratsuba(c,r,PX,Rg) _ZextPolyMultiplyKaratsuba(c,r,PX,Rg)
#define ZextPolyMultiplyKaratsubaLazy(c,r,PX,Rg) _ZextPolyMultiplyKaratsubaLazy(c,r,PX,Rg)
#define ZextPolyMultiplyKaratsubaVeryLazy(c,r,PX,Rg) _ZextPolyMultiplyKaratsubaVeryLazy(c,r,PX,Rg)

#define ZextPolyMultiplyScalar(r,p,e,Rg) _ZextPolyMultiplyScalar(r,p,e,Rg)
#define ZextPolyDot(r,p,q,R) _ZextPolyDot(r,p,q,R)

#define ZextPolyDivide(r,MX,PX,QX,Rg) _ZextPolyDivide(r,MX,PX,QX,Rg)
#define ZextPolyDivideLazy(r,MX,PX,QX,Rg) _ZextPolyDivideLazy(r,MX,PX,QX,Rg)
#define ZextPolyDivideVeryLazy(r,MX,PX,QX,Rg) _ZextPolyDivideVeryLazy(r,MX,PX,QX,Rg)
#define ZextPolyModuloVeryLazy(MX,PX,QX,k,Rg) _ZextPolyModuloVeryLazy(MX,PX,QX,k,Rg)

/* io */
#define ZextPolyReadFromString(r,s,base,Rg) ZENPolyReadFromString(r,s,base,Rg) 
#define ZextPolyPrintToString(r,base,Rg) ZENPolyPrintToString(r,base,Rg)
#define ZextPolyReadFromFile(r,file,base,Rg) ZENPolyReadFromFile(r,file,base,Rg) 
#define ZextPolyPrintToFile(file,r,base,Rg) ZENPolyPrintToFile(file,r,base,Rg)

#define ZextPolyGetFromString(r,s,Rg) ZENPolyReadFromString(r,s,10,Rg)
#define ZextPolyPutToString(r,Rg) ZENPolyPrintToString(r,10,Rg)
#define ZextPolyGetFromFile(r,file,Rg) ZENPolyReadFromFile(r,file,10,Rg) 
#define ZextPolyPutToFile(file,r,Rg) ZENPolyPrintToFile(file,r,10,Rg)


/* Rings
 *******/
extern int _ZextInit ___((ZENRing));
extern int _ZextClose ___((ZENRing));

extern int _Ze2mPrcInitEltMultiply ___((ZENRing));
extern int _Ze2mPrcCloseEltMultiply ___((ZENRing));

extern int _ZextPrcInitEltMultiply ___((ZENRing));
extern int _ZextPrcCloseEltMultiply ___((ZENRing));

extern int _ZextPrcInitTrace ___((ZENRing));
extern int _ZextPrcCloseTrace ___((ZENRing));

extern int _ZextPrcInitPolyRootsCanonical ___((ZENRing));
extern int _ZextPrcClosePolyRootsCanonical ___((ZENRing));

extern int _ZextPrcInitEltTabulate ___((ZENRing,int));

/* Procedure on elements
 ***********************/
/* Allocation */
extern int _ZextEltCreate ___((ZENElt *, ZENRing)); /* Create elements */
extern int _ZextEltCreateLarge ___((ZENElt *, ZENRing));
extern void _ZextEltFree ___((ZENElt, ZENRing)); /* Free elements */

/* Assigning */
extern void _ZextEltAssign ___((ZENElt *, ZENElt, ZENRing)); /* Assign an element with another */
extern void _ZextEltSetToZero ___((ZENElt *, ZENRing)); /* Set an element to zero */
extern void _ZextEltSetToOne ___((ZENElt *, ZENRing)); /* Set an element to one */
extern void _ZextEltSetRandom ___((ZENElt *, ZENRing)); /* Set randomly an element */
extern void _ZextEltSetNext ___((ZENElt *, ZENRing)); /* Set an element to the next */
extern int _ZextEltConvert ___((ZENElt *, ZENRing, ZENElt, ZENRing));

/* Test */
extern int _ZextEltAreEqual ___((ZENElt, ZENElt, ZENRing)); /* Is an element equal to another */
extern int _ZextEltIsZero ___((ZENElt, ZENRing)); /* Is an element equal to zero */
extern int _ZextEltIsOne ___((ZENElt, ZENRing)); /* Is an element equal to one */

/* Arithmetic */
extern void _ZextEltAdd ___((ZENElt *, ZENElt, ZENRing));	/* Addition of elements */
extern void _ZextEltNegate ___((ZENElt *, ZENElt, ZENRing));	/* Negation of elements */
extern void _ZextEltSubtract ___((ZENElt *, ZENElt, ZENRing));	/* Subtraction of elements */

extern void _ZextEltSquare ___((ZENElt *, ZENElt, ZENRing)); /* Squaring of elements */
extern void _ZextEltMultiply ___((ZENElt *, ZENElt, ZENElt, ZENRing)); /* Multiplication of elements */

extern void _Ze2mEltSquare ___((ZENElt*,ZENElt,ZENRing)); /* Squaring of elements */
extern void _Ze2mEltMultiply ___((ZENElt*,ZENElt,ZENElt,ZENRing)); /* Multiplication of elements */

extern void _ZextEltSquareRecip ___((ZENElt*,ZENElt,ZENRing)); /* Squaring of elements */
extern void _ZextEltMultiplyRecip ___((ZENElt*,ZENElt,ZENElt,ZENRing)); /* Multiplication of elements */

extern int _ZextEltInverse ___((ZENElt *, ZENElt, ZENRing)); /* Inverse of elements */
extern void _ZextEltTraceNaive ___((ZENElt *, ZENElt, ZENRing));	/* Trace of elements without precomputations*/
extern void _ZextEltTracePrc ___((ZENElt *, ZENElt, ZENRing));	/* Trace of elements with precomputations*/

/* io */
extern int _ZextEltReadFromString ___((ZENElt *, char *, int, ZENRing)); /* Read a polynomial from a string */
extern char* _ZextEltPrintToString ___((ZENElt, int, ZENRing)); /* Print a polynomial to a string */
extern int _ZextEltReadFromFile ___((ZENElt *, FILE *, int, ZENRing));	/* Read a polynomial from a file */
extern int _ZextEltPrintToFile ___((FILE *, ZENElt, int, ZENRing)); /* Print a polynomial to a file */
extern int _ZextEltGetFromString ___((ZENElt *, char *, ZENRing)); /* Get a polynomial from a string */
extern char* _ZextEltPutToString ___((ZENElt, ZENRing)); /* Put a polynomial to a string */
extern int _ZextEltGetFromFile ___((ZENElt *, FILE *, ZENRing)); /* Get a polynomial from a file */
extern int _ZextEltPutToFile ___((FILE *, ZENElt, ZENRing)); /* Put a polynomial to a file */

extern void _ZextEltToZ ___((BigNum,BigNumLength*,ZENElt,ZENRing)); /* Evaluate in $p$ an element */
extern void _ZextZToElt ___((ZENElt *,BigNum,BigNumLength,ZENRing)); /* Get the element from an evaluation in $p$ */

extern int _ZextEP ___((ZENElt, ZENRing));
extern int _ZextER ___((ZENElt *, ZENRing));

/* Procedures on polynomials
 ***************************/
/* Allocation */
extern ZENPoly _ZextPolyCreate ___((int, ZENRing)); /* Create polynomials */
extern ZENPoly _ZextPolyCreateLarge ___((int, ZENRing)); /* Create polynomials */
extern void _ZextPolyFree ___((ZENPoly, ZENRing)); /* Free polynomials */

/* Assigning */
extern void _ZextPolyAssign ___((ZENPoly, ZENPoly, ZENRing)); /* Assign a polynomial with another */
extern void _ZextPolyAssignLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Assign a polynomial with another */
extern void _ZextPolyLazyToVeryLazy ___((ZENPoly, ZENPoly, ZENRing)); /* Assign a polynomial with another */
extern void _ZextPolySetToZero ___((ZENPoly, ZENRing)); /* Set a polynomial to zero */
extern void _ZextPolySetToZeroLarge ___((ZENPoly, ZENRing)); /* Set a polynomial to zero */
extern void _ZextPolySetToXi ___((ZENPoly, int, ZENRing)); /* Set a polynomial to $X^i$ */
extern void _ZextPolySetRandom ___((ZENPoly, int, ZENRing)); /* Set randomly a polynomial */
extern void _ZextPolySetCoeff ___((ZENPoly, int, ZENElt, ZENRing)); /* Set a coefficient of a polynomial */
extern void _ZextPolyGetCoeff ___((ZENElt *, ZENPoly, int, ZENRing)); /* Get a coefficient of a polynomial */
extern ZENElt _ZextPolyGetCoeffPtr ___((ZENPoly, int, ZENRing)); /* Get a pointer on a coefficient of a polynomial */
extern void _ZextPolyUpdateDegree ___((ZENPoly, ZENRing)); /* Update the degree of a polynomial */

/* Test */
extern int _ZextPolyAreEqual ___((ZENPoly, ZENPoly, ZENRing)); /* Is a polynomial equal to another */
extern int _ZextPolyIsZero ___((ZENPoly, ZENRing)); /* Is a polynomial equal to zero */
extern int _ZextPolyIsXi ___((ZENPoly, int, ZENRing)); /* Is a polynomial equal to $X^i$ */

/* Arithmetic */
extern void _ZextPolyAdd ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZextPolyAddLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Addition of polynomials */
extern void _ZextPolyNegate ___((ZENPoly, ZENPoly, ZENRing)); /* Negation of polynomials */
extern void _ZextPolyNegateLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Negation of polynomials */
extern void _ZextPolySubtract ___((ZENPoly, ZENPoly, ZENRing)); /* Substraction of polynomials */
extern void _ZextPolySubtractLarge ___((ZENPoly, ZENPoly, ZENRing)); /* Substraction of polynomials */

extern int _ZextPolySquareChar2 ___((ZENPoly, ZENPoly, ZENRing)); /* Squaring of polynomials */

extern int _ZextPolySquareCharpPlainVeryLazy ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZextPolySquareCharpPlain ___((ZENPoly, ZENPoly, ZENRing));

extern int _ZextPolySquareKaratsuba ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZextPolySquareKaratsubaLazy ___((ZENPoly, ZENPoly, ZENRing));
extern int _ZextPolySquareKaratsubaVeryLazy ___((ZENPoly, ZENPoly, ZENRing));

extern int _ZextPolyMultiply ___((ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Multiplication of polynomials */
extern int _ZextPolyMultiplyLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZextPolyMultiplyVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

extern int _ZextPolyMultiplyPlain ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZextPolyMultiplyPlainVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

extern int _ZextPolyMultiplyKaratsuba ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZextPolyMultiplyKaratsubaLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));
extern int _ZextPolyMultiplyKaratsubaVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

extern void _ZextPolyMultiplyScalar ___((ZENPoly, ZENPoly, ZENElt, ZENRing)); /* Multiplication of polynomials */

extern void _ZextPolyDot  ___((ZENElt *, ZENPoly, ZENPoly, ZENRing)); /* Scalar product of 2 polynomials */
extern int _ZextPolyDivide ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Division of polynomials */
extern int _ZextPolyDivideLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Division of polynomials */
extern int _ZextPolyDivideVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, ZENRing)); /* Division of polynomials */
extern int _ZextPolyModuloVeryLazy ___((ZENPoly, ZENPoly, ZENPoly, int, ZENRing)); /* Lazy division of polynomials */

/* Procedures on series
 ***************************/
/* Allocation */
extern ZENSr _ZextSrCreate ___((int, ZENRing)); /* Create series */
extern void _ZextSrFree ___((ZENSr, ZENRing)); /* Free series */

/* Assigning */
extern void _ZextSrAssign ___((ZENSr, ZENSr, ZENRing)); /* Assign a valuation with another */
extern void _ZextSrSetToZero ___((ZENSr,ZENRing)); /* Set a valuation to zero */
extern void _ZextSrSetRandom ___((ZENSr,ZENRing)); /* Set randomly a valuation */
extern void _ZextSrSetCoeff ___((ZENSr, int, ZENElt, ZENRing)); /* Set a coefficient of a valuation */
extern void _ZextSrGetCoeff ___((ZENElt, ZENSr, int, ZENRing)); /* Get a coefficient of a valuation */
extern ZENElt _ZextSrGetCoeffPtr ___((ZENSr, int, ZENRing)); /* Get a pointer on a coefficient of a valuation */
extern void _ZextSrUpdateValuation ___((ZENSr, ZENRing)); /* Update the valuation of a valuation */

/* Test */
extern int _ZextSrAreEqual ___((ZENSr, ZENSr, ZENRing)); /* Is a valuation equal to another */
extern int _ZextSrIsZero ___((ZENSr, ZENRing)); /* Is a valuation equal to zero */

/* Arithmetic */
extern void _ZextSrAdd ___((ZENSr, ZENSr, ZENRing)); /* Addition of series */
extern void _ZextSrNegate ___((ZENSr, ZENSr, ZENRing)); /* Negation of series */
extern void _ZextSrSubtract ___((ZENSr, ZENSr, ZENRing)); /* Substraction of series */
extern int _ZextSrSquareChar2 ___((ZENSr, ZENSr, ZENRing)); /* Squaring of series in characteristic 2*/
extern int _ZextSrSquareCharp ___((ZENSr, ZENSr, ZENRing)); /* Squaring of series in characteristic p*/
extern int _ZextSrMultiply ___((ZENSr, ZENSr, ZENSr, ZENRing)); /* Multiplication of series */
extern void _ZextSrMultiplyScalar ___((ZENSr, ZENElt, ZENSr, ZENRing)); /* Multiplication of series by a scalar */
extern int _ZextSrDivide ___((ZENSr, ZENSr, ZENRing)); /* Division of series */

/* Procedures on matrix
 **********************/
/* Allocation */
extern int _ZextMatCreateParams ___((ZENMatTypeStruct,
			   Dim,Dim,
			   ZENMat *,
			   ZENRing));		/* Creates a matrix parameters
						   structure */
extern int __ZextMatCreate ___((ZENMat, ZENRing));
extern void _ZextMatFree ___((ZENMat, ZENRing));

/* matrix conversion */
extern int _ZextVect2Mat ___((ZENMat, ZENMat, ZENRing));
extern int _ZextMat2Vect ___((ZENMat, ZENMat, ZENRing));
extern int _Ze2mVect2Mat ___((ZENMat, ZENMat, ZENRing));
extern int _Ze2mMat2Vect ___((ZENMat, ZENMat, ZENRing));
extern int _ZepmVect2Mat ___((ZENMat, ZENMat, ZENRing));
extern int _ZepmMat2Vect ___((ZENMat, ZENMat, ZENRing));


/*  
   \TEX{
   
   \Section{Ring initialization}
   \input{zextinit}
   
   \Section{Procedures on elements}
   \input{zextelt_alloc}
   \input{zextelt_assign}
   \input{zextelt_test}
   \input{zextelt_add}
   \input{zextelt_multiply}
   \input{zextelt_inverse}
   \input{zextelt_trace}
   \input{zextelt_io}

   \Section{Procedures on polynomials}
   \input{zextpoly_alloc}
   \input{zextpoly_assign}
   \input{zextpoly_test}
   \input{zextpoly_add}
   \input{zextpoly_multiply}

   \Section{Procedures on matrices}
   \input{zextmat_create}
   \input{zextmat_convert}
   }
*/

#endif

