/*-*-LaTeX-*-
 * zen/inputs/doc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
  This is the documentation file of the whole ZEN library. There is
  no C code in this file, only LaTeX commands.


   \TEX{
\input{index}

\input{title}
\thispagestyle{headings}

\tableofcontents  

\Chapter{Introduction}
\input{intro}

\Chapter{Beginning with \zen}
\input{maths}

\Chapter{Understanding \zen}
\input{user}

\Chapter{Enumerating the functions of \zen}
\input{zen}
\input{zbn}

\Chapter{Implementation principles of |ZEN|.}
\Section{How to write a new arithmetic}
\input{implement}

\Section{Sub-directories of |ZEN|}
\input{init}


\Section{Generic functions}
  \Subsection{Ring initialization}
  \input{zerdoc}

  \Subsection{Extensions}
  \input{zextdoc}

  \Subsection{General functions}
  \input{zegdoc}

  \Subsection{Default functions}
  \input{zeddoc}

  \Section{Arithmetics}
  \Subsection{Modular rings}
  \input{zepdoc}

  \Subsection{Modular rings with small modulus}
  \input{zepsdoc}

  \Subsection{$\GF{2}$ case}
  \input{ze2doc}

  \Subsection{Rationals}
  \input{zefdoc}

  \Subsection{Clones} 

  The purpose of clones is to perform when it is possible some
  precomputations in order to speed up arithmetic. These
  precomputations are made upon initialization and can take some time.
  Therefore, cloning can be efficient only if a lot of operations are
  performed.
 
  \Subsubsection{Tabulated clone}
  \input{zetabdoc}

  \Subsubsection{Logarithm clone}
  \input{zelogdoc}

  \Subsubsection{Use of chinese remainder theorem}
  \input{zecdoc}



\Section{Testing the library} 

When adding some new arithmetics to |ZEN|, it is necessary to test
these new features. This is done using the standard |zentest| program.

\input{zentest}

\appendix

\Chapter{Installing \zen}
     \Section{The principle}

     You probably got the |ZEN| package via anonymous ftp in a file
     probably named something like |ZEN.x.y.tgz|. The compilation of
     |ZEN| needs the package |ZMAKE| which should be available via
     anonymous ftp too. Then, what you have to do is:
     \begin{enumerate}
     \item To decompress |ZEN.x.y.tgz| at the same level as |ZMAKE|
       with\\ \$ | gzip -d $<$ ZEN.x.y.tgz |$\vert$| tar xvf - |.
     \item To go in the |ZEN| directory with\\
       \$ | cd ZEN.x.y|.
     \item \label{compilation} To compile with\\
       \$ | make|.
     \item \label{documentation} To get the documentation with\\
       \$ | make doc|.
     \end{enumerate}

     At the end of step \ref{compilation}, a library file |libzen.a| is
     compiled in |lib/arch| where |arch| depends on your system and the
     file |zen.h| is in |include|. 
     Moreover, test executables are in |bin/arch|.
     
     At the end of step \ref{documentation}, the user manual is in
     the file |doc/dvi/doc.dvi|. Moreover, you can possibly get the
     ``advanced user manual'' in the file |doc/dvi/advanced.dvi| after
     |make adoc|.
     
     Figure~\ref{fig:archi} enumerates the currently tested
     architectures.  Usually, the operation sequence of
     figure~\ref{fig:compil} will produce the library without problems.
     
     \begin{figure}
       \begin{center}
         \begin{tabular}
           {cccc}
%           \hline
           Vendor & Processor & Operating system & Compilers \\
           \hline
%          \hline
           SUN    &  sparc    & Solaris          & cc,gcc \\
           \hline
           DEC    &  alpha    & OSF              & cc, gcc \\
           \hline
           HP     &  PA-RISC  & HP-UX            & cc, gcc \\
           \hline
           IBM    &  RS6000   & AIX              & cc, gcc \\
           \hline
           PC     &  486      & linux            & gcc \\
%           \cline{3-4}
                  &           & Windows 95/98/2000/NT & gcc (cygnus) \\
%           \hline
         \end{tabular}
       \end{center}
       \caption{Portability of \zen}
       \label{fig:archi}
     \end{figure}

\newpage     
     \Section{Configuring the compiler options}
     \input{site}
     
     \begin{figure}
       \begin{center}
         \begin{pspicture*}(-2.5,-15)(15,0)
           \rput(2,-1){\sc Operations sequence}
           \rput(9,-1){\sc Shell command}
           
           \psframe[shadow=true](-2,-3)(6,-2)
           \rput(2,-2.5){Local parameters configuration}
           \psframe(7,-3)(11,-2)
           \rput(9,-2.5){\begin{tabular}{c}
               File editing of \ZENF{site.h} \\
               in \ZENF{zmake} \end{tabular}}
           
           \psline{->}(2,-3)(2,-8)
           
           \psframe[shadow=true](-2,-9)(6,-8)
           \rput(2,-8.5){Library compilation}
           \psframe(7,-9)(11,-8)
           \rput(9,-8.5){\ZENF{make}}
           
           \psline[linestyle=dashed]{->}(2,-9)(2,-10)
           
           \psframe[shadow=true](-2,-11)(6,-10)
           \rput(2,-10.5){Cleaning unnecessary files}
           \psframe(7,-11)(11,-10)
           \rput(9,-10.5){\ZENF{make clean}}
           
           \psline[linestyle=dashed]{->}(2,-11)(2,-12)
           
           \psframe[shadow=true](-2,-13)(6,-12)
           \rput(2,-12.5){Removing all compiled files including library}
           \psframe(7,-13)(11,-12)
           \rput(9,-12.5){\ZENF{make tidy}}
         \end{pspicture*}
         
         
       \end{center}
       \caption{Compilation of the library.}
       \label{fig:compil}
     \end{figure}

%     \Section{Known problems}
%
%     We are aware of some problems in installing the library. How to
%     fix these problems can be found in the file |bugs.txt| on ftp
%     servers together with the package.
%
%     See also the note~\ref{zbncompil}, page~\pageref{zbncompil}, if
%     you have memory problems on your computer.

     \Section{Customized memory allocation functions}
     \input{memalloc}

     \input{biblio}
}
*/
