/*-*-LaTeX-*-
 * zen/inputs/intro.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
  This is a documentation file of the ZEN library.
  There is no C code in this file, only LaTeX commands.

\TEX{

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Introduction %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  Many computational problems need arithmetic operations in polynomial
  finite rings of $\Z/n\Z$ where $n$ is an integer ($n>1$).  Integer
  factorizations, primality testing are for instance such
  applications.  To solve them, programmers use general symbolic
  mathematical softwares or write specific programs (most of the time
  in C\cite{}).

  On the first hand, symbolic mathematical softwares (Maple\cite{},
  Mathematica\cite{},\ldots) handle with difficulty computations in
  finite fields. In the worst cases, such programs perform
  computations with rationals before finally reducing the objects
  modulo the characteristic $n$, in the best cases, such reductions
  are performs but extensions of a finite ring cannot be implemented.
  In any cases, applications written with such softwares are ten to
  hundred times slower than an ``ad hoc'' implementation in C.  On the
  other hand, optimized C libraries (|CESAR|, |Lidia|) deal only with
  one side of finite fields, mainly $\Z/n\Z$.

  We hardly believe we can keep the efficiency of these C libraries
  while working in any polynomial extension of $\Z/n\Z$.  We designed
  the |ZEN| library to perform efficient arithmetic operations in
  these sets.
  
  Via oriented object concepts programmed in C, you can work in the
  same way not only in any polynomial extension of $\Z/n\Z$, but also
  in any polynomial extension of another finite ring even if $n$ is
  not a prime or even if the polynomial which defines an extension is
  not irreducible. The current finite ring is an argument of any
  procedure of this library.  So, once a program is written for a
  given finite field, for instance $\Z/2\Z$, only few minor changes
  will be necessary to make it work in other finite fields, for
  instance $\Z/5\Z$, $\Z/18446744073709551629\Z$,
  $(\Z/2\Z)[t]/({t}^{10}+{t}^{3}+1)$,
  $((\Z/1753\Z)[t]({t}^{2}+7))[u]/(u^2+t)$,\ldots or even in rings,
  for instance in $\Z/1024\Z$, $(\Z/2\Z)[t]/({t}^{10}+1)$,\ldots In
  this later case, the functions of |ZEN| still works but exceptions
  are raised if inverses cannot be computed.

  To combine simplicity and efficiency, the procedures needed to
  handle elements, polynomials, matrices,\ldots are set at the running
  time while initializing the current ring because the computer
  structure of these objects depends on this ring. So, even if the
  syntax of our functions is always the same, the procedures
  dynamically called are functions of the ring.

  Two other original features of |ZEN| are the ``precomputation'' and
  ``clone'' concepts. In practice, it well known we can improve some
  algorithms at the expense of precomputations. Nevertheless,
  precomputations can take time! In |ZEN|, no precomputation is done
  by default but the user can perform some to speed up
  multiplications, exponentiations,\ldots. To that end, |ZEN| provides
  a procedure whose arguments are a ring and a flag to control
  the way precomputations are added to the ring. 

  It is well known too you can increase performances by changing the
  way you represent elements of a finite fields. A ring is defined by
  default by its polynomial basis in |ZEN|. But, some operations can be
  significantly speeded up if you change its representation. For
  instance, if you are working in a finite field, you can look for a
  generator and then working in the set of indexes. In |ZEN|, these
  representations are called ``clones'' and you can compute clones of
  any rings.  To that end, |ZEN| provides a procedure whose arguments
  are a ring and a flag to control which clone you want to
  compute. Obviously, you can then add precomputations to a clone.

  This library can be used at two levels.
   \begin{enumerate}
   \item For a current usage, the functions of the library
     can be used to perform operations on elements, polynomials,
     matrices, series and elliptic curves over every polynomial
     extension over $\Z/n\Z$ as described previously. Here, you only
     have to include |zen.h| and use the types and the functions of
     |ZEN| in your C sources and link your object files with the
     library |libzen.a|. These functionalities are described in this
     manual.
   
   \item For advanced users who really need to gain more efficiency,
     it is possible to replace procedure calls by macros in their own
     functions. But, such users have to know the internal data
     structures of the library and to write specific functions for
     each structure if they still want to handle any finite ring.
     These users are then sure that their applications are not
     penalized by inopportune procedure calls.
   \end{enumerate}

   Parts of this library was developped during the thesis of the
   authors~\cite{Flo,Rey}.
} */
