/*-*-LaTeX-*-
 * zen/inputs/maths.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
  This is a documentation file of the ZEN library. There is
  no C code in this file, only LaTeX commands.


   \TEX{
   We here suppose the reader familiar with the simple notions of algebra such as set, group, ring, field, equivalence relations, equivalence class, etc... The purpose of this section is to recall the construction of a finite field, and to make the parallel with the library functions. The detailed definitions of the notions used in this section can be found in any first degree algebra course.

   \Section{Modular rings}

   The basis of every construction of a finite field is a prime field $\quotient{\Z}{p\Z}$, with $p$ a prime integer.
   This field is the set of integers modulo $p$.
   More generally, for all integer $n$, the set of integers modulo $n$ is a finite ring.

   In C language, mathematical objects are represented by types. The \zen\ library therefore defines some types corresponding to the mathematical objects. THe first of these types is the |ZENRing| one which stands for a ring.
   In order to build a C representation |R| of the finite ring $R=\quotient{\Z}{n\Z}$, we need first to declare the variable |ZENRing R;| and then to build it using a C representation of $n$.
   As \zen\ is based upon |BigNum|, we use this representation for integers. Hence, we will need a couple |BigNum n; BigNumLength nl;| to represent $n$. We will describe more precisely what a |BigNum| is later, but for the moment, we can use the following function to set |n| to the value $n$: |ZBNReadFromString(&n,&nl,"1234567",10)|. This makes the couple |(n,nl)| represent $n=1234567$ in base 10.
   Now we can build a representation of $R=\quotient{\Z}{1234567\Z}$ using the intended function of \zen: |ZENBaseRingAlloc(R,n,nl)|.

   Once the ring initialized, one can use the mathematical objects defined on it, using the same type of programmation.
   For instance an element will be declared as a |ZENElt E;|, and allocated using |ZENEltAlloc(E,R)|.
   You can use this element in \zen\ functions, for instance set it to one: |ZENEltSetToOne(E,R)|.
   For efficiency reasons, there is no garbage collector in \zen. Hence, it is necessary to free the objects after use. For an element, the function to use is |ZENEltFree(E,R)|, and for a ring |ZENRingClose(R)|.

\Section{Ring extensions}
Let's recall that if a set of
cardinality $n$ is a finite field, then $n$ is a power of a prime
$n=p^m$, and there exists isomorphic representations of this set using
an irreducible polynomial $P$ over $\quotient{\Z}{p\Z}$.  More
generally, the quotient structure $\quotient{\quotient{\Z}{n\Z}[X]}{P(X)}$ is a finite field if and only if $n$ is prime and $P$ is irreducible.

One of the advantage of \zen, is to allow
easy use of polynomial extensions.
For instance, if we want to define a polynomial extension over the above |ZENRing R|, we will need to declare a polynomial |ZENPoly P|, define it over |R|, for instance with the function |ZENPolyReadFromString|, and use it to build the extension |ZENRing R2| with |ZENRingExtAlloc(R2,P,R)|.

Now the main advantage of \zen\ is that there is no change in the syntax of functions whether you work in a modular ring, an extension or a tower of extensions.
It is therefore possible to write generic programs, and choose the ring or field of definition dynamically when they are used, even if the mathematical structure differs.

\Section{A programming example}
As this is certainly the best way to understand how to make programs that use \zen, here is a small example.
A more complex and more interesting example is given in the \zenfact\ documentation.
     \input{example}

} */
