/*
 * zen/prgm/c/zenbench.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zen.h"

static char verbose;
extern char *optarg;
extern int optind, opterr;

#define RG_MAX 1000
#define NB_TESTED 1989		/* 1989 */
#define DEGREE_A 20
#define DEGREE_B 10

static ZENElt *ZENEltRandomArray(n, Rg)
     int n;
     ZENRing Rg;
{
  int i;
  ZENElt *a;

  a = (ZENElt *)ZENMalloc(sizeof(ZENElt)*n);
  for(i = 0; i < n; i++) {
    ZENEltCreate(a[i], Rg);
    ZENEltSetRandom(a[i], Rg);
  }
  return(a);
}

static ZENPoly *ZENPolyRandomArray(n, deg, Rg)
     int n, deg;
     ZENRing Rg;
{
  int i;
  ZENPoly *a;

  a = (ZENPoly *)ZENMalloc(sizeof(ZENPoly)*n);
  for(i = 0; i < n; i++) {
    a[i] = ZENPolyCreate(deg, Rg);
    ZENPolySetRandom(a[i], deg, Rg);
  }
  return(a);
}

static void ZENEltFreeArray(a, n, Rg)
     ZENElt *a;
     int n;
     ZENRing Rg;
{
  int i;

  for(i = 0; i < n; i++) {
    ZENEltFree(a[i], Rg);
  }
  ZENFree(a);
}

static void ZENPolyFreeArray(a, n, Rg)
     ZENPoly *a;
     int n;
     ZENRing Rg;
{
  int i;

  for(i = 0; i < n; i++) {
    ZENPolyFree(a[i], Rg);
  }
  ZENFree(a);
}

static char *Read(fin)
     FILE *fin;
{char *str,*str2;
 char line[1024];
 
 str = (char *)ZENMalloc(2*sizeof(char));
 sprintf(str,"\\");
 for(;;)
   {
     fscanf(fin, "%s", line);
     str2=str;
     str=(char *)ZENMalloc((strlen(str)+strlen(line))*sizeof(char));
     sprintf(str,"%s",str2);
     ZENFree(str2);
     sprintf(str+strlen(str)-1,"%s",line);
     if(str[strlen(str)-1]!='\\')
       break;
   }
 return(str);
}

/***********/

static ZENElt *ZENEltInitRandom(nb,R)
     int nb;
     ZENRing R;
{int i;
 ZENElt *ret;

 ret=(ZENElt *)(ZENMalloc(nb*sizeof(ZENElt)));
 
 if(!ret)
   exit(-1);
 
 for(i=0;i<nb;i++)
   {
     if(ZENEltCreate(ret[i],R))
       {
	 ZENError();
	 exit(-1);
       }
     ZENEltSetRandom(ret[i],R);
   }

 return(ret);
}

/***********/

static void ZENEltCloseRandom(ret,nb,R)
     ZENElt *ret;
     int nb;
     ZENRing R;
{int i;

 for(i=0;i<nb;i++)
   ZENEltFree(ret[i],R);

 ZENFree(ret);
}

/***********/

static BigNum *ZBNInitRandom(R,nb)
     int nb;
     ZENRing R;
{int i;
BigNumLength nl,bufl;
BigNum *ret,buf;

 nl=ZENRingSizeElt(R);
 ret=(BigNum *)(ZENMalloc(nb*sizeof(BigNum)));
 
 if(!ret)
   exit(-1);

 for(i=0;i<nb;i++)
   {
     ret[i]=ZBNC(nl);
     if(!ret[i])
       {
	 ZENError();
	 exit(-1);
       }
     ZBNSetRandom(ret[i],nl);
     bufl=ZBNModulo(&buf,ret[i],nl,ZENRingP(R),ZENRingPl(R));
     ZBNAssign(ret[i],buf,bufl);
     ZBNSetToZero(ret[i]+bufl,nl-bufl);
     ZBNF(buf);
   }

 return(ret);
}

/***********/

static BigNum *ZBNInitZero(nl,nb)
     int nb,nl;
{int i;
 BigNum *ret;

 ret=(BigNum *)(ZENMalloc(nb*sizeof(BigNum)));
 
 if(!ret)
   exit(-1);
 
 for(i=0;i<nb;i++)
   {
     ret[i]=ZBNC(nl);
     if(!ret[i])
       {
	 ZENError();
	 exit(-1);
       }
     ZBNSetToZero(ret[i],nl);
   }

 return(ret);
}

/***********/

static void ZBNCloseRandom(ret,nb)
     BigNum *ret;
     int nb;
{int i;

 for(i=0;i<nb;i++)
   ZBNF(ret[i]);

 ZENFree(ret);
}

/***********/

static void BenchMultiply(nb,R,Ropt)
     int nb;
     ZENRing R,Ropt;
{BigNum *am,*bm,*cm;
 double tt;
 int i;
 
 /* BigNum Multiplications */
 am=ZBNInitRandom(R,nb);
 bm=ZBNInitRandom(R,nb);
 cm=ZBNInitZero(2*ZENRingPl(R),nb);

 if(ZENPrcInitEltMultiplyPtr(Ropt) && (ZepRingKaraMLim(Ropt)>0))
   {
     printf("%d ZBNMultiplyKaratsuba(%d) : \n", nb, ZepRingKaraMLim(Ropt));
     tt = runtime();
     for(i = 0; i < nb; i++) {
       if (verbose) {
	 ZBNPrintToFile(stdout, am[i], ZENRingPl(R), 10);
	 printf(" * ");
	 ZBNPrintToFile(stdout, bm[i], ZENRingPl(R), 10);
	 printf(" -> ");
       }
       
       ZBNMultiplyKaratsubaBuffer(cm[i],am[i],ZENRingPl(R),bm[i],ZENRingPl(R),ZepRingKaraBuf(Ropt));
	    
       if (verbose) {
	 ZBNPrintToFile(stdout, cm[i], 2*ZENRingPl(R), 10);
	 printf(";\n");
       }
     }
     printf("Time = %2f ms/ops\n", (runtime()-tt)/nb);	  
   }
 
 printf("%d ZBNMultiply : \n", nb);
 tt = runtime();
 for(i = 0; i < nb; i++) {
   if (verbose) {
     ZBNPrintToFile(stdout, am[i], ZENRingPl(R), 10);
     printf(" * ");
     ZBNPrintToFile(stdout, bm[i], ZENRingPl(R), 10);
     printf(" -> ");
   }
   
   ZBNMultiply(cm[i],2*ZENRingPl(R),am[i],ZENRingPl(R),bm[i],ZENRingPl(R));
   
   if (verbose) {
     ZBNPrintToFile(stdout, cm[i], 2*ZENRingPl(R), 10);
     printf(";\n");
   }
 }
 
 printf("Time = %2f ms/ops\n", (runtime()-tt)/nb);	  

 ZBNCloseRandom(am,nb);
 ZBNCloseRandom(bm,nb);
 ZBNCloseRandom(cm,nb);
}
/***********/

static void BenchSquare(nb,R,Ropt)
     int nb;
     ZENRing R,Ropt;
{BigNum *am,*cm;
 double tt;
 int i;
 
 /* BigNum Multiplications */
 am=ZBNInitRandom(R,nb);
 cm=ZBNInitZero(2*ZENRingPl(R),nb);

 if(ZENPrcInitEltMultiplyPtr(Ropt) && (ZepRingKaraSLim(Ropt)>0))
   {
     printf("%d ZBNSquareKaratsuba(%d) : \n", nb, ZepRingKaraSLim(Ropt));
     tt = runtime();
     for(i = 0; i < nb; i++) {
       if (verbose) {
	 ZBNPrintToFile(stdout, am[i], ZENRingPl(R), 10);
	 printf(" ^ 2 -> ");
       }
       
       ZBNSquareKaratsubaBuffer(cm[i],am[i],ZENRingPl(R),ZepRingKaraBuf(Ropt));
	    
       if (verbose) {
	 ZBNPrintToFile(stdout, cm[i], 2*ZENRingPl(R), 10);
	 printf(";\n");
       }
     }
     printf("Time = %2f ms/ops\n", (runtime()-tt)/nb);	  
   }
 
 printf("%d ZBNSquare : \n", nb);
 tt = runtime();
 for(i = 0; i < nb; i++) {
   if (verbose) {
     ZBNPrintToFile(stdout, am[i], ZENRingPl(R), 10);
     printf(" ^ 2 -> ");
   }
   
   ZBNSquare(cm[i],2*ZENRingPl(R),am[i],ZENRingPl(R));
   
   if (verbose) {
     ZBNPrintToFile(stdout, cm[i], 2*ZENRingPl(R), 10);
     printf(";\n");
   }
 }
 
 printf("Time = %2f ms/ops\n", (runtime()-tt)/nb);	  
 
 ZBNCloseRandom(am,nb);
 ZBNCloseRandom(cm,nb);
}
/***********/
static void BenchZENEltExp(nb,R,Ropt)
     int nb;
     ZENRing R,Ropt;
{int i;
 double tt;
 ZENElt *aE;
 BigNum *eE;
   

 aE=ZENEltInitRandom(nb,R);
 eE=ZBNInitRandom(R,nb);
  
 printf("%d ZENEltExp : \n ", nb);
 tt = runtime();

 for(i = 0; i < nb; i++) {
   if (verbose) {
     ZENEltPrintToFile(stdout, aE[i], 10, R);
     printf(" ^ ");
     ZBNPrintToFile(stdout, eE[i], ZENRingPl(R), 10);
     printf(" -> ");
   }
   ZENEltExp(aE[i], eE[i], ZENRingPl(R), aE[i], R);
   if (verbose) {
     ZENEltPrintToFile(stdout, aE[i], 10, R);
     printf(";\n");
   }
 }

 printf("Time = %.2f ms/ops\n", (runtime()-tt)/nb);

 printf("%d ZENEltExp (Ropt) : \n ", nb);
 tt = runtime();

 for(i = 0; i < nb; i++) {
   if (verbose) {
     ZENEltPrintToFile(stdout, aE[i], 10, R);
     printf(" ^ ");
     ZBNPrintToFile(stdout, eE[i], ZENRingPl(R), 10);
     printf(" -> ");
   }
   ZENEltExp(aE[i], eE[i], ZENRingPl(R), aE[i], Ropt);
   if (verbose) {
     ZENEltPrintToFile(stdout, aE[i], 10, R);
     printf(";\n");
   }
 }

 printf("Time = %.2f ms/ops\n", (runtime()-tt)/nb);

 ZENEltCloseRandom(aE,nb,R);
 ZBNCloseRandom(eE,nb);
}

/***********/
/***********/
/***********/
/***********/

int main(argc, argv)
     int argc;
     char *argv[];
{
  char *str;
  int i, nb, j, rg, da, db;
  ZENRing Rg[RG_MAX],Rgopt[RG_MAX];
  ZENElt aI;
  ZENPoly P;
  ZENPrc prc;
  ZENCln cln;
  BigNum n;
  BigNumLength nl;
  double tt;
  char mflag, sflag, Mflag, Pflag, Sflag, Iflag, Tflag, Eflag;
  int c;
  FILE *fin;

  setbuf(stdout, NULL);

  verbose = 0; nb = NB_TESTED; da = DEGREE_A; db = DEGREE_B;
  mflag = sflag = Mflag = Pflag = Sflag = Iflag = Eflag = Tflag = 0;
  fin = stdin;
  while ((c = getopt(argc, argv, "msPMSITEn:vf:D:d:")) != EOF)
    switch (c) {

      /* An input file */
    case 'f':
      if (!(fin = fopen(optarg, "r"))) {
	fprintf(stderr, "Can't open %s, reading data on stdin\n", optarg);
	fin = stdin;
      }
      break;

      /* verbose */
    case 'v':
      verbose = 1;
      break;

      /* How many time ? */
    case 'n':
      nb = atoi(optarg);
      break;

      /* Polynomial arithmetic */
    case 'P':
      Pflag = 1;
      break;

    case 'D': /* largest degree */
      sscanf(optarg, "%d", &da);
      break;
      
    case 'd': /* smallest degree */
      sscanf(optarg, "%d", &db);
      break;

      /* Multiplication */
    case 'm':
      mflag = 1;
      break;

    case 'M':			/* ZENEltMultiply */
      Mflag = 1;
      break;

      /* Squaring */
    case 's':
      sflag = 1;
      break;

    case 'S':
      Sflag = 1;
      break;

      /* Tracing */
    case 'T':
      Tflag = 1;
      break;

      /* Inverse */
    case 'I':
      Iflag = 1;
      break;

      /* Exponentiation */
    case 'E':
      Eflag = 1;
      break;
    }

    /* Read on stdin the irreducible polynomial defining ZEN($2^n$) */
    printf("Enter the integer n which defines Z/nZ, (1 for aborting; end it by '\\' if you need more lines):\n");
    str=Read(fin);
    if (ZBNReadFromString(&n, &nl, str, 10) < 0) {
      ZENError();
      fprintf(stderr, "Error while reading the characteristic\n");
      exit(-1);
    }
    ZENFree(str);
    if ((ZBNDigit(n, 0) == 1) && (nl == 1)) {
      exit(-1);
    }
    
    printf("\nInitializing the ring R0 = Z/");
    if (ZBNPrint(n, nl)) {
      ZENError();
      fprintf(stderr, "Error while printing a bignum\n");
      exit(-1);
    }
    printf("Z\n");
    tt = runtime();
    if (!(Rg[0] = ZENBaseRingCreate(n, nl))) {
      ZENError();
      fprintf(stderr, "Error while initializing this ring\n");
      exit(-1);
    }
    ZBNF(n);
    printf("Initialization of the ring R0 done in %.2f s\n",
	   (runtime()-tt)/1000.);
    tt = runtime();
#if 1
    ZENClnSetAll(cln);
#endif
    if (!(Rgopt[0] = ZENRingClone(Rg[0], cln))) {
      ZENError();
      fprintf(stderr, "Error while initializing optimized ring\n");
      exit(-1);
    }
    printf("Initialization of the ring Ropt0 done in %.2f s\n",
	   (runtime()-tt)/1000.);

    tt = runtime();
    ZENPrcSetAll(prc);
    ZENRingAddPrc(Rgopt[0],prc);
    printf("All precomputations for the optimized ring R0 done in %.2f s\n",
	   (runtime()-tt)/1000.);
#if 1
    for(rg = 1; rg < RG_MAX; rg++) {
#else
      {rg = 1;
#endif
      printf("Enter the polynomial that will define the extension, or new integer for chinese (1 for aborting; end it by '\\' if you need more lines):\n");
      str=Read(fin);
#if 1
      if ((ZENPolyReadFromString(P, str, 10, Rg[rg-1]) < 0) ||
	  (ZENPolyDeg(P, Rg[rg-1]) == 0)) {
	
	if (ZBNReadFromString(&n, &nl, str, 10) < 0) {
	  rg--;
	  ZENFree(str);
	  break;
	}
	ZENFree(str);
	
	if ((ZBNDigit(n, 0) == 1) && (nl == 1)) {
#if 1
	  rg--;
#else
	  if(rg == 1)
	    {
	      rg--;
	      break;
	    }

	  printf("\nInitializing the ring R%d = Chinese(R)\n",rg);
	  tt = runtime();
	  if (!(Rg[rg] = ZENChineseRingCreate(rg, Rg))) {
	    ZENError();
	    fprintf(stderr, "Error while initializing this ring\n");
	    exit(-1);
	  }
	  printf("Initialization of the ring R%d done in %.2f s\n",rg,
		 (runtime()-tt)/1000.);
	  printf("\nInitializing the ring Ropt%d = Chinese(Ropt)\n",rg);
	  tt = runtime();
	  if (!(Rgopt[rg] = ZENChineseRingCreate(rg, Rgopt))) {
	    ZENError();
	    fprintf(stderr, "Error while initializing this ring\n");
	    exit(-1);
	  }
	  printf("Initialization of the ring Ropt%d done in %.2f s\n",rg,
		 (runtime()-tt)/1000.);
	  tt = runtime();
	  ZENPrcSetAll(prc);
	  ZENRingAddPrc(Rgopt[rg],prc);
	  printf("All precomputations for the optimized ring R0 done in %.2f s\n",
		 (runtime()-tt)/1000.);
#endif
	  break;
	}
    
	printf("\nInitializing the ring R%d = Z/",rg);
	if (ZBNPrint(n, nl)) {
	  ZENError();
	  fprintf(stderr, "Error while printing a bignum\n");
	  exit(-1);
	}
	printf("Z\n");
	tt = runtime();
	if (!(Rg[rg] = ZENBaseRingCreate(n, nl))) {
	  ZENError();
	  fprintf(stderr, "Error while initializing this ring\n");
	  exit(-1);
	}
	ZBNF(n);
	printf("Initialization of the ring R%d done in %.2f s\n",rg,
	       (runtime()-tt)/1000.);
	tt = runtime();
	if (!(Rgopt[rg] = ZENRingClone(Rg[rg], cln))) {
	  ZENError();
	  fprintf(stderr, "Error while initializing optimized ring\n");
	  exit(-1);
	}
	printf("Initialization of the ring Ropt%d done in %.2f s\n",rg,
	       (runtime()-tt)/1000.);

	tt = runtime();
	ZENPrcSetAll(prc);
	ZENRingAddPrc(Rgopt[rg],prc);
	printf("All precomputations for the optimized ring R0 done in %.2f s\n",
	       (runtime()-tt)/1000.);
	continue;
      }

      ZENFree(str);
      printf("\nInitializing the ring R%d = R%d", rg, rg-1);
      printf("[X]/(P(X)) with P(X) = ");
      ZENPolyPrintToFile(stdout, P, 10, Rg[rg-1]);
      printf("\n");
      tt = runtime();
#else
      printf("I am going to compute a random polynomial of degree %d\n", atoi(str));
      P =ZENPolyCreate(atoi(str), Rg[rg-1]);
      ZENPolySetRandom(P, atoi(str), Rg[rg-1]);
      printf("I computed a random polynomial of degree %d\n", ZENPolyDeg(P, Rg[rg-1]));
#endif
      if (!(Rg[rg] = ZENExtRingCreate(P, Rg[rg-1]))) {
	ZENError();
	fprintf(stderr, "Error while initilizing the ring\n");
	exit(-1);
      }
      ZENPolyFree(P, Rg[rg-1]);
#if 1
      printf("Initialization of the ring R%d done in %.2f s\n", rg, (runtime()-tt)/1000.);
      tt=runtime();
      if (!(Rgopt[rg] = ZENRingClone(Rg[rg],cln))) {
	ZENError();
	fprintf(stderr, "Error while initilizing the ring\n");
	exit(-1);
      }
      printf("Initialization of the ring Ropt%d done in %.2f s\n", rg, (runtime()-tt)/1000.);

#else
      Rgopt[rg] = Rg[rg];

      tt = runtime();
      ZENPrcSetAll(prc);
      ZENRingAddPrc(Rgopt[rg],prc);
      printf("All precomputations for the optimized ring Ropt%d done in %.2f s\n", rg,
	     (runtime()-tt)/1000.);
#endif
    }

    
      
    if (mflag) 
      BenchMultiply(nb,Rg[rg],Rgopt[rg]);
    
    
    if (sflag)
      BenchSquare(nb,Rg[rg],Rgopt[rg]);
    
    
    if (Mflag) {
      if (Pflag) {
	int i, j, snb, q, r;
	ZENPoly *a, *b, c;

	/* Multiplications */
	snb = 1+(int)sqrt((double)nb);
	c = ZENPolyCreate(da+db, Rg[rg]);
	
	printf("Computing 2 random arrays of %d ZENPolys\n", snb);
	a = ZENPolyRandomArray(snb, da, Rg[rg]);
	b = ZENPolyRandomArray(snb, db, Rg[rg]);
	q = nb/snb; r = nb%snb;
	
	tt = runtime();
	printf("%d ZENPolyMultiply(degree %d * degree %d) : \n", nb, da, db);
	for(i = 0; i < q; i++) {
	  for(j = 0; j < snb; j++) {
	    ZENPolyMultiply(c, a[i], b[j], Rg[rg]);
	  }
	}
	for(j = 0; j < r; j++) {
	  ZENPolyMultiply(c, a[i], b[j], Rg[rg]);
	}
	printf("Time = %2f s\n", (runtime()-tt)/1000.);

	ZENPolyFreeArray(a, snb, Rg[rg]);
	ZENPolyFreeArray(b, snb, Rg[rg]);
	ZENPolyFree(c, Rg[rg]);
      }
      else {
	int i, j, snb, q, r;
	ZENElt *a, *b, c;

	/* Multiplications */
	snb = 1+(int)sqrt((double)nb);
	ZENEltCreate(c, Rg[rg]);
	
	printf("Computing 2 random arrays of %d ZENElts\n", snb);
	a = ZENEltRandomArray(snb, Rg[rg]);
	b = ZENEltRandomArray(snb, Rg[rg]);
	q = nb/snb; r = nb%snb;
	
	tt = runtime();
	printf("%d ZENEltMultiply : \n", nb);
	for(i = 0; i < q; i++) {
	  for(j = 0; j < snb; j++) {
	    ZENEltMultiply(c, a[i], b[j], Rg[rg]);
	  }
	}
	for(j = 0; j < r; j++) {
	  ZENEltMultiply(c, a[i], b[j], Rg[rg]);
	}
	printf("Time = %2f s\n", (runtime()-tt)/1000.);
	
	ZENEltFreeArray(a, snb, Rg[rg]);
	ZENEltFreeArray(b, snb, Rg[rg]);
	ZENEltFree(c, Rg[rg]);
      }
    }
      
    if (Sflag) {
      if (Pflag) {
	int i, j, snb, q, r;
	ZENPoly *a, c;

	/* Squarings */
	snb = 1+(int)sqrt((double)nb);
	c = ZENPolyCreate(2*da, Rg[rg]);
	
	printf("Computing 2 random arrays of %d ZENPolys\n", snb);
	a = ZENPolyRandomArray(snb, da, Rg[rg]);
	q = nb/snb; r = nb%snb;
	
	tt = runtime();
	printf("%d ZENPolySquare (degree %d ^ 2): \n", nb, da);
	for(i = 0; i < q; i++) {
	  for(j = 0; j < snb; j++) {
	    ZENPolySquare(c, a[j], Rg[rg]);
	  }
	}
	for(j = 0; j < r; j++) {
	  ZENPolySquare(c, a[j], Rg[rg]);
	}
	printf("Time = %2f s\n", (runtime()-tt)/1000.);

	ZENPolyFreeArray(a, snb, Rg[rg]);
	ZENPolyFree(c, Rg[rg]);
      }
      else {
	/* Squarings */
	
	int i, j, snb, q, r;
	ZENElt *a;
	
	snb = 1+(int)sqrt((double)nb);
	
	printf("Computing 1 random array of %d ZENElts\n", snb);
	a = ZENEltRandomArray(snb, Rgopt[rg]);
	q = nb/snb; r = nb%snb;
	
	tt = runtime();
	printf("%d ZENEltSquare : \n", nb);
	for(i = 0; i < q; i++) {
	  for(j = 0; j < snb; j++) {
	    ZENEltSquare(a[i], a[i], Rgopt[rg]);
	  }
	}
	for(j = 0; j < r; j++) {
	  ZENEltSquare(a[i], a[i], Rgopt[rg]);
	}
	printf("Time = %2f s\n", (runtime()-tt)/1000.);
	
	ZENEltFreeArray(a, snb, Rgopt[rg]);
      }
    }
    
    if (Iflag) {
      if (Pflag) {
	int i, j, snb, q, r;
	ZENPoly *a, *b, quo, rem;

	/* Multiplications */
	snb = 1+(int)sqrt((double)nb);

	quo = ZENPolyCreate(da+db, Rg[rg]);
	rem = ZENPolyCreate(da+db, Rg[rg]);
	
	printf("Computing 2 random arrays of %d ZENPolys\n", snb);
	a = ZENPolyRandomArray(snb, da, Rg[rg]);
	b = ZENPolyRandomArray(snb, db, Rg[rg]);
	q = nb/snb; r = nb%snb;
	
	tt = runtime();
	printf("%d ZENPolyDivide (degree %d / degree %d) : \n", nb, da, db);
	for(i = 0; i < q; i++) {
	  for(j = 0; j < snb; j++) {
	    ZENPolyDivide(quo, rem, a[i], b[j], Rg[rg]);
	  }
	}
	for(j = 0; j < r; j++) {
	  ZENPolyDivide(quo, rem, a[i], b[j], Rg[rg]);
	}
	printf("Time = %2f s\n", (runtime()-tt)/1000.);

	ZENPolyFreeArray(a, snb, Rg[rg]);
	ZENPolyFreeArray(b, snb, Rg[rg]);
	ZENPolyFree(quo, Rg[rg]);
	ZENPolyFree(rem, Rg[rg]);
      }
      else {
	/* Inverses */
	
	tt = runtime();
	if (ZENEltCreate(aI , Rg[rg])) {
	  ZENError();
	  exit(-1);
	}
	
	printf("%d ZENEltInverse : \n", nb);
	for(i = 0; i < nb; i++) {
	  ZENEltSetRandom(aI, Rg[rg]);
	  if (verbose) {
	    ZENEltPrintToFile(stdout, aI, 10, Rg[rg]);
	    printf(" -> ");
	  }
	  ZENEltInverse(aI, aI, Rg[rg]);
	  if (verbose) {
	    ZENEltPrintToFile(stdout, aI, 10, Rg[rg]);
	    printf(";\n");
	  }
	}
	
	ZENEltFree(aI, Rg[rg]);
	printf("Time = %.2f s\n", (runtime()-tt)/1000.);
      }
    }
    
    if (Eflag)
      BenchZENEltExp(nb,Rg[rg],Rgopt[rg]);
    
    
    if (Tflag) {
      /* Traces */
      
      tt = runtime();
      if (ZENEltCreate(aI , Rg[rg])) {
	ZENError();
	exit(-1);
      }
      
      printf("%d ZENEltTrace : \ntr( ", nb);
      
      j=0;
      for(i = 0; i < nb; i++) {
	ZENEltSetRandom(aI, Rg[rg]);
	if (verbose) {
	  ZENEltPrintToFile(stdout, aI, 10, Rg[rg]);
	  printf(") -> ");
	}
	ZENEltTrace(aI, aI, Rg[rg]);
	if (verbose) {
	  ZENEltPrintToFile(stdout, aI, 10, Rg[rg]);
	  printf(" %d ;\n",j);
	}
      }
      
      ZENEltFree(aI, Rg[rg]);
      printf("Time = %.2f s\n", (runtime()-tt)/1000.);
    }
      
    /* Freeing memory */
    for(; rg >= 0; rg--) {
      printf("\nClosing the ring R%d\n", rg);
      tt = runtime();
      if (ZENRingClose(Rgopt[rg])) {
	ZENError();
	fprintf(stderr, "Error while closing optimized ring\n");
	exit(-1);
      }
      if (ZENRingClose(Rg[rg])) {
	ZENError();
	fprintf(stderr, "Error while closing this ring\n");
	exit(-1);
      }
      printf("These rings were succesfully closed in %.2f s\n", (runtime()-tt)/1000.);
    }

#if defined(i486icc) & defined(__DEBUG_ALLOC__)
  _dump_allocated(0);
#endif

  exit(0);
}
