/*
 * zen/prgm/c/zentest.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{ At the compilation time, a test program is compiled. To use
   it, you only have to type |bin/arch/zentest|. This program tests
   all the functions of the library (and so can take time ...).

   The five sets of procedures are tested. That is to say:
   \begin{itemize}
   \item Functions on elements.
   \item Functions on polynomials.
   \item Functions on matrices.
   \item Functions on series.
   \item Functions on elliptic curves.
   \end{itemize}
   The procedures of this program are not in |libzen.a| but can be
   good examples of procedures written with |ZEN| functions.  The same
   tests are performed for different finite rings following user
   options specified at the execution of the process. For a detailed
   description of these options, type |bin/arch/zentest -h|.
   } */
#include <stdio.h>
#include "zen.h"
#include "zef.h"		/* This inclusion is only for
				   the APPROXIMATE_RESULTS value */
#include "test.h"

#define FINITE_FIELD TRUE

extern void Usage ___(());
extern int ZENEltTest ___((ZENRing, int **, int));
extern int ZENPolyTest ___((ZENRing, int **, int));
extern int ZENMatTest ___((ZENRing, int **, int));

#include "zentest_data.h"


#define ConstantCopy(cst,data,nb_ln,nb_rg) if(1) \
 { cst=(int **)ZENMalloc(nb_ln*sizeof(int *));     \
   for(i=0;i<nb_ln;i++) {                       \
     cst[i]=(int *)ZENMalloc(nb_rg*sizeof(int));   \
     for(j=0;j<nb_rg;j++)                       \
       cst[i][j]=data[i][j];                    \
   } } else

int main(argc, argv)
     int argc;
     char *argv[];
{
  int **eltcst,**polycst,**matcst,**srcst,**eccst;
  char **def;
  int c, i, j, rg, nb_rg;
  int lib_flag, proc_flag;
  int zbnlib;
  int zen2lib, zen2lib0, zen2lib1, zen2lib2;
  int zepslib, zepslib0, zepslib1, zepslib2;
  int zetablib, zetablib0, zetablib1, zetablib2;
  int zeplib, zeplib0, zeplib1, zeplib2;
  int zeflib, zeflib0, zeflib1, zeflib2;
  int zeclib, zeclib0, zeclib1;
  int level[NB_RG];
  int eltproc, polyproc, matproc, srproc, ecproc;
  double tt, TT;
  BigNum n;
  BigNumLength nl;
  int first;
  ZENRing Rg[NB_RG], oldrg, RgClone;
  ZENPoly P1;
  ZENPrc prc;
  ZENCln cln;


  extern char *optarg;
  
  zbnsrandom(0);
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  eltcst = polycst = matcst = srcst = eccst = NULL;

  def = NULL;
  lib_flag = proc_flag = 1;

  zbnlib = 1;
  zen2lib = zen2lib0 = zen2lib1 = zen2lib2 = 1; 
  zepslib = zepslib0 = zepslib1 = zepslib2 = 1; 
  zetablib = zetablib0 = zetablib1 = zetablib2 = 1; 
  zeplib = zeplib0 = zeplib1 = zeplib2 = 1;
  zeflib = zeflib0 = !APPROXIMATE_RESULTS;
  zeflib1 = zeflib2 = 0;
  zeclib = zeclib0 = zeclib1 = 1;

  
  eltproc = polyproc = matproc = srproc = ecproc = 1;

  while ((c = getopt(argc, argv, "l:p:")) != EOF)
    switch (c) {
      
      /* The finite rings to test */
    case 'l':
      
      if (lib_flag) {
        lib_flag = 0;
	zbnlib = 0;
        zen2lib = zen2lib0 = zen2lib1 = zen2lib2 = 0;
        zepslib = zepslib0 = zepslib1 = zepslib2 = 0;
        zetablib = zetablib0 = zetablib1 = zetablib2 = 0;
        zeplib = zeplib0 = zeplib1 = zeplib2 = 0;
        zeflib = zeflib0 = zeflib1 = zeflib2 = 0;
        zeclib = zeclib0 = zeclib1 = 0;
      }
      switch(optarg[0]) {
      case 'z':
	zbnlib = 1;
	break;
      case 't':
        if (!strcmp(optarg, "t")) {
          zen2lib0 = zen2lib1 = zen2lib2 = 1;
          zen2lib = 1;
          break;
        }
        else if (!strcmp(optarg, "t0")) {
          zen2lib0 = 1;
          zen2lib = 1;
          break;
        }
        else if (!strcmp(optarg, "t1")) {
          zen2lib1 = 1;
          zen2lib = 1;
          break;
        }
        else if (!strcmp(optarg, "t2")) {
          zen2lib2 = 1;
          zen2lib = 1;
          break;
        }
        break;
        
      case 's':
        if (!strcmp(optarg, "s")) {
          zepslib0 = zepslib1 = zepslib2 = 1;
          zepslib = 1;
          break;
        }
        else if (!strcmp(optarg, "s0")) {
          zepslib0 = 1;
          zepslib = 1;
          break;
        }
        else if (!strcmp(optarg, "s1")) {
          zepslib1 = 1;
          zepslib = 1;
          break;
        }
        else if (!strcmp(optarg, "s2")) {
          zepslib2 = 1;
          zepslib = 1;
          break;
        }

      case 'a':
        if (!strcmp(optarg, "a")) {
          zetablib0 = zetablib1 = zetablib2 = 1;
          zetablib = 1;
          break;
        }
        else if (!strcmp(optarg, "a0")) {
          zetablib0 = 1;
          zetablib = 1;
          break;
        }
        else if (!strcmp(optarg, "a1")) {
          zetablib1 = 1;
          zetablib = 1;
          break;
        }
        else if (!strcmp(optarg, "a2")) {
          zetablib2 = 1;
          zetablib = 1;
          break;
        }

      case 'p':
        if (!strcmp(optarg, "p")) {
          zeplib0 = zeplib1 = zeplib2 = 1;
          zeplib = 1;
          break;
        }
        else if (!strcmp(optarg, "p0")) {
          zeplib0 = 1;
          zeplib = 1;
          break;
        }
        else if (!strcmp(optarg, "p1")) {
          zeplib1 = 1;
          zeplib = 1;
          break;
        }
        else if (!strcmp(optarg, "p2")) {
          zeplib2 = 1;
          zeplib = 1;
          break;
        }
      case 'f':
        if (!strcmp(optarg, "f")) {
          zeflib0 = 1;
	  zeflib1 = zeflib2 = 0;
          zeflib = 1;
          break;
        }
        else if (!strcmp(optarg, "f0")) {
          zeflib0 = 1;
          zeflib = 1;
          break;
        }
        else if (!strcmp(optarg, "f1")) {
          zeflib1 = 1;
          zeflib = 1;
          break;
        }
        else if (!strcmp(optarg, "f2")) {
          zeflib2 = 1;
          zeflib = 1;
          break;
        }
      case 'c':
        if (!strcmp(optarg, "c")) {
          zeclib0 = 1;
	  zeclib1 = 1;
          zeclib = 1;
          break;
        }
        else if (!strcmp(optarg, "c0")) {
          zeclib0 = 1;
          zeclib = 1;
          break;
        }
        else if (!strcmp(optarg, "c1")) {
          zeclib1 = 1;
          zeclib = 1;
          break;
        }
      default:
        Usage(argv[0]);
        exit(-1);
      }
      break;
      
      /* The type of procedures to test */
    case 'p':
      
      if (proc_flag) {
        proc_flag = 0;
        eltproc = polyproc = matproc = srproc = ecproc = 0;
      }
      switch(optarg[0]) {
      case 'e':
        eltproc = 1;
        break;
      case 'm':
        matproc = 1;
        break;
      case 'p':
        polyproc = 1;
        break;
      case 's':
        srproc = 1;
        break;
      case 'c':
        ecproc = 1;
        break;
      default:
        Usage(argv[0]);
        exit(-1);
      }
      break;

    case 'h':
    case '?':
      Usage(argv[0]);
      exit(-1);
    }
  /* Test loop */
#if ZEN_NO_TEST
  printf("Computing fast tests\n");
#else
  printf("Computing debugging tests\n");
#endif
  printf("===============\n");
  TT = runtime();

  if (zbnlib) {
   
    /* ZBN tests */
    printf("\n--------------------------\n");
    printf("Testing ZBN functions.\n");
    tt = runtime();
    if (ZENTestZBN()) {
      fprintf(stderr, "Error while testing ZBN operations\n");
      ZENError();
      exit(-1);
    }
    printf("\nZBN functions succesfully tested in %.2f s\n", (runtime()-tt)/1000.);
    printf("----------------------------------------------\n");
  }

  for(;;) {
    int ZEC_FLAG;
    if (zen2lib) {
      /* Rg = $Z/2Z$ */
      ZEC_FLAG=0;

      if ((level[0] = zen2lib0)) {nb_rg=1;}
      if ((level[1] = zen2lib1)) {nb_rg=2;}
      if ((level[2] = zen2lib2)) {nb_rg=3;}
      zen2lib = 0;
      nl = 1;

      ConstantCopy(      eltcst,ZENELT_ZE2_CSTS,NB_LN_ELT,NB_RG);
      ConstantCopy(      polycst,ZENPOLY_ZE2_CSTS,NB_LN_POLY,NB_RG);
      ConstantCopy(      matcst,ZENMAT_ZE2_CSTS,NB_LN_MAT,NB_RG);
      ConstantCopy(      srcst,ZENSR_ZE2_CSTS,NB_LN_SR,NB_RG);
      ConstantCopy(      eccst,ZENEC_ZE2_CSTS,NB_LN_EC,NB_RG);

      if (!(def = (char **)ZENMalloc(sizeof(char *)*NB_RG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < NB_RG; i++) {
        if (ZEN_ZE2_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZE2_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }
          strcpy(def[i], ZEN_ZE2_DEF[i]);
        }
        else {
          def[i] = NULL;
        }
      }
    }
    else if (zepslib) {
      /* Rg = $Z/pZ$, $p < 2^{BN_DIGIT_SIZE}$ */
      ZEC_FLAG=0;
            
      if ((level[0] = zepslib0)) {nb_rg=1;}
      if ((level[1] = zepslib1)) {nb_rg=2;}
      if ((level[2] = zepslib2)) {nb_rg=3;}
      zepslib = 0;
      nl = 1;

      ConstantCopy(      eltcst,ZENELT_ZEPS_CSTS,NB_LN_ELT,NB_RG);
      ConstantCopy(      polycst,ZENPOLY_ZEPS_CSTS,NB_LN_POLY,NB_RG);
      ConstantCopy(      matcst,ZENMAT_ZEPS_CSTS,NB_LN_MAT,NB_RG);
      ConstantCopy(      srcst,ZENSR_ZEPS_CSTS,NB_LN_SR,NB_RG);
      ConstantCopy(      eccst,ZENEC_ZEPS_CSTS,NB_LN_EC,NB_RG);

      if (!(def = (char **)ZENMalloc(sizeof(char *)*NB_RG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < NB_RG; i++) {
        if (ZEN_ZEPS_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZEPS_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }
          strcpy(def[i], ZEN_ZEPS_DEF[i]);
        }
        else {
          def[i] = NULL;
        }
      }       
    }
    else if (zeplib) {
      /* Rg = $Z/pZ$, $p < 2^{BN_DIGIT_SIZE}$ */
      ZEC_FLAG=0;
      
      if ((level[0] = zeplib0)) {nb_rg=1;}
      if ((level[1] = zeplib1)) {nb_rg=2;}
      if ((level[2] = zeplib2)) {nb_rg=3;}
      zeplib = 0;
      nl = 1;

      ConstantCopy(      eltcst,ZENELT_ZEP_CSTS,NB_LN_ELT,NB_RG);
      ConstantCopy(      polycst,ZENPOLY_ZEP_CSTS,NB_LN_POLY,NB_RG);
      ConstantCopy(      matcst,ZENMAT_ZEP_CSTS,NB_LN_MAT,NB_RG);
      ConstantCopy(      srcst,ZENSR_ZEP_CSTS,NB_LN_SR,NB_RG);
      ConstantCopy(      eccst,ZENEC_ZEP_CSTS,NB_LN_EC,NB_RG);

      if (!(def = (char **)ZENMalloc(sizeof(char *)*NB_RG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < NB_RG; i++) {
        if (ZEN_ZEP_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZEP_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }
          strcpy(def[i], ZEN_ZEP_DEF[i]);
        }
        else {
          def[i] = NULL;
        }
      }       
    }
    else if (zeflib) {
      /* Rg = $Z/pZ$, $p < 2^{BN_DIGIT_SIZE}$ */
      ZEC_FLAG=0;      

#if APPROXIMATE_RESULTS
      fprintf(stderr,"WARNING: APPROXIMATE_RESULTS may cause error in zentest \nbecause equality cannot be strictly defined.\n APPROXIMATE_LEVEL was set to %d\n",APPROXIMATE_LEVEL);
#endif
      if(zeflib1 || zeflib2)
	fprintf(stderr,"WARNING: Zef may cause errors when used in extension towers.\n");
      
      if ((level[0] = zeflib0)) {nb_rg=1;}
      if ((level[1] = zeflib1)) {nb_rg=2;}
      if ((level[2] = zeflib2)) {nb_rg=3;}
      zeflib = 0;
      nl = 1;

      ConstantCopy(      eltcst,ZENELT_ZEF_CSTS,NB_LN_ELT,NB_RG);
      ConstantCopy(      polycst,ZENPOLY_ZEF_CSTS,NB_LN_POLY,NB_RG);
      ConstantCopy(      matcst,ZENMAT_ZEF_CSTS,NB_LN_MAT,NB_RG);
      ConstantCopy(      srcst,ZENSR_ZEF_CSTS,NB_LN_SR,NB_RG);
      ConstantCopy(      eccst,ZENEC_ZEF_CSTS,NB_LN_EC,NB_RG);

      if (!(def = (char **)ZENMalloc(sizeof(char *)*NB_RG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < NB_RG; i++) {
        if (ZEN_ZEF_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZEF_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }
          strcpy(def[i], ZEN_ZEF_DEF[i]);
        }
        else {
          def[i] = NULL;
        }
      }       
    }
    else if (zetablib) {
      /* Rg = $Z/pZ$, $p < 2^{BN_DIGIT_SIZE}$ */
      ZEC_FLAG=0;
            
      if ((level[0] = zetablib0)) {nb_rg=1;}
      if ((level[1] = zetablib1)) {nb_rg=2;}
      if ((level[2] = zetablib2)) {nb_rg=3;}
      zetablib = 0;
      nl = 1;

      ConstantCopy(      eltcst,ZENELT_ZETAB_CSTS,NB_LN_ELT,NB_RG);
      ConstantCopy(      polycst,ZENPOLY_ZETAB_CSTS,NB_LN_POLY,NB_RG);
      ConstantCopy(      matcst,ZENMAT_ZETAB_CSTS,NB_LN_MAT,NB_RG);
      ConstantCopy(      srcst,ZENSR_ZETAB_CSTS,NB_LN_SR,NB_RG);
      ConstantCopy(      eccst,ZENEC_ZETAB_CSTS,NB_LN_EC,NB_RG);

      if (!(def = (char **)ZENMalloc(sizeof(char *)*NB_RG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < NB_RG; i++) {
        if (ZEN_ZETAB_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZETAB_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }
          strcpy(def[i], ZEN_ZETAB_DEF[i]);
        }
        else {
          def[i] = NULL;
        }
      }       
    }
    else if (zeclib) {
      /* Rg = $Z/pZ$, $p < 2^{BN_DIGIT_SIZE}$ */
      ZEC_FLAG=1;
      ecproc=0;

      if ((level[0] = zeclib0)) {nb_rg=1;}
      if ((level[1] = zeclib1)) {nb_rg=2;}
      zeclib = 0;
      nl = 1;

      ConstantCopy(      eltcst,ZENELT_ZEC_CSTS,NB_LN_ELT,ZEC_NB_RG);
      ConstantCopy(      polycst,ZENPOLY_ZEC_CSTS,NB_LN_POLY,ZEC_NB_RG);
      ConstantCopy(      matcst,ZENMAT_ZEC_CSTS,NB_LN_MAT,ZEC_NB_RG);
      ConstantCopy(      srcst,ZENSR_ZEC_CSTS,NB_LN_SR,ZEC_NB_RG);
      ConstantCopy(      eccst,ZENEC_ZEC_CSTS,NB_LN_EC,ZEC_NB_RG);

      if (!(def = (char **)ZENMalloc(sizeof(char *)*ZEC_NB_TOTRG))) {
        fprintf(stderr, "Malloc error\n");
        exit(-1);
      }
      for(i = 0; i < ZEC_NB_TOTRG; i++) {
        if (ZEN_ZEC_DEF[i]) {
          if (!(def[i] = (char *) ZENMalloc(sizeof(char)*(1+strlen(ZEN_ZEC_DEF[i]))))) {
            fprintf(stderr, "Malloc error\n");
            exit(-1);
          }
          strcpy(def[i], ZEN_ZEC_DEF[i]);
        }
        else {
          def[i] = NULL;
        }
      }       

    }
    else break;

    for(rg = 0; rg < nb_rg; rg++) {
      if(ZEC_FLAG)
	{
	  ZENRing PR[ZEC_NB_TOTRG];

	  switch(rg)
	    {
	    case 0:
	      for(i=0;i<ZEC_NB_BASE;i++)
		{
		  if (ZBNReadFromString(&n, &nl, def[i], 10) != strlen(def[i])) {
		    ZENError();
		    fprintf(stderr, "Error while reading a bignum\n");
		    exit(-1);
		  }
		  
		  printf("\nInitializing the ring PR%d = Z/",i);
		  if (ZBNPrint(n, nl)) {
		    ZENError();
		    fprintf(stderr, "Error while printing a bignum\n");
		    exit(-1);
		  }
		  printf("Z\n");
		  tt = runtime();
		  if (ZENBaseRingAlloc(PR[i], n, nl)) {
		    ZENError();
		    fprintf(stderr, "Error while initializing this ring\n");
		    exit(-1);
		  }
		  ZBNF(n);
		  tt=runtime()-tt;
		  tt /= (double)1000;
		  printf("Initialization of the ring PR%d done in %.2f s\n",i,tt);
		  ZENClnSetAll(cln);

		  printf("Computing a clone of the ring PR%d\n", i);
		  tt = runtime();
		  RgClone = ZENRingClone(PR[i], cln);
		  tt=runtime()-tt;
		  tt /= (double)1000;
		  printf("A clone of the ring PR%d done in %.2f s\n",i,tt);
		  PR[i]=RgClone;
		  /* Precomputations */
		  ZENPrcSetAll(prc);
#if FINITE_FIELD
		  ZENPrcSet(prc, ZENPRC_FINITE_FIELD);
#endif
		  printf("Precomputations for the ring PR%d\n", i);
		  tt = runtime();
		  ZENRingAddPrc(PR[i], prc);
		  printf("Precomputations of the ring PR%d done in %.2f s\n",
			 i, (runtime()-tt)/1000.);
		}
	      if (ZENRingChinese(Rg[0], ZEC_NB_BASE, PR)) {
		ZENError();
		fprintf(stderr, "Error while initializing this ring\n");
		exit(-1);
	      }	      
	      break;
	      
	    case 1:
	      i=ZEC_NB_BASE;
	      if (ZBNReadFromString(&n, &nl, def[i], 10) != strlen(def[0])) {
		ZENError();
		fprintf(stderr, "Error while reading a bignum\n");
		exit(-1);
	      }
		  
	      printf("\nInitializing the ring R0 = Z/");
	      if (ZBNPrint(n, nl)) {
		ZENError();
		fprintf(stderr, "Error while printing a bignum\n");
		exit(-1);
	      }
	      printf("Z\n");
	      tt = runtime();
	      if (ZENBaseRingAlloc(PR[0], n, nl)) {
		ZENError();
		fprintf(stderr, "Error while initializing this ring\n");
		exit(-1);
	      }
	      ZBNF(n);
	      printf("Initialization of the ring R0 done in %.2f s\n",
		     (runtime()-tt)/1000.);
	      for(i++;i<ZEC_NB_TOTRG;i++)
		{
		  printf("\nInitializing the ring PR%d = R0", i);
		  printf("[X]/(P(X)) with\nP(X) = %s\n", def[i]);
		  tt = runtime();
	      
		  if(ZENPolyReadFromString(P1, def[i], 10, PR[0]) < 0) {
		    ZENError();
		    fprintf(stderr, "Error while reading the definition polynomial\n");
		    exit(-1);
		  }
	      
		  if (ZENExtRingAlloc(PR[i], P1, PR[0])) {
		    ZENError();
		    fprintf(stderr, "Error while initilizing the ring\n");
		    exit(-1);
		  }
		  ZENPolyFree(P1,PR[0]);
		  printf("Initialization of the ring PR%d done in %.2f s\n",i,
			 (runtime()-tt)/1000.);
		  ZENClnSetAll(cln);

		  printf("Computing a clone of the ring PR%d\n", i);
		  tt = runtime();
		  RgClone = ZENRingClone(PR[i], cln);
		  printf("A clone of the ring PR%d done in %.2f s\n",
			 i, (runtime()-tt)/1000.);
		  PR[i]=RgClone;
		  /* Precomputations */
		  ZENPrcSetAll(prc);
#if FINITE_FIELD
		  ZENPrcSet(prc, ZENPRC_FINITE_FIELD);
#endif
		  printf("Precomputations for the ring PR%d\n", i);
		  tt = runtime();
		  ZENRingAddPrc(PR[i], prc);
		  printf("Precomputations of the ring PR%d done in %.2f s\n",
			 i, (runtime()-tt)/1000.);
		}
	      if (ZENRingChinese(Rg[1], ZEC_NB_EXT, PR+ZEC_NB_BASE+1)) {
		ZENError();
		fprintf(stderr, "Error while initializing this ring\n");
		exit(-1);
	      }	      
	      break;
	    }
	}
      else
	switch(rg)
	  {
	  case 0:
	    /* Initialization */
	    if (ZBNReadFromString(&n, &nl, def[0], 10) != strlen(def[0])) {
	      ZENError();
	      fprintf(stderr, "Error while reading a bignum\n");
	      exit(-1);
	    }
	    printf("\nInitializing the ring R0 = Z/");
	    if (ZBNPrint(n, nl)) {
	      ZENError();
	      fprintf(stderr, "Error while printing a bignum\n");
	      exit(-1);
	    }
	    printf("Z\n");
	    tt = runtime();
	    if (ZENBaseRingAlloc(Rg[0], n, nl)) {
	      ZENError();
	      fprintf(stderr, "Error while initializing this ring\n");
	      exit(-1);
	    }
	    ZBNF(n);
	    tt=runtime()-tt;
	    tt /= (double)1000;
	    printf("Initialization of the ring R0 done in %.2f s\n",tt);	    
	    break;
	    
	  default :
	    /* Put all the precomputation flags in prc */
	    ZENPrcSetAll(prc);
            
	    /* Initialization of the extension ring */
	    if (def[rg]) {
	      printf("\nInitializing the ring R%d = R%d", rg, rg-1);
	      printf("[X]/(P(X)) with\nP(X) = %s\n", def[rg]);
	      tt = runtime();
	      
	      if(ZENPolyReadFromString(P1, def[rg], 10, Rg[rg-1]) < 0) {
		ZENError();
		fprintf(stderr, "Error while reading the definition polynomial\n");
		exit(-1);
	      }
	      
	      if (ZENExtRingAlloc(Rg[rg], P1, Rg[rg-1])) {
		ZENError();
		fprintf(stderr, "Error while initilizing the ring\n");
		exit(-1);
	      }
	    }
	    else {
	      printf("\nInitializing the ring R%d = R%d", rg, rg-1);
	      printf("[X]/(P(X)) with degree(P(X)) = %d :\n", eltcst[DEG_ZEN_EXT][rg]);
	      tt = runtime();
	      if (!(P1 = ZENPolyCreate(eltcst[DEG_ZEN_EXT][rg], Rg[rg-1]))) {
		ZENError();
		fprintf(stderr, "Error while allocation P(X)\n");
		exit(-1);
	      }
	      ZENPolySetRandom(P1, eltcst[DEG_ZEN_EXT][rg], Rg[rg-1]);
	      ZENPolyPrintToFile(stdout,P1,10,Rg[rg-1]);
	      printf("\n");
	      
	      for(;;) {
		if (!ZENExtRingAlloc(Rg[rg], P1, Rg[rg-1])) 
		  break;
              
		if (ZENErrErrno != ZEN_NONMONIC) {
		  ZENError();
		  fprintf(stderr, "Error while initializing this ring\n");
		  exit(-1);
		}
		ZENPolySetRandom(P1, eltcst[DEG_ZEN_EXT][rg], Rg[rg-1]);
	      }
	    }
	    ZENPolyFree(P1, Rg[rg-1]);
	    printf("Initialization of the ring R%d done in %.2f s\n", rg, (runtime()-tt)/1000.);          
	  }
    
          
      for(first = 0;; first++) {	
	if (!first) {
	  if(ZEC_FLAG)
	    {
	      printf("Computing the modular ring of chinese ring R%d\n", rg);
	      RgClone=ZENRingCopy(ZENRingOrigin(Rg[rg]));
	      printf("A clone of the ring R%d done in %.2f s\n",
		     rg, (runtime()-tt)/1000.);
	    }
	  else
	    {
	      /* Clone */
	      ZENClnSetAll(cln);

	      printf("Computing a clone of the ring R%d\n", rg);
	      tt = runtime();
	      RgClone = ZENRingClone(Rg[rg], cln);
	      printf("A clone of the ring R%d done in %.2f s\n",
		     rg, (runtime()-tt)/1000.);
	    }
	  /* Precomputations */
	  ZENPrcSetAll(prc);
#if FINITE_FIELD
	  ZENPrcSet(prc, ZENPRC_FINITE_FIELD);
#endif
	  printf("Precomputations for the ring R%d\n", rg);
	  tt = runtime();
	  ZENRingAddPrc(Rg[rg], prc);
	  printf("Precomputations of the ring R%d done in %.2f s\n",
		 rg, (runtime()-tt)/1000.);
	}
	else if (first == 1) {
	  if (ZENRingArith(Rg[rg]) == ZENRingArith(RgClone)) {
	    printf("The clone is the same ring as R%d, I don't test it\n",
		   rg);

	    /* Freeing the clone */
	    printf("\nClosing the clone of ring R%d\n", rg);
	    tt = runtime();
	    if (ZENRingClose(RgClone)) {
	      fprintf(stderr, "Error while closing this ring\n");
	      ZENError();
	      exit(-1);
	    }
	    printf("This clone of ring was succesfully closed in %.2f s\n", (runtime()-tt)/1000.);
	    break;
	  }
	  oldrg = Rg[rg];
	  Rg[rg] = RgClone;
	}
	else if (first == 2) {

	  /* Freeing the clone */
	  printf("\nClosing the clone of ring R%d\n", rg);
	  tt = runtime();
	  if (ZENRingClose(RgClone)) {
	    fprintf(stderr, "Error while closing this ring\n");
	    ZENError();
	    exit(-1);
	  }
	  printf("This clone of ring was succesfully closed in %.2f s\n", (runtime()-tt)/1000.);
	  Rg[rg] = oldrg;
	  break;
	}
      
	if (level[rg] && eltproc) {
	  /* Element tests */
	  printf("\n--------------------------\n");
	  printf("Testing element functions.\n");
	  tt = runtime();
	  if (ZENTestElt(Rg[rg], eltcst, rg)) {
	    fprintf(stderr, "Error while testing element operations\n");
	    ZENError();
	    exit(-1);
	  }
	  printf("\nElement functions succesfully tested in %.2f s\n", (runtime()-tt)/1000.);
	  printf("----------------------------------------------\n");
	}
      
	if (level[rg] && polyproc) {
	  /* Polynomials tests */
	  printf("\n-----------------------------\n");
	  printf("Testing polynomial functions.\n");
	  tt = runtime();
	  if (ZENTestPoly(Rg[rg], polycst, rg)) {
	    fprintf(stderr, "Error while testing polynomial functions\n");
	    exit(-1);
	  }
	  printf("\nPolynomial functions succesfully tested in %.2f s\n", (runtime()-tt)/1000.);
	  printf("-------------------------------------------------\n");
	}    
	
	if (level[rg] && matproc) {
	  /* Matricial tests */
	  printf("\n-----------------------------\n");
	  printf("Testing matricial functions.\n");
	  tt = runtime();
	  if (ZENTestMat(Rg[rg], matcst, rg)) {
	    fprintf(stderr, "Error while testing matricial functions\n");
	    exit(-1);
	  }
	  printf("\nMatricial functions succesfully tested in %.2f s\n", (runtime()-tt)/1000.);
	  printf("-------------------------------------------------\n");
	}    
	
	if (level[rg] && srproc) {
	  /* Series tests */
	  printf("\n-----------------------------\n");
	  printf("Testing series functions.\n");
	  tt = runtime();
	  if (ZENTestSr(Rg[rg], srcst, rg)) {
	    fprintf(stderr, "Error while testing series functions\n");
	    exit(-1);
	  }
	  printf("\nSeries functions succesfully tested in %.2f s\n", (runtime()-tt)/1000.);
	  printf("-------------------------------------------------\n");
	}

	if (level[rg] && ecproc && !ZBNIsZero(ZENRingQ(Rg[rg]), ZENRingQl(Rg[rg]))) {
#if FINITE_FIELD
	  
	  /* Elliptic curve tests */
	  printf("\n-----------------------------\n");
	  printf("Testing elliptic curve functions.\n");
	  tt = runtime();
	  if (ZENTestEc(Rg[rg], eccst, rg)) {
	    fprintf(stderr, "Error while testing elliptic curves functions\n");
	    exit(-1);
	  }
	  printf("\nElliptic curve functions succesfully tested in %.2f s\n", (runtime()-tt)/1000.);
	  printf("-------------------------------------------------\n");
#else
	  printf("Elliptic curve functions are not checked,\nthis can only be tested in finite fields and the flag FINITE_FIELD was not set at the compilation time.\n");
#endif
	}
      }
    }
      
    for(rg--; rg >= 0; rg--) {
      
      /* Freeing the extension rings */
      printf("\nClosing the ring R%d\n", rg);
      tt = runtime();
      if(ZEC_FLAG)
	{
	  if (ZENRingClose(Rg[rg])) {
	    fprintf(stderr, "Error while closing this ring\n");
	    ZENError();
	    exit(-1);
	  }
	}
      else
	{
	  if (ZENRingClose(Rg[rg])) {
	    fprintf(stderr, "Error while closing this ring\n");
	    ZENError();
	    exit(-1);
	  }
	}
      printf("This ring was succesfully closed in %.2f s\n", (runtime()-tt)/1000.);
    }
    
    /* free constants */
    for(i = (ZEC_FLAG ? ZEC_NB_TOTRG : NB_RG); i--;) {
      if (def[i]) {	
        ZENFree(def[i]);
      }
    }
    ZENFree(def);
    if(eltcst) {
      for(i = 0; i < NB_LN_ELT; i++) {
	ZENFree(eltcst[i]);
      }
      ZENFree(eltcst);
      eltcst = NULL;
    }
    if(polycst) {
      for(i = 0; i < NB_LN_POLY; i++) {
	ZENFree(polycst[i]);
      }
      ZENFree(polycst);
      polycst = NULL;
    }
    if(matcst) {
      for(i = 0; i < NB_LN_MAT; i++) {
	ZENFree(matcst[i]);
      }
      ZENFree(matcst);
      matcst = NULL;
    }
    if(srcst) {
      for(i = 0; i < NB_LN_SR; i++) {
	ZENFree(srcst[i]);
      }
      ZENFree(srcst);
      srcst = NULL;
    }
    if(eccst) {
      for(i = 0; i < NB_LN_EC; i++) {
	ZENFree(eccst[i]);
      }
      ZENFree(eccst);
      eccst = NULL;
    }
  }
  printf("\nAll the tests correctly performed in %.2f s\n", (runtime()-TT)/1000.);

  
#if defined(i486icc) & defined(__DEBUG_ALLOC__)
  _dump_allocated(0);
#endif

  exit(0);
}

void Usage(arg)
     char *arg;
{
  printf("Usage: %s [-l(t|s)[val]] [-p(e|p|m)] [-help]\n", arg);
  printf("This program perform tests of the library zenq\n");
  printf("If no option is given, all the functions are tested\n");
  printf("-l <lib>: To perform tests only on function of the ring <lib>.\n");
  printf("<lib> can be,\n");
  printf("     z : for zbn.\n");
  printf("     t : for the subring Z/2Z.\n");
  printf("     s : for the subring Z/pZ, p small.\n");
  printf("     a : for tabulated clones.\n");
  printf("     p : for the subring Z/pZ.\n");
  printf("     f : for the ring Q.\n");
  printf("     c : for the chinese remainder construction.\n");
  printf(" [val] can be the level of extensions, ie 0, 1 or 2.\n");
  printf("-p <proc>: To perform tests only on the set of functions <proc>.\n");
  printf("<proc> can be,\n");
  printf("     e : for procedures on elements,\n");
  printf("     p : for procedures on polynomials.\n");
  printf("     m : for procedures on matrices.\n");
  printf("     s : for procedures on series.\n");
  printf("     c : for procedures on elliptic curves.\n");
}

/*
   Mathematical exceptions are verified with the following function which is not part of the library functions.
   
   \title{Test why a ZEN function returned |ZEN_NO_INVERSE|.}
   \input{A finite ring |Rg|.}
   \output{0 if |ZENRingFact(Rg)| was really a factor of a modulo, |ZENERR|
   otherwise.}

   */
int ZENTestEltNOINVCheck(Rg)
     ZENRing Rg;
{
#if 1
  printf("|");
#endif

  if (ZENRingPol(Rg)) {
    ZENPoly PX;
    BigNumLength kl;
    int ret;
    
    ZENEltToZ(ZENRingBNbuf(Rg),&kl,ZENRingFact(Rg),Rg);
    PX=ZENZToPoly(ZENRingBNbuf(Rg),kl,ZENRingDef(Rg));
    if (!PX) {
      return(ZERR);
    }
      
    if (ZENPolyDeg(PX,ZENRingDef(Rg))) {
      ZENPoly QX, RX;
      
      if (!(QX = ZENPolyCreate(ZENRingDeg(Rg), ZENRingDef(Rg)))) {
        ZENPolyFree(PX, ZENRingDef(Rg));
        return(ZERR);
      }

      if (!(RX = ZENPolyCreate(ZENRingDeg(Rg), ZENRingDef(Rg)))) {
        ZENPolyFree(PX, ZENRingDef(Rg));
        ZENPolyFree(QX, ZENRingDef(Rg));
        return(ZERR);
      }

      switch(ZENPolyDivide(QX, RX, ZENRingPol(Rg), PX, ZENRingDef(Rg))) {
      case ZENERR:
        ZENPolyFree(PX, ZENRingDef(Rg));
        ZENPolyFree(QX, ZENRingDef(Rg));
        ZENPolyFree(RX, ZENRingDef(Rg));
        return(ZERR);
      case ZEN_NO_INVERSE:
        if (ZENPolyDeg(QX, ZENRingDef(Rg))) {
	  ZENPolyFree(PX, ZENRingDef(Rg));
          ZENPolyFree(QX, ZENRingDef(Rg));
          ZENPolyFree(RX, ZENRingDef(Rg));
          return(ZERR);
        }
        ret = ZENTestEltNOINVCheck(ZENRingDef(Rg));
	ZENPolyFree(PX, ZENRingDef(Rg));
	ZENPolyFree(QX, ZENRingDef(Rg));
        ZENPolyFree(RX, ZENRingDef(Rg));
        return(ret);

      case ZEN_HAS_INVERSE:
        ZENPolyFree(PX, ZENRingDef(Rg));
	ZENPolyFree(QX, ZENRingDef(Rg));
        if (!ZENPolyIsZero(RX, ZENRingDef(Rg))) {
          ZENPolyFree(RX, ZENRingDef(Rg));
          return(ZERR);
        }
        ZENPolyFree(RX, ZENRingDef(Rg));
        return(0);
      }
    }
    else {
      ZENPolyFree(PX, ZENRingDef(Rg));
      return(ZENTestEltNOINVCheck(ZENRingDef(Rg)));
    }
  }
  else if (ZENRingPl(Rg) == 1) {
    BigNumLength al;
    
    ZENEltToZ(ZENRingBNbuf(Rg),&al,ZENRingFact(Rg),Rg);    
    if (!ZBNDigit(ZENRingBNbuf(Rg), 0)) {
      return(ZERR);
    }
    if (ZBNDigit(ZENRingP(Rg), 0)%ZBNDigit(ZENRingBNbuf(Rg),0)) {
      return(ZERR);
    }
    return(0);
  }
  else {
    BigNum g;
    BigNumLength al, gl;

    gl = ZENRingPl(Rg)+1;
    if (!(g = ZBNC(gl))) {
      return(ZERR);
    }
    ZBNAssign(g,  ZENRingP(Rg), ZENRingPl(Rg));
    ZBNDigit(g, ZENRingPl(Rg)) = 0;
    ZENEltToZ(ZENRingBNbuf(Rg),&al,ZENRingFact(Rg),Rg);    
    ZBNDivide(g, gl, ZENRingBNbuf(Rg), al);
    if (!ZBNIsZero(g, al)) {
      ZBNF(g);
      return(ZERR);
    }
    ZBNF(g);
    return(0);
  }
  return(ZERR);
}

