/*
 * zen/prgm/zmake.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "zmake.h"
#include "init.h"

char *H_FILES = "test";

char *FILES = "zbn_test elt_test mat_test poly_test sr_test ec_test";

char *ZENTEST = "zentest";

char *ZENBENCH = "zenbench";

int main(argc, argv)
     int argc;
     char *argv[];
{
  /*
   * Misc. variables
   ******************/
  char *GMP_LIB;
  char *ZEN_LIBRARY;
  char *ALL_FILES;

  /* 
   * Internal targets 
   ********************/
  int objtoar = 0;
  int zentestlink = 0; int zenbenchlink = 0;
  int clean_obj = 0; int clean_bin = 0;
  int clean_doc = 0; int clean_tex = 0; int clean_dvi = 0;
  int tidy_obj = 0; int tidy_bin = 0;
  int tidy_doc = 0;
  
  /* 
   * External targets
   *******************/
  int all = Target("all", argc, argv);

  int zentest = Target("zentest", argc, argv);
  int zenbench = Target("zenbench", argc, argv);

  int cc = Target("cc", argc, argv);

  int doc = Target("doc", argc, argv);
  int tex = Target("tex", argc, argv);
  int dvi = Target("dvi", argc, argv);

  int archtidy = Target("archtidy", argc, argv);
  int archclean = Target("archclean", argc, argv);
  int clean = Target("clean", argc, argv);
  int tidy = Target("tidy", argc, argv);

  /* 
   * Initialization
   *****************/
  zen_init(CURDIR, TOPDIR, ZCFDIR);
  zmake_init(ZEN_DEFINES, ZEN_INCLUDES, "", MathLibrary, argc, argv);

  ZEN_LIBRARY = concat(ZEN_LIB,
		       concat(PATHSEP,
			      concat(PREFIX,
				     concat(LIBZEN, ASUF))));

  ALL_FILES = concat(FILES,
		     concat(" ",
			    concat(ZENTEST,
				   concat(" ", ZENBENCH))));

#if USE_GMP
  GMP_LIB=" -lgmp";
#else
  GMP_LIB="";
#endif

  /* 
   * Targets
   ***********/

  /* default */
  TargetImply(defaults, all);

  /* all */
  TargetImply(all, zentest=zenbench);

  /* cc */
  TargetImply(zentest, cc=zentestlink);
  TargetImply(zenbench, cc=zenbenchlink);
  SrcToObjects(cc,
	       PRGM_OBJ, PREFIX,
	       PRGM_SRC, "", FILES,
	       ZEN_INC, "");

  /* zentest */
  SrcToObjects(zentestlink,
	       PRGM_OBJ, PREFIX,
	       PRGM_SRC, "", ZENTEST,
	       ZEN_INC, "");
  ObjectsToBinary(zentestlink,
		  PRGM_BIN, PREFIX, ZENTEST,
		  PRGM_OBJ, PREFIX, concat(FILES, concat(" ", ZENTEST)),
		  GMP_LIB, ZEN_LIBRARY);
  
  /* zenbench */
  SrcToObjects(zenbenchlink,
	       PRGM_OBJ, PREFIX,
	       PRGM_SRC, "", ZENBENCH,
	       ZEN_INC, "");
  ObjectsToBinary(zenbenchlink,
		  PRGM_BIN, PREFIX, ZENBENCH,
		  PRGM_OBJ, PREFIX, ZENBENCH,
		  GMP_LIB, ZEN_LIBRARY);

  /* doc */
  TargetImply(doc, dvi);
  TargetImply(dvi, tex);

  /* tex */
  SrcToTex(tex,
	   PRGM_TEX, "",
	   PRGM_SRC, "", ALL_FILES, CSUF,
	   PRGM_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   PRGM_TEX, "",
	   PRGM_INC, "", H_FILES, HSUF,
	   PRGM_BIN, "", IPT, PARSER);

  /* dvi */
  TexToDvi(dvi,
	   PRGM_DVI, "",
	   PRGM_TEX, "", ALL_FILES,
	   ZEN_INPUTS);
  TexToDvi(dvi,
	   PRGM_DVI, "",
	   PRGM_TEX, "", H_FILES,
	   ZEN_INPUTS);

  /* Cleaning */
  TargetImply(tidy, tidy_obj=tidy_bin=tidy_doc);
  TargetImply(archtidy, clean_obj=clean_doc=clean_bin);
  TargetImply(clean, tidy_obj=clean_tex);
  TargetImply(archclean, clean_obj=clean_tex);
  TargetImply(clean_doc, clean_dvi=clean_tex);
 
  CleaningRules(clean_obj, PRGM_OBJ, PREFIX, ALL_FILES, OSUF);
  CleaningRules(clean_bin, PRGM_BIN, PREFIX, ZENTEST, ESUF);
  CleaningRules(clean_bin, PRGM_BIN, PREFIX, ZENBENCH, ESUF);

  CleaningRules(clean_tex, PRGM_TEX, "", ALL_FILES, ".tex");
  TidyingRules(clean_tex, PRGM_TEX);
  CleaningRules(clean_dvi, PRGM_DVI, "", ALL_FILES, ".dvi");
  CleaningRules(clean_dvi, PRGM_DVI, "", H_FILES, ".dvi");
  CleaningRules(clean_tex|clean_dvi, PRGM_DOC, "", "", "");

  TidyingRules(tidy_obj, PRGM_OBT);
  TidyingRules(tidy_bin, PRGM_BNT);
  TidyingRules(tidy_doc, PRGM_DOC);

  /* 
   * Closing
   **********/
  zmake_close();

  exit(0);
}
