/*
 * zen/sys/c/error.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "sys.h"

zen_err zen_error = {ZEN_LIB_UNKNOWN, 0, 0};/*
   \TEX{
   When |OptimizingCode| is set to |NO| in file |specif.def|, the following flag is set by default :
   } 
   \CiteC */
int debugflag=1;		/* \EndCiteC */

/* The following definitions must be global for SunOS 4 cc compiler !!! Why ? (Florent) */

char *_zen_function[] = {
  "Unknown",			/* 0 */
  "Weight",
  "GcdPlain",
  "GcdBinary",
  "EeaPlain",
  "EeaLehmer",			/* 5 */
  "Gcd",
  "Eea",
  "PrintToString",
  "ReadFromString",
  "PutToString",		/* 10 */
  "GetFromString",
  "PrintToFile",
  "ReadFromFile",		
  "PutToFile",
  "GetFromFile",		/* 15 */
  "Init",
  "Close",
  "Modulo",
  "IsASquare",
  "SquareRoot",			/* 20 */
  "Inverse",
  "AddPrc",
  "ExpMakePrc",
  "MultiplyKaratsuba",
  "SquareKaratsuba",		/* 25 */
  "EltToZ",
  "ZToElt",
  "KaratsubaMultiplyCutoff",
  "KaratsubaSquareCutoff",
  "PrcInitKaratsuba",		/* 30 */
  "Alloc",
  "Sqrt",
  "PrcInitEltMultiply",
  "Factorial",
  "Convert",			/* 35 */
  "EltExp",
  "ZextPrcInitTrace",
    
  "PolyCreate",			/* 101 */
  "PolyPrintToString",
  "PolyReadFromString",
  "PolyPrintToFile",		
  "PolyReadFromFile",		/* 105 */
  "PolyPutToString",
  "PolyGetFromString",		
  "PolyPutToFile",
  "PolyGetFromFile",
  "PolyCreateFromFile",		/* 110 */
  "PolyToZ",
  "ZTOPoly",
    
  "MatAdd",			/* 201 */
  "MatNegate",
  "MatSubtract",
  "MatAssign",
  "_MatCreate",			/* 205 */
  "MatCreateParams",
  "MatGetBloc",
  "MatSetBloc",			
  "MatGetDigit",
  "MatSetDigit",		/* 210 */
  "MatMultiply",		
  "MatWinograd",
  "MatSetToZero",
  "MatSetToOne",
  "MatSetRandom",		/* 215 */
  "MatPermuteBn",
  "MatPermuteDigit",
  "MatMultiplyLeftPermutation",
  "MatMultiplyRightPermutation",
  "MatGaussPlain",		/* 220 */
  "MatGaussInverse",
  "MatInverse",
  "MatKernel",
  "MatPrintToString",
  "MatReadFromString",		/* 225 */
  "MatPrintToFile",		
  "MatReadFromFile",		
  "MatPutToString",
  "MatGetFromString",		
  "MatPutToFile",		/* 230 */
  "MatGetFromFile",
  "MatMultiplySameTypePermuteBn",
  "Ze2MatMultiplySameTypePermutationDigit",
  "Ze2MatMultiplyOppositeTypePermutationDigit",
  "MatGaussTrigInvertible",	/* 235 */
  "MatMultiplyPermutations",
  "MatRank",
  "MatSetSubMat",
  "MatGetSubMat",
  "MatConvert",			/* 240 */
  "Vect2Mat",
  "Mat2Vect",

  "PermutationCreate",		/* 301 */
  "PermutationAssign",
  "PermutationSetToZero",
  "PermutationSetToOne",
  "PermutationSetRandom",	/* 305 */
  "PermutationGetBloc",
  "PermutationSetDigit",
  "PermutationGetBloc",
  "PermutationSetDigit",
  "PermutationPermuteBn",	/* 310 */
  "PermutationIsOne",

  "SrCreate",			/* 401 */
  "SrPrintToString",
  "SrReadFromString",
  "SrPrintToFile",		
  "SrReadFromFile",		/* 405 */
  "SrPutToString",
  "SrGetFromString",		
  "SrPutToFile",
  "SrGetFromFile",
  "SrCreateFromFile",		/* 410 */

  "EcCreate",			/* 501 */
  "EcPtCreate",
  "EcPtPrintToString",
  "EcPtReadFromString",
  "EcPtPrintToFile",		/* 505 */	
  "EcPtReadFromFile",
  "EcPtPutToString",
  "EcPtGetFromString",		
  "EcPtPutToFile",
  "EcPtGetFromFile",		/* 510 */		
  "EcPtMultiply",
  "EcPtAdd"
};

char *_zen_message[] = {
  "Unknown error",		/* 0 */
  "Malloc error",		
  "Sprintf error",
  "Sscanf error",
  "Fprintf error",
  "Fscanf error",		/* 5 */
  "Syntax error",
  "Ring freed before its extensions",
  "Strcat error",
  "Non inversibble highest coefficient in the definition polynomial",
  "Matrix dimensions error",	/* 10 */
  "Uncorrect matrix dimension",
  "Permutation operation on non permutation matrix",
  "Assigning a permutation matrix with non permutation",
  "Uncorrect type of matrix",
  "Matrix should be invertible",	/* 15 */
  "Division by 0",
  "Too big ZENRing for tabulating",
  "This ZENRing is not an extension",
  "This arithmetic is unavailable",
  "Wrong input",		/* 20 */
  "Non inversible element",
  "This ZENRing is not a finite field"
};


char *_zen_library[] = {
  "Unknown",
  "ZBN", 
  "ZBZ",
  "ZE2",
  "ZEPS",
  "ZEXT",			/* 5 */
  "ZEG",
  "ZED",
  "ZEP",
  "ZETAB",
  "ZELOG",			/* 10 */
  "ZER",
  "ZEF",
  "ZEC",
  "ZEM"
};


/* \TEX{
   This flag activates automatic printing of error messages whenever a system  error occurs.
   This is useful for debugging small programs in which it is somehow pedantic to test all returned values.
   \title{Writing a message describing an error}
   \side{An error message is printed on the standard error output.}
   \note{The message depends on |zen_error|.}   
   }
   */
void ZENError ()
{  
  int fct, index;
  
  fct=zen_error.fct;
  index = 0;
  if (fct > ____ZEN_BASE) 
    index += ____ZEN_BASE - 100;
  
  if (fct > 100 + ____ZEN_POLYN) 
    index += ____ZEN_POLYN - 100;
  
  if (fct > 200 + ____ZEN_MAT) 
    index += ____ZEN_MAT - 100;
  
  if (fct > 300 + ____ZEN_PERM) 
    index += ____ZEN_PERM - 100;
  
  if (fct > 400 + ____ZEN_SR) 
    index += ____ZEN_SR - 100;
  
  if (fct > 500 + ____ZEN_EC) 
    index += ____ZEN_EC - 100;
  
  fct += index;
  
  fprintf(stderr,"%s (%s) : %s\n",
	  _zen_library[zen_error.lib],
	  _zen_function[fct],
	  _zen_message[zen_error.err]);
  /*  ZENSetError(ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_UNKNOWN);*/
}

/*
   \TEX{
   \title{Debugging function}
   \output{|ZENERR|}
   \side{If |debugflag| is non zero, prints an error message.}
   \note{This function can be used to set breakpoint in a debugger. It is never called when the library is compiled in optimized mode.}
   } 
   */
int _ZENERR()
{
  if(debugflag)
    ZENError();
  return(ZENERR);
}

/*
   \TEX{
   \title{Debugging function}
   \output{|ZENNULL| pointer.}
   \side{If |debugflag| is non zero, prints an error message.}
   \note{This function can be used to set breakpoint in a debugger. It is never called when the library is compiled in optimized mode.}
   } 
   */
void *_ZENNULL()
{
  if(debugflag)
    ZENError();
  return(ZENNULL);
}
