/*
 * zen/sys/c/rand.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "zbn.h"

/* \TEX{
   \Subsection{Random assignment}
   } */

/* \TEX{
   In the same spirit as for allocation functions, we allow customized
   random functions.
   } */

#if 0
/* \TEX{
   \title{Pseudo-Random Generator (PRG) initialization}
   \input{A string |seed| of length |seed_len| which serves as seed
   for the internal PRG.}
   \note{The PRG behavior is determined by this seed. The programs
   included in |ZEN| use this function to initialize the PRG. By
   default, this function calls the native |ZENNativeSrand|
   function. This can be overwritten by |ZENSetRandomFunctions|.}
   } */
void zsrandom(seed, seed_len)
     unsigned char *seed; 
     size_t seed_len; {}

/* \TEX{ 
   \title{Pseudo-Random Generation}
   \output{A pseudo-random |unsigned char|.}
   \note{This function is used all over the |ZEN| and |ZENFACT|
   libraries when random generation is needed. By default, this
   function calls the native |ZENNativeRand| function. This can be
   overwritten by |ZENSetRandomFunctions|.}
   } */
unsigned char zrandom(); {}
#endif

/* Internal RC4 variables */
static unsigned char RC4Table[256];
static unsigned char RC4i, RC4j;
unsigned char RC4Init = 0;

/* Internal RC4 functions */
/* \TEX{
   \title{Native random initialization}
   \input{A seed |seed| of length |seed_len| for the |ZENNativeRand|
   function.}
   } */
void ZENNativeSrand(seed, seed_len)
     unsigned char *seed;
     size_t seed_len;
{
  unsigned char i, j;
  unsigned char tmp;
  
  i = 0;
  do {
    RC4Table[i] = i;
    i++;
  } while(i);
  i = j = 0;
  do {
    tmp = i < seed_len ? seed[i] : 0;
    j += RC4Table[i] + tmp;
    
    tmp = RC4Table[i];
    RC4Table[i] = RC4Table[j]; 
    RC4Table[j] = tmp;

    i++;
  }  while(i);
  RC4i = 0;
  RC4j = 0;
  RC4Init = 1;
}

/* \TEX{
   \title{Native random function}
   \output{One pseudo-random |unsigned char|}
   \note{By now, this function is based on RC4.}
   } */
unsigned char ZENNativeRand()
{
  unsigned char tmp;

  if (!RC4Init) {ZENNativeSrand(NULL, 0);}
  
  RC4i++;
  RC4j += RC4Table[RC4i];
  
  tmp = RC4Table[RC4i];
  RC4Table[RC4i] = RC4Table[RC4j]; 
  RC4Table[RC4j] = tmp;
  
  tmp = RC4Table[RC4i]+RC4Table[RC4j];
  return(RC4Table[tmp]);
}

void          (* _zsrandom) ___((unsigned char *, size_t)) = ZENNativeSrand;
unsigned char (* _zrandom)  ___((void))                    = ZENNativeRand;

/* \TEX{
   \title{Setting custom pseudo-random functions}
   \input{An initialization function of the pseudo-random generator
   and the two random functions performing |srandom|, and |random|.}
   \output{If non-|NULL|, the initialization pointer function is
   called and its output is returned. Otherwise, the return output is
   0.}
   \side{The internal pseudo-random functions |zbnrandom| and
   |zbnsrandom| are customized.}
   \note{This function should be called at the very beginning of a
   program. At least, it should be called before the first call to
   |zbnsrandom|, as it will determine the subsequent behavior of
   |zbnrandom|.}
   } */
int ZENSetRandomFunctions(in,sr,ra)
     int (* in)           ___((void));
     void (* sr)          ___((unsigned char *, size_t));
     unsigned char (* ra) ___((void));
{
  int ret;

  if(in)
    ret=in();
  else
    ret=0;

  if(sr)
    _zsrandom=sr;
  if(ra)
    _zrandom=ra;

  return(ret);
}
