/*
 * zen/sys/c/runtime.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \def\srcfile{runtime.c}
   Accurate timings are necessary to compare the algorithms implemented
   in this package. This is done by the function |runtime()|.
   }
*/
#include <stdio.h>
#include <time.h>
#include <limits.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/param.h>
#ifndef HZ
# define HZ CLK_TCK
#endif

#include "sys.h"

/* \TEX{
   \title{Getting system times.}
   \output{The user time (in $10^{-3}$s) of the process that calls
   this function is returned.}
   } */
double runtime ()
{
#ifdef i486icc
  double time1;
  
  time1 = (double)clock();
  time1 = 1000*time1/CLOCKS_PER_SEC;
  
  return(time1);
#else
# ifdef SystemV4
  struct rusage r;
  struct timeval t;
  double ret;
  
  getrusage(0, &r);
  t = r.ru_utime;

  ret=1000.0* ((double)t.tv_sec) + ((double)t.tv_usec)/1000.0;
  return(ret);  
# else  
  struct tms t;
  double ret;
  
  times(&t);
  ret=t.tms_utime*((double)1000.)/HZ;
  return(ret);
# endif
#endif
}

