/*
 * zen/sys/zmake.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "zmake.h"
#include "init.h"

char *FILES = "memalloc error rand runtime";
char *H_FILES = "sys";

int main(argc, argv)
     int argc;
     char *argv[];
{
  /*
   * Misc. variables
   *****************/
  char *SYS_PREFIX;

  /* 
   * Internal targets 
   ********************/
  int objtoar = 0;
  int clean_obj = 0; int clean_lib = 0;
  int clean_doc = 0; int clean_tex = 0; int clean_dvi = 0;
  int tidy_lib = 0; int tidy_obj = 0; int tidy_bin = 0;
  int tidy_doc = 0;
  
  /* 
   * External targets
   *******************/
  int all = Target("all", argc, argv);

  int cc = Target("cc", argc, argv);

  int doc = Target("doc", argc, argv);
  int tex = Target("tex", argc, argv);
  int dvi = Target("dvi", argc, argv);

  int archtidy = Target("archtidy", argc, argv);
  int archclean = Target("archclean", argc, argv);
  int clean = Target("clean", argc, argv);
  int tidy = Target("tidy", argc, argv);

  /* 
   * Initialization
   *****************/
  zen_init(CURDIR, TOPDIR, ZCFDIR);
  zmake_init(ZEN_DEFINES, ZEN_INCLUDES, "", "", argc, argv);

  SYS_PREFIX = concat(PREFIX, PSYS);

  /* 
   * Targets
   ***********/

  /* default */
  TargetImply(defaults, cc);

  /* all */
  TargetImply(all, cc=doc);

  /* cc */
  TargetImply(cc, objtoar);
  SrcToObjects(cc,
	       SYS_OBJ, SYS_PREFIX,
	       SYS_SRC, "", FILES,
	       ZEN_INC, "");

  /* ar */
  ObjectsToArList(objtoar,
		  SYS_LIB, SYS_PREFIX, "", OBJ_TOAR,
		  SYS_OBJ, SYS_PREFIX, FILES);

  /* doc */
  TargetImply(doc, dvi);
  TargetImply(dvi, tex);

  /* tex */
  SrcToTex(tex,
	   SYS_TEX, "",
	   SYS_SRC, "", FILES, CSUF,
	   SYS_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   SYS_TEX, "",
	   ZMAKE_DIR, "", PARSER, HSUF,
	   SYS_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   SYS_TEX, "",
	   SYS_INC, "", H_FILES, HSUF,
	   SYS_BIN, "", IPT, PARSER);

  /* dvi */
  TexToDvi(dvi,
	   SYS_DVI, "",
	   SYS_TEX, "", FILES,
	   ZEN_INPUTS);
  TexToDvi(dvi,
	   SYS_DVI, "",
	   SYS_TEX, "", PARSER,
	   ZEN_INPUTS);
  TexToDvi(dvi,
	   SYS_DVI, "",
	   SYS_TEX, "", H_FILES,
	   ZEN_INPUTS);

  /* Cleaning */
  TargetImply(tidy, tidy_obj=tidy_lib=tidy_bin=tidy_doc);
  TargetImply(archtidy, archclean=clean_doc);
  TargetImply(clean, clean_lib=tidy_obj=clean_tex);
  TargetImply(archclean, clean_obj=clean_lib=clean_tex);
  TargetImply(clean_doc, clean_dvi=clean_tex);
  TargetImply(tidy_obj, clean_obj=clean_lib);
  
  CleaningRules(clean_obj, SYS_OBJ, SYS_PREFIX, FILES, OSUF);

  CleaningRules(clean_lib, SYS_LIB, "", SYS_PREFIX, OBJ_TOAR);

  TidyingRules(clean_tex, SYS_TEX);
  CleaningRules(clean_tex, SYS_BIN, "", PARSER, ESUF);

  CleaningRules(clean_dvi, SYS_DVI, "", FILES, ".dvi");
  CleaningRules(clean_dvi, SYS_DVI, "", PARSER, ".dvi");
  CleaningRules(clean_dvi, SYS_DVI, "", H_FILES, ".dvi");
  CleaningRules(clean_tex|clean_dvi, SYS_DOC, "", "", "");

  TidyingRules(tidy_obj, SYS_OBT);
  TidyingRules(tidy_lib, SYS_LBT);
  TidyingRules(tidy_bin, SYS_BNT);
  TidyingRules(tidy_doc, SYS_DOC);

  /* 
   * Closing
   **********/
  zmake_close();

  exit(0);
}
