/*
 * zen/zbn/c/add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

#ifdef USE_GMP
# include <gmp.h>
#endif

/* \TEX{
   \Subsection{Addition}   
   } */

/* \TEX{
   \title{Incrementation}
   \input{An allocated |BigNum| and its length, and a carry in that must
   be 0 or 1.}
   \output{The carry out.}
   \side{$|(m,ml)| += |carry|$}
   \note{If |ml| is zero, nothing is done and the returned value is |carry|.}
   } */
BigNumDigit ZBNAddCarry(m, ml, carry)
     BigNum m;
     BigNumLength ml;
     BigNumDigit carry;
{
  if (!carry) {return(CST_BNDIGIT(0));}
  if (ml) {
    m += ml;  ml *= CST_BNLENGTH(-1);
    do {
      if (++m[ml]) {break;}
      ml++;
    } while(ml);
  }
  return(ml ? CST_BNDIGIT(0) : CST_BNDIGIT(1));
}

#ifndef USE_GMP
/* \TEX{
   \title{Addition}
   \input{Two allocated |BigNum|s, their lengths and a carry in that
   must be be 0 or 1.}
   \output{The carry out}
   \side{$|(m,ml)| += |(n,nl)| + |carry|$.}
   \note{One must have $|nl| \le |ml|$.  If $|nl| = 0$, then
   |ZBNAdd(m,ml,n,0,c)| behaves like |ZBNAddCarry(m,ml,c)|.}
   } */
BigNumDigit ZBNAdd (m, ml, n, nl, carry)
     BigNum m;
     BigNumLength ml;
     BigNum n;
     BigNumLength nl;
     BigNumDigit carry;
{
  BigNumDigit l, c;

  m += nl;
  n += nl;
  nl *= CST_BNLENGTH(-1);
  ml += nl;

  switch((int)(nl %8)) {
  case -7:
    l = n[nl]; c = carry; carry = CST_BNDIGIT(0);
    zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
    zbnadd(carry, m[nl], carry, m[nl], CST_BNDIGIT(0), l); 
    nl++;
  case -6:
    l = n[nl]; c = carry; carry = CST_BNDIGIT(0);
    zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
    zbnadd(carry, m[nl], carry, m[nl], CST_BNDIGIT(0), l); 
    nl++;
  case -5:
    l = n[nl]; c = carry; carry = CST_BNDIGIT(0);
    zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
    zbnadd(carry, m[nl], carry, m[nl], CST_BNDIGIT(0), l); 
    nl++;
  case -4:
    l = n[nl]; c = carry; carry = CST_BNDIGIT(0);
    zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
    zbnadd(carry, m[nl], carry, m[nl], CST_BNDIGIT(0), l); 
    nl++;
  case -3:
    l = n[nl]; c = carry; carry = CST_BNDIGIT(0);
    zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
    zbnadd(carry, m[nl], carry, m[nl], CST_BNDIGIT(0), l); 
    nl++;
  case -2:
    l = n[nl]; c = carry; carry = CST_BNDIGIT(0);
    zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
    zbnadd(carry, m[nl], carry, m[nl], CST_BNDIGIT(0), l); 
    nl++;
  case -1:
    l = n[nl]; c = carry; carry = CST_BNDIGIT(0);
    zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
    zbnadd(carry, m[nl], carry, m[nl], CST_BNDIGIT(0), l); 
    nl++;
  }

  if (nl) {
    do {
      l = n[nl]; c = CST_BNDIGIT(0);
      zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
      zbnadd(c, m[nl], c, m[nl], CST_BNDIGIT(0), l);

      l = n[nl+1]; carry = CST_BNDIGIT(0);
      zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
      zbnadd(carry, m[nl+1], carry, m[nl+1], CST_BNDIGIT(0), l);

      l = n[nl+2]; c = CST_BNDIGIT(0);
      zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
      zbnadd(c, m[nl+2], c, m[nl+2], CST_BNDIGIT(0), l);

      l = n[nl+3]; carry = CST_BNDIGIT(0);
      zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
      zbnadd(carry, m[nl+3], carry, m[nl+3], CST_BNDIGIT(0), l);
      
      l = n[nl+4]; c = CST_BNDIGIT(0);
      zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
      zbnadd(c, m[nl+4], c, m[nl+4], CST_BNDIGIT(0), l);

      l = n[nl+5]; carry = CST_BNDIGIT(0);
      zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
      zbnadd(carry, m[nl+5], carry, m[nl+5], CST_BNDIGIT(0), l);

      l = n[nl+6]; c = CST_BNDIGIT(0);
      zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
      zbnadd(c, m[nl+6], c, m[nl+6], CST_BNDIGIT(0), l);

      l = n[nl+7]; carry = CST_BNDIGIT(0);
      zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
      zbnadd(carry, m[nl+7], carry, m[nl+7], CST_BNDIGIT(0), l);
      
      nl += 8;
    } while(nl);
  }

  if (ml && carry) {
    m += ml;  ml *= CST_BNLENGTH(-1);
    do {
      if (++m[ml]) {break;}
      ml++;
    } while(ml);
    carry = ml ? CST_BNDIGIT(0) : CST_BNDIGIT(1);
  }
  return(carry);
}
#else
BigNumDigit ZBNAdd (m, ml, n, nl, carry)
     BigNum m;
     BigNumLength ml;
     BigNum n;
     BigNumLength nl;
     BigNumDigit carry;
{
  return((nl) ? 
	 ((ml) <= (nl) ? 
	  (BigNumDigit)(mpn_add_n(m, m, n, ml)+
			mpn_add_1(m, m, ml, carry)) 
	  : 
	  (BigNumDigit)(mpn_add_1(m+nl, m+nl, ml-nl,
				  mpn_add_n(m, m, n, nl)+
				  mpn_add_1(m, m, nl, carry))))
	 : ZBNAddCarry(m, ml, carry));
    
}
#endif

/* \TEX{
   \Subsubsection{Subtraction}   
   } */

/* \TEX{
   \title{Additive inverse}
   \input{An allocated |BigNum|}
   \side{$|(n,nl)| \leftarrow 2^{|SIZE_BLOC|\times|nl|}-1-|(n,nl)|$} 
   \note{If |nl| equals zero, does nothing.}
   } */
void ZBNComplement(n, nl)
     BigNum n;
     BigNumLength nl;
{
  if (!nl) {return;}
  n += nl; nl *= CST_BNLENGTH(-1);
  do {
    n[nl] = ~n[nl];
    nl++;
  } while (nl);
}

/* \TEX{
   \title{Decrementation}
   \input{An allocated |BigNum|, its length and a borrow in that must
   be 0 or 1.}
   \output{The borrow out.}
   \side{$|(m,ml)| -= (1-|carry|)$}
   \note{If |ml| is zero, nothing is done and the return value is
   |carry|.}
   } */
BigNumDigit ZBNSubtractBorrow(m, ml, carry)
     BigNum m;
     BigNumLength ml;
     BigNumDigit carry;
{
  if (carry) {return(CST_BNDIGIT(1));}
  if (ml) {
    m += ml;  ml *= CST_BNLENGTH(-1);
    do {
      if (m[ml]--) {break;}
      ml++;
    } while(ml);
  }
  return(ml ? CST_BNDIGIT(1) : CST_BNDIGIT(0));
}

#ifndef USE_GMP
/* \TEX{
   \title{Subtraction}
   \input{Two allocated |BigNum|s, their length and a borrow in that
   must be be 0 or 1.}
   \output{The borrow out.}
   \side{$|(m,ml)| -= |(n,nl)| + (1-|carry|)$}
   \note{One must have $|nl|\le |ml|$. If $|nl|=0$, then
   |ZBNSubtract(m,ml,n,0,b)| behaves like
   |ZBNSubtractBorrow(m,ml,b)|.}
   } */
BigNumDigit ZBNSubtract(m, ml, n, nl, carry)
     BigNum m;
     BigNumLength ml;
     BigNum n;
     BigNumLength nl;
     BigNumDigit carry;
{
  BigNumDigit c, l;
  
  m += nl;
  n += nl;
  nl *= CST_BNLENGTH(-1);
  ml += nl;

  switch((int)(nl %8)) {
  case -7:
    l = ~n[nl]; c = CST_BNDIGIT(0);
    zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
    zbnadd(c, m[nl], c, m[nl], CST_BNDIGIT(0), l);
    carry = c;
    nl++;
  case -6:
    l = ~n[nl]; c = CST_BNDIGIT(0);
    zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
    zbnadd(c, m[nl], c, m[nl], CST_BNDIGIT(0), l);
    carry = c;
    nl++;
  case -5:
    l = ~n[nl]; c = CST_BNDIGIT(0);
    zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
    zbnadd(c, m[nl], c, m[nl], CST_BNDIGIT(0), l);
    carry = c;
    nl++;
  case -4:
    l = ~n[nl]; c = CST_BNDIGIT(0);
    zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
    zbnadd(c, m[nl], c, m[nl], CST_BNDIGIT(0), l);
    carry = c;
    nl++;
  case -3:
    l = ~n[nl]; c = CST_BNDIGIT(0);
    zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
    zbnadd(c, m[nl], c, m[nl], CST_BNDIGIT(0), l);
    carry = c;
    nl++;
  case -2:
    l = ~n[nl]; c = CST_BNDIGIT(0);
    zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
    zbnadd(c, m[nl], c, m[nl], CST_BNDIGIT(0), l);
    carry = c;
    nl++;
  case -1:
    l = ~n[nl]; c = CST_BNDIGIT(0);
    zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
    zbnadd(c, m[nl], c, m[nl], CST_BNDIGIT(0), l);
    carry = c;
    nl++;
  }
  
  if (nl) {
    do {
      l = ~n[nl]; c = CST_BNDIGIT(0);
      zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
      zbnadd(c, m[nl], c, m[nl], CST_BNDIGIT(0), l);

      l = ~n[nl+1]; carry = CST_BNDIGIT(0);
      zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
      zbnadd(carry, m[nl+1], carry, m[nl+1], CST_BNDIGIT(0), l);

      l = ~n[nl+2]; c = CST_BNDIGIT(0);
      zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
      zbnadd(c, m[nl+2], c, m[nl+2], CST_BNDIGIT(0), l);

      l = ~n[nl+3]; carry = CST_BNDIGIT(0);
      zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
      zbnadd(carry, m[nl+3], carry, m[nl+3], CST_BNDIGIT(0), l);

      l = ~n[nl+4]; c = CST_BNDIGIT(0);
      zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
      zbnadd(c, m[nl+4], c, m[nl+4], CST_BNDIGIT(0), l);

      l = ~n[nl+5]; carry = CST_BNDIGIT(0);
      zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
      zbnadd(carry, m[nl+5], carry, m[nl+5], CST_BNDIGIT(0), l);

      l = ~n[nl+6]; c = CST_BNDIGIT(0);
      zbnadd(c, l, c, l, CST_BNDIGIT(0), carry);
      zbnadd(c, m[nl+6], c, m[nl+6], CST_BNDIGIT(0), l);

      l = ~n[nl+7]; carry = CST_BNDIGIT(0);
      zbnadd(carry, l, carry, l, CST_BNDIGIT(0), c);
      zbnadd(carry, m[nl+7], carry, m[nl+7], CST_BNDIGIT(0), l);

      nl += 8;
    } while(nl);
  }

  if (ml && !carry) {
    m += ml;  ml *= CST_BNLENGTH(-1);
    do {
      if (m[ml]--) {break;}
      ml++;
    } while(ml);
    carry = ml ? CST_BNDIGIT(1) : CST_BNDIGIT(0);
  }
  return(carry);
}
#else
BigNumDigit ZBNSubtract(m, ml, n, nl, carry)
     BigNum m;
     BigNumLength ml;
     BigNum n;
     BigNumLength nl;
     BigNumDigit carry;
{
  return((nl) ? 
	 ((ml) <= (nl) ? 
	  ((BigNumDigit)(1)-(BigNumDigit)(mpn_sub_n(m, m, n, ml)+
					  mpn_sub_1(m, m, ml, (BigNumDigit)(1-(carry))))) 
	  : 
	  ((BigNumDigit)(1)-(BigNumDigit)(mpn_sub_1((m)+(nl), (m)+(nl), ((ml)-(nl)),
						    mpn_sub_n(m, m, n, (nl))+
						    mpn_sub_1(m, m, (nl), (BigNumDigit)(1-(carry)))))))
	 : ZBNSubtractBorrow(m,ml,carry));
}
#endif
