/*
 * zen/zbn/c/assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

/* \TEX{
   \Subsection{Assigning}
   } */

/* \TEX{
   \title{Set a |BigNum| to zero}
   \input{An allocated |BigNum| and its length.}
   \side{The |BigNum| is filled with zeros.}
   } */
void ZBNSetToZero (n, nl)
     BigNum n;
     BigNumLength nl;
{
  if (!nl) {return;}
  n += nl; nl *= CST_BNLENGTH(-1);
  do {
    n[nl] = CST_BNDIGIT(0);
    nl++;
  } while (nl);
}

/* \TEX{
   \title{Set a |BigNum| to one}
   \input{An allocated |BigNum| and its length.}
   \side{The |BigNum| is filled with zeros execept the first digit
   which is set to one.}
   } */
void ZBNSetToOne (n, nl)
     BigNum n;
     BigNumLength nl;
{
  if (nl <= CST_BNLENGTH(0)) {return;}
  n[0] = CST_BNDIGIT(1);
  n += nl; nl--;
  if (!nl) {return;}
  nl *= CST_BNLENGTH(-1);
  do {
    n[nl] = CST_BNDIGIT(0);
    nl++;
  } while (nl);
}

/* \TEX{
   \title{Assigning |BigNum|s}
   \input{Two allocated |BigNum|s and a length.}
   \side{The |BigNum| |m| is filled with |(n,nl)|.}
   \note{All kinds of overlapping are possible. No side effect if |nl|
   is zero.}
   } */
void ZBNAssign(m, n, nl)
     BigNum m;
     BigNum n;
     BigNumLength nl;
{
  if (!nl) {return;}
  if (m < n) {
    m += nl; n += nl; nl *= CST_BNLENGTH(-1);
    do {
      m[nl] = n[nl];
      nl++;
    } while(nl);
  }
  else {
    nl--;
    do {
      m[nl] = n[nl];
      nl--;
    } while(nl >= CST_BNLENGTH(0));
  }
}
