/*
 * zen/zbn/c/binary.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

/* \TEX{
   \Subsection{Binary operations}
   } */

/* \TEX{
   \title{And}
   \input{Two allocated |BigNum|s and a length.}
   \side{$|(m,nl)| = |(m,nl)|$ and $|(n,nl)|$.}
   \note{One must have the allocated size of $|m|$ which must be
   greater (or equal) than |nl|.}
   } */
void ZBNAnd(m, n, nl)
     BigNum m, n;
     BigNumLength nl;
{
  if (!nl) {return;}
  n += nl; m += nl; nl *= CST_BNLENGTH(-1);
  do {
    m[nl] &= n[nl];
    nl++;
  } while (nl);
}

/* \TEX{
   \title{Or}
   \input{Two allocated |BigNum|s and a length.}
   \side{$|(m,nl)| = |(m,nl)|$ or $|(n,nl)|$.}
   \note{One must have the allocated size of $|m|$ which must be
   greater than |nl|.}
   } */
void ZBNOr(m, n, nl)
     BigNum m, n;
     BigNumLength nl;
{ 
  if (!nl) {return;}
  n += nl; m += nl; nl *= CST_BNLENGTH(-1);
  do {
    m[nl] |= n[nl];
    nl++;
  } while (nl);
}

/* \TEX{
   \title{Xor}
   \input{Two allocated |BigNum|s and a length.}
   \side{$|(m,nl)| = |(m,nl)|$ xor $|(n,nl)|$.}
   \note{One must have the allocated size of $|m|$ which must be
   greater than |nl|.}
   } */
void ZBNXor(m, n, nl)
     BigNum m, n;
     BigNumLength nl;
{
  if (!nl) {return;}
  n += nl; m += nl; nl *= CST_BNLENGTH(-1);
  do {
    m[nl] ^= n[nl];
    nl++;
  } while (nl);
}

