/*
 * zen/zbn/c/digits.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

/* \TEX{   
   \Subsection{Basic functions on |BigNumDigit|s}

   We defined the following macros to help manipulating a |BigNum|.
   } */

#if 0
/*\TEX{
  \title{Reduction modulo a bloc}
  \input{An integer |x|.}
  \output{The remainder of $|x| / |SIZE_BLOC|$, that is to say $|x|
  \bmod |SIZE_BLOC|$ }
  } */
BigNumLength modSizeBloc(x)
     BigNumLength x;{}
#endif

#if 0
/* \TEX{
   \title{Number of blocs}
   \input{An integer |x|.}
   \output{The quotient $|x| / |SIZE_BLOC|$ }
   } */
BigNumLength divSizeBloc(x)
     BigNumLength x;{}
#endif

#if 0
/* \TEX[
   The two following functions have an equivalent meaning, once
   replaced |SIZE_BLOC| by |SIZE_CHAR|.
   \title{Reduction modulo a char}
   \input{An integer}
   \output{The remainder of $|x| / |SIZE_CHAR|$, that is to say $|x|
   \bmod |SIZE_CHAR|$ }
   } */
BigNumLength modSizeChar(x)
     BigNumLength x;{}
#endif

#if 0
/* \TEX[
   \title{Number of chars}
   \input{An integer}
   \output{The quotient $|x| / |SIZE_CHAR|$ }
   } */
BigNumLength divSizeChar(x)
     BigNumLength x;{}
#endif

#if 0
/* \TEX{
   \title{Shifting 1}
   \input{An integer |x|.}
   \output{The |BigNumDigit| representing $2^{|x|} $ if
   $|x|<|SIZE_BLOC|$, 0 otherwise.}
   } */
BigNumDigit BlocExp2(x)
     unsigned int x; {}
#endif

#if 0
/* \TEX{
   \title{Getting least significant digit of a |BigNum|}
   \input{An allocated |BigNum|.}
   \output{The least significant digit of |n|.}
   } */
BigNumDigit ZBNGetDigit(n)
     BigNum n; {}
#endif

#if 0
/* \TEX{
   \title{Assigning least significant digit of a |BigNum|}
   \input{An allocated |BigNum| and a |BigNumDigit|.}
   \side{The least significant digit of |n| is set to |d|.}
   } */
void ZBNSetDigit(n,d)
     BigNum n;
     BigNumDigit d;{}
#endif

#if 0
/* \TEX{
   \title{Getting i$^{\rm th}$ digit of a |BigNum|}
   \input{An integer and an allocated |BigNum|.}
   \output{The |i|th digit of |n|.}
   \note{The 0th digit is the least significant one. If |n| is of size
   |nl|, the (|nl|-1)th is the most significant one.  This is a macro
   that can be a lvalue in an assignment.}
   } */
BigNumDigit ZBNDigit(n,i)
     BigNum n;
     int i; {}
#endif

/* \TEX{
   \title{Number of unsignificant bits in a digit}
   \input{A digit |d|.}
   \output{The number |k| of most significant bits set to zero in |d|.}
   \note{If |d| is zero, $|k|=|SIZE_BLOC|$. }
   } */
BigNumLength ZBNNumLeadingZeroBitsInDigit(d)
     BigNumDigit d;
{
  BigNumLength b;
  BigNumDigit e;
 
  if (!d) {
    return(SIZE_BLOC);
  }
 
  b = -1;
  do {
    b++;
    e = d; d <<= 1;
  } while (d > e);
  return(b);
}

#if 0
/* \TEX{
   \title{Number of significative bits}
   \input{A digit |d|.}
   \output{The number |k| of significative bits in |d|.}
   \note{If |d| is zero, $|k|=0$. }
   } */
unsigned int ZBNNumBitsInDigit(d)
     BigNumDigit d; {}
#endif

/* \TEX{
   \title{Significant length of a |BigNum|}
   \input{An allocated |BigNum| and its length.}
   \output{The number of significant digits of |(n,nl)|.}
   \note{If |(n,nl)| is zero, returns 1.}
   } */
BigNumLength ZBNNumDigits(n, nl)
     BigNum n;
     BigNumLength nl;
{
  if (!nl) {return(CST_BNLENGTH(1));}
  nl--;
  while(nl && !n[nl]) {nl--;}
  return(nl+1);
}

