/*
 * zen/zbn/c/karamul.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "kara_defaults.h"

#ifndef KARAm_LIM
static BigNumLength KARAm_LIM=1;
# define KKL 0
#else
# define KKL KARAm_LIM
#endif

#if ZBN_KARA_OPTS & MK3
# define StopKara(N) (((N)&((-1)^1)) <= KARAm_LIM)
#else
# define StopKara(N) ((N) <= KARAm_LIM)
#endif

static void ZBNMK1 ___((BigNum,BigNum,BigNumLength,BigNum,BigNumLength,BigNum));
static void ZBNMK2 ___((BigNum,BigNum,BigNum,BigNumLength,BigNum));
static void ZBNMK3 ___((BigNum,BigNum,BigNum,BigNumLength,BigNum));

#if KKL < 4
  static void ZBNMK1_4 ___((BigNum,BigNum,BigNumLength,BigNum,BigNum));
  static void ZBNMK2_4 ___((BigNum,BigNum,BigNum,BigNum));
#else
# define ZBNMK1_4(x,a,al,b,buf) ZBNMultiplyPlain(x,(al)+4,a,al,b,(BigNumLength)4)
# define ZBNMK2_4(x,a,b,buf) ZBNMultiplyPlain(x,(BigNumLength)8,a,(BigNumLength)4,b,(BigNumLength)4)
#endif

#if KKL < 3
  static void ZBNMK1_3 ___((BigNum,BigNum,BigNumLength,BigNum,BigNum));
  static void ZBNMK3_3 ___((BigNum,BigNum,BigNum,BigNum));
#else
# define ZBNMK1_3(x,a,al,b,buf) ZBNMultiplyPlain(x,(al)+3,a,al,b,(BigNumLength)3)
# define ZBNMK3_3(x,a,b,buf) ZBNMultiplyPlain(x,(BigNumLength)6,a,(BigNumLength)3,b,(BigNumLength)3)
#endif

#if KKL < 2
  static void _ZBNMK1_2 ___((BigNum,BigNum,BigNumLength,BigNum,BigNum));
  static void _ZBNMK2_2 ___((BigNum,BigNum,BigNum,BigNum));
# ifdef KARAm_LIM
#  define ZBNMK1_2(x,a,al,b,buf) _ZBNMK1_2(x,a,al,b,buf) 
#  define ZBNMK2_2(x,a,b,buf) _ZBNMK2_2(x,a,b,buf)
# else
#  define ZBNMK1_2(x,a,al,b,buf) if(StopKara(2)) ZBNMultiplyPlain(x,(al)+2,a,al,b,(BigNumLength)2); else _ZBNMK1_2(x,a,al,b,buf)
#  define ZBNMK2_2(x,a,b,buf) if(StopKara(2)) ZBNMultiplyPlain(x,(BigNumLength)4,a,(BigNumLength)2,b,(BigNumLength)2); else _ZBNMK2_2(x,a,b,buf)
# endif
#else
# define ZBNMK1_2(x,a,al,b,buf) ZBNMultiplyPlain(x,(al)+2,a,al,b,(BigNumLength)2)
# define ZBNMK2_2(x,a,b,buf) ZBNMultiplyPlain(x,(BigNumLength)4,a,(BigNumLength)2,b,(BigNumLength)2)
#endif

#if ZBN_KARA_OPTS & DAss
# define _DblAssign(buf,a,b,lh,l2) if(1) { ZBNAssign(buf,a,lh); ZBNAssign(buf+lh,b,lh); ZBNSetToZero(buf+l2,l2); } else
#else
# define _DblAssign(buf,a,b,lh,l2)  if(1) {BigNum pt1,pt2,pt3,pta,ptb,ptlim; \
   ptlim = pt3 = (lh) + (pt2 = lh + (pt1 = buf)); pta = a; ptb = b; \
   do { *(pt1++)= *(pta++); *(pt2++)= *(ptb++); *(pt3++)= 0; *(pt3++) = 0;} while (pt2 < ptlim); \
 } else
#endif


#if ZBN_KARA_OPTS & MSub
# define _MSub(buf,lh,l2,a,b,N,sgn) if(1) { \
   if(ZBNSubtract(buf,lh,a,N,(BigNumDigit)1)) sgn=0; else { \
     _CompAdd(buf,lh); sgn= -1; } \
     if(ZBNSubtract(buf+lh,lh,b,N,(BigNumDigit)1)==0) { \
	 _CompAdd(buf+lh,lh); sgn= ~sgn; } } else
#else
# define _MSub(buf,lh,l2,a,b,N,sgn) if(1) { \
	   if(ZBNSubtract(buf,lh,a,N,(BigNumDigit)1)) { \
     if(ZBNSubtract(buf+lh,lh,b,N,(BigNumDigit)1)) sgn=0; \
     else { _CompAdd(buf+lh,lh); sgn=1; } \
   } else { if(ZBNSubtract(buf+lh,lh,b,N,(BigNumDigit)1)) { \
         _CompAdd(buf,lh); sgn=1; \
       } else { \
         ZBNComplement(buf,l2); \
         _AddCarry(buf,lh); \
         _AddCarry(buf+lh,lh); \
				 sgn=0; } } } else
#endif

#if KARA_DEBUG
#  define _Verification(x,xl,a,al,b,bl,str)  BigNum bufinit,verif;  BigNumDigit vret; \
 if(KARA_DEBUG > 1) fprintf(stderr,"Entering %s...\n",str); \
 bufinit=ZBNC(xl); ZBNAssign(bufinit,x,xl); \
 verif=ZBNC(xl); ZBNAssign(verif,bufinit,xl); vret=ZBNMultiplyPlain(verif,xl,a,al,b,bl)
# define __Verification(x,xl,a,al,b,bl,N,ret,str) \
  if(vret != (ret)) fprintf(stderr,"%s: vret:= %d - %d ;\n",str,vret,ret); \
  if(ZBNCompare(verif,xl,x,xl)) \
    fprintf(stderr,"%s: Different result\n",str); \
  if((KARA_DEBUG >= 4) || (ZBNCompare(verif,xl,x,xl))) { \
     fprintf(stderr,"%s: ",str); \
     ZBNPrintToFile(stderr,bufinit,xl,10); \
     fprintf(stderr,"+("); \
     if(N) { ZBNPrintToFile(stderr,a,N,10); \
     fprintf(stderr," + 2^%d * ",(N)*SIZE_BLOC); } \
     ZBNPrintToFile(stderr,a+N,(al)-(N),10); \
     fprintf(stderr,") * ("); \
     if(N) { ZBNPrintToFile(stderr,b,N,10); \
     fprintf(stderr," + 2^%d * ",(N)*SIZE_BLOC); }\
     ZBNPrintToFile(stderr,(b)+(N),(bl)-(N),10); \
     fprintf(stderr,") - "); \
     ZBNPrintToFile(stderr,x,xl,10); \
     fprintf(stderr,";\n "); \
     ZBNPrintToFile(stderr,verif,xl,10); \
     fprintf(stderr,"\n "); } \
  ZBNF(verif);  if(KARA_DEBUG > 1) fprintf(stderr,"Exiting %s...\n",str)
#else
# define _Verification(x,xl,a,al,b,bl,str)
# define __Verification(x,xl,a,al,b,bl,N,ret,str) 
#endif

#if KARA_DEBUG > 1
# define _Interm(a,b,buf,N,lh) \
 fprintf(stderr,"ahl :="); \
 ZBNPrintToFile(stderr,buf,lh,10); \
 fprintf(stderr," - "); \
 ZBNPrintToFile(stderr,a,N,10); \
 fprintf(stderr,";\nbhl :="); \
 ZBNPrintToFile(stderr,(buf)+(lh),lh,10); \
 fprintf(stderr," - "); \
 ZBNPrintToFile(stderr,b,N,10); \
 fprintf(stderr,";\n")
# define __Interm(buf,lh) \
 fprintf(stderr,"ahl -"); \
 ZBNPrintToFile(stderr,buf,lh,10); \
 fprintf(stderr,";\nbhl -"); \
 ZBNPrintToFile(stderr,(buf)+(lh),lh,10); \
 fprintf(stderr,";\n")
# define ___Interm(x,xl,N) \
 fprintf(stderr,"x := "); \
 ZBNPrintToFile(stderr,x,(N)<<1,10); \
 fprintf(stderr," + 2^%d * ",((N)<<1)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,x+((N)<<1),(xl)-((N)<<1),10); \
 fprintf(stderr,";\n"); \
 fprintf(stderr,"x := x + 2^%d * (",(N)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,x,((N)<<1),10); \
 fprintf(stderr," + "); \
 ZBNPrintToFile(stderr,(x)+((N)<<1),(xl)-((N)<<1),10); \
 fprintf(stderr,");\nx  - (")
# define ____Interm(x,xl,N,buf,l2,sgn) \
 ZBNPrintToFile(stderr,x,N,10); \
 fprintf(stderr," + 2^%d * ",(N)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,(x)+(N),N,10); \
 fprintf(stderr," + 2^%d * ",2*(N)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,(x)+2*(N),N,10); \
 fprintf(stderr," + 2^%d * ",3*(N)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,(x)+3*(N),N,10); \
 fprintf(stderr,");\n"); \
 fprintf(stderr,"ahl * bhl - "); \
 ZBNPrintToFile(stderr,(buf)+(l2),l2,10); \
 fprintf(stderr,";\n"); \
 if(sgn) fprintf(stderr,"x := x + 2^%d * ",(N)*SIZE_BLOC); \
 else fprintf(stderr,"x := x - 2^%d * ",(N)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,(buf)+(l2),l2,10); \
 fprintf(stderr,";\n")
#else
# define _Interm(a,b,buf,N,lh)
# define __Interm(buf,lh)
# define ___Interm(x,xl,N)
# define ____Interm(x,xl,N,buf,l2,sgn)
#endif

#ifndef KARAm_LIM
/* \TEX{
   \title{Karatsuba's multiplication of two |BigNum|s.}
   \input{Three allocated |BigNum|s with $|xl| \ge |al|+|bl|$ and the
   limit of size under which |ZBNMultiplyPlain| must be used.}
   \output{The possible carry out, or |ZENERR| if an error occured.}
   \side{|x += a * b|}
   \note{A buffer is allocated and freed.}
   } */
BigNumDigit ZBNMultiplyKaratsuba(x,xl,a,al,b,bl,lim)
     BigNum x,a,b;
     BigNumLength xl,al,bl,lim;
{
#else
BigNumDigit ZBNMultiplyKaratsuba(x,xl,a,al,b,bl)
     BigNum x,a,b;
     BigNumLength xl,al,bl;
{
#endif
 BigNum tmp,z;
 BigNumLength tmpl;
 BigNumDigit ret;

 _Verification(x,xl,a,al,b,bl,"ZBNMultiplyKaratsuba");

 al=ZBNNumDigits(a,al);
 bl=ZBNNumDigits(b,bl);
 
 if(bl>al)
   {
     tmp=a;     tmpl=al;
     a=b;       al=bl;
     b=tmp;     bl=tmpl;
   }

#ifndef KARAm_LIM
 KARAm_LIM=lim;
#endif
 
 if(StopKara(bl))
#if PURE_KARA
   ret=ZBNMultiplyDigit(x,xl,a,al,*b);
#else
   ret=ZBNMultiplyPlain(x,xl,a,al,b,bl);
#endif
 else
   {
     tmpl=ZBNSizeBufKaraM(bl);
     tmp=ZBNC(tmpl);
     
     if(!tmp)
       {
	 ZENSetError(ZEN_LIB_ZBN,ZEN_MULTKARA,ZEN_ALLOC);
	 return(ZERR);
       }

     tmpl=al+bl;
     z=ZBNC(tmpl);
     if(!z)
       {
	 ZENSetError(ZEN_LIB_ZBN,ZEN_MULTKARA,ZEN_ALLOC);
	 return(ZERR);
       }
     
     ZBNSetToZero(z,tmpl);

     if (bl == 4) {
       if(al == bl)
	 ZBNMK2_4(z,a,b,tmp);
       else
	 ZBNMK1_4(z,a,al,b,tmp);
     }
     else if (bl == 3) {
       if(StopKara(2))
	 {
	   ret=ZBNMultiplyPlain(x,xl,a,al,b,(BigNumLength)3);
	   goto fin;
	 }
       else if(al == bl)
	 ZBNMK3_3(z,a,b,tmp);
       else
	 ZBNMK1_3(z,a,al,b,tmp);
     }
     else if (bl == 2) {
       if(al == bl)
	 ZBNMK2_2(z,a,b,tmp);
       else
	 ZBNMK1_2(z,a,al,b,tmp);
     }
     else {
       if(al == bl) {
	 if(al & 1)
	   ZBNMK3(z,a,b,al>>1,tmp);
	 else
	   ZBNMK2(z,a,b,al>>1,tmp);
       }
       else
	 ZBNMK1(z,a,al,b,bl,tmp);
     }
     ret=ZBNAdd(x,xl,z,tmpl,(BigNumDigit)0);
   fin:
     ZBNF(z);
     ZBNF(tmp);
   }

 __Verification(x,xl,a,al,b,bl,0,ret,"ZBNMultiplyKaratsuba");
 return(ret);
}

#ifndef KARAm_LIM
/* \TEX{
   \title{Karatsuba's multiplication}
   \input{Four |BigNum|s: $|x|=0$ of size greater or equal than
   $|al|+|bl|$, a buffer of size at least |ZBNSizeBufKaraM(bl)|, and
   the limit of size under which |ZBNMultiply| must be used. }
   \side{$|x|=|a|\times|b|$}
   } */
void ZBNMultiplyKaratsubaBuffer(x,a,al,b,bl,buf,lim)
     BigNum x,a,b,buf;
     BigNumLength al,bl,lim;
{
#else
void ZBNMultiplyKaratsubaBuffer(x,a,al,b,bl,buf)
     BigNum x,a,b,buf;
     BigNumLength al,bl;
{
#endif

  _Verification(x,al+bl,a,al,b,bl,"ZBNMultiplyKaratsubaBuffer");

#ifndef KARAm_LIM
  KARAm_LIM=lim;
#endif

  al=ZBNNumDigits(a,al);
  bl=ZBNNumDigits(b,bl);

  if(bl>al)
    {
      if(StopKara(al))
#if PURE_KARA
	ZBNMultiplyDigit(x,al+bl,b,bl,*a);
#else
	ZBNMultiplyPlain(x,al+bl,b,bl,a,al);
#endif      
      else if (al == 4) {
	ZBNMK1_4(x,b,bl,a,buf);
      }
      else if (al == 3) {
	if(StopKara(2))
	  ZBNMultiplyPlain(x,bl+3,b,bl,a,(BigNumLength)3);
	else
	  ZBNMK1_3(x,b,bl,a,buf);
      }
      else if (al == 2) {
	ZBNMK1_2(x,b,bl,a,buf);
      }
      else {
	ZBNMK1(x,b,bl,a,al,buf);
      }
    }
  else
    {
      if(StopKara(bl))
#if PURE_KARA
	ZBNMultiplyDigit(x,al+bl,a,al,*b);
#else
	ZBNMultiplyPlain(x,al+bl,a,al,b,bl);
#endif      
      else
	{
	  if(bl<al) {
	    if (bl == 4) {
		ZBNMK1_4(x,a,al,b,buf);
	    }
	    else if (bl == 3) {
	      if(StopKara(2))
		ZBNMultiplyPlain(x,al+3,a,al,b,(BigNumLength)3);
	      else
		ZBNMK1_3(x,a,al,b,buf);
	    }
	    else if (bl == 2) {
	      ZBNMK1_2(x,a,al,b,buf);
	    }
	    else {
	      ZBNMK1(x,a,al,b,bl,buf);
	    }
	  }
	  else {
	    if (al == 4) {
	      ZBNMK2_4(x,a,b,buf);
	    }
	    else if (al == 3) {
	      if(StopKara(2))
		ZBNMultiplyPlain(x,(BigNumLength)6,a,(BigNumLength)3,b,(BigNumLength)3);
	      else
		ZBNMK3_3(x,a,b,buf);
	    }
	    else if (al == 2) {
	      ZBNMK2_2(x,a,b,buf);
	    }
	    else {
	      if(al & 1)
		ZBNMK3(x,a,b,al>>1,buf);
	      else
		ZBNMK2(x,a,b,al>>1,buf);
	    }
	  }
	}
    }

  __Verification(x,al+bl,a,al,b,bl,0,0,"ZBNMultiplyKaratsubaBuffer");
}



/* \TEX{
   \title{Size of buffer needed for Karatsuba's multiplication}
   \input{The length of smallest operand.}
   \output{The length of the buffer needed in Karatsuba's routines}
   } */
BigNumLength ZBNSizeBufKaraM(l)
     BigNumLength l;
{BigNumLength ret=0;
 
 for(;;)
   {
     l = (l&1) + (l>>1);
     ret += l << 2;
     if(l<=KARAm_LIM)
       return(ret);
   }
}

/************************************************/

/* One must have al != bl
   x must be set to zero and of size >= al+bl
   Side effect: x = a*b
   Buffer must be allocated.
*/
static void ZBNMK1(x,a,al,b,bl,buf)
     BigNum x,a,b,buf;
     BigNumLength al,bl;
{BigNum tmp;
 BigNumLength tmpl;
 void (*MK)();

 if(bl & 1)
   MK=ZBNMK3;
 else
   MK=ZBNMK2;
 
 MK(x,a,b,bl>>1,buf);

 
 do
   {
     al -= bl;
     a += bl;
     x += bl;

     if (al<bl)
       {
	 if(StopKara(al))
	   {
#if PURE_KARA
	     ZBNMultiplyDigit(x,1+bl,b,bl,*a);
#else
	     ZBNMultiplyPlain(x,al+bl,b,bl,a,al);
#endif	     
	     return;
	   }

	 tmp=a;     tmpl=al;
	 a=b;       al=bl;
	 b=tmp;     bl=tmpl;

	 if (bl == 4) {
	   ZBNMK1_4(x,a,al,b,buf);
	 }
	 else if (bl == 3) {
	   if(StopKara(2))
	     ZBNMultiplyPlain(x,al+3,a,al,b,(BigNumLength)3);
	   else
	     ZBNMK1_3(x,a,al,b,buf);
	   return;
	 }
	 else if (bl == 2) {
	   ZBNMK1_2(x,a,al,b,buf);
	   return;
	 }
	 
         if(bl & 1)
           MK=ZBNMK3;
         else
           MK=ZBNMK2;
       }
     
     MK(x,a,b,bl>>1,buf);
          
   } while(al != bl);
}


/* sizes of a and b are equal to 2*N. Size of x must be equal at least to 4*N
   x must be set to zero
   Side effect: x=a*b
   Buffer must be allocated.
*/
static void ZBNMK2(x,a,b,N,buf)
     BigNum x,a,b,buf;
     BigNumLength N;
{BigNumLength l;
 BigNumDigit sgn;

 _Verification(x,N<<2,a,N<<1,b,N<<1,"ZBNMK2");

 l=N<<1;

 _DblAssign(buf,a+N,b+N,N,l);

 _Interm(a,b,buf,N,N);

 _MSub(buf,N,l,a,b,N,sgn);
 
 __Interm(buf,N);

 if(StopKara(N))
   {
#if PURE_KARA			/* N=1 l=2 */
     ZBNMultiplyDigit(x,(BigNumLength)2,a,(BigNumLength)1,*b);
     ZBNMultiplyDigit(x+2,(BigNumLength)2,a+1,(BigNumLength)1,b[1]);
     ZBNMultiplyDigit(buf+2,(BigNumLength)2,buf,(BigNumLength)1,buf[1]);
#else
     ZBNMultiplyPlain(x,l,a,N,b,N);
     ZBNMultiplyPlain(x+l,l,a+N,N,b+N,N);
     ZBNMultiplyPlain(buf+l,l,buf,N,buf+N,N);
#endif     
   }
 else
   {
     BigNumLength n;
     BigNum bl;
     
     bl=buf+(l<<1);

     if (N == 4) {
       ZBNMK2_4(x,a,b,bl);
       ZBNMK2_4(x+8,a+4,b+4,bl);
       ZBNMK2_4(buf+8,buf,buf+4,bl);
     }
     else if (N == 3) {
       if(StopKara(2)) {
	 ZBNMultiplyPlain(x,(BigNumLength)6,a,(BigNumLength)3,b,(BigNumLength)3);
	 ZBNMultiplyPlain(x+6,(BigNumLength)6,a+3,(BigNumLength)3,b+3,(BigNumLength)3);
	 ZBNMultiplyPlain(buf+6,(BigNumLength)6,buf,(BigNumLength)3,buf+3,(BigNumLength)3);
       }
       else {
	 ZBNMK3_3(x,a,b,bl);
	 ZBNMK3_3(x+6,a+3,b+3,bl);
	 ZBNMK3_3(buf+6,buf,buf+3,bl);
       }
     }
     else if (N == 2) {
       ZBNMK2_2(x,a,b,bl);
       ZBNMK2_2(x+4,a+2,b+2,bl);
       ZBNMK2_2(buf+4,buf,buf+2,bl);
     }
     else {
       n= N>>1;
       
       if(N & 1) {
	 ZBNMK3(x,a,b,n,bl);
	 ZBNMK3(x+l,a+N,b+N,n,bl);
	 ZBNMK3(buf+l,buf,buf+N,n,bl);
       }
       else {
	 ZBNMK2(x,a,b,n,bl);
	 ZBNMK2(x+l,a+N,b+N,n,bl);
	 ZBNMK2(buf+l,buf,buf+N,n,bl);
       }
     }
   }
 
 ___Interm(x,N<<2,N);

 _Mix(x,N,l,l,N);

 ____Interm(x,N<<2,N,buf,N<<1,sgn);

 if(sgn)
   _Add(x+N,l+N,buf+l,l,(BigNumDigit)0);
 else
   _Sub(x+N,l+N,buf+l,l,(BigNumDigit)1);

 __Verification(x,N<<2,a,N<<1,b,N<<1,N,0,"ZBNMK2");
}

/* sizes of a and b are equal to 2*N+1. Size of x must be equal at least to 4*N+2
   x must be set to zero
   Side effect: x=a*b
   Buffer must be allocated.
*/

static void ZBNMK3(x,a,b,N,buf)
     BigNum x,a,b,buf;
     BigNumLength N;
#if ZBN_KARA_OPTS & MK3
{
  BigNumLength l,lt,dlt;

  _Verification(x,(N<<2)+2,a,(N<<1)+1,b,(N<<1)+1,"ZBNMK3");
  
  dlt=(lt=(l=(N<<1))+1)+1;

  if(StopKara(N))
    ZBNMultiplyPlain(x,lt<<1,a,lt,b,lt);
  else
    {
      if (N == 2) {
	ZBNMK2_4(x,a,b,buf);
      }
      else if (N == 1) {
	ZBNMK2_2(x,a,b,buf);
      }
      else {
	ZBNMK2(x,a,b,N,buf);
      }
      ZBNMultiplyDigit((x)+(l),dlt,a,lt,(b)[l]); 
      ZBNMultiplyDigit((x)+(l),dlt,b,l,(a)[l]);
    }
}
#else
{BigNumLength l,lh,l2;
 BigNumDigit sgn;

 _Verification(x,(N<<2)+2,a,(N<<1)+1,b,(N<<1)+1,"ZBNMK3");

 l=N<<1;
 lh=N+1;
 l2=lh<<1;
 
 _DblAssign(buf,a+N,b+N,lh,l2);

 _Interm(a,b,buf,N,lh);

 _MSub(buf,lh,l2,a,b,N,sgn);

 __Interm(buf,lh);

#if PURE_KARA
 {
   BigNumLength n;
   BigNum bl;
   
   bl=buf+(l2<<1);
    
   if (N == 4) {
     /* N=4 lh=5 l=8 l2=10 n=2 */
     ZBNMK2_4(x,a,b,bl);
     
     ZBNMK3(x+8,a+4,b+4,(BigNumLength)2,bl);
     ZBNMK3(buf+10,buf,buf+5,(BigNumLength)2,bl);
   }
   else if (N == 3) {
     /* N=3 lh=4 l=6 l2=8 n=2 */
     ZBNMK3_3(x,a,b,bl);
     
     ZBNMK2_4(x+6,a+3,b+3,bl);
     ZBNMK2_4(buf+8,buf,buf+4,bl);
   }
   else if (N == 2) {
     /* N=2 lh=3 l=4 l2=6 n=1 */
     ZBNMK2_2(x,a,b,bl);
     ZBNMK3_3(x+4,a+2,b+2,bl);
     ZBNMK3_3(buf+6,buf,buf+3,bl);
   }
   else if (N == 1) {
     /* N=1 lh=2 l=2 l2=4 n=1 */
     ZBNMultiplyDigit(x,(BigNumLength)2,a,(BigNumLength)1,*b);
     
     ZBNMK2_2(x+2,a+1,b+1,bl);
     ZBNMK2_2(buf+4,buf,buf+2,bl);
   }
   else {
     n= lh>>1;
     if(lh & 1) {
       ZBNMK2(x,a,b,n,bl);
       ZBNMK3(x+l,a+N,b+N,n,bl);
       ZBNMK3(buf+l2,buf,buf+lh,n,bl);
     }
     else {
       ZBNMK3(x,a,b,N>>1,bl);    
       ZBNMK2(x+l,a+N,b+N,n,bl);
       ZBNMK2(buf+l2,buf,buf+lh,n,bl);
     }
   }
 }
#else
 if(StopKara(lh)) {
   ZBNMultiplyPlain(x,l,a,N,b,N);
   ZBNMultiplyPlain(x+l,l2,a+N,lh,b+N,lh);
   ZBNMultiplyPlain(buf+l2,l2,buf,lh,buf+lh,lh);
 }
 else
   {
     BigNumLength n;
     BigNum bl;

     bl=buf+(l2<<1);
     
     if (N == 4) {
       if(StopKara(4))
	 ZBNMultiplyPlain(x,(BigNumLength)8,a,(BigNumLength)4,b,(BigNumLength)4);
       else
	 ZBNMK2_4(x,a,b,bl);
       
       ZBNMK3(x+8,a+4,b+4,(BigNumLength)2,bl);
       ZBNMK3(buf+10,buf,buf+5,(BigNumLength)2,bl);
     }
     else if (N == 3) {
       if(StopKara(2))
	 ZBNMultiplyPlain(x,(BigNumLength)6,a,(BigNumLength)3,b,(BigNumLength)3);
       else
	 ZBNMK3_3(x,a,b,bl);

       ZBNMK2_4(x+6,a+3,b+3,bl);
       ZBNMK2_4(buf+8,buf,buf+4,bl);
     }
     else if (N == 2) { 
       if(StopKara(2)) {
	 ZBNMultiplyPlain(x,(BigNumLength)4,a,(BigNumLength)2,b,(BigNumLength)2);
	 ZBNMultiplyPlain(x+4,(BigNumLength)6,a+2,(BigNumLength)3,b+2,(BigNumLength)3);
	 ZBNMultiplyPlain(buf+6,(BigNumLength)6,buf,(BigNumLength)3,buf+3,(BigNumLength)3);
       }
       else {
	 ZBNMK2_2(x,a,b,bl);
	 
	 ZBNMK3_3(x+4,a+2,b+2,bl);
	 ZBNMK3_3(buf+6,buf,buf+3,bl);
       }
     }
     else if (N == 1) { 
       ZBNMultiplyDigit(x,l,a,(BigNumLength)1,*b);
       ZBNMK2_2(x+2,a+1,b+1,bl);
       ZBNMK2_2(buf+4,buf,buf+2,bl);
     }
     else {
       n= lh>>1;
       if(lh & 1) {
	 if(StopKara(N))
	   ZBNMultiplyPlain(x,l,a,N,b,N);
	 else
	   ZBNMK2(x,a,b,n,bl);
	 ZBNMK3(x+l,a+N,b+N,n,bl);
	 ZBNMK3(buf+l2,buf,buf+lh,n,bl);
       }
       else {
	 if(StopKara(N))
	   ZBNMultiplyPlain(x,l,a,N,b,N);
	 else
	   ZBNMK3(x,a,b,N>>1,bl);
	 ZBNMK2(x+l,a+N,b+N,n,bl);
	 ZBNMK2(buf+l2,buf,buf+lh,n,bl);
       }  
     }
   }
#endif

 ___Interm(x,(N<<2)+2,N);
 
 _Mix(x,N,l,l2,lh+1);

 ____Interm(x,(N<<2)+2,N,buf,l2,sgn);

 if(sgn)
   _Add(x+N,l2+N,buf+l2,l2,(BigNumDigit)0);
 else
   _Sub(x+N,l2+N,buf+l2,l2,(BigNumDigit)1);
        
 __Verification(x,(N<<2)+2,a,(N<<1)+1,b,(N<<1)+1,N,0,"ZBNMK3");
}
#endif /* MK3 */


/* sizes of a and b are equal to 2. Size of x must be equal at least to 4
   x must be set to zero
   Side effect: x=a*b
   Buffer must be allocated.
*/

static void _ZBNMK2_2(x,a,b,buf)
     BigNum x,a,b,buf;
{BigNumDigit sgn;

 _Verification(x,4,a,2,b,2,"ZBNMK2_2");

 buf[2]=buf[3]=sgn=0;

#if KARA_DEBUG > 1
 fprintf(stderr,"ahl :=");
 ZBNPrintToFile(stderr,a+1,1,10);
 fprintf(stderr," - ");
 ZBNPrintToFile(stderr,a,1,10);
 fprintf(stderr,";\nbhl :=");
 ZBNPrintToFile(stderr,b+1,1,10);
 fprintf(stderr," - ");
 ZBNPrintToFile(stderr,b,1,10);
 fprintf(stderr,";\n");
#endif

#define IMK2_2(bufl,a) if(1) { \
	(buf)[bufl]=(a)[1];			 \
	if(Sub2Digit((buf)[bufl],(a)[0])) { (buf)[bufl]= 1 + ~((buf)[bufl]); sgn= ~sgn; }} else

  {
    BigNumDigit save;
    
    IMK2_2(0,a);
    IMK2_2(1,b);
  }

 __Interm(buf,1);
 
 ZBNMultiply2Digit(x,a,b);
 ZBNMultiply2Digit(x+2,a+1,b+1);
 ZBNMultiply2Digit(buf+2,buf,buf+1);

 ___Interm(x,4,1);

 if(Add2Digit(x[1],x[2]))
   x[3]++;
 x[2]=x[1];
 if(Add2Digit(x[2],x[3]))
   x[3]++;
 if(Add2Digit(x[1],x[0]) && IS_0(++(x[2])) )
   x[3]++;

 ____Interm(x,4,1,buf,2,sgn);

 if(sgn)
   {
     if(Add2Digit(x[1],buf[2]) && IS_0(++(x[2])))
       x[3]++;

     if(Add2Digit(x[2],buf[3]))
       x[3]++;
   }
 else
   {
     BigNumDigit save;

     if(Sub2Digit(x[1],buf[2]) && IS_0((x[2])--))
       x[3]--;

     if(Sub2Digit(x[2],buf[3]))
       x[3]--;
   }
 
 __Verification(x,4,a,2,b,2,1,0,"ZBNMK2_2");
}

/* sizes of a and b are equal to 4. Size of x must be equal at least to 8
   Pure Karatsuba recursion...
   x must be set to zero
   Side effect: x=a*b
   Buffer must be allocated.
   */
#ifndef ZBNMK2_4
static void ZBNMK2_4(x,a,b,buf)
     BigNum x,a,b,buf;
{BigNumDigit sgn,*pt;

 _Verification(x,8,a,4,b,4,"ZBNMK2_4"); 
 pt=buf; 
 *(pt++)=a[2]; 
 *(pt++)=a[3]; 
 *(pt++)=b[2]; 
 *(pt++)=b[3]; 
 *(pt++)=0; 
 *(pt++)=0; 
 *(pt++)=0; 
 *(pt++)=sgn=0; 
 _Interm(a,b,buf,2,2);
 
#define IMK2_4(pt,a) if(1) { \
	if(Sub2Digit((pt)[0],(a)[0]) && (IS_0(((pt)[1])--))) { sgn= ~sgn; \
	     if(IS_0((pt)[0])) (pt)[1]=(~((-1) - (a)[1])+1); \
	     else { (pt)[0]= ~((pt)[0]) + 1; (pt)[1]=(~((-1) - (a)[1])) ; } } \
	else if(Sub2Digit((pt)[1],(a)[1])) { sgn=~sgn; \
	   if(IS_0((pt)[0])) (pt)[1]=(~((pt)[1])) + 1; \
	   else { (pt)[0]=~(pt)[0]+1; (pt)[1]=~((pt)[1]); } } } else

   { 
     BigNumDigit save; 
     IMK2_4(buf,a); 
     pt=buf+2; 
     IMK2_4(pt,b); 
   } 
 __Interm(buf,2); 


 ZBNMK2_2(x,a,b,buf+8);
 ZBNMK2_2(x+4,a+2,b+2,buf+8);
 ZBNMK2_2(buf+4,buf,buf+2,buf+8);

 ___Interm(x,8,2); 
 _Mix(x,(BigNumLength)2,(BigNumLength)4,(BigNumLength)4,(BigNumLength)2); 
 ____Interm(x,8,2,buf,4,sgn); 
 if(sgn) 
   _Add(x+2,(BigNumLength)6,buf+4,(BigNumLength)4,(BigNumDigit)0); 
 else 
   _Sub(x+2,(BigNumLength)6,buf+4,(BigNumLength)4,(BigNumDigit)1); 
 __Verification(x,8,a,4,b,4,2,0,"ZBNMK2_4");

}
#endif
/* sizes of a and b are equal to 3. Size of x must be equal at least to 6
   x must be set to zero
   Side effect: x=a*b
   Buffer must be allocated.
*/
#ifndef ZBNMK3_3
static void ZBNMK3_3(x,a,b,buf)
     BigNum x,a,b,buf;
{
  _Verification(x,6,a,3,b,3,"ZBNMK3_3");
  
  
  ZBNMK2_2(x,a,b,buf);

  ZBNMultiplyDigit(x+2,(BigNumLength)4,a,(BigNumLength)3,b[2]);
  ZBNMultiplyDigit(x+2,(BigNumLength)4,b,(BigNumLength)2,a[2]);

  __Verification(x,6,a,3,b,3,1,0,"ZBNMK3_3");
}
#endif

/* One must have al > 2
   Size of b is 2
   x must be set to zero and of size >= al+2
   Side effect: x = a*b
   Buffer must be allocated.
*/
static void _ZBNMK1_2(x,a,al,b,buf)
     BigNum x,a,b,buf;
     BigNumLength al;
{

 ZBNMK2_2(x,a,b,buf);

 do
   {
     al -= 2;
     a += 2;
     x += 2;

     if (al<2)
       {
	 ZBNMultiplyDigit(x,(BigNumLength)3,b,(BigNumLength)2,*a);
	 return;
       }
     ZBNMK2_2(x,a,b,buf);
          
   } while(al != 2);
}

/* One must have al > 3
   Size of b is 3
   x must be set to zero and of size >= al+3
   Side effect: x = a*b
   Buffer must be allocated.
*/
#ifndef ZBNMK1_3
static void ZBNMK1_3(x,a,al,b,buf)
     BigNum x,a,b,buf;
     BigNumLength al;
{BigNum tmp;
 BigNumLength tmpl;

 
 ZBNMK3_3(x,a,b,buf);
   
 do
   {
     al -= 3;
     a += 3;
     x += 3;

     if (al<3)
       {
	 if (al == 1) {
	   ZBNMultiplyDigit(x,(BigNumLength)4,b,(BigNumLength)3,*a);
	   return;
	 }
	 if (al == 2) {
	   ZBNMK1_2(x,b,(BigNumLength)3,a,buf);
	   return;
	 }
       }

     ZBNMK3_3(x,a,b,buf);
   } while(al != 4);
}
#endif

/* One must have al > 4
   Size of b is 4
   x must be set to zero and of size >= al+4
   Side effect: x = a*b
   Buffer must be allocated.
*/
#ifndef ZBNMK1_4
static void ZBNMK1_4(x,a,al,b,buf)
     BigNum x,a,b,buf;
     BigNumLength al;
{BigNum tmp;
 BigNumLength tmpl;

 
 ZBNMK2_4(x,a,b,buf);

 do
   {
     al -= 4;
     a += 4;
     x += 4;

     if (al<4)
       {
	 if (al == 1) {
	   ZBNMultiplyDigit(x,(BigNumLength)5,b,(BigNumLength)4,*a);
	   return;
	 }
	 if (al == 2) {
	   ZBNMK1_2(x,b,(BigNumLength)4,a,buf);
	   return;
	 }
	 if (al == 3) {
	   if(StopKara(2))
	     ZBNMultiplyPlain(x,(BigNumLength)7,b,(BigNumLength)4,a,(BigNumLength)3);
	   else
	     ZBNMK1_3(x,b,(BigNumLength)4,a,buf);
	   return;
	 }
       }
     ZBNMK2_4(x,a,b,buf);
   } while(al != 4);
}
#endif
