/*
 * zen/zbn/c/karasqu.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "kara_defaults.h"

#ifndef KARAs_LIM
static BigNumLength KARAs_LIM=1;
# define KKL 0
#else
# define KKL KARAs_LIM
#endif

#define StopKara(N) ((N) <= KARAs_LIM)

static void ZBNSK1 ___((BigNum,BigNum,BigNumLength,BigNum));
static void ZBNSK2 ___((BigNum,BigNum,BigNumLength,BigNum));

#if KKL < 4
  static void ZBNSK1_4 ___((BigNum,BigNum,BigNum));
#else
# define ZBNSK1_4(x,a,buf) ZBNSquarePlain(x,(BigNumLength)8,a,(BigNumLength)4)
#endif

#if KKL < 3
  static void ZBNSK2_3 ___((BigNum,BigNum,BigNum));
#else
# define ZBNSK2_3(x,a,buf) ZBNSquarePlain(x,(BigNumLength)6,a,(BigNumLength)3)
#endif

#if KKL < 2
  static void _ZBNSK1_2 ___((BigNum,BigNum,BigNum));
# ifdef KARAs_LIM
#  define ZBNSK1_2(x,a,buf) _ZBNSK1_2(x,a,buf) 
# else
#  define ZBNSK1_2(x,a,buf) if(StopKara(2)) ZBNSquarePlain(x,(BigNumLength)4,a,(BigNumLength)2); else _ZBNSK1_2(x,a,buf) 
# endif
#else
# define ZBNSK1_2(x,a,buf) ZBNSquarePlain(x,(BigNumLength)4,a,(BigNumLength)2)
#endif

#if ZBN_KARA_OPTS & SAss
# define _SglAssign(buf,a,lh,l2) if(1) { ZBNAssign(buf,a,lh); ZBNSetToZero(buf+lh,l2); } else
#else
# define _SglAssign(buf,a,lh,l2)  if(1) {BigNum pt1,pt2,pta,ptlim; \
   ptlim = pt2 = (lh) + (pt1 = buf); pta = a; \
   do { *(pt1++)= *(pta++); *(pt2++)= 0; *(pt2++) = 0; } while (pt1 < ptlim); \
 } else
#endif

#if KARA_DEBUG
#  define _Verification(x,xl,a,al,str)  BigNum bufinit,verif;  BigNumDigit vret; \
 if(KARA_DEBUG > 1) fprintf(stderr,"Entering %s...\n",str); \
 bufinit=ZBNC(xl); ZBNAssign(bufinit,x,xl); \
 verif=ZBNC(xl); ZBNAssign(verif,bufinit,xl); vret=ZBNSquarePlain(verif,xl,a,al)
# define __Verification(x,xl,a,al,N,ret,str) \
  if(vret != (ret)) fprintf(stderr,"%s: vret:= %d - %d ;\n",str,vret,ret); \
  if(ZBNCompare(verif,xl,x,xl)) \
    fprintf(stderr,"%s: Different result\n",str); \
  if((KARA_DEBUG >= 4) || (ZBNCompare(verif,xl,x,xl))) { \
     fprintf(stderr,"%s: ",str); \
     ZBNPrintToFile(stderr,bufinit,xl,10); \
     fprintf(stderr,"+("); \
     if(N) { ZBNPrintToFile(stderr,a,N,10); \
     fprintf(stderr," + 2^%d * ",(N)*SIZE_BLOC); } \
     ZBNPrintToFile(stderr,a+N,(al)-(N),10); \
     fprintf(stderr,") ^ 2 - "); \
     ZBNPrintToFile(stderr,x,xl,10); \
     fprintf(stderr,";\n "); \
     ZBNPrintToFile(stderr,verif,xl,10); \
     fprintf(stderr,"\n "); } \
  ZBNF(verif); ZBNF(bufinit); \
  if(KARA_DEBUG > 1) fprintf(stderr,"Exiting %s...\n",str)
#else
# define _Verification(x,xl,a,al,str)
# define __Verification(x,xl,a,al,N,ret,str) 
#endif
#if KARA_DEBUG > 1
# define _Interm(a,buf,N,lh) \
 fprintf(stderr,"ahl :="); \
 ZBNPrintToFile(stderr,buf,lh,10); \
 fprintf(stderr," - "); \
 ZBNPrintToFile(stderr,a,N,10); \
 fprintf(stderr,";\n")
# define __Interm(buf,lh) \
 fprintf(stderr,"abs(ahl) -"); \
 ZBNPrintToFile(stderr,buf,lh,10); \
 fprintf(stderr,";\n")
# define ___Interm(x,xl,N) \
 fprintf(stderr,"x := "); \
 ZBNPrintToFile(stderr,x,(N)<<1,10); \
 fprintf(stderr," + 2^%d * ",((N)<<1)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,x+((N)<<1),(xl)-((N)<<1),10); \
 fprintf(stderr,";\n"); \
 fprintf(stderr,"x := x + 2^%d * (",(N)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,x,((N)<<1),10); \
 fprintf(stderr," + "); \
 ZBNPrintToFile(stderr,(x)+((N)<<1),(xl)-((N)<<1),10); \
 fprintf(stderr,");\nx  - (")
# define ____Interm(x,xl,N,buf,l2) \
 ZBNPrintToFile(stderr,x,N,10); \
 fprintf(stderr," + 2^%d * ",(N)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,(x)+(N),N,10); \
 fprintf(stderr," + 2^%d * ",2*(N)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,(x)+2*(N),N,10); \
 fprintf(stderr," + 2^%d * ",3*(N)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,(x)+3*(N),N,10); \
 fprintf(stderr,");\n"); \
 fprintf(stderr,"ahl ^ 2 - "); \
 ZBNPrintToFile(stderr,(buf)+(l2),l2,10); \
 fprintf(stderr,";\n"); \
 fprintf(stderr,"x := x - 2^%d * ",(N)*SIZE_BLOC); \
 ZBNPrintToFile(stderr,(buf)+(l2),l2,10); \
 fprintf(stderr,";\n")
#else
# define _Interm(a,buf,N,lh)
# define __Interm(buf,N)
# define ___Interm(x,xl,N)
# define ____Interm(x,xl,N,buf,l2)
#endif

#ifndef KARAs_LIM
/* \TEX{
   \title{Karatsuba's squaring of two |BigNum|s.}
   \input{Two allocated |BigNum|s with $|xl| \ge 2|al|$ and the limit
   of size under which |ZBNSquare| must be used.}
   \output{The possible carry out, or |ZENERR| if an error occured.}
   \side{|x += a * a|}
   \note{A buffer is allocated and freed.}
   } */
BigNumDigit ZBNSquareKaratsuba(x,xl,a,al,lim)
     BigNum x,a;
     BigNumLength xl,al,lim;
{
#else
BigNumDigit ZBNSquareKaratsuba(x,xl,a,al)
     BigNum x,a;
     BigNumLength xl,al;
{
#endif
 BigNum z,tmp;
 BigNumLength tmpl;
 BigNumDigit ret;

 _Verification(x,xl,a,al,"ZBNSquareKaratsuba");

#ifndef KARAs_LIM
 KARAs_LIM=lim;
#endif
 
 if(StopKara(al))
#if PURE_KARA
   ret=ZBNMultiplyDigit(x,xl,a,(BigNumLength)1,a[0]);
#else
   ret=ZBNSquarePlain(x,xl,a,al);
#endif 
 else
   {
     tmpl=ZBNSizeBufKaraS(al);
     tmp=ZBNC(tmpl);
     
     if(!tmp)
       {
	 ZENSetError(ZEN_LIB_ZBN,ZEN_SQUKARA,ZEN_ALLOC);
	 return(ZERR);
       }
     
     z=ZBNC(xl);
     if(!z)
       {
	 ZENSetError(ZEN_LIB_ZBN,ZEN_SQUKARA,ZEN_ALLOC);
	 return(ZERR);
       }
     
     ZBNSetToZero(z,xl);

     if (al == 2) {
       ZBNSK1_2(z,a,tmp);
     }
     else if (al == 3) {
       ZBNSK2_3(z,a,tmp);
     }
     else if (al == 4) {
       ZBNSK1_4(z,a,tmp);
     }
     else {
       if(al&1)
	 ZBNSK2(z,a,al>>1,tmp);
       else
	 ZBNSK1(z,a,al>>1,tmp);
     }
     ret=ZBNAdd(x,xl,z,xl,(BigNumDigit)0);
     ZBNF(z);
     ZBNF(tmp);
   }

 __Verification(x,xl,a,al,0,ret,"ZBNSquareKaratsuba");

 return(ret);
}

#ifndef KARAs_LIM
/* \TEX{
   \title{Karatsuba's squaring}
   \input{Three |BigNum|s: $|x|=0$ of size greater or equal than $2\times|al|$, a buffer of size at least |ZBNSizeBufKaraM(bl)|, and the limit of size under which |ZBNMultiply| must be used. }
   \side{$|x|=|a|^2$}
   } */
void ZBNSquareKaratsubaBuffer(x,a,al,buf,lim)
     BigNum x,a,buf;
     BigNumLength al,lim;
{
#else
void ZBNSquareKaratsubaBuffer(x,a,al,buf)
     BigNum x,a,buf;
     BigNumLength al;
{
#endif
 _Verification(x,al<<1,a,al,"ZBNSquareKaratsubaBuffer");

#ifndef KARAs_LIM
  KARAs_LIM=lim;
#endif
  if(StopKara(al))
#if PURE_KARA
    ZBNSquare1Digit(x,a);
#else
    ZBNSquarePlain(x,al<<1,a,al);
#endif
  else {
    if (al == 2) {
      ZBNSK1_2(x,a,buf);
    }
    else if (al == 3) {
      ZBNSK2_3(x,a,buf);
    }
    else if (al == 4) {
      ZBNSK1_4(x,a,buf);
    }
    else {
      if(al & 1)
	ZBNSK2(x,a,al>>1,buf);
      else
	ZBNSK1(x,a,al>>1,buf);
    }
  }
  __Verification(x,al<<1,a,al,0,0,"ZBNSquareKaratsubaBuffer");
}

/* \TEX{
   \title{Size of buffer needed for Karatsuba's squaring}
   \input{The length of smallest operand and the limit under which standard multiplication is performed.}
   \output{The length of the buffer needed in Karatsuba's routines}
   } */
BigNumLength ZBNSizeBufKaraS(l)
     BigNumLength l;
{BigNumLength ret=0;
 
 for(;;)
   {
     l = (l&1) + (l>>1);
     ret += l << 2;
     if(l<=KARAs_LIM)
       return(ret);
   }
}

/******************************************************************/

/* sizes of a is equal to 2*N. Size of x must be equal at least to 4*N
   x must be set to zero
   Side effect: x=a*a
   Buffer must be allocated.
*/

static void ZBNSK1(x,a,N,buf)
     BigNum x,a,buf;
     BigNumLength N;
{BigNumLength l;

 _Verification(x,N<<2,a,N<<1,"ZBNSK1");

 l=N<<1;


 _SglAssign(buf,a+N,N,l);

 _Interm(a,buf,N,N);
 
 if(!(ZBNSubtract(buf,N,a,N,(BigNumDigit)1)))
   _CompAdd(buf,N);

 __Interm(buf,N);

 if(StopKara(N))
   {
#if PURE_KARA  
     ZBNSquare1Digit(x,a);
     ZBNSquare1Digit(x+2,a+N);
     ZBNSquare1Digit(buf+1,buf);
#else
     ZBNSquarePlain(x,l,a,N);
     ZBNSquarePlain(x+l,l,a+N,N);
     ZBNSquarePlain(buf+N,l,buf,N);
#endif
   }
 else
   {
     BigNumLength n;
     BigNum bl;

     bl=buf+N+l;
     
     if (N == 2) {
       ZBNSK1_2(x,a,bl);
       ZBNSK1_2(x+4,a+2,bl);
       ZBNSK1_2(buf+2,buf,bl);
     }
     else if (N == 3) {
       ZBNSK2_3(x,a,bl);
       ZBNSK2_3(x+6,a+3,bl);
       ZBNSK2_3(buf+3,buf,bl);
     }
     else if (N == 4) {
       ZBNSK1_4(x,a,bl);
       ZBNSK1_4(x+8,a+4,bl);
       ZBNSK1_4(buf+4,buf,bl);	
     }
     else {   
       n=N>>1;
       if(N&1)
	 {
	   ZBNSK2(x,a,n,bl);
	   ZBNSK2(x+l,a+N,n,bl);
	   ZBNSK2(buf+N,buf,n,bl);
	 }
       else
	 {
	   ZBNSK1(x,a,n,bl);
	   ZBNSK1(x+l,a+N,n,bl);
	   ZBNSK1(buf+N,buf,n,bl);
	 }
     }
   }

 ___Interm(x,N<<2,N);

 _Mix(x,N,l,l,N);

 ____Interm(x,N<<2,N,buf,N<<1);

 _Sub(x+N,l+N,buf+N,l,(BigNumDigit)1);
 
 __Verification(x,N<<2,a,N<<1,N,0,"ZBNSK1");
}

/* sizes of a is equal to 2. Size of x must be equal at least to 4
   x must be set to zero
   Side effect: x=a*a
   Buffer must be allocated.
*/

static void _ZBNSK1_2(x,a,buf)
     BigNum x,a,buf;
{
 _Verification(x,4,a,2,"ZBNSK1_2");

  {
    BigNumDigit save;

    buf[0]=a[1];		 
    if(Sub2Digit(buf[0],a[0]))
      buf[0]= 1 + ~(buf[0]);
  }
#if KARA_DEBUG > 1
  fprintf(stderr,"ahl :=");
  ZBNPrintToFile(stderr,buf,1,10);
  fprintf(stderr,";\n");
#endif
  
  buf[1]=buf[2]=0;
  
  ZBNSquare1Digit(x,a);
  ZBNSquare1Digit(x+2,a+1);
  ZBNSquare1Digit(buf+1,buf);

  ___Interm(x,4,1);

  if(Add2Digit(x[1],x[2]))
    x[3]++;
  x[2]=x[1];
  if(Add2Digit(x[2],x[3]))
    x[3]++;
  if(Add2Digit(x[1],x[0]) && IS_0(++(x[2])) )
    x[3]++;

 ____Interm(x,4,1,buf,2);

  {
    BigNumDigit save;
    
    if(Sub2Digit(x[1],buf[1]) && IS_0((x[2])--))
      x[3]--;
    
    if(Sub2Digit(x[2],buf[2]))
      x[3]--;
  }

 __Verification(x,4,a,2,1,0,"ZBNSK1_2");
}

/* sizes of a is equal to 4. Size of x must be equal at least to 8
   x must be set to zero
   Side effect: x=a*a
   Buffer must be allocated.
*/

#ifndef ZBNSK1_4
static void ZBNSK1_4(x,a,buf)
     BigNum x,a,buf;
{BigNumDigit *pt;

 _Verification(x,8,a,4,"ZBNSK1_4"); 
 pt=buf; 
 *(pt++)=a[2]; 
 *(pt++)=a[3]; 
 *(pt++)=0; 
 *(pt++)=0; 
 *(pt++)=0; 
 *(pt++)=0; 
 _Interm(a,buf,2,2); 
 { 
   BigNumDigit save; 
   if(Sub2Digit(buf[0],a[0]) && (IS_0((buf[1])--))) 
     {  
       if(IS_0(buf[0])) 
	 buf[1]=(~((-1) - a[1])+1);  
       else 
	 { 
	   buf[0]= ~(buf[0]) + 1; 
	   buf[1]=(~((-1) - a[1])) ; 
	 } 
     }  
   else if(Sub2Digit(buf[1],a[1])) 
     { 
       if(IS_0(buf[0])) 
	 buf[1]=(~(buf[1])) + 1; 
       else 
	 { 
	   buf[0]=~(buf[0])+1; 
	   buf[1]=~(buf[1]); 
	 } 
     } 
 } 
 __Interm(buf,2);

 
 ZBNSK1_2(x,a,buf+6);
 ZBNSK1_2(x+4,a+2,buf+6);
 ZBNSK1_2(buf+2,buf,buf+6);
 
 ___Interm(x,8,2); 
 _Mix(x,(BigNumLength)2,(BigNumLength)4,(BigNumLength)4,(BigNumLength)2); 
 ____Interm(x,8,2,buf,4); 
 _Sub(x+2,(BigNumLength)6,buf+2,(BigNumLength)4,(BigNumDigit)1); 
 __Verification(x,8,a,4,2,0,"ZBNSK1_4");
   
}
#endif

/* sizes of a is equal to 3. Size of x must be equal at least to 6
   x must be set to zero
   Side effect: x=a*a
   Buffer must be allocated.
*/
#ifndef ZBNSK2_3
static void ZBNSK2_3(x,a,buf)
     BigNum x,a,buf;
{
 _Verification(x,6,a,3,"ZBNSK2_3");

 ZBNSK1_2(x,a,buf);
 
#if SK2_NOSQ
 ZBNMultiplyDigit(x+2,(BigNumLength)4,a,(BigNumLength)3,a[2]);  
 ZBNMultiplyDigit(x+2,(BigNumLength)4,a,(BigNumLength)2,a[2]); 
#else
 ZBNSetToZero(buf,(BigNumLength)4); 
 ZBNMultiplyDigit(buf,(BigNumLength)3,a,(BigNumLength)2,a[2]); 
 ZBNShiftLeft(buf,(BigNumLength)4,1); 
 ZBNAdd(x+2,(BigNumLength)4,buf,(BigNumLength)4,(BigNumDigit)0); 
 ZBNSquare1Digit(x+4,a+2); 
#endif
   
 __Verification(x,6,a,3,1,0,"ZBNSK2_3");
}
#endif

/* sizes of a is equal to 2*N+1. Size of x must be equal at least to 4*N+2
   x must be set to zero
   Side effect: x=a*a
   Buffer must be allocated.
*/
static void ZBNSK2(x,a,N,buf)
     BigNum x,a,buf;
     BigNumLength N;
#if ZBN_KARA_OPTS & SK2
{
  BigNumLength l,lt,dlt;
  _Verification(x,(N<<2)+2,a,(N<<1)+1,"ZBNSK2");

  dlt=((lt=(l=(N<<1))+1)<<1);

  if (N == 1) {
    ZBNSK1_2(x,a,buf);
  }
  else if (N == 2) {
    ZBNSK1_4(x,a,buf);
  }
  else {
    ZBNSK1(x,a,N,buf);
  }
# if SK2_NOSQ
  ZBNMultiplyDigit((x)+(l),dlt,a,lt,a[l]); 
  ZBNMultiplyDigit((x)+(l),dlt,a,l,a[l]);
# else
  ZBNSetToZero(buf,lt+1);
  ZBNMultiplyDigit(buf,lt,a,l,a[l]);
  ZBNShiftLeft(buf,lt+1,1);
  ZBNAdd(x+l,dlt,buf,lt+1,(BigNumDigit)0);
  buf[0] = buf[1] = (BigNumDigit)0;
  ZBNSquare1Digit(x+(l<<1), a+l);
# endif
  __Verification(x,(N<<2)+2,a,(N<<1)+1,N,0,"ZBNSK2");
}
#else  /* SK2 */
{
  BigNumLength l,lh,l2;

  _Verification(x,(N<<2)+2,a,(N<<1)+1,"ZBNSK2");

  l=N<<1;
  lh=N+1;
  l2=lh<<1;
  
  _SglAssign(buf,a+N,lh,l2);
  
  _Interm(a,buf,N,lh);
  
  if(!(ZBNSubtract(buf,lh,a,N,(BigNumDigit)1)))
    _CompAdd(buf,lh);
  
  __Interm(buf,lh);
  
# if PURE_KARA
  {
    BigNumLength n;
    BigNum bl;
    
    bl=buf+lh+l2;
    
    if (N == 4) {
      ZBNSK1_4(x,a,bl);
      
      ZBNSK2(x+8,a+4,(BigNumLength)2,bl);
      ZBNSK2(buf+5,buf,(BigNumLength)2,bl);
    }
    else if (N == 3) {
      ZBNSK2_3(x,a,bl);
      ZBNSK1_4(x+6,a+3,bl);
      ZBNSK1_4(buf+4,buf,bl);
    }
    else if (N == 2) {
      ZBNSK1_2(x,a,bl);
      ZBNSK2_3(x+4,a+2,bl);
      ZBNSK2_3(buf+3,buf,bl);
    }
    else if (N == 1) {
      /* N=1 lh=2 l=2 l2=4 n=1 */
      ZBNSquare1Digit(x,a);
      
      ZBNSK1_2(x+2,a+1,bl);
      ZBNSK1_2(buf+2,buf,bl);
    }
    else {
      n= lh>>1;
      if(lh & 1)
	{
	  ZBNSK1(x,a,n,bl);
	  
	  ZBNSK2(x+l,a+N,n,bl);
	  ZBNSK2(buf+lh,buf,n,bl);
	}
      else
	{
	  ZBNSK2(x,a,N>>1,bl);
	  
	  ZBNSK1(x+l,a+N,n,bl);
	  ZBNSK1(buf+lh,buf,n,bl);
	}
    }
  }
# else
  if(StopKara(lh))
    {
      ZBNSquarePlain(x,l,a,N);
      ZBNSquarePlain(x+l,l2,a+N,lh);
      ZBNSquarePlain(buf+lh,l2,buf,lh);
    }
  else
    {
      BigNumLength n;
      BigNum bl;
      
      bl=buf+lh+l2;
      
    if (N == 4) {
      if(StopKara(4))
	ZBNSquarePlain(x,(BigNumLength)8,a,(BigNumLength)4);
      else
	ZBNSK1_4(x,a,bl);
      
      ZBNSK2(x+8,a+4,(BigNumLength)2,bl);
      ZBNSK2(buf+5,buf,(BigNumLength)2,bl);
    }
    else if (N == 3) {
      if(StopKara(3))
	ZBNSquarePlain(x,(BigNumLength)6,a,(BigNumLength)3);
      else
	ZBNSK2_3(x,a,bl);
      
      ZBNSK1_4(x+6,a+3,bl);
      ZBNSK1_4(buf+4,buf,bl);
    }
    else if (N == 2) {
      ZBNSK1_2(x,a,bl);
      ZBNSK2_3(x+4,a+2,bl);
      ZBNSK2_3(buf+3,buf,bl);
    }
    else if (N == 1) {
      ZBNSquare1Digit(x,a);
      ZBNSK1_2(x+2,a+1,bl);
      ZBNSK1_2(buf+2,buf,bl);
    }
    else {
      n= lh>>1;
      if(lh & 1)
	{
	  if(StopKara(N))
	    ZBNSquarePlain(x,l,a,N);
	  else
	    ZBNSK1(x,a,n,bl);
	  
	  ZBNSK2(x+l,a+N,n,bl);
	  ZBNSK2(buf+lh,buf,n,bl);
	}
      else
	{
	  if(StopKara(N))
	    ZBNSquarePlain(x,l,a,N);
	  else
	    ZBNSK2(x,a,N>>1,bl);
	  
	  ZBNSK1(x+l,a+N,n,bl);
	  ZBNSK1(buf+lh,buf,n,bl);
	}
	}
    }
# endif
  
  ___Interm(x,(N<<2)+2,N);
  
  _Mix(x,N,l,l2,lh+1);
  
  ____Interm(x,(N<<2)+2,N,buf,(N<<1)+2);
  
  _Sub(x+N,l2+N,buf+lh,l2,(BigNumDigit)1);
  
  __Verification(x,(N<<2)+2,a,(N<<1)+1,N,0,"ZBNSK2");
}
#endif /* SK2 */

