/*
 * zen/zbn/c/karatest.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "kara_defaults.h"

char *Txt[]={
  "Karatsuba's bounds.",
  "-------------------",
  "",
  "This program tests the Karatsuba procedures and sets the limit of use",
  "as compared with standard algorithms. This setting may differ from",
  "machine to machine, even under the same architecture.",
  "",
  "The program tests sizes using dichotomic way. Therefore, it can",
  "take a while. The result is then incorporated in new C source files for",
  "recompilation.",
  "",
  NULL};

char *TxtSet[]={
  "Karatsuba's cutoff.",
  "-------------------",
  "",
  "This program tests the Karatsuba procedures and sets the cutoff to",
  "the optimal size for a given machine. This setting may differ from",
  "machine to machine, even under the same architecture.",
  "",
  "The program tests a number of sizes and takes the average optimal cutof.",
  "The result is then incorporated in new C source files for recompilation.",
  "",
  NULL};

# define NBSIZES 40
int karatestsizes[NBSIZES]={16,20,23,24,25,30,31,32,33,34,35,40,47,48,49,50,55,60,63,64,65,75,100,125,128,150,175,200,225,250,256,275,300,384,500,512,760,768,1000,1024};

/*******************************************************************/
#define PrintTxt(txt) PrintfTxt(stdout,txt)
static void PrintfTxt(fd,txt)
     char *txt[];
     FILE *fd;
{int i=0;

 while(txt[i] != NULL)
   fprintf(fd,"%s\n",txt[i++]);
}

/*******************************************************************/
static int SetLimit(fct, ret, min)
     BigNumLength (*fct) ___((BigNumLength));
     int ret, min;
{
  int max,i,j,k;

  for(j = 0, max = (min<<1);
      ;
      max <<= 1) {

    fprintf(stderr, "Size %d (%d <= size <= %d):\n", max, min, max);
    i = (int)fct((BigNumLength)max);
    fprintf(stderr,"\n");

    if(i != ret) {
      if(j == 0) {
	j = 1;
	continue;
      }
      else {
	break;
      }
    }
    if(j == 0) {
      min = max;
    }
  }
  
  for(k = -1;;) {
    j = (max+min)/2;
    if ((j == k) || (max <= min)) {
      return(j);
    }

    fprintf(stderr,"Size %d (%d <= size <= %d):\n", j, min, max);
    i = (int)fct((BigNumLength)j);
    fprintf(stderr,"\n");

    if (i != ret) {
      max = j;
    }
    else if (i >= 0) {
      min = j;
    }
    k = j;
  }   
}

/**********************************************************************/
static int SetCutoff(fct)
     BigNumLength (*fct) ___((BigNumLength));
{
  int i,n,t,ret;

  for(i=n=t=0;i<NBSIZES;i++)
    {
      fprintf(stderr, "Iteration %d/%d (size %d):\n", i+1, NBSIZES, karatestsizes[i]);
      ret=(int)fct((BigNumLength)karatestsizes[i]);
      fprintf(stderr, "\n");
      if(ret>0)
	{
	  for(ret>>=1;ret;ret>>=1)
	    t++;
	  n++;
	}
    }
  i = t/n;
  if((t%n)>=(n/2))
    i++;
  i=BlocExp2(i);
  fprintf(stderr,"         ----> Karatsuba's cutoff: %d BigNumDigits (%d bits).\n\n",
	  i, i*SIZE_BLOC);
  return(i);

}

/*******************************************************************/  
int main(argc,argv)
     int argc;
     char **argv;
{
  int ret;

#ifdef KARA_LIM_SET
  PrintfTxt(stderr,TxtSet);

  ret=SetCutoff(ZBNKaratsubaMultiplyCutoff);
  if(ret>0)
    printf("#define KARAm_LIM %d\n",ret);
  ret=SetCutoff(ZBNKaratsubaSquareCutoff);
  if(ret>0)
    printf("#define KARAs_LIM %d\n",ret);
#else
  PrintfTxt(stderr,Txt);

  printf("#define KARAm_LIM %d\n", KARAm_LIM);
  printf("#define KARAs_LIM %d\n", KARAs_LIM);
  ret=SetLimit(ZBNKaratsubaMultiplyCutoff,0,1);
  if(ret>0)
    printf("#define KARAm_BOUND %d\n",ret);
  fprintf(stderr,"      ----> Karatsuba's multiplication used from %d BigNumDigits (%d bits).\n\n", ret, ret*SIZE_BLOC);
  ret=SetLimit(ZBNKaratsubaSquareCutoff,0,1);
  if(ret>0)
    printf("#define KARAs_BOUND %d\n",ret);
  fprintf(stderr,"      ----> Karatsuba's squaring used from %d BigNumDigits (%d bits).\n\n",ret, ret*SIZE_BLOC);
#endif

  exit(0);
}
