/*
 * zen/zbn/c/log.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include <math.h>
#include "zbn.h"

/* \TEX{
   \Subsection{Logarithms}
   } */

/* \TEX{
   \title{Logarithm in base $e$ of a |BigNum|}
   \input{An allocated |BigNum| |n| of length |nl|.}
   \output{The logarithm in base $e$ of (|n|, |nl|)}.   
   } */
double ZBNLog(n, nl)
     BigNum n;
     BigNumLength nl;
{
  double lnz;
  BigNumDigit d;
#ifdef M_LN2
# undef M_LN2
#endif
#define M_LN2	                0.69314718055994530942

  if(nl == 1)
    lnz = log((double)ZBNGetDigit(n));
  else{
    /* n < (b[nl-1]+1)*B^(nl-1) */
    d = 1 + ZBNGetDigit(n+(nl-1));
    if(!d){
      /* case where d=2^BN_DIGIT_SIZE-1 */
      lnz = ((double)nl*BN_DIGIT_SIZE*M_LN2);
    }
    else{
      lnz = ((double)(nl-1)*BN_DIGIT_SIZE*M_LN2);
      lnz += log((double)(d+1));
    }
  }
  return(lnz);
}
