/*
 * zen/zbn/c/mul-new.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

# define zbnadd(c1, c0, b0)			\
{						\
  BigNumDigit _foo0 = c0;			\
  BigNumDigit _foo1 = c1;			\
						\
__asm__("addl %2,%1\nadcl $0x0,%0"		\
	: "=g,g"  ((BigNumDigit)(c1)),		\
	  "=g,r" ((BigNumDigit)(c0))		\
	: "r,g"   ((BigNumDigit)(b0)),		\
	"0,0"  ((BigNumDigit)(_foo1)),		\
	  "1,1" ((BigNumDigit)(_foo0)));		\
  }
/* \TEX{
   \Subsection{Multiplication}
   } */

#if 0
/* (h, m ) = (0, m) + n * d  + (0, c) */
#define zbnmulladd(h, m, n, d, c)							\
{											\
  int __dummy;										\
__asm__ const ("mull %3\n addl %5, %%eax\n adcl $0,%1\n addl %%eax,%0\n adcl $0,%1"	\
	       : "=g"      ((BigNumDigit)(m)),						\
	         "=&d"       ((BigNumDigit)(c)),					\
	         "=a"       ((BigNumDigit)(__dummy))					\
	       : "rm"       ((BigNumDigit)(d)),						\
	         "2"        ((BigNumDigit)(n)),						\
	         "ri"       ((BigNumDigit)(c)));					\
}

static BigNumDigit ZBNMultiplyEven(m, n, nl, d)
     BigNum m, n;
     BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit c;

  c = 0;
  do {
    zbnmulladd(c, m[nl], n[nl], d, c);
    zbnmulladd(c, m[nl+1], n[nl+1], d, c);
    nl += 2; 
  } while (nl);

  return(c);
}

static BigNumDigit ZBNMultiplyOdd(m, n, nl, d)
     BigNum m, n;
     BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit c;

  c = 0;
  zbnmulladd(c, m[nl], n[nl], d, c);
  nl++;

  if (nl) {
    do {
      zbnmulladd(c, m[nl], n[nl], d, c);
      zbnmulladd(c, m[nl+1], n[nl+1], d, c);
      nl += 2;
    } while (nl);
  }
  return(c);
}

#else
static BigNumDigit ZBNMultiplyEven(m, n, nl, d)
     BigNum m, n;
     BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit l, h, c;

  c = 0;
  do {
    l = n[nl];
    zbnmul(h, l, l, d);
    zbnadd(h, l, c);
    zbnadd(h, m[nl], l);
    
    l = n[nl+1];
    zbnmul(c, l, l, d);
    zbnadd(c, l, h);
    zbnadd(c, m[nl+1], l);

    nl += 2; 
  } while (nl);

  return(c);
}

static BigNumDigit ZBNMultiplyOdd(m, n, nl, d)
     BigNum m, n;
     BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit l, h;
  BigNumDigit c;

  zbnmul(c, l, n[nl], d);
  zbnadd(c, m[nl], l);
  nl++;

  if (nl) {
    do {
      zbnmul(h, l, n[nl], d);
      zbnadd(h, l, c);
      zbnadd(h, m[nl], l);

      zbnmul(c, l, n[nl+1], d);
      zbnadd(c, l, h);
      zbnadd(c, m[nl+1], l);

      nl += 2;
    } while (nl);
  }
  return(c);
}

#endif

#define ZBNMultiplyDigitHeart(m, n, d, c, h, l, nl)	\
do {							\
  l = n[-1];						\
  zbnmul(h, l, l, d);					\
  zbnadd(h, l, c);					\
  zbnadd(h, m[-1], l);					\
} while(0)

static BigNumDigit ZBNMultiplyDigit1(m, n, d)
     BigNum m;
     BigNum n;
     BigNumDigit d;
{
  BigNumDigit l, c;

  l = n[-1]; zbnmul(c, l, l, d); zbnadd(c, m[-1], l);
  return(c);
}

static BigNumDigit ZBNMultiplyDigit2(m, n, d)
     BigNum m;
     BigNum n;
     BigNumDigit d;
{
  BigNumDigit l, h, c;

  l = n[-2]; zbnmul(h, l, l, d); zbnadd(h, m[-2], l);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -1);
  return(c);
}

static BigNumDigit ZBNMultiplyDigit3(m, n, d)
     BigNum m;
     BigNum n;
     BigNumDigit d;
{
  BigNumDigit l, h, c;

  l = n[-3]; zbnmul(h, l, l, d); zbnadd(h, m[-3], l);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -2);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -1);
  return(h);
}

static BigNumDigit ZBNMultiplyDigit4(m, n, d)
     BigNum m;
     BigNum n;
     BigNumDigit d;
{
  BigNumDigit l, h, c;

  l = n[-4]; zbnmul(h, l, l, d); zbnadd(h, m[-4], l);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -3);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -2);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -1);
  return(c);
}

static BigNumDigit ZBNMultiplyDigit5(m, n, d)
     BigNum m;
     BigNum n;
     BigNumDigit d;
{
  BigNumDigit l, h, c;

  l = n[-5]; zbnmul(h, l, l, d); zbnadd(h, m[-5], l);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -4);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -3);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -2);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -1);
  return(h);
}

static BigNumDigit ZBNMultiplyDigit6(m, n, d)
     BigNum m;
     BigNum n;
     BigNumDigit d;
{
  BigNumDigit l, h, c;

  l = n[-6]; zbnmul(h, l, l, d); zbnadd(h, m[-6], l);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -5);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -4);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -3);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -2);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -2);
  return(c);
}

static BigNumDigit ZBNMultiplyDigit7(m, n, d)
     BigNum m;
     BigNum n;
     BigNumDigit d;
{
  BigNumDigit l, h, c;

  l = n[-7]; zbnmul(h, l, l, d); zbnadd(h, m[-7], l);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -6);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -5);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -4);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -3);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -2);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -1);
  return(h);
}

static BigNumDigit ZBNMultiplyDigit8(m, n, d, c)
     BigNum m;
     BigNum n;
     BigNumDigit d;
     BigNumDigit c;
{
  BigNumDigit l, h;

  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -8);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -7);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -6);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -5);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -4);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -3);
  ZBNMultiplyDigitHeart(m, n, d, h, c, l, -2);
  ZBNMultiplyDigitHeart(m, n, d, c, h, l, -1);
  return(h);
}

/* \TEX{
   \title{Multiplication by a digit}
   \input{Two allocated |BigNum|s and a digit.}
   \output{A carry out.}
   \side{$|(n,nl)| += |(m,ml)| \times |d|$.}
   \note{One must have $|nl| \ge |ml|+1$.  If |ml| equals zero, the
   return is 0 and nothing is done.  }
   } */
BigNumDigit ZBNMultiplyDigit(m, ml, n, nl, d)
     BigNum m; BigNumLength ml;
     BigNum n; BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit carry, c;

  if (!d) {
    return(CST_BNDIGIT(0));
  }

  if (d == CST_BNDIGIT(1)) {
    return(ZBNAdd(m, ml, n, nl, CST_BNDIGIT(0)));
  }

  if (nl) {
    m += nl; n += nl;
    nl *= CST_BNLENGTH(-1); ml += nl;

    switch(nl%8) {
    case CST_BNLENGTH(-7):
      carry = ZBNMultiplyDigit7(m, n, d);
      nl += 7;
      break;
    case CST_BNLENGTH(-6):
      carry = ZBNMultiplyDigit6(m, n, d);
      nl += 6;
      break;
    case CST_BNLENGTH(-5):
      carry = ZBNMultiplyDigit5(m, n, d);
      nl += 5;
      break;
    case CST_BNLENGTH(-4):
      carry = ZBNMultiplyDigit4(m, n, d);
      nl += 4;
      break;
    case CST_BNLENGTH(-3):
      carry = ZBNMultiplyDigit3(m, n, d);
      nl += 3;
      break;
    case CST_BNLENGTH(-2):
      carry = ZBNMultiplyDigit2(m, n, d);
      nl += 2;
      break;
    case CST_BNLENGTH(-1):
      carry = ZBNMultiplyDigit1(m, n, d);
      nl += 1;
      break;
    }
    
    if (nl) {
      do {
	carry = ZBNMultiplyDigit8(&(m[nl]), &(n[nl]), d, carry);
	nl += 8;
      } while(nl);
    }

    d = carry;
    zbnadd(carry, m[0], d);
    m += ml; ml--;
  }
  else {
    m += ml;
  }
  if (ml && carry) {
    ml *= CST_BNLENGTH(-1);
    do {
      if (++m[ml]) {break;}
      ml++;
    } while(ml);
    carry = ml ? 0 : 1;
  }
  return(carry);
}

#if 0
/* \TEX{
   \title{Multiplication and Add}
   \input{Three allocated |BigNum|s and their lengths}
   \output{A carry out.}
   \side{$|(p,pl)| += |(m,ml)| \times |(n,nl)|$.}
   \note{One must have $|pl| \ge |ml|+|nl|$. If |nl| equals zero,
   there is no side effect and the carry out is 0.}
   } */
BigNumDigit ZBNMultiply(p, pl, m, ml, n, nl)
     BigNum p, m, n;
     BigNumLength pl, ml, nl; {}
#endif
BigNumDigit ZBNMultiplyPlain(p, pl, m, ml, n, nl)
     BigNum p; BigNumLength pl;
     BigNum m; BigNumLength ml;
     BigNum n; BigNumLength nl;
{
  BigNumDigit carry;
  
  carry = 0;
  while (nl-- > 0) {
    carry += ZBNMultiplyDigit(p, pl, m, ml, *n);
    p++; n++; pl--;
  }
  return(carry);
}

/* \TEX{
   \title{Multiplication}
   \input{Three |BigNum|s and their lengths}
   \side{$|(n,nl)| \leftarrow |(a,al)| \times |(b,bl)|$.}
   \note{One must have $|nl| \ge |al|+|bl|$, |al| and |bl| non zeros.}
   } */
void ZBNMult(n, a, al, b, bl)
     BigNum n, a, b;
     BigNumLength al, bl;
{
  ZBNSetToZero(n, al+bl);
  ZBNMultiply(n, al+bl, a, al, b, bl);
}

#if 0
/* \TEX{
   \title{Squaring and Add}
   \input{Two allocated |BigNum|s and their lengths}
   \output{A carry out.}
   \side{$|(n,nl)| += |(a,al)|^2$.}
   \note{One must have $|nl| \ge 2|al|$. If |al| equals zero, there
   is no side effect and the carry out is 0.}
   } */
BigNumDigit ZBNSquare(n, nl, a, al)
     BigNum n, a;
     BigNumLength nl,al; {}
#endif
BigNumDigit ZBNSquarePlain(p, pl, n, nl)
     BigNum p; BigNumLength pl;
     BigNum n; BigNumLength nl;
{
  BigNumDigit carry;
  BigNumLength ml;
  BigNum m;
  
  m = n; ml = nl; carry = 0;
  while (nl-- > 0) {
    carry += ZBNMultiplyDigit(p, pl, m, ml, *n);
    p++, n++, pl--;
  }
  return(carry);
}

/* \TEX{
   \title{Squaring}
   \input{Two |BigNum|s and a length}
   \side{$|(n,nl)| \leftarrow |(a,al)|^2 $.}
   \note{One must have |n| distinct from |a|, $|nl| \ge 2|al|$, and
   $|al| > 0$.}
   } */
void ZBNSqu(n, a, al)
     BigNum n, a;
     BigNumLength al;
{
  ZBNSetToZero(n, al<<1);
  ZBNSquare(n, al<<1, a, al);
}


static BigNumDigit ZBNMultiplyEven(m, n, nl, d)
     BigNum m, n;
     BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit l, h, c;

  c = 0;
  do {
    l = n[nl];
    zbnmul(h, l, l, d);
    zbnadd(h, l, c);
    zbnadd(h, m[nl], l);
    
    l = n[nl+1];
    zbnmul(c, l, l, d);
    zbnadd(c, l, h);
    zbnadd(c, m[nl+1], l);

    nl += 2; 
  } while (nl);

  return(c);
}

static BigNumDigit ZBNMultiplyOdd(m, n, nl, d)
     BigNum m, n;
     BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit l, h;
  BigNumDigit c;

  zbnmul(c, l, n[nl], d);
  zbnadd(c, m[nl], l);
  nl++;

  if (nl) {
    do {
      zbnmul(h, l, n[nl], d);
      zbnadd(h, l, c);
      zbnadd(h, m[nl], l);

      zbnmul(c, l, n[nl+1], d);
      zbnadd(c, l, h);
      zbnadd(c, m[nl+1], l);

      nl += 2;
    } while (nl);
  }
  return(c);
}
