/*
 * zen/zbn/c/shift.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

 
#ifdef USE_GMP
# include <gmp.h>
#endif

/* \TEX{
   \Subsubsection{Shifting}
   } */

#ifndef USE_GMP
# if (SIZE_BLOC == 32) && !defined(sunossparc)
typedef BigNumDigit (*ZBNShiftLeftType) ___((BigNum, BigNumLength));

#define ZBNShiftLeftTemplate(name, n, nl, l)	\
static BigNumDigit name(n, nl)				\
     BigNum n;					\
     BigNumLength nl;				\
{						\
  BigNumDigit e, d;				\
						\
  if ((!nl)) {return(CST_BNDIGIT(0));}	\
  d = 0; n += nl; nl *= CST_BNLENGTH(-1);	\
  do {						\
    e = n[nl] >> (SIZE_BLOC-l);			\
    n[nl] <<= l;				\
    n[nl] |= d;					\
    d = e;					\
    nl++;					\
  } while(nl);					\
  return(d);					\
}

static BigNumDigit zbnslt0(n, nl) BigNum n; BigNumLength nl; {return(CST_BNDIGIT(0));}
ZBNShiftLeftTemplate(zbnslt1, n, nl, 1)
ZBNShiftLeftTemplate(zbnslt2, n, nl, 2)
ZBNShiftLeftTemplate(zbnslt3, n, nl, 3)
ZBNShiftLeftTemplate(zbnslt4, n, nl, 4)
ZBNShiftLeftTemplate(zbnslt5, n, nl, 5)
ZBNShiftLeftTemplate(zbnslt6, n, nl, 6)
ZBNShiftLeftTemplate(zbnslt7, n, nl, 7)
ZBNShiftLeftTemplate(zbnslt8, n, nl, 8)
ZBNShiftLeftTemplate(zbnslt9, n, nl, 9)
ZBNShiftLeftTemplate(zbnslt10, n, nl, 10)
ZBNShiftLeftTemplate(zbnslt11, n, nl, 11)
ZBNShiftLeftTemplate(zbnslt12, n, nl, 12)
ZBNShiftLeftTemplate(zbnslt13, n, nl, 13)
ZBNShiftLeftTemplate(zbnslt14, n, nl, 14)
ZBNShiftLeftTemplate(zbnslt15, n, nl, 15)
ZBNShiftLeftTemplate(zbnslt16, n, nl, 16)
ZBNShiftLeftTemplate(zbnslt17, n, nl, 17)
ZBNShiftLeftTemplate(zbnslt18, n, nl, 18)
ZBNShiftLeftTemplate(zbnslt19, n, nl, 19)
ZBNShiftLeftTemplate(zbnslt20, n, nl, 20)
ZBNShiftLeftTemplate(zbnslt21, n, nl, 21)
ZBNShiftLeftTemplate(zbnslt22, n, nl, 22)
ZBNShiftLeftTemplate(zbnslt23, n, nl, 23)
ZBNShiftLeftTemplate(zbnslt24, n, nl, 24)
ZBNShiftLeftTemplate(zbnslt25, n, nl, 25)
ZBNShiftLeftTemplate(zbnslt26, n, nl, 26)
ZBNShiftLeftTemplate(zbnslt27, n, nl, 27)
ZBNShiftLeftTemplate(zbnslt28, n, nl, 28)
ZBNShiftLeftTemplate(zbnslt29, n, nl, 29)
ZBNShiftLeftTemplate(zbnslt30, n, nl, 30)
ZBNShiftLeftTemplate(zbnslt31, n, nl, 31)

static ZBNShiftLeftType ZBNShiftLeftTabular[32] = {
  zbnslt0, zbnslt1, zbnslt2, zbnslt3, zbnslt4, zbnslt5, zbnslt6, zbnslt7,
  zbnslt8, zbnslt9, zbnslt10, zbnslt11, zbnslt12, zbnslt13, zbnslt14, zbnslt15,
  zbnslt16, zbnslt17, zbnslt18, zbnslt19, zbnslt20, zbnslt21, zbnslt22, zbnslt23,
  zbnslt24, zbnslt25, zbnslt26, zbnslt27, zbnslt28, zbnslt29, zbnslt30, zbnslt31
};

/* \TEX{
   \Subsubsection{Shifting a |BigNum|}
   } */

/* \TEX{
   \title{Shifting left of a small amount}
   \input{An allocated |BigNum|, its length and a positive integer
   $0 \le |l| < |SIZE_BLOC|$.}
   \output{The digit shifted out}
   \side{$|(n,nl)| <<= |l|$}
   \note{If |l| equals zero, nothing is done and the returned value
   is 0.}
   } */
BigNumDigit ZBNShiftLeft(n, nl, l)
     BigNum n;
     BigNumLength nl;
     BigNumLength l;
{
  return((ZBNShiftLeftTabular[l])(n, nl));
}
# else
BigNumDigit ZBNShiftLeft(n, nl, l)
     BigNum n;
     BigNumLength nl;
     BigNumLength l;
{
  BigNumLength r;
  BigNumDigit e, d;

  if ((!nl) || (!l)) {return(CST_BNDIGIT(0));}
  r = SIZE_BLOC-l;
  d = 0; n += nl; nl *= CST_BNLENGTH(-1);
  do {
    e = n[nl] >> r;
    n[nl] <<= l;
    n[nl] |= d;
    d = e;
    nl++;
  } while(nl);
  return(d);
}
# endif
#else
BigNumDigit ZBNShiftLeft(n, nl, l)
     BigNum n;
     BigNumLength nl;
     BigNumLength l;
{
  return((l) ? (BigNumDigit)(mpn_lshift(n, n, nl, l)) : (BigNumDigit)(0));
}
#endif

#ifndef USE_GMP
# if (SIZE_BLOC == 32) && !defined(sunossparc)
typedef BigNumDigit (*ZBNShiftRightType)(BigNum, BigNumLength);

#define ZBNShiftRightTemplate(name, n, nl, r)	\
static BigNumDigit name(n, nl)			\
     BigNum n;					\
     BigNumLength nl;				\
{						\
  BigNumDigit d, e;				\
						\
  if ((!nl) || (!r)) {return(CST_BNDIGIT(0));}	\
  d = 0;					\
  nl --;					\
  do {						\
    e = n[nl] << (SIZE_BLOC-r);			\
    n[nl] >>= r;				\
    n[nl] |= d;					\
    d = e;					\
    nl--;					\
  } while(nl >= 0);				\
  return(d);					\
}

static BigNumDigit zbnsrt0(n, nl) BigNum n; BigNumLength nl; {return(CST_BNDIGIT(0));}
ZBNShiftRightTemplate(zbnsrt1, n, nl, 1)
ZBNShiftRightTemplate(zbnsrt2, n, nl, 2)
ZBNShiftRightTemplate(zbnsrt3, n, nl, 3)
ZBNShiftRightTemplate(zbnsrt4, n, nl, 4)
ZBNShiftRightTemplate(zbnsrt5, n, nl, 5)
ZBNShiftRightTemplate(zbnsrt6, n, nl, 6)
ZBNShiftRightTemplate(zbnsrt7, n, nl, 7)
ZBNShiftRightTemplate(zbnsrt8, n, nl, 8)
ZBNShiftRightTemplate(zbnsrt9, n, nl, 9)
ZBNShiftRightTemplate(zbnsrt10, n, nl, 10)
ZBNShiftRightTemplate(zbnsrt11, n, nl, 11)
ZBNShiftRightTemplate(zbnsrt12, n, nl, 12)
ZBNShiftRightTemplate(zbnsrt13, n, nl, 13)
ZBNShiftRightTemplate(zbnsrt14, n, nl, 14)
ZBNShiftRightTemplate(zbnsrt15, n, nl, 15)
ZBNShiftRightTemplate(zbnsrt16, n, nl, 16)
ZBNShiftRightTemplate(zbnsrt17, n, nl, 17)
ZBNShiftRightTemplate(zbnsrt18, n, nl, 18)
ZBNShiftRightTemplate(zbnsrt19, n, nl, 19)
ZBNShiftRightTemplate(zbnsrt20, n, nl, 20)
ZBNShiftRightTemplate(zbnsrt21, n, nl, 21)
ZBNShiftRightTemplate(zbnsrt22, n, nl, 22)
ZBNShiftRightTemplate(zbnsrt23, n, nl, 23)
ZBNShiftRightTemplate(zbnsrt24, n, nl, 24)
ZBNShiftRightTemplate(zbnsrt25, n, nl, 25)
ZBNShiftRightTemplate(zbnsrt26, n, nl, 26)
ZBNShiftRightTemplate(zbnsrt27, n, nl, 27)
ZBNShiftRightTemplate(zbnsrt28, n, nl, 28)
ZBNShiftRightTemplate(zbnsrt29, n, nl, 29)
ZBNShiftRightTemplate(zbnsrt30, n, nl, 30)
ZBNShiftRightTemplate(zbnsrt31, n, nl, 31)

static ZBNShiftRightType ZBNShiftRightTabular[32] = {
  zbnsrt0, zbnsrt1, zbnsrt2, zbnsrt3, zbnsrt4, zbnsrt5, zbnsrt6, zbnsrt7,
  zbnsrt8, zbnsrt9, zbnsrt10, zbnsrt11, zbnsrt12, zbnsrt13, zbnsrt14, zbnsrt15,
  zbnsrt16, zbnsrt17, zbnsrt18, zbnsrt19, zbnsrt20, zbnsrt21, zbnsrt22, zbnsrt23,
  zbnsrt24, zbnsrt25, zbnsrt26, zbnsrt27, zbnsrt28, zbnsrt29, zbnsrt30, zbnsrt31
};

/* \TEX{
   \title{Shifting right of a small amount}
   \input{An allocated |BigNum|, its length and a positive integer
   $0 \le |l| < |SIZE_BLOC|$.}
   \output{The digit shifted out}
   \side{$|(n,nl)| >>= |l|$}
   \note{If |l| equals zero, nothing is done and the returned value
   is 0.}
   } */
BigNumDigit ZBNShiftRight(n, nl, l)
     BigNum n;
     BigNumLength nl;
     BigNumLength l;
{
  return((ZBNShiftRightTabular[l])(n, nl));
}
# else
BigNumDigit ZBNShiftRight(n, nl, r)
     BigNum n;
     BigNumLength nl;
     BigNumLength r;
{
  BigNumLength l;
  BigNumDigit d, e;

  if ((!nl) || (!r)) {return(CST_BNDIGIT(0));}
  l = SIZE_BLOC-r;
  d = 0;
  nl --;
  do {
    e = n[nl] << l;
    n[nl] >>= r;
    n[nl] |= d;
    d = e;
    nl--;
  } while(nl >= 0);
  return(d);
}
# endif
#else
BigNumDigit ZBNShiftRight(n, nl, r)
     BigNum n;
     BigNumLength nl;
     BigNumLength r;
{
  return((r) ? (BigNumDigit)(mpn_rshift(n, n, nl, r)) : (BigNumDigit)(0));
}
#endif

/* \TEX{
   \title{Shifting left}
   \input{Two allocated |BigNum|s, a pointer on a length, and an
   integer |nnn|.}
   \side{$|r| = |n| << |nnn|$}
   \note{The size of |r| must be at least $|nl|+|divSizeBloc(nnn)|$,
   and such that the result of shift can be stored. Hence, an
   allocated size of $|nl|+|divSizeBloc(nnn)|+1$ is always
   sufficient. The |BigNum| |n| can be part of |r|. The length of |r|
   is stored in |*p_rl|. The pointer |p_rl| can be the address of the
   external value of |nl|.}
   } */
void ZBNAnyShiftLeft(r, p_rl, n, nl, nnn)
     BigNum r,n;
     BigNumLength *p_rl,nl;
     unsigned int nnn;
{
  BigNumDigit out;
  BigNumLength ds=0;

  if(nnn >= SIZE_BLOC) {    
    ds=divSizeBloc(nnn);
    ZBNAssign(r+ds,n,nl);
    ZBNSetToZero(r,ds);
    r += ds;
    nnn = modSizeBloc(nnn);
  }
  else {
    ZBNAssign(r,n,nl);
  }
  out=ZBNShiftLeft(r,nl,nnn);
  if(out) {
    ZBNDigit(r,nl)=out;
    ds++;
  }
  *p_rl = nl+ ds;
}

/* \TEX{
   \title{Shifting right.}
   \input{An allocated |BigNum| |n| of |nl| |BigNumDigits| and an
   integer |nnn|.}
   \side{$|n| >>= |nnn|$}
   \note{Low degree bits are lost}
   } */
void ZBNAnyShiftRight(n,nl,nnn)
     BigNum n;
     BigNumLength nl;
     unsigned int nnn;
{
  if(nnn >= SIZE_BLOC) {
    BigNumLength ds;
    
    ds=divSizeBloc(nnn);
    nl -= ds;
    ZBNAssign(n,n+ds,nl);
    ZBNSetToZero(n+nl,ds);
  }
  ZBNShiftRight(n,nl,modSizeBloc(nnn));
}
