/*
 * zen/zbn/c/test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

/* \TEX{   
   \Subsection{Comparisons}
   } */

/* \TEX{
   \title{Test if a big integer value is equal to zero}
   \input{An allocated |BigNum| and its length.}
   \output{The predicate |(n,nl) == 0|}
   } */
int ZBNIsZero (n, nl)
     BigNum n;
     BigNumLength nl;
{
  if (!nl) {return(1);}
  n += nl; nl *= CST_BNLENGTH(-1);
  do {
    if (n[nl]) {return(0);}
    nl++;
  } while(nl);
  return(1);
}

/* \TEX{
   \title{Test if a big integer value is equal to one}
   \input{An allocated |BigNum| and its length.}
   \output{The predicate |(n,nl) == 1|}
   } */
int ZBNIsOne (n, nl)
     BigNum n;
     BigNumLength nl;
{
  if (nl <= CST_BNLENGTH(0)) {return(1);}
  if (n[0] != CST_BNDIGIT(1)) {return(0);}
  n += nl; nl--;
  if (!nl) {return(1);}
  nl *= CST_BNLENGTH(-1);
  do {
    if (n[nl]) {return(0);}
    nl++;
  } while(nl);
  return(1);
}

/* \TEX{
   \title{Comparison of bignums}
   \input{Two allocated |BigNum|s and their lengths}
   \output{
   \begin{itemize}
   \item $1$ if $|m| > |n|$;
   \item zero if $|m| = |n|$;
   \item $-1$ if $|m| < |n|$.
   \end{itemize}}
   } */
int ZBNCompare(m, ml, n, nl)
     BigNum m;
     BigNumLength ml;
     BigNum n;
     BigNumLength nl;
{
  if (ml == nl) {
    if (!ml) {return(0);}
    n--; m--;
    do {
      if (m[ml] > n[ml]) {return(1);}
      if (m[ml] < n[ml]) {return(-1);}
      ml--;
    } while (ml);
  }
  else if (ml > nl) {
    m += nl-1;
    ml -= nl;
    do {
      if (m[ml]) {return(1);}
      ml--;
    } while (ml);
    m -= nl; n--;
    do {
      if (m[nl] > n[nl]) {return(1);}
      if (m[nl] < n[nl]) {return(-1);}
      nl--;
    } while (nl);
  }
  else {
    n += ml-1;
    nl -= ml;
    do {
      if (n[nl]) {return(-1);}
      nl--;
    } while (nl);
    n -= ml; m--;
    do {
      if (m[ml] > n[ml]) {return(1);}
      if (m[ml] < n[ml]) {return(-1);}
      ml--;
    } while (ml);
  }
  return(0);
}

#if 0
/* \TEX{
   \title{Test if two bignums are equal}
   \input{Two allocated |BigNum|s and their lengths}
   \output{The predicate |(n,nl) == (m,ml)|}
   } */
int ZBNAreEqual(m, ml, n, nl)
     BigNum n, m;
     BigNumLength nl, ml; {}
#endif

/* \TEX{
   \title{Comparison of significant bignums}
   \input{Two bignums}
   \output{
   \begin{itemize}
   \item $1$ if $|m| > |n|$;
   \item zero if $|m| = |n|$;
   \item $-1$ if $|m| < |n|$.
   \end{itemize}}
   \note{One must have $|m[ml-1]|\neq 0$ and $|n[nl-1]|\neq 0$.}
   } */
int ZBNCompareLazy(m, ml, n, nl)
     BigNum m, n;
     BigNumLength ml, nl;
{
  if (ml != nl) {return (ml > nl ? 1 : -1);}
  if (!ml) {return(0);}
  n--; m--;
  do {
    if (m[ml] > n[ml]) {return(1);}
    if (m[ml] < n[ml]) {return(-1);}
    ml--;
  } while (ml);
  return(0);
}
