/*
 * zen/zbn/c/weight.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

#if !LOW_MEMORY
# include "zbn_weight_data.h"
#endif

/* \TEX{
   \Subsection{Hamming weight of a |BigNum|}
   } */

/* \TEX{
   \title{Number of bits of a digit}
   \input{A |BigNumDigit|}
   \output{The number of non zero bits in |d|}
   \note{This function is only provided for compatibility. Using
   directly the macro |zbnwght| is probably slightly faster.}
   } */
BigNumLength ZBNWeightDigitFast(d)
     BigNumDigit d;
{
  BigNumLength w;

  zbnwght(w, d);
  return(w);
}

/* \______TEX{
   \title{Hamming weight of a |BigNumDigit|.}
   \input{A BigNum |n|.}
   \output{The number of 1 in the binary representation of |*n| or
   |ZENERR| if an error occored.}
   \side{The first time this function is called, precomputations are
   done. To disallocate them, call this function with |n|=NULL.}
   } */
BigNumLength ZBNWeightDigit(n)
     BigNum n;
{
  BigNumLength w;

  zbnwght(w, *n);
  return(w);
}

/* \TEX{
   \title{Weight of a BigNum}
   \input{An allocated |BigNum| |n| of size |nl|.}
   \output{The number of 1 in the binary representation of |n|.}
   } */
BigNumLength ZBNWeight(n, nl)
     BigNum n;
     BigNumLength nl;
{
  BigNumLength w, wp;

  w = 0;
  if (nl) {
    n += nl; nl *= CST_BNLENGTH(-1);
    do {
      zbnwght(wp, n[nl]);
      w += wp;
      nl++;
    }
    while(nl);
  }
  return(w);
}



