/*
 * zen/zbn/zmake.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "zmake.h"
#include "init.h"

char *SYS_FILES = "memalloc error rand runtime";
char *C_FILES = "random digits assign test binary shift add mul div gcd  modulo io log sqrt weight";
char *KARA_FILES = "karamul karasqu karacut";
char *H_FILES = "zbn zbnconfig";

char *KARALIM = "karalim";
char *KARACUT = "karacut";
char *KARADEF = "karadefs";

char *KARATEST= "karatest";

char *AddPrefix(prefix, files)
     char *prefix;
     char *files;
{
  char *file, *next, *ptr;

  ptr = concat("", "");
  for(next = ExpandedFileName(&file, files, PATHSEP);
      next;
      next = ExpandedFileName(&file, next, PATHSEP)) {
    append(&ptr, " "); append(&ptr, prefix); append(&ptr, file);    
    ZFree(file);
  }
  return(ptr);
}


int main(argc, argv)
     int argc;
     char *argv[];
{
  /*
   * Misc. variables
   *****************/
  char *GMP_DEFINES, *GMP_LIB;
  char *ZBN_PREFIX;
  char *SYS_PREFIX;
  char *KARA_LIM_FILES, *KARA_CUT_FILES, *KARA_ALL_FILES, *LIMINC, *CUTINC;
  char *ZEN_LIBRARY;

 /* 
   * Internal targets 
   ********************/
  int objtoar = 0;
  int clean_obj = 0; int clean_lib = 0;
  int clean_doc = 0; int clean_tex = 0; int clean_dvi = 0;
  int tidy_lib = 0; int tidy_obj = 0; int tidy_bin = 0;
  int tidy_doc = 0;
  
  /* 
   * External targets
   *******************/
  int all = Target("all", argc, argv);

  int cc = Target("cc", argc, argv);

  int doc = Target("doc", argc, argv);
  int tex = Target("tex", argc, argv);
  int dvi = Target("dvi", argc, argv);

  int archtidy = Target("archtidy", argc, argv);
  int archclean = Target("archclean", argc, argv);
  int clean = Target("clean", argc, argv);
  int tidy = Target("tidy", argc, argv);

  /* 
   * Initialization
   *****************/
  zen_init(CURDIR, TOPDIR, ZCFDIR);
  zmake_init(ZEN_DEFINES, ZEN_INCLUDES, "", MathLibrary, argc, argv);
  
  ZBN_PREFIX = concat(PREFIX, PZBN);
  SYS_PREFIX = concat(PREFIX, PSYS);

  KARA_LIM_FILES = concat(C_FILES, concat(" ", AddPrefix(KARALIM, concat(KARA_FILES, concat(" ", KARATEST)))));
  KARA_CUT_FILES = concat(C_FILES, concat(" ", AddPrefix(KARACUT, concat(KARA_FILES, concat(" ", KARATEST)))));
  KARA_ALL_FILES = concat(C_FILES, concat(" ", concat(KARA_FILES, concat(" ", KARATEST))));
  
  LIMINC = concat(ZBN_OBJ, concat("/", concat(PREFIX, KARALIM)));
  CUTINC = concat(ZBN_OBJ, concat("/", concat(PREFIX, KARACUT)));
  
  ZEN_LIBRARY = concat(ZEN_LIB,
		       concat(PATHSEP,
			      concat(PREFIX,
				     concat(LIBZEN, ASUF))));

#if USE_GMP
  GMP_DEFINES=" -DUSE_GMP";
  GMP_LIB=" -lgmp";
#else
  GMP_DEFINES="";
  GMP_LIB="";
#endif
  /* 
   * Targets
   ***********/

  /* default */
  TargetImply(defaults, cc);

  /* all */
  TargetImply(all, cc=doc);

  /* cc */
  TargetImply(cc, objtoar);
  SrcToObjects(cc,
	       ZBN_OBJ, ZBN_PREFIX,
	       ZBN_SRC, "", C_FILES,
	       ZEN_INC, GMP_DEFINES);

  /* cc, karalim */
  SrcToObjects(cc,
	       ZBN_OBJ, concat(ZBN_PREFIX, KARALIM),
	       ZBN_SRC, "", KARA_FILES,
	       ZEN_INC, " -DKARA_LIM_SET");
  SrcToObjects(cc,
	       ZBN_OBJ, concat(ZBN_PREFIX, KARALIM),
	       ZBN_SRC, "", KARATEST,
	       ZEN_INC, " -DKARA_LIM_SET");
  if (ObjectsToBinary2(cc,
		      ZBN_BIN, concat(PREFIX, KARALIM), KARATEST,
		      SYS_OBJ, SYS_PREFIX, SYS_FILES,
		      ZBN_OBJ, ZBN_PREFIX, KARA_LIM_FILES,
		      GMP_LIB, "")) {
    char *command;
    
    command = concat(ZBN_BIN, "/");
    append(&command, PREFIX);
    append(&command, KARALIM);
    append(&command, KARATEST);
    append(&command, ESUF);
    ExecuteCommand(command,
		   NULL, NULL, NULL, NULL,
		   LIMINC, "", KARADEF, HSUF);
    ZFree(command);
  }

  /* cc, karacut */
  SrcToObjects(cc,
	       ZBN_OBJ, concat(ZBN_PREFIX, KARACUT),
	       ZBN_SRC, "", KARA_FILES,
	       concat(LIMINC, concat(" ", ZEN_INC)),
	       concat(" -DKARA_CUT_SET", concat(" -I", LIMINC)));
  SrcToObjects(cc,
	       ZBN_OBJ, concat(ZBN_PREFIX, KARACUT),
	       ZBN_SRC, "", KARATEST,
	       ZEN_INC,
	       concat(" -DKARA_CUT_SET", concat(" -I", LIMINC)));
  if (ObjectsToBinary2(cc,
		      ZBN_BIN, concat(PREFIX, KARACUT), KARATEST,
		      SYS_OBJ, SYS_PREFIX, SYS_FILES,
		      ZBN_OBJ, ZBN_PREFIX, KARA_CUT_FILES,
		      GMP_LIB, "")) {
    char *command;
    
    command = concat(ZBN_BIN, "/");
    append(&command, PREFIX);
    append(&command, KARACUT);
    append(&command, KARATEST);
    append(&command, ESUF);
    ExecuteCommand(command,
		   NULL, NULL, NULL, NULL,
		   CUTINC, "", KARADEF, HSUF);
    ZFree(command);
  }
  
  /* cc, kara */
  SrcToObjects(cc,
	       ZBN_OBJ, ZBN_PREFIX,
	       ZBN_SRC, "", KARA_ALL_FILES,
	       concat(CUTINC, concat(" ", ZEN_INC)),
	       concat(" -DKARA_CUT_SET", concat(" -I", CUTINC)));

  /* ar */
  ObjectsToArList(objtoar,
		  ZBN_LIB, ZBN_PREFIX, "", OBJ_TOAR,
		  ZBN_OBJ, ZBN_PREFIX, concat(C_FILES, concat(" ", KARA_FILES)));

  /* doc */
  TargetImply(doc, dvi);
  TargetImply(dvi, tex);

  /* tex */
  SrcToTex(tex,
	   ZBN_TEX, PZBN,
	   ZBN_SRC, "", C_FILES, CSUF,
	   ZBN_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   ZBN_TEX, PZBN,
	   ZBN_SRC, "", KARA_FILES, CSUF,
	   ZBN_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   ZBN_TEX, "",
	   ZBN_INC, "", H_FILES, HSUF,
	   ZBN_BIN, "", IPT, PARSER);

  /* dvi */
  TexToDvi(dvi,
	   ZBN_DVI, PZBN,
	   ZBN_TEX, PZBN, C_FILES,
	   ZEN_INPUTS);
  TexToDvi(dvi,
	   ZBN_DVI, PZBN,
	   ZBN_TEX, PZBN, KARA_FILES,
	   ZEN_INPUTS);
  TexToDvi(dvi,
	   ZBN_DVI, "",
	   ZBN_TEX, "", H_FILES,
	   ZEN_INPUTS);

  /* Cleaning */
  TargetImply(tidy, tidy_obj=tidy_lib=tidy_bin=tidy_doc);
  TargetImply(archtidy, archclean=clean_doc);
  TargetImply(clean, clean_lib=tidy_obj=clean_tex);
  TargetImply(archclean, clean_obj=clean_lib=clean_tex);
  TargetImply(clean_doc, clean_dvi=clean_tex);
  TargetImply(tidy_obj, clean_obj=clean_lib);
  
  TidyingRules(clean_obj, LIMINC); TidyingRules(clean_obj, CUTINC);
  CleaningRules(clean_obj, ZBN_OBJ, ZBN_PREFIX, KARA_LIM_FILES, OSUF);
  CleaningRules(clean_obj, ZBN_OBJ, ZBN_PREFIX, KARA_CUT_FILES, OSUF);
  CleaningRules(clean_obj, ZBN_OBJ, ZBN_PREFIX, KARA_ALL_FILES, OSUF);
  CleaningRules(clean_obj, ZBN_BIN, concat(PREFIX, KARACUT), KARATEST, ESUF);
  CleaningRules(clean_obj, ZBN_BIN, concat(PREFIX, KARALIM), KARATEST, ESUF);

  CleaningRules(clean_lib, ZBN_LIB, "", ZBN_PREFIX, OBJ_TOAR);

  CleaningRules(clean_tex, ZBN_TEX, PZBN, C_FILES, ".tex");
  CleaningRules(clean_tex, ZBN_TEX, PZBN, KARA_FILES, ".tex");
  CleaningRules(clean_tex, ZBN_TEX, "", H_FILES, ".tex");

  CleaningRules(clean_dvi, ZBN_DVI, PZBN, C_FILES, ".dvi");
  CleaningRules(clean_dvi, ZBN_DVI, PZBN, KARA_FILES, ".dvi");
  CleaningRules(clean_dvi, ZBN_DVI, "", H_FILES, ".dvi");

  CleaningRules(clean_tex|clean_dvi, ZBN_DOC, "", "", "");

  TidyingRules(tidy_obj, ZBN_OBT);
  TidyingRules(tidy_lib, ZBN_LBT);
  TidyingRules(tidy_bin, ZBN_BNT);
  TidyingRules(tidy_doc, ZBN_DOC);

  /* 
   * Closing
   **********/
  zmake_close();

  exit(0);
}
