/*
 * zen/ze2/c/mat_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   
   \Subsection{Tests on a matrix}
   }
   */


#include <stdio.h>
#include "ze2.h"
/* \TEX[
   \title{Test of equality}
   \input{Two matrices and a |Ze2Ring|}
   \output{1 if the matrices are equal, 0 otherwise}
   }
   */

int _Ze2MatAreEqual(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c,dl,ml;

 
 if(Ze2MatAreSameType(A,B,R))
   return(Ze2MatSameTypeAreEqual(A,B,R));

 l=Ze2MatNbBn(A,R)-1;
 dl=divSizeBloc(l);
 ml=modSizeBloc(l);
 while(l>=0)
   {
     for(c=Ze2MatNbBloc(A,R)-1;c>=0;c--)
       if(Ze2MatLnCoeffFast(B,c,dl,ml,R)!=Ze2MatLnCoeff(A,l,c,R))
	 return(0);
     l--;
     ml--;
     if(ml<0)
       {
	 dl--;
	 ml=SIZE_BLOC-1;
       } 
   }
 return(1);
}

/* \TEX[
   \title{Test of equality to zero}
   \input{A matrix and a |Ze2Ring|}
   \output{1 if the matrix is null, 0 otherwise}
   }
   */

int _Ze2MatIsZero(A,R)
     ZENMat A;
     ZENRing R;
{
 return(Ze2MatIsZero(A,R));
}

/* \TEX[
   \title{Test of equality to identity}
   \input{A matrix and a |Ze2Ring|}
   \output{1 if the matrix is identity, 0 otherwise}
   \note{The matrix is checked to be square.} 
   }
   */

int _Ze2MatIsOne(A,R)
     ZENMat A;
     ZENRing R;
{int l,nb;
 BigNumDigit El,*ptr;

 if(Ze2MatNbBn(A,R) != Ze2MatNbDigit(A,R))
   return(0);


 l=0;
 El=1;
 ptr=Ze2MatFirst(A,R);
 nb=Ze2MatNbBloc(A,R)-1;
 for(;;)
   {
     if(*ptr != El)
       return(0);
     ptr++;
     
     l++;
     if(l<Ze2MatNbBn(A,R))
       {
	 El<<=1;
	 if(!El)
	   {
	     El=1;
	     if(!ZBNIsZero(ptr,nb+1))
	       return(0);
	     ptr += nb+1;
	   }
	 else
	   {
	     if(nb)
	       {
		 if(!ZBNIsZero(ptr,nb))
		   return(0);
		 ptr += nb;
	       }
	   }
     
       }
     else
       break;
   }
 return(1);
}



