/*
 * zen/ze2/c/permutation_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Permutation assignments}

   These functions are general ones. They are directly written with |Ze2Elt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "ze2.h"

/*
   \TEX{
   \title{Assignment.}
   \input{A |Ze2Mat| |A| and a permutation matrix |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   
int _Ze2Permutation2Matrix(A,B,R)
     ZENMat A,B;
     ZENRing R;
{
  int l;
  
  ZENTest(!(Ze2MatIsPermutation(B,R)),
	  ZEN_NOTPERM,ZEN_PERMASSIGN,ZEN_LIB_ZE2, ZERR);
  ZENTest(Ze2MatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_PERMASSIGN,ZEN_LIB_ZE2, ZERR);
  ZENTest((Ze2MatNbBn(A,R) != Ze2MatNbBn(B,R))
	  ||(Ze2MatNbDigit(A,R) != Ze2MatNbDigit(B,R)),
	  ZEN_MAT_DIM,ZEN_PERMASSIGN,ZEN_LIB_ZE2, ZERR);
  
  Ze2MatSetToZero(A,R);
  
  if(Ze2MatAreSameType(A,B,R))
    for(l=Ze2MatNbBn(A,R)-1;l>=0;l--)
      Ze2MatLnChangeCoeff(A,l,ZENPermutationBn(B,l,R),R);
  else
    for(l=Ze2MatNbBn(A,R)-1;l>=0;l--)
      Ze2MatLnChangeCoeff(A,ZENPermutationBn(B,l,R),l,R);
  return(0);
}







