/*
 * zen/ze2/c/poly_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial additions}
   
   One function is provided to do additions: |_Ze2PolyAdd| to add a
   |ZENPoly| to another.
   }
   */
#include <stdio.h>
#include "ze2.h"

/* \TEX[
   \title{Addition.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| += |PX|.}
   }
   */
void _Ze2PolyAdd(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int ds, ms, i, shift;

  if (Ze2PolyDeg(RX, Rg) >= Ze2PolyDeg(PX, Rg)) {
    ds = divSizeBloc(Ze2PolyDeg(PX, Rg));
    ms = modSizeBloc(Ze2PolyDeg(PX, Rg));
    shift = SIZE_BLOC-ms-1;
    for(i = 0; i < ds; i++) {
      Ze2PolyGetDigit(RX, i, Rg) ^= Ze2PolyGetDigit(PX, i, Rg);
    }
    Ze2PolyGetDigit(RX, i, Rg) ^=
      (Ze2PolyGetDigit(PX, i, Rg) << shift) >> shift;
  }
  else {
    ds = divSizeBloc(Ze2PolyDeg(RX, Rg));
    ms = modSizeBloc(Ze2PolyDeg(RX, Rg));
    for(i = 0; i < ds; i++) {
      Ze2PolyGetDigit(RX, i, Rg) ^= Ze2PolyGetDigit(PX, i, Rg);
    }
    shift = SIZE_BLOC-1-ms;
    Ze2PolyGetDigit(RX, ds, Rg) <<= shift;
    Ze2PolyGetDigit(RX, ds, Rg) >>= shift;
    Ze2PolyGetDigit(RX, ds, Rg) ^= Ze2PolyGetDigit(PX, ds, Rg);
    for(i = ds+1; i <= divSizeBloc(Ze2PolyDeg(PX, Rg)); i++) {
      Ze2PolyGetDigit(RX, i, Rg) = Ze2PolyGetDigit(PX, i, Rg);
    }
    Ze2PolyDeg(RX, Rg) = Ze2PolyDeg(PX, Rg);
  }
  Ze2PolyUpdateDegree(RX, Rg);
}

