/*
 * zen/ze2/c/poly_alloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   
   
   \Subsection{Polynomial creation}
   
   |_Ze2PolyCreate| allocates one |ZENPoly| and |_Ze2PolyFree| frees then it.
   
   
   }
   */
#include <stdio.h>
#include "ze2.h"

/* \TEX[
   \title{Creation.}
   \input{A finite ring |R| and a degree |deg|.}
   \output{An allocated |ZENPoly| set to 0 with |deg| coefficients allocated. }
   }
   */
ZENPoly _Ze2PolyCreate(deg, Rg)
     int deg;
     ZENRing Rg;
{
    ZENPoly PX;
    int ds;

    if (!(PX = ZENPolyAllocStruct(Rg))) {
      ZENSetError(ZEN_LIB_ZE2, ZEN_POLYCREATE, ZEN_ALLOC);
      return(ZNULL);
    }
    ds = 1+divSizeBloc(deg);
    if (!(Ze2PolyDigit(PX, Rg) = ZBNC(ds))) {
      ZENSetError(ZEN_LIB_ZE2, ZEN_POLYCREATE, ZEN_ALLOC);
      ZENFree(PX);
      return(ZNULL);
    }
    Ze2PolyDeg(PX, Rg) = 0;
    Ze2PolyLgt(PX, Rg) = ds*SIZE_BLOC-1;
    return(PX);
}

/* \TEX[
   \title{Free.}
   \input{a |ZENPoly| |PX| of a finite ring |Rg| already allocated with
   |_Ze2PolyCreate|.}
   \side{|PX| is freed.}
   }
 f  */
void _Ze2PolyFree(PX,Rg)
     ZENPoly PX;
     ZENRing Rg;
{
  ZBNF(Ze2PolyDigit(PX, Rg));
  ZENFree(PX);
}

