/*
 * zen/ze2/c/poly_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial multiplication}
   
   Two functions are provided to do multiplications: |_Ze2PolySquare| to square a
   |ZENPoly| and |_Ze2PolyMultiply| to compute the multiplication of two |ZENPoly|.
   }
   */

#include <stdio.h>
#include <ze2.h>

#define USE_MACROS 1

#include "ze2_polymult_data.h"
#define Ze2PreMultFast(a,b) PREZENMULT_FAST_[a][b]

#include "ze2_polydiv_data.h"
#define Ze2PreDivFast(a,b) PREZENDIV_FAST_[a][b]

/*********************  Ze2PolyCutInChars *************/
#define Ze2PolyCutInChars(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, a) \
  Ze2PolyCutInCharsMacro(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, a) 

#if 1
# if SIZE_BLOC == 64
				/* Word of 64 bits */
#define Ze2PolyCutInCharsMacro(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, a)	\
if (1) {									\
      sa0 = a;									\
  } else
# else
				/* Word of 32 bits */
#define Ze2PolyCutInCharsMacro(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, a)	\
if (1) {									\
      sa0 = a;									\
      sa3 = sa0 >> LN2PREZENMULT;						\
      sa0 &= ((1<<LN2PREZENMULT)-1);						\
      sa1 = sa3 & ((1<<LN2PREZENMULT)-1);					\
      sa3 >>= LN2PREZENMULT;							\
      sa2 = sa3 & ((1<<LN2PREZENMULT)-1);					\
      sa3 >>= LN2PREZENMULT;							\
} else
# endif
#else
# if SIZE_BLOC == 64
				/* Word of 64 bits */
#define Ze2PolyCutInCharsMacro(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, a)	\
if (1) {									\
      sa0 = a;									\
      sa7 = sa0 >> LN2PREZENMULT;						\
      sa0 &= ((1<<LN2PREZENMULT)-1);						\
      sa1 = sa7 & ((1<<LN2PREZENMULT)-1);					\
      sa7 >>= LN2PREZENMULT;							\
      sa2 = sa7 & ((1<<LN2PREZENMULT)-1);					\
      sa7 >>= LN2PREZENMULT;							\
      sa3 = sa7 & ((1<<LN2PREZENMULT)-1);					\
      sa7 >>= LN2PREZENMULT;							\
      sa4 = sa7 & ((1<<LN2PREZENMULT)-1);					\
      sa7 >>= LN2PREZENMULT;							\
      sa5 = sa7 & ((1<<LN2PREZENMULT)-1);					\
      sa7 >>= LN2PREZENMULT;							\
      sa6 = sa7 & ((1<<LN2PREZENMULT)-1);					\
      sa7 >>= LN2PREZENMULT;							\
  } else
# else
				/* Word of 32 bits */
#define Ze2PolyCutInCharsMacro(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, a)	\
if (1) {									\
      sa0 = a;									\
      sa3 = sa0 >> LN2PREZENMULT;						\
      sa0 &= ((1<<LN2PREZENMULT)-1);						\
      sa1 = sa3 & ((1<<LN2PREZENMULT)-1);					\
      sa3 >>= LN2PREZENMULT;							\
      sa2 = sa3 & ((1<<LN2PREZENMULT)-1);					\
      sa3 >>= LN2PREZENMULT;							\
  }										\
} else
# endif
#endif

/*********************  Ze2PolyMultiplyChars *************/
#if !USE_MACROS
#define Ze2PolyMultiplyChars(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7) \
  Ze2PolyMultiplyCharsProc(&(cc), &(dd), sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7) 

void Ze2PolyMultiplyCharsProc(p_cc, p_dd,
			      sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7,
			      sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7)
     BigNumDigit *p_cc, *p_dd;
     BigNumDigit sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7;
     BigNumDigit sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7;
{
# if SIZE_BLOC == 64
#  if 1
    /* Word of 64 bits */
    register BigNumDigit h, l, t, b, a, cc, dd;

    cc = *p_cc;
    dd = *p_dd;

    b = sa0; a = sb0;
    h = 0; l = a;
    
    t = b & (BigNumDigit)1;	/* bit 1 */
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 2 */
    h = a >> (BN_DIGIT_SIZE-1); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 3 */
    h = a >> (BN_DIGIT_SIZE-2); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 4 */
    h = a >> (BN_DIGIT_SIZE-3); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 5 */
    h = a >> (BN_DIGIT_SIZE-4); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 6 */
    h = a >> (BN_DIGIT_SIZE-5); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 7 */
    h = a >> (BN_DIGIT_SIZE-6); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 8 */
    h = a >> (BN_DIGIT_SIZE-7); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 9 */
    h = a >> (BN_DIGIT_SIZE-8); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 10 */
    h = a >> (BN_DIGIT_SIZE-9); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 11 */
    h = a >> (BN_DIGIT_SIZE-10); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 12 */
    h = a >> (BN_DIGIT_SIZE-11); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 13 */
    h = a >> (BN_DIGIT_SIZE-12); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 14 */
    h = a >> (BN_DIGIT_SIZE-13); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 15 */
    h = a >> (BN_DIGIT_SIZE-14); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 16 */
    h = a >> (BN_DIGIT_SIZE-15); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 17 */
    h = a >> (BN_DIGIT_SIZE-16); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 18 */
    h = a >> (BN_DIGIT_SIZE-17); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 19 */
    h = a >> (BN_DIGIT_SIZE-18); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 20 */
    h = a >> (BN_DIGIT_SIZE-19); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 21 */
    h = a >> (BN_DIGIT_SIZE-20); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 22 */
    h = a >> (BN_DIGIT_SIZE-21); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 23 */
    h = a >> (BN_DIGIT_SIZE-22); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 24 */
    h = a >> (BN_DIGIT_SIZE-23); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 25 */
    h = a >> (BN_DIGIT_SIZE-24); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 26 */
    h = a >> (BN_DIGIT_SIZE-25); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 27 */
    h = a >> (BN_DIGIT_SIZE-26); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 28 */
    h = a >> (BN_DIGIT_SIZE-27); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 29 */
    h = a >> (BN_DIGIT_SIZE-28); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 30 */
    h = a >> (BN_DIGIT_SIZE-29); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 31 */
    h = a >> (BN_DIGIT_SIZE-30); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 32 */
    h = a >> (BN_DIGIT_SIZE-31); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 33 */
    h = a >> (BN_DIGIT_SIZE-32); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 34 */
    h = a >> (BN_DIGIT_SIZE-33); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 35 */
    h = a >> (BN_DIGIT_SIZE-34); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 36 */
    h = a >> (BN_DIGIT_SIZE-35); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 37 */
    h = a >> (BN_DIGIT_SIZE-36); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 38 */
    h = a >> (BN_DIGIT_SIZE-37); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 39 */
    h = a >> (BN_DIGIT_SIZE-38); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 40 */
    h = a >> (BN_DIGIT_SIZE-39); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 41 */
    h = a >> (BN_DIGIT_SIZE-40); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 42 */
    h = a >> (BN_DIGIT_SIZE-41); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 43 */
    h = a >> (BN_DIGIT_SIZE-42); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 44 */
    h = a >> (BN_DIGIT_SIZE-43); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 45 */
    h = a >> (BN_DIGIT_SIZE-44); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 46 */
    h = a >> (BN_DIGIT_SIZE-45); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 47 */
    h = a >> (BN_DIGIT_SIZE-46); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 48 */
    h = a >> (BN_DIGIT_SIZE-47); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 49 */
    h = a >> (BN_DIGIT_SIZE-48); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 50 */
    h = a >> (BN_DIGIT_SIZE-49); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 51 */
    h = a >> (BN_DIGIT_SIZE-50); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 52 */
    h = a >> (BN_DIGIT_SIZE-51); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 53 */
    h = a >> (BN_DIGIT_SIZE-52); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 54 */
    h = a >> (BN_DIGIT_SIZE-53); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 55 */
    h = a >> (BN_DIGIT_SIZE-54); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 56 */
    h = a >> (BN_DIGIT_SIZE-55); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 57 */
    h = a >> (BN_DIGIT_SIZE-56); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 58 */
    h = a >> (BN_DIGIT_SIZE-57); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 59 */
    h = a >> (BN_DIGIT_SIZE-58); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 60 */
    h = a >> (BN_DIGIT_SIZE-59); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 61 */
    h = a >> (BN_DIGIT_SIZE-60); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 62 */
    h = a >> (BN_DIGIT_SIZE-61); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 63 */
    h = a >> (BN_DIGIT_SIZE-62); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}
    b >>= 1; t = b & 1;		/* bit 64 */
    h = a >> (BN_DIGIT_SIZE-63); l <<= 1;
    if(t) {dd ^= h; cc ^= l;}			\

    *p_cc = cc;
    *p_dd = dd;
#  else
    /* Word of 64 bits */                                                    
    register BigNumDigit t, cc, dd;

    cc = *p_cc;
    dd = *p_dd;

    cc ^= (Ze2PreMultFast(sa0,sb0));
    cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb1) << LN2PREZENMULT;
    cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb2) << (LN2PREZENMULT*2);
    cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb3) << (LN2PREZENMULT*3);
    cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb4) << (LN2PREZENMULT*4);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb5) << (LN2PREZENMULT*5);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb6) << (LN2PREZENMULT*6);
      t = (BigNumDigit)Ze2PreMultFast(sa0,sb7);
      cc ^= t << (LN2PREZENMULT*7);
      dd ^= (t >> LN2PREZENMULT);
									     
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb0) << LN2PREZENMULT;
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb1) << (LN2PREZENMULT*2);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb2) << (LN2PREZENMULT*3);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb3) << (LN2PREZENMULT*4);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb4) << (LN2PREZENMULT*5);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb5) << (LN2PREZENMULT*6);
      t = (BigNumDigit)Ze2PreMultFast(sa1,sb6);
      cc ^= t << (LN2PREZENMULT*7);
      dd ^= t >> LN2PREZENMULT;
      dd ^= Ze2PreMultFast(sa1,sb7);
      									     
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb0) << (LN2PREZENMULT*2);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb1) << (LN2PREZENMULT*3);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb2) << (LN2PREZENMULT*4);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb3) << (LN2PREZENMULT*5);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb4) << (LN2PREZENMULT*6);
      t = (BigNumDigit)Ze2PreMultFast(sa2,sb5);
      cc ^= t << (LN2PREZENMULT*7);
      dd ^= t >> LN2PREZENMULT;
      dd ^= Ze2PreMultFast(sa2,sb6);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa2,sb7) << LN2PREZENMULT;
      									     
      cc ^= (BigNumDigit)Ze2PreMultFast(sa3,sb0) << (LN2PREZENMULT*3);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa3,sb1) << (LN2PREZENMULT*4);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa3,sb2) << (LN2PREZENMULT*5);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa3,sb3) << (LN2PREZENMULT*6);
      t = (BigNumDigit)Ze2PreMultFast(sa3,sb4);
      cc ^= t << (LN2PREZENMULT*7);
      dd ^= t >> LN2PREZENMULT;
      dd ^= Ze2PreMultFast(sa3,sb5);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa3,sb6) << LN2PREZENMULT;
      dd ^= (BigNumDigit)Ze2PreMultFast(sa3,sb7) << (LN2PREZENMULT*2);
									     
      cc ^= (BigNumDigit)Ze2PreMultFast(sa4,sb0) << (LN2PREZENMULT*4);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa4,sb1) << (LN2PREZENMULT*5);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa4,sb2) << (LN2PREZENMULT*6);
      t = (BigNumDigit)Ze2PreMultFast(sa4,sb3);
      cc ^= t << (LN2PREZENMULT*7);
      dd ^= t >> LN2PREZENMULT;
      dd ^= Ze2PreMultFast(sa4,sb4);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa4,sb5) << LN2PREZENMULT;
      dd ^= (BigNumDigit)Ze2PreMultFast(sa4,sb6) << (LN2PREZENMULT*2);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa4,sb7) << (LN2PREZENMULT*3);
    									     
      cc ^= (BigNumDigit)Ze2PreMultFast(sa5,sb0) << (LN2PREZENMULT*5);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa5,sb1) << (LN2PREZENMULT*6);
      t = (BigNumDigit)Ze2PreMultFast(sa5,sb2);
      cc ^= t << (LN2PREZENMULT*7);
      dd ^= t >> LN2PREZENMULT;
      dd ^= Ze2PreMultFast(sa5,sb3);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa5,sb4) << LN2PREZENMULT;
      dd ^= (BigNumDigit)Ze2PreMultFast(sa5,sb5) << (LN2PREZENMULT*2);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa5,sb6) << (LN2PREZENMULT*3);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa5,sb7) << (LN2PREZENMULT*4);
									     
      cc ^= (BigNumDigit)Ze2PreMultFast(sa6,sb0) << (LN2PREZENMULT*6);
      t = (BigNumDigit)Ze2PreMultFast(sa6,sb1);
      cc ^= t << (LN2PREZENMULT*7);
      dd ^= t >> LN2PREZENMULT;
      dd ^= Ze2PreMultFast(sa6,sb2);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa6,sb3) << LN2PREZENMULT;
      dd ^= (BigNumDigit)Ze2PreMultFast(sa6,sb4) << (LN2PREZENMULT*2);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa6,sb5) << (LN2PREZENMULT*3);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa6,sb6) << (LN2PREZENMULT*4);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa6,sb7) << (LN2PREZENMULT*5);
									     
      t = (BigNumDigit)Ze2PreMultFast(sa7,sb0);
      cc ^= (t << (LN2PREZENMULT*7));
      dd ^= t >> LN2PREZENMULT;
      dd ^= Ze2PreMultFast(sa7,sb1);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa7,sb2) << LN2PREZENMULT;
      dd ^= (BigNumDigit)Ze2PreMultFast(sa7,sb3) << (LN2PREZENMULT*2);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa7,sb4) << (LN2PREZENMULT*3);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa7,sb5) << (LN2PREZENMULT*4);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa7,sb6) << (LN2PREZENMULT*5);
      dd ^= ((BigNumDigit)Ze2PreMultFast(sa7,sb7) << (LN2PREZENMULT*6));


      *p_cc = cc;
      *p_dd = dd;
#  endif
# else
      BigNumDigit cc, dd;
									     
      /* Word of 32 bits */    						     
      cc = (*p_cc)^(Ze2PreMultFast(sa0,sb0));
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb1) << LN2PREZENMULT;
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb2) << (LN2PREZENMULT*2);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb3) << (LN2PREZENMULT*3);
      dd = (*p_dd)^(Ze2PreMultFast(sa0,sb3) >> LN2PREZENMULT);
   									     
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb0) << LN2PREZENMULT;
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb1) << (LN2PREZENMULT*2);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb2) << (LN2PREZENMULT*3);
      dd ^= Ze2PreMultFast(sa1,sb2) >> LN2PREZENMULT;
      dd ^= Ze2PreMultFast(sa1,sb3);
   									     
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb0) << (LN2PREZENMULT*2);
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb1) << (LN2PREZENMULT*3);
      dd ^= Ze2PreMultFast(sa2,sb1) >> LN2PREZENMULT;
      dd ^= Ze2PreMultFast(sa2,sb2);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa2,sb3) << LN2PREZENMULT;
   									     
      *p_cc = cc^((BigNumDigit)Ze2PreMultFast(sa3,sb0) << (LN2PREZENMULT*3));
      dd ^= Ze2PreMultFast(sa3,sb0) >> LN2PREZENMULT;
      dd ^= Ze2PreMultFast(sa3,sb1);
      dd ^= (BigNumDigit)Ze2PreMultFast(sa3,sb2) << LN2PREZENMULT;
      *p_dd = dd^((BigNumDigit)Ze2PreMultFast(sa3,sb3) << (LN2PREZENMULT*2));
# endif  
}
#else
# if 1
     /* No need of tricky things for 64 bit computers !! */
#define Ze2PolyMultiplyChars(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7) \
  Ze2PolyMultiplyCharsMacro(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7) 

#  if SIZE_BLOC == 64
				/* Word of 64 bits */
#define Ze2PolyMultiplyCharsMacro(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7)			     \
if (1) {					\
  register BigNumDigit h, l, t, b, a;		\
						\
  b = sa0; a = sb0;				\
  h = 0; l = a;					\
						\
  t = b & (BigNumDigit)1;	/* bit 1 */	\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 2 */	\
  h = a >> (BN_DIGIT_SIZE-1); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 3 */	\
  h = a >> (BN_DIGIT_SIZE-2); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 4 */	\
  h = a >> (BN_DIGIT_SIZE-3); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 5 */	\
  h = a >> (BN_DIGIT_SIZE-4); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 6 */	\
  h = a >> (BN_DIGIT_SIZE-5); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 7 */	\
  h = a >> (BN_DIGIT_SIZE-6); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 8 */	\
  h = a >> (BN_DIGIT_SIZE-7); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 9 */	\
  h = a >> (BN_DIGIT_SIZE-8); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 10 */	\
  h = a >> (BN_DIGIT_SIZE-9); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 11 */	\
  h = a >> (BN_DIGIT_SIZE-10); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 12 */	\
  h = a >> (BN_DIGIT_SIZE-11); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 13 */	\
  h = a >> (BN_DIGIT_SIZE-12); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 14 */	\
  h = a >> (BN_DIGIT_SIZE-13); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 15 */	\
  h = a >> (BN_DIGIT_SIZE-14); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 16 */	\
  h = a >> (BN_DIGIT_SIZE-15); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 17 */	\
  h = a >> (BN_DIGIT_SIZE-16); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 18 */	\
  h = a >> (BN_DIGIT_SIZE-17); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 19 */	\
  h = a >> (BN_DIGIT_SIZE-18); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 20 */	\
  h = a >> (BN_DIGIT_SIZE-19); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 21 */	\
  h = a >> (BN_DIGIT_SIZE-20); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 22 */	\
  h = a >> (BN_DIGIT_SIZE-21); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 23 */	\
  h = a >> (BN_DIGIT_SIZE-22); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 24 */	\
  h = a >> (BN_DIGIT_SIZE-23); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 25 */	\
  h = a >> (BN_DIGIT_SIZE-24); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 26 */	\
  h = a >> (BN_DIGIT_SIZE-25); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 27 */	\
  h = a >> (BN_DIGIT_SIZE-26); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 28 */	\
  h = a >> (BN_DIGIT_SIZE-27); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 29 */	\
  h = a >> (BN_DIGIT_SIZE-28); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 30 */	\
  h = a >> (BN_DIGIT_SIZE-29); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 31 */	\
  h = a >> (BN_DIGIT_SIZE-30); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 32 */	\
  h = a >> (BN_DIGIT_SIZE-31); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 33 */	\
  h = a >> (BN_DIGIT_SIZE-32); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 34 */	\
  h = a >> (BN_DIGIT_SIZE-33); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 35 */	\
  h = a >> (BN_DIGIT_SIZE-34); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 36 */	\
  h = a >> (BN_DIGIT_SIZE-35); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 37 */	\
  h = a >> (BN_DIGIT_SIZE-36); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 38 */	\
  h = a >> (BN_DIGIT_SIZE-37); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 39 */	\
  h = a >> (BN_DIGIT_SIZE-38); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 40 */	\
  h = a >> (BN_DIGIT_SIZE-39); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 41 */	\
  h = a >> (BN_DIGIT_SIZE-40); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 42 */	\
  h = a >> (BN_DIGIT_SIZE-41); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 43 */	\
  h = a >> (BN_DIGIT_SIZE-42); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 44 */	\
  h = a >> (BN_DIGIT_SIZE-43); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 45 */	\
  h = a >> (BN_DIGIT_SIZE-44); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 46 */	\
  h = a >> (BN_DIGIT_SIZE-45); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 47 */	\
  h = a >> (BN_DIGIT_SIZE-46); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 48 */	\
  h = a >> (BN_DIGIT_SIZE-47); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 49 */	\
  h = a >> (BN_DIGIT_SIZE-48); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 50 */	\
  h = a >> (BN_DIGIT_SIZE-49); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 51 */	\
  h = a >> (BN_DIGIT_SIZE-50); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 52 */	\
  h = a >> (BN_DIGIT_SIZE-51); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 53 */	\
  h = a >> (BN_DIGIT_SIZE-52); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 54 */	\
  h = a >> (BN_DIGIT_SIZE-53); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 55 */	\
  h = a >> (BN_DIGIT_SIZE-54); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 56 */	\
  h = a >> (BN_DIGIT_SIZE-55); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 57 */	\
  h = a >> (BN_DIGIT_SIZE-56); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 58 */	\
  h = a >> (BN_DIGIT_SIZE-57); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 59 */	\
  h = a >> (BN_DIGIT_SIZE-58); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 60 */	\
  h = a >> (BN_DIGIT_SIZE-59); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 61 */	\
  h = a >> (BN_DIGIT_SIZE-60); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 62 */	\
  h = a >> (BN_DIGIT_SIZE-61); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 63 */	\
  h = a >> (BN_DIGIT_SIZE-62); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
  b >>= 1; t = b & 1;		/* bit 64 */	\
  h = a >> (BN_DIGIT_SIZE-63); l <<= 1;		\
  if(t) {dd ^= h; cc ^= l;}			\
} else
#  else
				/* Word of 32 bits */  
#define Ze2PolyMultiplyCharsMacro(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7) \
if (1) {									\
      cc ^= (Ze2PreMultFast(sa0,sb0));						\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb1) << LN2PREZENMULT;		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb2) << (LN2PREZENMULT*2);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb3) << (LN2PREZENMULT*3);		\
      dd ^= (Ze2PreMultFast(sa0,sb3) >> LN2PREZENMULT);				\
										\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb0) << LN2PREZENMULT;		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb1) << (LN2PREZENMULT*2);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb2) << (LN2PREZENMULT*3);		\
      dd ^= Ze2PreMultFast(sa1,sb2) >> LN2PREZENMULT;				\
      dd ^= Ze2PreMultFast(sa1,sb3);						\
										\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb0) << (LN2PREZENMULT*2);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb1) << (LN2PREZENMULT*3);		\
      dd ^= Ze2PreMultFast(sa2,sb1) >> LN2PREZENMULT;				\
      dd ^= Ze2PreMultFast(sa2,sb2);						\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa2,sb3) << LN2PREZENMULT;		\
										\
      cc ^= ((BigNumDigit)Ze2PreMultFast(sa3,sb0) << (LN2PREZENMULT*3));	\
      dd ^= Ze2PreMultFast(sa3,sb0) >> LN2PREZENMULT;				\
      dd ^= Ze2PreMultFast(sa3,sb1);						\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa3,sb2) << LN2PREZENMULT;		\
      dd ^= ((BigNumDigit)Ze2PreMultFast(sa3,sb3) << (LN2PREZENMULT*2));	\
} else
#  endif
# else
#define Ze2PolyMultiplyChars(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7) \
  Ze2PolyMultiplyCharsMacro(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7) 

#  if SIZE_BLOC == 64
				/* Word of 64 bits */
#define Ze2PolyMultiplyCharsMacro(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7) \
if (1) {									\
										\
      register BigNumDigit t;							\
										\
      cc ^= (Ze2PreMultFast(sa0,sb0));						\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb1) << LN2PREZENMULT;		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb2) << (LN2PREZENMULT*2);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb3) << (LN2PREZENMULT*3);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb4) << (LN2PREZENMULT*4);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb5) << (LN2PREZENMULT*5);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb6) << (LN2PREZENMULT*6);		\
      t = (BigNumDigit)Ze2PreMultFast(sa0,sb7);					\
      cc ^= t << (LN2PREZENMULT*7);						\
      dd ^= (t >> LN2PREZENMULT);						\
										\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb0) << LN2PREZENMULT;		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb1) << (LN2PREZENMULT*2);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb2) << (LN2PREZENMULT*3);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb3) << (LN2PREZENMULT*4);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb4) << (LN2PREZENMULT*5);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb5) << (LN2PREZENMULT*6);		\
      t = (BigNumDigit)Ze2PreMultFast(sa1,sb6);					\
      cc ^= t << (LN2PREZENMULT*7);						\
      dd ^= t >> LN2PREZENMULT;							\
      dd ^= Ze2PreMultFast(sa1,sb7);						\
										\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb0) << (LN2PREZENMULT*2);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb1) << (LN2PREZENMULT*3);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb2) << (LN2PREZENMULT*4);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb3) << (LN2PREZENMULT*5);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb4) << (LN2PREZENMULT*6);		\
      t = (BigNumDigit)Ze2PreMultFast(sa2,sb5);					\
      cc ^= t << (LN2PREZENMULT*7);						\
      dd ^= t >> LN2PREZENMULT;							\
      dd ^= Ze2PreMultFast(sa2,sb6);						\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa2,sb7) << LN2PREZENMULT;		\
										\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa3,sb0) << (LN2PREZENMULT*3);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa3,sb1) << (LN2PREZENMULT*4);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa3,sb2) << (LN2PREZENMULT*5);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa3,sb3) << (LN2PREZENMULT*6);		\
      t = (BigNumDigit)Ze2PreMultFast(sa3,sb4);					\
      cc ^= t << (LN2PREZENMULT*7);						\
      dd ^= t >> LN2PREZENMULT;							\
      dd ^= Ze2PreMultFast(sa3,sb5);						\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa3,sb6) << LN2PREZENMULT;		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa3,sb7) << (LN2PREZENMULT*2);		\
										\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa4,sb0) << (LN2PREZENMULT*4);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa4,sb1) << (LN2PREZENMULT*5);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa4,sb2) << (LN2PREZENMULT*6);		\
      t = (BigNumDigit)Ze2PreMultFast(sa4,sb3);					\
      cc ^= t << (LN2PREZENMULT*7);						\
      dd ^= t >> LN2PREZENMULT;							\
      dd ^= Ze2PreMultFast(sa4,sb4);						\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa4,sb5) << LN2PREZENMULT;		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa4,sb6) << (LN2PREZENMULT*2);		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa4,sb7) << (LN2PREZENMULT*3);		\
										\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa5,sb0) << (LN2PREZENMULT*5);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa5,sb1) << (LN2PREZENMULT*6);		\
      t = (BigNumDigit)Ze2PreMultFast(sa5,sb2);					\
      cc ^= t << (LN2PREZENMULT*7);						\
      dd ^= t >> LN2PREZENMULT;							\
      dd ^= Ze2PreMultFast(sa5,sb3);						\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa5,sb4) << LN2PREZENMULT;		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa5,sb5) << (LN2PREZENMULT*2);		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa5,sb6) << (LN2PREZENMULT*3);		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa5,sb7) << (LN2PREZENMULT*4);		\
										\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa6,sb0) << (LN2PREZENMULT*6);		\
      t = (BigNumDigit)Ze2PreMultFast(sa6,sb1);					\
      cc ^= t << (LN2PREZENMULT*7);						\
      dd ^= t >> LN2PREZENMULT;							\
      dd ^= Ze2PreMultFast(sa6,sb2);						\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa6,sb3) << LN2PREZENMULT;		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa6,sb4) << (LN2PREZENMULT*2);		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa6,sb5) << (LN2PREZENMULT*3);		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa6,sb6) << (LN2PREZENMULT*4);		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa6,sb7) << (LN2PREZENMULT*5);		\
										\
      t = (BigNumDigit)Ze2PreMultFast(sa7,sb0);					\
      cc ^= (t << (LN2PREZENMULT*7));						\
      dd ^= t >> LN2PREZENMULT;							\
      dd ^= Ze2PreMultFast(sa7,sb1);						\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa7,sb2) << LN2PREZENMULT;		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa7,sb3) << (LN2PREZENMULT*2);		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa7,sb4) << (LN2PREZENMULT*3);		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa7,sb5) << (LN2PREZENMULT*4);		\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa7,sb6) << (LN2PREZENMULT*5);		\
      dd ^= ((BigNumDigit)Ze2PreMultFast(sa7,sb7) << (LN2PREZENMULT*6));	\
} else
#  else
				/* Word of 32 bits */
#define Ze2PolyMultiplyCharsMacro(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7) \
if (1) {									\
      cc ^= (Ze2PreMultFast(sa0,sb0));						\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb1) << LN2PREZENMULT;		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb2) << (LN2PREZENMULT*2);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa0,sb3) << (LN2PREZENMULT*3);		\
      dd ^= (Ze2PreMultFast(sa0,sb3) >> LN2PREZENMULT);				\
										\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb0) << LN2PREZENMULT;		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb1) << (LN2PREZENMULT*2);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa1,sb2) << (LN2PREZENMULT*3);		\
      dd ^= Ze2PreMultFast(sa1,sb2) >> LN2PREZENMULT;				\
      dd ^= Ze2PreMultFast(sa1,sb3);						\
										\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb0) << (LN2PREZENMULT*2);		\
      cc ^= (BigNumDigit)Ze2PreMultFast(sa2,sb1) << (LN2PREZENMULT*3);		\
      dd ^= Ze2PreMultFast(sa2,sb1) >> LN2PREZENMULT;				\
      dd ^= Ze2PreMultFast(sa2,sb2);						\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa2,sb3) << LN2PREZENMULT;		\
										\
      cc ^= ((BigNumDigit)Ze2PreMultFast(sa3,sb0) << (LN2PREZENMULT*3));	\
      dd ^= Ze2PreMultFast(sa3,sb0) >> LN2PREZENMULT;				\
      dd ^= Ze2PreMultFast(sa3,sb1);						\
      dd ^= (BigNumDigit)Ze2PreMultFast(sa3,sb2) << LN2PREZENMULT;		\
      dd ^= ((BigNumDigit)Ze2PreMultFast(sa3,sb3) << (LN2PREZENMULT*2));	\
} else
#  endif
# endif
#endif

/*************************** Ze2PolySquareDigit *********************/
#if !USE_MACROS
#  define Ze2PolySquareDigit(c, d, a, Rg) \
          _Ze2PolySquareDigitProc(&(c), &(d), a, Rg)
void _Ze2PolySquareDigitProc(c, d, a, Rg) 
     BigNumDigit *c, *d, a;
     ZENRing Rg;
{
# if SIZE_BLOC == 64

    /* Word of 64 bits */
    register BigNumDigit ap, app, cc, dd;
    
    ap = a & ((1<<LN2PREZENMULT)-1);    
    cc = Ze2PreMultFast(ap,ap);

    app = a >> LN2PREZENMULT;
    ap = app & ((1<<LN2PREZENMULT)-1);   
    cc ^= (BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*2);

    app >>= LN2PREZENMULT;
    ap = app & ((1<<LN2PREZENMULT)-1);   
    cc ^= (BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*4);

    app >>= LN2PREZENMULT;
    ap = app & ((1<<LN2PREZENMULT)-1);   
    *c = cc ^ ((BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*6));

    app >>= LN2PREZENMULT;
    ap = app & ((1<<LN2PREZENMULT)-1);   
    dd = Ze2PreMultFast(ap,ap);

    app >>= LN2PREZENMULT;
    ap = app & ((1<<LN2PREZENMULT)-1);   
    dd ^= (BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*2);

    app >>= LN2PREZENMULT;
    ap = app & ((1<<LN2PREZENMULT)-1);   
    dd ^= (BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*4);

    app >>= LN2PREZENMULT;
    *d = dd ^ ((BigNumDigit)Ze2PreMultFast(app,app) << (LN2PREZENMULT*6));   
# else
    /* Word of 32 bits */
    register BigNumDigit ap, app, cc, dd;

    ap = a & ((1<<LN2PREZENMULT)-1);    
    cc = Ze2PreMultFast(ap,ap);

    app = a >> LN2PREZENMULT;
    ap = app & ((1<<LN2PREZENMULT)-1);
    *c = cc ^ ((BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*2));

    app >>= LN2PREZENMULT;
    ap = app & ((1<<LN2PREZENMULT)-1);
    dd = Ze2PreMultFast(ap,ap);

    app >>= LN2PREZENMULT;
    *d = dd ^ ((BigNumDigit)Ze2PreMultFast(app,app) << (LN2PREZENMULT*2));
  } 
# endif
}
#else
#  define Ze2PolySquareDigit(c, d, a, Rg) \
          Ze2PolySquareDigitMacro(c, d, a, Rg)
# if SIZE_BLOC == 64
				/* Word of 64 bits */
#define Ze2PolySquareDigitMacro(c, d, a, Rg)					\
if(1) {										\
    register BigNumDigit cc, dd, ap, app;					\
										\
    ap = a & ((1<<LN2PREZENMULT)-1);						\
    cc = Ze2PreMultFast(ap,ap);							\
										\
    app = a >> LN2PREZENMULT;							\
    ap = app & ((1<<LN2PREZENMULT)-1);						\
    cc ^= (BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*2);		\
										\
    app >>= LN2PREZENMULT;							\
    ap = app & ((1<<LN2PREZENMULT)-1);						\
    cc ^= (BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*4);		\
										\
    app >>= LN2PREZENMULT;							\
    ap = app & ((1<<LN2PREZENMULT)-1);						\
    c = cc ^ ((BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*6));		\
										\
    app >>= LN2PREZENMULT;							\
    ap = app & ((1<<LN2PREZENMULT)-1);						\
    dd = Ze2PreMultFast(ap,ap);							\
										\
    app >>= LN2PREZENMULT;							\
    ap = app & ((1<<LN2PREZENMULT)-1);						\
    dd ^= (BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*2);		\
										\
    app >>= LN2PREZENMULT;							\
    ap = app & ((1<<LN2PREZENMULT)-1);						\
    dd ^= (BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*4);		\
										\
    app >>= LN2PREZENMULT;							\
    d = dd ^ ((BigNumDigit)Ze2PreMultFast(app,app) << (LN2PREZENMULT*6));	\
} else
# else
				/* Word of 32 bits */
#define Ze2PolySquareDigitMacro(c, d, a, Rg)					\
if(1) {										\
    register BigNumDigit ap, app, cc, dd;					\
										\
    ap = (a) & ((1<<LN2PREZENMULT)-1);						\
    cc = Ze2PreMultFast(ap,ap);							\
										\
    app = (a) >> LN2PREZENMULT;							\
    ap = app & ((1<<LN2PREZENMULT)-1);						\
    c = cc ^ ((BigNumDigit)Ze2PreMultFast(ap,ap) << (LN2PREZENMULT*2));		\
										\
    app >>= LN2PREZENMULT;							\
    ap = app & ((1<<LN2PREZENMULT)-1);						\
    dd = Ze2PreMultFast(ap,ap);							\
										\
    app >>= LN2PREZENMULT;							\
    d = dd ^ ((BigNumDigit)Ze2PreMultFast(app,app) << (LN2PREZENMULT*2));	\
} else
# endif
#endif

/****************** Ze2PolyNormalizeDigit ***************/
#if !USE_MACROS
#  define Ze2PolyNormalizeDigit(r, p, q, Rg) \
          _Ze2PolyNormalizeDigitProc(&(r), p, q, Rg)
void _Ze2PolyNormalizeDigitProc(q, a, b, Rg) 
     BigNumDigit *q, a, b;
     ZENRing Rg;
{
# if SIZE_BLOC == 64
    register BigNumDigit qpp, qp, ap, app, bb, b7, b6, b5, b4, b3, b2, b1, b0;

    b0 = b & ((1<<LN2PREZENMULT)-1);
    app = b >> LN2PREZENMULT;
    b1 = app & ((1<<LN2PREZENMULT)-1);
    app >>= LN2PREZENMULT;
    b2 = app & ((1<<LN2PREZENMULT)-1);
    app >>= LN2PREZENMULT;
    b3 = app & ((1<<LN2PREZENMULT)-1);
    app >>= LN2PREZENMULT;
    b4 = app & ((1<<LN2PREZENMULT)-1);
    app >>= LN2PREZENMULT;
    b5 = app & ((1<<LN2PREZENMULT)-1);
    app >>= LN2PREZENMULT;
    b6 = app & ((1<<LN2PREZENMULT)-1);
    b7 = app >> LN2PREZENMULT;
    bb = b7 -(1 << (LN2PREZENMULT-1));

    qp = Ze2PreDivFast(a >> (LN2PREZENMULT*7),bb);
    ap = a ^ ((Ze2PreMultFast(qp,b0)) >> (LN2PREZENMULT-1));
    ap ^= Ze2PreMultFast(qp,b1)<<1;
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b2) << (1+LN2PREZENMULT);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b3) << (1+LN2PREZENMULT*2);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b4) << (1+LN2PREZENMULT*3);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b5) << (1+LN2PREZENMULT*4);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b6) << (1+LN2PREZENMULT*5);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b7) << (1+LN2PREZENMULT*6);
    qpp = qp;

    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*6),bb);
    ap ^= Ze2PreMultFast(qp,b1) >> (LN2PREZENMULT-1);
    ap ^= Ze2PreMultFast(qp,b2) << 1;
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b3) << (1+LN2PREZENMULT);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b4) << (1+LN2PREZENMULT*2);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b5) << (1+LN2PREZENMULT*3);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b6) << (1+LN2PREZENMULT*4);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b7) << (1+LN2PREZENMULT*5);
    qpp <<= LN2PREZENMULT;
    qpp |= qp;

    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*5),bb);
    ap ^= Ze2PreMultFast(qp,b2) >> (LN2PREZENMULT-1);
    ap ^= Ze2PreMultFast(qp,b3) << 1;
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b4) << (1+LN2PREZENMULT);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b5) << (1+LN2PREZENMULT*2);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b6) << (1+LN2PREZENMULT*3);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b7) << (1+LN2PREZENMULT*4);
    qpp <<= LN2PREZENMULT;
    qpp |= qp;

    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*4),bb);
    ap ^= Ze2PreMultFast(qp,b3) >> (LN2PREZENMULT-1);
    ap ^= Ze2PreMultFast(qp,b4) << 1;
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b5) << (1+LN2PREZENMULT);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b6) << (1+LN2PREZENMULT*2);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b7) << (1+LN2PREZENMULT*3);
    qpp <<= LN2PREZENMULT;
    qpp |= qp;

    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*3),bb);
    ap ^= Ze2PreMultFast(qp,b4) >> (LN2PREZENMULT-1);
    ap ^= Ze2PreMultFast(qp,b5) << 1;
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b6) << (1+LN2PREZENMULT);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b7) << (1+LN2PREZENMULT*2);
    qpp <<= LN2PREZENMULT;
    qpp |= qp;

    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*2),bb);
    ap ^= Ze2PreMultFast(qp,b5) >> (LN2PREZENMULT-1);
    ap ^= Ze2PreMultFast(qp,b6) << 1;
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b7) << (1+LN2PREZENMULT);
    qpp <<= LN2PREZENMULT;
    qpp |= qp;

    qp = Ze2PreDivFast(ap >> LN2PREZENMULT,bb);
    ap ^= Ze2PreMultFast(qp,b6) >> (LN2PREZENMULT-1);
    ap ^= Ze2PreMultFast(qp,b7) << 1;
    qpp <<= LN2PREZENMULT;
    qpp |= qp;

    qp = Ze2PreDivFast(ap,bb);
    qpp <<= LN2PREZENMULT;
    qpp |= qp;

    *q = qpp;
# else
    register BigNumDigit qpp, qp, ap, app, bb, b3, b2, b1, b0;

    b0 = b & ((1<<LN2PREZENMULT)-1);
    app = b >> LN2PREZENMULT;
    b1 = app & ((1<<LN2PREZENMULT)-1);
    app >>= LN2PREZENMULT;
    b2 = app & ((1<<LN2PREZENMULT)-1);
    app >>= LN2PREZENMULT;
    b3 = app & ((1<<LN2PREZENMULT)-1);
    bb = b3 -(1 << (LN2PREZENMULT-1));

    qp = Ze2PreDivFast(a >> (LN2PREZENMULT*3),bb);
    ap = a ^ ((Ze2PreMultFast(qp,b0)) >> (LN2PREZENMULT-1));
    ap ^= Ze2PreMultFast(qp,b1)<<1;
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b2) << (1+LN2PREZENMULT);
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b3) << (1+LN2PREZENMULT*2);
    qpp = qp;

    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*2),bb);
    ap ^= Ze2PreMultFast(qp,b1) >> (LN2PREZENMULT-1);
    ap ^= Ze2PreMultFast(qp,b2) << 1;
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b3) << (1+LN2PREZENMULT);
    qpp <<= LN2PREZENMULT;
    qpp |= qp;

    qp = Ze2PreDivFast(ap >> LN2PREZENMULT,bb);
    ap ^= Ze2PreMultFast(qp,b2) >> (LN2PREZENMULT-1);
    ap ^= Ze2PreMultFast(qp,b3) << 1;
    qpp <<= LN2PREZENMULT;
    qpp |= qp;

    qp = Ze2PreDivFast(ap,bb);
    qpp <<= LN2PREZENMULT;
    qpp |= qp;

    *q = qpp;
# endif
}
#else
#  define Ze2PolyNormalizeDigit(r, p, q, Rg) \
          Ze2PolyNormalizeDigitMacro(r, p, q, Rg)
# if SIZE_BLOC == 64
#define Ze2PolyNormalizeDigitMacro(q, a, b, Rg)					\
if(1) {										\
    register BigNumDigit qpp, qp, ap, app, bb, b7, b6, b5, b4, b3, b2, b1, b0;	\
										\
    b0 = b & ((1<<LN2PREZENMULT)-1);						\
    app = b >> LN2PREZENMULT;							\
    b1 = app & ((1<<LN2PREZENMULT)-1);						\
    app >>= LN2PREZENMULT;							\
    b2 = app & ((1<<LN2PREZENMULT)-1);						\
    app >>= LN2PREZENMULT;							\
    b3 = app & ((1<<LN2PREZENMULT)-1);						\
    app >>= LN2PREZENMULT;							\
    b4 = app & ((1<<LN2PREZENMULT)-1);						\
    app >>= LN2PREZENMULT;							\
    b5 = app & ((1<<LN2PREZENMULT)-1);						\
    app >>= LN2PREZENMULT;							\
    b6 = app & ((1<<LN2PREZENMULT)-1);						\
    b7 = app >> LN2PREZENMULT;							\
    bb = b7 -(1 << (LN2PREZENMULT-1));						\
										\
    qp = Ze2PreDivFast(a >> (LN2PREZENMULT*7),bb);				\
    ap = a ^ ((Ze2PreMultFast(b0,qp)) >> (LN2PREZENMULT-1));			\
    ap ^= Ze2PreMultFast(qp,b1)<<1;						\
    ap ^= (BigNumDigit)Ze2PreMultFast(b2,qp) << (1+LN2PREZENMULT);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b3,qp) << (1+LN2PREZENMULT*2);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b4,qp) << (1+LN2PREZENMULT*3);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b5,qp) << (1+LN2PREZENMULT*4);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b6,qp) << (1+LN2PREZENMULT*5);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b7,qp) << (1+LN2PREZENMULT*6);		\
    qpp = qp;									\
										\
    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*6),bb);				\
    ap ^= Ze2PreMultFast(b1,qp) >> (LN2PREZENMULT-1);				\
    ap ^= Ze2PreMultFast(b2,qp) << 1;						\
    ap ^= (BigNumDigit)Ze2PreMultFast(b3,qp) << (1+LN2PREZENMULT);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b4,qp) << (1+LN2PREZENMULT*2);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b5,qp) << (1+LN2PREZENMULT*3);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b6,qp) << (1+LN2PREZENMULT*4);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b7,qp) << (1+LN2PREZENMULT*5);		\
    qpp <<= LN2PREZENMULT;							\
    qpp |= qp;									\
										\
    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*5),bb);				\
    ap ^= Ze2PreMultFast(b2,qp) >> (LN2PREZENMULT-1);				\
    ap ^= Ze2PreMultFast(b3,qp) << 1;						\
    ap ^= (BigNumDigit)Ze2PreMultFast(b4,qp) << (1+LN2PREZENMULT);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b5,qp) << (1+LN2PREZENMULT*2);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b6,qp) << (1+LN2PREZENMULT*3);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b7,qp) << (1+LN2PREZENMULT*4);		\
    qpp <<= LN2PREZENMULT;							\
    qpp |= qp;									\
										\
    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*4),bb);				\
    ap ^= Ze2PreMultFast(b3,qp) >> (LN2PREZENMULT-1);				\
    ap ^= Ze2PreMultFast(b4,qp) << 1;						\
    ap ^= (BigNumDigit)Ze2PreMultFast(b5,qp) << (1+LN2PREZENMULT);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b6,qp) << (1+LN2PREZENMULT*2);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b7,qp) << (1+LN2PREZENMULT*3);		\
    qpp <<= LN2PREZENMULT;							\
    qpp |= qp;									\
										\
    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*3),bb);				\
    ap ^= Ze2PreMultFast(b4,qp) >> (LN2PREZENMULT-1);				\
    ap ^= Ze2PreMultFast(b5,qp) << 1;						\
    ap ^= (BigNumDigit)Ze2PreMultFast(b6,qp) << (1+LN2PREZENMULT);		\
    ap ^= (BigNumDigit)Ze2PreMultFast(b7,qp) << (1+LN2PREZENMULT*2);		\
    qpp <<= LN2PREZENMULT;							\
    qpp |= qp;									\
										\
    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*2),bb);				\
    ap ^= Ze2PreMultFast(b5,qp) >> (LN2PREZENMULT-1);				\
    ap ^= Ze2PreMultFast(b6,qp) << 1;						\
    ap ^= (BigNumDigit)Ze2PreMultFast(b7,qp) << (1+LN2PREZENMULT);		\
    qpp <<= LN2PREZENMULT;							\
    qpp |= qp;									\
										\
    qp = Ze2PreDivFast(ap >> LN2PREZENMULT,bb);					\
    ap ^= Ze2PreMultFast(b6,qp) >> (LN2PREZENMULT-1);				\
    ap ^= Ze2PreMultFast(b7,qp) << 1;						\
    qpp <<= LN2PREZENMULT;							\
    qpp |= qp;									\
										\
    qp = Ze2PreDivFast(ap,bb);							\
    qpp <<= LN2PREZENMULT;							\
    qpp |= qp;									\
										\
    q = qpp;									\
} else
# else
#define Ze2PolyNormalizeDigitMacro(q, a, b, Rg)				\
if(1) {									\
    register BigNumDigit qpp, qp, ap, app, bb, b3, b2, b1, b0;		\
									\
    b0 = b & ((1<<LN2PREZENMULT)-1);					\
    app = b >> LN2PREZENMULT;						\
    b1 = app & ((1<<LN2PREZENMULT)-1);					\
    app >>= LN2PREZENMULT;						\
    b2 = app & ((1<<LN2PREZENMULT)-1);					\
    app >>= LN2PREZENMULT;						\
    b3 = app & ((1<<LN2PREZENMULT)-1);					\
    bb = b3 -(1 << (LN2PREZENMULT-1));					\
									\
    qp = Ze2PreDivFast(a >> (LN2PREZENMULT*3),bb);			\
    ap = a ^ ((Ze2PreMultFast(qp,b0)) >> (LN2PREZENMULT-1));		\
    ap ^= Ze2PreMultFast(qp,b1)<<1;					\
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b2) << (1+LN2PREZENMULT);	\
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b3) << (1+LN2PREZENMULT*2);	\
    qpp = qp;								\
									\
    qp = Ze2PreDivFast(ap >> (LN2PREZENMULT*2),bb);			\
    ap ^= Ze2PreMultFast(qp,b1) >> (LN2PREZENMULT-1);			\
    ap ^= Ze2PreMultFast(qp,b2) << 1;					\
    ap ^= (BigNumDigit)Ze2PreMultFast(qp,b3) << (1+LN2PREZENMULT);	\
    qpp <<= LN2PREZENMULT;						\
    qpp |= qp;								\
									\
    qp = Ze2PreDivFast(ap >> LN2PREZENMULT,bb);				\
    ap ^= Ze2PreMultFast(qp,b2) >> (LN2PREZENMULT-1);			\
    ap ^= Ze2PreMultFast(qp,b3) << 1;					\
    qpp <<= LN2PREZENMULT;						\
    qpp |= qp;								\
									\
    qp = Ze2PreDivFast(ap,bb);						\
    qpp <<= LN2PREZENMULT;						\
    qpp |= qp;								\
									\
    q = qpp;								\
} else
# endif
#endif

/***************** Ze2mNormalize *******************/
#if !USE_MACROS
#  define Ze2mNormalize(a, b, c, d, Rg) \
     Ze2mNormalizeProc(&a, &b, &c, &d, Rg)

void Ze2mNormalizeProc(a, b, c, d, Rg)
     BigNumDigit *a, *b, *c, *d;
     ZENRing Rg;
{
# if SIZE_BLOC == 64
    BigNumDigit digit; 
    
    digit = ((*c >> (Ze2mRingNbBits(Rg)-1)) ^ 
	     (*d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1))) >> ((LN2PREZENMULT<<1)*3); 
    *a ^= (BigNumDigit)Ze2mRingPreDivFast(Rg)[digit]<<((LN2PREZENMULT<<1)*3); 
    *b ^= Ze2mRingPreDivFast(Rg)[digit]>>((LN2PREZENMULT<<1)); 
 
    digit = ((*c >> (Ze2mRingNbBits(Rg)-1)) ^ 
	     (*d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1))) >> ((LN2PREZENMULT<<1)*2); 
    *a ^= (BigNumDigit)Ze2mRingPreDivFast(Rg)[digit &
					     ((1<<(LN2PREZENMULT<<1))-1)] <<((LN2PREZENMULT<<1)*2); 
    
    digit = ((*c >> (Ze2mRingNbBits(Rg)-1)) ^ 
	     (*d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1))) >> 
	       ((LN2PREZENMULT<<1)); 
    *a ^= (BigNumDigit)Ze2mRingPreDivFast(Rg)[digit & ((1<<(LN2PREZENMULT<<1))-1)] 
      <<((LN2PREZENMULT<<1)); 
     
    digit = ((*c >> (Ze2mRingNbBits(Rg)-1)) ^ 
	     (*d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1))); 
    *a ^= Ze2mRingPreDivFast(Rg)[digit & ((1<<(LN2PREZENMULT<<1))-1)]; 
# else
    BigNumDigit digit; 
 
    digit = ((*c >> (Ze2mRingNbBits(Rg)-1)) ^ 
	     (*d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1))) >> 
	       ((LN2PREZENMULT<<1)); 
    *a ^= (BigNumDigit)Ze2mRingPreDivFast(Rg)[digit]<<((LN2PREZENMULT<<1)); 
    *b ^= (BigNumDigit)Ze2mRingPreDivFast(Rg)[digit]>>((LN2PREZENMULT<<1)); 
     
    digit = ((*c >> (Ze2mRingNbBits(Rg)-1)) ^ 
	     (*d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1))); 
    *a ^= Ze2mRingPreDivFast(Rg)[digit & ((1<<(LN2PREZENMULT<<1))-1)]; 
# endif
  *d = 0; 
  *c &= ((BigNumDigit)1<<(Ze2mRingNbBits(Rg)-1))-1; 
}

#else
#  define Ze2mNormalize(a, b, c, d, Rg) \
     Ze2mNormalizeMacro(a, b, c, d, Rg)

# if SIZE_BLOC == 64
#define Ze2mNormalizeMacro(a, b, c, d, Rg)						\
if(1){											\
    BigNumDigit digit;									\
											\
    digit = ((c >> (Ze2mRingNbBits(Rg)-1)) ^						\
	     (d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1))) >>				\
	       ((LN2PREZENMULT<<1)*3);							\
    a ^= (BigNumDigit)Ze2mRingPreDivFast(Rg)[digit]<<((LN2PREZENMULT<<1)*3);		\
    b ^= Ze2mRingPreDivFast(Rg)[digit]>>((LN2PREZENMULT<<1));				\
											\
    digit = ((c >> (Ze2mRingNbBits(Rg)-1)) ^						\
	     (d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1))) >>				\
	       ((LN2PREZENMULT<<1)*2);							\
    a ^= (BigNumDigit)Ze2mRingPreDivFast(Rg)[digit & ((1<<(LN2PREZENMULT<<1))-1)]	\
      <<((LN2PREZENMULT<<1)*2);								\
											\
    digit = ((c >> (Ze2mRingNbBits(Rg)-1)) ^						\
	     (d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1))) >>				\
	       ((LN2PREZENMULT<<1));							\
    a ^= (BigNumDigit)Ze2mRingPreDivFast(Rg)[digit & ((1<<(LN2PREZENMULT<<1))-1)]	\
      <<((LN2PREZENMULT<<1));								\
											\
    digit = ((c >> (Ze2mRingNbBits(Rg)-1)) ^						\
	     (d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1)));					\
    a ^= Ze2mRingPreDivFast(Rg)[digit & ((1<<(LN2PREZENMULT<<1))-1)];			\
											\
  d = 0;										\
  c &= ((BigNumDigit)1<<(Ze2mRingNbBits(Rg)-1))-1;					\
}else
# else
#define Ze2mNormalizeMacro(a, b, c, d, Rg)					\
if(1){										\
    BigNumDigit digit;								\
										\
    digit = ((c >> (Ze2mRingNbBits(Rg)-1)) ^					\
	     (d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1))) >>			\
	       ((LN2PREZENMULT<<1));						\
    a ^= (BigNumDigit)Ze2mRingPreDivFast(Rg)[digit]<<((LN2PREZENMULT<<1));	\
    b ^= (BigNumDigit)Ze2mRingPreDivFast(Rg)[digit]>>((LN2PREZENMULT<<1));	\
										\
    digit = ((c >> (Ze2mRingNbBits(Rg)-1)) ^					\
	     (d << (SIZE_BLOC-Ze2mRingNbBits(Rg)+1)));				\
    a ^= Ze2mRingPreDivFast(Rg)[digit & ((1<<(LN2PREZENMULT<<1))-1)];		\
										\
  d = 0;									\
  c &= ((BigNumDigit)1<<(Ze2mRingNbBits(Rg)-1))-1;				\
} else
# endif
#endif

/* \TEX[
   \title{Squaring.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE| for compatibility.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _Ze2PolySquare(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int ds, shift;
  BigNum h, l, a;
  BigNumDigit tmp, sa;

  /* Cleaning garbage high bits of PX */
  ds = divSizeBloc(Ze2PolyDeg(PX, Rg));
  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(PX, Rg));
  Ze2PolyGetDigit(PX, ds, Rg) <<= shift;
  Ze2PolyGetDigit(PX, ds, Rg) >>= shift;

  /* Setting the degree of RX */
  Ze2PolyDeg(RX, Rg) =  Ze2PolyDeg(PX, Rg)<<1;

  /* Where is the highest word of the result ? */
  if (divSizeBloc(Ze2PolyDeg(RX, Rg)) < (ds<<1)+1) {
    h = &tmp;
    tmp = 0;			/* for purify !!! */
    l = Ze2PolyDigit(RX, Rg)+(ds<<1);
  }
  else {
    h = Ze2PolyDigit(RX, Rg)+(ds<<1)+1;
    l = h-1;
  }

  /* Usual algorithm to square one polynomial */
  if ((a = Ze2PolyDigit(PX, Rg)+ds) >= Ze2PolyDigit(PX, Rg)) {
    Ze2PolySquareDigit(*l, *h, *a, Rg);
    
    h = l-1; l -= 2;
    while(a > Ze2PolyDigit(PX, Rg)) {
      a--;
      sa = *a;
      Ze2PolySquareDigit(*l, *h, sa, Rg);
      h -= 2; l -= 2;
    }
  }
  return(ZEN_HAS_INVERSE);
}

/* \TEX[
   \title{Scalar product.}
   \input{Two polynomials |PX|, |RX| and a pointer to a |ZENElt| |e| of a
   finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE| for compatibility.}
   \side{$|*e| = \sum p_iq_i$}
   }
   */
void _Ze2PolyDot(e, PX, QX, Rg)
     ZENElt *e;
     ZENPoly PX, QX;
     ZENRing Rg;
{
  BigNumDigit S;
  int dp, dq, ds, i, shift;
  
  dp = divSizeBloc(Ze2PolyDeg(PX, Rg));
  dq = divSizeBloc(Ze2PolyDeg(QX, Rg));
  
  if (dp < dq) {
    
    /* Cleaning garbage high bits of PX */
    shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(PX, Rg));
    Ze2PolyGetDigit(PX, dp, Rg) <<= shift;
    Ze2PolyGetDigit(PX, dp, Rg) >>= shift;

    ds = dp;
  }
  else if (dp == dq) {
    /* Cleaning garbage high bits of PX */
    shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(PX, Rg));
    Ze2PolyGetDigit(PX, dp, Rg) <<= shift;
    Ze2PolyGetDigit(PX, dp, Rg) >>= shift;

    /* Cleaning garbage high bits of QX */
    shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(QX, Rg));
    Ze2PolyGetDigit(QX, dq, Rg) <<= shift;
    Ze2PolyGetDigit(QX, dq, Rg) >>= shift;

    ds = dp;
  }
  else {
    /* Cleaning garbage high bits of QX */
    shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(QX, Rg));
    Ze2PolyGetDigit(QX, dq, Rg) <<= shift;
    Ze2PolyGetDigit(QX, dq, Rg) >>= shift;

    ds = dq;
  }

  S = Ze2PolyGetDigit(PX, 0, Rg) & Ze2PolyGetDigit(QX, 0, Rg);
  for(i = 1; i <= ds; i++) {
    S ^= Ze2PolyGetDigit(PX, i, Rg) & Ze2PolyGetDigit(QX, i, Rg);
  }

  for(i = BN_DIGIT_SIZE>>1; i > 0; i >>= 1) {
    S ^= (S>>i);
  }
  Ze2Elt2Int(*e, Rg) = S & 1;
}

/* \TEX[
   \title{Evaluation.}
   \input{A polynomial |PX| and two elements |f| and |e| of a
   finite ring |Rg|.}
   \side{$|f| = |PX|(e)$.}
   }
   */
void _Ze2PolyEval(f, PX, e, Rg)
     ZENElt *f, e;
     ZENPoly PX;
     ZENRing Rg;
{
  int i;

  if (Ze2EltIsZero(e, Rg)) {
    if(Ze2PolyGetDigit(PX, 0, Rg)&1)
      Ze2EltSetToOne(*f, Rg);
    else
      Ze2EltSetToZero(*f, Rg);
  }
  else {
    int dp, shift;
    BigNumDigit c;
    
    /* Cleaning garbage high bits of PX */
    dp = divSizeBloc(Ze2PolyDeg(PX, Rg));
    shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(PX, Rg));
    c = Ze2PolyGetDigit(PX, dp, Rg) << shift;
    for(i = dp-1; i >= 0; i--) {
      c ^= Ze2PolyGetDigit(PX, i, Rg);
    }
    Ze2Elt2Int(*f, Rg) = ZBNWeightDigitFast(c)&1;
  }
}

/* \TEX[
   \title{Multiplication by a scalar.}
   \input{Two |ZENPoly| |RX|, |PX| and a |ZENElt| |e| of a finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE| for compatibility.}
   \side{$|RX| = |PX|\times|e|$.}
   }
   */
void _Ze2PolyMultiplyScalar(RX, PX, e, Rg)
     ZENPoly RX, PX;
     ZENElt e;
     ZENRing Rg;
{
  if (Ze2EltIsZero(e, Rg)) {
    Ze2PolySetToZero(RX, Rg);
  }
  else {
    Ze2PolyAssign(RX, PX, Rg);
  }
}

/* \TEX[
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE| for compatibility.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$.}
   }
   */
int _Ze2PolyMultiply(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int dsa, dsb, shift, deg;
  BigNum a, b, h, k, l, p, q;
  BigNumDigit sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7;
  BigNumDigit sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7;
  BigNumDigit cc, dd, tmp;

  /* If one argument is null => EASY */
  if (Ze2PolyIsZero(PX, Rg) || Ze2PolyIsZero(QX, Rg)) {
    Ze2PolySetToZero(RX, Rg);
    return(ZEN_HAS_INVERSE);
  }
  
  /* Cleaning garbage high bits of PX */
  dsa = divSizeBloc(Ze2PolyDeg(PX, Rg));
  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(PX, Rg));
  Ze2PolyGetDigit(PX, dsa, Rg) <<= shift;
  Ze2PolyGetDigit(PX, dsa, Rg) >>= shift;

  /* Cleaning garbage high bits of QX */
  dsb = divSizeBloc(Ze2PolyDeg(QX, Rg));
  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(QX, Rg));
  Ze2PolyGetDigit(QX, dsb, Rg) <<= shift;
  Ze2PolyGetDigit(QX, dsb, Rg) >>= shift;

  /* Setting the degree of RX */
  Ze2PolyDeg(RX, Rg) = 
    Ze2PolyDeg(PX, Rg)+Ze2PolyDeg(QX, Rg);
  
  /* Usual algorithm to multiply 2 polynomials */
  b = Ze2PolyDigit(QX, Rg)+dsb;
    
  /* Where is the highest word of the result ? */
  k = Ze2PolyDigit(RX, Rg)+dsb+1;
  if (divSizeBloc(Ze2PolyDeg(RX, Rg)) < dsa+dsb+1) {
    ZBNSetToZero(k, dsa);
    k += dsa;
    h = &tmp;
    tmp = 0;			/* for purify !!! */
    l = k-1;
  }
  else {
    ZBNSetToZero(k, dsa+1);
    k += dsa;
    h = k;
    l = k-1;
  }
  Ze2PolyCutInChars(sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7, *b);
  
  a = Ze2PolyDigit(PX, Rg)+dsa;
  if (a > Ze2PolyDigit(PX, Rg)) {
    Ze2PolyCutInChars(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, *a);
    cc = *l; dd = *h;
    Ze2PolyMultiplyChars(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7);
    *l = cc; *h = dd;
    h = l; l--;
    
    for(a--; a > Ze2PolyDigit(PX, Rg); a--) {
      Ze2PolyCutInChars(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, *a);
      cc = *l; dd = *h;
      Ze2PolyMultiplyChars(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7);
      *l = cc; *h = dd;
      h = l; l--;
    }
  }
  *l = 0;
  Ze2PolyCutInChars(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, *a);
  cc = *l; dd = *h;
  Ze2PolyMultiplyChars(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7);
  *l = cc; *h = dd;

  while(b > Ze2PolyDigit(QX, Rg)) {
    b--;
    h = --k;
    l = h-1;
    Ze2PolyCutInChars(sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7, *b);
    for(a = Ze2PolyDigit(PX, Rg)+dsa;
	a > Ze2PolyDigit(PX, Rg);
	a--) {
      Ze2PolyCutInChars(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, *a);
      cc = *l; dd = *h;
      Ze2PolyMultiplyChars(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7);
      *l = cc; *h = dd;
      h = l; l--;
    }
    *l = 0;
    Ze2PolyCutInChars(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, *a);
    cc = *l; dd = *h;
    Ze2PolyMultiplyChars(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7);
    *l = cc; *h = dd;
  }
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Lazy division.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder. We must have an allocated size of |PX| $\ge$ |SIZE_CHAR-1-modSizeChar(Ze2PolyDeg(QX, Rg))|.}
   \output{|ZEN_HAS_INVERSE|.}
   \note{|PX| is destroyed.}
   }
   */
int _Ze2PolyDivideLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int dsa, dsb, ds, dsr, shift, i;
  ZENPoly buf;
  BigNum h, l, a, b;
  BigNumDigit sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7;
  BigNumDigit sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7;
  BigNumDigit pb, sc, lp, hpp;

  /* easy cases */
  if (Ze2PolyDeg(PX, Rg) < Ze2PolyDeg(QX, Rg)) {
    Ze2PolySetToZero(RX, Rg);
    Ze2PolyAssign(MX, PX, Rg);
    return(ZEN_HAS_INVERSE);
  }

  /* Cleaning garbage high bits of PX */
  dsa = divSizeBloc(Ze2PolyDeg(PX, Rg));
  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(PX, Rg));
  Ze2PolyGetDigit(PX, dsa, Rg) <<= shift;
  Ze2PolyGetDigit(PX, dsa, Rg) >>= shift;

  /* Shifting QX left as much as possible */
  dsb = divSizeBloc(Ze2PolyDeg(QX, Rg));
  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(QX, Rg));
  ZBNShiftLeft(Ze2PolyDigit(QX, Rg), dsb+1, shift);

  if ((ds = divSizeBloc(Ze2PolyDeg(PX, Rg)+shift)) > dsa) {
    Ze2PolyGetDigit(PX, ds, Rg) = 0;
  }
  ZBNShiftLeft(Ze2PolyDigit(PX, Rg), ds+1, shift);

  pb = Ze2PolyGetDigit(QX, dsb, Rg);
  for(a = Ze2PolyDigit(PX, Rg)+ds;
      a > Ze2PolyDigit(PX, Rg)+dsb;
      a--) {
    Ze2PolyNormalizeDigit(sc, *a, pb, Rg);
    Ze2PolyCutInChars(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sc);
    h = a+1; l = a;
    b = Ze2PolyDigit(QX, Rg)+dsb+1;
    while(b > Ze2PolyDigit(QX, Rg)) {
      b--;
      h = l; l--;
      Ze2PolyCutInChars(sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7, *b);

      lp = hpp = 0;
      Ze2PolyMultiplyChars(lp, hpp, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7);
      *l ^= lp<<1;
      *h ^= (hpp<<1)|(lp>>(SIZE_BLOC-1));
    }
    *a = sc;
  }

  /* Restoring QX */
  ZBNShiftRight(Ze2PolyDigit(QX, Rg), dsb+1, shift);

  /* Getting the quotient */
  Ze2PolyDeg(RX,Rg) =  
    Ze2PolyDeg(PX, Rg) >= Ze2PolyDeg(QX, Rg)
      ?
	Ze2PolyDeg(PX, Rg)-Ze2PolyDeg(QX, Rg)
	  :
	    0;
  dsr = divSizeBloc(Ze2PolyDeg(RX, Rg));
  Ze2PolyGetDigit(RX, dsr, Rg) = 0;
  ZBNAssign(Ze2PolyDigit(RX, Rg), Ze2PolyDigit(PX, Rg)+dsb+1, ds > dsb ? ds-dsb : 0);
  ZBNShiftLeft(Ze2PolyDigit(RX, Rg), dsr+1, 1);

  /* Getting the remainder */
  Ze2PolyGetDigit(MX, 0, Rg) = 0;
  ds = ds > dsb ? dsb : ds;
  ZBNShiftRight(Ze2PolyDigit(PX, Rg), ds+1, shift);
  if ((ds == dsb) &&
      (Ze2PolyGetDigit(PX, dsb, Rg) & ((BigNumDigit)1 << (SIZE_BLOC-1-shift)))) {
    Ze2PolyGetDigit(RX, 0, Rg) |= 1;
    for(i = 0; i <= dsb; i++) {
      Ze2PolyGetDigit(PX, i, Rg) ^= Ze2PolyGetDigit(QX, i, Rg);
    }
  }
  Ze2PolyDeg(PX, Rg) =  (ds+1)*SIZE_BLOC-1;
  Ze2PolyUpdateDegree(PX, Rg);
  Ze2PolyAssign(MX, PX, Rg);
  
  return(ZEN_HAS_INVERSE);
}

/* \TEX[
   \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| or |ZENERR| if an error occured}
   }
   */
int _Ze2PolyDivide(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int ret, shift;
  ZENPoly Pbuf;

  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(QX, Rg));
  if (!(Pbuf = ZENPolyCreate(Ze2PolyDeg(PX, Rg)+shift, Rg))) {
    return(ZERR);
  }
  Ze2PolyAssign(Pbuf, PX, Rg);
  ret = Ze2PolyDivideLazy(RX, MX, Pbuf, QX, Rg);
  Ze2PolyFree(Pbuf, Rg);
  return(ret);
}

/* \TEX[
   \title{Lazy mudulo computation.}
   \input{A polynomial |PX| to divide by
   |QX|$=X^{\mbox{degree}(QX)}+\sum_{i=0}^{|k|}q_iX^i$ in a finite
   ring |Rg|. |MX| will be the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| for an error}
   \note{|PX| is destroyed.}
   }
   */
int _Ze2PolyModuloLazy(MX, PX, QX, k, Rg)
     ZENPoly MX, PX, QX;
     int k;
     ZENRing Rg;
{
  int dsa, dsb, ds, dsr, dsk, shift, i;
  ZENPoly buf;
  BigNum h, l, a, b;
  BigNumDigit pb, sc, lp, hpp;
  BigNumDigit sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7;
  BigNumDigit sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7;

  /* Cleaning garbage high bits of PX */
  dsa = divSizeBloc(Ze2PolyDeg(PX, Rg));
  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(PX, Rg));
  Ze2PolyGetDigit(PX, dsa, Rg) <<= shift;
  Ze2PolyGetDigit(PX, dsa, Rg) >>= shift;

  /* Shifting QX left as much as possible */
  dsb = divSizeBloc(Ze2PolyDeg(QX, Rg));
  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(QX, Rg));
  ZBNShiftLeft(Ze2PolyDigit(QX, Rg), dsb+1, shift);

  if ((ds = divSizeBloc(Ze2PolyDeg(PX, Rg)+shift)) > dsa) {
    Ze2PolyGetDigit(PX, ds, Rg) = 0;
  }
  ZBNShiftLeft(Ze2PolyDigit(PX, Rg), ds+1, shift);
  dsk = divSizeBloc(k+shift);

  pb = Ze2PolyGetDigit(QX, dsb, Rg);
  for(a = Ze2PolyDigit(PX, Rg)+ds;
      a > Ze2PolyDigit(PX, Rg)+dsb;
      a--) {
    Ze2PolyNormalizeDigit(sc, *a, pb, Rg);
    Ze2PolyCutInChars(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sc);
    h = a - dsb + dsk +1; l = h -1;
    b = Ze2PolyDigit(QX, Rg)+dsk+1;
    while(b > Ze2PolyDigit(QX, Rg)) {
      b--; h = l; l--;

      Ze2PolyCutInChars(sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7, *b);
      lp = hpp = 0;
      Ze2PolyMultiplyChars(lp, hpp, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7);
      *l ^= lp<<1;
      *h ^= (hpp<<1)|(lp>>(SIZE_BLOC-1));
    }
    *a = sc;
  }

  /* Restoring QX */
  ZBNShiftRight(Ze2PolyDigit(QX, Rg), dsb+1, shift);

  /* Getting the remainder */
  ds = ds > dsb ? dsb : ds;
  ZBNShiftRight(Ze2PolyDigit(PX, Rg), ds+1, shift);
  if ((ds == dsb) &&
      (Ze2PolyGetDigit(PX, dsb, Rg) & ((BigNumDigit)1 << (SIZE_BLOC-1-shift)))) {
    for(i = 0; i <= dsb; i++) {
      Ze2PolyGetDigit(PX, i, Rg) ^= Ze2PolyGetDigit(QX, i, Rg);
    }
  }
  Ze2PolyDeg(PX, Rg) =  (ds+1)*SIZE_BLOC-1;
  Ze2PolyUpdateDegree(PX, Rg);
  Ze2PolyAssign(MX, PX, Rg);

  return(ZEN_HAS_INVERSE);
}

#define RINGDEF ZENRingDef(R)

/* \TEX{
   \title{Multiplication in $\GF{2^m}$ with precomputations.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |R|.}
   \side{$|*p_c| = a\times b$.}
   }
   */
void _Ze2mEltMultiply(p_ec, ea, eb, R)
     ZENElt *p_ec, ea, eb;
     ZENRing R;
{
#define PX ZENElt2Pol(ea, R)
#define QX ZENElt2Pol(eb, R)
#define RX ZENRingPbuf(R)

  int dsa, dsb, ds, dsr, i, shift;
  BigNum b, k, p, q, pph, ppl;
  
  /* If one argument is 0 or 1 => EASY */
  if (!Ze2PolyDeg(PX, RINGDEF)) {
    if (Ze2PolyGetDigit(PX, 0, RINGDEF) & 1) {
      ZENEltAssign(*p_ec, eb, R);
    }
    else {
      ZENEltSetToZero(*p_ec, R);
    }
    return;
  }
  else if (!Ze2PolyDeg(QX, RINGDEF)) {
    if (Ze2PolyGetDigit(QX, 0, RINGDEF) & 1) {
      ZENEltAssign(*p_ec, ea, R);
    }
    else {
      ZENEltSetToZero(*p_ec, R);
    }
    return;
  }

  /* Cleaning garbage high bits of PX */
  dsa = divSizeBloc(Ze2PolyDeg(PX, Rg));
  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(PX, Rg));
  Ze2PolyGetDigit(PX, dsa, Rg) <<= shift;
  Ze2PolyGetDigit(PX, dsa, Rg) >>= shift;

  /* Cleaning garbage high bits of QX */
  dsb = divSizeBloc(Ze2PolyDeg(QX, Rg));
  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(QX, Rg));
  Ze2PolyGetDigit(QX, dsb, Rg) <<= shift;
  Ze2PolyGetDigit(QX, dsb, Rg) >>= shift;

  Ze2PolyDeg(RX, RINGDEF) = Ze2PolyDeg(PX, RINGDEF)+Ze2PolyDeg(QX, RINGDEF);
  ds = divSizeBloc(Ze2PolyDeg(RX, RINGDEF));

  
  /* Usual algorithm to multiply 2 polynomials */
  k = Ze2PolyDigit(RX, RINGDEF);
  ZBNSetToZero(Ze2PolyDigit(RX, RINGDEF), dsa+2);
  for(b = Ze2PolyDigit(QX, RINGDEF);
      b <= Ze2PolyDigit(QX, RINGDEF)+dsb;
      b++) {
    register BigNumDigit sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7;
    register int j;
    register BigNum h, l, a;
    BigNumDigit dd;

    Ze2PolyCutInChars(sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7, *b);    
    l = k;
    h = ++k;
    dd = *l;
    for(a = Ze2PolyDigit(PX, RINGDEF);
	a <= Ze2PolyDigit(PX, RINGDEF)+dsa;
	a++) {
      register BigNumDigit sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7;
      BigNumDigit cc;
      
      Ze2PolyCutInChars(sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, *a);
      cc = dd;
      dd = *h;
      Ze2PolyMultiplyChars(cc, dd, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, sb0, sb1, sb2, sb3, sb4, sb5, sb6, sb7);
      *l = cc;
      l = h;
      h++;
    }
    *l = dd;
    *h = 0;
  }

  /* Usual algorithm to get the modulo of a polynomial */
  dsr = divSizeBloc(ZENRingDeg(R));
  for(i = ds+1; i <= 2*dsr+1; i++) {
    ZENPolyGetDigit(RX, i, RINGDEF) = 0;
  }
  
  for(pph = ZENPolyDigit(RX, RINGDEF)+2*dsr+1, p = pph-dsr;
      p > ZENPolyDigit(RX, RINGDEF);
      pph--, p--) {
    q = p-1; ppl = pph-1;
    Ze2mNormalize(*q, *p, *ppl, *pph, R);
  }
  Ze2PolyDeg(RX, RINGDEF) = Ze2PolyDeg(RX, RINGDEF) < ZENRingDeg(R) ? 
    Ze2PolyDeg(RX, RINGDEF) : ZENRingDeg(R)-1;
  Ze2PolyUpdateDegree(RX, RINGDEF);
  Ze2PolyAssign(ZENElt2Pol(*p_ec, R), RX, RINGDEF);  
}

/* \TEX{
   \title{Squaring in $\GF{2^m}$ with Precomputations.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   }
   */
void _Ze2mEltSquare(p_ec, ea, R)
     ZENElt *p_ec, ea;
     ZENRing R;
{
  int dsa, ds, dsr, i, shift;
  BigNum a, b, h, k, l, p, q, pph, ppl;
  BigNumDigit sa, sb, tmp;
  
  /* If one argument is 0 or 1 => EASY */
  if (!Ze2PolyDeg(PX, RINGDEF)) {
    if (Ze2PolyGetDigit(PX, 0, RINGDEF) & 1) {
      ZENEltSetToOne(*p_ec, R);
    }
    else {
      ZENEltSetToZero(*p_ec, R);
    }
    return;
  }

  /* size of PX and RX*/
  dsa = divSizeBloc(Ze2PolyDeg(PX, RINGDEF));
  Ze2PolyDeg(RX, RINGDEF) = 2*Ze2PolyDeg(PX, RINGDEF);
  ds = 2*dsa+1;

  /* Cleaning garbage high bits of PX */
  shift = SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(PX, Rg));
  Ze2PolyGetDigit(PX, dsa, Rg) <<= shift;
  Ze2PolyGetDigit(PX, dsa, Rg) >>= shift;

  /* Usual algorithm to multiply 2 polynomials */
  l = Ze2PolyDigit(RX, RINGDEF)+ds+1;
  
  a = Ze2PolyDigit(PX, RINGDEF)+dsa+1;
  while(a > Ze2PolyDigit(PX, RINGDEF)) {
    a--; sa = *a;
    h = l-1; l = h-1;

    Ze2PolySquareDigit(*l, *h, sa, R);
  }

  dsr = divSizeBloc(ZENRingDeg(R));
  for(i = ds+1; i <= 2*dsr+1; i++) {
    ZENPolyGetDigit(RX, i, RINGDEF) = 0;
  }
  
  for(pph = ZENPolyDigit(RX, RINGDEF)+2*dsr+1, p = pph-dsr;
      p > ZENPolyDigit(RX, RINGDEF);
      pph--, p--) {
    q = p-1; ppl = pph-1;
    Ze2mNormalize(*q, *p, *ppl, *pph, R);
  }
  Ze2PolyDeg(RX, RINGDEF) = Ze2PolyDeg(RX, RINGDEF) < ZENRingDeg(R) ? 
    Ze2PolyDeg(RX, RINGDEF) : ZENRingDeg(R);
  Ze2PolyUpdateDegree(RX, RINGDEF);
  Ze2PolyAssign(ZENElt2Pol(*p_ec, R), RX, RINGDEF);  
}
