/*
 * zen/ze2/c/poly_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial tests}
   
   Three functions are provided to do tests: |_Ze2PolyAreEqual| to test if two
   |ZENPoly| are equal, |_Ze2PolyIsZero| to test if a |ZENPoly| is equal to
   zero, |_Ze2PolyIsXi| to test if a |ZENPoly| is $X^i$.
   }
   */
#include <stdio.h>
#include "ze2.h"

/* \TEX[
   \title{Equal.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{|RX| == |PX|.}
   }
   */
int _Ze2PolyAreEqual(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int ms, ds;

  if (Ze2PolyDeg(RX, Rg) != Ze2PolyDeg(PX, Rg)) return(0);  

  if ((ds = divSizeBloc(Ze2PolyDeg(RX, Rg))) &&
      !ZBNAreEqual(Ze2PolyDigit(RX, Rg), ds, Ze2PolyDigit(PX, Rg), ds)) {
    return(0);
  }

  ms = modSizeBloc(Ze2PolyDeg(RX, Rg));
  if ((Ze2PolyGetDigit(RX, ds, Rg) << (SIZE_BLOC-1-ms)) !=
      (Ze2PolyGetDigit(PX, ds, Rg) << (SIZE_BLOC-1-ms))) return(0);

  return(1);
}

/* \TEX[
   \title{Is zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \output{|RX| == 0.}
   }
   */
int _Ze2PolyIsZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  return(!Ze2PolyDeg(RX, Rg) && !(Ze2PolyGetDigit(RX, 0, Rg) & 1));
}

/* \TEX[
   \title{Is $X^i$.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg| and an integer $i$
   smaller than the degree of |RX|.}
   \output{|RX| == $X^i$.}
   }
   */
int _Ze2PolyIsXi(RX, i, Rg)
     ZENPoly RX;
     int i;
     ZENRing Rg;
{
  int d, m, j;

  if (Ze2PolyDeg(RX, Rg) != i) {
    return(0);
  }

  d = divSizeBloc(i);
  m = modSizeBloc(i);
  for(j = 0; j < d; j++) {
    if (Ze2PolyGetDigit(RX, j, Rg)) {
      return(0);
    }
  }
  if ((Ze2PolyGetDigit(RX, j, Rg) << (SIZE_BLOC-m-1)) !=
      ((BigNumDigit)1 << (SIZE_BLOC-1))) {
    return(0);
  }
  return(1);
}
