/*
 * zen/zec/c/elt_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element additions}
   
   |_ZecEltAdd| enable us to add a |ZENElt| to another.
   }
   */
#include <stdio.h>
#include "zec.h"

/* \TEX[
   \title{Addition.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |Rg|.}
   \side{|*p_c| += |a|.}
   }
   */
void _ZecEltAdd(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltAdd(ZecElt2Elt(*p_c,i,Rg),ZecElt2Elt(a,i,Rg),ZecRingSubring(Rg,i));
}

/* \TEX[
   \title{Negation.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |Rg|.}
   \side{|*p_c| = -|a|.}
   }
   */
void _ZecEltNegate(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltNegate(ZecElt2Elt(*p_c,i,Rg),ZecElt2Elt(a,i,Rg),ZecRingSubring(Rg,i));
}


/* \TEX[
   \title{Subtract.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |Rg|.}
   \side{|*p_c| -= |a|.}
   }
   */
void _ZecEltSubtract(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltSubtract(ZecElt2Elt(*p_c,i,Rg),ZecElt2Elt(a,i,Rg),ZecRingSubring(Rg,i));
}
