/*
 * zen/zec/c/elt_alloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element creation}
   
   |_ZecEltCreate| allocates one |ZENElt| and |_ZecEltFree| frees then it.
   }
   */
#include <stdio.h>
#include "zec.h"

/* \TEX[
   \title{Creation.}
   \input{A finite ring |Rg| and a pointer on a |ZENElt|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   }
   */
int _ZecEltCreate(a, Rg)
     ZENElt *a;
     ZENRing Rg;
{
  int i;

  i=ZecRingNbS(Rg);
  if((ZecElt2Elts(*a,Rg)=(ZENElt *)ZENMalloc(i*sizeof(ZENElt)))==NULL)
    {
      ZENSetError(ZEN_LIB_ZEC,ZEN_MALLOC,ZEN_ALLOC);
      return(ZERR);
    }
  for(;i--;)
    {
      if(ZENEltCreate(ZecElt2Elt(*a,i,Rg),ZecRingSubring(Rg,i)))
	{
	  for(i++;i<ZecRingNbS(Rg);i++)
	    ZENEltFree(ZecElt2Elt(*a,i,Rg),ZecRingSubring(Rg,i));
	  ZENFree(ZecElt2Elts(*a,Rg));
	  return(ZERR);
	}
    }
  return(0);
}

/* \TEX{
   \title{Free.}
   \input{a |ZENElt| |a| of a finite ring |Rg| already allocated with
   |_ZecEltCreate|.}
   \side{|a| is freed.}
   }
   */
void _ZecEltFree(a, Rg)
     ZENElt a;
     ZENRing Rg;
{ 
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltFree(ZecElt2Elt(a,i,Rg),ZecRingSubring(Rg,i));
  ZENFree(ZecElt2Elts(a,Rg));
}

