/*
 * zen/zec/c/elt_exp.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element exponentiation}
   
   }
   */
#include <stdio.h>
#include "zec.h"

/* \TEX{
   \title{Exponentiation.}
   \input{A pointer on an element |R| and an element |P| of a finite ring |Rg|, a bignum (|k|, |kl|).}
   \output{0 if no error occurs, |ZENERR| otherwise.}
   \side{$|*R| = |P|^{(|k|, |kl|)}$.}
   \note{One must have $|*(k+kl-1)| \neq 0$. The exponentiation is made in each subring.}
   }
   */
int _ZecEltExp (R, k, kl, P, Rg)
     ZENElt *R, P;
     BigNum k;
     BigNumLength kl;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    if(ZENEltExp(ZecElt2Elt(*R,i,Rg),k,kl,ZecElt2Elt(P,i,Rg),ZecRingSubring(Rg,i)))
      return(ZERR);
  return(0);
}

/* \TEX{
   \title{Adding precomputations.}
   \input{A finite ring |R|.} 
   \output{|ZENERR| if an allocation error occured, 0 otherwise.}
   \side{|_ZecEltExp()| can be very fast if all the subrings are finite fields. Precomputations are NOT recursively done on these subrings.}
   }
   */
int _ZecPrcInitEltExp(R)
     ZENRing R;
{
  fprintf(stderr,"Chinese remainder exponentiation...\n");
  ZENEltExpPtr(R) = _ZecEltExp;
  return(0);
}

/* \TEX[
   \title{Freeing precomputations.}
   \input{A finite ring |R|.} 
   \side{Buffers for |_ZENEltExpBlock| are freed.}
   }
   */
int _ZecPrcCloseEltExp(R)
     ZENRing R;
{  
  ZENEltExpPtr(R) = _ZENEltExpPlain;
  return(ZEN_HAS_INVERSE);
}
