/*
 * zen/zec/c/elt_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element inversion}
   
   |_ZecEltInverse| computes the inverse of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zec.h"

/* \TEX{
   \title{Inverse.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|*p_c| = 1/a$ if |a| has an inverse, ZENRingFact(Rg) is NOT
   filled with a factor of the modulo otherwise.} 
   }
   */
int _ZecEltInverse(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  int i;

  if (ZecEltIsZero(a, Rg))
    {
      ZENSetError(ZEN_LIB_ZEC, ZEN_INVERSE, ZEN_DIVBYZERO);
      return(ZERR);
    }

  for(i=ZecRingNbS(Rg);i--;)
    {
      if(ZENEltIsZero(ZecElt2Elt(a,i,Rg),ZecRingSubring(Rg,i)))
	{
	  BigNum p;
	  BigNumLength pl;
	  
	  switch(ZENRingArith(ZENRingOrigin(Rg)))
	    {
	    case ZEN_LIB_ZEXT:
	      p=ZENPolyToZ(&pl,
			   ZENRingPol(ZecRingSubring(Rg,i)),
			   ZENRingDef(ZecRingSubring(Rg,i)));
	      if(p==NULL)
		return(ZERR);
	      ZENZToElt(ZENRingFact(Rg),p,pl,Rg);	      
	      ZBNF(p);
	      return(ZEN_NO_INVERSE);
	    case ZEN_LIB_ZEP:
	    case ZEN_LIB_ZEPS:
	      ZENZToElt(ZENRingFact(Rg),
			ZENRingP(ZecRingSubring(Rg,i)),
			ZENRingPl(ZecRingSubring(Rg,i)),
			Rg);	      
	      return(ZEN_NO_INVERSE);
	    default:
	      return(ZERR);
	    }
	}
      switch(ZENEltInverse(ZecElt2Elt(*p_c,i,Rg),ZecElt2Elt(a,i,Rg),ZecRingSubring(Rg,i)))
	{
	case ZENERR:
	  return(ZERR);
	case ZEN_HAS_INVERSE:
	  break;
	case ZEN_NO_INVERSE:
	  return(ZEN_NO_INVERSE);
	}
    }
  return(ZEN_HAS_INVERSE);
}
