/*
 * zen/zec/c/elt_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Element multiplication}
   
   Two functions are provided to do multiplications: |_ZecEltSquare| to square a
   |ZENElt| and |_ZecEltMultiply| to compute the multiplication of two |ZENElt|.
   }
   */
#include <stdio.h>
#include "zec.h"

/* \TEX[
   \title{Squaring.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   }
   */
void _ZecEltSquare(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltSquare(ZecElt2Elt(*p_c,i,Rg),ZecElt2Elt(a,i,Rg),ZecRingSubring(Rg,i));
}

/* \TEX[
   \title{Multiplication.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \side{$|*p_c| = a\times b$.}
   }
   */
void _ZecEltMultiply(p_c, a, b, Rg)
     ZENElt *p_c, a, b;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltMultiply(ZecElt2Elt(*p_c,i,Rg),ZecElt2Elt(a,i,Rg),ZecElt2Elt(b,i,Rg),ZecRingSubring(Rg,i));
}



