/*
 * zen/zec/c/elt_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element tests}
   
   Three functions are provided to do tests: |_ZecEltAreEqual| to test if two
   |ZENElt| are equal, |_ZecEltIsZero| to test if a |ZENelt| is equal to
   zero, |_ZecEltIsOne| to test if a |ZENelt| is equal to one.
   }
   */
#include <stdio.h>
#include "zec.h"

/* \TEX[
   \title{Equal.}
   \input{Two |ZENElt| |a| and |b| of a finite ring |Rg|.}
   \output{|a| == |b|.}
   }
   */
int _ZecEltAreEqual(a, b, Rg)
     ZENElt a, b;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    if(!ZENEltAreEqual(ZecElt2Elt(a,i,Rg),ZecElt2Elt(b,i,Rg),ZecRingSubring(Rg,i)))
      return(0);
  return(1);
}

/* \TEX[
   \title{Is zero.}
   \input{A |ZENElt| |a| of a finite ring |Rg|.}
   \output{|a| == 0.}
   }
   */
int _ZecEltIsZero(a, Rg)
     ZENElt a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    if(!ZENEltIsZero(ZecElt2Elt(a,i,Rg),ZecRingSubring(Rg,i)))
      return(0);
  return(1);
}

/* \TEX[
   \title{Is one.}
   \input{ |ZENElt| |a| of a finite ring |Rg|.}
   \output{|a| == 1.}
   }
   */
int _ZecEltIsOne(a, Rg)
     ZENElt a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    if(!ZENEltIsOne(ZecElt2Elt(a,i,Rg),ZecRingSubring(Rg,i)))
      return(0);
  return(1);
}
