/*
 * zen/zec/c/elt_trace.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element trace}
   
   |_ZecEltTrace| computes the trace of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zec.h"

/* \TEX[
   \title{Trace.}
   \input{A |ZENElt| |a| of a finite ring |Rg|.}
   }
   */
void _ZecEltTrace(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltTrace(ZecElt2Elt(*p_c,i,Rg),ZecElt2Elt(a,i,Rg),ZecRingSubring(Rg,i));
}

/* \TEX[
   \title{Find an element of trace one.}
   \input{A finite extension |Rg|.}
   \output{|ZENERR| if an error occured, |ZEN_NO_INVERSE| if an inverse was
   inpossible to compute, |ZEN_HAS_INVERSE| otherwise.} 
   \side{The finite field |Rg| is updated. In particular, an element of
   trace one is stored.}
   }
    */
int _ZecPrcInitPolyRootsCanonical(Rg)
     ZENRing Rg;
{
#if 0
  ZENPrc prc;
  BigNumDigit C[1];
  int i;

  ZENPrcSetNone(prc);
  ZENPrcSet(prc, ZENPRC_POLY_ROOTS_CANONICAL);
  ZENPrcSet(prc, ZENPRC_FINITE_FIELD);
  switch (ZENRingAddPrc(ZecRingOrigin(Rg), prc)) {
  case -1:
    return(ZERR);
  case ZEN_NO_INVERSE:
    return(ZEN_NO_INVERSE);
  case ZEN_HAS_INVERSE:
    if (!ZENPrcIsSet(ZENRingPrcp(ZecRingOrigin(Rg)),
		     ZENPRC_POLY_ROOTS_CANONICAL)) {
      return(ZERR);
    }
    break;
  }
  
  ZENEltToZ(C, &i, ZENRingTheta(ZecRingOrigin(Rg)), ZecRingOrigin(Rg));
  ZecElt2Int(ZENRingTheta(Rg), Rg) = C[0];
#endif
  return(ZEN_HAS_INVERSE);
}


/* \TEX[
   \title{Free the element of trace 1.}
   \input{A finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE|}.
   \side{For compatibility purposes.}
   }
   */
int _ZecPrcClosePolyRootsCanonical(Rg)
     ZENRing Rg;
{
  return(ZEN_HAS_INVERSE);
}
