/*
 * zen/zed/zmake.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "zmake.h"
#include "init.h"

char *C_FILES="doc prc poly_alloc mat_assign mat_create sr_alloc";

char *H_FILES="zed";

int main(argc, argv)
     int argc;
     char *argv[];
{
  /*
   * Misc. variables
   *****************/
  char *ZED_PREFIX;

 /* 
   * Internal targets 
   ********************/
  int objtoar = 0;
  int clean_obj = 0; int clean_lib = 0;
  int clean_doc = 0; int clean_tex = 0; int clean_dvi = 0;
  int tidy_lib = 0; int tidy_obj = 0; int tidy_bin = 0;
  int tidy_doc = 0;
  
  /* 
   * External targets
   *******************/
  int all = Target("all", argc, argv);

  int cc = Target("cc", argc, argv);

  int doc = Target("doc", argc, argv);
  int tex = Target("tex", argc, argv);
  int dvi = Target("dvi", argc, argv);

  int archtidy = Target("archtidy", argc, argv);
  int archclean = Target("archclean", argc, argv);
  int clean = Target("clean", argc, argv);
  int tidy = Target("tidy", argc, argv);

  /* 
   * Initialization
   *****************/
  zen_init(CURDIR, TOPDIR, ZCFDIR);
  zmake_init(ZEN_DEFINES, ZEN_INCLUDES, "", "", argc, argv);
  
  ZED_PREFIX = concat(PREFIX, PZED);

  /* 
   * Targets
   ***********/

  /* default */
  TargetImply(defaults, cc);

  /* all */
  TargetImply(all, cc=doc);

  /* cc */
  TargetImply(cc, objtoar);
  SrcToObjects(cc,
	       ZED_OBJ, ZED_PREFIX,
	       ZED_SRC, "", C_FILES,
	       ZEN_INC, "");

  /* ar */
  ObjectsToArList(objtoar,
		  ZEN_LIB, ZED_PREFIX, "", OBJ_TOAR,
		  ZED_OBJ, ZED_PREFIX, C_FILES);

  /* doc */
  TargetImply(doc, dvi);
  TargetImply(dvi, tex);

  /* tex */
  SrcToTex(tex,
	   ZED_TEX, PZED,
	   ZED_SRC, "", C_FILES, CSUF,
	   ZED_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   ZED_TEX, "",
	   ZED_INC, "", H_FILES, HSUF,
	   ZED_BIN, "", IPT, PARSER);

  /* dvi */
  TexToDvi(dvi,
	   ZED_DVI, PZED,
	   ZED_TEX, PZED, C_FILES,
	   ZEN_INPUTS);
  TexToDvi(dvi,
	   ZED_DVI, "",
	   ZED_TEX, "", H_FILES,
	   ZEN_INPUTS);

  /* Cleaning */
  TargetImply(tidy, tidy_obj=tidy_lib=tidy_bin=tidy_doc);
  TargetImply(archtidy, archclean=clean_doc);
  TargetImply(clean, clean_lib=tidy_obj=clean_tex);
  TargetImply(archclean, clean_obj=clean_lib=clean_tex);
  TargetImply(clean_doc, clean_dvi=clean_tex);
  TargetImply(tidy_obj, clean_obj=clean_lib);
  

  CleaningRules(clean_obj, ZED_OBJ, ZED_PREFIX, C_FILES, OSUF);
  CleaningRules(clean_lib, ZED_LIB, "", ZED_PREFIX, OBJ_TOAR);
  TidyingRules(clean_tex, ZED_TEX);
  CleaningRules(clean_dvi, ZED_DVI, "", C_FILES, ".dvi");
  CleaningRules(clean_tex|clean_dvi, ZED_DOC, "", "", "");

  TidyingRules(tidy_obj, ZED_OBT);
  TidyingRules(tidy_lib, ZED_LBT);
  TidyingRules(tidy_bin, ZED_BNT);
  TidyingRules(tidy_doc, ZED_DOC);

  /* 
   * Closing
   **********/
  zmake_close();

  exit(0);
}
