/*
 * zen/zef/c/elt_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element additions}
   }
   */
#include <stdio.h>
#include <zef.h>

static int __Unsimplify2(a,e,ZEF,R)
     ZENElt a;
     int e,ZEF;
     ZENRing R;
{
  BigNumLength nl;

  if(ZefEltFlag(a,R)&ZEF)
    {
      if(e)
	{
	  nl=ZefEltMl(a,ZEF,R)+1+divSizeBloc(e);
	  if(ZefEltMlm(a,ZEF,R)<nl)
	    {
	      BigNum buf;
	      
	      buf=ZBNC(nl);
	      if(buf==NULL)
		{
		  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
		  return(ZERR);
		}
	      ZBNAssign(buf,ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R));
	      ZefEltMlm(a,ZEF,R)=nl;
	      ZBNF(ZefEltM(a,ZEF,R));
	      ZefEltM(a,ZEF,R)=buf;
	    }
	  ZBNSetToZero(ZefEltM(a,ZEF,R)+ZefEltMl(a,ZEF,R),
		       ZefEltMlm(a,ZEF,R)-ZefEltMl(a,ZEF,R));
	  ZBNAnyShiftLeft(ZefEltM(a,ZEF,R),&(ZefEltMl(a,ZEF,R)),
			  ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R),e);
	  ZefEltMl(a,ZEF,R)=ZBNNumDigits(ZefEltM(a,ZEF,R),
					 ZefEltMl(a,ZEF,R));
	}
    }
  else
    {
      nl=1+divSizeBloc(e);
      if(ZefEltMlm(a,ZEF,R)<nl)
	{
	  if(ZefEltMlm(a,ZEF,R))
	    ZBNF(ZefEltM(a,ZEF,R));
	  ZefEltM(a,ZEF,R)=ZBNC(nl);
	  if(ZefEltM(a,ZEF,R)==NULL)
	    {
	      ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	      return(ZERR);
	    }
	  ZefEltMlm(a,ZEF,R)=nl;
	}
      ZBNSetToZero(ZefEltM(a,ZEF,R),nl-1);
      ZefEltMl(a,ZEF,R)=nl;
      ZefEltM(a,ZEF,R)[nl-1]=BlocExp2(modSizeBloc(e));
      ZefEltFlag(a,R) |= ZEF;
      ZefEltFlag(a,R) |= ZEF_TOEXP;
    }
  if(e)
    {
      ZefEltFlag(a,R) |= ZEF_TOSIMPLIFY;
      ZefEltE(a,Rg) += (ZEF == ZEF_INTEGER ? -e : e);
    }
  else
    ZefEltFlag(a,R) |= ZEF_TOPGCD;
  return(0);
}

int _ZefEltUnsimplify(a,Rg)
     ZENElt a;
     ZENRing Rg;
{
  if(ZefEltE(a,Rg)<0)
    {
      if(__Unsimplify2(a,-ZefEltE(a,Rg),ZEF_INVERSE,Rg))
	return(ZERR);
    }
  else if(ZefEltE(a,Rg)>0)
    {
      if(__Unsimplify2(a,ZefEltE(a,Rg),ZEF_INTEGER,Rg))
	return(ZERR);
    }
  return(0);
}

static int __Unsimplify(a,c,Rg)
     ZENElt a,c;
     ZENRing Rg;
{
  int m,M;

  if(ZefEltE(a,Rg)<ZefEltE(c,Rg))
    {
      m=ZefEltE(a,Rg);
      M=ZefEltE(c,Rg);
    }
  else
    {
      M=ZefEltE(a,Rg);
      m=ZefEltE(c,Rg);
    }

  if(M<0)
    {
      if(ZefEltE(a,Rg)<M)
	{
	  if(__Unsimplify2(a,M-ZefEltE(a,Rg),ZEF_INVERSE,Rg))
	    return(ZERR);
	}
      else if(ZefEltE(c,Rg)<M)
	{
	  if(__Unsimplify2(c,M-ZefEltE(c,Rg),ZEF_INVERSE,Rg))
	    return(ZERR);
	}
    }
  else if(m>0)
    {
      if(ZefEltE(a,Rg)>m)
	{
	  if(__Unsimplify2(a,ZefEltE(a,Rg)-m,ZEF_INTEGER,Rg))
	    return(ZERR);
	}
      else if(ZefEltE(c,Rg)>m)
	{
	  if(__Unsimplify2(c,ZefEltE(c,Rg)-m,ZEF_INTEGER,Rg))
	    return(ZERR);
	}
    }
  else
    {
      if(ZefEltE(c,Rg) < 0)
	{
	  if(__Unsimplify2(a,ZefEltE(a,Rg),ZEF_INTEGER,Rg))
	    return(ZERR);
	  if(__Unsimplify2(c,-ZefEltE(c,Rg),ZEF_INVERSE,Rg))
	    return(ZERR);
	}
      else if(ZefEltE(a,Rg) < 0)
	{
	  if(__Unsimplify2(a,-ZefEltE(a,Rg),ZEF_INVERSE,Rg))
	    return(ZERR);
	  if(__Unsimplify2(c,ZefEltE(c,Rg),ZEF_INTEGER,Rg))
	    return(ZERR);
	}
      else
	{
	  if(__Unsimplify2(a,ZefEltE(a,Rg),ZEF_INTEGER,Rg))
	    return(ZERR);
	  if(__Unsimplify2(c,ZefEltE(c,Rg),ZEF_INTEGER,Rg))
	    return(ZERR);
	}
    }
  return(0);
}
static int __MulBn(c,a,Rg)
     ZENElt c,a;
     ZENRing Rg;
{
  BigNumLength nl;
  BigNum buf;
  
  nl=ZefEltNl(c,Rg)+ZefEltDl(a,Rg);
  if(ZefEltNlm(c,Rg)<nl)
    {
      buf=ZBNC(nl);
      if(buf==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	  return(ZERR);
	}
      ZBNSetToZero(buf,nl);
      if(ZBNMultiply(buf,nl,
		     ZefEltN(c,Rg),ZefEltNl(c,Rg),
		     ZefEltD(a,Rg),ZefEltDl(a,Rg)))
	return(ZERR);
      ZBNF(ZefEltN(c,Rg));
      ZefEltN(c,Rg)=buf;
      ZefEltNl(c,Rg)=ZefEltNlm(c,Rg)=nl;
    }
  else
    {
      buf=ZBNC(ZefEltNl(c,Rg));
      if(buf==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	  return(ZERR);
	}
      ZBNAssign(buf,ZefEltN(c,Rg),ZefEltNl(c,Rg));
      ZBNSetToZero(ZefEltN(c,Rg),nl);
      if(ZBNMultiply(ZefEltN(c,Rg),nl,
		     buf,ZefEltNl(c,Rg),
		     ZefEltD(a,Rg),ZefEltDl(a,Rg)))
	return(ZERR);
      ZefEltNl(c,Rg)=nl;
      ZBNF(buf);
    }
  return(0);
}

static int __AddSub(c,a,op,Rg)
     ZENElt c,a;
     int op;
     ZENRing Rg;
{
  BigNumLength nl;

  if(__Unsimplify(a,c,Rg))
    return(ZERR);

  if((ZefEltType(a,Rg)|ZefEltType(c,Rg))!=ZEF_INTEGER)
    {
      BigNum buf;
      
      if(__Unsimplify2(a,0,ZEF_INTEGER,Rg))
	return(ZERR);
      if(__Unsimplify2(a,0,ZEF_INVERSE,Rg))
	return(ZERR);
      if(__Unsimplify2(c,0,ZEF_INTEGER,Rg))
	return(ZERR);
      if(__Unsimplify2(c,0,ZEF_INVERSE,Rg))
	return(ZERR);
      if(__MulBn(c,a,Rg))
	return(ZERR);
      if(__MulBn(a,c,Rg))
	return(ZERR);
      nl=ZefEltDl(a,Rg)+ZefEltDl(c,Rg);
      buf=ZBNC(nl);
      if(buf==NULL)
	return(ZERR);
      ZBNSetToZero(buf,nl);
      if(ZBNMultiply(buf,nl,
		     ZefEltD(c,Rg),ZefEltDl(c,Rg),
		     ZefEltD(a,Rg),ZefEltDl(a,Rg)))
	return(ZERR);
      ZBNF(ZefEltD(c,Rg));
      ZefEltD(c,Rg)=buf;
      ZefEltDl(c,Rg)=ZefEltDlm(c,Rg)=nl;
    }
  else {
    if(__Unsimplify2(a,0,ZEF_INTEGER,Rg))
      return(ZERR);
    if(__Unsimplify2(c,0,ZEF_INTEGER,Rg))
      return(ZERR);      
  }

  nl=max(ZefEltNl(c,Rg),ZefEltNl(a,Rg))+1;
  if(ZefEltNlm(c,Rg)<nl)
    {
      BigNum buf;
      
      buf=ZBNC(nl);
      if(buf==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	  return(ZERR);
	}
      ZBNAssign(buf,ZefEltN(c,Rg),ZefEltNl(c,Rg));
      ZBNF(ZefEltN(c,Rg));
      ZefEltN(c,Rg)=buf;
      ZefEltNlm(c,Rg)=nl;
    }

  ZBNSetToZero(ZefEltN(c,Rg)+ZefEltNl(c,Rg),ZefEltNlm(c,Rg)-ZefEltNl(c,Rg));
  if(op)
    {
      switch(ZBNCompare(ZefEltN(c,Rg),ZefEltNl(c,Rg),
			ZefEltN(a,Rg),ZefEltNl(a,Rg)))
	{
	case BN_GT:
	  ZBNSubtract(ZefEltN(c,Rg),ZefEltNl(c,Rg),
		      ZefEltN(a,Rg),ZefEltNl(a,Rg),
		      1);
	  ZefEltFlag(c,R) |= ZEF_TOSIMPLIFY;
	  return(0);
	case BN_LT:
	  ZBNSubtract(ZefEltN(a,Rg),ZefEltNl(a,Rg),
		      ZefEltN(c,Rg),ZefEltNl(a,Rg),
		      1);
	  ZBNF(ZefEltN(c,Rg));
	  ZefEltN(c,Rg)=ZefEltN(a,Rg);
	  ZefEltNl(c,Rg)=ZefEltNl(a,Rg);
	  ZefEltNlm(c,Rg)=ZefEltNlm(a,Rg);
	  ZefEltNlm(a,Rg)=0;	/* to prevent unauthorized freeing */
	  ZefEltFlag(c,R) |= ZEF_TOSIMPLIFY;
	  return(1);
	}
      ZefEltN(c,Rg)[0]=0;
      ZefEltNl(c,Rg)=1;
      ZefEltSetType(c,ZEF_INTEGER,Rg);
    }
  else
    {
      if(ZBNAdd(ZefEltN(c,Rg),ZefEltNlm(c,Rg),
		ZefEltN(a,Rg),ZefEltNl(a,Rg),
		(BigNumDigit)0))
	return(ZERR);
      ZefEltNl(c,Rg)=ZBNNumDigits(ZefEltN(c,Rg),ZefEltNlm(c,Rg));
      ZefEltFlag(c,Rg) |= ZEF_TOSIMPLIFY;
    }
  return(0);
}
#define __Subtract(a,c,Rg) __AddSub(a,c,1,Rg)
#define __Add(a,c,Rg) __AddSub(a,c,0,Rg)

/* \TEX[
   \title{Addition.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |Rg|.}
   \side{|*p_c| += |a|.}
   \note{One can have |*p_c| = |a|.}
   }
   */
void _ZefEltAdd(p_c, b, Rg)
     ZENElt *p_c, b;
     ZENRing Rg;
{
  ZENElt a;
#if VERBOSE
  printf("(");
  VERBOSEFCT(*p_c,Rg);
  printf("+(");
  VERBOSEFCT(b,Rg);
  printf(")) - (");
#endif
#if SIMPLIFY			/* Double call is necessary : some adjustments can be performed
				   after the simplification. */
  if(ZefEltSimplify(b,SIMPLIFY_LIM,Rg))
    return;
# if AFTER_PGCD
  if(ZefEltSimplify(b,SIMPLIFY_LIM,Rg))
    return;
# endif
#endif
  
  if(ZefEltCreate(a,Rg))
    return;
  ZefEltAssign(a,b,Rg);

#if SIMPLIFY
  if(ZefEltSimplify(*p_c,SIMPLIFY_LIM,Rg))
    return;
# if AFTER_PGCD
  if(ZefEltSimplify(*p_c,SIMPLIFY_LIM,Rg))
    return;
# endif
#endif
  
  if(ZefEltSign(*p_c,Rg)^ZefEltSign(a,Rg))
    {
      if(__Subtract(*p_c,a,Rg))
	ZefEltNeg(*p_c,Rg);
    }
  else
    __Add(*p_c,a,Rg);
  ZefEltFree(a,Rg);
#if VERBOSE
  VERBOSEFCT(*p_c,Rg);
  printf(");\n");
#endif
}

/* \TEX[
   \title{Negation.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |Rg|.}
   \side{|*p_c| = - |a|.}
   \note{One can have |*p_c| = |a|.}
   }
   */
void _ZefEltNegate(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
#if VERBOSE
  printf("-(");
  VERBOSEFCT(a,Rg);
  printf(") - (");
#endif
  ZefEltAssign(*p_c,a,Rg);
  ZefEltNeg(*p_c,Rg);
#if VERBOSE
  VERBOSEFCT(*p_c,Rg);
  printf(");\n");
#endif
}

/* \TEX[
   \title{Subtraction.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{|*p_c| -= |a|.}
   \note{One can have |*p_c| = |a|.}
   }
   */
void _ZefEltSubtract(p_c, b, Rg)
     ZENElt *p_c, b;
     ZENRing Rg;
{
  ZENElt a;

#if VERBOSE
  printf("(");
  VERBOSEFCT(*p_c,Rg);
  printf("-(");
  VERBOSEFCT(b,Rg);
  printf(")) - (");
#endif

#if SIMPLIFY
  if(ZefEltSimplify(b,SIMPLIFY_LIM,Rg))
    return;
# if AFTER_PGCD
  if(ZefEltSimplify(b,SIMPLIFY_LIM,Rg))
    return;
# endif
#endif
  
  if(ZefEltCreate(a,Rg))
    return;
  ZefEltAssign(a,b,Rg);

  if(ZefEltSimplify(*p_c,SIMPLIFY_LIM,Rg))
    return;
  
  if(ZefEltSign(*p_c,Rg)^ZefEltSign(a,Rg))
    __Add(*p_c, a, Rg);
  else
    {
      if(__Subtract(*p_c, a, Rg))
	ZefEltNeg(*p_c,Rg);
    }
  ZefEltFree(a,Rg);
#if VERBOSE
  VERBOSEFCT(*p_c,Rg);
  printf(");\n");
#endif

}

