/*
 * zen/zef/c/elt_alloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element creation}
   
   |_ZefEltCreate| allocates one |ZENElt| and |_ZefEltFree| frees then it.
   }
   */
#include <stdio.h>
#include "zef.h"

/* \TEX[
   \title{Creation.}
   \input{A finite ring |Rg| and a pointer on a |ZENElt|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \note{The initialized fraction values one.}
   }
   */
int _ZefEltCreate(a, Rg)
     ZENElt *a;
     ZENRing Rg;
{
  ZENElt2Frac(*a,Rg)=(struct zenfrac *)(ZENMalloc(sizeof(struct zenfrac)));
  if(ZENElt2Frac(*a,Rg)==NULL)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
      return(ZERR);
    }
  ZefEltFlag(*a,Rg)=0;
  ZefEltNlm(*a,Rg)=0;
  ZefEltDlm(*a,Rg)=0;
  ZefEltE(*a,Rg)=0;
  return(0);
}

/* \TEX[
   \title{Free.}
   \input{a |ZENElt| |a| of a finite ring |Rg| already allocated with
   |_ZefEltCreate|.}
   \side{|a| is freed.}
   }
   */
void _ZefEltFree(a,Rg)
     ZENElt a;
     ZENRing Rg;
{
  if(ZefEltNlm(a,Rg))
    ZBNF(ZefEltN(a,Rg));
  if(ZefEltDlm(a,Rg))
    ZBNF(ZefEltD(a,Rg));
  ZENFree(ZENElt2Frac(a,Rg));
  return;
}
