/*
 * zen/zef/c/elt_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element assignment}
   }
   */
#include <stdio.h>
#include "zef.h"

static int __Assign(a,b,ZEF,Rg)
     ZENElt a, b;
     int ZEF;
     ZENRing Rg;
{
  if(ZefEltFlag(b,Rg)&ZEF)
    {
      if(ZefEltMlm(a,ZEF,Rg)<ZefEltMl(b,ZEF,Rg))
	{
	  if(ZefEltMlm(a,ZEF,Rg))
	    ZBNF(ZefEltM(a,ZEF,Rg));
	  ZefEltMlm(a,ZEF,Rg)=ZefEltMl(b,ZEF,Rg);
	  ZefEltM(a,ZEF,Rg)=ZBNC(ZefEltMlm(a,ZEF,Rg));
	  if(ZefEltM(a,ZEF,Rg)==NULL)
	    {
	      ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	      return(ZERR);
	    }
	}
      ZBNAssign(ZefEltM(a,ZEF,Rg),ZefEltM(b,ZEF,Rg),ZefEltMl(b,ZEF,Rg));
      ZefEltMl(a,ZEF,Rg)=ZefEltMl(b,ZEF,Rg);
    }
  return(0);
}
/* \TEX[
   \title{Assign.}
   \input{A pointer |pa| on a |ZENElt| and a |ZENElt| |b| of a finite ring |Rg|.}
   \side{|*pa| is filled with |b|.}
   }
   */
void _ZefEltAssign(pa, b, Rg)
     ZENElt *pa, b;
     ZENRing Rg;
{
  if(ZENElt2Frac(*pa,Rg) != ZENElt2Frac(b,Rg))
    {
      ZefEltFlag(*pa,Rg)=ZefEltFlag(b,Rg);
      ZefEltE(*pa,Rg)=ZefEltE(b,Rg);
      if(__Assign(*pa,b,ZEF_INTEGER,Rg))
	return;
      if(__Assign(*pa,b,ZEF_INVERSE,Rg))
	return;
    }
}

/* \TEX[
   \title{Set to zero.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set to zero.}
   }
   */
void _ZefEltSetToZero(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{
  ZefEltE(*pa,Rg)=0;

  if(ZefEltNlm(*pa,Rg)==0)
    {
      if(ZefEltDlm(*pa,Rg)>0)
	{
	  ZefEltN(*pa,Rg)=ZefEltD(*pa,Rg);
	  ZefEltNlm(*pa,Rg)=ZefEltDlm(*pa,Rg);
	  ZefEltDlm(*pa,Rg)=0;
	}
      else
	{
	  ZefEltN(*pa,Rg)=ZBNC(1);
	  if(ZefEltN(*pa,Rg)==NULL)
	    {
	      ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	      return;
	    }
	  ZefEltNlm(*pa,Rg)=1;
	}
    }
	    
  ZefEltN(*pa,Rg)[0]=0;
  ZefEltNl(*pa,Rg)=1;      
  ZefEltFlag(*pa,Rg)=ZEF_INTEGER;
}

/* \TEX[
   \title{Set to one.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set to one.}
   }
   */
void _ZefEltSetToOne(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{
  ZefEltE(*pa,Rg)=0;
  ZefEltFlag(*pa,Rg)=ZEF_ONE;
}

static int __SetRandom(a,ZEF,Rg)
     ZENElt a;
     int ZEF;
     ZENRing Rg;
{
  if(ZefEltMlm(a,ZEF,Rg)==0)
    {
      ZefEltM(a,ZEF,Rg)=ZBNC(1);
      if(ZefEltM(a,ZEF,Rg)==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	  return(ZERR);
	}
      ZefEltMlm(a,ZEF,Rg)=1;
    }
  return(0); 
}
     
/* \TEX{
   \title{Set to random.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set randomly.}
   \note{A random fraction has no sense. This function is intended for testing purposes. Hence, the form of the result is peculiar. Do not use this function if you want true randomness. }
   }
   */
void _ZefEltSetRandom(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{
  if(__SetRandom(*pa,ZEF_INTEGER,Rg))
    return;
  if(__SetRandom(*pa,ZEF_INVERSE,Rg))
    return;

  ZefEltE(*pa,Rg)=zbnrandom()%11-5; /* -5 <= e <= +5 */
  ZBNSetRandom(ZefEltN(*pa,Rg),ZefEltNlm(*pa,Rg));
#if 1
  ZefEltN(*pa,Rg)[0] &= 255;	/* Don't worry ! That's small fractions
				   but they will grow rapidly
				   in extensions ! */
#endif
  ZefEltNl(*pa,Rg)=ZBNNumDigits(ZefEltN(*pa,Rg),ZefEltNlm(*pa,Rg));
  do
    {
      ZBNSetRandom(ZefEltD(*pa,Rg),ZefEltDlm(*pa,Rg));
#if 1
      ZefEltD(*pa,Rg)[0] &= 255;
#endif
      ZefEltDl(*pa,Rg)=ZBNNumDigits(ZefEltD(*pa,Rg),ZefEltDlm(*pa,Rg));
    } while(ZBNIsZero(ZefEltD(*pa,Rg),ZefEltDl(*pa,Rg)));
  
  ZefEltFlag(*pa,Rg)=ZEF_TOSIMPLIFY | ((zbnrandom()%2)*ZEF_NEGATIVE);
  ZefEltSetType(*pa,ZEF_FRAC,Rg);
  ZefEltAdjustType(*pa,Rg);
  return;
}

/* \TEX{
   \title{Enumerate.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is incremented by one.}
   \note{Enumerating the fractions is a Sysyphe's work. This function exists for testing purposes.}
   }
   */
void _ZefEltSetNext(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{
  ZefEltSetToOne(ZENRingHbuf(Rg),Rg);
  ZefEltAdd(*pa,ZENRingHbuf(Rg),Rg);
}

/* \TEX[
   \title{Conversion.}
   \input{A pointer |pa| on a |ZENElt| of a rational ring |Rg|, an element |b| of a ring |R|.}
   \side{|*pa| is set to |b|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   }
   */
int _ZefEltConvert(pa, Rg, b, R)
     ZENElt *pa,b;
     ZENRing Rg,R;
{
  BigNum p;
  BigNumLength pl;
  
  switch(ZENRingArith(R)) {
    
    /* Fractions
     *************/
  case ZEN_LIB_ZEF:
    ZefEltAssign(*pa, b, R);
    return(ZEN_HAS_INVERSE);
    
    /* Specific cases where things can be done faster
     *************************************************/
  case ZEN_LIB_ZE2:
    if(ZENEltIsZero(b, R))
      ZefEltSetToZero(*pa, R);
    else
      ZefEltSetToOne(*pa, R);
    return(ZEN_HAS_INVERSE);
    
    /* Default
     **********/
  default:
    
    /* R is a prime finite field ? */
    if (!ZENRingDeg(R)) {
      pl = ZENRingQl(R);
      if (!(p = ZBNC(pl))) {
	ZENSetError(ZEN_LIB_ZEF, ZEN_CONVERT, ZEN_ALLOC);
	return(ZERR);
      }
      ZENEltToZ(p, &pl, b, R);
      ZefZToElt(*pa, p, pl, Rg);
      ZBNF(p);
      break;
    }

    /* Otherwise, what can we do ? */
    ZENSetError(ZEN_LIB_ZEF, ZEN_CONVERT, ZEN_INPUT);
    return(ZERR);
  }
    
  return(ZEN_HAS_INVERSE);
}






