/*
 * zen/zef/c/elt_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{   
   \Subsection{Element inversion}
   
   |_ZefEltInverse| computes the inverse of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zef.h"

/* \TEX[
   \title{Inverse.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |R|.}
   \output{|ZENERR| if an error occured, |ZEN_HAS_INVERSE| otherwise}
   \side{$|*p_c| = 1/a$ if |a| is non zero.} 
   }
   */
int _ZefEltInverse(p_c, a, R)
     ZENElt *p_c, a;
     ZENRing R;
{
  BigNum tmp;
  int tmpl;
  
#if VERBOSE
  printf("(");
  VERBOSEFCT(a,R);
  printf(")^(-1) - (");
#endif
  if(ZefEltIsZero(a,R))
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_INVERSE,ZEN_DIVBYZERO);
      return(ZERR);
    }

  ZefEltAssign(*p_c,a,R);
  ZefEltE(*p_c,R)= -ZefEltE(*p_c,R);
  ZefEltFlag(*p_c,R)=ZefEltFlag(*p_c,R);
  if(ZefEltType(*p_c,R))
    {
      tmp=ZefEltN(*p_c,R);
      ZefEltN(*p_c,R)=ZefEltD(*p_c,R);
      ZefEltD(*p_c,R)=tmp;
      tmpl=ZefEltNl(*p_c,R);
      ZefEltNl(*p_c,R)=ZefEltDl(*p_c,R);
      ZefEltDl(*p_c,R)=tmpl;
      tmpl=ZefEltNlm(*p_c,R);
      ZefEltNlm(*p_c,R)=ZefEltDlm(*p_c,R);
      ZefEltDlm(*p_c,R)=tmpl;
      
      if(ZefEltType(*p_c,R)<ZEF_FRAC)
	ZefEltFlag(*p_c,R) ^= ZEF_FRAC;
    }
#if VERBOSE
  VERBOSEFCT(*p_c,R);
  printf(");\n");
#endif
  return(ZEN_HAS_INVERSE);
}


