/*
 * zen/zef/c/elt_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element multiplication}
   }
   */
#include <stdio.h>
#include <zef.h>

static int __Square(c, a, ZEF, R)
     ZENElt c, a;
     int ZEF;
     ZENRing R;
{
  if(ZefEltFlag(a,R)&ZEF)
    {
      BigNum buf;
      BigNumLength bufl;
      
      bufl=ZefEltMl(a,ZEF,R);
      buf=ZBNC(bufl);
      if(buf==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	  return(ZERR);
	}
      ZBNAssign(buf,ZefEltM(a,ZEF,R),bufl);
      if(ZefEltMlm(c,ZEF,R)<2*bufl)
	{
	  if(ZefEltMlm(c,ZEF,R))
	    ZBNF(ZefEltM(c,ZEF,R));
	  ZefEltMlm(c,ZEF,R)=2*bufl;
	  ZefEltM(c,ZEF,R)=ZBNC(ZefEltMlm(c,ZEF,R));
	  if(ZefEltM(c,ZEF,R)==NULL)
	    {
	      ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	      return(ZERR);
	    }
	}

      ZBNSetToZero(ZefEltM(c,ZEF,R),ZefEltMlm(c,ZEF,R));
      if(ZBNSquare(ZefEltM(c,ZEF,R),ZefEltMlm(c,ZEF,R),buf,bufl))
	return(ZERR);
      ZefEltMl(c,ZEF,R)=ZBNNumDigits(ZefEltM(c,ZEF,R),ZefEltMlm(c,ZEF,R));
      ZBNF(buf);
    }
  return(0);
}
/* \TEX[
   \title{Squaring.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   \note{One can have |*p_c| = |a|.}
   }
   */
void _ZefEltSquare(p_c, a, R)
     ZENElt *p_c, a;
     ZENRing R;
{
#if VERBOSE
  printf("(");
  VERBOSEFCT(a,R);
  printf(")^2 - (");
#endif
  
  ZefEltE(*p_c,R) = ZefEltE(a,R)*2;
  ZefEltFlag(*p_c,R) = ZefEltFlag(a,R) & (~ZEF_NEGATIVE);
  if(ZefEltType(a,R))    
    if(!__Square(*p_c,a,ZEF_INTEGER,R))
      __Square(*p_c,a,ZEF_INVERSE,R);
  
#if VERBOSE
  VERBOSEFCT(*p_c,R);
  printf(");\n");
#endif
}

static int __Multiply(c, a, b, flag, ZEF, R)
     ZENElt c, a, b;
     int flag,ZEF;
     ZENRing R;
{
  if(flag & ZEF)
    {
      if(!(ZefEltType(b,R) & ZEF))
	{
	  if(ZefEltMlm(c,ZEF,R)<ZefEltMl(a,ZEF,R))
	    {
	      if(ZefEltMlm(c,ZEF,R))
		ZBNF(ZefEltM(c,ZEF,R));
	      ZefEltMlm(c,ZEF,R)=ZefEltMl(a,ZEF,R);
	      ZefEltM(c,ZEF,R)=ZBNC(ZefEltMlm(c,ZEF,R));
	      if(ZefEltM(c,ZEF,R)==NULL)
		{
		  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
		  return(ZERR);
		}
	    }
	  ZBNAssign(ZefEltM(c,ZEF,R),ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R));
	  ZefEltMl(c,ZEF,R)=ZefEltMl(a,ZEF,R);
	}
      else if(!(ZefEltType(a,R) & ZEF))
	{
	  if(ZefEltMlm(c,ZEF,R)<ZefEltMl(b,ZEF,R))
	    {
	      if(ZefEltMlm(c,ZEF,R))
		ZBNF(ZefEltM(c,ZEF,R));
	      ZefEltMlm(c,ZEF,R)=ZefEltMl(b,ZEF,R);
	      ZefEltM(c,ZEF,R)=ZBNC(ZefEltMlm(c,ZEF,R));
	      if(ZefEltM(c,ZEF,R)==NULL)
		{
		  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
		  return(ZERR);
		}
	    }
	  ZBNAssign(ZefEltM(c,ZEF,R),ZefEltM(b,ZEF,R),ZefEltMl(b,ZEF,R));
	  ZefEltMl(c,ZEF,R)=ZefEltMl(b,ZEF,R);
	}
      else
	{
	  BigNum buf;
	  int bufl;
	  
	  bufl=ZefEltMl(b,ZEF,R);
	  buf=ZBNC(bufl);
	  if(buf==NULL)
	    {
	      ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	      return(ZERR);
	    }
	  ZBNAssign(buf,ZefEltM(b,ZEF,R),bufl);
	  if(ZefEltMlm(c,ZEF,R)<ZefEltMl(a,ZEF,R)+ZefEltMl(b,ZEF,R))
	    {
	      if(ZefEltMlm(c,ZEF,R))
		ZBNF(ZefEltM(c,ZEF,R));
	      ZefEltMlm(c,ZEF,R)=ZefEltMl(a,ZEF,R)+ZefEltMl(b,ZEF,R);
	      ZefEltM(c,ZEF,R)=ZBNC(ZefEltMlm(c,ZEF,R));
	      if(ZefEltM(c,ZEF,R)==NULL)
		{
		  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
		  return(ZERR);
		}
	    }
	  ZBNSetToZero(ZefEltM(c,ZEF,R),ZefEltMlm(c,ZEF,R));
	  if(ZBNMultiply(ZefEltM(c,ZEF,R),ZefEltMlm(c,ZEF,R),
			     ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R),
			     buf,bufl))
	    return(ZERR);
	  ZBNF(buf);
	  ZefEltMl(c,ZEF,R)=ZBNNumDigits(ZefEltM(c,ZEF,R),
					    ZefEltMlm(c,ZEF,R));
	}
    }
  return(0);
}
/* \TEX[
   \title{Multiplication.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |R|.}
   \side{$|*p_c| = a\times b$.}
   \note{One can have |*p_c| = |b|.}
   }
   */
void _ZefEltMultiply(p_c, a, b, R)
     ZENElt *p_c, a, b;
     ZENRing R;
{
  int flag;

#if VERBOSE
  VERBOSEFCT(a,R);
  printf("*(");
  VERBOSEFCT(b,R);
  printf(") - (");
#endif

  ZefEltE(*p_c,R) = ZefEltE(a,R)+ZefEltE(b,R);
  flag=ZefEltSign(a,R)^ZefEltSign(b,R);

  if(!ZefEltType(a,R))
    flag |= ZefEltSimplifyFlag(b,R);
  else
    flag |= ZEF_TOPGCD | (ZEF_TOEXP & (ZefEltFlag(a,R) | ZefEltFlag(b,R)));

  flag |= ZefEltType(a,R) | ZefEltType(b,R);

  if(!(flag & ZEF_FRAC))
    ZefEltFlag(*p_c,R)=flag;
  else if(!__Multiply(*p_c, a, b, flag, ZEF_INTEGER, R))
    if(!__Multiply(*p_c, a, b, flag, ZEF_INVERSE, R))
      ZefEltFlag(*p_c,R)=flag;
#if VERBOSE
  VERBOSEFCT(*p_c,R);
  printf(");\n");
#endif

}










