/*
 * zen/zef/c/elt_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element tests}
   Note that tests can have false answers if approximation is set.
   }
   */
#include <stdio.h>
#include "zef.h"

#if APPROXIMATE_RESULTS
static void __Approximate(a,ZEF,R)
     ZENElt a;
     int ZEF;
     ZENRing R;
{
  if(ZefEltFlag(a,R)&ZEF)
    {
      ZefEltMl(a,ZEF,R)=ZBNNumDigits(ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R));
      if(ZefEltMl(a,ZEF,R) > APPROXIMATE_LEVEL)
	{
	  int nl;
	  nl=(ZefEltMl(a,ZEF,R)-APPROXIMATE_LEVEL)*SIZE_BLOC
	    -ZBNNumLeadingZeroBitsInDigit(ZefEltM(a,ZEF,R)[ZefEltMl(a,ZEF,R)-1]);
	  if((ZefEltM(a,ZEF,R)[divSizeBloc(nl-1)]>>modSizeBloc(nl-1))&1)
	    {			/* we approximate the least significant
				   digit */
	      ZBNAnyShiftRight(ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R),nl);
	      ZefEltMl(a,ZEF,R)=APPROXIMATE_LEVEL;
	      if(ZBNAddCarry(ZefEltM(a,ZEF,R),APPROXIMATE_LEVEL,1))
		{		/* This is the very special case of 111...111
				   in which case the approx is     1000...000
				   */
		  ZefEltMl(a,ZEF,R)=ZefEltM(a,ZEF,R)[0]=1;
		  nl+=APPROXIMATE_LEVEL*SIZE_BLOC;
		}
	      else
		ZefEltFlag(a,R) |= ZEF_TOSIMPLIFY;
	    }
	  else
	    {
	      ZBNAnyShiftRight(ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R),nl);
	      ZefEltMl(a,ZEF,R)=APPROXIMATE_LEVEL;
	      ZefEltFlag(a,R) |= ZEF_TOSIMPLIFY;
	    }
	  ZefEltE(a,R) += (ZEF == ZEF_INTEGER ? nl : -nl);
	}
    }
}
#endif
static int __Simplify(a,ZEF,R)
     ZENElt a;
     int ZEF;
     ZENRing R;
{
  BigNumLength nl;
  if(ZefEltType(a,R) & ZEF)
    {
      ZefEltMl(a,ZEF,R)=ZBNNumDigits(ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R));
      nl=0;
      while((nl<ZefEltMl(a,ZEF,R))&&(ZefEltM(a,ZEF,R)[nl]==0))
	nl++;
      if(nl!=ZefEltMl(a,ZEF,R))
	{
	  BigNumDigit Enb;

	  if(nl>0)
	    {
	      ZefEltE(a,R) += (ZEF==ZEF_INTEGER ? nl*SIZE_BLOC : -nl*SIZE_BLOC);
	      ZefEltMl(a,ZEF,R) -= nl;
	      ZBNAssign(ZefEltM(a,ZEF,R),ZefEltM(a,ZEF,R)+nl,ZefEltMl(a,ZEF,R));
	    }
	  nl=0;
	  Enb=1;
	  while((nl<SIZE_BLOC)
		&&((ZefEltM(a,ZEF,R)[0] & Enb)==0))
	    {
	      nl++;
	      Enb <<= 1;
	    }
	  if(nl)
	    {
	      ZefEltE(a,R) += (ZEF==ZEF_INTEGER ? nl : -nl);
	      ZBNShiftRight(ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R),nl);
	      ZefEltMl(a,ZEF,R)=ZBNNumDigits(ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R));
	    }
	  if(ZBNIsOne(ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R)))
	    ZefEltFlag(a,R) ^= ZEF;
	}
      return(ZefEltMl(a,ZEF,R));
    }
  return(0);
}

static int __Div(a,ZEF,g,gl,R)
     ZENElt a;
     int ZEF;
     BigNum g;
     BigNumLength gl;
     ZENRing R;
{
  BigNum div;
  int dl;

  dl=ZefEltMl(a,ZEF,R)+1;
  div=ZBNC(dl);
  if(div==NULL)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
      return(ZERR);
    }
  ZBNAssign(div,ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R));
  div[dl-1]=0;
  ZBNDivide(div,dl,g,gl);
  dl=ZBNNumDigits(div+gl,dl-gl);
  ZBNAssign(ZefEltM(a,ZEF,R),div+gl,dl);
  ZefEltMl(a,ZEF,R)=dl;
  ZBNF(div);
  return(0);      
}
/*
   \TEX{
   \title{Simplification}
   \input{A |ZENElt| representing a fraction, the minimal size for simplification and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The fraction is simplified.}
   } 
   */
int _ZefEltSimplify(a,lim,R)
     ZENElt a;
     int lim;
     ZENRing R;
{
  if(!ZefEltSimplifyFlag(a,R))
    return(0);

  if(!ZefEltType(a,R))
    {
      ZefEltFlag(a,R) &= ~(ZEF_TOSIMPLIFY);
      return(0);
    }

  if((ZefEltFlag(a,R) & ZEF_INTEGER)
     &&(ZBNIsZero(ZefEltN(a,R),ZefEltNl(a,R))))
    {
      ZefEltFlag(a,R) &= ~(ZEF_TOSIMPLIFY | ZEF_NEGATIVE | ZEF_INVERSE);
      ZefEltNl(a,R)=1;
      ZefEltE(a,R) = 0;
      return(0);
    }

#if APPROXIMATE_RESULTS
# if AFTER_PGCD
  if((ZefEltType(a,R) != ZEF_FRAC) ||
     (!(ZefEltFlag(a,R) & ZEF_TOPGCD)) )
# endif
    {
      __Approximate(a,ZEF_INTEGER,R);
      __Approximate(a,ZEF_INVERSE,R);
    }
#endif

  if(ZefEltFlag(a,R) & ZEF_TOEXP)
    {
      lim -= __Simplify(a,ZEF_INTEGER,R);
      lim -= __Simplify(a,ZEF_INVERSE,R);
      ZefEltFlag(a,R) ^= ZEF_TOEXP;
    }
  else
    {
      if(ZefEltType(a,R) & ZEF_INTEGER)
	lim -= ZefEltNl(a,R);
      if(ZefEltType(a,R) & ZEF_INVERSE)
	lim -= ZefEltDl(a,R);
    }      

  if(lim>=0)
    return(0);
  if(!(ZefEltFlag(a,R)&ZEF_TOPGCD))    
    return(0);
    
				/* GCD computation */
  if(ZefEltType(a,R)==ZEF_FRAC)
    {
      if(ZBNAreEqual(ZefEltN(a,R),ZefEltNl(a,R),ZefEltD(a,R),ZefEltDl(a,R)))
	ZefEltFlag(a,R) &= ~(ZEF_TOPGCD | ZEF_FRAC);  	
      else
	{
	  BigNum g;
	  BigNumLength gl;

	  gl=max(ZefEltNl(a,R),ZefEltDl(a,R));
	  g=ZBNC(gl);
	  if(g==NULL)
	    {
	      ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	      return(ZERR);
	    }
	  if (ZBNGcd(g, &gl, ZefEltN(a,R),ZefEltNl(a,R),ZefEltD(a,R),ZefEltDl(a,R))) {
	    return(ZERR);
	  }
	  if((gl!=0) || (g[0]!=1))
	    {
	      if(__Div(a,ZEF_INTEGER,g,gl,R))
		return(ZERR);
	      if(__Div(a,ZEF_INVERSE,g,gl,R))
		return(ZERR);
	    }
	  ZBNF(g);
      
	  ZefEltAdjustType(a,R);
	  ZefEltFlag(a,R) ^= ZEF_TOPGCD;  

#if AFTER_PGCD
	  __Approximate(a,ZEF_INTEGER,R);
	  __Approximate(a,ZEF_INVERSE,R);  
#endif
	}
    }
  else
    ZefEltFlag(a,R) ^= ZEF_TOPGCD;  
  return(0);
}

static int __Equal(a,b,ZEF,R)
     ZENElt a,b;
     int ZEF;
     ZENRing R;
{
  if(ZefEltType(a,R) & ZEF)
    {
      if(ZefEltType(b,R) & ZEF)
	return(ZBNAreEqual(ZefEltM(a,ZEF,R),ZefEltMl(a,ZEF,R),
			   ZefEltM(b,ZEF,R),ZefEltMl(b,ZEF,R)));
      return(0);
    }
  else
    {
      if(ZefEltType(b,R) & ZEF)
	return(0);
      return(1);
    }
}

/* \TEX[
   \title{Equal.}
   \input{Two |ZENElt| |a| and |b| of a finite ring |R|.}
   \output{|a == b| or |ZENERR| if an error occured.}
   }
   */
int _ZefEltAreEqual(a,b,R)
     ZENElt a, b;
     ZENRing R;
{
  if(ZefEltSimplify(a,0,R))
    return(ZERR);
  if(ZefEltSimplify(b,0,R))
    return(ZERR);

  if(ZefEltE(a,R)!=ZefEltE(b,R))
    return(0);
  if(ZefEltSign(a,R)!=ZefEltSign(b,R))
    return(0);
  if(!__Equal(a,b,ZEF_INTEGER,R))
    return(0);
  if(!__Equal(a,b,ZEF_INVERSE,R))
    return(0);
  return(1);
}

/* \TEX[
   \title{Is zero.}
   \input{A |ZENElt| |a| of a finite ring |R|.}
   \output{|a == 0| or |ZENERR| if an error occured.}
   }
   */
int _ZefEltIsZero(a, R)
     ZENElt a;
     ZENRing R;
{
  if(ZefEltType(a,R) & ZEF_INTEGER)
    {
      ZefEltNl(a,R)=ZBNNumDigits(ZefEltN(a,R),ZefEltNl(a,R));
      if(ZefEltNl(a,R)!=1)
	return(0);

      if(ZefEltN(a,R)[0])
	return(0);
      ZefEltSetType(a,ZEF_INTEGER,R);
      ZefEltE(a,R)=0;
      return(1);
    }
  return(0);
}

/* \TEX[
   \title{Is one.}
   \input{ |ZENElt| |a| of a finite ring |R|.}
   \output{|a| == 1.}
   }
   */
int _ZefEltIsOne(a, R)
     ZENElt a;
     ZENRing R;
{
  if(ZefEltSimplify(a,(int)(ZefEltNlm(a,R)+ZefEltDlm(a,R)),R))	/* only the exponent */
    return(ZERR);

  if(ZefEltE(a,R))
    return(0);

  if(ZefEltSign(a,R))
    return(0);

  switch(ZefEltType(a,R))
    {
    case ZEF_ONE:
      return(1);
    case ZEF_INTEGER:
      if(ZBNIsOne(ZefEltN(a,R),ZefEltNl(a,R)))
	{
	  ZefEltSetType(a,ZEF_ONE,R);
	  return(1);
	}
      return(0);
    case ZEF_INVERSE:
      if(ZBNIsOne(ZefEltD(a,R),ZefEltDl(a,R)))
	{
	  ZefEltSetType(a,ZEF_ONE,R);
	  return(1);
	}
      return(0);
    case ZEF_FRAC:
      if(ZBNAreEqual(ZefEltN(a,R),ZefEltNl(a,R),ZefEltD(a,R),ZefEltDl(a,R)))
	{
	  ZefEltSetType(a,ZEF_ONE,R);
	  return(1);
	}
      return(0);
    }
  return(ZERR);
}
