/*
 * zen/zef/c/elt_trace.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element trace}
   This has no sense, except for compatibility
   }
   */
#include <stdio.h>
#include "zef.h"

/* \TEX[
   \title{Trace.}
   \input{A |ZENElt| |a| of a finite ring |Rg|.}
   }
   */
void _ZefEltTrace(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZefEltTrace(*p_c, a, Rg);
}


/* \TEX[
   \title{Computing an element of trace 1.}
   \input{A finite ring |Rg|.} 
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| otherwise.}
   \side{|ZENRingTheta(Rg)| is allocated and filled with 1.}
   }
   */
int _ZefPrcInitPolyRootsCanonical(Rg)
     ZENRing Rg;
{
  if (ZefEltCreate(ZENRingTheta(Rg), Rg)) {
    ZENSetError(ZEN_LIB_ZEF, ZEN_ADDPRC, ZEN_ALLOC);
    return(ZERR);
  }
  ZefEltSetToOne(ZENRingTheta(Rg), Rg);

  return(ZEN_HAS_INVERSE);
}

/* \TEX[
   \title{Desallocate the element of trace 1.}
   \input{A finite ring |Rg|.} 
   \output{|ZEN_HAS_INVERSE|.}
   \side{|ZENRingTheta(Rg)| is desallocated.}
   }
   */
int _ZefPrcClosePolyRootsCanonical(Rg)
     ZENRing Rg;
{
  ZefEltFree(ZENRingTheta(Rg), Rg);
  
  return(ZEN_HAS_INVERSE);
}
