/*
 * zen/zef/c/init.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   }
   */
#include <stdio.h>
#include "zef.h"
#include "zeg.h"
#include "zed.h"


/* \TEX{
   \title{Initialization.}
   \input{A partially defined |ZENRing|.}
   \output{|ZENERR| if an error occured, $|SR_GEN|\vert |MAT_GEN|$ otherwise.}
   \side{The initialization of fraction field is completed.}
   \note{This procedure is called by |ZENBaseRingCreate| and should not be called otherwise.}
   }
   */
int _ZefInit(R)
     ZENRing R;
{
  ZENRingArith(R) = ZEN_LIB_ZEF;

  /* Procedure on rings
   **********************/
  ZENPrcInitEltMultiplyPtr(R) = NULL;
  ZENPrcCloseEltMultiplyPtr(R) = NULL;

  ZENPrcInitPolyRootsCanonicalPtr(R) = NULL;
  ZENPrcClosePolyRootsCanonicalPtr(R) = NULL;

  ZENPrcInitEltExpPtr(R) = _ZENPrcInitEltExp;
  ZENPrcCloseEltExpPtr(R) = _ZENPrcCloseEltExp;

  /* Procedure on elements
   ****************************/
  /* Allocation */
  ZENEltCreatePtr(R) = _ZefEltCreate;	/* Create elements */
  ZENEltCreateLargePtr(R) = _ZefEltCreate;	/* Create elements */
  ZENEltFreePtr(R) = _ZefEltFree;	/* Free elements */

  /* Assigning */
  ZENEltAssignPtr(R) = _ZefEltAssign;	/* Assign an element with another */
  ZENEltSetToZeroPtr(R) = _ZefEltSetToZero; /* Set an element to zero */
  ZENEltSetToOnePtr(R) = _ZefEltSetToOne; /* Set an element to one */
  ZENEltSetRandomPtr(R) = _ZefEltSetRandom; /* Set randomly an element */
  ZENEltSetNextPtr(R) = _ZefEltSetNext; /* Set an element to the next */

  /* Test */
  ZENEltAreEqualPtr(R) = _ZefEltAreEqual; /* Is an element equal to another */
  ZENEltIsZeroPtr(R) = _ZefEltIsZero;	/* Is an element equal to zero */
  ZENEltIsOnePtr(R) = _ZefEltIsOne;	/* Is an element equal to one */
  ZENEltIsASquarePtr(R) = _ZENEltIsASquareCharp; /* Is an element a square ? */

  /* Arithmetic */
  ZENEltAddPtr(R) = _ZefEltAdd;	/* Addition of elements */
  ZENEltNegatePtr(R) = _ZefEltNegate;	/* Negation of an element */
  ZENEltSubtractPtr(R) = _ZefEltSubtract; /* Substraction of elements */
  ZENEltSquarePtr(R) = _ZefEltSquare;	/* Squaring of elements */
  ZENEltMultiplyPtr(R) = _ZefEltMultiply; /* Multiplication of elements */
  ZENEltInversePtr(R) = _ZefEltInverse; /* Inverse of elements */
  ZENEltTracePtr(R) = _ZefEltTrace;		/* Trace of elements */
  ZENEltSquareRootPtr(R) = _ZENEltSquareRootCharp; /* Square root of an element */
  ZENEltExpPtr(R) = _ZENEltExpPlain;	/* Exponentiation */

  /* io */
  ZENEltReadFromStringPtr(R) = _ZefEltReadFromString;	/* Read an element from a string */
  ZENEltPrintToStringPtr(R) = _ZefEltPrintToString; /* Print an element to a string */
  ZENEltReadFromFilePtr(R) = _ZefEltReadFromFile; /* Read an element from a file */
  ZENEltPrintToFilePtr(R) = _ZefEltPrintToFile; /* Print an element to a file */
  ZENEltGetFromStringPtr(R) = _ZefEltGetFromString; /* Get an element from a string */
  ZENEltPutToStringPtr(R) = _ZefEltPutToString; /* Put an element to a string */
  ZENEltGetFromFilePtr(R) = _ZefEltGetFromFile; /* Get an element from a file */
  ZENEltPutToFilePtr(R) = _ZefEltPutToFile; /* Put an element to a file */
  
  ZENZToEltPtr(R) = _ZefZToElt; /* Evaluate in $p$ an element */
  ZENEltToZPtr(R) = _ZefEltToZ; /* Get the element from an evaluation in $p$ */
  ZENEltConvertPtr(R) = _ZefEltConvert;
  
  ZENMatCreateParamsPtr(R) = _ZedMatCreateParams;	
  ZENMatCreatePtr(R) = __ZedMatCreate;

  return(POLY_GEN | SR_GEN | MAT_GEN);
}

/* \TEX[
   \title{Closing.}
   \input{A partially freed |ZENRing|.}
   \output{0.}
   }
   */
int _ZefClose(R)		
     ZENRing R;
{
  return(0);
}
