/*
 * zen/zeg/c/ec_alloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Elliptic curves}
   }
   */
#include <stdio.h>
#include "zeg.h"

/* \TEX{
   \title{Allocation of an elliptic curve.}
   \input{A finite ring |Rg|.}
   \output{The elliptic curve initialized or NULL if an error occured.}
   }
   */
ZENEc _ZENEcCreate(Rg)
     ZENRing Rg;
{
  ZENEc E;

  if (!(E = ZENEcAllocStruct(Rg))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECCREATE, ZEN_ALLOC);
    return(NULL);
  }

  /*The sub extension */
  ZENEcRing(E) = Rg;

  /* The invariants and the buffers */
  if (ZENEltCreate(ZENEcD(E), Rg)) {
    goto error;
  }
  if (ZENEltCreate(ZENEcJ(E), Rg)) {
    ZENEltFree(ZENEcD(E), Rg);
    goto error;
  }
  if (!(ZENEcXY(E)= ZENEcPtAllocStruct(E))) {
    ZENEltFree(ZENEcD(E), Rg);
    ZENEltFree(ZENEcJ(E), Rg);
    goto error;
  }
  if (ZENEltCreate(ZENEcX(E), Rg)) {
    ZENEltFree(ZENEcD(E), Rg);
    ZENEltFree(ZENEcJ(E), Rg);
    ZENFree(ZENEcXY(E));
    goto error;
  }
  if (ZENEltCreate(ZENEcY(E), Rg)) {
    ZENEltFree(ZENEcD(E), Rg);
    ZENEltFree(ZENEcJ(E), Rg);
    ZENEltFree(ZENEcX(E), Rg);
    ZENFree(ZENEcXY(E));
    goto error;
  }
  if (ZENEltCreate(ZENEcZ(E), Rg)) {
    ZENEltFree(ZENEcD(E), Rg);
    ZENEltFree(ZENEcJ(E), Rg);
    ZENEltFree(ZENEcX(E), Rg);
    ZENEltFree(ZENEcY(E), Rg);
    ZENFree(ZENEcXY(E));
    goto error;
  }
  if (ZENEltCreate(ZENEcDb(E), Rg)) {
    ZENEltFree(ZENEcZ(E), Rg);
    ZENEltFree(ZENEcD(E), Rg);
    ZENEltFree(ZENEcJ(E), Rg);
    ZENEltFree(ZENEcX(E), Rg);
    ZENEltFree(ZENEcY(E), Rg);
    ZENFree(ZENEcXY(E));
    goto error;
  }

  /* The parameters of the curve */
  if (ZENEltCreate(ZENEcA1(E), Rg)) {
    goto error0;
  }
  if (ZENEltCreate(ZENEcA2(E), Rg)) {
    ZENEltFree(ZENEcA1(E), Rg);
    goto error0;
  }
  if (ZENEltCreate(ZENEcA3(E), Rg)) {
    ZENEltFree(ZENEcA1(E), Rg);
    ZENEltFree(ZENEcA2(E), Rg);
    goto error0;
  }
  if (ZENEltCreate(ZENEcA4(E), Rg)) {
    ZENEltFree(ZENEcA1(E), Rg);
    ZENEltFree(ZENEcA2(E), Rg);
    ZENEltFree(ZENEcA3(E), Rg);
    goto error0;
  }
  if (ZENEltCreate(ZENEcA6(E), Rg)) {
    ZENEltFree(ZENEcA1(E), Rg);
    ZENEltFree(ZENEcA2(E), Rg);
    ZENEltFree(ZENEcA3(E), Rg);
    ZENEltFree(ZENEcA4(E), Rg);
    goto error0;
  }

  /* The canonical curve */
  if (ZENEltCreate(ZENEcB1(E), ZENEcRing(E))) {
    goto error1;
  }
  if (ZENEltCreate(ZENEcB2(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcB1(E), ZENEcRing(E));
    goto error1;
  }
  if (ZENEltCreate(ZENEcB3(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcB1(E), ZENEcRing(E));
    ZENEltFree(ZENEcB2(E), ZENEcRing(E));
    goto error1;
  }
  if (ZENEltCreate(ZENEcB4(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcB1(E), ZENEcRing(E));
    ZENEltFree(ZENEcB2(E), ZENEcRing(E));
    ZENEltFree(ZENEcB3(E), ZENEcRing(E));
    goto error1;
  }
  if (ZENEltCreate(ZENEcB6(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcB1(E), ZENEcRing(E));
    ZENEltFree(ZENEcB2(E), ZENEcRing(E));
    ZENEltFree(ZENEcB3(E), ZENEcRing(E));
    ZENEltFree(ZENEcB4(E), ZENEcRing(E));
    goto error1;
  }

  /* The isomorphism */
  if (ZENEltCreate(ZENEcAd(E), ZENEcRing(E))) {
    goto error2;
  }
  if (ZENEltCreate(ZENEcBd(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcAd(E), ZENEcRing(E));
    goto error2;
  }
  if (ZENEltCreate(ZENEcCd(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcBd(E), ZENEcRing(E));
    ZENEltFree(ZENEcAd(E), ZENEcRing(E));
    goto error2;
  }
  if (ZENEltCreate(ZENEcDd(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcCd(E), ZENEcRing(E));
    ZENEltFree(ZENEcBd(E), ZENEcRing(E));
    ZENEltFree(ZENEcAd(E), ZENEcRing(E));
    goto error2;
  }
  if (ZENEltCreate(ZENEcEd(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcDd(E), ZENEcRing(E));
    ZENEltFree(ZENEcCd(E), ZENEcRing(E));
    ZENEltFree(ZENEcBd(E), ZENEcRing(E));
    ZENEltFree(ZENEcAd(E), ZENEcRing(E));
    goto error2;
  }
  
  /* The inverse isomorphism */
  if (ZENEltCreate(ZENEcAi(E), ZENEcRing(E))) {
    goto error3;
  }
  if (ZENEltCreate(ZENEcBi(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcAi(E), ZENEcRing(E));
    goto error3;
  }
  if (ZENEltCreate(ZENEcCi(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcBi(E), ZENEcRing(E));
    ZENEltFree(ZENEcAi(E), ZENEcRing(E));
    goto error3;
  }
  if (ZENEltCreate(ZENEcDi(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcCi(E), ZENEcRing(E));
    ZENEltFree(ZENEcBi(E), ZENEcRing(E));
    ZENEltFree(ZENEcAi(E), ZENEcRing(E));
    goto error3;
  }
  if (ZENEltCreate(ZENEcEi(E), ZENEcRing(E))) {
    ZENEltFree(ZENEcDi(E), ZENEcRing(E));
    ZENEltFree(ZENEcCi(E), ZENEcRing(E));
    ZENEltFree(ZENEcBi(E), ZENEcRing(E));
    ZENEltFree(ZENEcAi(E), ZENEcRing(E));
    goto error3;
  }
  
  return(E);

 error3:
  ZENEltFree(ZENEcAd(E), Rg);
  ZENEltFree(ZENEcBd(E), Rg);
  ZENEltFree(ZENEcCd(E), Rg);
  ZENEltFree(ZENEcEd(E), Rg);
  ZENEltFree(ZENEcDd(E), Rg);

 error2:
  ZENEltFree(ZENEcB1(E), Rg);
  ZENEltFree(ZENEcB2(E), Rg);
  ZENEltFree(ZENEcB3(E), Rg);
  ZENEltFree(ZENEcB4(E), Rg);
  ZENEltFree(ZENEcB6(E), Rg);

 error1:
  ZENEltFree(ZENEcA1(E), Rg);
  ZENEltFree(ZENEcA2(E), Rg);
  ZENEltFree(ZENEcA3(E), Rg);
  ZENEltFree(ZENEcA4(E), Rg);
  ZENEltFree(ZENEcA6(E), Rg);

 error0:
  ZENEltFree(ZENEcD(E), Rg);
  ZENEltFree(ZENEcDb(E), Rg);
  ZENEltFree(ZENEcJ(E), Rg);
  ZENEltFree(ZENEcX(E), Rg);
  ZENEltFree(ZENEcY(E), Rg);
  ZENEltFree(ZENEcZ(E), Rg);
  ZENFree(ZENEcXY(E));

 error:
  return(NULL);
}


/* \TEX{
   \title{Initialization of an elliptic curve.}
   \input{Five parameters |a1|,|a2|, |a3|, |a4|, |a6| of an elliptic
   curve the equation of which is $Y^2+a_1XY+a_3Y=X^3+a_2X^2+a4_X+a6$.}
   \output{|ZEN_HAS_INVERSE| if the curve was succesfully initialized,
   |ZEN_NO_INVERSE| if a factor of the definition polynomial was
   found, |ZENERR| otherwise.}
   \note{Some precomputations can be done there. |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENEcInitialize(E, a1, a2, a3, a4, a6)
     ZENEc E;
     ZENElt a1;
     ZENElt a2;
     ZENElt a3;
     ZENElt a4;
     ZENElt a6;
{
  ZENElt d2, d4, d6, d8, t;
  int ret;
  ZENRing Rg;

  Rg = ZENEcRing(E);
  ret = ZEN_HAS_INVERSE;

  /* First, allocation */
  if (ZENEltCreate(d2, Rg)) {
    return(ZERR);
  }
  if (ZENEltCreate(d4, Rg)) {
    ZENEltFree(d2, Rg);
    return(ZERR);
  }
  if (ZENEltCreate(d6, Rg)) {
    ZENEltFree(d2, Rg);
    ZENEltFree(d4, Rg);
    return(ZERR);
  }
  if (ZENEltCreate(d8, Rg)) {
    ZENEltFree(d2, Rg);
    ZENEltFree(d4, Rg);
    ZENEltFree(d6, Rg);
    return(ZERR);
  }
  if (ZENEltCreate(t, Rg)) {
    ZENEltFree(d2, Rg);
    ZENEltFree(d4, Rg);
    ZENEltFree(d6, Rg);
    ZENEltFree(d8, Rg);
    return(ZERR);
  }
  
  /* a1, a2, a3, a4, a6 */
  ZENEltAssign(ZENEcA1(E), a1, Rg);
  ZENEltAssign(ZENEcA2(E), a2, Rg);
  ZENEltAssign(ZENEcA3(E), a3, Rg);
  ZENEltAssign(ZENEcA4(E), a4, Rg);
  ZENEltAssign(ZENEcA6(E), a6, Rg);

  /* d2 */
  ZENEltSquare(d2, ZENEcA1(E), Rg);
  ZENEltAdd(d2, ZENEcA2(E), Rg);
  ZENEltAdd(d2, ZENEcA2(E), Rg);
  ZENEltAdd(d2, ZENEcA2(E), Rg);
  ZENEltAdd(d2, ZENEcA2(E), Rg);

  /* d4 */
  ZENEltMultiply(d4, ZENEcA1(E), ZENEcA3(E), Rg);
  ZENEltAdd(d4, ZENEcA4(E), Rg);
  ZENEltAdd(d4, ZENEcA4(E), Rg);

  /* d6 */  
  ZENEltSquare(d6, ZENEcA3(E), Rg);
  ZENEltAdd(d6, ZENEcA6(E), Rg);
  ZENEltAdd(d6, ZENEcA6(E), Rg);
  ZENEltAdd(d6, ZENEcA6(E), Rg);
  ZENEltAdd(d6, ZENEcA6(E), Rg);

  /* d8 */
  ZENEltSquare(d8, ZENEcA1(E), Rg);
  ZENEltMultiply(d8, ZENEcA6(E), d8, Rg);
  
  ZENEltMultiply(t, ZENEcA2(E), ZENEcA6(E), Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(d8, t, Rg);
  
  ZENEltMultiply(t, ZENEcA1(E), ZENEcA3(E), Rg);
  ZENEltMultiply(t, ZENEcA4(E), t, Rg);
  ZENEltSubtract(d8, t, Rg);

  ZENEltSquare(t, ZENEcA3(E), Rg);
  ZENEltMultiply(t, ZENEcA2(E), t, Rg);
  ZENEltAdd(d8, t, Rg);

  ZENEltSquare(t, ZENEcA4(E), Rg);
  ZENEltSubtract(d8, t, Rg);

  /* discriminant */
  ZENEltMultiply(ZENEcD(E), d2, d4, Rg);
  ZENEltMultiply(ZENEcD(E), d6, ZENEcD(E), Rg);
  ZENEltAssign(t, ZENEcD(E), Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(ZENEcD(E), t, Rg);
  
  ZENEltSquare(d6, d6, Rg);
  ZENEltAssign(t, d6, Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(t, d6, Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(t, d6, Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(t, d6, Rg);
  ZENEltSubtract(ZENEcD(E), t, Rg);

  ZENEltSquare(t, d4, Rg);
  ZENEltMultiply(t, d4, t, Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltAdd(t, t, Rg);
  ZENEltSubtract(ZENEcD(E), t, Rg);

  ZENEltSquare(t, d2, Rg);
  ZENEltMultiply(t, d8, t, Rg);
  ZENEltSubtract(ZENEcD(E), t, Rg);

  if (!ZENEltIsZero(ZENEcD(E), Rg)) {

    /* Elliptic curve
     ******************/

    /* invariant */
    ZENEltSquare(ZENEcJ(E), d2, Rg);
    ZENEltAssign(t, d4, Rg);
    ZENEltAdd(t, t, Rg);
    ZENEltAdd(t, d4, Rg);
    ZENEltAdd(t, t, Rg);
    ZENEltAdd(t, t, Rg);
    ZENEltAdd(t, t, Rg);
    ZENEltSubtract(ZENEcJ(E), t, Rg);

    ZENEltSquare(t, ZENEcJ(E), Rg);
    ZENEltMultiply(ZENEcJ(E), t, ZENEcJ(E), Rg);
    
    if ((ret = ZENEltInverse(t, ZENEcD(E), Rg)) != ZEN_HAS_INVERSE) {
      goto end;
    }
    
    ZENEltMultiply(ZENEcJ(E), t, ZENEcJ(E), Rg);

    /* Values by default */
    ZENEltAssign(ZENEcB1(E), ZENEcA1(E), Rg);
    ZENEltAssign(ZENEcB2(E), ZENEcA2(E), Rg);
    ZENEltAssign(ZENEcB3(E), ZENEcA3(E), Rg);
    ZENEltAssign(ZENEcB4(E), ZENEcA4(E), Rg);
    ZENEltAssign(ZENEcB6(E), ZENEcA6(E), Rg);

    ZENEltSetToOne(ZENEcAd(E), ZENEcRing(E));
    ZENEltSetToZero(ZENEcBd(E), ZENEcRing(E));
    ZENEltSetToZero(ZENEcCd(E), ZENEcRing(E));
    ZENEltSetToOne(ZENEcDd(E), ZENEcRing(E));
    ZENEltSetToZero(ZENEcEd(E), ZENEcRing(E));

    ZENEltSetToOne(ZENEcAi(E), ZENEcRing(E));
    ZENEltSetToZero(ZENEcBi(E), ZENEcRing(E));
    ZENEltSetToZero(ZENEcCi(E), ZENEcRing(E));
    ZENEltSetToOne(ZENEcDi(E), ZENEcRing(E));
    ZENEltSetToZero(ZENEcEi(E), ZENEcRing(E));

    ZENEltAssign(ZENEcDb(E), ZENEcD(E), ZENEcRing(E));

    /* Initialization following the characteristic */
    if (ZENRingPl(Rg) == 1) {
      if (ZBNDigit(ZENRingP(Rg), 0) == 2) {
	ret = ZENEcInitializeChar2(E);
      }
      else if (ZBNDigit(ZENRingP(Rg), 0) == 3) {
	ret = ZENEcInitializeChar3(E);
      }
      else {
	ret = ZENEcInitializeCharp(E);
      }
    }
    else {
      ret = ZENEcInitializeCharp(E);
    }
  }
  else {
    /* Singular curve
     ******************/

    /* Allocation */
    E->PtCreate = _ZENEcPtCreate;
    E->PtFree = _ZENEcPtFree;
    
    /* Assigning */
    E->PtSetToZero = _ZENEcPtSetToZero;
    E->PtSetNext = _ZENEcPtSetNext;
    E->PtAssign = _ZENEcPtAssign;
    E->PtGetX = _ZENEcPtGetX;
    E->PtGetY = _ZENEcPtGetY;
    E->PtSetXY = _ZENEcPtSetXY;
    E->PtGeneral2Canonical = _ZENEcPtAssign;
    E->PtCanonical2General = _ZENEcPtAssign;

    /* Test */
    E->PtIsZero = _ZENEcPtIsZero;
    E->PtIsOnEc = _ZENEcPtIsOnEc;
    E->PtAreEqual = _ZENEcPtAreEqual;
    
    /* Arithmetic */
    E->PtPreAdd = _ZENEcPtPreAdd;
    E->PtPostAdd = _ZENEcPtPostAddAllChar;
    
    E->PtPreDouble = _ZENEcPtPreDoubleAllChar;
    E->PtPostDouble = _ZENEcPtPostDoubleAllChar;
    
    E->PtNegate = _ZENEcPtNegateAllChar;
  }

  /* Freeing memory */
  end:
  ZENEltFree(d2, Rg);
  ZENEltFree(d4, Rg);
  ZENEltFree(d6, Rg);
  ZENEltFree(d8, Rg);
  ZENEltFree(t, Rg);
  
  return(ret);
}

/* \TEX{
   \title{Reducing an elliptic curve to its canonical representation
   in fields of characteristic 2.}
   \input{An elliptic the equation of which is
   $Y^2+a_1XY+a_3Y=X^3+a_2X^2+a4_X+a6$.} 
   \output{|ZEN_HAS_INVERSE| if the curve was succesfully initialized,
   |ZEN_NO_INVERSE| if a factor of the definition polynomial was
   found, |ZERR| otherwise.}
   \side{If the invariant of the curve is 0, we compute a substitution
   to get an equation of the form $Y^2+a_3Y=X^3+a4_X+a6$, otherwise we
   get an equation of the form $Y^2+XY=X^3+a4_X+a6$.}
   \note{Some precomputations can be done there. |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENEcInitializeChar2(E)
     ZENEc E;
{
  int ret;

  if (ZENEltIsZero(ZENEcJ(E), ZENEcRing(E))) {
    
    if (!ZENEltIsZero(ZENEcA1(E), ZENEcRing(E)) ||
	!ZENEltIsZero(ZENEcA2(E), ZENEcRing(E))) {

      /*
     * That means a1 = 0, 
     * 
     * We set Y = y;
     *    and X = x+a2;
     * 
     * That is to say y = Y;
     *            and x = X+a2;
     * 
     * And we get
     * y^2 + a3*y = x^3+(a2^2+a4)*x+a6+a4*a2;
     *
     * Delta = b3^4;
     */    
    
    /* B4 */
      ZENEltSquare(ZENEcB4(E), ZENEcA2(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB4(E), ZENEcA4(E),ZENEcRing(E) );

    /* B6 */
      ZENEltMultiply(ZENEcB6(E), ZENEcA4(E), ZENEcA2(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcA6(E), ZENEcRing(E));

    /* B3 */
      ZENEltAssign(ZENEcB3(E), ZENEcA3(E), ZENEcRing(E));

      /* The others */
      ZENEltSetToZero(ZENEcB1(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcB2(E), ZENEcRing(E));

      ZENEltSetToOne(ZENEcAd(E), ZENEcRing(E));
      ZENEltAssign(ZENEcBd(E), ZENEcA2(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcCd(E), ZENEcRing(E));
      ZENEltSetToOne(ZENEcDd(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcEd(E), ZENEcRing(E));

      ZENEltSetToOne(ZENEcAi(E), ZENEcRing(E));
      ZENEltAssign(ZENEcBi(E), ZENEcA2(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcCi(E), ZENEcRing(E));
      ZENEltSetToOne(ZENEcDi(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcEi(E), ZENEcRing(E));

    /* Db */
      ZENEltSquare(ZENEcDb(E), ZENEcB3(E), ZENEcRing(E));
      ZENEltSquare(ZENEcDb(E), ZENEcDb(E), ZENEcRing(E));
    }
    
    /* Allocation */
    E->PtCreate = _ZENEcPtCreate;
    E->PtFree = _ZENEcPtFree;
    
    /* Assigning */
    E->PtSetToZero = _ZENEcPtSetToZero;
    E->PtSetNext = _ZENEcPtSetNextChar23p;
    E->PtAssign = _ZENEcPtAssign;
    E->PtGetX = _ZENEcPtGetXChar23p;
    E->PtGetY = _ZENEcPtGetYChar23p;
    E->PtSetXY = _ZENEcPtSetXYChar23p;
    E->PtGeneral2Canonical = _ZENEcPtGeneral2Canonical;
    E->PtCanonical2General = _ZENEcPtCanonical2General;
    
    /* Test */
    E->PtIsZero = _ZENEcPtIsZero;
    E->PtIsOnEc = _ZENEcPtIsOnEcChar23p;
    E->PtAreEqual = _ZENEcPtAreEqual;

    /* Arithmetic */
    E->PtPreAdd = _ZENEcPtPreAdd;
    E->PtPostAdd = _ZENEcPtPostAddChar2J0;

    E->PtPreDouble = _ZENEcPtPreDoubleChar2J0;
    E->PtPostDouble = _ZENEcPtPostDoubleChar2J0;

    E->PtNegate = _ZENEcPtNegateChar2J0;
  }
  else {

    if (!ZENEltIsOne(ZENEcA1(E), ZENEcRing(E)) ||
	!ZENEltIsZero(ZENEcA3(E), ZENEcRing(E)) ||
	!ZENEltIsZero(ZENEcA4(E), ZENEcRing(E))) {

      /*
     * We set Y = a1^3*y+(a4*a1^2+a3^2)/a1^3;
     *    and X = a1^2*x+a3/a1;
     * 
     * That is to say y = 1/a1^3*Y+1/a1^4*a4+1/a1^6*a3^2;
     *            and x = 1/a1^2*X+1/a1^3*a3;
     * 
     * And we get
     * y^2 + x*y = x^3 
     * 	    +(1/a1^3*a3+1/a1^2*a2)*x^2
     *             +1/a1^6*a6+1/a1^8*a2*a3^2+1/a1^9*a3^3+1/a1^7*a3*a4
     *             +1/a1^8*a4^2+1/a1^12*a3^4;
     *
     * Delta = b6;
     */

    /* The isomorphism
     ******************/
      if ((ret = ZENEltInverse(ZENEcBd(E), ZENEcA1(E), ZENEcRing(E)))
	  != ZEN_HAS_INVERSE) { /* 1/a1 */
	return(ret);
      }

      /* Ai */
      ZENEltSquare(ZENEcAi(E), ZENEcBd(E), ZENEcRing(E)); /*  1/a1^2 */

      /* Di */
      ZENEltMultiply(ZENEcDi(E), ZENEcAi(E), ZENEcBd(E), ZENEcRing(E)); /*  1/a1^3 */
    
      /* B6 */
      ZENEltSquare(ZENEcAd(E), ZENEcDi(E), ZENEcRing(E)); /* 1/a1^6 */
      ZENEltMultiply(ZENEcB6(E), ZENEcAd(E), ZENEcA6(E), ZENEcRing(E)); /* 1/a1^6*a6 */

      ZENEltSquare(ZENEcEi(E), ZENEcA3(E), ZENEcRing(E)); /* a3^2 */
      ZENEltMultiply(ZENEcEi(E), ZENEcAd(E), ZENEcEi(E), ZENEcRing(E)); /* 1/a1^6*a3^2 */

      ZENEltMultiply(ZENEcAd(E), ZENEcBd(E), ZENEcAd(E), ZENEcRing(E)); /* 1/a1^7 */
      ZENEltMultiply(ZENEcB2(E), ZENEcA3(E), ZENEcAd(E), ZENEcRing(E)); /* 1/a1^7*a3 */
      ZENEltMultiply(ZENEcB2(E), ZENEcA4(E), ZENEcB2(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* +1/a1^7*a3*a4 */
    
      ZENEltMultiply(ZENEcAd(E), ZENEcBd(E), ZENEcAd(E), ZENEcRing(E)); /* 1/a1^8 */
      ZENEltSquare(ZENEcB2(E), ZENEcA3(E), ZENEcRing(E)); /* a3^2 */
      ZENEltMultiply(ZENEcB2(E), ZENEcAd(E), ZENEcB2(E), ZENEcRing(E)); /* 1/a1^8*a3^2 */
      ZENEltMultiply(ZENEcB2(E), ZENEcA2(E), ZENEcB2(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* +1/a1^8*a3^2*a2 */

      ZENEltSquare(ZENEcB2(E), ZENEcA4(E), ZENEcRing(E)); /* a4^2 */
      ZENEltMultiply(ZENEcB2(E), ZENEcAd(E), ZENEcB2(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* +1/a1^8*a4^2 */
    
      ZENEltMultiply(ZENEcAd(E), ZENEcBd(E), ZENEcAd(E), ZENEcRing(E)); /* 1/a1^9 */
      ZENEltSquare(ZENEcB2(E), ZENEcA3(E), ZENEcRing(E)); /* a3^2 */
      ZENEltMultiply(ZENEcB2(E), ZENEcA3(E), ZENEcB2(E), ZENEcRing(E)); /* a3^3 */
      ZENEltMultiply(ZENEcB2(E), ZENEcAd(E), ZENEcB2(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* +1/a1^9*a3^3 */
    
      ZENEltMultiply(ZENEcAd(E), ZENEcDi(E), ZENEcAd(E), ZENEcRing(E)); /* 1/a1^12 */
      ZENEltSquare(ZENEcB2(E), ZENEcA3(E), ZENEcRing(E)); /* a3^2 */
      ZENEltSquare(ZENEcB2(E), ZENEcB2(E), ZENEcRing(E)); /* a3^4 */
      ZENEltMultiply(ZENEcB2(E), ZENEcAd(E), ZENEcB2(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* +1/a1^12*a3^4 */
    
    /* B2 */
      ZENEltMultiply(ZENEcB2(E), ZENEcA3(E), ZENEcDi(E), ZENEcRing(E)); /* 1/a1^3*a3 */
      ZENEltMultiply(ZENEcDd(E), ZENEcA2(E), ZENEcAi(E), ZENEcRing(E)); /* 1/a1^2*a2 */
      ZENEltAdd(ZENEcB2(E), ZENEcDd(E), ZENEcRing(E));
    
      /* Ei */
      ZENEltSquare(ZENEcAd(E), ZENEcAi(E), ZENEcRing(E)); /* 1/a1^4 */
      ZENEltMultiply(ZENEcAd(E), ZENEcA4(E), ZENEcAd(E), ZENEcRing(E)); /* 1/a1^4*a4 */
      ZENEltAdd(ZENEcEi(E), ZENEcAd(E), ZENEcRing(E));

      /* Ed */
      ZENEltMultiply(ZENEcEd(E), ZENEcA4(E), ZENEcBd(E), ZENEcRing(E)); /* a4/a1 */
      ZENEltSquare(ZENEcAd(E), ZENEcA3(E), ZENEcRing(E)); /* a3^2 */
      ZENEltMultiply(ZENEcAd(E), ZENEcDi(E), ZENEcAd(E), ZENEcRing(E)); /* 1/a1^3*a3^2 */
      ZENEltAdd(ZENEcEd(E), ZENEcAd(E), ZENEcRing(E));
    
    /* Bi */
      ZENEltMultiply(ZENEcBi(E), ZENEcA3(E), ZENEcDi(E), ZENEcRing(E)); /* a3/a1^3 */
    
      /* Ad */
      ZENEltSquare(ZENEcAd(E), ZENEcA1(E), ZENEcRing(E)); /* a1^2 */

      /* Dd */
      ZENEltMultiply(ZENEcDd(E), ZENEcA1(E), ZENEcAd(E), ZENEcRing(E)); /* a1^3 */

      /* Bd */
      ZENEltMultiply(ZENEcBd(E), ZENEcA3(E), ZENEcBd(E), ZENEcRing(E)); /* a1^3 */

      /* The others */
      ZENEltSetToOne(ZENEcB1(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcB3(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcB4(E), ZENEcRing(E));

      ZENEltSetToZero(ZENEcCd(E), ZENEcRing(E));

      ZENEltSetToZero(ZENEcCi(E), ZENEcRing(E));

      /* Db */
      ZENEltAssign(ZENEcDb(E), ZENEcB6(E), ZENEcRing(E));
    }

    /* Allocation */
    E->PtCreate = _ZENEcPtCreate;
    E->PtFree = _ZENEcPtFree;
    
    /* Assigning */
    E->PtSetToZero = _ZENEcPtSetToZero;
    E->PtSetNext = _ZENEcPtSetNextChar23p;
    E->PtAssign = _ZENEcPtAssign;
    E->PtGetX = _ZENEcPtGetXChar23p;
    E->PtGetY = _ZENEcPtGetYChar23p;
    E->PtSetXY = _ZENEcPtSetXYChar23p;
    E->PtGeneral2Canonical = _ZENEcPtGeneral2Canonical;
    E->PtCanonical2General = _ZENEcPtCanonical2General;
    
    /* Test */
    E->PtIsZero = _ZENEcPtIsZero;
    E->PtIsOnEc = _ZENEcPtIsOnEcChar23p;
    E->PtAreEqual = _ZENEcPtAreEqual;
    
    /* Arithmetic */
    E->PtPreAdd = _ZENEcPtPreAdd;
    E->PtPostAdd = _ZENEcPtPostAddChar2J1;

    E->PtPreDouble = _ZENEcPtPreDoubleChar2J1;
    E->PtPostDouble = _ZENEcPtPostDoubleChar2J1;

    E->PtNegate = _ZENEcPtNegateChar2J1;
  }

  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Reducing an elliptic curve to its canonical representation
   in fields of characteristic 3.}
   \input{An elliptic the equation of which is
   $Y^2+a_1XY+a_3Y=X^3+a_2X^2+a4_X+a6$.} 
   \output{|ZEN_HAS_INVERSE| if the curve was succesfully initialized,
   |ZEN_NO_INVERSE| if a factor of the definition polynomial was
   found, |ZERR| otherwise.}
   \side{If the invariant of the curve is 0, we compute a substitution
   to get an equation of the form $Y^2=X^3+a4_X+a6$, otherwise we
   get an equation of the form $Y^2=X^3+a2_X^2+a6$.}
   \note{Some precomputations can be done there. |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENEcInitializeChar3(E)
     ZENEc E;
{ 
  int ret;
 
  if (ZENEltIsZero(ZENEcJ(E), ZENEcRing(E))) {

    if (!ZENEltIsZero(ZENEcA1(E), ZENEcRing(E)) ||
	!ZENEltIsZero(ZENEcA3(E), ZENEcRing(E)) ||
	!ZENEltIsZero(ZENEcA2(E), ZENEcRing(E))) {
      /*
     * So a2 = 2*a1^2, 
     * 
     * We set Y = 2*y+a1*x+a3;
     *    and X = x;
     * 
     * Or equivalently y = 2*Y+a1*X+a3;
     *             and x = X;
     * 
     * And we get
     * y^2 = x^3+(2*a3*a1+a4)*x+a3^2+a6
     *
     * Delta = -b4^3;
     */

    /* B4 */
      ZENEltMultiply(ZENEcB4(E), ZENEcA1(E), ZENEcA3(E),  ZENEcRing(E)); /* a1*a3 */
      ZENEltAdd(ZENEcB4(E), ZENEcB4(E), ZENEcRing(E)); /* 2*a1*a3 */
      ZENEltAdd(ZENEcB4(E), ZENEcA4(E), ZENEcRing(E));

      /* B6 */
      ZENEltSquare(ZENEcB6(E), ZENEcA3(E), ZENEcRing(E));	/* a3^2 */
      ZENEltAdd(ZENEcB6(E), ZENEcA6(E), ZENEcRing(E));

    /* The others */
      ZENEltSetToZero(ZENEcB1(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcB2(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcB3(E), ZENEcRing(E));

      ZENEltSetToOne(ZENEcAd(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcBd(E), ZENEcRing(E));
      ZENEltAssign(ZENEcCd(E), ZENEcA1(E), ZENEcRing(E));
      ZENEltSetToOne(ZENEcDd(E), ZENEcRing(E));
      ZENEltNegate(ZENEcDd(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAssign(ZENEcEd(E), ZENEcA3(E), ZENEcRing(E));

      ZENEltSetToOne(ZENEcAi(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcBi(E), ZENEcRing(E));
      ZENEltAssign(ZENEcCi(E), ZENEcA1(E), ZENEcRing(E));
      ZENEltSetToOne(ZENEcDi(E), ZENEcRing(E));
      ZENEltNegate(ZENEcDi(E), ZENEcDi(E), ZENEcRing(E));
      ZENEltAssign(ZENEcEi(E), ZENEcA3(E), ZENEcRing(E));

      /* Db */
      ZENEltSquare(ZENEcDb(E), ZENEcB4(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDb(E), ZENEcB4(E), ZENEcDb(E),  ZENEcRing(E));
      ZENEltNegate(ZENEcDb(E), ZENEcDb(E), ZENEcRing(E));
    }

    /* Allocation */
    E->PtCreate = _ZENEcPtCreate;
    E->PtFree = _ZENEcPtFree;
    
    /* Assigning */
    E->PtSetToZero = _ZENEcPtSetToZero;
    E->PtSetNext = _ZENEcPtSetNextChar23p;
    E->PtAssign = _ZENEcPtAssign;
    E->PtGetX = _ZENEcPtGetXChar23p;
    E->PtGetY = _ZENEcPtGetYChar23p;
    E->PtSetXY = _ZENEcPtSetXYChar23p;
    E->PtGeneral2Canonical = _ZENEcPtGeneral2Canonical;
    E->PtCanonical2General = _ZENEcPtCanonical2General;
    
    /* Test */
    E->PtIsZero = _ZENEcPtIsZero;
    E->PtIsOnEc = _ZENEcPtIsOnEcChar23p;
    E->PtAreEqual = _ZENEcPtAreEqual;

    /* Arithmetic */
    E->PtPreAdd = _ZENEcPtPreAdd;
    E->PtPostAdd = _ZENEcPtPostAddChar3J0;

    E->PtPreDouble = _ZENEcPtPreDoubleChar3J0;
    E->PtPostDouble = _ZENEcPtPostDoubleChar3J0;

    E->PtNegate = _ZENEcPtNegateChar3;
  }
  else {

    if (!ZENEltIsZero(ZENEcA1(E), ZENEcRing(E)) ||
	!ZENEltIsZero(ZENEcA3(E), ZENEcRing(E)) ||
	!ZENEltIsZero(ZENEcA4(E), ZENEcRing(E))) {
      /*
     * We set Y = 2*y+a1*x+2/(a1^2+a2)*a3*a1^2+a1/(a1^2+a2)*a4+a3;
     *    and X = x+(2*a3*a1+a4)/(a1^2+a2);
     * 
     * Or equivalently y = 2*Y+a1*X+a3;
     *             and x = X+(a3*a1+2*a4)/(a1^2+a2);
     * 
     * We get:
     * y^2 = x^3+
     *       x^2*(a1^2+a2)+
     *       a6+a3^2+1/(a1^2+a2)^2*a3^2*a1^4+1/(a1^2+a2)*a3^2*a1^2+
     *       a1^2/(a1^2+a2)^2*a4^2+1/(a1^2+a2)^3*a4^3+1/(a1^2+a2)*a4^2+
     *       1/(a1^2+a2)^2*a3*a1^3*a4+a1/(a1^2+a2)*a4*a3+a2/(a1^2+a2)^2*a3^2*a1^2+
     *       a2/(a1^2+a2)^2*a3*a1*a4+2/(a1^2+a2)^3*a3^3*a1^3+a2/(a1^2+a2)^2*a4^2
     *
     * Delta = b6;
     */

      ZENEltSquare(ZENEcCi(E), ZENEcA1(E), ZENEcRing(E)); /* a1^2 */
      ZENEltMultiply(ZENEcDi(E), ZENEcA1(E), ZENEcCi(E), ZENEcRing(E)); /*  a1^3  */
      ZENEltSquare(ZENEcEi(E), ZENEcCi(E), ZENEcRing(E)); /* a1^4 */

      /* B2 */
      ZENEltAssign(ZENEcB2(E), ZENEcCi(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB2(E), ZENEcA2(E), ZENEcRing(E)); /* a1^2+a2 */

    /* B6 */
      if ((ret = ZENEltInverse(ZENEcB1(E), ZENEcB2(E), ZENEcRing(E)))
	  != ZEN_HAS_INVERSE) {	/* 1/(a1^2+a2) */
	return(ret);
      }
      ZENEltSquare(ZENEcB3(E), ZENEcB1(E), ZENEcRing(E)); /*  1/(a1^2+a2)^2  */
      ZENEltMultiply(ZENEcB4(E), ZENEcB3(E), ZENEcB1(E), ZENEcRing(E));
      /*  1/(a1^2+a2)^3  */

      ZENEltSquare(ZENEcAi(E), ZENEcA3(E), ZENEcRing(E)); /* a3^2 */
      ZENEltMultiply(ZENEcBi(E), ZENEcA3(E), ZENEcAi(E), ZENEcRing(E)); /*  a3^3  */
    
      ZENEltSquare(ZENEcAd(E), ZENEcA4(E), ZENEcRing(E)); /* a4^2 */
      ZENEltMultiply(ZENEcCd(E), ZENEcA4(E), ZENEcAd(E), ZENEcRing(E)); /*  a4^3  */
    
      ZENEltMultiply(ZENEcB6(E), ZENEcB3(E), ZENEcAi(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcB6(E), ZENEcEi(E), ZENEcB6(E), ZENEcRing(E)); /* 1/(a1^2+a2)^2*a3^2*a1^4 */
    
      ZENEltMultiply(ZENEcDd(E), ZENEcB1(E), ZENEcAi(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcCi(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E)); /* +1/(a1^2+a2)*a3^2*a1^2 */
    
      ZENEltMultiply(ZENEcDd(E), ZENEcB3(E), ZENEcCi(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcAd(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E)); /* +a1^2/(a1^2+a2)^2*a4^2 */
    
      ZENEltMultiply(ZENEcDd(E), ZENEcCd(E), ZENEcB4(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E)); /* +1/(a1^2+a2)^3*a4^3 */
    
      ZENEltMultiply(ZENEcDd(E), ZENEcB1(E), ZENEcAd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E)); /* +1/(a1^2+a2)*a4^2 */

      ZENEltMultiply(ZENEcDd(E), ZENEcB3(E), ZENEcA3(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcA4(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcDi(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E)); /* +1/(a1^2+a2)^2*a3*a1^3*a4 */

      ZENEltMultiply(ZENEcDd(E), ZENEcB1(E), ZENEcA1(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcA3(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcA4(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E)); /* +a1/(a1^2+a2)*a4*a3 */

      ZENEltMultiply(ZENEcDd(E), ZENEcB3(E), ZENEcA2(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcAi(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcCi(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E)); /* +a2/(a1^2+a2)^2*a3^2*a1^2 */

      ZENEltMultiply(ZENEcDd(E), ZENEcB3(E), ZENEcA2(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcA3(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcA1(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcA4(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E)); /* +a2/(a1^2+a2)^2*a3*a1*a4 */

      ZENEltMultiply(ZENEcDd(E), ZENEcB4(E), ZENEcBi(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcDi(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E)); /* +2/(a1^2+a2)^3*a3^3*a1^3 */

      ZENEltMultiply(ZENEcDd(E), ZENEcB3(E), ZENEcA2(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcAd(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcB6(E), ZENEcDd(E), ZENEcRing(E)); /* +a2/(a1^2+a2)^2*a4^2 */

      ZENEltAdd(ZENEcB6(E), ZENEcA6(E), ZENEcRing(E)); /* +a6 */

      ZENEltAdd(ZENEcB6(E), ZENEcAi(E), ZENEcRing(E)); /* +a3^2 */

    /* Ed */
      ZENEltMultiply(ZENEcEd(E), ZENEcB1(E), ZENEcA3(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcEd(E), ZENEcCi(E), ZENEcEd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcEd(E), ZENEcEd(E), ZENEcRing(E)); /* 2/(a1^2+a2)*a3*a1^2 */

      ZENEltMultiply(ZENEcDd(E), ZENEcB1(E), ZENEcA1(E), ZENEcRing(E));
      ZENEltMultiply(ZENEcDd(E), ZENEcA4(E), ZENEcDd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcEd(E), ZENEcDd(E), ZENEcRing(E)); /* +a1/(a1^2+a2)*a4 */ 

      ZENEltAdd(ZENEcEd(E), ZENEcA3(E), ZENEcRing(E)); /* +a3 */
    
      /* Bd and Bi */
      ZENEltMultiply(ZENEcBi(E), ZENEcA3(E), ZENEcA1(E), ZENEcRing(E));
      ZENEltAssign(ZENEcBd(E), ZENEcBi(E), ZENEcRing(E));
    
      ZENEltAdd(ZENEcBi(E), ZENEcA4(E), ZENEcRing(E));
      ZENEltAdd(ZENEcBi(E), ZENEcA4(E), ZENEcRing(E)); /* a3*a1+2*a4 */
    
      ZENEltAdd(ZENEcBd(E), ZENEcBd(E), ZENEcRing(E));
      ZENEltAdd(ZENEcBd(E), ZENEcA4(E), ZENEcRing(E)); /* 2*a3*a1+a4 */

      ZENEltMultiply(ZENEcBi(E), ZENEcB1(E), ZENEcBi(E), ZENEcRing(E)); /* (2*a3*a1+a4)/(a1^2+a2) */
      ZENEltMultiply(ZENEcBd(E), ZENEcB1(E), ZENEcBd(E), ZENEcRing(E)); /* (a3*a1+2*a4)/(a1^2+a2) */
    
    /* The others */
      ZENEltSetToZero(ZENEcB1(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcB3(E), ZENEcRing(E));
      ZENEltSetToZero(ZENEcB4(E), ZENEcRing(E));

      ZENEltSetToOne(ZENEcAd(E), ZENEcRing(E));
      ZENEltAssign(ZENEcCd(E), ZENEcA1(E), ZENEcRing(E));
      ZENEltSetToOne(ZENEcDd(E), ZENEcRing(E));
      ZENEltNegate(ZENEcDd(E), ZENEcDd(E), ZENEcRing(E));

      ZENEltSetToOne(ZENEcAi(E), ZENEcRing(E));
      ZENEltAssign(ZENEcCi(E), ZENEcA1(E), ZENEcRing(E));
      ZENEltSetToOne(ZENEcDi(E), ZENEcRing(E));
      ZENEltNegate(ZENEcDi(E), ZENEcDi(E), ZENEcRing(E));
      ZENEltAssign(ZENEcEi(E), ZENEcA3(E), ZENEcRing(E));

    /* Db */
      ZENEltAssign(ZENEcDb(E), ZENEcB6(E), ZENEcRing(E));
    }

    /* Allocation */
    E->PtCreate = _ZENEcPtCreate;
    E->PtFree = _ZENEcPtFree;
    
    /* Assigning */
    E->PtSetToZero = _ZENEcPtSetToZero;
    E->PtSetNext = _ZENEcPtSetNextChar23p;
    E->PtAssign = _ZENEcPtAssign;
    E->PtGetX = _ZENEcPtGetXChar23p;
    E->PtGetY = _ZENEcPtGetYChar23p;
    E->PtSetXY = _ZENEcPtSetXYChar23p;
    E->PtGeneral2Canonical = _ZENEcPtGeneral2Canonical;
    E->PtCanonical2General = _ZENEcPtCanonical2General;
    
    /* Test */
    E->PtIsZero = _ZENEcPtIsZero;
    E->PtIsOnEc = _ZENEcPtIsOnEcChar23p;
    E->PtAreEqual = _ZENEcPtAreEqual;
    
    /* Arithmetic */
    E->PtPreAdd = _ZENEcPtPreAdd;
    E->PtPostAdd = _ZENEcPtPostAddChar3J1;

    E->PtPreDouble = _ZENEcPtPreDoubleChar3J1;
    E->PtPostDouble = _ZENEcPtPostDoubleChar3J1;

    E->PtNegate = _ZENEcPtNegateChar3;
  }

  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Reducing an elliptic curve to its canonical representation
   in fields of characteristic greater than 3.}
   \input{An elliptic the equation of which is
   $Y^2+a_1XY+a_3Y=X^3+a_2X^2+a4_X+a6$.} 
   \output{|ZEN_HAS_INVERSE| if the curve was succesfully initialized,
   |ZEN_NO_INVERSE| if a factor of the definition polynomial was
   found, |ZERR| otherwise.}
   \side{We compute a substitution to get an equation of the form
   $Y^2=X^3+a4_X+a6$.}
   \note{Some precomputations can be done there. |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENEcInitializeCharp(E)
     ZENEc E;
{
  BigNumDigit C[1];
  int ret;

  /*
   * On pose Y = 1/216*y-1/72*a1*x+1/24*a1^3+1/6*a1*a2-1/2*a3;
   *      et X = 1/36*x-1/12*a1^2-1/3*a2;
   * 
   * ou vis versa x = 36*X+3*a1^2+12*a2;
   *              y = 216*Y+108*a1*X+108*a3;
   * 
   * Et on obtient l'equation:
   * 
   * y^2 = x^3
   *       -(216*a1^2*a2+432*a2^2-1296*a4+27*a1^4-648*a3*a1)*x
   *       -(-46656*a6-3456*a2^3+7776*a1*a2*a3+15552*a4*a2
   *         -11664*a3^2-648*a1^4*a2+1944*a1^3*a3-2592*a1^2*a2^
   *         2-54*a1^6+3888*a4*a1^2)
   *
   * Delta = -16*(4*b4^3+27*b6^2)
   */

  if (!ZENEltIsZero(ZENEcA1(E), ZENEcRing(E)) ||
      !ZENEltIsZero(ZENEcA2(E), ZENEcRing(E)) ||
      !ZENEltIsZero(ZENEcA3(E), ZENEcRing(E))) {
    C[0] = 2;
    ZENZToElt(ZENEcB1(E), C, 1, ZENEcRing(E));   /* 2 */

    C[0] = 3;
    ZENZToElt(ZENEcB3(E), C, 1, ZENEcRing(E));  /* 3 */

    if ((ret = ZENEltInverse(ZENEcB2(E), ZENEcB1(E), ZENEcRing(E)))
	!= ZEN_HAS_INVERSE) {	/* 1/2 */
      return(ret);
    }
    if ((ret = ZENEltInverse(ZENEcB4(E), ZENEcB3(E), ZENEcRing(E)))
	!= ZEN_HAS_INVERSE) {	/* 1/3 */
      return(ret);
    }
  
    /* Ad */
    ZENEltSquare(ZENEcAd(E), ZENEcB4(E), ZENEcRing(E));
    ZENEltSquare(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* 1/4 */
    ZENEltMultiply(ZENEcAd(E), ZENEcB6(E), ZENEcAd(E), ZENEcRing(E)); /* 1/36 */

    /* Bd */
    ZENEltSquare(ZENEcBd(E), ZENEcA1(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcBd(E), ZENEcB4(E), ZENEcBd(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcBd(E), ZENEcB6(E), ZENEcBd(E), ZENEcRing(E)); /* 1/12*a1^2 */

    ZENEltMultiply(ZENEcB6(E), ZENEcB4(E), ZENEcA2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcBd(E), ZENEcB6(E), ZENEcRing(E)); /* +1/3*a2 */

    ZENEltNegate(ZENEcBd(E), ZENEcBd(E), ZENEcRing(E));

  /* Dd */
    ZENEltSquare(ZENEcCd(E), ZENEcB4(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcCd(E), ZENEcB2(E), ZENEcCd(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcCd(E), ZENEcB2(E), ZENEcCd(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcCd(E), ZENEcB2(E), ZENEcCd(E), ZENEcRing(E)); /* 1/72 */

    ZENEltMultiply(ZENEcDd(E), ZENEcB4(E), ZENEcCd(E), ZENEcRing(E)); /* 1/216 */

    /* Cd */
    ZENEltMultiply(ZENEcCd(E), ZENEcA1(E), ZENEcCd(E), ZENEcRing(E)); /* 1/72*a1 */

    ZENEltNegate(ZENEcCd(E), ZENEcCd(E), ZENEcRing(E));

  /* Ed */
    ZENEltMultiply(ZENEcEd(E), ZENEcB2(E), ZENEcB4(E), ZENEcRing(E)); /* 1/6 */
  
    ZENEltSquare(ZENEcB6(E), ZENEcA1(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB6(E), ZENEcA1(E), ZENEcB6(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB6(E), ZENEcEd(E), ZENEcB6(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB6(E), ZENEcB2(E), ZENEcB6(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB6(E), ZENEcB2(E), ZENEcB6(E), ZENEcRing(E)); /* 1/24*a1^3 */

    ZENEltMultiply(ZENEcEd(E), ZENEcA1(E), ZENEcEd(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcEd(E), ZENEcA2(E), ZENEcEd(E), ZENEcRing(E)); /*1/6*a1*a2 */
    ZENEltAdd(ZENEcEd(E), ZENEcB6(E), ZENEcRing(E)); /* +1/24*a1^3 */

    ZENEltMultiply(ZENEcB6(E), ZENEcB2(E),  ZENEcA3(E),  ZENEcRing(E));
    ZENEltSubtract(ZENEcEd(E), ZENEcB6(E), ZENEcRing(E)); /* -1/2*a3 */

    /* Ai */
    ZENEltSquare(ZENEcAi(E), ZENEcB3(E), ZENEcRing(E));
    ZENEltSquare(ZENEcB6(E), ZENEcB1(E), ZENEcRing(E)); /* 4 */
    ZENEltMultiply(ZENEcAi(E), ZENEcB6(E), ZENEcAi(E), ZENEcRing(E)); /* 36 */

    /* Bi */
    ZENEltMultiply(ZENEcBi(E), ZENEcB3(E), ZENEcB6(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcBi(E), ZENEcA2(E), ZENEcBi(E), ZENEcRing(E)); /* 12*a2 */

    ZENEltSquare(ZENEcB6(E), ZENEcA1(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB6(E), ZENEcB3(E), ZENEcB6(E), ZENEcRing(E));
    ZENEltAdd(ZENEcBi(E), ZENEcB6(E), ZENEcRing(E)); /* +3*a1^2 */

    /* Ei */
    ZENEltMultiply(ZENEcDi(E), ZENEcAi(E), ZENEcB3(E), ZENEcRing(E)); /* 108 */
    ZENEltMultiply(ZENEcEi(E), ZENEcA3(E), ZENEcDi(E), ZENEcRing(E)); /* 108*a3 */
  
  /* Ci */
    ZENEltMultiply(ZENEcCi(E), ZENEcDi(E), ZENEcA1(E), ZENEcRing(E)); /* 108*a1 */
  
    /* Di */
    ZENEltMultiply(ZENEcDi(E), ZENEcB1(E), ZENEcDi(E), ZENEcRing(E)); /* 216 */
  
    /* B6 */
    ZENEltMultiply(ZENEcB6(E), ZENEcDi(E), ZENEcAi(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB6(E), ZENEcB3(E), ZENEcB6(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB6(E), ZENEcB1(E), ZENEcB6(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB6(E), ZENEcA6(E), ZENEcB6(E), ZENEcRing(E)); /* 46656*a6 */
  
    ZENEltSquare(ZENEcB2(E), ZENEcA2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcDi(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* +3456*a2^3  */

    ZENEltMultiply(ZENEcB2(E), ZENEcDi(E), ZENEcAi(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltSubtract(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* -7776*a1*a2*a3 */

    ZENEltMultiply(ZENEcB2(E), ZENEcDi(E), ZENEcAi(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA4(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltSubtract(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E));	/* +15552*a4*a2 */

    ZENEltSquare(ZENEcB2(E), ZENEcAi(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E));	/* +11664*a3^2 */

    ZENEltSquare(ZENEcB2(E), ZENEcA1(E), ZENEcRing(E));
    ZENEltSquare(ZENEcB2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcDi(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* +648*a1^4*a2 */

    ZENEltSquare(ZENEcB2(E), ZENEcA1(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcDi(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltSubtract(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* -1944*a1^3*a3 */
  
    ZENEltSquare(ZENEcB2(E), ZENEcA1(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcAi(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcAi(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* +2592*a1^2*a2^2 */

    ZENEltSquare(ZENEcB2(E), ZENEcA1(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltSquare(ZENEcB2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* +54*a1^6 */
  
    ZENEltSquare(ZENEcB2(E), ZENEcA1(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA4(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcAi(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcAi(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltSubtract(ZENEcB6(E), ZENEcB2(E), ZENEcRing(E)); /* -3888*a4*a1^2 */

  /* B4 */
    ZENEltSquare(ZENEcB4(E), ZENEcA1(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB4(E), ZENEcA2(E), ZENEcB4(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB4(E), ZENEcDi(E), ZENEcB4(E), ZENEcRing(E)); /* 216*a1^2*a2 */
  
    ZENEltSquare(ZENEcB2(E), ZENEcA2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcDi(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB4(E), ZENEcB2(E), ZENEcRing(E)); /* +432*a2^2 */
  
    ZENEltMultiply(ZENEcB2(E), ZENEcDi(E), ZENEcB3(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA4(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltSubtract(ZENEcB4(E), ZENEcB2(E), ZENEcRing(E)); /* -1296*a4 */
  
    ZENEltSquare(ZENEcB2(E), ZENEcA1(E), ZENEcRing(E));
    ZENEltSquare(ZENEcB2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB4(E), ZENEcB2(E), ZENEcRing(E)); /* 27*a1^4 */
  
    ZENEltMultiply(ZENEcB2(E), ZENEcDi(E), ZENEcB3(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA1(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcA3(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltSubtract(ZENEcB4(E), ZENEcB2(E), ZENEcRing(E));	/* -648*a3*a1 */

    ZENEltNegate(ZENEcB4(E), ZENEcB4(E), ZENEcRing(E));

  /* Db */
    ZENEltSquare(ZENEcB1(E), ZENEcB6(E), ZENEcRing(E));
    ZENEltAssign(ZENEcB2(E), ZENEcB1(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB2(E), ZENEcB1(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB2(E), ZENEcB1(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB2(E), ZENEcB1(E), ZENEcRing(E));
    ZENEltNegate(ZENEcDb(E), ZENEcB2(E), ZENEcRing(E));

    ZENEltSquare(ZENEcB2(E), ZENEcB4(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcB2(E), ZENEcB4(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(ZENEcB2(E), ZENEcB2(E), ZENEcRing(E));
    ZENEltSubtract(ZENEcDb(E), ZENEcB2(E), ZENEcRing(E));

    ZENEltAdd(ZENEcDb(E), ZENEcDb(E), ZENEcRing(E));
    ZENEltAdd(ZENEcDb(E), ZENEcDb(E), ZENEcRing(E));
    ZENEltAdd(ZENEcDb(E), ZENEcDb(E), ZENEcRing(E));
    ZENEltAdd(ZENEcDb(E), ZENEcDb(E), ZENEcRing(E));

    /* The others */
    ZENEltSetToZero(ZENEcB1(E), ZENEcRing(E));
    ZENEltSetToZero(ZENEcB2(E), ZENEcRing(E));
    ZENEltSetToZero(ZENEcB3(E), ZENEcRing(E));
  }

  /* Allocation */
  E->PtCreate = _ZENEcPtCreate;
  E->PtFree = _ZENEcPtFree;
  
  /* Assigning */
  E->PtSetToZero = _ZENEcPtSetToZero;
  E->PtSetNext = _ZENEcPtSetNextChar23p;
  E->PtAssign = _ZENEcPtAssign;
  E->PtGetX = _ZENEcPtGetXChar23p;
  E->PtGetY = _ZENEcPtGetYChar23p;
  E->PtSetXY = _ZENEcPtSetXYChar23p;
  E->PtGeneral2Canonical = _ZENEcPtGeneral2Canonical;
  E->PtCanonical2General = _ZENEcPtCanonical2General;

  /* Test */
  E->PtIsZero = _ZENEcPtIsZero;
  E->PtIsOnEc = _ZENEcPtIsOnEcChar23p;
  E->PtAreEqual = _ZENEcPtAreEqual;

  /* Arithmetic */
  E->PtPreAdd = _ZENEcPtPreAdd;
  E->PtPostAdd = _ZENEcPtPostAddCharp;

  E->PtPreDouble = _ZENEcPtPreDoubleCharp;
  E->PtPostDouble = _ZENEcPtPostDoubleCharp;

  E->PtNegate = _ZENEcPtNegateCharp;

  return(ZEN_HAS_INVERSE);  
}

/* \TEX{
   \title{Desallocating}
   \input{An elliptic curve |E|.}
   \side{|E| is freed.}
   }
   */
void _ZENEcFree(E)
     ZENEc E;
{

  ZENEltFree(ZENEcB1(E), ZENEcRing(E));
  ZENEltFree(ZENEcB2(E), ZENEcRing(E));
  ZENEltFree(ZENEcB3(E), ZENEcRing(E));
  ZENEltFree(ZENEcB4(E), ZENEcRing(E));
  ZENEltFree(ZENEcB6(E), ZENEcRing(E));

  ZENEltFree(ZENEcAd(E), ZENEcRing(E));
  ZENEltFree(ZENEcBd(E), ZENEcRing(E));
  ZENEltFree(ZENEcCd(E), ZENEcRing(E));
  ZENEltFree(ZENEcDd(E), ZENEcRing(E));
  ZENEltFree(ZENEcEd(E), ZENEcRing(E));

  ZENEltFree(ZENEcAi(E), ZENEcRing(E));
  ZENEltFree(ZENEcBi(E), ZENEcRing(E));
  ZENEltFree(ZENEcCi(E), ZENEcRing(E));
  ZENEltFree(ZENEcDi(E), ZENEcRing(E));
  ZENEltFree(ZENEcEi(E), ZENEcRing(E));

  ZENEltFree(ZENEcA1(E), ZENEcRing(E));
  ZENEltFree(ZENEcA2(E), ZENEcRing(E));
  ZENEltFree(ZENEcA3(E), ZENEcRing(E));
  ZENEltFree(ZENEcA4(E), ZENEcRing(E));
  ZENEltFree(ZENEcA6(E), ZENEcRing(E));

  ZENEltFree(ZENEcD(E), ZENEcRing(E));
  ZENEltFree(ZENEcDb(E), ZENEcRing(E));
  ZENEltFree(ZENEcJ(E), ZENEcRing(E));
  ZENEltFree(ZENEcX(E), ZENEcRing(E));
  ZENEltFree(ZENEcY(E), ZENEcRing(E));
  ZENEltFree(ZENEcZ(E), ZENEcRing(E));
  ZENFree(ZENEcXY(E));
  ZENFree(E);
}

