/*
 * zen/zeg/c/ecpt_add_char2.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Elliptic curve points additions}
   
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Last step for adding 2 points in fields of characteristic 2 with non null invariant.}
   \input{Three allocated points |R|, |P| and |Q| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| must be correctly initialized.}
   }
   */
void _ZENEcPtPostAddChar2J1(R, P, Q, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;
{
  ZENEltAssign(ZENEcY(E), ZENEcPtY(Q, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcY(E), ZENEcPtY(P, E), ZENEcRing(E));

  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  
  ZENEltSquare(ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcB2(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcPtX(Q, E), ZENEcRing(E));

  ZENEltAssign(ZENEcPtX(R, E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcPtX(R, E), ZENEcX(E), ZENEcRing(E));

  ZENEltMultiply(ZENEcX(E), ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcPtY(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcPtX(R, E), ZENEcRing(E));
  ZENEltAssign(ZENEcPtY(R, E), ZENEcX(E), ZENEcRing(E)); 
}


/* \TEX{
   \title{First step for doubling 1 point in fields of characteristic 2 with non null invariant.}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| are set.}
   }
   */
void _ZENEcPtPreDoubleChar2J1(P, E)
     ZENEcPt P;
     ZENEc E;
{
  ZENEltAssign(ZENEcY(E), ZENEcPtX(P, E), ZENEcRing(E));
}

/* \TEX{
   \title{Last step for doubling 1 point in fields of characteristic 2 with non null invariant.}
   \input{Two allocated point |R| and |P| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| must be correctly initialized.}
   }
   */
void _ZENEcPtPostDoubleChar2J1(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;
{
  ZENEltMultiply(ZENEcY(E), ZENEcPtY(P, E), ZENEcY(E), ZENEcRing(E));
  ZENEltAdd(ZENEcY(E), ZENEcPtX(P, E), ZENEcRing(E));

  ZENEltAssign(ZENEcX(E), ZENEcPtX(P, E), ZENEcRing(E));

  ZENEltSquare(ZENEcPtX(R,E), ZENEcY(E), ZENEcRing(E));
  ZENEltAdd(ZENEcPtX(R,E), ZENEcY(E), ZENEcRing(E));
  ZENEltAdd(ZENEcPtX(R,E), ZENEcB2(E), ZENEcRing(E));

  ZENEltAdd(ZENEcX(E), ZENEcPtX(R,E), ZENEcRing(E));
  ZENEltMultiply(ZENEcX(E), ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcPtY(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcPtX(R, E), ZENEcRing(E));
  ZENEltAssign(ZENEcPtY(R, E), ZENEcX(E), ZENEcRing(E));  
}

/* \TEX{
   \title{Opposite in fields of characteristic 2 with non null invariant.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|.}
   \side{$|R| = -|P|$.} 
   }
   */
void _ZENEcPtNegateChar2J1(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;
{
  ZENEcPtAssign(R, P, E);
  if (!ZENEcPtIsZero(R, E)) {
    ZENEltAdd(ZENEcPtY(R, E), ZENEcPtX(P, E), ZENEcRing(E));
    ZENEltNegate(ZENEcPtY(R, E), ZENEcPtY(R, E), ZENEcRing(E));
  }
}

/* \TEX{
   \title{Last step for adding 2 points in fields of characteristic 2 with null invariant.}
   \input{Three allocated points |R|, |P| and |Q| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| must be correctly initialized.}
   }
   */
void _ZENEcPtPostAddChar2J0(R, P, Q, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;
{  
  ZENEltAssign(ZENEcY(E), ZENEcPtY(Q, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcY(E), ZENEcPtY(P, E), ZENEcRing(E));

  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcY(E), ZENEcRing(E));  

  ZENEltSquare(ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcPtX(Q, E), ZENEcRing(E));

  ZENEltAssign(ZENEcPtX(R, E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcPtX(R, E), ZENEcX(E), ZENEcRing(E));

  ZENEltMultiply(ZENEcX(E), ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcPtY(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcB3(E), ZENEcRing(E));
  ZENEltAssign(ZENEcPtY(R, E), ZENEcX(E), ZENEcRing(E));
}

/* \TEX{
   \title{First step for doubling 1 point in fields of characteristic 2 with null invariant.}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| are set.}
   }
   */
void _ZENEcPtPreDoubleChar2J0(P, E)
     ZENEcPt P;
     ZENEc E;
{
  ZENEltAssign(ZENEcY(E), ZENEcB3(E), ZENEcRing(E));
}

/* \TEX{
   \title{Last step for doubling 1 point in fields of characteristic 2 with null invariant.}
   \input{Two allocated point |R| and |P| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| must be correctly initialized.}
   }
   */
void _ZENEcPtPostDoubleChar2J0(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;
{
  ZENEltSquare(ZENEcX(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcB4(E), ZENEcRing(E));
  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  
  ZENEltAssign(ZENEcX(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltSquare(ZENEcPtX(R,E), ZENEcY(E), ZENEcRing(E));

  ZENEltAdd(ZENEcX(E), ZENEcPtX(R,E), ZENEcRing(E));
  ZENEltMultiply(ZENEcX(E), ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcPtY(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcB3(E), ZENEcRing(E));
  ZENEltAssign(ZENEcPtY(R, E), ZENEcX(E), ZENEcRing(E));
}

/* \TEX{
   \title{Opposite in fields of characteristic 2 with null invariant.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|.}
   \side{$|R| = -|P|$.} 
   }
   */
void _ZENEcPtNegateChar2J0(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;
{
  ZENEcPtAssign(R, P, E);

  if (!ZENEcPtIsZero(R, E)) {
    ZENEltAdd(ZENEcPtY(R, E), ZENEcB3(E), ZENEcRing(E));
    ZENEltNegate(ZENEcPtY(R, E), ZENEcPtY(R, E), ZENEcRing(E));
  }
}
