/*
 * zen/zeg/c/ecpt_add_charp.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Elliptic curve points additions in fields of characteristic greater than 3 }
   
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{First step for doubling 1 point in fields of characteristic greater than 3.}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| are set.}
   }
   */
void _ZENEcPtPreDoubleCharp(P, E)
     ZENEcPt P;
     ZENEc E;
{
  ZENEltAssign(ZENEcY(E), ZENEcPtY(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcY(E), ZENEcPtY(P, E), ZENEcRing(E));
}

/* \TEX{
   \title{Last step for adding 2 points in finite fields of charateristic greater than 3.}
   \input{Three allocated points |R|, |P| and |Q| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| must be correctly initialized.}
   }
   */
void _ZENEcPtPostAddCharp(R, P, Q, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;
{
  ZENEltAssign(ZENEcY(E), ZENEcPtY(Q, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcY(E), ZENEcPtY(P, E), ZENEcRing(E));

  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcY(E), ZENEcRing(E));    

  ZENEltSquare(ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  ZENEltSubtract(ZENEcX(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcX(E), ZENEcPtX(Q, E), ZENEcRing(E));
  
  ZENEltAssign(ZENEcPtX(R, E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcPtX(R, E), ZENEcX(E), ZENEcRing(E));
  ZENEltMultiply(ZENEcPtX(R, E), ZENEcY(E), ZENEcPtX(R, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcPtX(R, E), ZENEcPtY(P, E), ZENEcRing(E));
  
  ZENEltAssign(ZENEcPtY(R, E), ZENEcPtX(R, E), ZENEcRing(E));
  ZENEltAssign(ZENEcPtX(R, E), ZENEcX(E), ZENEcRing(E));
}

/* \TEX{
   \title{Last step for doubling 1 point in finite fields of charateristic greater than 3.}
   \input{Two allocated point |R| and |P| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| must be correctly initialized.}
   }
   */
void _ZENEcPtPostDoubleCharp(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;
{
  ZENEltSquare(ZENEcZ(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAssign(ZENEcX(E), ZENEcZ(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcZ(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcZ(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcB4(E), ZENEcRing(E));

  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcY(E), ZENEcRing(E));
    
  ZENEltSquare(ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  ZENEltSubtract(ZENEcX(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcX(E), ZENEcPtX(P, E), ZENEcRing(E));
  
  ZENEltAssign(ZENEcPtX(R, E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcPtX(R, E), ZENEcX(E), ZENEcRing(E));
  ZENEltMultiply(ZENEcPtX(R, E), ZENEcY(E), ZENEcPtX(R, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcPtX(R, E), ZENEcPtY(P, E), ZENEcRing(E));

  ZENEltAssign(ZENEcPtY(R, E), ZENEcPtX(R, E), ZENEcRing(E));
  ZENEltAssign(ZENEcPtX(R, E), ZENEcX(E), ZENEcRing(E));
}

/* \TEX{
   \title{Opposite in fields of characteristic graeter than 3.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|.}
   \side{$|R| = -|P|$.} 
   }
   */
void _ZENEcPtNegateCharp(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;
{
  ZENEcPtAssign(R, P, E);
  if (!ZENEcPtIsZero(R, E)) {
    ZENEltNegate(ZENEcPtY(R, E), ZENEcPtY(R, E), ZENEcRing(E));
  }
}
