/*
 * zen/zeg/c/ecpt_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Points on elliptic curves tests}
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Is a point equal to another point ?}
   \input{Two allocated points |P| and |Q| of an elliptic curve |E|.}
   \output{The predicate $|P|=|Q|$}
   }
   */
int _ZENEcPtAreEqual(P, Q, E)
     ZENEcPt P, Q;
     ZENEc E;
{
  if (!ZENEcPtZ(P, E) && !ZENEcPtZ(Q, E)) {
    return(1);
  }
  return((ZENEcPtZ(P, E) == ZENEcPtZ(Q, E)) &&
	 (ZENEltAreEqual(ZENEcPtX(P, E), ZENEcPtX(Q, E), ZENEcRing(E))) &&
	 (ZENEltAreEqual(ZENEcPtY(P, E), ZENEcPtY(Q, E), ZENEcRing(E))));
}

/* \TEX{
   \title{Is a point on the curve ?}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \output{The predicate $|P|\in|E|$}
   }
   */
int _ZENEcPtIsOnEc(P, E)
     ZENEcPt P;
     ZENEc E;
{
  if (ZENEcPtIsZero(P, E)) {
    return(1);
  }

  ZENEltSquare(ZENEcY(E), ZENEcPtY(P, E), ZENEcRing(E));
  
  ZENEltMultiply(ZENEcX(E),
		 ZENEcPtX(P, E),
		 ZENEcPtY(P, E),
		 ZENEcRing(E));
  ZENEltMultiply(ZENEcX(E),
		 ZENEcA1(E),
		 ZENEcX(E),
		 ZENEcRing(E));
  ZENEltAdd(ZENEcY(E), ZENEcX(E), ZENEcRing(E));

  ZENEltMultiply(ZENEcX(E),
		 ZENEcA3(E),
		 ZENEcPtY(P, E),
		 ZENEcRing(E));
  ZENEltAdd(ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  
  ZENEltMultiply(ZENEcX(E),
		 ZENEcPtX(P, E),
		 ZENEcPtX(P, E),
		 ZENEcRing(E));
  ZENEltMultiply(ZENEcX(E),
		 ZENEcPtX(P, E),
		 ZENEcX(E),
		 ZENEcRing(E));
  ZENEltSubtract(ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  
  ZENEltSquare(ZENEcX(E),
	       ZENEcPtX(P, E),
	       ZENEcRing(E));
  ZENEltMultiply(ZENEcX(E),
		 ZENEcA2(E),
		 ZENEcX(E),
		 ZENEcRing(E));
  ZENEltSubtract(ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  
  ZENEltMultiply(ZENEcX(E),
		 ZENEcA4(E),
		 ZENEcPtX(P, E),
		 ZENEcRing(E));
  ZENEltSubtract(ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  
  ZENEltSubtract(ZENEcY(E), ZENEcA6(E), ZENEcRing(E));
  
  return(ZENEltIsZero(ZENEcY(E), ZENEcRing(E)));
}

/* \TEX{
   \title{Is a point on the curve ?}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \output{The predicate $|P|\in|E|$}
   }
   */
int _ZENEcPtIsOnEcChar23p(P, E)
     ZENEcPt P;
     ZENEc E;
{
  if (ZENEcPtIsZero(P, E)) {
    return(1);
  }

  ZENEltSquare(ZENEcY(E), ZENEcPtY(P, E), ZENEcRing(E));
  
  ZENEltMultiply(ZENEcX(E),
		 ZENEcPtX(P, E),
		 ZENEcPtY(P, E),
		 ZENEcRing(E));
  ZENEltMultiply(ZENEcX(E),
		 ZENEcB1(E),
		 ZENEcX(E),
		 ZENEcRing(E));
  ZENEltAdd(ZENEcY(E), ZENEcX(E), ZENEcRing(E));

  ZENEltMultiply(ZENEcX(E),
		 ZENEcB3(E),
		 ZENEcPtY(P, E),
		 ZENEcRing(E));
  ZENEltAdd(ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  
  ZENEltMultiply(ZENEcX(E),
		 ZENEcPtX(P, E),
		 ZENEcPtX(P, E),
		 ZENEcRing(E));
  ZENEltMultiply(ZENEcX(E),
		 ZENEcPtX(P, E),
		 ZENEcX(E),
		 ZENEcRing(E));
  ZENEltSubtract(ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  
  ZENEltSquare(ZENEcX(E),
	       ZENEcPtX(P, E),
	       ZENEcRing(E));
  ZENEltMultiply(ZENEcX(E),
		 ZENEcB2(E),
		 ZENEcX(E),
		 ZENEcRing(E));
  ZENEltSubtract(ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  
  ZENEltMultiply(ZENEcX(E),
		 ZENEcB4(E),
		 ZENEcPtX(P, E),
		 ZENEcRing(E));
  ZENEltSubtract(ZENEcY(E), ZENEcX(E), ZENEcRing(E));
  
  ZENEltSubtract(ZENEcY(E), ZENEcB6(E), ZENEcRing(E));
  
  return(ZENEltIsZero(ZENEcY(E), ZENEcRing(E)));
}

/* \TEX{
   \title{Is a point equal to zero ?}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \output{The predicate $|P|=0$}
   }
   */
int _ZENEcPtIsZero(P, E)
     ZENEcPt P;
     ZENEc E;
{
  return(ZENEcPtZ(P, E) ? 0 : 1);
}
