/*
 * zen/zeg/c/elt_trace.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Absolute trace}
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{``Absolute Trace''.}
   \input{Two allocated elements |e| and |f| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_HAS_INVERSE| otherwise.}
   \side{$|f| = \mbox{Tr}_{|Rg|/(Z/pZ)}(|e|)$ where $p$ is the
   characteristic of |Rg|. If |Rg|=$\Z/p\Z$, $|f|=|e|$.}
   }
   */
int _ZENEltAbsoluteTrace(e, f, Rg)
     ZENElt *e, f;
     ZENRing Rg;
{
  ZENElt t, u;
  int deg;
  ZENRing rg, last_rg;
  BigNum n;
  BigNumLength nl;
  
  if (ZENEltCreate(t, Rg)) {
    return(ZERR);
  }
  ZENEltTrace(t, f, Rg);
  deg = 1;
  for(last_rg = Rg, rg = ZENRingDef(last_rg);
      rg;
      last_rg = rg, rg = ZENRingDef(last_rg)) {

    deg *= ZENRingDeg(last_rg);
    if (!(n = ZBNC(ZENRingQl(last_rg)))) {
      ZENEltFree(t, last_rg);
      return(ZERR);
    }
    ZENEltToZ(n, &nl, t, last_rg);
    
    if (ZBNCompareLazy(n, nl, ZENRingQ(rg), ZENRingQl(rg)) != BN_LT ) {
      ZENEltFree(t, last_rg);
      ZBNF(n);
      return(ZERR);
    }
    ZENEltFree(t, last_rg);
    if (ZENEltCreate(u, rg)) {
      ZBNF(n);
      return(ZERR);
    }
    ZENZToElt(u, n, nl, rg);
    ZBNF(n);
    
    ZENEltTrace(u, u, rg);
    t = u;
  }
  
  if (!(n = ZBNC(ZENRingQl(last_rg)))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_UNKNOWN, ZEN_ALLOC);
    ZENEltFree(t, last_rg);
    return(ZERR);
  }
  ZENEltToZ(n, &nl, t, last_rg);
  ZENZToElt(*e, n, nl, Rg);
  ZBNF(n);
  ZENEltFree(t, last_rg);

  return(0);
}
