/*
 * zen/zeg/c/mat_create.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Allocating a matrix}
   }
   */


#include <stdio.h>
#include "zen.h"
/* \TEX{
   \title{Allocating a matrix.}
   \input{The dimensions of the desired matrix and a |ZENRing|.}
   \output{NULL if an error occured, a |ZENMat| of the desired dimensions otherwise.}
   \note{The type of the resulting matrix minimize the memory allocated. This function is always called. It calls the two pointer functions |ZENMatCreateParams| and |_R_ZENMatCreate|.}
    }
  */   
ZENMat _ZENMatCreate(l,c,R)
     Dim l,c;
     ZENRing R;
{
  ZENMat pms1,pms2;

  ZENCheckError(ZENMatCreateParams(ZENMatTypeRow,l,c,&pms1,R),NULL);
  if(ZENMatNbBnd(pms1,R)==0)
    {
      ZENCheckError(_R_ZENMatCreate(pms1,R),NULL);
      return(pms1);
    }

  ZENCheckError(ZENMatCreateParams(ZENMatTypeCol,l,c,&pms2,R),NULL);
  if(ZENMatNbBnd(pms1,R)>ZENMatNbBnd(pms2,R))
    {
      ZENFree(pms1);
      ZENCheckError(_R_ZENMatCreate(pms2,R),NULL);
      return(pms2);
    }
  else
    {
      ZENFree(pms2);
      ZENCheckError(_R_ZENMatCreate(pms1,R),NULL);
      return(pms1);
    }
}
 
/* \TEX[
   \title{Copying a matrix.}
   \input{A |ZENMat| and a |ZENRing|.}
   \output{NULL if an error occured, a new allocated |ZENMat| equal to the initial one.}
    }
  */   
ZENMat _ZENMatCopy(M,R)
     ZENMat M;
     ZENRing R;
{
  ZENMat pms;

  if(ZENMatIsPermutation(M,R))
    {
      ZENCheckError((pms=ZENPermutationCreate(ZENMatNbRow(M,R),ZENMatType(M,R),R))==NULL,NULL);
      ZBNAssign(ZENMatFirst(pms,R),ZENMatFirst(M,R),ZENMatNbBnd(M,R));
    }
  else
    {
      ZENCheckError(ZENMatCreateParams(ZENMatType(M,R),
				       ZENMatNbRow(M,R),ZENMatNbCol(M,R),
				       &pms,R),NULL);
      ZENCheckError(_R_ZENMatCreate(pms,R),NULL);
      ZENCheckError(ZENMatAssign(pms,M,R),NULL);
    }
  return(pms);
}

/* \TEX[
   \title{Allocating a matrix of type |ZENMatTypeRow|}
   \input{The dimensions of the desired matrix and a |ZENRing|.}
   \output{NULL if an error occured, a |ZENMat| of the desired dimensions otherwise.}
   }
   */   
ZENMat _ZENMatRowCreate(l,c,R)
     Dim l,c;
     ZENRing R;
{
  ZENMat pms;
  
  ZENCheckError(ZENMatCreateParams(ZENMatTypeRow,l,c,&pms,R),NULL);
  ZENCheckError(_R_ZENMatCreate(pms,R),NULL);
  return(pms);
}
 

/* \TEX[
   \title{Allocating a matrix of type |ZENMatTypeCol|}
   \input{The dimensions of the desired matrix and a |ZENRing|.}
   \output{NULL if an error occured, a |ZENMat| of the desired dimensions otherwise.}
   }
   */   
ZENMat _ZENMatColCreate(l,c,R)
     Dim l,c;
     ZENRing R;
{
  ZENMat pms;
  
  ZENCheckError(ZENMatCreateParams(ZENMatTypeCol,l,c,&pms,R),NULL);
  ZENCheckError(_R_ZENMatCreate(pms,R),NULL);
  return(pms);
}
 
