/*
 * zen/zeg/c/mat_matrix.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial operations}

   Due to the structure of a |ZENMat|, some operations have to be chosen at initialisation of the matrix according to its type. Hence repetition functions of the |ZENRing| functions are necessary.
   
   }
   */
#include <stdio.h>
#include "zen.h"

/*
   \TEX[
   \title{Assignment.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int __ZENMatAssign(A,B,R)
     ZENMat A,B;
     ZENRing R;
{
  return(_R_ZENMatAssign(A,B,R));
}

/*
   \TEX[
   \title{Set to one}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are set to zero except for the coefficients with same indices on rows and columns that are set to one.}
   \note{There is no error if the matrix is not square.}
   }
   */
   
int __ZENMatSetToOne(A,R)
     ZENMat A;
     ZENRing R;
{
  return(_R_ZENMatSetToOne(A,R));
}

/*
   \TEX[
   \title{Equality to identity}
   \input{A |ZENMat| and a |ZENRing|}
   \output{|M == |Identity.}
   \note{The matrix is checked to be square.}
   }
   */

int __ZENMatIsOne(A,R)
     ZENMat A;
     ZENRing R;
{
  return(_R_ZENMatIsOne(A,R));
}
/*
   \TEX[
   \title{Set to random}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are randomly set.}
   }
   
   */
   
int __ZENMatSetRandom(A,R)
     ZENMat A;
     ZENRing R;
{
  return(_R_ZENMatSetRandom(A,R));
}

/*
   \TEX[
   \title{Get a bloc}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed bloc of |A|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
int __ZENMatGetBloc(p_E,A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENElt *p_E;
     ZENRing R;
{
  return(ZENMatGetBloc(*p_E,A,l,c,R));
}
/*
   \TEX[
   \title{Get a bloc pointer}
   \input{A |ZENMat| |A|, two indexes and a finite ring |R|.}
   \output{The |ZENElt| indexed bloc of |A|}
   \note{Except for advanced user, |GetCoeffPtr| functions should be preferred}
   }
   
   */
   
ZENElt __ZENMatGetBlocPtr(A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENRing R;
{
  return(ZENMatGetBlocPtr(A,l,c,R));
}
/*
   \TEX[
   \title{Set a bloc}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The indexed bloc of |A| is set to |E|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
int __ZENMatSetBloc(A,l,c,E,R)
     ZENMat A;
     Dim l,c;
     ZENElt E;
     ZENRing R;
{
  return(ZENMatSetBloc(A,l,c,E,R));
}

/*
   \TEX[
   \title{Get a digit}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed digit of |A|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
int __ZENMatGetDigit(p_E,A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENElt *p_E;
     ZENRing R;
{
  return(ZENMatGetDigit(*p_E,A,l,c,R));
}

/*
   \TEX[
   \title{Get a bloc pointer}
   \input{A |ZENMat| |A|, two indexes and a finite ring |R|.}
   \output{The |ZENElt| indexed digit of |A|}
   \note{Except for advanced user, |GetCoeffPtr| functions should be preferred}
   }
   
   */
   
ZENElt __ZENMatGetDigitPtr(A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENRing R;
{
  return(ZENMatGetDigitPtr(A,l,c,R));
}

/*
   \TEX[
   \title{Set a bloc}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The indexed bloc of |A| is set to |E|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
int __ZENMatSetDigit(A,l,c,E,R)
     ZENMat A;
     Dim l,c;
     ZENElt E;
     ZENRing R;
{  
  return(ZENMatSetDigit(A,l,c,E,R));
}

/*
   \TEX[
   
   \title{Permute two |Bn|s}
   \input{A |ZENMat| |A|, two indexes and a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The specified |Bns| of |A| are exchanged}
   \note{Except for advanced user, |PermuteRow| and |PermuteCol| functions should be preferred}
   }
   
   */
   
int __ZENMatPermuteBn(A,l1,l2,R)
     ZENMat A;
     Dim l1,l2;
     ZENRing R;
{
  return(_R_ZENMatPermuteBn(A,l1,l2,R));
}

/*
   \TEX[
   \title{Permute two |digit|s}
   \input{A |ZENMat| |A|, two indexes and a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The specified |digits| of |A| are exchanged}
   \note{Except for advanced user, |PermuteRow| and |PermuteCol| functions should be preferred}
   }   
   */
   

int __ZENMatPermuteDigit(A,l1,l2,R)
     ZENMat A;
     Dim l1,l2;
     ZENRing R;
{
  return(ZENMatPermuteDigit(A,l1,l2,R));
}
