/*
 * zen/zeg/c/mat_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   
   \Subsection{Tests on a matrix}
   }
   */


#include <stdio.h>
#include "zen.h"
/* \TEX[
   \title{Test of equality}
   \input{Two matrices and a |ZENRing|}
   \output{1 if the matrices are equal, 0 otherwise}
   }
   */

int _ZENMatAreEqual(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;
 
 if(ZENMatAreSameType(A,B,R))
   {
     if((ZENMatNbBn(A,R) != ZENMatNbBn(B,R))
	||(ZENMatNbDigit(A,R) != ZENMatNbDigit(B,R)))
       return(0);
     for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZENMatNbBloc(A,R)-1;c>=0;c--)
	 if(!(ZENEltAreEqual(ZENMatGetBlocPtr(A,l,c,R),ZENMatGetBlocPtr(B,l,c,R),R)))
	   return(0);
     return(1);
   }
 else
   {
     if((ZENMatNbBn(A,R) != ZENMatNbDigit(B,R))
	||(ZENMatNbDigit(A,R) != ZENMatNbBn(B,R)))
       return(0);
     for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZENMatNbBloc(A,R)-1;c>=0;c--)
	 if(!(ZENEltAreEqual(ZENMatGetBlocPtr(A,l,c,R),ZENMatGetBlocPtr(B,c,l,R),R)))
	   return(0);
     return(1);
   }
  
}

/* \TEX[
   \title{Test of equality to zero}
   \input{A matrix and a |ZENRing|}
   \output{1 if the matrix is null, 0 otherwise}
   }
   */

int _ZENMatIsZero(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;
 
 for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZENMatNbBloc(A,R)-1;c>=0;c--)
     if(!(ZENEltIsZero(ZENMatGetBlocPtr(A,l,c,R),R)))
       return(0);
 return(1);
}

/* \TEX[
   \title{Test of equality to identity}
   \input{A matrix and a |ZENRing|}
   \output{1 if the matrix is identity, 0 otherwise}
   \note{The matrix is checked to be square.} 
   }
   */

int _ZENMatIsOne(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;

 if(ZENMatNbBn(A,R) != ZENMatNbDigit(A,R))
   return(0);
 
 for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZENMatNbBn(A,R)-1;c>=0;c--)
     if(l==c)
       {
	 if(!(ZENEltIsOne(ZENMatGetCoeffPtr(A,l,c,R),R)))
	   return(0);
       }
     else
       {
	 if(!(ZENEltIsZero(ZENMatGetCoeffPtr(A,l,c,R),R)))
	   return(0);
       }
 return(1); 
}


