/*
 * zen/zeg/c/permutation_ops.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Permutation operations}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zen.h"
#include "zeg.h"

/*
   \TEX{
   \title{Multiplication.}
   \input{Three permutations matrix |X|, |A| and |B|.}
   \output{|ZENERR| if an error occurred, 0 otherwise}
   \side{$|X|= |A| \times |B|$.}
   }
   */
   
int _ZENMatMultiplyPermutations(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{
  char cas;

  ZENTest(!((ZENMatIsPermutation(X,R))
	    &&(ZENMatIsPermutation(A,R))
	    &&(ZENMatIsPermutation(B,R))),
	  ZEN_ISPERM,ZEN_MATMULPERM,ZEN_LIB_ZEG, ZERR);
  
  ZENTest((ZENMatNbRow(A,R) != ZENMatNbRow(X,R))
	  ||(ZENMatNbCol(A,R) != ZENMatNbRow(B,R))
	  ||(ZENMatNbCol(B,R) != ZENMatNbCol(X,R)),
	  ZEN_MAT_DIM,ZEN_MATMULTIPLY,ZEN_LIB_ZEG, ZERR);
  cas=(ZENMatIsRowType(X,R)<<2)|(ZENMatIsRowType(A,R)<<1)|(ZENMatIsRowType(B,R));
  
  switch(cas)
    {
    case _CCC_ :
      ZENMatMultiplyPermCCC(X,A,B,R);
      break;
    case _CCR_ :
      ZENMatMultiplyPermCCR(X,A,B,R);
      break;
    case _CRC_ :
      ZENMatMultiplyPermCRC(X,A,B,R);
      break;
    case _CRR_ :
      ZENMatMultiplyPermCRR(X,A,B,R);
      break;
    case _RCC_ :
      ZENMatMultiplyPermRCC(X,A,B,R);
      break;
    case _RCR_ :
      ZENMatMultiplyPermRCR(X,A,B,R);
      break;
    case _RRC_ :
      ZENMatMultiplyPermRRC(X,A,B,R);
      break;
    case _RRR_ :
      ZENMatMultiplyPermRRR(X,A,B,R);
      break;
    }
  return(0);
}

/*
   \TEX{
   \title{Multiplication.}
   \input{Three |ZENMat| |X|, |A| and |B| of $\Z/2\Z$ of types |R|, |R| and |R|.}
   \side{$|X|= |A| \times |B|$.}
   }
   */

void _ZENMatMultiplyPermRRR(X,A,B,R) /* CCC CRC RRC */
     ZENMat X,A,B;
     ZENRing R;
{
  int l;

  for(l=ZENMatNbBn(X,R)-1;l>=0;l--)
    ZENPermutationBn(X,l,R)=ZENPermutationBn(B,ZENPermutationBn(A,l,R),R);
}

/*
   \TEX{
   \title{Multiplication.}
   \input{Three |ZENMat| |X|, |A| and |B| of $\Z/2\Z$ of types |R|, |C| and |C|.}
   \side{$|X|= |A| \times |B|$.}
   }
   */

void _ZENMatMultiplyPermRCC(X,A,B,R) /* CRR */
     ZENMat X,A,B;
     ZENRing R;
{
  int c;
  
  for(c=ZENMatNbBn(B,R)-1;c>=0;c--)
    ZENPermutationBn(X,ZENPermutationBn(A,ZENPermutationBn(B,c,R),R),R)=(Dim)c;
}

/*
   \TEX{
   \title{Multiplication.}
   \input{Three |ZENMat| |X|, |A| and |B| of $\Z/2\Z$ of types |C|, |C| and |R|.}
   \side{$|X|= |A| \times |B|$.}
   }
   */
void _ZENMatMultiplyPermCCR(X,A,B,R) /* RCR */
     ZENMat X,A,B;
     ZENRing R;
{
  int l;

  for(l=ZENMatNbBn(B,R)-1;l>=0;l--)
    ZENPermutationBn(X,ZENPermutationBn(B,l,R),R)=ZENPermutationBn(A,l,R);
}
