/*
 * zen/zeg/c/poly_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial additions}
   
   One function is provided to do additions: |_ZENPolyAdd| to add a
   |ZENPoly| to another.
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Addition.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| += |PX|.}
   }
   */
void _ZENPolyAdd(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  ZENElt elt;
  int mindeg, i;

  if (ZENPolyDeg(RX, Rg) < ZENPolyDeg(PX, Rg)) {
    mindeg = ZENPolyDeg(RX, Rg);
    ZENPolyDeg(RX, Rg) =  ZENPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZENPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    elt = ZENPolyGetCoeffPtr(RX, i, Rg);
    ZENEltAdd(elt, ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
    ZENPolySetCoeff(RX, i, elt, Rg);
  }
  for(i = mindeg+1; i <= ZENPolyDeg(PX, Rg); i++) {
    ZENPolySetCoeff(RX, i, ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
  }
  ZENPolyUpdateDegree(RX, Rg);
}

/* \TEX{
   \title{Negation.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| = -|PX|.}
   }
   */
void _ZENPolyNegate(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  ZENElt elt;
  int i;

  for(i = 0; i <= ZENPolyDeg(PX, Rg); i++) {
    elt = ZENPolyGetCoeffPtr(RX, i, Rg);
    ZENEltNegate(elt, ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
    ZENPolySetCoeff(RX, i, elt, Rg);
  }
  ZENPolyDeg(RX, Rg) =  ZENPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Subtraction.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| -= |PX|.}
   }
   */
void _ZENPolySubtract(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  ZENElt elt;
  int i, mindeg;

  if (ZENPolyDeg(RX, Rg) < ZENPolyDeg(PX, Rg)) {
    mindeg = ZENPolyDeg(RX, Rg);
    ZENPolyDeg(RX, Rg) =  ZENPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZENPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    elt = ZENPolyGetCoeffPtr(RX, i, Rg);
    ZENEltSubtract(elt, ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
    ZENPolySetCoeff(RX, i, elt, Rg);
  }
  for(i = mindeg+1; i <= ZENPolyDeg(PX, Rg); i++) {
    elt = ZENPolyGetCoeffPtr(RX, i, Rg);
    ZENEltNegate(elt, ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
    ZENPolySetCoeff(RX, i, elt, Rg);
  }
  ZENPolyUpdateDegree(RX, Rg);
}
