/*
 * zen/zeg/c/poly_derivative.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Derivation of a polynomial}
   } 
   */
#include <stdio.h>
#include "zen.h"

/*
   \TEX{
   \title{Compute the derivative of a polynomial}
   \input{A pointer on a |ZENPoly| and a |ZENPoly| over a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{$|D|(X)=|P|'(X)$.}
   \note{One can have |D|=|P|.}
   } 
   */
int _ZENPolyDerive(D,P,R)
     ZENPoly D,P;
     ZENRing R;
{int deg,dg;
 BigNumDigit dm;
 ZENElt cf,elt;

 deg=ZENPolyDeg(P,R);
 if(deg==0) {
   ZENPolySetToZero(D,R);
   return(0);
 }
    
 if(ZENEltCreate(elt,R))
   return(ZERR);
 if(ZENEltCreate(cf,R))
   return(ZERR);


 dm=1;
 dg=1;
 ZENPolyDeg(D,R)=0;
 do {
   if((ZENRingPl(R)==1)
      &&(dm == *(ZENRingP(R))))
     dm=0;

   if(dm)
     {
       ZENZToElt(elt,&dm,1,R);
       ZENEltMultiply(cf,elt,ZENPolyGetCoeffPtr(P,dg,R),R);
       ZENPolySetCoeff(D,dg-1,cf,R);
       if(!ZENEltIsZero(cf,R))
	 ZENPolyDeg(D,R)=dg-1;
     }
   else
     {
       ZENZToElt(elt,&dm,1,R);
       ZENPolySetCoeff(D,dg-1,elt,R);
     }

   dm++;
   dg++;
 } while (dg <= deg);

 ZENEltFree(elt,R);
 ZENEltFree(cf,R);
 return(0);
}
