/*
 * zen/zeg/c/poly_resultant.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial resultant}

   }
   */
#include <stdio.h>
#include "zen.h"

/* Used by ZENPolyResultant. From P. Gaudry. */
static int ZENPolyPseudoDivide(Q, R, A, B, Rg)
     ZENPoly Q, R, A, B;
     ZENRing Rg;
{
  ZENPoly QQ, RR, SS;
  ZENElt aux, d;
  BigNumDigit e;
  
  if (ZENPolyIsZero(B, Rg)) 
    return(ZERR);

  if (!(RR=ZENPolyCopy(A, Rg))) {
    return(ZERR);
  }
  if (ZENPolyAlloc(QQ, ZENPolyDeg(A, Rg), Rg)) {
    ZENPolyFree(RR, Rg);
    return(ZERR);
  }
  if (ZENPolyAlloc(SS, ZENPolyDeg(A, Rg), Rg)) {
    ZENPolyFree(RR, Rg); ZENPolyFree(QQ, Rg);
    return(ZERR);
  }
  if (ZENEltAlloc(aux, Rg)) {
    ZENPolyFree(RR, Rg); ZENPolyFree(QQ, Rg); ZENPolyFree(SS, Rg);
    return(ZERR);
  }
  if (ZENEltAlloc(d, Rg)) {
    ZENPolyFree(RR, Rg); ZENPolyFree(QQ, Rg); ZENPolyFree(SS, Rg);
    ZENEltFree(aux, Rg);
    return(ZERR);
  }

  ZENPolySetToZero(QQ, Rg);
  e=ZENPolyDeg(A, Rg)-ZENPolyDeg(B, Rg)+1;
  ZENEltAssign(d,ZENPolyGetCoeffPtr(B, ZENPolyDeg(B, Rg), Rg), Rg); 
  while (ZENPolyDeg(RR, Rg) >= ZENPolyDeg(B, Rg)) {
    ZENPolySetToXi(SS, ZENPolyDeg(RR, Rg)-ZENPolyDeg(B, Rg), Rg);
    ZENEltAssign(aux, ZENPolyGetCoeffPtr(RR, ZENPolyDeg(RR, Rg), Rg), Rg);
    ZENPolySetCoeff(SS, ZENPolyDeg(RR, Rg)-ZENPolyDeg(B, Rg), aux, Rg);
    ZENPolyMultiplyScalar(QQ, QQ, d, Rg);
    ZENPolyAdd(QQ, SS, Rg);
    ZENPolyMultiplyScalar(RR, RR, d, Rg);
    ZENPolyMultiply(SS, B, SS, Rg);
    ZENPolySubtract(RR, SS, Rg);
    e--;
  }
  ZENEltExp(d, &e, 1, d, Rg);
  ZENPolyMultiplyScalar(Q, QQ, d, Rg);
  ZENPolyMultiplyScalar(R, RR, d, Rg);

  ZENPolyFree(QQ, Rg); ZENPolyFree(RR, Rg); ZENPolyFree(SS, Rg);
  ZENEltFree(aux, Rg); ZENEltFree(d, Rg);

  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Resultant of 2 polynomials.}
   \input{Three polynomials |*Res|, |A| and |B| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{|*Res|$(X) = resultant($|A|$(X), $|B|$(X))$ if |ZEN_HAS_INVERSE| is
   returned,  |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   \note{Computation of the resultant of two polynomials by the
   sub-resultant algorithm. From P. Gaudry.}
   }
   */
int _ZENPolyResultant(Res, A, B, Rg)
     ZENElt *Res;
     ZENPoly A,B;
     ZENRing Rg;
{
  ZENPoly AA, BB, R;
  ZENElt g, h, aux;
  int s, d, i, ret;

  if (ZENPolyIsZero(A, Rg) || ZENPolyIsZero(B, Rg)) {
    ZENEltSetToZero(*Res, Rg);
    return(ZEN_HAS_INVERSE);
  }
  s=1;
  if (ZENPolyDeg(A, Rg)<ZENPolyDeg(B, Rg)) {
    if (!(AA=ZENPolyCopy(B, Rg))) {
      return(ZERR);
    }
    if (!(BB=ZENPolyCopy(A, Rg))) {
      ZENPolyFree(AA, Rg);
      return(ZERR);
    }
    if ((ZENPolyDeg(AA, Rg)%2 == 1) && (ZENPolyDeg(BB, Rg)%2 == 1)) {s=-1;}
  } else {
    if (!(AA=ZENPolyCopy(A, Rg))) {
      return(ZERR);
    }
    if (!(BB=ZENPolyCopy(B, Rg))) {
      ZENPolyFree(AA, Rg);
      return(ZERR);
    }
  }
  if (ZENPolyAlloc(R, ZENPolyDeg(AA, Rg), Rg)) {
    ZENPolyFree(AA, Rg); ZENPolyFree(BB, Rg);
  }
  if (ZENEltAlloc(g, Rg)) {
    ZENPolyFree(AA, Rg); ZENPolyFree(BB, Rg); ZENPolyFree(R, Rg);
  }
  if (ZENEltAlloc(h, Rg)) {
    ZENPolyFree(AA, Rg); ZENPolyFree(BB, Rg); ZENPolyFree(R, Rg);
    ZENEltFree(g, Rg);
  }
  if (ZENEltAlloc(aux, Rg)) {
    ZENPolyFree(AA, Rg); ZENPolyFree(BB, Rg); ZENPolyFree(R, Rg);
    ZENEltFree(g, Rg); ZENEltFree(h, Rg);
  }

  ZENEltSetToOne(g, Rg);
  ZENEltSetToOne(h, Rg);
  while (ZENPolyDeg(BB, Rg)>0) {
    d=ZENPolyDeg(AA, Rg)-ZENPolyDeg(BB, Rg);
    if ((ZENPolyDeg(AA, Rg)%2 == 1) && (ZENPolyDeg(BB, Rg)%2 == 1))
    s=-s;
    ZENPolyPseudoDivide(AA, R, AA, BB, Rg);
    ZENPolyAssign(AA, BB, Rg);
    ZENPolyAssign(BB, R, Rg);
    ZENEltAssign(aux, h, Rg);
    for (i=1; i<d; i++) 
      ZENEltMultiply(aux, h, aux, Rg);
    ZENEltMultiply(aux, g, aux, Rg);
    if ((ret = ZENEltInverse(aux, aux, Rg)) != ZEN_HAS_INVERSE) {
      goto end;
    }
    ZENPolyMultiplyScalar(BB, BB, aux, Rg);
    ZENEltAssign(g, ZENPolyGetCoeffPtr(AA, ZENPolyDeg(AA, Rg), Rg), Rg);
    ZENEltSetToOne(aux, Rg);
    for (i=1; i<d; i++)
      ZENEltMultiply(aux, h, aux, Rg);
    if ((ret = ZENEltInverse(aux, aux, Rg)) != ZEN_HAS_INVERSE) {
      goto end;
    }
    for (i=1; i<=d; i++) 
      ZENEltMultiply(aux, g, aux, Rg);
    ZENEltAssign(h, aux, Rg);
  }
  d=ZENPolyDeg(AA, Rg);
  ZENEltSetToOne(aux, Rg);
  for (i=1; i<d; i++)
    ZENEltMultiply(aux, h, aux, Rg);
    if ((ret = ZENEltInverse(aux, aux, Rg)) != ZEN_HAS_INVERSE) {
      goto end;
    }
  ZENEltAssign(g, ZENPolyGetCoeffPtr(BB, ZENPolyDeg(BB, Rg), Rg), Rg);
  for (i=1; i<=d; i++) 
    ZENEltMultiply(aux, g, aux, Rg);
  if (s==-1)
    ZENEltNegate(aux, aux, Rg);
  ZENEltAssign(*Res, aux, Rg);

  ret = ZEN_HAS_INVERSE;

 end:
  ZENEltFree(g, Rg); ZENEltFree(h, Rg); ZENEltFree(aux, Rg);
  ZENPolyFree(AA, Rg); ZENPolyFree(BB, Rg); ZENPolyFree(R, Rg);
  return(ret);
}

