/*
 * zen/zeg/c/poly_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial tests}
   
   Three functions are provided to do tests: |_ZENPolyAreEqual| to test if two
   |ZENPoly| are equal, |_ZENPolyIsZero| to test if a |ZENPoly| is equal to
   zero and , |_ZENPolyIsXi| to test if a |ZENPoly| is equal to $X^i$.
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Equal.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{|RX| == |PX|.}
   }
   */
int _ZENPolyAreEqual(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  if (ZENPolyDeg(RX, Rg) != ZENPolyDeg(PX, Rg)) {
    return(0);
  }

  for(i = 0; i <= ZENPolyDeg(RX, Rg); i++) {
    if (!ZENEltAreEqual(ZENPolyGetCoeffPtr(RX, i, Rg),
			ZENPolyGetCoeffPtr(PX, i, Rg),
			Rg)) {
      return(0);
    }
  }
  return(1);
}

/* \TEX{
   \title{Is zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \output{|RX| == 0.}
   }
   */
int _ZENPolyIsZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  return(!ZENPolyDeg(RX, Rg) &&
	 ZENEltIsZero(ZENPolyGetCoeffPtr(RX, 0, Rg), Rg));
}

/* \TEX{
   \title{Is a polynomial equal to $X^i$.}
   \input{A polynomial |p| of a finite ring |Rg| and a degree |i|.}
   \output{$|p| == X^i$.}
   }
   */
int _ZENPolyIsXi (p, deg, Rg)
     ZENPoly p;
     int deg;
     ZENRing Rg;
{
  int i;

  if (ZENPolyDeg(p, Rg) != deg) {
    return(0);
  }

  for(i = 0; i < ZENPolyDeg(p, R); i++) {
    if (!ZENEltIsZero(ZENPolyGetCoeffPtr(p, i, Rg), Rg)) {
      return(0);
    }
  }
  if (!ZENEltIsOne(ZENPolyGetCoeffPtr(p, i, Rg), Rg)) {
    return(0);
  }
  
  return(1);
}
