/*
 * zen/zeg/c/sr_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Series additions}
   
   One function is provided to do additions: |_ZENSrAdd| to add a
   |ZENSr| to another. You can also compute the opposite of a series
   with |_ZENSrNegate| or subtract a series from another with
   |_ZENSrSubtract|.
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Addition.}
   \input{Two allocated series |RX| and |PX| of a finite ring |Rg|.}
   \side{$|RX| += |PX|$.}
   }
   */
void _ZENSrAdd(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;
{
  ZENElt elt;
  int i;

  
  ZENSrDeg(RX, Rg) =
    ZENSrDeg(RX, Rg) < ZENSrDeg(PX, Rg) ?
      ZENSrDeg(RX, Rg) : ZENSrDeg(PX, Rg);
  
  if (ZENSrVal(PX, Rg) >= ZENSrVal(RX, Rg)) {
    for(i = ZENSrVal(PX, Rg); i <= ZENSrDeg(RX, Rg); i++) {
      elt = ZENSrGetCoeffPtr(RX, i, Rg);
      ZENEltAdd(elt, ZENSrGetCoeffPtr(PX, i, Rg), Rg);
      ZENSrSetCoeff(RX, i, elt, Rg);
    }
    ZENSrUpdateValuation(RX, Rg);
  }
  else {
    if (ZENSrDeg(PX, Rg) >= ZENSrVal(RX, Rg)) {
      for(i = ZENSrDeg(RX, Rg); i >= ZENSrVal(RX, Rg); i--) {
	ZENSrSetCoeff(RX,
		      i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg),
		      ZENSrGetCoeffPtr(RX, i, Rg),
		      Rg);
	elt = ZENSrGetCoeffPtr(RX, i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg), Rg);
	ZENEltAdd(elt, ZENSrGetCoeffPtr(PX, i, Rg), Rg);
	ZENSrSetCoeff(RX, i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg), elt, Rg);
      }
      for(i = ZENSrVal(PX, Rg); i < ZENSrVal(RX, Rg); i++) {
	ZENSrSetCoeff(RX,
		      i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg),
		      ZENSrGetCoeffPtr(PX, i, Rg),
		      Rg);
      }
    }
    else {
      for(i = ZENSrVal(PX, Rg); i <= ZENSrDeg(RX, Rg); i++) {
	ZENSrSetCoeff(RX,
		      i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg),
		      ZENSrGetCoeffPtr(PX, i, Rg),
		      Rg);
      }
    }
    ZENSrVal(RX, Rg) = ZENSrVal(PX, Rg);
  }
}

/* \TEX{
   \title{Negation.}
   \input{Two allocated series |RX| and |PX| of a finite ring |Rg|.}
   \side{$|RX| = -|PX|$.}
   }
   */
void _ZENSrNegate(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;
{
  ZENElt elt;
  int i;

  ZENSrDeg(RX, Rg) =  ZENSrDeg(PX, Rg);
  ZENSrVal(RX, Rg) =  ZENSrVal(PX, Rg);
  for(i = ZENSrVal(PX, Rg); i <= ZENSrDeg(PX, Rg); i++) {
    elt = ZENSrGetCoeffPtr(PX, i, Rg);
    ZENEltNegate(elt, elt, Rg);
    ZENSrSetCoeff(RX, i, elt, Rg);
  }
}

/* \TEX{
   \title{Subtraction.}
   \input{Two allocated series |RX| and |PX| of a finite ring |Rg|.}
   \side{$|RX| -= |PX|$.}
   }
   */
void _ZENSrSubtract(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;
{
  ZENElt elt;
  int i;

  ZENSrDeg(RX, Rg) =
    ZENSrDeg(RX, Rg) < ZENSrDeg(PX, Rg) ?
      ZENSrDeg(RX, Rg) : ZENSrDeg(PX, Rg);
  
  if (ZENSrVal(PX, Rg) >= ZENSrVal(RX, Rg)) {
    for(i = ZENSrVal(PX, Rg); i <= ZENSrDeg(RX, Rg); i++) {
      elt = ZENSrGetCoeffPtr(RX, i, Rg);
      ZENEltSubtract(elt, ZENSrGetCoeffPtr(PX, i, Rg), Rg);
      ZENSrSetCoeff(RX, i, elt, Rg);
    }
    ZENSrUpdateValuation(RX, Rg);
  }
  else {
    if (ZENSrDeg(PX, Rg) >= ZENSrVal(RX, Rg)) {
      for(i = ZENSrDeg(RX, Rg); i >= ZENSrVal(RX, Rg); i--) {
	ZENSrSetCoeff(RX,
		      i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg),
		      ZENSrGetCoeffPtr(RX, i, Rg),
		      Rg);
	elt = ZENSrGetCoeffPtr(RX, i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg), Rg);
	ZENEltSubtract(elt, ZENSrGetCoeffPtr(PX, i, Rg), Rg);
	ZENSrSetCoeff(RX, i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg), elt, Rg);
      }
      for(i = ZENSrVal(PX, Rg); i < ZENSrVal(RX, Rg); i++) {
	elt = ZENSrGetCoeffPtr(RX,
			       i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg),
			       Rg);
	ZENEltNegate(elt, ZENSrGetCoeffPtr(PX, i, Rg), Rg);
	ZENSrSetCoeff(RX,
		      i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg),
		      elt,
		      Rg);
      }
    }
    else {
      for(i = ZENSrVal(PX, Rg); i < ZENSrDeg(RX, Rg); i++) {
	elt = ZENSrGetCoeffPtr(RX,
			       i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg),
			       Rg);
	ZENEltNegate(elt, ZENSrGetCoeffPtr(PX, i, Rg), Rg);
	ZENSrSetCoeff(RX,
		      i-ZENSrVal(PX, Rg)+ZENSrVal(RX, Rg),
		      elt,
		      Rg);
      }
    }
    ZENSrVal(RX, Rg) = ZENSrVal(PX, Rg);
  }
}
