/*
 * zen/zeg/c/sr_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Series tests}
   
   Two functions are provided to do tests: |_ZENSrAreEqual| to test if two
   |ZENSr| are equal, |_ZENSrIsZero| to test if a |ZENSr| is equal to
   zero.
   }
   */
#include <stdio.h>
#include "zen.h"


/* \TEX{
   \title{Equality.}
   \input{Two allocated series |RX| and |PX| of a finite ring |Rg|.}
   \output{The predicate $|RX| = |PX|$.}
   }
   */
int _ZENSrAreEqual(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;
{
  int i;

  if (ZENSrDeg(RX, Rg) != ZENSrDeg(PX, Rg)) {
    return(0);
  }
  if (ZENSrVal(RX, Rg) != ZENSrVal(PX, Rg)) {
    return(0);
  }

  for(i = ZENSrVal(RX, Rg); i <= ZENSrDeg(RX, Rg); i++) {
    if (!ZENEltAreEqual(ZENSrGetCoeffPtr(RX, i, Rg),
			ZENSrGetCoeffPtr(PX, i, Rg),
			Rg)) {
      return(0);
    }
  }
  return(1);
}


/* \TEX{
   \title{Is zero.}
   \input{An allocated series |RX| of a finite ring |Rg|.}
   \output{The predicate $|RX| = 0$.}
   }
   */
int _ZENSrIsZero(RX, Rg)
     ZENSr RX;
     ZENRing Rg;
{
  int i;

  for(i = ZENSrVal(RX, Rg); i <= ZENSrDeg(RX, Rg); i++) {
    if (!ZENEltIsZero(ZENSrGetCoeffPtr(RX, i, Rg), Rg)) {
      return(0);
    }
  }
  return(1);
}
