/*
 * zen/zeg/zmake.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "zmake.h"
#include "init.h"

char *C_FILES="doc ec_alloc ec_io ecpt_add ecpt_add_char2 ecpt_add_char3 ecpt_add_charp ecpt_alloc ecpt_assign ecpt_multiply ecpt_test elt_assign elt_exp elt_sqrt elt_trace debug clone_io mat_add mat_assign mat_create mat_gauss mat_inverse mat_io mat_matrix mat_multiply mat_permute mat_test mat_convert permutation_assign permutation_create permutation_ops poly_add poly_assign poly_derivative poly_roots poly_resultant poly_gcd poly_io poly_multiply poly_test sr_add sr_assign sr_io sr_multiply sr_test";

char *H_FILES="zen zentypes zeg";

int main(argc, argv)
     int argc;
     char *argv[];
{
  /*
   * Misc. variables
   *****************/
  char *ZEG_PREFIX;

 /* 
   * Internal targets 
   ********************/
  int objtoar = 0;
  int clean_obj = 0; int clean_lib = 0;
  int clean_doc = 0; int clean_tex = 0; int clean_dvi = 0;
  int tidy_lib = 0; int tidy_obj = 0; int tidy_bin = 0;
  int tidy_doc = 0;
  
  /* 
   * External targets
   *******************/
  int all = Target("all", argc, argv);

  int cc = Target("cc", argc, argv);

  int doc = Target("doc", argc, argv);
  int tex = Target("tex", argc, argv);
  int dvi = Target("dvi", argc, argv);

  int archtidy = Target("archtidy", argc, argv);
  int archclean = Target("archclean", argc, argv);
  int clean = Target("clean", argc, argv);
  int tidy = Target("tidy", argc, argv);

  /* 
   * Initialization
   *****************/
  zen_init(CURDIR, TOPDIR, ZCFDIR);
  zmake_init(ZEN_DEFINES, ZEN_INCLUDES, "", "", argc, argv);
  
  ZEG_PREFIX = concat(PREFIX, PZEG);

  /* 
   * Targets
   ***********/

  /* default */
  TargetImply(defaults, cc);

  /* all */
  TargetImply(all, cc=doc);

  /* cc */
  TargetImply(cc, objtoar);
  SrcToObjects(cc,
	       ZEG_OBJ, ZEG_PREFIX,
	       ZEG_SRC, "", C_FILES,
	       ZEN_INC, "");

  /* ar */
  ObjectsToArList(objtoar,
		  ZEG_LIB, ZEG_PREFIX, "", OBJ_TOAR,
		  ZEG_OBJ, ZEG_PREFIX, C_FILES);

  /* doc */
  TargetImply(doc, dvi);
  TargetImply(dvi, tex);

  /* tex */
  SrcToTex(tex,
	   ZEG_TEX, PZEG,
	   ZEG_SRC, "", C_FILES, CSUF,
	   ZEG_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   ZEG_TEX, "",
	   ZEG_INC, "", H_FILES, HSUF,
	   ZEG_BIN, "", IPT, PARSER);

  /* dvi */
  TexToDvi(dvi,
	   ZEG_DVI, PZEG,
	   ZEG_TEX, PZEG, C_FILES,
	   ZEN_INPUTS);
  TexToDvi(dvi,
	   ZEG_DVI, "",
	   ZEG_TEX, "", H_FILES,
	   ZEN_INPUTS);

  /* Cleaning */
  TargetImply(tidy, tidy_obj=tidy_lib=tidy_bin=tidy_doc);
  TargetImply(archtidy, archclean=clean_doc);
  TargetImply(clean, clean_lib=tidy_obj=clean_tex);
  TargetImply(archclean, clean_obj=clean_lib=clean_tex);
  TargetImply(clean_doc, clean_dvi=clean_tex);
  TargetImply(tidy_obj, clean_obj=clean_lib);
  
  CleaningRules(clean_obj, ZEG_OBJ, ZEG_PREFIX, C_FILES, OSUF);
  CleaningRules(clean_lib, ZEG_LIB, "", ZEG_PREFIX, OBJ_TOAR);
  TidyingRules(clean_tex, ZEG_TEX);
  CleaningRules(clean_dvi, ZEG_DVI, "", C_FILES, ".dvi");
  CleaningRules(clean_tex|clean_dvi, ZEG_DOC, "", "", "");

  TidyingRules(tidy_obj, ZEG_OBT);
  TidyingRules(tidy_lib, ZEG_LBT);
  TidyingRules(tidy_bin, ZEG_BNT);
  TidyingRules(tidy_doc, ZEG_DOC);

  /* 
   * Closing
   **********/
  zmake_close();

  exit(0);
}
