/*
 * zen/zelog/c/elt_alloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element creation}
   
   |_ZelogEltCreate| allocates one |ZENElt| and |_ZelogEltFree| frees then it.
   }
   */
#include <stdio.h>
#include "zelog.h"

/* \TEX{
   \title{Creation.}
   \input{A finite ring |Rg| and a pointer on a |ZENElt|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   }
   */
int _ZelogEltCreate(a, Rg)
     ZENElt *a;
     ZENRing Rg;
{
  return(ZelogEltCreate(*a, Rg));
}

/* \TEX{
   \title{Free.}
   \input{a |ZENElt| |a| of a finite ring |Rg| already allocated with
   |_ZelogEltCreate|.}
   \side{|a| is freed.}
   }
   */
void _ZelogEltFree(a, Rg)
     ZENElt a;
     ZENRing Rg;
{ 
  ZelogEltFree(a, Rg);
}

