/*
 * zen/zelog/c/elt_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Element assignment}
   
   One can set a |ZENElt| to zero or one with |_ZelogEltSetToZero| or
   |_ZelogEltSetToOne|. Then we can assign a |ZENElt| with an other with
   |_ZelogEltAssign|. Furthermore, we can set randomly a |ZENElt| with
   |_ZelogEltSetRandom| or enumerate the elements of a finite ring with
   |_ZelogEltSetNext|.
   }
   */
#include <stdio.h>
#include "zelog.h"

/* \TEX{
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| and a |ZENElt| |b| of a finite ring |Rg|.}
   \side{|*p_a| is filled with |b|.}
   }
   */
void _ZelogEltAssign(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  ZelogEltAssign(*p_a, b, Rg);
}

/* \TEX{
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| of a clone |Rg| and a |ZENElt| |b| of the original ring of |Rg|.}
   \side{|*p_a| is filled with |b|.}
   }
   */

void _ZelogEltToClone(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  BigNumLength i=1;
  BigNumDigit C[1];

  ZENEltToZ(C, &i, b, ZelogRingOrigin(Rg));
  ZelogElt2Int(*p_a, Rg) = ZelogLogTable(Rg, C[0]);
}

/* \TEX{
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| of the original ring of |Rg| and a |ZENElt| |b| of |Rg|.}
   \side{|*p_a| is filled with |b|.}
   }
   */
void _ZelogCloneToElt(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  ZENEltAssign(*p_a, ZelogConvTable(Rg, ZelogElt2Int(b, Rg)), ZelogRingOrigin(Rg));
}

/* \TEX{
   \title{Set to zero.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set to zero..}
   }
   */
void _ZelogEltSetToZero(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZelogEltSetToZero(*p_a, Rg);
}

/* \TEX{
   \title{Set to one.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set to one.}
   }
   */
void _ZelogEltSetToOne(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZelogEltSetToOne(*p_a, Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set randomly.}
   }
   */
void _ZelogEltSetRandom(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZelogEltSetRandom(*p_a, Rg);
}

/* \TEX{
   \title{Enumerate.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set randomly.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \note{After a number of call of this function equal to the number of elements in the finite ring, we obtain the same element.}
   }
   */
void _ZelogEltSetNext(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZelogEltSetNext(*p_a, Rg);
}

