/*
 * zen/zelog/c/elt_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{input/output}
   
   Conversions between the internal representation and the ASCII
   representation of a |ZENElt| are done by |_ZelogEltPrintToString|. |_ZelogEltPrintToFile|
   allows to write a |ZENElt| in a file.
   One can moreover write to and then read from a file a |ZENelt| with |_ZelogEltPutToFile|
   and |_ZelogEltReadFromFile|.
   }
   */
#include <stdio.h>
#include "zelog.h"

/* \TEX{
   \title{Evaluating an element as a multivariate polynomial over $Z$.}
   \input{A Bignum |p| of allocated size greater or equal than
   |ZENRingQl(Rg)|, a pointer |p_pl| on the real size of |p|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|p| is filled with the value of |G| considered as a
   multivariate polynomial in which we substitute all the variables
   with 2, |*p_pl| contains the real size of |p|.}
   } */
void _ZelogEltToZ(p, p_pl, G, Rg)
     BigNum p;
     BigNumLength *p_pl;
     ZENElt G;
     ZENRing Rg;
{
  BigNumDigit C[1];

  if (!ZelogElt2Int(G, Rg)) {
    ZENEltSetToZero(ZENRingHbuf(ZelogRingOrigin(Rg)), ZelogRingOrigin(Rg));
  }
  else {
    C[0] = ZelogElt2Int(G, Rg)-1;
    ZENEltExpPlain(ZENRingHbuf(ZelogRingOrigin(Rg)), C, (BigNumLength)1, ZelogGenerator(Rg), ZelogRingOrigin(Rg));
  }
  
  ZENEltToZ(p, p_pl, ZENRingHbuf(ZelogRingOrigin(Rg)), ZelogRingOrigin(Rg));
}

/* \TEX{
   \title{Getting an element from its evaluation as a multivariate
   polynomial over $Z$.}
   \input{A Bignum |p| of size |pl|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|G| is filled with the element whose evaluation considered as a
   multivariate polynomial in which we substitute all the variables
   with 2 is equal to (|p|, |pl|).}
   } */
void _ZelogZToElt(G, p, pl, Rg)
     BigNum p;
     BigNumLength pl;
     ZENElt *G;
     ZENRing Rg;
{
  ZelogZToElt(*G, p, pl, Rg);
}
