/*
 * zen/zelog/c/elt_trace.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element trace}
   
   |_ZelogEltTrace| computes the trace of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zelog.h"

/* \TEX{
   \title{Trace.}
   \input{A |ZENElt| |a| of a finite ring |Rg|.}
   }
   */
void _ZelogEltTrace(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZelogEltTrace(*p_c, a, Rg);
}

/* \TEX{
   \title{Find an element of trace one.}
   \input{A finite extension |Rg|.}
   \output{|ZENERR| if an error occured, |ZEN_NO_INVERSE| if an inverse was
   inpossible to compute, |ZEN_HAS_INVERSE| otherwise.} 
   \side{The finite field |Rg| is updated. In particular, an element of
   trace one is stored.}
   }
   */
int _ZelogPrcInitPolyRootsCanonical(Rg)
     ZENRing Rg;
{
  ZENPrc prc;
  BigNumDigit C[1];
  BigNumLength i;

  ZENPrcSetNone(prc);
  ZENPrcSet(prc, ZENPRC_POLY_ROOTS_CANONICAL);
  ZENPrcSet(prc, ZENPRC_FINITE_FIELD);
  switch (ZENRingAddPrc(ZelogRingOrigin(Rg), prc)) {
  case -1:
    return(ZERR);
  case ZEN_NO_INVERSE:
    return(ZEN_NO_INVERSE);
  case ZEN_HAS_INVERSE:
    if (!ZENPrcIsSet(ZENRingPrcp(ZelogRingOrigin(Rg)),
		     ZENPRC_POLY_ROOTS_CANONICAL)) {
      return(ZERR);
    }
    break;
  }
  
  ZENEltToZ(C, &i, ZENRingTheta(ZelogRingOrigin(Rg)), ZelogRingOrigin(Rg));
  ZelogElt2Int(ZENRingTheta(Rg), Rg) = ZelogLogTable(Rg, C[0]);
  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Free the element of trace 1.}
   \input{A finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE|}.
   \side{For compatibility purposes.}
   }
   */
int _ZelogPrcClosePolyRootsCanonical(Rg)
     ZENRing Rg;
{
  return(ZEN_HAS_INVERSE);
}
