/*
 * zen/zelog/c/mat_convert.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "zelog.h"

#define ZCTPG(R,i) ZENElt2Pol(ZelogConvTable(R,i),ZENRingOrigin(R))
#define ZelogCloneToGF2(R,i) (ZENPolyGetDigit(ZCTPG(R,i),0,ZENRingDef(R)))
#define ZelogGF2ToClone(R,i) ((unsigned short)ZelogLogTable(R,i))

/*
   \TEX{
   \Subsection{Matricial conversion}
   }
   */

/*
   \TEX{
   \title{Extension conversion}
   \input{Two matrices |ZENMat| and a tabulated extension |ZENRing| over $\Z/2\Z$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |m| is assigned with |M|.}
   } 
   */
int _Zelog2mMat2Vect(M,m,R)
     ZENMat M,m;
     ZENRing R;
{int l,c,cc,cl,n;
 BigNum buf;
 BigNum ptm;
 unsigned short *ptM;
 ZENRing RO;
 unsigned int mask;
 
 ZENTest((ZENRingDef(R)==NULL) ||
	 (ZENRingOrigin(R)==NULL) ||
	 (ZENRingArith(ZENRingDef(R)) != ZEN_LIB_ZE2),
	 ZEN_NOT_EXT,ZEN_MAT2VECT,ZEN_LIB_ZELOG, ZERR);
 RO=ZENRingOrigin(R);

 ZENTest(ZENMatIsPermutation(M,R)
	 ||ZENMatIsPermutation(m,ZENRingDef(RO)),
	 ZEN_ISPERM,ZEN_MAT2VECT,ZEN_LIB_ZELOG, ZERR);

 
 n=ZENPolyDeg(ZENRingPol(RO),ZENRingDef(RO));
 mask=BlocExp2(n)-1;
 
 buf=ZBNC(ZENMatNbBloc(m,ZENRingDef(RO)));
 ZENTest(buf == NULL,
	 ZEN_ALLOC,ZEN_MAT2VECT,ZEN_LIB_ZELOG, ZERR);

 if(ZENMatIsRowType(m,ZENRingDef(RO)))
   {
     ZENTest((ZENMatNbRow(M,R) != ZENMatNbBn(m,ZENRingDef(RO)))
	     ||(ZENMatNbCol(M,R)*n != ZENMatNbDigit(m,ZENRingDef(RO))),
	     ZEN_MAT_DIM,ZEN_MAT2VECT,ZEN_LIB_ZELOG, ZERR);
     cl=ZENMatNbCol(M,R);
   }
 else
   {
     ZENTest((ZENMatNbCol(M,R) != ZENMatNbBn(m,ZENRingDef(RO)))
	     ||(ZENMatNbRow(M,R)*n != ZENMatNbDigit(m,ZENRingDef(RO))),
	     ZEN_MAT_DIM,ZEN_MAT2VECT,ZEN_LIB_ZELOG, ZERR);
     cl=ZENMatNbRow(M,R);
   }
   
 ptm=ZENMatFirst(m,ZENRingDef(RO));
 ptM=(unsigned short *)ZENMatFirst(M,R);
 
 if(ZENMatAreSameType(m,M,R))
   for(l=ZENMatNbBn(m,ZENRingDef(R));
       l--;
       ptm += ZENMatNbBloc(m,ZENRingDef(R)))
     {       
       cc=ZENMatNbBloc(m,ZENRingDef(R));
       ZBNAssign(buf,ptm,cc);
       for(c=cl,cc=cc*SIZE_BLOC-1;c--;ptM++,cc-=n)
	 {
	   *ptM = ZelogGF2ToClone(R,(*buf) & mask);
	   ZBNShiftRight(buf,divSizeBloc(cc)+1,n);
	 }
     }
 else
   for(l=ZENMatNbBn(m,ZENRingDef(R));
       l--;
       ptm += ZENMatNbBloc(m,ZENRingDef(R)))
     {
       cc=ZENMatNbBloc(m,ZENRingDef(R));
       ZBNAssign(buf,ptm,cc);
       for(c=cl,cc=cc*SIZE_BLOC-1;c--;ptM+=ZENMatNbDigit(M,R),cc-=n)
	 {
	   *ptM = ZelogGF2ToClone(R,(*buf) & mask);
	   ZBNShiftRight(buf,divSizeBloc(cc)+1,n);
	 }
       ptM -= ZENMatNbBn(M,R)*ZENMatNbDigit(M,R)-1;
     }
 ZBNF(buf);
 return(0);
}

/*     
   \TEX{
   \title{Extension conversion}
   \input{Two matrices |ZENMat| and a tabulated extension |ZENRing| over $\Z/2\Z$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |M| is assigned with |m|.}
   } 
   */
int _Zelog2mVect2Mat(m,M,R)
     ZENMat M,m;
     ZENRing R;
{int cl,l,c,n,cc;
 BigNum ptm;
 unsigned short *ptM;
 unsigned int mask;
 ZENRing RO;

 
 ZENTest((ZENRingDef(R)==NULL) ||
	 (ZENRingOrigin(R)==NULL) ||
	 (ZENRingArith(ZENRingDef(R)) != ZEN_LIB_ZE2),
	 ZEN_NOT_EXT,ZEN_VECT2MAT,ZEN_LIB_ZELOG, ZERR);
 RO=ZENRingOrigin(R);
 
 ZENTest(ZENMatIsPermutation(M,R)
	 ||ZENMatIsPermutation(m,ZENRingDef(RO)),
	 ZEN_ISPERM,ZEN_VECT2MAT,ZEN_LIB_ZELOG, ZERR);
     
 n=ZENPolyDeg(ZENRingPol(RO),ZENRingDef(RO));
 mask=BlocExp2(n)-1;

 if(ZENMatIsRowType(m,ZENRingDef(RO)))
   {
     ZENTest((ZENMatNbRow(M,R) != ZENMatNbBn(m,ZENRingDef(RO)))
	     ||(ZENMatNbCol(M,R)*n != ZENMatNbDigit(m,ZENRingDef(RO))),
	     ZEN_MAT_DIM,ZEN_VECT2MAT,ZEN_LIB_ZELOG, ZERR);
     cl=ZENMatNbCol(M,R);     
   }
 else
   {
     ZENTest((ZENMatNbRow(M,R)*n != ZENMatNbDigit(m,ZENRingDef(RO)))
	     ||(ZENMatNbCol(M,R) != ZENMatNbBn(m,ZENRingDef(RO))),
	     ZEN_MAT_DIM,ZEN_VECT2MAT,ZEN_LIB_ZELOG, ZERR);
     cl=ZENMatNbRow(M,R);
   }

 ptm=ZENMatFirst(m,ZENRingDef(RO))+ZENMatNbBnd(m,ZENRingDef(RO));
 ptM=(unsigned short *)ZENMatFirst(M,R)+ZENMatNbBn(M,R)*ZENMatNbDigit(M,R);

 if(ZENMatAreSameType(m,M,R))
   for(l=ZENMatNbBn(m,ZENRingDef(R)),
	 ptm -= ZENMatNbBloc(m,ZENRingDef(R)),
	 ptM--;
       l--;
       ptm -= ZENMatNbBloc(m,ZENRingDef(R)))
     {
       *ptm = ZelogCloneToGF2(R,(*ptM)) & mask;
	 
       for(c=cl-1,ptM--,cc=2*n-1;c--;ptM--,cc+=n)
	 {
	   ZBNShiftLeft(ptm,divSizeBloc(cc)+1,n);
	   *ptm |= ZelogCloneToGF2(R,(*ptM)) & mask;
	 }
     }
 else
   for(l=ZENMatNbBn(m,ZENRingDef(R)),
	 ptm -= ZENMatNbBloc(m,ZENRingDef(R)),
	 ptM--;
       l--;
       ptm -= ZENMatNbBloc(m,ZENRingDef(R)))
     {
       *ptm = ZelogCloneToGF2(R,(*ptM)) & mask;
       for(c=cl-1,ptM-=ZENMatNbDigit(M,RO),cc=2*n-1;
	   c--;
	   ptM-=ZENMatNbDigit(M,RO),cc+=n)
	 {
	   ZBNShiftLeft(ptm,divSizeBloc(cc)+1,n);
	   *ptm |= ZelogCloneToGF2(R,(*ptM)) & mask;
	 }
       ptM += ZENMatNbBn(M,RO)*ZENMatNbDigit(M,RO)-1;
     }
 return(0);
}
