/*
 * zen/zelog/c/mat_create.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Allocating a matrix}
   }
   */


#include <stdio.h>
#include "zelog.h"
#include "zed.h"
#include "zeg.h"

/* \TEX{
   \title{Computation of the matrix parameters}
   \input{A |ZENMatTypeStruct|, the dimensions of the desired matrix, a pointer on a |ZENMat| and a |ZENRing|.}
   \output{0 if no error occured, |ZENERR| otherwise.}
   \side{The |ZENMat| structure is allocated and the parameters computed according to the type of the desired matrix, but the coefficients are not allocated.}
   \note{This function should not be directly used (see |ZelogMatCreate|)}
   }
   */   

int _ZelogMatCreateParams(ts,l,c,p_pms,R)
     Dim l,c;
     ZENMatTypeStruct ts;
     ZENMat *p_pms;
     ZENRing R;
{
  *p_pms=(ZENMat)ZENMalloc(sizeof(struct zenmat));

  ZENTest(*p_pms==NULL,
	  ZEN_ALLOC,ZEN_MATCREPMS,ZEN_LIB_ZELOG, ZERR);
  ZENTest((l<=0)||(c<=0),
	  ZEN_MAT_DIM0,ZEN_MATCREPMS,ZEN_LIB_ZELOG, ZERR);
 
  ZENMatFreePtr(*p_pms)=_ZedsMatFree;
  ZENMatAssignPtr(*p_pms)=__ZENMatAssign;
  ZENMatSetToOnePtr(*p_pms)=__ZENMatSetToOne;
  ZENMatIsOnePtr(*p_pms)=__ZENMatIsOne;
  ZENMatSetRandomPtr(*p_pms)=__ZENMatSetRandom;
  ZENMatPermuteBnPtr(*p_pms)=__ZENMatPermuteBn;
  
  ZelogMatSize(*p_pms,R)=0;
  ZelogMatNbPtr(*p_pms,R)=0;
  
  switch(ts)
    {
    case ZENMatTypeRow:
      ZelogMatType(*p_pms,R)=ZENMatTypeRow;
      ZelogMatNbBn(*p_pms,R)=l;
      ZelogMatNbDigit(*p_pms,R)=c;
      ZelogMatNbBloc(*p_pms,R)=divSizeBloc(c*SIZE_SHORT-1)+1;
      ZelogMatNbBnd(*p_pms,R)= l * ZelogMatNbBloc(*p_pms,R);

      ZENMatGetCoeffNewPtr(*p_pms)=__ZENMatGetBloc;
      ZENMatGetCoeffPtrPtr(*p_pms)=__ZENMatGetBlocPtr;
      ZENMatSetCoeffPtr(*p_pms)=__ZENMatSetBloc;     
      ZENMatPermuteRowPtr(*p_pms)=__ZENMatPermuteBn;     
      ZENMatPermuteColPtr(*p_pms)=__ZENMatPermuteDigit;     
      break;
      
    case ZENMatTypeCol:
      ZelogMatType(*p_pms,R)=ZENMatTypeCol;
      ZelogMatNbBn(*p_pms,R)=c;
      ZelogMatNbDigit(*p_pms,R)=l;
      ZelogMatNbBloc(*p_pms,R)=divSizeBloc(l*SIZE_SHORT-1)+1;
      ZelogMatNbBnd(*p_pms,R)= c * ZelogMatNbBloc(*p_pms,R);

      ZENMatGetCoeffNewPtr(*p_pms)=__ZENMatGetDigit;
      ZENMatGetCoeffPtrPtr(*p_pms)=__ZENMatGetDigitPtr;
      ZENMatSetCoeffPtr(*p_pms)=__ZENMatSetDigit;     
      ZENMatPermuteRowPtr(*p_pms)=__ZENMatPermuteDigit;     
      ZENMatPermuteColPtr(*p_pms)=__ZENMatPermuteBn;     
      break;

    default:
      ZENTest(1,
	      ZEN_UNKNOWN,ZEN_MATCREPMS,ZEN_LIB_ZELOG, ZERR);
    }
  return(0);
}

