/*
 * zen/zelog/c/mat_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial multiplications}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zelog.h"
/*
   \TEX{
   \title{Multiplication.}
   \input{Three |ZENMat| |X|, |A| and |B| of a finite ring |R|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{$|X|= |A| \times |B|$.}
   \note{Algorithm used is the general one.}
   }
   */
   
int _ZelogMatMultiplyPlain(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{int l,c,k;
 ZENElt S,g,e1,e2;
 ZENTest(ZelogMatIsPermutation(A,R)||ZelogMatIsPermutation(B,R)||ZelogMatIsPermutation(X,R),
	 ZEN_ISPERM,ZEN_MATMULTIPLY,ZEN_LIB_ZELOG, ZERR);
 ZENTest((ZelogMatNbRow(A,R) != ZelogMatNbRow(X,R))
	 ||(ZelogMatNbCol(A,R) != ZelogMatNbRow(B,R))
	 ||(ZelogMatNbCol(B,R) != ZelogMatNbCol(X,R)),
	 ZEN_MAT_DIM,ZEN_MATMULTIPLY,ZEN_LIB_ZELOG, ZERR);

 for(l=ZelogMatNbRow(X,R)-1;l>=0;l--)
   for(c=ZelogMatNbCol(X,R)-1;c>=0;c--)
     {
       k=ZelogMatNbCol(A,R)-1;
       e1=ZelogMatGetCoeffPtr(A,l,k,R);
       e2=ZelogMatGetCoeffPtr(B,k,c,R);
       ZelogEltMultiply(S,e1,e2,R);
       for(k--;k>=0;k--)
         {
	   e1=ZelogMatGetCoeffPtr(A,l,k,R);
	   e2=ZelogMatGetCoeffPtr(B,k,c,R);
	   ZENEltMultiply(g,e1,e2,R);
           ZENEltAdd(S,g,R);
         }
       ZelogMatSetCoeff(X,l,c,S,R);
     }
 return(0);
}

/*
   \TEX{
   \title{Multiplication of scalar}
   \input{A |ZENMat|, a |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{All the elements of |A| are multiplied by |b|.}
   \note{|A| cannot be a permutation.}
   }
   */
int _ZelogMatMultiplyScalar(A,b,R)
     ZENMat A;
     ZENElt b;
     ZENRing R;
{
  int l,c;

  ZENTest(ZelogMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATMULTIPLY,ZEN_LIB_ZELOG, ZERR);
     
  for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
    for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
      ZelogAnyEltMultiply(ZelogMatBloc(A,l,c,R),ZelogElt2Int(b,R),ZelogMatBloc(A,l,c,R),R);
  
  return(0);
}

