/*
 * zen/zelog/c/mat_permute.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial permutations}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions.
   }
   */
#include <stdio.h>
#include "zelog.h"


/*
   \TEX{
   \title{Permute two |Bn|s}
   \input{A |ZENMat| |A|, two indexes and a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The specified |Bns| of |A| are exchanged}
   \note{Except for advanced user, |PermuteRow| and |PermuteCol| functions should be preferred}
   }
   
   */
   
int _ZelogMatPermuteBn(A,l1,l2,R)
     ZENMat A;
     Dim l1,l2;
     ZENRing R;
{int c;
 ZENElt S;
 
 ZENTest(ZelogMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATPERMUTEBNS,ZEN_LIB_ZELOG, ZERR);
 ZENTest((l1<0)||(l2<0)||(l1>=ZelogMatNbBn(A,R))||(l2>=ZelogMatNbBn(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATPERMUTEBNS,ZEN_LIB_ZELOG, ZERR);

 for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
   {
     ZelogAnyEltAssign(ZelogElt2Int(S,R),ZelogMatBloc(A,l1,c,R),R);
     ZelogAnyEltAssign(ZelogMatBloc(A,l1,c,R),ZelogMatBloc(A,l2,c,R),R);
     ZelogAnyEltAssign(ZelogMatBloc(A,l2,c,R),ZelogElt2Int(S,R),R);
   }
 return(0);
}

/*
   \TEX{
   \title{Permute two |digit|s}
   \input{A |ZENMat| |A|, two indexes and a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The specified |digits| of |A| are exchanged}
   \note{Except for advanced user, |PermuteRow| and |PermuteCol| functions should be preferred}
   }   
   */
   

int _ZelogMatPermuteDigit(A,l1,l2,R)
     ZENMat A;
     Dim l1,l2;
     ZENRing R;
{int c;
 ZENElt S;
 
 ZENTest(ZelogMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATPERMUTEDG,ZEN_LIB_ZELOG, ZERR);
 ZENTest((l1<0)||(l2<0)||(l1>=ZelogMatNbDigit(A,R))||(l2>=ZelogMatNbDigit(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATPERMUTEDG,ZEN_LIB_ZELOG, ZERR);

 for(c=ZelogMatNbBn(A,R)-1;c>=0;c--)
   {
     ZelogAnyEltAssign(ZelogElt2Int(S,R),ZelogMatBloc(A,c,l1,R),R);
     ZelogAnyEltAssign(ZelogMatBloc(A,c,l1,R),ZelogMatBloc(A,c,l2,R),R);
     ZelogAnyEltAssign(ZelogMatBloc(A,c,l2,R),ZelogElt2Int(S,R),R);
   }
 return(0);
}

