/*
 * zen/zelog/c/mat_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   
   \Subsection{Tests on a matrix}
   }
   */


#include <stdio.h>
#include "zelog.h"
/* \TEX{
   \title{Test of equality}
   \input{Two matrices and a |ZENRing|}
   \output{1 if the matrices are equal, 0 otherwise}
   }
   */

int _ZelogMatAreEqual(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;
 
 if(ZelogMatAreSameType(A,B,R))
   return(ZelogMatSameTypeAreEqual(A,B,R));
 
 if((ZelogMatNbBn(A,R) != ZelogMatNbDigit(B,R))
    ||(ZelogMatNbDigit(A,R) != ZelogMatNbBn(B,R)))
   return(0);
 for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
     if(!(ZelogAnyEltAreEqual(ZelogMatBloc(A,l,c,R),ZelogMatBloc(B,c,l,R),R)))
       return(0);
 return(1);
}

/* \TEX{
   \title{Test of equality to zero}
   \input{A matrix and a |ZENRing|}
   \output{1 if the matrix is null, 0 otherwise}
   }
   */

int _ZelogMatIsZero(A,R)
     ZENMat A;
     ZENRing R;
{
  return(ZelogMatIsZero(A,R));
}

/* \TEX{
   \title{Test of equality to identity}
   \input{A matrix and a |ZENRing|}
   \output{1 if the matrix is identity, 0 otherwise}
   \note{The matrix is checked to be square.} 
   }
   */

int _ZelogMatIsOne(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;

 if(ZelogMatNbBn(A,R) != ZelogMatNbDigit(A,R))
   return(0);
 
 for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZelogMatNbBn(A,R)-1;c>=0;c--)
     if(l==c)
       {
	 if(!(ZelogAnyEltIsOne(ZelogMatBloc(A,l,c,R),R)))
	   return(0);
       }
     else
       {
	 if(!(ZelogAnyEltIsZero(ZelogMatBloc(A,l,c,R),R)))
	   return(0);
       }
 return(1); 
}


