/*
 * zen/zelog/c/poly_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial additions}
   
   One function is provided to do additions: |_ZelogPolyAdd| to add a
   |ZENPoly| to another.
   }
   */
#include <stdio.h>
#include "zelog.h"

/* \TEX{
   \title{Addition.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| += |PX|.}
   }
   */
void _ZelogPolyAdd(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;

  if (ZelogPolyDeg(RX, Rg) < ZelogPolyDeg(PX, Rg)) {
    mindeg = ZelogPolyDeg(RX, Rg);
    ZelogPolyDeg(RX, Rg) =  ZelogPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZelogPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZelogAnyEltAdd(ZelogPolyGetCof(RX, i, Rg), ZelogPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZelogPolyDeg(PX, Rg); i++) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i, Rg),
		   ZelogPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZelogPolyUpdateDegree(RX, Rg);
}

/* \TEX{
   \title{Negation.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| = -|PX|.}
   }
   */
void _ZelogPolyNegate(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZelogPolyDeg(PX, Rg); i++) {
    ZelogAnyEltNegate(ZelogPolyGetCof(RX, i, Rg), ZelogPolyGetCof(PX, i, Rg), Rg);
  }
  ZelogPolyDeg(RX, Rg) =  ZelogPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Subtraction.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| -= |PX|.}
   }
   */
void _ZelogPolySubtract(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, mindeg;

  if (ZelogPolyDeg(RX, Rg) < ZelogPolyDeg(PX, Rg)) {
    mindeg = ZelogPolyDeg(RX, Rg);
    ZelogPolyDeg(RX, Rg) =  ZelogPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZelogPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZelogAnyEltSubtract(ZelogPolyGetCof(RX, i, Rg), ZelogPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZelogPolyDeg(PX, Rg); i++) {
    ZelogAnyEltNegate(ZelogPolyGetCof(RX, i, Rg),
		   ZelogPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZelogPolyUpdateDegree(RX, Rg);
}
