/*
 * zen/zelog/c/poly_alloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
  \TEX{

   \Subsection{Polynomial creation}
   
   |_ZelogPolyCreate| allocates one |ZENPoly| and |_ZelogPolyFree| frees then it.
   }
   */
#include <stdio.h>
#include "zelog.h"

/* \TEX{
   \title{Creation.}
   \input{A finite ring |Rg| and a degree |deg|.}
   \output{An allocated |ZENPoly| set to 0 with |deg| coefficients allocated. }
   }
   */
ZENPoly _ZelogPolyCreate(deg, Rg)
     int deg;
     ZENRing Rg;
{
  ZENPoly PX;
  int lt;

  
  if (!(PX = ZENPolyAllocStruct(Rg))) {
    ZENSetError(ZEN_LIB_ZELOG, ZEN_POLYCREATE, ZEN_ALLOC);
    return(NULL);
  }
  
  lt=divSizeBloc((deg+1)*SIZE_SHORT)+1;
  if (!(ZelogPolyCof(PX, Rg) = (unsigned short *)ZENMalloc(lt*sizeof(BigNumDigit)))) {
    ZENFree(PX);
    ZENSetError(ZEN_LIB_ZELOG, ZEN_POLYCREATE, ZEN_ALLOC);
    return(NULL);
  }
  ZelogPolyLgt(PX, Rg) = lt*sizeof(BigNumDigit)/sizeof(unsigned short) - 1;
  return(PX);
}

/* \TEX{
   \title{Free.}
   \input{a |ZENPoly| |PX| of a finite ring |Rg| already allocated with
   |_ZelogPolyCreate|.}
   \side{|PX| is freed.}
   }
   */
void _ZelogPolyFree(PX,Rg)
     ZENPoly PX;
     ZENRing Rg;
{
  ZENFree(ZelogPolyCof(PX, Rg));
  ZENFree(PX);
}
