/*
 * zen/zelog/c/poly_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial tests}
   
   Three functions are provided to do tests: |_ZelogPolyAreEqual| to test if two
   |ZENPoly| are equal, |_ZelogPolyIsZero| to test if a |ZENPoly| is equal to
   zero and , |_ZelogPolyIsXi| to test if a |ZENPoly| is equal to $X^i$.
   }
   */
#include <stdio.h>
#include "zelog.h"

/* \TEX{
   \title{Equal.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{|RX| == |PX|.}
   }
   */
int _ZelogPolyAreEqual(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  if (ZelogPolyDeg(RX, Rg) != ZelogPolyDeg(PX, Rg)) return(0);

  for(i = 0; i <= ZelogPolyDeg(RX, Rg); i++) {
    if (!ZelogAnyEltAreEqual(ZelogPolyGetCof(RX, i, Rg),
			 ZelogPolyGetCof(PX, i, Rg),
			 Rg)) {
      return(0);
    }
  }
  return(1);
}

/* \TEX{
   \title{Is zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \output{|RX| == 0.}
   }
   */
int _ZelogPolyIsZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  return(!ZelogPolyDeg(RX, Rg) &&
	 ZelogAnyEltIsZero(ZelogPolyGetCof(RX, 0, Rg), Rg));
}

/* \TEX{
   \title{Is a polynomial equal to $X^i$.}
   \input{A polynomial |p| of a finite ring |Rg| and a degree |i|.}
   \output{$|p| == X^i$.}
   }
   */
int _ZelogPolyIsXi (p, deg, Rg)
     ZENPoly p;
     int deg;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZelogPolyDeg(p, Rg) != deg) {
    return(0);
  }

  for(i = 0; i < ZelogPolyDeg(p, R); i++) {
    elt = ZelogPolyGetCoeffPtr(p, i, Rg);
    if (!ZelogEltIsZero(elt, Rg)) {
      return(0);
    }
  }
  elt = ZelogPolyGetCoeffPtr(p, i, Rg);
  if (!ZelogEltIsOne(elt, Rg)) {
    return(0);
  }
  
  return(1);
}
