/*
 * zen/zelog/zmake.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "zmake.h"
#include "init.h"

char *C_FILES="doc elt_add elt_alloc elt_assign elt_inverse elt_io elt_multiply elt_test elt_trace init mat_add mat_assign mat_convert mat_create mat_gauss mat_inverse mat_multiply mat_permute mat_test poly_add poly_alloc poly_assign poly_multiply poly_test";

char *H_FILES="zelog";

int main(argc, argv)
     int argc;
     char *argv[];
{
  /*
   * Misc. variables
   *****************/
  char *ZELOG_PREFIX;

 /* 
   * Internal targets 
   ********************/
  int objtoar = 0;
  int clean_obj = 0; int clean_lib = 0;
  int clean_doc = 0; int clean_tex = 0; int clean_dvi = 0;
  int tidy_lib = 0; int tidy_obj = 0; int tidy_bin = 0;
  int tidy_doc = 0;
  
  /* 
   * External targets
   *******************/
  int all = Target("all", argc, argv);

  int cc = Target("cc", argc, argv);

  int doc = Target("doc", argc, argv);
  int tex = Target("tex", argc, argv);
  int dvi = Target("dvi", argc, argv);

  int archtidy = Target("archtidy", argc, argv);
  int archclean = Target("archclean", argc, argv);
  int clean = Target("clean", argc, argv);
  int tidy = Target("tidy", argc, argv);

  /* 
   * Initialization
   *****************/
  zen_init(CURDIR, TOPDIR, ZCFDIR);
  zmake_init(ZEN_DEFINES, ZEN_INCLUDES, "", "", argc, argv);
  
  ZELOG_PREFIX = concat(PREFIX, PZELOG);

  /* 
   * Targets
   ***********/

  /* default */
  TargetImply(defaults, cc);

  /* all */
  TargetImply(all, cc=doc);

  /* cc */
  TargetImply(cc, objtoar);
  SrcToObjects(cc,
	       ZELOG_OBJ, ZELOG_PREFIX,
	       ZELOG_SRC, "", C_FILES,
	       ZEN_INC, "");

  /* ar */
  ObjectsToArList(objtoar,
		  ZELOG_LIB, ZELOG_PREFIX, "", OBJ_TOAR,
		  ZELOG_OBJ, ZELOG_PREFIX, C_FILES);

  /* doc */
  TargetImply(doc, dvi);
  TargetImply(dvi, tex);

  /* tex */
  SrcToTex(tex,
	   ZELOG_TEX, PZELOG,
	   ZELOG_SRC, "", C_FILES, CSUF,
	   ZELOG_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   ZELOG_TEX, "",
	   ZELOG_INC, "", H_FILES, HSUF,
	   ZELOG_BIN, "", IPT, PARSER);

  /* dvi */
  TexToDvi(dvi,
	   ZELOG_DVI, PZELOG,
	   ZELOG_TEX, PZELOG, C_FILES,
	   ZEN_INPUTS);
  TexToDvi(dvi,
	   ZELOG_DVI, "",
	   ZELOG_TEX, "", H_FILES,
	   ZEN_INPUTS);

  /* Cleaning */
  TargetImply(tidy, tidy_obj=tidy_lib=tidy_bin=tidy_doc);
  TargetImply(archtidy, archclean=clean_doc);
  TargetImply(clean, clean_lib=tidy_obj=clean_tex);
  TargetImply(archclean, clean_obj=clean_lib=clean_tex);
  TargetImply(clean_doc, clean_dvi=clean_tex);
  TargetImply(tidy_obj, clean_obj=clean_lib);
  
  CleaningRules(clean_obj, ZELOG_OBJ, ZELOG_PREFIX, C_FILES, OSUF);
  CleaningRules(clean_lib, ZELOG_LIB, "", ZELOG_PREFIX, OBJ_TOAR);
  TidyingRules(clean_tex, ZELOG_TEX);
  CleaningRules(clean_dvi, ZELOG_DVI, "", C_FILES, ".dvi");
  CleaningRules(clean_tex|clean_dvi, ZELOG_DOC, "", "", "");

  TidyingRules(tidy_obj, ZELOG_OBT);
  TidyingRules(tidy_lib, ZELOG_LBT);
  TidyingRules(tidy_bin, ZELOG_BNT);
  TidyingRules(tidy_doc, ZELOG_DOC);

  /* 
   * Closing
   **********/
  zmake_close();

  exit(0);
}
