/*
 * zen/zem/c/doc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*-*-LaTeX-*-
  This is the documentation file of the whole ZEN library. There is
  no C code in this file, only LaTeX commands.


   \TEX{
Montgomery's idea is implemented in \zen. The following description is largely inspired from~\cite[pages 133--135]{theseMor}.

We now assume that we want to work in a modular ring $\quotient{\Z}{N\Z}$, with $N$ an odd integer.
Let $R$ be an integer greater than $N$ and prime with $N$.
Then, the mapping
$$
\begin{array}{rcrcl}
\hat{\phi} & : & \quotient{\Z}{N\Z} & \rightarrow & \quotient{\Z}{N\Z} \\
  &   & x & \mapsto     & \hat{x}=Rx \bmod N
\end{array}
$$
is invertible and we call $\hat{x}$ the $N$-rsidu of $x$.

It is always possible to find $u$ and $v$ two integers such that
$$
Ru-vN=1,
$$
with
$$
0<u<N
\mbox{ and }
0<v<R.
$$

\subparagraph*{Theorem}
Let $z$ be an integer between 0 and $RN$, the following algorithm computes $zR^{-1}mod N$.
\begin{enumerate}
\item Compute $m=((z \bmod R) v) \bmod R$.
\item Let $t=(z+mN)/R$.
\item If $t$ is greater than $N$, subtract $N$.
\item Return $t$.
\end{enumerate}

\subparagraph*{Proof}
We have
\begin{eqnarray*}
m = zv \bmod R & \Leftrightarrow & mN = z(vN) \bmod R, \cr
& \Leftrightarrow & mN = -z \bmod R, \cr
& \Leftrightarrow & mN + z = \alpha R, \cr
\end{eqnarray*}
with $\alpha$ an integer.
Hence,  $t$ in step 2 is an integer.
Furthermore, $tR = z \bmod N$.
As $0\le z<RN$ and $0\le m < R$, we have, after step 2, $0\le t<2N$.
Hence, the returned value after step 3 is $zR^{-1} \bmod N$.

We denote  $\hat{\phi}^{-1}$ the above procedure.

Montgomery's idea is to replace modular oerations on integers by operations on $N$-rsidus.
In this case we have
$$
\hat{x} \hat{\times} \hat{y} = \hat{\phi}^{-1}(\hat{x}\times \hat{y}),
$$
because
$$
(xy)R = (xR)(yR)R^{-1}.
$$
On the other hand, we have
$$
\hat{(x^{-1})} = (\hat{x})^{-1} \hat{R},
$$
because
$$
(x^{-1})R= (xR)^{-1}R^2.
$$

\paragraph*{Implementation in \zen}
\
In order to use Montgomery's representation, one has to use a clone. The implementation follows what precedes, except that the inversion is faster using the following procedure:
\begin{enumerate}
\item Convert the  $N$-rsidu in the original modular ring.
\item Inversion in the original ring.
\item Re-conversion to obtain the $N$-rsidu.
\end{enumerate}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\begin{figure}
\begin{center}
\begin{tabular}{@{$\vert$ }c@{ $\vert$ }c@{ $\vert$ }c@{ $\vert$ }c@{ $\vert$ }c@{ $\vert$ }c@{ $\vert$}}\hline
Size (bits)               & 128   & 256   & 512  & 768  & 1024 \\ \hline
\zen (standard ring)      & 0.027 & 0.13  & 0.76 & 2.27 & 5.0  \\ \hline
\zen (clone + precomputations)   & 0.013 & 0.08  & 0.53 & 1.65 & 3.8  \\ \hline
\end{tabular}
\end{center}
\caption[Modular exponentiation $a^b \bmod c$]{modular exponentiation  $a^b \bmod c$ on sparc II (times given in seconds)}
\label{fig:montgomery}
\end{figure}
\paragraph*{Performances}
\
The table~\ref{fig:montgomery} shows the improvement obtained by Montgomery's reduction on the classical example of modular exponentiation. The observed gain is in the range 25 to~40~\%.

\subparagraph*{Note}
Montgomery's idea applies only for odd modulus. For even ones, one has to use chinese remainder theorem to get rid of even factors.



}
*/
