/*
 * zen/zem/c/elt_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{   
   \Subsection{Element inversion}
   
   |_ZemEltInverse| computes the inverse of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zem.h"

/* \TEX[
   \title{Inverse.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |R|.}
   \output{|ZENERR| if an error occured, |ZEN_HAS_INVERSE| if |a| has an inverse, |ZEN_NO_INVERSE| otherwise}
   \side{$|*p_c| = 1/a$ if |a| has an inverse; otherwise |ZENRingFact(R)| is filled with the gcd of |a| with the modulo.} 
   }
   */
int _ZemEltInverse(p_c, a, R)
     ZENElt *p_c, a;
     ZENRing R;
{
  int ret;
  
  ZemCloneToElt(ZENRingHbuf(R),a,R);
  if((ret=ZepEltInverse(ZENRingJbuf(R),ZENRingHbuf(R),ZENRingOrigin(R))))
    return(ret);
  ZemEltToClone(*p_c,ZENRingJbuf(R),R);
  return(ZEN_HAS_INVERSE);
}

