/*
 * zen/zem/c/elt_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{input/output}
   }
   */
#include <stdio.h>
#include "zem.h"

/* \TEX[
   \title{Evaluating an element as a multivariate polynomial over $Z$.}
   \input{A Bignum |p| of allocated size greater or equal than
   |ZENRingQl(Rg)|, a pointer |p_pl| on the real size of |p|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|p| is filled with the value of |G| considered as a
   multivariate polynomial in which we substitute all the variables
   with the characteristic, |*p_pl| contains the real size of |p|.}
   } */
void _ZemEltToZ(p, p_pl, G, Rg)
     BigNum p;
     BigNumLength *p_pl;
     ZENElt G;
     ZENRing Rg;
{
  ZBNAssign(ZemRingBn(Rg),ZENElt2Bn(G,Rg),ZENRingSizeElt(Rg));
  ZBNSetToZero(ZemRingBn(Rg)+ZENRingSizeElt(Rg),
	       ZemRingBnl(Rg)-ZENRingSizeElt(Rg));
  *p_pl=_ZemEltRedMod(ZemRingBn(Rg),Rg);
  ZBNAssign(p,ZemRingBn(Rg),*p_pl);
}

/* \TEX[
   \title{Getting an element from its evaluation as a multivariate
   polynomial over $Z$.}
   \input{A Bignum |p| of size |pl|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|G| is filled with the element whose evaluation considered as a
   multivariate polynomial in which we substitute all the variables
   with the characteristic is equal to (|p|, |pl|).}
   } */
void _ZemZToElt(G, p, pl, Rg)
     BigNum p;
     BigNumLength pl;
     ZENElt *G;
     ZENRing Rg;
{
  BigNumLength bnl;
  
  ZBNSetToZero(ZemRingBn(Rg),bnl=ZENRingPl(Rg));
  ZBNAssign(ZemRingBn(Rg)+bnl,p,pl);
  ZBNSetToZero(ZemRingBn(Rg)+bnl+pl,ZemRingBnl(Rg)-bnl-pl);
  ZBNDivide(ZemRingBn(Rg),ZemRingBnl(Rg),ZENRingP(Rg),ZENRingPl(Rg));
  bnl=ZBNNumDigits(ZemRingBn(Rg),ZENRingPl(Rg));
  ZBNAssign(ZENElt2Bn(*G,Rg),ZemRingBn(Rg),bnl);
  ZBNSetToZero(ZENElt2Bn(*G,Rg)+bnl,ZENRingSizeElt(Rg)-bnl);
}
