/*
 * zen/zem/c/elt_trace.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element trace}
   
   |_ZemEltTrace| computes the trace of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zem.h"

/* \TEX[
   \title{Computing an element of trace 1.}
   \input{A finite ring |Rg|.} 
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| otherwise.}
   \side{|ZENRingTheta(Rg)| is allocated and filled with 1.}
   }
   */
int _ZemPrcInitPolyRootsCanonical(Rg)
     ZENRing Rg;
{
  if (ZemEltCreate(ZENRingTheta(Rg), Rg)) {
    ZENSetError(ZEN_LIB_ZEM, ZEN_ADDPRC, ZEN_ALLOC);
    return(ZERR);
  }
  ZemEltSetToOne(ZENRingTheta(Rg), Rg);

  return(ZEN_HAS_INVERSE);
}

/* \TEX[
   \title{Desallocate the element of trace 1.}
   \input{A finite ring |Rg|.} 
   \output{|ZEN_HAS_INVERSE|.}
   \side{|ZENRingTheta(Rg)| is desallocated.}
   }
   */
int _ZemPrcClosePolyRootsCanonical(Rg)
     ZENRing Rg;
{
  ZemEltFree(ZENRingTheta(Rg), Rg);
  
  return(ZEN_HAS_INVERSE);
}
