/*
 * zen/zem/c/poly_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsubsection{Polynomial assignment}
   
   }
   */
#include <stdio.h>
#include "zem.h"

#define ZBNPolyGetCof(P,i,R) ZENElt2Bn(ZemPolyGetCof((P),(i),(R)),(R))

/* \TEX[
   \title{Set to $X^i$.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg| allocated at least for
   degree $|deg|$.}
   \side{|RX| is set to $X^{|deg|}$.}
   }
   */
void _ZemPolySetToXi(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZemPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZemEltSetToZero(ZemPolyGetCof(RX, i, Rg), Rg);
  }
  ZemEltSetToOne(ZemPolyGetCof(RX, deg, Rg), Rg);
}

/* For internal use only, cf. poly_multiply. */
void _ZemPolyLazyToVeryLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZemPolyDeg(PX, Rg); i++) {
    ZBNSetToZero(ZBNPolyGetCof(RX, i, Rg), ZENRingSizeElt(Rg));
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg)+ZENRingSizeElt(Rg),
	      ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg));
    ZBNSetDigit(ZBNPolyGetCof(RX, i, Rg)+2*ZENRingSizeElt(Rg), 0);
    ZBNSetDigit(ZBNPolyGetCof(RX, i, Rg)+2*ZENRingSizeElt(Rg)+1, 0);
  }
  for(; i <= ZemPolyLgt(RX, Rg); i++) {
    ZBNSetToZero(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg)+1);
    
  }
  ZemPolyDeg(RX, Rg) =  ZemPolyDeg(PX, Rg);
}

