/*
 * zen/zem/c/poly_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsubsection{Polynomial tests}
   
   Three functions are provided to do tests: |_ZepPolyAreEqual| to test if two
   |ZENPoly| are equal, |_ZepPolyIsZero| to test if a |ZENPoly| is equal to
   zero and , |_ZepPolyIsXi| to test if a |ZENPoly| is equal to $X^i$.
   }
   */
#include <stdio.h>
#include "zem.h"

/* \TEX[
   \title{Is a polynomial equal to $X^i$.}
   \input{A polynomial |p| of a finite ring |Rg| and a degree |i|.}
   \output{$|p| == X^i$.}
   }
   */
int _ZemPolyIsXi (p, deg, Rg)
     ZENPoly p;
     int deg;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZemPolyDeg(p, Rg) != deg) {
    return(0);
  }

  for(i = 0; i < ZemPolyDeg(p, R); i++) {
    elt = ZemPolyGetCoeffPtr(p, i, Rg);
    if (!ZemEltIsZero(elt, Rg)) {
      return(0);
    }
  }
  elt = ZemPolyGetCoeffPtr(p, i, Rg);
  if (!ZemEltIsOne(elt, Rg)) {
    return(0);
  }
  
  return(1);
}
