/*
 * zen/zep/c/elt_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element additions}
   |_ZepEltAdd| enable us to add a |ZENElt| to another.
   }
   */
#include <stdio.h>
#include <zep.h>

/* \TEX{
   \title{Addition.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |Rg|.}
   \side{|*p_c| += |a|.}
   \note{Standard representation is used. One can have |*p_c| = |a|.}
   }
   */
void _ZepEltAdd(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZBNModAdd(ZENElt2Bn(*p_c,Rg), ZENRingSizeElt(Rg),
	    ZENElt2Bn(a,Rg), ZENRingSizeElt(Rg),
	    ZENRingP(Rg), ZENRingPl(Rg));
}

/* \TEX{
   \title{Addition.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |Rg|.}
   \side{|*p_c| += |a|.}
   \note{Standard representation is used. One can have |*p_c| = |a|.
   This function is used when the modulo is a power of $2^{|SIZE_BLOC|}$}
   }
   */
void _ZepEltAddMisc(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZepEltAddMisc(*p_c, a, Rg);
}

/* \TEX{
   \title{Negation.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |Rg|.}
   \side{|*p_c| = - |a|.}
   \note{Standard representation is used. One can have |*p_c| = |a|.}
   }
   */
void _ZepEltNegate(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZBNModNegate(ZepRingNbuf(Rg),ZENRingSizeElt(Rg),
	       ZENElt2Bn(a,Rg),ZENRingSizeElt(Rg),
	       ZepRingP1(Rg),ZENRingSizeElt(Rg));
  ZBNAssign(ZENElt2Bn(*p_c,Rg),
	    ZepRingNbuf(Rg),ZENRingSizeElt(Rg));
}

/* \TEX{
   \title{Subtraction.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{|*p_c| -= |a|.}
   \note{Standard representation is used. One can have |*p_c| = |a|.}
   }
   */
void _ZepEltSubtract(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZBNAssign(ZepRingNbuf(Rg),
	    ZENElt2Bn(*p_c,Rg),ZENRingSizeElt(Rg));
  ZBNModSubtract(ZepRingNbuf(Rg),ZENRingSizeElt(Rg),
		 ZENElt2Bn(a,Rg),ZENRingSizeElt(Rg),
		 ZepRingP1(Rg),ZENRingSizeElt(Rg));
  ZBNAssign(ZENElt2Bn(*p_c,Rg),
	    ZepRingNbuf(Rg),ZENRingSizeElt(Rg));
}

