/*
 * zen/zep/c/elt_alloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element creation}
   
   |_ZepEltCreate| allocates one |ZENElt| and |_ZepEltFree| frees then it.
   }
   */
#include <stdio.h>
#include "zep.h"

/* \TEX{
   \title{Creation.}
   \input{A finite ring |Rg| and a pointer on a |ZENElt|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   }
   */
int _ZepEltCreate(a, Rg)
     ZENElt *a;
     ZENRing Rg;
{
  return(ZepEltCreate(*a, Rg));
}

int _ZepEltCreateLarge(a, Rg)
     ZENElt *a;
     ZENRing Rg;
{
  return(ZepEltCreateLarge(*a, Rg));
}

/* \TEX{
   \title{Free.}
   \input{a |ZENElt| |a| of a finite ring |Rg| already allocated with
   |_ZepEltCreate|.}
   \side{|a| is freed.}
   }
   */
void _ZepEltFree(a,Rg)
     ZENElt a;
     ZENRing Rg;
{ 
  ZepEltFree(a, Rg);
  return;
}
