/*
 * zen/zep/c/elt_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{   
   \Subsection{Element inversion}
   
   |_ZepEltInverse| computes the inverse of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zep.h"

/* \TEX{
   \title{Inverse.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |R|.}
   \output{|ZENERR| if an error occured, |ZEN_HAS_INVERSE| if |a| has an inverse, |ZEN_NO_INVERSE| otherwise}
   \side{$|*p_c| = 1/a$ if |a| has an inverse; otherwise |ZENRingFact(R)| is filled with the gcd of |a| with the modulo.} 
   }
   */
int _ZepEltInverse(p_c, a, R)
     ZENElt *p_c, a;
     ZENRing R;
{BigNumLength gl,il;

 gl=ZENRingPl(R);
 il=ZENRingSizeElt(R);

 if(ZepEltIsZero(a,R))
   {
     ZENSetError(ZEN_LIB_ZEP,ZEN_INVERSE,ZEN_DIVBYZERO);
     return(ZERR);
   }
 
 switch(ZBNEea(ZENElt2Bn(ZENRingFact(R),R),&gl,
	       ZENElt2Bn(*p_c,R),&il,
	       ZENRingP(R),ZENRingPl(R),
	       ZENElt2Bn(a,R),ZBNNumDigits(ZENElt2Bn(a,R),
					   ZENRingSizeElt(R))))
   {
   case ZBN_HAS_INVERSE:
     ZBNSetToZero(ZENElt2Bn(*p_c,R)+il,ZENRingSizeElt(R)-il);
     return(ZEN_HAS_INVERSE);

   case ZBN_NO_INVERSE:
     ZBNSetToZero(ZENElt2Bn(ZENRingFact(R),R)+gl,ZENRingSizeElt(R)-gl);
     ZENSetError(ZEN_LIB_ZEP,ZEN_INVERSE,ZEN_ELT_NOTINV);
     return(ZEN_NO_INVERSE);

   default:
     return(ZERR);
   }
}

/* \TEX{
   \title{Inverse.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |R|.}
   \output{|ZENERR| if an error occured, |ZEN_HAS_INVERSE| if |a| has an inverse, |ZEN_NO_INVERSE| otherwise}
   \side{$|*p_c| = 1/a$ if |a| has an inverse; otherwise |ZENRingFact(R)| is filled with the gcd of |a| with the modulo.}
   \note{This function is used when the modulo is a power of $2^{|SIZE_BLOC|}$}
   }
   */
int _ZepEltInverseMisc(p_c, a, R)
     ZENElt *p_c, a;
     ZENRing R;
{BigNumLength gl,il;

 gl=il=ZENRingPl(R);
 switch(ZBNEea(ZENElt2Bn(ZENRingFact(R),R),&gl,
	       ZepRingIbuf(R),&il,
	       ZENRingP(R),ZENRingPl(R),
	       ZENElt2Bn(a,R),ZENRingSizeElt(R)))
   {
   case ZBN_HAS_INVERSE:
     ZBNAssign(ZENElt2Bn(*p_c,R),
	       ZepRingIbuf(R),
	       il = ZBNNumDigits(ZepRingIbuf(R), il));
     ZBNSetToZero(ZENElt2Bn(*p_c,R)+il,ZENRingSizeElt(R)-il);
     return(ZEN_HAS_INVERSE);

   case ZBN_NO_INVERSE:
     ZBNSetToZero(ZENElt2Bn(ZENRingFact(R),R)+gl,ZENRingSizeElt(R)-gl);
     ZENSetError(ZEN_LIB_ZEP,ZEN_INVERSE,ZEN_ELT_NOTINV);
     return(ZEN_NO_INVERSE);

   default:
     return(ZERR);
   }
 return(ZEN_NO_INVERSE);
}
