/*
 * zen/zep/c/elt_trace.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element trace}
   
   |_ZepEltTrace| computes the trace of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zep.h"

/* \TEX{
   \title{Trace.}
   \input{A |ZENElt| |a| of a finite ring |Rg|.}
   }
   */
void _ZepEltTrace(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZepEltTrace(*p_c, a, Rg);
}


/* \TEX{
   \title{Computing an element of trace 1.}
   \input{A finite ring |Rg|.} 
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| otherwise.}
   \side{|ZENRingTheta(Rg)| is allocated and filled with 1.}
   }
   */
int _ZepPrcInitPolyRootsCanonical(Rg)
     ZENRing Rg;
{
  if (ZepEltCreate(ZENRingTheta(Rg), Rg)) {
    ZENSetError(ZEN_LIB_ZEP, ZEN_ADDPRC, ZEN_ALLOC);
    return(ZERR);
  }
  ZepEltSetToOne(ZENRingTheta(Rg), Rg);

  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Desallocate the element of trace 1.}
   \input{A finite ring |Rg|.} 
   \output{|ZEN_HAS_INVERSE|.}
   \side{|ZENRingTheta(Rg)| is desallocated.}
   }
   */
int _ZepPrcClosePolyRootsCanonical(Rg)
     ZENRing Rg;
{
  ZepEltFree(ZENRingTheta(Rg), Rg);
  
  return(ZEN_HAS_INVERSE);
}
