/*
 * zen/zep/c/mat_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial assignments}

   }
   */
#include <stdio.h>
#include "zep.h"

/*
   \TEX{
   
   \title{Assignment.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _ZepMatAssign(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;

 ZENTest(ZepMatIsPermutation(A,R)
	 ||ZepMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATASSIGN,ZEN_LIB_ZEP, ZERR);
 
 if(ZepMatAreSameType(A,B,R))
   ZepMatSameTypeAssign(A,B,R);
 else
   {
     ZENTest((ZepMatNbBn(A,R) != ZepMatNbDigit(B,R))
	     ||(ZepMatNbDigit(A,R) != ZepMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATASSIGN,ZEN_LIB_ZEP, ZERR);
     
     for(l=ZepMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZepMatNbDigit(A,R)-1;c>=0;c--)
	 ZepEltAssign(ZENMatBloc(A,l,c,R),ZENMatBloc(B,c,l,R),R);
   }
 return(0);
}

/*
   \TEX{
   \title{Set to zero}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are set to zero.}
   }
   */
   
int _ZepMatSetToZero(A,R)
     ZENMat A;
     ZENRing R;
{
 ZENTest(ZepMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETZERO,ZEN_LIB_ZEP, ZERR);

 ZepMatSetToZero(A,R);
 
 return(0);
}

/*
   \TEX{
   \title{Set to one}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are set to zero except for the coefficients with same indices on rows and columns that are set to one.}
   \note{There is no error if the matrix is not square.}
   }
   */
   
int _ZepMatSetToOne(A,R)
     ZENMat A;
     ZENRing R;
{int l;

 ZENTest(ZepMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETONE,ZEN_LIB_ZEP, ZERR);

 ZepMatSetToZero(A,R);
#define min(a,b) ((a)<(b)?a:b)
 l=min(ZepMatNbBn(A,R),ZepMatNbDigit(A,R))-1;
 for(;l>=0;l--)
   ZENEltSetToOne(ZENMatBloc(A,l,l,R),R);
     
 return(0);
}


