/*
 * zen/zep/c/mat_create.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Allocating a matrix}
   }
   */


#include <stdio.h>
#include "zep.h"
#include "zed.h"

/* \TEX{
   \title{Computation of the matrix parameters}
   \input{A |ZENMatTypeStruct|, the dimensions of the desired matrix, a pointer on a |ZENMat| and a |ZENRing|.}
   \output{0 if no error occured, |ZENERR| otherwise.}
   \side{The |ZENMat| structure is allocated and the parameters computed according to the type of the desired matrix. Array is allocated.}
   \note{This function should not be directly used (see |ZENMatCreate|). If chosen in a |ZENRing| initialization  routine, |_ZedGetBloc|, |_ZedGetBlocPtr| and |_ZedSetBloc| must also be chosen.}
   }
   */   

int _ZepMatCreateParams(ts,l,c,p_pms,R)
     Dim l,c;
     ZENMatTypeStruct ts;
     ZENMat *p_pms;
     ZENRing R;
{
  int ret;

  ret=_ZedMatCreateParams(ts,l,c,p_pms,R);
  if(ret==ZENERR)
    {
      ZENSetError(ZEN_LIB_ZEP,ZEN_MATCREPMS,ZEN_UNKNOWN);
      return(ZERR);
    }
  ZENMatSize(*p_pms,R)=ZENRingSizeElt(R);
  ZENMatNbPtr(*p_pms,R)=l*c;
  ZENMatNbBnd(*p_pms,R)=ZENMatNbPtr(*p_pms,R)*ZENRingSizeElt(R);
  return(ret);
}
