/*
 * zen/zep/c/mat_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   
   \Subsection{Tests on a matrix}
   }
   */


#include <stdio.h>
#include "zep.h"
/* \TEX{
   \title{Test of equality}
   \input{Two matrices and a |ZENRing|}
   \output{1 if the matrices are equal, 0 otherwise}
   }
   */

int _ZepMatAreEqual(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;
 
 if(ZepMatAreSameType(A,B,R))
   return(ZepMatSameTypeAreEqual(A,B,R));
 
 if((ZepMatNbBn(A,R) != ZepMatNbDigit(B,R))
    ||(ZepMatNbDigit(A,R) != ZepMatNbBn(B,R)))
   return(0);
 for(l=ZepMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZepMatNbDigit(A,R)-1;c>=0;c--)
     if(!(ZepEltAreEqual(ZENMatBloc(A,l,c,R),ZENMatBloc(B,c,l,R),R)))
       return(0);
 return(1);
}

/* \TEX{
   \title{Test of equality to zero}
   \input{A matrix and a |ZENRing|}
   \output{1 if the matrix is null, 0 otherwise}
   }
   */

int _ZepMatIsZero(A,R)
     ZENMat A;
     ZENRing R;
{
  return(ZepMatIsZero(A,R));
}

/* \TEX{
   \title{Test of equality to identity}
   \input{A matrix and a |ZENRing|}
   \output{1 if the matrix is identity, 0 otherwise}
   \note{The matrix is checked to be square.} 
   }
   */

int _ZepMatIsOne(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;

 if(ZepMatNbBn(A,R) != ZepMatNbDigit(A,R))
   return(0);
 
 for(l=ZepMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZepMatNbBn(A,R)-1;c>=0;c--)
     if(l==c)
       {
	 if(!(ZENEltIsOne(ZENMatBloc(A,l,c,R),R)))
	   return(0);
       }
     else
       {
	 if(!(ZepEltIsZero(ZENMatBloc(A,l,c,R),R)))
	   return(0);
       }
 return(1); 
}


