/*
 * zen/zep/c/poly_alloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsubsection{Polynomial creation}
   
   |_ZepPolyCreate| allocates one |ZENPoly| and |_ZepPolyFree| frees then it.
   }
   */
#include <stdio.h>
#include "zep.h"

/* \TEX{
   \title{Creation.}
   \input{A finite ring |Rg| and a degree |deg|.}
   \output{An allocated |ZENPoly| set to 0 with |deg| coefficients allocated. }
   }
   */
ZENPoly _ZepPolyCreate(deg, Rg)
     int deg;
     ZENRing Rg;
{
  ZENPoly PX;
  int i;

  
  if (!(PX = ZENPolyAllocStruct(Rg))) {
    ZENSetError(ZEN_LIB_ZEP, ZEN_POLYCREATE, ZEN_ALLOC);
    return(ZNULL);
  }
  if (!(ZepPolyCof(PX, Rg) = (ZENElt *)ZENMalloc(sizeof(ZENElt)*(deg+1)))) {
    ZENFree(PX);
    ZENSetError(ZEN_LIB_ZEP, ZEN_POLYCREATE, ZEN_ALLOC);
    return(ZNULL);
  }
  for(i = 0; i <= deg; i++) {
    if (ZepEltCreate(ZepPolyGetCof(PX, i, Rg), Rg)) {
      for(i--; i>= 0; i--) {
	ZepEltFree(ZepPolyGetCof(PX, i, Rg), Rg);
      }
      ZENFree(ZepPolyCof(PX, Rg));
      ZENFree(PX);
      return(ZNULL);
    }
  }
  ZepPolyLgt(PX, Rg) = deg;
  return(PX);
}

ZENPoly _ZepPolyCreateLarge(deg, Rg)
     int deg;
     ZENRing Rg;
{
  ZENPoly PX;
  int i;

  
  if (!(PX = ZENPolyAllocStruct(Rg))) {
    ZENSetError(ZEN_LIB_ZEP, ZEN_POLYCREATE, ZEN_ALLOC);
    return(ZNULL);
  }
  if (!(ZepPolyCof(PX, Rg) = (ZENElt *)ZENMalloc(sizeof(ZENElt)*(deg+1)))) {
    ZENFree(PX);
    ZENSetError(ZEN_LIB_ZEP, ZEN_POLYCREATE, ZEN_ALLOC);
    return(ZNULL);
  }
  for(i = 0; i <= deg; i++) {
    if (ZepEltCreateLarge(ZepPolyGetCof(PX, i, Rg), Rg)) {
      for(i--; i>= 0; i--) {
	ZepEltFree(ZepPolyGetCof(PX, i, Rg), Rg);
      }
      ZENFree(ZepPolyCof(PX, Rg));
      ZENFree(PX);
      return(ZNULL);
    }
  }
  ZepPolyLgt(PX, Rg) = deg;
  return(PX);
}

/* \TEX{
   \title{Free.}
   \input{a |ZENPoly| |PX| of a finite ring |Rg| already allocated with
   |_ZepPolyCreate|.}
   \side{|PX| is freed.}
   }
   */
void _ZepPolyFree(PX,Rg)
     ZENPoly PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZepPolyLgt(PX, Rg); i++) {
    ZepEltFree(ZepPolyGetCof(PX, i, Rg), Rg);
  }
  ZENFree(ZepPolyCof(PX, Rg));
  ZENFree(PX);
}


