/*
 * zen/zep/c/poly_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsubsection{Polynomial tests}
   
   Three functions are provided to do tests: |_ZepPolyAreEqual| to test if two
   |ZENPoly| are equal, |_ZepPolyIsZero| to test if a |ZENPoly| is equal to
   zero and , |_ZepPolyIsXi| to test if a |ZENPoly| is equal to $X^i$.
   }
   */
#include <stdio.h>
#include "zep.h"

/* \TEX{
   \title{Equal.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{|RX| == |PX|.}
   }
   */
int _ZepPolyAreEqual(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  if (ZepPolyDeg(RX, Rg) != ZepPolyDeg(PX, Rg)) return(0);

  for(i = 0; i <= ZepPolyDeg(RX, Rg); i++) {
    if (!ZepEltAreEqual(ZepPolyGetCof(RX, i, Rg),
			 ZepPolyGetCof(PX, i, Rg),
			 Rg)) {
      return(0);
    }
  }
  return(1);
}

/* \TEX{
   \title{Is zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \output{|RX| == 0.}
   }
   */
int _ZepPolyIsZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  return(!ZepPolyDeg(RX, Rg) &&
	 ZepEltIsZero(ZepPolyGetCof(RX, 0, Rg), Rg));
}

/* \TEX{
   \title{Is a polynomial equal to $X^i$.}
   \input{A polynomial |p| of a finite ring |Rg| and a degree |i|.}
   \output{$|p| == X^i$.}
   }
   */
int _ZepPolyIsXi (p, deg, Rg)
     ZENPoly p;
     int deg;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZepPolyDeg(p, Rg) != deg) {
    return(0);
  }

  for(i = 0; i < ZepPolyDeg(p, R); i++) {
    elt = ZepPolyGetCoeffPtr(p, i, Rg);
    if (!ZepEltIsZero(elt, Rg)) {
      return(0);
    }
  }
  elt = ZepPolyGetCoeffPtr(p, i, Rg);
  if (!ZepEltIsOne(elt, Rg)) {
    return(0);
  }
  
  return(1);
}
