/*
 * zen/zeps/c/elt_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element inversion}
   
   |_ZepsEltInverse| computes the inverse of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zeps.h"

/* \TEX{
   \title{Inverse.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|*p_c| = 1/a$ if |a| has an inverse, ZENRingFact(Rg) is
   filled with a factor of the modulo otherwise.} 
   }
   */
int _ZepsEltInverse(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  BigNumDigit x, u0, b, q, pg, inv;
  int d;

  if (!ZepsElt2Int(a, Rg)) {
    ZENSetError(ZEN_LIB_ZEPS, ZEN_INVERSE, ZEN_DIVBYZERO);
    return(ZERR);
  }

  x = ZepsElt2Int(a, Rg);
  u0 = 1;
  inv = 0;
  d = -1;
  pg = ZBNDigit(ZENRingP(Rg), 0);
  while(pg > 1){
    d = -d;
    q = x / pg;
    b = u0 + inv * q;
    u0 = inv;
    inv = b;
    b = pg;
    pg = x - q * pg;
    x = b;
  }
  if(pg == 1){
    if(d < 0) {
      ZepsElt2Int(*p_c, Rg) = ZBNDigit(ZENRingP(Rg), 0) - inv;
    }
    else {
      ZepsElt2Int(*p_c, Rg) = inv;
    }
    return(ZEN_HAS_INVERSE);
  }

  ZepsElt2Int(ZENRingFact(Rg), Rg) = b;
  ZENSetError(ZEN_LIB_ZEPS,ZEN_INVERSE,ZEN_ELT_NOTINV);
  return(ZEN_NO_INVERSE);
}
