/*
 * zen/zeps/c/mat_convert.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zeps.h"

/*
   \TEX{
   \Subsection{Matrix conversion}
   The following functions are intended to allow conversion of matrices between an extension and its definition |ZENRing|. In all the following functions, matrix |M| is defined over the extension |R| and |m| is defined over the underlevel extension |ZENRingDef(R)|.

   \title{Extension conversion}
   \input{Two matrices |ZENMat| and a non tabulated extension |ZENRing| over $\Z/p\Z$ with $p$ small.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |m| is assigned with |M|.}
   } 
   */
int _ZepmMat2Vect(M,m,R)
     ZENMat M,m;
     ZENRing R;
{int c,l;
 ZENElt *ptm,*ptM;

 ZENTest(ZENRingDef(R)==NULL,
	 ZEN_NOT_EXT,ZEN_MAT2VECT,ZEN_LIB_ZEPS, ZERR);

 ZENTest(ZENMatIsPermutation(M,R)
	 ||ZENMatIsPermutation(m,ZENRingDef(R)),
	 ZEN_ISPERM,ZEN_MAT2VECT,ZEN_LIB_ZEPS, ZERR);
 
 ZENTest(ZENMatNbBnd(M,R) != ZENMatNbBnd(m,ZENRingDef(R)),
	 ZEN_MAT_DIM,ZEN_MAT2VECT,ZEN_LIB_ZEPS, ZERR);

 if(ZENMatAreSameType(m,M,R))
   ZBNAssign(ZENMatFirst(M,R),ZENMatFirst(m,ZENRingDef(R)),ZENMatNbBnd(M,R));
 else
   {
     ptm=(ZENElt *)ZENMatFirst(m,ZENRingDef(R));
     ptM=(ZENElt *)ZENMatFirst(M,R);
     for(l=ZENMatNbBn(M,R);
	 l--;
	 ptm -= ZENMatNbBn(m,ZENRingDef(R))*ZENMatNbDigit(m,ZENRingDef(R)) - ZENMatSize(M,R))
       for(c=ZENMatNbDigit(M,R);
	   c--;
	   ptM += ZENMatSize(M,R),ptm += ZENMatNbBloc(m,ZENRingDef(R)))

	 ZBNAssign((BigNum)ptM,(BigNum)ptm,(BigNumLength)ZENMatSize(M,R));
   }
 ptM=(ZENElt *)ZENMatFirst(M,R)-1;
 for(l=ZENMatNbBn(M,R)*ZENMatNbDigit(M,R);l--;ptM--)
   {
     ZENPolyDeg(ZENElt2Pol(*ptM,R),
		ZENRingDef(R))=ZENPolyLgt(ZENElt2Pol(*ptM,R),
					  ZENRingDef(R));
     ZENPolyUpdateDegree(ZENElt2Pol(*ptM,R),ZENRingDef(R));
   }
 return(0);
}

/*     
   \TEX{
   \title{Extension conversion}
   \input{Two matrices |ZENMat| and a non tabulated extension |ZENRing| over $\Z/p\Z$ with $p$ small.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |M| is assigned with |m|.}
   } 
   */
int _ZepmVect2Mat(m,M,R)
     ZENMat M,m;
     ZENRing R;
{int l,c;
 ZENElt *pt,*ptm,*ptM;

 ZENTest(ZENRingDef(R)==NULL,
	 ZEN_NOT_EXT,ZEN_VECT2MAT,ZEN_LIB_ZEPS, ZERR);

 ZENTest(ZENMatIsPermutation(M,R)
	 ||ZENMatIsPermutation(m,ZENRingDef(R)),
	 ZEN_ISPERM,ZEN_VECT2MAT,ZEN_LIB_ZEPS, ZERR);
     
 ZENTest(ZENMatNbBnd(M,R) != ZENMatNbBnd(m,ZENRingDef(R)),
	 ZEN_MAT_DIM,ZEN_MAT2VECT,ZEN_LIB_ZEPS, ZERR);

 ptm=(ZENElt *)ZENMatFirst(m,ZENRingDef(R));
 ptM=(ZENElt *)ZENMatFirst(M,R);
 pt=(ZENElt *)ZENMatFirst(M,R)-ZENMatNbBn(M,R)*ZENMatNbDigit(M,R);
 if(ZENMatAreSameType(m,M,R))
   for(l=ZENMatNbBn(M,R)*ZENMatNbDigit(M,R);l--;pt++)
     {
       ZBNAssign((BigNum)ptm,(BigNum)ptM,(BigNumLength)(1 + ZENPolyDeg(ZENElt2Pol(*pt,R),ZENRingDef(R))));
       ZBNSetToZero((BigNum)(ptm) + 1 + ZENPolyDeg(ZENElt2Pol(*pt,R),ZENRingDef(R)),
		    ZENPolyLgt(ZENElt2Pol(*pt,R),ZENRingDef(R))
		    -ZENPolyDeg(ZENElt2Pol(*pt,R),ZENRingDef(R)));
       ptm += ZENMatSize(M,R);
       ptM += ZENMatSize(M,R);
     }
 else
   for(l=ZENMatNbBn(M,R);
       l--;
       ptm -= ZENMatNbBn(m,ZENRingDef(R))*ZENMatNbDigit(m,ZENRingDef(R)) - ZENMatSize(M,R))
     for(c=ZENMatNbDigit(M,R);
	 c--;
	 pt++,ptm+=ZENMatNbBloc(m,ZENRingDef(R)))
       {
	 ZBNAssign((BigNum)ptm,(BigNum)ptM,(BigNumLength)(1 + ZENPolyDeg(ZENElt2Pol(*pt,R),ZENRingDef(R))));
	 ZBNSetToZero((BigNum)(ptm) + 1 + ZENPolyDeg(ZENElt2Pol(*pt,R),ZENRingDef(R)),
		      ZENPolyLgt(ZENElt2Pol(*pt,R),ZENRingDef(R))
		      -ZENPolyDeg(ZENElt2Pol(*pt,R),ZENRingDef(R)));
	 ptM += ZENMatSize(M,R);
       }
 return(0);
}

