/*
 * zen/zeps/c/poly_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial additions}
   
   One function is provided to do additions: |_ZepsPolyAdd| to add a
   |ZENPoly| to another.
   }
   */
#include <stdio.h>
#include "zeps.h"

/* \TEX{
   \title{Addition.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| += |PX|.}
   }
   */
void _ZepsPolyAdd(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;

  if (ZepsPolyDeg(RX, Rg) < ZepsPolyDeg(PX, Rg)) {
    mindeg = ZepsPolyDeg(RX, Rg);
    ZepsPolyDeg(RX, Rg) =  ZepsPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZepsPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZepsEltAdd(ZepsPolyGetCof(RX, i, Rg), ZepsPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZepsPolyDeg(PX, Rg); i++) {
    ZepsEltAssign(ZepsPolyGetCof(RX, i, Rg),
		  ZepsPolyGetCof(PX, i, Rg),
		  Rg);
  }
  ZepsPolyUpdateDegree(RX, Rg);
}

/* \TEX{
   \title{Negation.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| = -|PX|.}
   }
   */
void _ZepsPolyNegate(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZepsPolyDeg(PX, Rg); i++) {
    ZepsEltNegate(ZepsPolyGetCof(RX, i, Rg), ZepsPolyGetCof(PX, i, Rg), Rg);
  }
  ZepsPolyDeg(RX, Rg) =  ZepsPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Subtraction.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| -= |PX|.}
   }
   */
void _ZepsPolySubtract(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, mindeg;

  if (ZepsPolyDeg(RX, Rg) < ZepsPolyDeg(PX, Rg)) {
    mindeg = ZepsPolyDeg(RX, Rg);
    ZepsPolyDeg(RX, Rg) =  ZepsPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZepsPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZepsEltSubtract(ZepsPolyGetCof(RX, i, Rg), ZepsPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZepsPolyDeg(PX, Rg); i++) {
    ZepsEltNegate(ZepsPolyGetCof(RX, i, Rg),
		   ZepsPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZepsPolyUpdateDegree(RX, Rg);
}
