/*
 * zen/zeps/c/poly_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial tests}
   
   Three functions are provided to do tests: |_ZepsPolyAreEqual| to test if two
   |ZENPoly| are equal, |_ZepsPolyIsZero| to test if a |ZENPoly| is equal to
   zero and , |_ZepsPolyIsXi| to test if a |ZENPoly| is equal to $X^i$.
   }
   */
#include <stdio.h>
#include "zeps.h"

/* \TEX{
   \title{Equal.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{|RX| == |PX|.}
   }
   */
int _ZepsPolyAreEqual(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  if (ZepsPolyDeg(RX, Rg) != ZepsPolyDeg(PX, Rg)) return(0);

  for(i = 0; i <= ZepsPolyDeg(RX, Rg); i++) {
    if (!ZepsEltAreEqual(ZepsPolyGetCof(RX, i, Rg),
			 ZepsPolyGetCof(PX, i, Rg),
			 Rg)) {
      return(0);
    }
  }
  return(1);
}

/* \TEX{
   \title{Is zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \output{|RX| == 0.}
   }
   */
int _ZepsPolyIsZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  return(!ZepsPolyDeg(RX, Rg) &&
	 ZepsEltIsZero(ZepsPolyGetCof(RX, 0, Rg), Rg));
}

/* \TEX{
   \title{Is a polynomial equal to $X^i$.}
   \input{A polynomial |p| of a finite ring |Rg| and a degree |i|.}
   \output{$|p| == X^i$.}
   }
   */
int _ZepsPolyIsXi (p, deg, Rg)
     ZENPoly p;
     int deg;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZepsPolyDeg(p, Rg) != deg) {
    return(0);
  }

  for(i = 0; i < ZepsPolyDeg(p, R); i++) {
    elt = ZepsPolyGetCoeffPtr(p, i, Rg);
    if (!ZepsEltIsZero(elt, Rg)) {
      return(0);
    }
  }
  elt = ZepsPolyGetCoeffPtr(p, i, Rg);
  if (!ZepsEltIsOne(elt, Rg)) {
    return(0);
  }
  
  return(1);
}
