/*
 * zen/zer/c/base_ring.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   \Section{Primitive ring initialization}
   }
   */
#include <stdio.h>
#include <zen.h>
#include <zer.h>

/********************************************************/

/* \TEX{
   \title{Initialization of a prime ring.}
   \input{A modulo (|n|, |nl|).} 
   \output{A ring |R| if everything was succesfully initialized, |ZENNULL| otherwise.}
   \note{If $N=0$, the |ZENRing| is a representation of the infinite field $\Q$. This feature is experimental.}
   }
   */
ZENRing _ZENBaseRingCreate(n, nl)
     BigNum n;
     BigNumLength nl;
{ZENRing R;
 int ret;
 
 nl = ZBNNumDigits(n, nl);

 if ((R = __ZENRingCreate(n,nl)) == ZENNULL)
   return(ZNULL);
 
 ZENRingQl(R) = nl;
 if (!(ZENRingQ(R) = ZBNC(nl))) {
   ZBNF(ZENRingP(R)); ZENFree(R);
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
   return(ZNULL);
 }

 ZBNAssign(ZENRingQ(R), n, nl);
 ZBNSubtractBorrow(ZENRingQ(R), nl, 0);
 nl = ZBNNumDigits(ZENRingQ(R), nl);
 ZENRingSizeElt(R) = nl; 
 ZENRingSZ(R) = 2*nl+1;
 if (!(ZENRingP1(R) = ZBNC(ZENRingSizeElt(R)))) {
   ZENFree(R);
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
   return(ZNULL);
 }
 ZBNAssign(ZENRingP1(R), ZENRingQ(R), ZENRingSizeElt(R));
 ZBNAssign(ZENRingQ(R), n, ZENRingQl(R));

 ZENRingPol(R) = NULL;
 ZENRingDef(R) = NULL;
 ZENRingOrigin(R) =  NULL;

 ZENRingVar(R)=0;
 
 /* Procedure on rings
  **********************/
 ZENPrcInitTracePtr(R) = NULL;
 ZENPrcCloseTracePtr(R) = NULL;
 
 /* We now complete the construction */
 
 ret=ZENERR-1;
 ZENSetError(ZEN_LIB_ZEP, ZEN_INIT, ZEN_UNAVAILABLE);
#if ZE2_AVAILABLE || ZEF_AVAILABLE || ZEPS_AVAILABLE
 if (ZENRingPl(R) == 1)		/* And not nl == 1 (take care of P=2^BLOCK_SIZE) */
   {
# if ZE2_AVAILABLE
     if(ZBNDigit(n, 0) == 2)
       ret=_Ze2Init(R);
     else
# endif
       {
# if ZEF_AVAILABLE
	 if(ZBNDigit(n, 0) == 0)
	   ret=_ZefInit(R);
	 else
# endif
# if ZEPS_AVAILABLE
       ret=_ZepsInit(R)
# endif
	 ;
       }
   }
 else
#endif

   {
#if ZEP_AVAILABLE
     ret=_ZepInit(R);
#endif
   }
 if(ret<0)
   {
     if(ret == ZENERR) 
       ZENFree(R);
     return(ZNULL);
   }
 
 if(ret & POLY_GEN)
   __ZENPolyGeneric(R);
 if(ret & MAT_GEN)
   __ZENMatGeneric(R);
 if (ZENRingArith(R) == ZEN_LIB_ZEF) {
   ZENMatMultiplyPtr(R) = _ZENMatMultiplyPlain; 
 }
 if(ret & SR_GEN)
   __ZENSrGeneric(R);

 /* Procedure on elements
  ****************************/
 /* Assignation */
 ZENEltToClonePtr(R) = ZENEltAssignPtr(R);
 ZENCloneToEltPtr(R) = ZENEltAssignPtr(R);

 /* Procedure on matrices
  ****************************/
  /* Matrix conversion */
  ZENVect2MatPtr(R) = ZENMatAssignPtr(R);
  ZENMat2VectPtr(R) = ZENMatAssignPtr(R);


 if(__ZENRingInitBuffers(R)) {
   ZBNF(ZENRingP(R));
   ZBNF(ZENRingQ(R));
   ZENFree(R);
   return(ZNULL);
 }
 
 return(R);
}

/**********************************************************************/
int __ZENBaseClose(R)
     ZENRing R;
{
  switch(ZENRingArith(R)) {
  case ZEN_LIB_ZE2:
    if(_Ze2Close(R))
      return(ZERR);
    break;
  case ZEN_LIB_ZEP:
    if(_ZepClose(R))
      return(ZERR);
    break;
  case ZEN_LIB_ZEF:
    if(_ZefClose(R))
      return(ZERR);
    break;
  case ZEN_LIB_ZEPS:
    if(_ZepsClose(R))
      return(ZERR);
    break;
  default:
    return(ZERR);
  }
  return(0);
}

