/*
 * zen/zer/c/test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

 /* \TEX{
   We provide a test in order to check if two rings are mathematically identical. This is different from the simpler test of equality between pointers. However, two isomorphic rings are not equal under this test.
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Testing whether two rings are equal.}
   \input{Two |ZENRings|.}
   \output{|ZENERR| if an error occured, 1 if the two |ZENRings| are the same, 0 otherwise.}
   \note{A clone is equal to its original ring.}
   }
   */
int _ZENRingAreEqual(R1,R2)
     ZENRing R1,R2;  
{
  switch(ZENRingArith(R1))
    {
    case ZEN_LIB_ZETAB:
    case ZEN_LIB_ZELOG:
    case ZEN_LIB_ZEC:
    case ZEN_LIB_ZEM:
      R1=ZENRingOrigin(R1);
    default:
      ;
    }
  switch(ZENRingArith(R2))
    {
    case ZEN_LIB_ZETAB:
    case ZEN_LIB_ZELOG:
    case ZEN_LIB_ZEC:
    case ZEN_LIB_ZEM:
      R2=ZENRingOrigin(R2);
    default:
      ;
    }

  switch(ZENRingArith(R1))
    {
    case ZEN_LIB_ZEF:
      return(ZENRingArith(R2)==ZEN_LIB_ZEF ? 1 : 0 );

    case ZEN_LIB_ZE2:
    case ZEN_LIB_ZEPS:
    case ZEN_LIB_ZEP:
      switch(ZENRingArith(R2))
	{
	case ZEN_LIB_ZE2:
	case ZEN_LIB_ZEPS:
	case ZEN_LIB_ZEP:
	  return(ZBNAreEqual(ZENRingP(R1),ZENRingPl(R1),
			     ZENRingP(R2),ZENRingPl(R2)));
	default:
	  return(0);
	}
    case ZEN_LIB_ZEXT:
      if(ZENRingArith(R2)!=ZEN_LIB_ZEXT)
	return(0);
      if(!ZENRingAreEqual(ZENRingDef(R1),ZENRingDef(R2)))
	return(0);
      else
	{
	  BigNum p1,p2;
	  int ret;
	  BigNumLength pl1,pl2;

	  p1=ZENPolyToZ(&pl1,ZENRingPol(R1),ZENRingDef(R1));
	  if(p1==ZENNULL)
	    return(ZERR);
	  p2=ZENPolyToZ(&pl2,ZENRingPol(R2),ZENRingDef(R2));	  
	  if(p2==ZENNULL)
	    {
	      ZBNF(p1);
	      return(ZERR);
	    }
	  ret=ZBNAreEqual(p1,pl1,p2,pl2);
	  ZBNF(p1);
	  ZBNF(p2);
	  return(ret);
	}
    default:
      ZENSetError(ZEN_LIB_ZER,ZEN_UNKNOWN,ZEN_UNKNOWN);
      return(ZERR);
    }
}

