/*
 * zen/zetab/c/elt_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Element assignment}
   
   One can set a |ZENElt| to zero or one with |_ZetabEltSetToZero| or
   |_ZetabEltSetToOne|. Then we can assign a |ZENElt| with an other with
   |_ZetabEltAssign|. Furthermore, we can set randomly a |ZENElt| with
   |_ZetabEltSetRandom| or enumerate the elements of a finite ring with
   |_ZetabEltSetNext|.
   }
   */
#include <stdio.h>
#include "zetab.h"

/* \TEX{
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| and a |ZENElt| |b| of a finite ring |Rg|.}
   \side{|*p_a| is filled with |b|.}
   }
   */
void _ZetabEltAssign(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  ZetabEltAssign(*p_a, b, Rg);
}

/* \TEX{
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| of a clone |Rg| and a |ZENElt| |b| of the original ring of |Rg|.}
   \side{|*p_a| is filled with |b|.}
   }
   */
void _ZetabEltToClone(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  BigNumLength i=1;
  BigNumDigit C[1];

  ZENEltToZ(C, &i, b, ZetabRingOrigin(Rg));
  ZetabElt2Int(*p_a, Rg) = C[0];
}

/* \TEX{
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| of the original ring of |Rg| and a |ZENElt| |b| of |Rg|.}
   \side{|*p_a| is filled with |b|.}
   }
   */
void _ZetabCloneToElt(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  ZENEltAssign(*p_a, ZetabConvTable(Rg, ZetabElt2Int(b, Rg)), ZetabRingOrigin(Rg));
}

/* \TEX{
   \title{Set to zero.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set to zero..}
   }
   */
void _ZetabEltSetToZero(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZetabEltSetToZero(*p_a, Rg);
}

/* \TEX{
   \title{Set to one.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set to one.}
   }
   */
void _ZetabEltSetToOne(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZetabEltSetToOne(*p_a, Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set randomly.}
   }
   */
void _ZetabEltSetRandom(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZetabEltSetRandom(*p_a, Rg);
}

/* \TEX{
   \title{Enumerate.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set randomly.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \note{After a number of call of this function equal to the number of elements in the finite ring, we obtain the same element.}
   }
   */
void _ZetabEltSetNext(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZetabEltSetNext(*p_a, Rg);
}

