/*
 * zen/zetab/c/elt_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element inversion}
   
   |_ZetabEltInverse| computes the inverse of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zetab.h"

/* \TEX{
   \title{Inverse.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|*p_c| = 1/a$ if |a| has an inverse, ZENRingFact(Rg) is NOT
   filled with a factor of the modulo otherwise.} 
   }
   */

int _ZetabEltInverse(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZENElt G;
  int ret;

  if (ZetabEltIsZero(a, Rg)) {
    ZENSetError(ZEN_LIB_ZETAB, ZEN_INVERSE, ZEN_DIVBYZERO);
    return(ZERR);
  }

  if (!(ZetabElt2Int(*p_c, Rg) =
	ZetabInvTable(Rg, ZetabElt2Int(a, Rg)))) {

    if (ZENEltCreate(G, ZetabRingOrigin(Rg))) {
      ZENSetError(ZEN_LIB_ZETAB, ZEN_INVERSE, ZEN_ALLOC);
      return(ZERR);
    }

    ZENCloneToElt(G, a, Rg);
    ret = ZENEltInverse(G, G, ZetabRingOrigin(Rg));
    
    ZENEltFree(G, ZetabRingOrigin(Rg));
    return(ret);
  }

  return(ZEN_HAS_INVERSE);
}

int __ZetabEltInverse(p_r,a,Rg)
     BigNumDigit *p_r, a;
     ZENRing Rg;
{
  int ret;
  ZENElt R, A;

  ZetabElt2Int(A, Rg) = a;
  ret = _ZetabEltInverse(&R, A, Rg);
  *p_r = ZetabElt2Int(R, Rg);
  return(ret);
}


