/*
 * zen/zetab/c/mat_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial additions}
   }
   */
#include <stdio.h>
#include "zetab.h"

/*
   \TEX{
   \title{Addition.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| += |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _ZetabMatAdd(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;

 ZENTest(ZetabMatIsPermutation(A,R)
	 ||ZetabMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATADD,ZEN_LIB_ZETAB, ZERR);

 if(ZetabMatAreSameType(A,B,R))
   {

     ZENTest((ZetabMatNbBn(A,R) != ZetabMatNbBn(B,R))
	     ||(ZetabMatNbDigit(A,R) != ZetabMatNbDigit(B,R)),
	     ZEN_MAT_DIM,ZEN_MATADD,ZEN_LIB_ZETAB, ZERR);

     for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
	 ZetabAnyEltAdd(ZetabMatBloc(A,l,c,R),ZetabMatBloc(B,l,c,R),R);
   }
 else
   {

     ZENTest((ZetabMatNbBn(A,R) != ZetabMatNbDigit(B,R))
	     ||(ZetabMatNbDigit(A,R) != ZetabMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATADD,ZEN_LIB_ZETAB, ZERR);
     
     for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
	 ZetabAnyEltAdd(ZetabMatBloc(A,l,c,R),ZetabMatBloc(B,c,l,R),R);

   }

 return(0);
}

   
/*
   \TEX{
   \title{Addition of scalar}
   \input{A |ZENMat|, a |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{All the elements of |A| are incremented by |b|.}
   \note{|A| cannot be a permutation.}
   }
   */
int _ZetabMatAddScalar(A,b,R)
     ZENMat A;
     ZENElt b;
     ZENRing R;
{
  int l,c;

  ZENTest(ZetabMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATADD,ZEN_LIB_ZETAB, ZERR);

     
  for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
    for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
      ZetabAnyEltAdd(ZetabMatBloc(A,l,c,R),ZetabElt2Int(b,R),R);
  
  return(0);
}

 /* \TEX{
   \title{Negation.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = -|B| if no error occured, |zen_errno| is set otherwise.}
   }
   */

int _ZetabMatNegate(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;

 ZENTest(ZetabMatIsPermutation(A,R)
	 ||ZetabMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATNEGATE,ZEN_LIB_ZETAB, ZERR);

 if(ZetabMatAreSameType(A,B,R))
   {

     ZENTest((ZetabMatNbBn(A,R) != ZetabMatNbBn(B,R))
	     ||(ZetabMatNbDigit(A,R) != ZetabMatNbDigit(B,R)),
	     ZEN_MAT_DIM,ZEN_MATNEGATE,ZEN_LIB_ZETAB, ZERR);

     for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
	   ZetabAnyEltNegate(ZetabMatBloc(A,l,c,R),ZetabMatBloc(B,l,c,R),R);

   }
 else
   {

     ZENTest((ZetabMatNbBn(A,R) != ZetabMatNbDigit(B,R))
	     ||(ZetabMatNbDigit(A,R) != ZetabMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATNEGATE,ZEN_LIB_ZETAB, ZERR);

     for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
	 ZetabAnyEltNegate(ZetabMatBloc(A,l,c,R),ZetabMatBloc(B,c,l,R),R);
	 
   }

 return(0);
}

/*
   \TEX{
   \title{Subtraction.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| -= |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _ZetabMatSubtract(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;

 ZENTest(ZetabMatIsPermutation(A,R)
	 ||ZetabMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATSUBTRACT,ZEN_LIB_ZETAB, ZERR);

 if(ZetabMatAreSameType(A,B,R))
   {

     ZENTest((ZetabMatNbBn(A,R) != ZetabMatNbBn(B,R))
	     ||(ZetabMatNbDigit(A,R) != ZetabMatNbDigit(B,R)),
	     ZEN_MAT_DIM,ZEN_MATSUBTRACT,ZEN_LIB_ZETAB, ZERR);

     for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
	 ZetabAnyEltSubtract(ZetabMatBloc(A,l,c,R),ZetabMatBloc(B,l,c,R),R);
   }
 else
   {

     ZENTest((ZetabMatNbBn(A,R) != ZetabMatNbDigit(B,R))
	     ||(ZetabMatNbDigit(A,R) != ZetabMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATSUBTRACT,ZEN_LIB_ZETAB, ZERR);

     for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
	 ZetabAnyEltSubtract(ZetabMatBloc(A,l,c,R),ZetabMatBloc(B,c,l,R),R);
   }

 return(0);
}

   
