/*
 * zen/zetab/c/mat_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial multiplications}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zetab.h"
/*
   \TEX{
   \title{Multiplication.}
   \input{Three |ZENMat| |X|, |A| and |B| of a finite ring |R|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{$|X|= |A| \times |B|$.}
   \note{Algorithm used is the general one.}
   }
   */
   
int _ZetabMatMultiplyPlain(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{int l,c,k;
 ZENElt S,g,e1,e2;
 ZENTest(ZetabMatIsPermutation(A,R)||ZetabMatIsPermutation(B,R)||ZetabMatIsPermutation(X,R),
	 ZEN_ISPERM,ZEN_MATMULTIPLY,ZEN_LIB_ZETAB, ZERR);
 ZENTest((ZetabMatNbRow(A,R) != ZetabMatNbRow(X,R))
	 ||(ZetabMatNbCol(A,R) != ZetabMatNbRow(B,R))
	 ||(ZetabMatNbCol(B,R) != ZetabMatNbCol(X,R)),
	 ZEN_MAT_DIM,ZEN_MATMULTIPLY,ZEN_LIB_ZETAB, ZERR);

 for(l=ZetabMatNbRow(X,R);l--;)
   for(c=ZetabMatNbCol(X,R);c--;)
     {
       k=ZetabMatNbCol(A,R)-1;
       e1 = ZetabMatGetCoeffPtr(A,l,k,R);
       e2 = ZetabMatGetCoeffPtr(B,k,c,R);
       ZetabEltMultiply(S, e1, e2, R);
       for(;k--;)
         {
	   e1 = ZetabMatGetCoeffPtr(A,l,k,R);
	   e2 = ZetabMatGetCoeffPtr(B,k,c,R);
	   ZENEltMultiply(g,e1,e2,R);
           ZENEltAdd(S,g,R);
         }
       ZetabMatSetCoeff(X,l,c,S,R);
     }
 return(0);
}

/*
   \TEX{
   \title{Multiplication of scalar}
   \input{A |ZENMat|, a |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{All the elements of |A| are multiplied by |b|.}
   \note{|A| cannot be a permutation.}
   }
   */
int _ZetabMatMultiplyScalar(A,b,R)
     ZENMat A;
     ZENElt b;
     ZENRing R;
{
  int l,c;

  ZENTest(ZetabMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATMULTIPLY,ZEN_LIB_ZETAB, ZERR);
     
  for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
    for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
      ZetabAnyEltMultiply(ZetabMatBloc(A,l,c,R),ZetabElt2Int(b,R),ZetabMatBloc(A,l,c,R),R);
  
  return(0);
}

