/*
 * zen/zetab/c/mat_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   
   \Subsection{Tests on a matrix}
   }
   */


#include <stdio.h>
#include "zetab.h"
/* \TEX{
   \title{Test of equality}
   \input{Two matrices and a |ZENRing|}
   \output{1 if the matrices are equal, 0 otherwise}
   }
   */

int _ZetabMatAreEqual(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;
 
 if(ZetabMatAreSameType(A,B,R))
   return(ZetabMatSameTypeAreEqual(A,B,R));
 
 if((ZetabMatNbBn(A,R) != ZetabMatNbDigit(B,R))
    ||(ZetabMatNbDigit(A,R) != ZetabMatNbBn(B,R)))
   return(0);
 for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
     if(!(ZetabAnyEltAreEqual(ZetabMatBloc(A,l,c,R),ZetabMatBloc(B,c,l,R),R)))
       return(0);
 return(1);
}

/* \TEX{
   \title{Test of equality to zero}
   \input{A matrix and a |ZENRing|}
   \output{1 if the matrix is null, 0 otherwise}
   }
   */

int _ZetabMatIsZero(A,R)
     ZENMat A;
     ZENRing R;
{
  return(ZetabMatIsZero(A,R));
}

/* \TEX{
   \title{Test of equality to identity}
   \input{A matrix and a |ZENRing|}
   \output{1 if the matrix is identity, 0 otherwise}
   \note{The matrix is checked to be square.} 
   }
   */

int _ZetabMatIsOne(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;

 if(ZetabMatNbBn(A,R) != ZetabMatNbDigit(A,R))
   return(0);
 
 for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZetabMatNbBn(A,R)-1;c>=0;c--)
     if(l==c)
       {
	 if(!(ZetabAnyEltIsOne(ZetabMatBloc(A,l,c,R),R)))
	   return(0);
       }
     else
       {
	 if(!(ZetabAnyEltIsZero(ZetabMatBloc(A,l,c,R),R)))
	   return(0);
       }
 return(1); 
}


