/*
 * zen/zetab/c/poly_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial additions}
   
   One function is provided to do additions: |_ZetabPolyAdd| to add a
   |ZENPoly| to another.
   }
   */
#include <stdio.h>
#include "zetab.h"

/* \TEX{
   \title{Addition.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| += |PX|.}
   }
   */
void _ZetabPolyAdd(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;

  if (ZetabPolyDeg(RX, Rg) < ZetabPolyDeg(PX, Rg)) {
    mindeg = ZetabPolyDeg(RX, Rg);
    ZetabPolyDeg(RX, Rg) =  ZetabPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZetabPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZetabAnyEltAdd(ZetabPolyGetCof(RX, i, Rg), ZetabPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZetabPolyDeg(PX, Rg); i++) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i, Rg),
		   ZetabPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZetabPolyUpdateDegree(RX, Rg);
}

/* \TEX{
   \title{Negation.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| = -|PX|.}
   }
   */
void _ZetabPolyNegate(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZetabPolyDeg(PX, Rg); i++) {
    ZetabAnyEltNegate(ZetabPolyGetCof(RX, i, Rg), ZetabPolyGetCof(PX, i, Rg), Rg);
  }
  ZetabPolyDeg(RX, Rg) =  ZetabPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Subtraction.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| -= |PX|.}
   }
   */
void _ZetabPolySubtract(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, mindeg;

  if (ZetabPolyDeg(RX, Rg) < ZetabPolyDeg(PX, Rg)) {
    mindeg = ZetabPolyDeg(RX, Rg);
    ZetabPolyDeg(RX, Rg) =  ZetabPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZetabPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZetabAnyEltSubtract(ZetabPolyGetCof(RX, i, Rg), ZetabPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZetabPolyDeg(PX, Rg); i++) {
    ZetabAnyEltNegate(ZetabPolyGetCof(RX, i, Rg),
		   ZetabPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZetabPolyUpdateDegree(RX, Rg);
}
