/*
 * zen/zetab/c/poly_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial tests}
   
   Three functions are provided to do tests: |_ZetabPolyAreEqual| to test if two
   |ZENPoly| are equal, |_ZetabPolyIsZero| to test if a |ZENPoly| is equal to
   zero and , |_ZetabPolyIsXi| to test if a |ZENPoly| is equal to $X^i$.
   }
   */
#include <stdio.h>
#include "zetab.h"

/* \TEX{
   \title{Equal.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{|RX| == |PX|.}
   }
   */
int _ZetabPolyAreEqual(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  if (ZetabPolyDeg(RX, Rg) != ZetabPolyDeg(PX, Rg)) return(0);

  for(i = 0; i <= ZetabPolyDeg(RX, Rg); i++) {
    if (!ZetabAnyEltAreEqual(ZetabPolyGetCof(RX, i, Rg),
			 ZetabPolyGetCof(PX, i, Rg),
			 Rg)) {
      return(0);
    }
  }
  return(1);
}

/* \TEX{
   \title{Is zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \output{|RX| == 0.}
   }
   */
int _ZetabPolyIsZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  return(!ZetabPolyDeg(RX, Rg) &&
	 ZetabAnyEltIsZero(ZetabPolyGetCof(RX, 0, Rg), Rg));
}

/* \TEX{
   \title{Is a polynomial equal to $X^i$.}
   \input{A polynomial |p| of a finite ring |Rg| and a degree |i|.}
   \output{$|p| == X^i$.}
   }
   */
int _ZetabPolyIsXi (p, deg, Rg)
     ZENPoly p;
     int deg;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZetabPolyDeg(p, Rg) != deg) {
    return(0);
  }

  for(i = 0; i < ZetabPolyDeg(p, R); i++) {
    elt = ZetabPolyGetCoeffPtr(p, i, Rg);
    if (!ZetabEltIsZero(elt, Rg)) {
      return(0);
    }
  }
  elt = ZetabPolyGetCoeffPtr(p, i, Rg);
  if (!ZetabEltIsOne(elt, Rg)) {
    return(0);
  }
  
  return(1);
}
