/*
 * zen/zext/c/elt_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element inversion}
   
   |_ZextEltInverse| computes the inverse of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zext.h"

/* \TEX{
   \title{Inverse.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|*p_c| = 1/a$ if |a| has an inverse, the gcd of |a| with the
   modulo otherwise.} 
   }
   */
int _ZextEltInverse(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  int ret;

  if(ZextEltIsZero(a,Rg))
    {
      ZENSetError(ZEN_LIB_ZEXT,ZEN_INVERSE,ZEN_DIVBYZERO);
      return(ZERR);
    }
  switch(ret = ZENPolyExtGcd(ZextElt2Pol(*p_c, Rg),
			     ZextElt2Pol(a, Rg),
			     ZENRingPol(Rg),
			     ZENRingDef(Rg))) {
  case -2:
    ZENEltAssign(ZENRingFact(Rg), *p_c, Rg);
    ZENSetError(ZEN_LIB_ZEXT,ZEN_INVERSE,ZEN_ELT_NOTINV);
    ret = ZEN_NO_INVERSE;
    break;
  case ZEN_NO_INVERSE:
    ZENPolyDeg(ZENElt2Pol(ZENRingFact(Rg), Rg), Rg) = 0;
    ZENPolySetCoeff(ZENElt2Pol(ZENRingFact(Rg), Rg),
		    0,
		    ZENRingFact(ZENRingDef(Rg)),
		    ZENRingDef(Rg));
    break;
  }
  return(ret);
}
