/*
 * zen/zext/c/init.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zext.h"
#include "zeg.h"

/* \TEX{
   \title{Initialization.}
   \input{A partially defined |ZENRing|.}
   \output{$|SR_GEN|\vert |MAT_GEN|$ otherwise.}
   \note{This procedure is called by |ZENExtRingCreate| and should not be called otherwise.}
   }
   */
int _ZextInit(R)
     ZENRing R;
{
  ZENRingArith(R) = ZEN_LIB_ZEXT;
  
  /* Procedure on rings
   **********************/
  if (ZBNIsZero(ZENRingQ(R), ZENRingQl(R))) {
    ZENPrcInitEltMultiplyPtr(R) = NULL;
    ZENPrcCloseEltMultiplyPtr(R) = NULL;

    ZENPrcInitTracePtr(R) = NULL;
    ZENPrcCloseTracePtr(R) = NULL;
    
    ZENPrcInitPolyRootsCanonicalPtr(R) = NULL;
    ZENPrcClosePolyRootsCanonicalPtr(R) = NULL;
    
    ZENPrcInitEltExpPtr(R) = NULL;
    ZENPrcCloseEltExpPtr(R) = NULL;
  }
  else {
    if ((ZENRingArith(ZENRingDef(R)) == ZEN_LIB_ZE2) &&
	(ZENRingSubDeg(R) < (LN2PREZENMULT<<1))) {
      ZENPrcInitEltMultiplyPtr(R) = _Ze2mPrcInitEltMultiply;
      ZENPrcCloseEltMultiplyPtr(R) = _Ze2mPrcCloseEltMultiply;
    }
    else {
      ZENPrcInitEltMultiplyPtr(R) = _ZextPrcInitEltMultiply;
      ZENPrcCloseEltMultiplyPtr(R) = _ZextPrcCloseEltMultiply;
    }
    
    ZENPrcInitTracePtr(R) = _ZextPrcInitTrace;
    ZENPrcCloseTracePtr(R) = _ZextPrcCloseTrace;
    
    ZENPrcInitPolyRootsCanonicalPtr(R) = _ZextPrcInitPolyRootsCanonical;
    ZENPrcClosePolyRootsCanonicalPtr(R) = _ZextPrcClosePolyRootsCanonical;
    
    ZENPrcInitEltExpPtr(R) = _ZENPrcInitEltExp;
    ZENPrcCloseEltExpPtr(R) = _ZENPrcCloseEltExp;
  }

  /* Procedure on elements
   ****************************/
  /* Allocation */
  ZENEltCreatePtr(R) = _ZextEltCreate;	/* Create elements */
  ZENEltCreateLargePtr(R) = _ZextEltCreateLarge;
  ZENEltFreePtr(R) = _ZextEltFree;	/* Free elements */

  /* Assigning */
  ZENEltAssignPtr(R) = _ZextEltAssign;	/* Assign an element with another */
  ZENEltSetToZeroPtr(R) = _ZextEltSetToZero; /* Set an element to zero */
  ZENEltSetToOnePtr(R) = _ZextEltSetToOne; /* Set an element to one */
  ZENEltSetRandomPtr(R) = _ZextEltSetRandom; /* Set randomly an element */
  ZENEltSetNextPtr(R) = _ZextEltSetNext; /* Set an element to the next */
  ZENEltToClonePtr(R) = ZENEltAssignPtr(R);
  ZENCloneToEltPtr(R) = ZENEltAssignPtr(R);
  ZENEltConvertPtr(R) = _ZextEltConvert;

  /* Test */
  ZENEltAreEqualPtr(R) = _ZextEltAreEqual; /* Is an element equal to another */
  ZENEltIsZeroPtr(R) = _ZextEltIsZero;	/* Is an element equal to zero */
  ZENEltIsOnePtr(R) = _ZextEltIsOne;	/* Is an element equal to one */
  if ((ZENRingPl(R) == 1) && (ZBNDigit(ZENRingP(R), 0) == 2)) {
    ZENEltIsASquarePtr(R) = _ZENEltIsASquareChar2; /* Is an element a square in finite fields with characteristic 2 ? */
  }
  else {
    ZENEltIsASquarePtr(R) = _ZENEltIsASquareCharp; /* Is an element a square in finite fields with odd characteristic? */
  }


  /* Arithmetic */
  ZENEltAddPtr(R) = _ZextEltAdd;	/* Addition of elements */
  ZENEltNegatePtr(R) = _ZextEltNegate;	/* Negation of an element */
  ZENEltSubtractPtr(R) = _ZextEltSubtract; /* Substraction of elements */
  ZENEltSquarePtr(R) = _ZextEltSquare;	/* Squaring of elements */
  ZENEltMultiplyPtr(R) = _ZextEltMultiply; /* Multiplication of elements */
  ZENEltInversePtr(R) = _ZextEltInverse; /* Inverse of elements */
  ZENEltTracePtr(R) = _ZextEltTraceNaive;		/* Trace of elements */
  if ((ZENRingPl(R) == 1) && (ZBNDigit(ZENRingP(R), 0) == 2)) {
    ZENEltSquareRootPtr(R) = _ZENEltSquareRootChar2; /* Square root of an element in finite fields with characteristic 2 */
  }
  else {
    ZENEltSquareRootPtr(R) = _ZENEltSquareRootCharp; /* Square root of an element in finite fields with odd characteristic */
  }
  ZENEltExpPtr(R) = _ZENEltExpPlain;	/* Exponentiation */


  /* io */
  ZENEltReadFromStringPtr(R) = _ZextEltReadFromString;	/* Read an element from a string */
  ZENEltPrintToStringPtr(R) = _ZextEltPrintToString; /* Print an element to a string */
  ZENEltReadFromFilePtr(R) = _ZextEltReadFromFile; /* Read an element from a file */
  ZENEltPrintToFilePtr(R) = _ZextEltPrintToFile; /* Print an element to a file */
  ZENEltGetFromStringPtr(R) = _ZextEltGetFromString; /* Get an element from a string */
  ZENEltPutToStringPtr(R) = _ZextEltPutToString; /* Put an element to a string */
  ZENEltGetFromFilePtr(R) = _ZextEltGetFromFile; /* Get an element from a file */
  ZENEltPutToFilePtr(R) = _ZextEltPutToFile; /* Put an element to a file */
  
  ZENZToEltPtr(R) = _ZextZToElt; /* Evaluate in $p$ an element */
  ZENEltToZPtr(R) = _ZextEltToZ; /* Get the element from an evaluation in $p$ */

  /* Procedures on polynomials
   *******************************/
  /* Allocation */
  ZENPolyCreatePtr(R) = _ZextPolyCreate; /* Create polynomials */
  ZENPolyCreateLargePtr(R) = _ZextPolyCreateLarge; /* Create polynomials */
  ZENPolyFreePtr(R) = _ZextPolyFree;	/* Free polynomials */

  /* Assigning */
  ZENPolyAssignPtr(R) = _ZextPolyAssign; /* Assign a polynomial with another */
  ZENPolyAssignLargePtr(R) = _ZextPolyAssignLarge;
  ZENPolyLazyToVeryLazyPtr(R) = _ZextPolyLazyToVeryLazy;
  ZENPolySetToZeroPtr(R) = _ZextPolySetToZero;	/* Set a polynomial to zero */
  ZENPolySetToZeroLargePtr(R) = _ZextPolySetToZeroLarge;	/* Set a polynomial to zero */
  ZENPolySetToXiPtr(R) = _ZextPolySetToXi;	/* Set a polynomial to X^i */
  ZENPolySetRandomPtr(R) = _ZextPolySetRandom;	/* Set randomly a polynomial */
  ZENPolySetCoeffPtr(R) = _ZextPolySetCoeff; /* Set a coefficient of a polynomial */
  ZENPolyGetCoeffNewPtr(R) = _ZextPolyGetCoeff; /* Get a coefficient of a polynomial */
  ZENPolyGetCoeffPtrPtr(R) = _ZextPolyGetCoeffPtr; /* Get a pointer to coefficient of a polynomial */
  ZENPolyUpdateDegreePtr(R) = _ZextPolyUpdateDegree; /* Get a coefficient of a polynomial */

  /* Test */
  ZENPolyAreEqualPtr(R) = _ZextPolyAreEqual; /* Is a polynomial equal to another */
  ZENPolyIsZeroPtr(R) = _ZextPolyIsZero; /* Is a polynomial equal to zero */
  ZENPolyIsXiPtr(R) = _ZextPolyIsXi; /* Is a polynomial equal to $X^i$ */

  /* Arithmetic */
  ZENPolyAddPtr(R) = _ZextPolyAdd;	/* Addition of polynomials */
  ZENPolyAddLargePtr(R) = _ZextPolyAddLarge;
  ZENPolyNegatePtr(R) = _ZextPolyNegate; /* Negation of polynomials */
  ZENPolyNegateLargePtr(R) = _ZextPolyNegateLarge; /* Negation of polynomials */
  ZENPolySubtractPtr(R) = _ZextPolySubtract; /* Substraction of polynomials */
  ZENPolySubtractLargePtr(R) = _ZextPolySubtractLarge; /* Substraction of polynomials */

  if ((ZENRingPl(R) == 1) && (ZBNDigit(ZENRingP(R), 0) == 2)) {
    ZENPolySquarePtr(R) = _ZextPolySquareChar2; /* Squaring of polynomials in characteristic 2 */
    ZENPolySquareLazyPtr(R) = _ZextPolySquareChar2;
    ZENPolySquareVeryLazyPtr(R) = _ZextPolySquareChar2;
  }
  else {
    ZENPolySquarePtr(R) = _ZextPolySquareKaratsuba; /* Squaring of polynomials in characteristic p */
    ZENPolySquareLazyPtr(R) = _ZextPolySquareKaratsubaLazy;
    ZENPolySquareVeryLazyPtr(R) = _ZextPolySquareKaratsubaVeryLazy;
  }

  ZENPolyMultiplyPtr(R) = _ZextPolyMultiply; /* Multiplication of polynomials */
  ZENPolyMultiplyLazyPtr(R) = _ZextPolyMultiplyLazy;
  ZENPolyMultiplyVeryLazyPtr(R) = _ZextPolyMultiplyVeryLazy;

  ZENPolyMultiplyScalarPtr(R) = _ZextPolyMultiplyScalar; /* Multiplication of polynomials by a scalar */
  ZENPolyDotPtr(R) = _ZextPolyDot; /* Scalar multiplication of polynomials */
  ZENPolyEvalPtr(R) = _ZENPolyEval; /* Evaluation of a polynomial */
  ZENPolyDividePtr(R) = _ZextPolyDivide; /* Division of polynomials */
  ZENPolyDivideLazyPtr(R) = _ZextPolyDivideLazy; /* Lazy division of polynomials */
  ZENPolyDivideVeryLazyPtr(R) = _ZextPolyDivideVeryLazy; /* Lazy division of polynomials */
  ZENPolyModuloLazyPtr(R) = _ZextPolyModuloVeryLazy; /* Lazy modulo of polynomials */
  ZENPolyModuloVeryLazyPtr(R) = _ZextPolyModuloVeryLazy; /* Lazy modulo of polynomials */
  ZENPolyDerivePtr(R) = _ZENPolyDerive;
  if ((ZENRingPl(R) == 1) && (ZBNDigit(ZENRingP(R), 0) == 2)) {
    ZENPolyRootsDegree2Ptr(R) = _ZENPolyRootsDegree2Char2; /* Roots of polynomials of degree 2 in finite fields with characteristic 2 */
  }
  else {
    ZENPolyRootsDegree2Ptr(R) = _ZENPolyRootsDegree2Charp; /* Roots of polynomials of degree 2 in finite fields with odd characteristic */
  }
  ZENMatCreateParamsPtr(R) = _ZextMatCreateParams;
  ZENMatCreatePtr(R) = __ZextMatCreate;

  return(SR_GEN | MAT_GEN);
}

/* \TEX{
   \title{Closing.}
   \input{A |ZENRing|.}
   \output{0.}
   \note{This procedure is called by |ZENRingClose| and should not be called otherwise.   }
   }
   */
int _ZextClose(R)		
     ZENRing R;
{
  return(0);
}
