/*
 * zen/zext/c/mat_convert.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zext.h"

/*
   \TEX{
   \Subsection{Matrix conversion}
   The following functions are intended to allow conversion of matrices between an extension and its definition |ZENRing|. In all the following functions, matrix |M| is defined over the extension |R| and |m| is defined over the underlevel extension |ZENRingDef(R)|.

   \title{Extension conversion}
   \input{Two matrices |ZENMat| and a non tabulated extension |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |m| is assigned with |M|.}
   } 
   */
int _ZextMat2Vect(M,m,R)
     ZENMat M,m;
     ZENRing R;
{int l,c,k;
 ZENElt *ptm,*ptM;

 ZENTest(ZENRingDef(R)==NULL,
	 ZEN_NOT_EXT,ZEN_MAT2VECT,ZEN_LIB_ZEXT, ZERR);

 ZENTest(ZENMatIsPermutation(M,R)
	 ||ZENMatIsPermutation(m,ZENRingDef(R)),
	 ZEN_ISPERM,ZEN_MAT2VECT,ZEN_LIB_ZEXT, ZERR);
 
 ZENTest(ZENMatNbBnd(M,R) != ZENMatNbBn(m,ZENRingDef(R))*ZENMatNbDigit(m,ZENRingDef(R)),
	 ZEN_MAT_DIM,ZEN_MAT2VECT,ZEN_LIB_ZEXT, ZERR);

 ptm=(ZENElt *)ZENMatFirst(m,ZENRingDef(R))-ZENMatNbPtr(m,ZENRingDef(R));
 ptM=(ZENElt *)ZENMatFirst(M,R);
 if(ZENMatAreSameType(m,M,R))
   for(l=ZENMatNbBnd(M,R);l--;ptm++,ptM++)
     ZENEltAssign(*ptM,*ptm,ZENRingDef(R));
 else
   for(l=ZENMatNbBn(M,R);
       l--;
       ptm -= ZENMatNbBn(m,ZENRingDef(R))*ZENMatNbDigit(m,ZENRingDef(R)) - ZENMatSize(M,R))
     for(c=ZENMatNbDigit(M,R);
	 c--;
	 ptm += -ZENMatSize(M,R) + ZENMatNbBloc(m,ZENRingDef(R)))
       for(k= ZENMatSize(M,R);k--;ptM++,ptm++)
	 ZENEltAssign(*ptM,*ptm,ZENRingDef(R));

 
 ptM=(ZENElt *)ZENMatFirst(M,R)-1;
 for(l=ZENMatNbBn(M,R)*ZENMatNbDigit(M,R);l--;ptM--)
   {
     ZENPolyDeg(ZENElt2Pol(*ptM,R),
		ZENRingDef(R))=ZENPolyLgt(ZENElt2Pol(*ptM,R),
					  ZENRingDef(R));
     ZENPolyUpdateDegree(ZENElt2Pol(*ptM,R),ZENRingDef(R));
   }
 return(0);
}

/*     
   \TEX{
   \title{Extension conversion}
   \input{Two matrices |ZENMat| and a non tabulated extension |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |M| is assigned with |m|.}
   } 
   */
int _ZextVect2Mat(m,M,R)
     ZENMat M,m;
     ZENRing R;
{int l,k,c;
 ZENElt *pt,*ptm,*ptM;

 ZENTest(ZENRingDef(R)==NULL,
	 ZEN_NOT_EXT,ZEN_VECT2MAT,ZEN_LIB_ZEXT, ZERR);

 ZENTest(ZENMatIsPermutation(M,R)
	 ||ZENMatIsPermutation(m,ZENRingDef(R)),
	 ZEN_ISPERM,ZEN_VECT2MAT,ZEN_LIB_ZEXT, ZERR);
     
 ZENTest(ZENMatNbBnd(M,R) != ZENMatNbBn(m,ZENRingDef(R))*ZENMatNbDigit(m,ZENRingDef(R)),
	 ZEN_MAT_DIM,ZEN_MAT2VECT,ZEN_LIB_ZEXT, ZERR);

 ptm=(ZENElt *)ZENMatFirst(m,ZENRingDef(R))-ZENMatNbPtr(m,ZENRingDef(R));
 ptM=(ZENElt *)ZENMatFirst(M,R);
 pt=(ZENElt *)ZENMatFirst(M,R)-ZENMatNbPtr(M,R);
 if(ZENMatAreSameType(m,M,R))
   for(l=ZENMatNbBn(M,R)*ZENMatNbDigit(M,R);l--;pt++)
     {
       for(k=0;k<=ZENPolyDeg(ZENElt2Pol(*pt,R),ZENRingDef(R));k++,ptm++,ptM++)
	 ZENEltAssign(*ptm,*ptM,ZENRingDef(R));
       for(;k<=ZENPolyLgt(ZENElt2Pol(*pt,R),ZENRingDef(R));k++,ptm++,ptM++)
	 ZENEltSetToZero(*ptm,ZENRingDef(R));
     }
 else
   for(l=ZENMatNbBn(M,R);
       l--;
       ptm -= ZENMatNbBn(m,ZENRingDef(R))*ZENMatNbDigit(m,ZENRingDef(R)) - ZENMatSize(M,R))
     for(c=ZENMatNbDigit(M,R);
	 c--;
	 pt++,ptm+=ZENMatNbBloc(m,ZENRingDef(R)) - ZENMatSize(M,R))
       {
	 for(k=0;
	     k<=ZENPolyDeg(ZENElt2Pol(*pt,R),ZENRingDef(R));
	     k++,ptm++,ptM++)
	   ZENEltAssign(*ptm,*ptM,ZENRingDef(R));
	 for(;k<=ZENPolyLgt(ZENElt2Pol(*pt,R),ZENRingDef(R));k++,ptm++,ptM++)
	   ZENEltSetToZero(*ptm,ZENRingDef(R));
       }
 return(0);
}


/*
   \TEX{
   \title{Extension conversion}
   \input{Two matrices |ZENMat| and a non tabulated extension |ZENRing| over $\Z/2\Z$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |M| defined over the extrension |R| is assigned with |m| defined over $\Z/2\Z$.}
   } 
   */
int _Ze2mMat2Vect(M,m,R)
     ZENMat M,m;
     ZENRing R;
{
  int l,c,k,cl;
  BigNum buf;
  BigNumLength cc,n;
  ZENElt *ptm, *ptM, *pt;
 
  ZENTest(ZENRingDef(R)==NULL,
	  ZEN_NOT_EXT,ZEN_MAT2VECT,ZEN_LIB_ZE2, ZERR);
  
  ZENTest(ZENMatIsPermutation(M,R)
	  ||ZENMatIsPermutation(m,ZENRingDef(R)),
	  ZEN_ISPERM,ZEN_MAT2VECT,ZEN_LIB_ZE2, ZERR);

 
  n=ZENPolyDeg(ZENRingPol(R),ZENRingDef(R));
  k=divSizeBloc(n)+1;
  buf=ZBNC(ZENMatNbBloc(m,ZENRingDef(R)));
  ZENTest(buf == NULL,
	  ZEN_ALLOC,ZEN_MAT2VECT,ZEN_LIB_ZE2, ZERR);

  if(ZENMatIsRowType(m,ZENRingDef(R)))
    {
      ZENTest((ZENMatNbRow(M,R) != ZENMatNbBn(m,ZENRingDef(R)))
	      ||(ZENMatNbCol(M,R)*n != ZENMatNbDigit(m,ZENRingDef(R))),
	      ZEN_MAT_DIM,ZEN_MAT2VECT,ZEN_LIB_ZE2, ZERR);
      cl=ZENMatNbCol(M,R);
    }
  else
    {
      ZENTest((ZENMatNbCol(M,R) != ZENMatNbBn(m,ZENRingDef(R)))
	      ||(ZENMatNbRow(M,R)*n != ZENMatNbDigit(m,ZENRingDef(R))),
	      ZEN_MAT_DIM,ZEN_MAT2VECT,ZEN_LIB_ZE2, ZERR);
      cl=ZENMatNbRow(M,R);
    }
   
  ptm=(ZENElt *)ZENMatFirst(m,ZENRingDef(R));
  ptM=(ZENElt *)ZENMatFirst(M,R);
  pt=ptM-ZENMatNbPtr(M,R);
 
  if(ZENMatAreSameType(m,M,R))
    for(l=ZENMatNbBn(m,ZENRingDef(R));
	l--;
	ptm += ZENMatNbBloc(m,ZENRingDef(R)))
      {
	cc=ZENMatNbBloc(m,ZENRingDef(R));
	ZBNAssign(buf,(BigNum)ptm,cc);
	for(c=cl;c--;ptM+=k,pt++)
	  {
	    ZBNAssign((BigNum)ptM,buf,(BigNumLength)k);
	    ZENPolyDeg(ZENElt2Pol(*pt,R),ZENRingDef(R))=n-1;
	    ZENPolyUpdateDegree(ZENElt2Pol(*pt,R),ZENRingDef(R));
	    ZBNAnyShiftRight(buf,cc,n);
	    cc=divSizeBloc(c*n)+1;
	  }
      }
  else
    for(l=ZENMatNbBn(m,ZENRingDef(R));
	l--;
	ptm += ZENMatNbBloc(m,ZENRingDef(R)))
      {
	cc=ZENMatNbBloc(m,ZENRingDef(R));
	ZBNAssign(buf,(BigNum)ptm,cc);
	for(c=cl;c--;ptM+=k*ZENMatNbDigit(M,R),pt+=ZENMatNbDigit(M,R))
	  {
	    ZBNAssign((BigNum)ptM,buf,(BigNumLength)k);
	    ZENPolyDeg(ZENElt2Pol(*pt,R),ZENRingDef(R))=n-1;
	    ZENPolyUpdateDegree(ZENElt2Pol(*pt,R),ZENRingDef(R));
	    ZBNAnyShiftRight(buf,cc,n);
	    cc=divSizeBloc(c*n)+1;
	  }
	pt-=ZENMatNbBn(M,R)*ZENMatNbDigit(M,R)-1;
	ptM-=ZENMatNbBn(M,R)*k*ZENMatNbDigit(M,R)-k;
      }
  ZBNF(buf);
  return(0);
}

/*     
   \TEX{
   \title{Extension conversion}
   \input{Two matrices |ZENMat| and a non tabulated extension |ZENRing| over $\Z/2\Z$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |M| is assigned with |m|.}
   } 
   */
int _Ze2mVect2Mat(m,M,R)
     ZENMat M,m;
     ZENRing R;
{
  int cl,l,c,n,k;
  ZENElt *ptm,*ptM;
 
  ZENTest(ZENRingDef(R)==NULL,
	  ZEN_NOT_EXT,ZEN_VECT2MAT,ZEN_LIB_ZE2, ZERR);
  
  ZENTest(ZENMatIsPermutation(M,R)
	  ||ZENMatIsPermutation(m,ZENRingDef(R)),
	  ZEN_ISPERM,ZEN_VECT2MAT,ZEN_LIB_ZE2, ZERR);
  
  n=ZENPolyDeg(ZENRingPol(R),ZENRingDef(R));
  k=divSizeBloc(n)+1;
  
  if(ZENMatIsRowType(m,ZENRingDef(R)))
    {
      ZENTest((ZENMatNbRow(M,R) != ZENMatNbBn(m,ZENRingDef(R)))
	      ||(ZENMatNbCol(M,R)*n != ZENMatNbDigit(m,ZENRingDef(R))),
	      ZEN_MAT_DIM,ZEN_VECT2MAT,ZEN_LIB_ZE2, ZERR);
      cl=ZENMatNbCol(M,R);     
    }
  else
    {
      ZENTest((ZENMatNbRow(M,R)*n != ZENMatNbDigit(m,ZENRingDef(R)))
	      ||(ZENMatNbCol(M,R) != ZENMatNbBn(m,ZENRingDef(R))),
	      ZEN_MAT_DIM,ZEN_VECT2MAT,ZEN_LIB_ZE2, ZERR);
      cl=ZENMatNbRow(M,R);
    }
  
  ptm=(ZENElt *)ZENMatFirst(m,ZENRingDef(R))+ZENMatNbBnd(m,ZENRingDef(R));
  ptM=(ZENElt *)ZENMatFirst(M,R);
  if(ZENMatAreSameType(m,M,R))
    {
      if(cl==1)
	{
	  for(l=ZENMatNbBn(m,ZENRingDef(R)),
	      ptm -= ZENMatNbBloc(m,ZENRingDef(R)),
	      ptM--;
	      l--;
	      ptm -= ZENMatNbBloc(m,ZENRingDef(R)),
	      ptM--)	   
	    ZBNAssign((BigNum)ptm,
		      ZENPolyDigit(ZENElt2Pol(*ptM,R),ZENRingDef(R)),
		      (BigNumLength)divSizeBloc(ZENPolyDeg(ZENElt2Pol(*ptM,R),ZENRingDef(R)))+1);
	}
      else
	{
	  ZENPoly P;
	  P=ZENPolyCreate(ZENMatNbDigit(m,ZENRingDef(R))-1,ZENRingDef(R));
	  ZENTest(!P,ZEN_ALLOC,ZEN_MAT2VECT,ZEN_LIB_ZE2, ZERR);
	  
	  for(l=ZENMatNbBn(m,ZENRingDef(R)),
	      ptm -= ZENMatNbBloc(m,ZENRingDef(R)),
	      ptM--;
	      l--;
	      ptm -= ZENMatNbBloc(m,ZENRingDef(R)))
	    {
	      int ds, shift;

	      ZENPolyAssign(P,ZENElt2Pol(*ptM,R),ZENRingDef(R));

	      /* Cleaning garbage high bits of P */
	      ds = divSizeBloc(ZENPolyDeg(P, ZENRingDef(R)));
	      shift = SIZE_BLOC-1-modSizeBloc(ZENPolyDeg(P, ZENRingDef(R)));
	      ZENPolyGetDigit(P, ds, ZENRingDef(R)) <<= shift;
	      ZENPolyGetDigit(P, ds, ZENRingDef(R)) >>= shift;

	      for(c=cl-1,ptM--;c--;ptM--)
		{
		  if(!ZENPolyIsZero(P,ZENRingDef(R))) /* P *= X^n */
		    {
		      BigNumLength bufl;

		      bufl=1+divSizeBloc(ZENPolyDeg(P,ZENRingDef(R)));
		      ZBNAnyShiftLeft(ZENPolyDigit(P,ZENRingDef(R)),
				      &bufl,
				      ZENPolyDigit(P,ZENRingDef(R)),
				      bufl,
				      n);
		      ZENPolyDeg(P,ZENRingDef(R))+=n;
		    }
		  ZENPolyAdd(P,ZENElt2Pol(*ptM,R),ZENRingDef(R));
		}
	      ZBNAssign((BigNum)ptm,
			ZENPolyDigit(P,ZENRingDef(R)),
			(BigNumLength)divSizeBloc(ZENPolyDeg(P,ZENRingDef(R)))+1);
	    }
	  ZENPolyFree(P,ZENRingDef(R));
	}
    }
  else
    {
      if(cl==1)
	{
	  for(l=ZENMatNbBn(m,ZENRingDef(R)),
	      ptm -= ZENMatNbBloc(m,ZENRingDef(R)),
	      ptM--;
	      l--;
	      ptm -= ZENMatNbBloc(m,ZENRingDef(R)))
	    {
	      ZBNAssign((BigNum)ptm,
			ZENPolyDigit(ZENElt2Pol(*ptM,R),ZENRingDef(R)),
			(BigNumLength)divSizeBloc(ZENPolyDeg(ZENElt2Pol(*ptM,R),ZENRingDef(R)))+1);
	      ptM += (ZENMatNbBn(M,R)-1)*ZENMatNbDigit(M,R)-1;
	    }
	}
      else
	{
	  ZENPoly P;
	  P=ZENPolyCreate(ZENMatNbDigit(m,ZENRingDef(R))-1,ZENRingDef(R));
	  ZENTest(!P,ZEN_ALLOC,ZEN_MAT2VECT,ZEN_LIB_ZE2, ZERR);
	  
	  for(l=ZENMatNbBn(m,ZENRingDef(R)),
	      ptm -= ZENMatNbBloc(m,ZENRingDef(R)),
	      ptM--;
	      l--;
	      ptm -= ZENMatNbBloc(m,ZENRingDef(R)))
	    {
	      int ds, shift;

	      ZENPolyAssign(P,ZENElt2Pol(*ptM,R),ZENRingDef(R));
	      /* Cleaning garbage high bits of P */
	      ds = divSizeBloc(ZENPolyDeg(P, ZENRingDef(R)));
	      shift = SIZE_BLOC-1-modSizeBloc(ZENPolyDeg(P, ZENRingDef(R)));
	      ZENPolyGetDigit(P, ds, ZENRingDef(R)) <<= shift;
	      ZENPolyGetDigit(P, ds, ZENRingDef(R)) >>= shift;

	      for(c=cl-1,ptM-=ZENMatNbDigit(M,R);
		  c--;
		  ptM-=ZENMatNbDigit(M,R))
		{
		  if(!ZENPolyIsZero(P,ZENRingDef(R))) /* P *= X^n */
		    {
		      BigNumLength bufl;

		      bufl=1+divSizeBloc(ZENPolyDeg(P,ZENRingDef(R)));
		      ZBNAnyShiftLeft(ZENPolyDigit(P,ZENRingDef(R)),
				      &bufl,
				      ZENPolyDigit(P,ZENRingDef(R)),
				      bufl,
				      n);
		      ZENPolyDeg(P,ZENRingDef(R))+=n;
		    }
		  ZENPolyAdd(P,ZENElt2Pol(*ptM,R),ZENRingDef(R));
		}
	      ZBNAssign((BigNum)ptm,
			ZENPolyDigit(P,ZENRingDef(R)),
			(BigNumLength)divSizeBloc(ZENPolyDeg(P,ZENRingDef(R)))+1);
	      ptM += ZENMatNbBn(M,R)*ZENMatNbDigit(M,R)-1;
	    }
	  ZENPolyFree(P,ZENRingDef(R));
	}
    }
  return(0);
}
