/*
 * zen/zext/c/poly_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial tests}
   
   Three functions are provided to do tests: |_ZextPolyAreEqual| to test if two
   |ZENPoly| are equal, |_ZextPolyIsZero| to test if a |ZENPoly| is equal to
   zero and , |_ZextPolyIsXi| to test if a |ZENPoly| is equal to $X^i$.
   }
   */
#include <stdio.h>
#include "zext.h"

/* \TEX{
   \title{Equal.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{|RX| == |PX|.}
   }
   */
int _ZextPolyAreEqual(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  if (ZextPolyDeg(RX, Rg) != ZextPolyDeg(PX, Rg)) return(0);

  for(i = 0; i <= ZextPolyDeg(RX, Rg); i++) {
    if (!ZextEltAreEqual(ZextPolyGetCof(RX, i, Rg),
			 ZextPolyGetCof(PX, i, Rg),
			 Rg)) {
      return(0);
    }
  }
  return(1);
}

/* \TEX{
   \title{Is zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \output{|RX| == 0.}
   }
   */
int _ZextPolyIsZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  return(!ZextPolyDeg(RX, Rg) &&
	 ZextEltIsZero(ZextPolyGetCof(RX, 0, Rg), Rg));
}

/* \TEX{
   \title{Is a polynomial equal to $X^i$.}
   \input{A polynomial |p| of a finite ring |Rg| and a degree |i|.}
   \output{$|p| == X^i$.}
   }
   */
int _ZextPolyIsXi (p, deg, Rg)
     ZENPoly p;
     int deg;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZextPolyDeg(p, Rg) != deg) {
    return(0);
  }

  for(i = 0; i < ZextPolyDeg(p, R); i++) {
    elt = ZextPolyGetCoeffPtr(p, i, Rg);
    if (!ZextEltIsZero(elt, Rg)) {
      return(0);
    }
  }
  elt = ZextPolyGetCoeffPtr(p, i, Rg);
  if (!ZextEltIsOne(elt, Rg)) {
    return(0);
  }
  
  return(1);
}
