/*
 * zen/zmake.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "zmake.h"
#include "init.h"

char *TEX_FILES = "intro user example advanced doc implement title biblio maths";

char *INIT_FILE = "init";

char *SITE_FILE = "site";

char *UDOC = "doc";

char *ADOC = "advanced";

int main(argc, argv)
     int argc;
     char *argv[];
{
  /*
   * Misc. variables
   ******************/
  char *ZEN_ALLSUBDIRS;

  /* 
   * Internal Targets
   ******************/
  int sys_ar = 0;
  int zbn_ar = 0;
  int zeg_ar = 0;
  int zed_ar = 0;
  int zem_ar = 0;
  int zec_ar = 0;
  int zetab_ar = 0;
  int zelog_ar = 0;
  int ze2_ar = 0;
  int zeps_ar = 0;
  int zep_ar = 0;
  int zef_ar = 0;
  int zext_ar = 0;
  int zer_ar = 0;

  int tidy_lib = 0;
  int tidy_bin = 0;
  int tidy_doc = 0;

  int clean_doc = 0;
  int clean_dvi = 0;
  int clean_tex = 0;
  int clean_obj = 0;
  int clean_lib = 0;

  int ltex = 0;
  int ludvi = 0;
  int ladvi = 0;

  /* 
   * External targets
   *******************/
  int all = Target("all", argc, argv);
  int zen = Target("zen", argc, argv);
  int cc = Target("cc", argc, argv);
  int ar = Target("ar", argc, argv);
  int ld = Target("ld", argc, argv);

  int sys = Target("sys", argc, argv);
  int zbn = Target("zbn", argc, argv);
  int zeg = Target("zeg", argc, argv);
  int zed = Target("zed", argc, argv);
  int zem = Target("zem", argc, argv);
  int zec = Target("zec", argc, argv);
  int zetab = Target("zetab", argc, argv);
  int zelog = Target("zelog", argc, argv);
  int ze2 = Target("ze2", argc, argv);
  int zeps = Target("zeps", argc, argv);
  int zep = Target("zep", argc, argv);
  int zef = Target("zef", argc, argv);
  int zext = Target("zext", argc, argv);
  int zer = Target("zer", argc, argv);
  int prgm = Target("prgm", argc, argv);

  int doc = Target("doc", argc, argv);
  int udoc = Target("udoc", argc, argv);
  int adoc = Target("adoc", argc, argv);
  int fdoc = Target("fdoc", argc, argv);

  int tex = Target("tex", argc, argv);
  int dvi = Target("dvi", argc, argv);

  int archtidy = Target("archtidy", argc, argv);
  int archclean = Target("archclean", argc, argv);
  int clean = Target("clean", argc, argv);
  int tidy = Target("tidy", argc, argv);

  int none = Target("none", argc, argv);

  /* 
   * Initialization
   *****************/
  zen_init(CURDIR, TOPDIR, ZCFDIR);
  zmake_init("", ZEN_INCLUDES, "", "", argc, argv);
  
  ZEN_ALLSUBDIRS = concat(ZEN_SUBDIRS,
		   concat(" ", PPRGM));

  /* 
   * Targets
   **********/

  /* default */
  TargetImply(defaults, zen);

  /* all */
  TargetImply(all, zen);

  /* zen */
  TargetImply(zen, cc=ar=ld);

  /* cc */
  TargetImply(cc, sys=zbn=zeg=zed=zem=zec=zetab=zelog=ze2=zeps=zep=zef=zext=zer);
  TargetImply(cc, ar);

  /* ar */
  TargetImply(ar, sys_ar=zbn_ar=zeg_ar=zed_ar=zem_ar=zec_ar=zetab_ar);
  TargetImply(ar, zelog_ar=ze2_ar=zeps_ar=zep_ar=zef_ar=zext_ar=zer_ar);
  IsTheLibraryStillHere(ar,
			ZEN_LIB, PREFIX, LIBZEN, ASUF,
			ZEN_LIB, PREFIX, ZEN_SUBDIRS, AR_TOLIB);

  /* ld */
  TargetImply(ld, prgm);

  /* Targets for cc */
  NamedTargetSubdirs(sys, PSYS, "cc", argc, argv);
  ArListToLibrary(sys_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PSYS, AR_TOLIB,
		  ZEN_LIB, PREFIX, PSYS, OBJ_TOAR);

  NamedTargetSubdirs(zbn, PZBN, "cc", argc, argv);
  ArListToLibrary(zbn_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZBN, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZBN, OBJ_TOAR);

  NamedTargetSubdirs(zeg, PZEG, "cc", argc, argv);
  ArListToLibrary(zeg_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZEG, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZEG, OBJ_TOAR);

  NamedTargetSubdirs(zed, PZED, "cc", argc, argv);
  ArListToLibrary(zed_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZED, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZED, OBJ_TOAR);

  NamedTargetSubdirs(zem, PZEM, "cc", argc, argv);
  ArListToLibrary(zem_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZEM, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZEM, OBJ_TOAR);

  NamedTargetSubdirs(zec, PZEC, "cc", argc, argv);
  ArListToLibrary(zec_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZEC, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZEC, OBJ_TOAR);

  NamedTargetSubdirs(zetab, PZETAB, "cc", argc, argv);
  ArListToLibrary(zetab_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZETAB, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZETAB, OBJ_TOAR);

  NamedTargetSubdirs(zelog, PZELOG, "cc", argc, argv);
  ArListToLibrary(zelog_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZELOG, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZELOG, OBJ_TOAR);

  NamedTargetSubdirs(ze2, PZE2, "cc", argc, argv);
  ArListToLibrary(ze2_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZE2, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZE2, OBJ_TOAR);

  NamedTargetSubdirs(zeps, PZEPS, "cc", argc, argv);
  ArListToLibrary(zeps_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZEPS, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZEPS, OBJ_TOAR);

  NamedTargetSubdirs(zep, PZEP, "cc", argc, argv);
  ArListToLibrary(zep_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZEP, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZEP, OBJ_TOAR);

  NamedTargetSubdirs(zef, PZEF, "cc", argc, argv);
  ArListToLibrary(zef_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZEF, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZEF, OBJ_TOAR);

  NamedTargetSubdirs(zext, PZEXT, "cc", argc, argv);
  ArListToLibrary(zext_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZEXT, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZEXT, OBJ_TOAR);

  NamedTargetSubdirs(zer, PZER, "cc", argc, argv);
  ArListToLibrary(zer_ar,
		  ZEN_LIB, PREFIX, LIBZEN, ASUF,
		  ZEN_LIB, PREFIX, PZER, AR_TOLIB,
		  ZEN_LIB, PREFIX, PZER, OBJ_TOAR);

  /* Targets for ld */
  NamedTargetSubdirs(prgm, PPRGM, "all", argc, argv);

  /* doc */
  TargetImply(doc, udoc);

  TargetImply(udoc, tex=ludvi);
  TargetImply(adoc, tex=ladvi);
  TargetImply(fdoc, dvi);
  TargetImply(dvi, tex=ludvi=ladvi);

  /* fdoc */
  NamedTargetSubdirs(fdoc, ZEN_ALLSUBDIRS, "doc", argc, argv);

  /* tex */
  NamedTargetSubdirs(tex, ZEN_ALLSUBDIRS, "tex", argc, argv);
  SrcToTex(tex,
	   ZEN_TEX, "",
	   IPT, "", TEX_FILES, CSUF,
	   PRGM_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   ZEN_TEX, "",
	   ZEN_TOP, "", INIT_FILE, HSUF,
	   PRGM_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   ZEN_TEX, "",
	   ZMAKE_DIR, "", SITE_FILE, HSUF,
	   PRGM_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   ZEN_TEX, "",
	   IPT, "", UDOC, CSUF,
	   PRGM_BIN, "", IPT, PARSER);
  SrcToTex(tex,
	   ZEN_TEX, "",
	   IPT, "", ADOC, CSUF,
	   PRGM_BIN, "", IPT, PARSER);

  /* dvi */
  TargetImply(dvi, ladvi=ludvi);
  NamedTargetSubdirs(dvi, ZEN_ALLSUBDIRS, "dvi", argc, argv);
  TexToDvi(dvi,
	   ZEN_DVI, "",
	   ZEN_TEX, "", TEX_FILES,
	   ZEN_INPUTS);
  TexToDvi(dvi,
	   ZEN_DVI, "",
	   ZEN_TEX, "", INIT_FILE,
	   ZEN_INPUTS);
  TexToDvi(dvi,
	   ZEN_DVI, "",
	   ZEN_TEX, "", SITE_FILE,
	   ZEN_INPUTS);
  TexToDvi(ludvi,
	   ZEN_DVI, "",
	   ZEN_TEX, "", UDOC,
	   ZEN_INPUTS);
  TexToDvi(ladvi,
	   ZEN_DVI, "",
	   ZEN_TEX, "", ADOC,
	   ZEN_INPUTS);

  /* Cleaning */
  TargetImply(tidy, tidy_lib=tidy_bin=tidy_doc);
  TargetImply(archtidy, clean_obj=clean_lib=clean_doc);
  TargetImply(archclean, clean_obj=clean_tex);
  TargetImply(clean, clean_obj=clean_tex);
  TargetImply(clean_doc, clean_dvi=clean_tex);

  NamedTargetSubdirs(tidy, ZEN_ALLSUBDIRS, "tidy", argc, argv);
  NamedTargetSubdirs(clean, ZEN_ALLSUBDIRS, "clean", argc, argv);
  NamedTargetSubdirs(archtidy, ZEN_ALLSUBDIRS, "archtidy", argc, argv);
  NamedTargetSubdirs(archclean, ZEN_ALLSUBDIRS, "archclean", argc, argv);

  CleaningRules(clean_dvi, ZEN_TEX, "", TEX_FILES, ".dvi");
  CleaningRules(clean_dvi, ZEN_TEX, "", INIT_FILE, ".dvi");
  CleaningRules(clean_dvi, ZEN_TEX, "", SITE_FILE, ".dvi");
  TidyingRules(clean_tex, ZEN_TEX);
  CleaningRules(clean_lib, ZEN_LIB, PREFIX, LIBZEN, ASUF);
  CleaningRules(clean_obj, ZEN_LIB, PREFIX, ZEN_SUBDIRS, AR_TOLIB);
  
  TidyingRules(tidy_lib, ZEN_LBT);
  TidyingRules(tidy_bin, ZEN_BNT);
  TidyingRules(tidy_doc, ZEN_DOC);

  /* 
   * Closing
   **********/
  zmake_close();

  exit(0);
}
