/*
 * zenfact/c/factors.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zen.h"
#include "zenfact.h"

/* \TEX{
   All the following procedures to factorize a polynomial in $GF(2^n)$
   will return a list of polynomials. After describing functions to
   manipulate such lists, we give the procedure to factorize polynomials.
   }
   */

/*
  \TEX{
  A list of polynomial is a pointer on a struct which contains a
  polynomial, its multiplicity and a pointer on an other ceil of the
  list. We can append a polynomial at the beginning of a list with
  |Gf2PolyListAppendFirst| or remove a polynomial at the beginning of the
  list |Gf2PolyListAppendFirst|.
  }
  */

#if 0
/*
  \TEX{
  \title{Creating a list of polynomials.}
  \input{A list of polynomials |lfact|.}
  \side{|lfact| is allocated}
  }
  */
void ZENPolyFactCreate(lfact, Rg)
     ZENPolyFact lfact;
     ZENRing Rg;{}
#endif

#if 0
/*
  \TEX{
  \title{Freeing a list of polynomials.}
  \input{A list of polynomials |lfact|.}
  \side{|lfact| is freed}
  }
  */
void ZENPolyFactFree(lfact, Rg)
     ZENPolyFact lfact;
     ZENRing Rg;{}
#endif
void _ZENPolyFactFree(p_lfact, Rg)
     ZENPolyFact *p_lfact;
     ZENRing Rg;
{
  int i;

  if (*p_lfact) {
    if ((*p_lfact)->lgt) {
      for(i = 0; i < (*p_lfact)->nb; i++) {
	ZENPolyFree((*p_lfact)->LP[i].P, Rg);
	if ((*p_lfact)->LP[i].Xq) {
	  ZENPolyFree((*p_lfact)->LP[i].Xq, Rg);
	}
      }
      ZENFree((*p_lfact)->LP);
    }
    ZENFree(*p_lfact);
    *p_lfact = NULL;
  }
}

#if 0
/*
  \TEX{
  \title{Freeing a polynomial structure.}
  \input{A polynomial structure |list|.}
  \side{|list| is allocated}
  }
  */
void ZENPolyListFree(list, Rg)
     ZENPolyList list;
     ZENRing Rg;{}
#endif
#if 0
/*
  \TEX{
  \title{Insert a polynomial into a list.}
  \input{A polynomial |P| and a multiplicity |mult|.}
  \side{(|P|, |mult|) is inserted in |lfact|.}
  pointer to |P|.}
  \output{-1 if an error occured, 0 otherwise.}
  }
  */
int ZENPolyFactAppend(lfact, list, Rg)
     ZENPolyFact lfact;
     ZENPolyList list;
     ZENRing Rg;{}
#endif
int _ZENPolyFactAppend(p_lfact, list, Rg)
     ZENPolyFact *p_lfact;
     ZENPolyList list;
     ZENRing Rg;
{
  int i, index;

  if (!(*p_lfact)) {
    MallocAT(*p_lfact, 1, struct zenpolyfact, ;, return(-1));
    (*p_lfact)->lgt = 0;    
  }
  
  if (!(*p_lfact)->lgt) {
    MallocAT((*p_lfact)->LP, 1, struct zenpolylist, ;, return(-1));
    (*p_lfact)->lgt = 1;
    (*p_lfact)->nb = 0;
  }
  
  if ((*p_lfact)->lgt == (*p_lfact)->nb) {
    ZENPolyList *ptr;
    
    (*p_lfact)->lgt++;
    MallocAT(ptr, (*p_lfact)->lgt, ZENPolyList, ;, return(-1));
    for(i = 0; i < (*p_lfact)->nb; i++) {
      ptr[i] = (*p_lfact)->LP[i];
    }
    ZENFree((*p_lfact)->LP);
    (*p_lfact)->LP = ptr;
  }
  
  for(index = 0; index < (*p_lfact)->nb; index++) {
    if (ZENPolyDeg((*p_lfact)->LP[index].P, Rg) >= ZENPolyDeg(list.P, Rg)) {
      break;
    }
  }
  for(i = index;
      (i  < (*p_lfact)->nb) &&
      (ZENPolyDeg((*p_lfact)->LP[i].P, Rg) == ZENPolyDeg(list.P, Rg));
      i++) {
    if (ZENPolyAreEqual((*p_lfact)->LP[i].P, list.P, Rg)) {
      index = i;
      break;
    }
  }
  (*p_lfact)->nb++;
  for(i = (*p_lfact)->nb-1; i >= index+1; i--) {
    (*p_lfact)->LP[i] = (*p_lfact)->LP[i-1];
  }
  (*p_lfact)->LP[index] = list;

  return(0);
}

#if 0
/*
  \TEX{
  \title{Removing a polynomial of a list.}
  \input{A list |lfact| and a pointer on a multiplicity |p_mult|
  and a polynomial |p_P|.}
  \output{-1 if |index| is greater than the size of the list,
  0 otherwise.}
  \side{The list is updated and |*p_P|, |*p_m| are initialized
  if no error occured.}
  }
  */
int ZENPolyFactRemove(list, lfact, index, Rg)
     ZENPolyList list;
     ZENPolyFact lfact;
     int index;
     ZENRing Rg;{}
#endif
int _ZENPolyFactRemove(p_list, p_lfact, index, Rg)
     ZENPolyList *p_list;
     ZENPolyFact *p_lfact;
     int index;
     ZENRing Rg;
{
  int i;

  if (!(*p_lfact) || (index >= (*p_lfact)->nb)) {
    return(-1);
  }

  *p_list = (*p_lfact)->LP[index];
  for(i = index+1; i < (*p_lfact)->nb; i++) {
    (*p_lfact)->LP[i-1] = (*p_lfact)->LP[i];
  }
  (*p_lfact)->nb--;

  if (!((*p_lfact)->nb)) {
    ZENPolyFactFree(*p_lfact, Rg);
  }
  return(0);
}

/*
  \TEX{
  Three main steps must be done to factorize a polynomial |P|:
  \begin{itemize}
  \item The square free step done by |ZENPolySFFactor|,
  \item The distict degree step done by |ZENPolyDDFactor|,
  \item The equal degree step done by |ZENPolyEDFactor|.
  \end{itemize}
  }
  */

#if 0
/*
  \TEX{
  \title{Factorization of a polynomial.}
  \input{A polynomial |P| to factorize.}
  \side{The factors of |P| are put in |p_lfact|.}
  \note{.}
  }
  */
int ZENPolyFactor(lfact, P, Rg)
     ZENPolyFact lfact;
     ZENPoly P;
     ZENRing Rg;{}
#endif
int _ZENPolyFactor(p_lfact, P, Rg)
     ZENPolyFact *p_lfact;
     ZENPoly P;
     ZENRing Rg;
{
  int ret;
  ZENPolyFact SFlfact, DDlfact, EDlfact;
  ZENPolyList SFlist, DDlist, EDlist;
  
  ZENPolyFactCreate(SFlfact, Rg);
#define mbf_F0 ZENPolyFactFree(SFlfact, Rg)
  ret = ZENPolySFFactor(SFlfact, P, Rg);
  CheckReturn(ret != ZEN_HAS_INVERSE, mbf_F0, return(ret));

  for(;;) {
    ret = ZENPolyFactRemove(SFlist, SFlfact, 0, Rg);
    CheckReturn(ret, mbf_F0, break);

    ZENPolyFactCreate(DDlfact, Rg);
#define mbf_F1 ZENPolyFactFree(DDlfact, Rg)
    ret = ZENPolyDDFactor(DDlfact, SFlist.P, Rg);
    ZENPolyListFree(SFlist, Rg);
    CheckReturn(ret != ZEN_HAS_INVERSE, mbf_F1; mbf_F0, return(ret)); 

    for(;;) {
      ret = ZENPolyFactRemove(DDlist, DDlfact, 0, Rg);
      CheckReturn(ret, mbf_F1, break);
      
      ZENPolyFactCreate(EDlfact, Rg);
#define mbf_F2 ZENPolyFactFree(EDlfact, Rg)
      ret = ZENPolyEDFactor(EDlfact, DDlist.P, DDlist.Xq, DDlist.k, Rg);
      ZENPolyListFree(DDlist, Rg);
      CheckReturn(ret != ZEN_HAS_INVERSE, mbf_F2; mbf_F1; mbf_F0, return(ret)); 
      for(;;) {
	ret = ZENPolyFactRemove(EDlist, EDlfact, 0, Rg);
	CheckReturn(ret, mbf_F2, break);
	EDlist.m *= DDlist.m*SFlist.m;

	ret = ZENPolyMakeMonic(EDlist.P, EDlist.P, Rg);
	CheckReturn(ret != ZEN_HAS_INVERSE, mbf_F2; mbf_F1; mbf_F0, return(ret));

	ret = ZENPolyFactAppend(*p_lfact, EDlist, Rg);
	CheckReturn(ret, mbf_F2; mbf_F1; mbf_F0, return(ZERR));
      }
    }
  }
  return(ZEN_HAS_INVERSE);
}


/*
  \TEX{
  The square free factorization of a polynomial $P(X)$ is a decomposition
  $P=P_1P_2^{2}P_3^{3}\ldots P_k^k$, with $P_1, \ldots,P_k$ square free
  and $(P_i, P_j) = 1$. The following algorithm is described in
  \cite{GeCzLa93}.
  }
  */

#if 0
/*
  \TEX{
  \title{Square free factorization of a polynomial.}
  \input{A polynomial |P| to factorize.}
  \output{}
  \side{The factors are put in |lfact|.}
  }
  */
int ZENPolySFFactor(lfact, P, Rg)
     ZENPolyFact lfact;
     ZENPoly P;
     ZENRing Rg;{}
#endif
int _ZENPolySFFactor(p_lfact, P, Rg)
     ZENPolyFact *p_lfact;
     ZENPoly P;
     ZENRing Rg;
{
  ZENPoly b, w, y, z;
  ZENPolyFact lfact;
  ZENPolyList list;
  int ret;

  ZENPolyAT(b, ZENPolyDeg(P, Rg), Rg, ;, return(ZERR));
#define mbf_SF0 ZENPolyFree(b, Rg)
  ZENPolyAT(w, ZENPolyDeg(P, Rg), Rg, mbf_SF0, return(ZERR));
#define mbf_SF1 ZENPolyFree(w, Rg); mbf_SF0
  ZENPolyAT(y, ZENPolyDeg(P, Rg), Rg, mbf_SF1, return(ZERR));
#define mbf_SF2 ZENPolyFree(y, Rg); mbf_SF1

  if (!ZENPolyDeg(P, Rg)) {
    list.m = 1; list.P = y; list.Xq = NULL;
    ZENPolyAssign(list.P, P, Rg);
    mbf_SF1;
    CheckReturn(ZENPolyFactAppend(*p_lfact, list, Rg), mbf_SF2, return(ZERR));
    return(ZEN_HAS_INVERSE);
  }

  list.m = 1;
  CheckReturn(ZENPolyDerive(b, P, Rg), mbf_SF2, return(ZERR));
  if (!ZENPolyIsZero(b, Rg)) {

    ret = ZENPolyGcd(b, P, b, Rg);
    CheckReturn(ret != ZEN_HAS_INVERSE, mbf_SF2, return(ret));
    
    ret = ZENPolyDivide(w, y, P, b, Rg);
    CheckReturn(ret != ZEN_HAS_INVERSE, mbf_SF2, return(ret));
    
    while(!(ZENPolyIsXi(w, 0, Rg))) {
      ret = ZENPolyGcd(y, w, b, Rg);
      CheckReturn(ret != ZEN_HAS_INVERSE, mbf_SF2, return(ret));
      
      if ((ZENPolyDeg(w, Rg)-ZENPolyDeg(y, Rg)) > 0) {
	ZENPolyAT(z, ZENPolyDeg(w, Rg)-ZENPolyDeg(y, Rg), Rg, mbf_SF2, return(ZERR));
	ret = ZENPolyDivide(z, w, w, y, Rg);
	CheckReturn(ret != ZEN_HAS_INVERSE, ZENPolyFree(z, Rg); mbf_SF2, return(ret));
	list.P = z; list.Xq = NULL; list.k = 0;
	CheckReturn(ZENPolyFactAppend(*p_lfact, list, Rg), mbf_SF2, return(ZERR));
      }
      list.m++;
      ret = ZENPolyDivide(b, w, b, y, Rg);
      CheckReturn(ret != ZEN_HAS_INVERSE, mbf_SF2, return(ret));

      ZENPolyAssign(w, y, Rg);
    }

    if (!ZENPolyDeg(b, Rg)) {
      mbf_SF2;
      return(ZEN_HAS_INVERSE);
    }
  }
  else {
    ZENPolyAssign(b, P, Rg);
  }
  CheckReturn(ZENPolyPRoot(b, b, Rg), mbf_SF2, return(ZERR));
  
  ZENPolyFactCreate(lfact, Rg);
  ret = ZENPolySFFactor(lfact, b, Rg);
  CheckReturn(ret != ZEN_HAS_INVERSE,
	      ZENPolyFactFree(lfact, Rg); mbf_SF2,
	      return(ret));
  
  for(;;) {
    CheckReturn(ZENPolyFactRemove(list, lfact, 0, Rg),;,break);
    list.m *= ZBNDigit(ZENRingP(Rg), 0);
    CheckReturn(ZENPolyFactAppend(*p_lfact, list, Rg),
		ZENPolyFactFree(lfact, Rg); mbf_SF2, return(ZERR));
  }
  return(ZEN_HAS_INVERSE);
}


/*
  \TEX{
  Once got a square free polynomial $P(X)$, we split the polynomial
  $P(X)$ in polynomaials of distinct degrees. This is the distinct
  degree factorization step. In fact, we get the factor of degree $d$ of
  $P(X)$ by $\gcd(X^{{2^n}^d}-X, P(X))$. Each factor is then completely
  factorized by |ZENPolyEDFactor| if the flag |edf| equals to 1. This
  function stops once a factor is found if the flag |of| is set.
  }
  */

#if 0
/*
  \TEX{
  \title{Distinct degree factorization of a polynomial.}
  \input{A polynomial |P| to factorize.}
  \side{The factors are put in |*p_lfact|.}
  }
  */
int ZENPolyDDFactor(lfact, P, Rg)
     ZENPolyFact lfact;
     ZENPoly P;
     ZENRing Rg;{}
#endif
int _ZENPolyDDFactor(p_lfact, P, Rg)
     ZENPolyFact *p_lfact;
     ZENPoly P;
     ZENRing Rg;
{
  ZENPolyList list;
  ZENPoly *Xqi, I, G, X;
  int nbXqi, i, pow, ret;
  ZENElt Xpq, Xq;
  ZENRing KRg;
  BigNum tmp, ptr;
  BigNumLength tmpl;
  ZENPrc prc;

  list.m = 1;
  if (ZENPolyDeg(P, Rg) < 2) {

    /* Polynomials of small degree are quite easy to factor !!! */
    ZENPolyAT(list.P, ZENPolyDeg(P, Rg), Rg, ;, return(ZERR));
    ZENPolyAssign(list.P, P, Rg);
    list.Xq = NULL; list.k = ZENPolyDeg(P, Rg);
    CheckReturn(ZENPolyFactAppend(*p_lfact, list, Rg),
		ZENPolyListFree(list, Rg),
		return(ZERR));
    return(ZEN_HAS_INVERSE);
  }


  /* Allcation of polynomials */
  ZENPolyAT(G, ZENPolyDeg(P, Rg), Rg, ;, return(ZERR));
#define mbf_DD0 ZENPolyFT(G, Rg)

  X = ZENZToPoly(ZENRingQ(Rg), ZENRingQl(Rg), Rg);
  CheckReturn(!X, mbf_DD0, return(ZERR));
#define mbf_DD1 ZENPolyFree(X, Rg); mbf_DD0

  MallocAT(Xqi, ZENPolyDeg(P, Rg), ZENPoly, mbf_DD1, return(ZERR));
  for(i = 0; i < (nbXqi = ZENPolyDeg(P, Rg)); i++) {Xqi[i] = NULL;}
#define mbf_DD2 for(i = 0; i < nbXqi; i++) {ZENPolyFT(Xqi[i], Rg);};; ZENFree(Xqi); mbf_DD1

  ZBNAT(tmp, ZENRingQl(Rg)*ZENPolyDeg(P, Rg), mbf_DD2, return(ZERR));
#define mbf_DD2a ZENFree(tmp); mbf_DD2

  /* Allocation of the extension */
  KRg = ZENExtRingCreate(P, Rg);
  CheckReturn(!KRg, mbf_DD2a, return(ZERR));
#define mbf_DD3 ZENRingClose(KRg); mbf_DD2a
  if (ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_ELT_MULTIPLY)) {
    ZENPrcSetNone(prc); 
    ZENPrcSet(prc, ZENPRC_ELT_MULTIPLY);
    ZENRingAddPrc(KRg, prc);
  }

  ZENEltAT(Xpq, KRg, mbf_DD3, return(ZERR));
#define mbf_DD4 ZENEltFree(Xpq, KRg); mbf_DD3

  ZENEltAT(Xq, KRg, mbf_DD4, return(ZERR));
#define mbf_DD5 ZENEltFree(Xq, KRg); mbf_DD4

  /* X^q mod P(X) */
  ZENZToElt(Xpq, ZENRingQ(Rg), ZENRingQl(Rg), KRg);
  ZENEltExp(Xpq, ZENRingQ(Rg), ZENRingQl(Rg), Xpq, KRg);

  ZENEltToZ(tmp, &tmpl, Xpq, KRg);
  Xqi[1] = ZENZToPoly(tmp, tmpl, Rg);
  CheckReturn(!Xqi[1], mbf_DD5, return(ZERR));

  /* gcd(X^q-X, P(X)) */
  ZENPolySubtract(Xqi[1], X, Rg);
  ret = ZENPolyGcd(G, ZENRingPol(KRg), Xqi[1], Rg);
  ZENPolyAdd(Xqi[1], X, Rg);
  CheckReturn(ret != ZEN_HAS_INVERSE, mbf_DD5, return(ret));

  if (ZENPolyDeg(G, Rg)) {
    
    /* A factor found */
    ZENPolyAT(list.P, ZENPolyDeg(G, Rg), Rg, mbf_DD5, return(ZERR));
#define mbf_DD6  ZENPolyFree(list.P, Rg); mbf_DD5
    ZENPolyAT(list.Xq, ZENPolyDeg(Xqi[1], Rg), Rg, mbf_DD6; mbf_DD6, return(ZERR));
#define mbf_DD7  ZENPolyFree(P, Rg); mbf_DD6

    ZENPolyAssign(list.P, G, Rg);
    ZENPolyDivide(G, list.Xq, P, G, Rg);
    CheckReturn(ret != ZEN_HAS_INVERSE, mbf_DD7, return(ret));
    list.k = 1;
    ret = ZENPolyDivide(list.Xq, list.Xq, Xqi[1], list.P, Rg);
    CheckReturn(ret != ZEN_HAS_INVERSE, mbf_DD7, return(ret));
    CheckReturn(ZENPolyFactAppend(*p_lfact, list, Rg), mbf_DD7, return(ZERR));

    CheckReturn(!ZENPolyDeg(G, Rg), mbf_DD5, return(ZEN_HAS_INVERSE));

    /* Freeing memory */
    ZENEltFree(Xpq, KRg); ZENEltFree(Xq, KRg); ZENRingClose(KRg);

    /* Allocation of the new extension */
    KRg = ZENExtRingCreate(G, Rg);
    CheckReturn(!KRg, mbf_DD2a, return(ZERR));
    if (ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_ELT_MULTIPLY)) {
      ZENPrcSetNone(prc); 
      ZENPrcSet(prc, ZENPRC_ELT_MULTIPLY);
      ZENRingAddPrc(KRg, prc);
    }
    ZENEltAT(Xpq, KRg, mbf_DD3, return(ZERR));
    ZENEltAT(Xq, KRg, mbf_DD4, return(ZERR));

    /* Updating some polynomials */
    ret = ZENPolyDivide(Xqi[1], Xqi[1], Xqi[1], ZENRingPol(KRg), Rg);
    CheckReturn(ret != ZEN_HAS_INVERSE, mbf_DD5, return(ret));
    ptr = ZENPolyToZ(&tmpl, Xqi[1], Rg);
    CheckReturn(!ptr, mbf_DD5, return(ZERR));
    ZENZToElt(Xpq, ptr, tmpl, KRg);
    ZBNF(ptr);
  }
  if (ZENRingDeg(KRg) < 4) {

    /* Polynomials of small degree are still quite easy to factor !!!  */
    ZENPolyAT(list.P, ZENRingDeg(KRg), Rg, mbf_DD5, return(ZERR));
    ZENPolyAT(list.Xq, ZENRingDeg(KRg)-1, Rg, mbf_DD6, return(ZERR));
    ZENPolyAssign(list.P, ZENRingPol(KRg), Rg);
    list.k = ZENRingDeg(KRg);
    ZENPolyAssign(list.Xq, Xqi[1], Rg);
    CheckReturn(ZENPolyFactAppend(*p_lfact, list, Rg), mbf_DD7, return(ZERR));
    mbf_DD5;
    return(ZEN_HAS_INVERSE);
  }

  /* Then, we compute X^(2*q), X^(3*q),... mod P(X)$ */
  ZENEltAssign(Xq, Xpq, KRg);
  for(i = 2; i < ZENRingDeg(KRg); i++) {
    
    ZENEltMultiply(Xq, Xpq, Xq, KRg);
    ZENEltToZ(tmp, &tmpl, Xq, KRg);
    Xqi[i] = ZENZToPoly(tmp, tmpl, Rg);
    CheckReturn(!Xqi[i], mbf_DD5, return(ZERR));
  }

  /* and finally $X^{q^d}$ for $d=2\ldots$ */
  ZENPolyAT(P, ZENRingDeg(KRg), Rg, mbf_DD5, return(ZERR));
  ZENPolyAssign(P, ZENRingPol(KRg), Rg);
  ZENPolyAT(Xqi[0], ZENRingDeg(KRg), Rg, mbf_DD5, return(ZERR));
  ZENPolyAT(I, ZENRingDeg(KRg), Rg, mbf_DD5, return(ZERR));
  ZENPolyAssign(I, Xqi[1], Rg);
#define mbf_DD8 ZENPolyFree(P, Rg); ZENPolyFree(I, Rg); mbf_DD5
  for(i = 2; i <= ZENPolyDeg(P, Rg)/2; i++) {
    
    ZENPolyDeg(G, Rg) = 0;
    ZENPolySetCoeff(G, 0, ZENPolyGetCoeffPtr(I, 0, Rg), Rg);
    for(pow = 1; pow <= ZENPolyDeg(I, Rg); pow++) {
      ZENPolyMultiplyScalar(Xqi[0], Xqi[pow], ZENPolyGetCoeffPtr(I, pow, Rg), Rg);
      ZENPolyAdd(G, Xqi[0], Rg);
    }
    ZENPolyAssign(I, G, Rg);

    /* gcd(X^q-X, P(X)) */
    ZENPolySubtract(I, X, Rg);
    ret = ZENPolyGcd(G, P, I, Rg);
    ZENPolyAdd(I, X, Rg);
    CheckReturn(ret != ZEN_HAS_INVERSE, mbf_DD8, return(ret));

    if (ZENPolyDeg(G, Rg)) {
      int j;

      /* A factor found */      
      ZENPolyAT(list.P, ZENPolyDeg(G, Rg), Rg, mbf_DD8, return(ZERR));
#define mbf_DD9  ZENPolyFree(list.P, Rg); mbf_DD8
      ZENPolyAT(list.Xq, ZENPolyDeg(P, Rg), Rg, mbf_DD9, return(ZERR));
#define mbf_DD10  ZENPolyFree(list.Xq, Rg); mbf_DD9

      ZENPolyAssign(list.P, G, Rg);
      list.k = i;
      ret = ZENPolyDivide(P, list.Xq, P, G, Rg);
      CheckReturn(ret != ZEN_HAS_INVERSE, mbf_DD10, return(ret));
      ret = ZENPolyDivide(list.Xq, list.Xq, Xqi[1], G, Rg);
      CheckReturn(ret != ZEN_HAS_INVERSE, mbf_DD10, return(ret));
      CheckReturn(ZENPolyFactAppend(*p_lfact, list, Rg), mbf_DD10, return(ZERR));

      CheckReturn(!ZENPolyDeg(P, Rg), mbf_DD8, return(ZEN_HAS_INVERSE));

      /* Updating some polynomials */
      ret = ZENPolyDivide(I, I, I, P, Rg);
      CheckReturn(ret != ZEN_HAS_INVERSE, mbf_DD8, return(ret));
      for(j= 1; j < ZENPolyDeg(P, Rg); j++) {
	ret = ZENPolyDivide(Xqi[j], Xqi[j], Xqi[j], P, Rg);
	CheckReturn(ret != ZEN_HAS_INVERSE, mbf_DD8, return(ret));
      }
    }
  }
  if (ZENPolyDeg(P, Rg)) { 
    ZENPolyAT(list.P, ZENPolyDeg(P, Rg), Rg, mbf_DD8, return(ZERR));
    ZENPolyAssign(list.P, P, Rg);
    list.k = ZENPolyDeg(P, Rg); list.Xq = Xqi[1]; Xqi[1] = NULL;
    CheckReturn(ZENPolyFactAppend(*p_lfact, list, Rg), mbf_DD10, return(ZERR));
  }
  mbf_DD8;
  return(ZEN_HAS_INVERSE);
}

/*
  \TEX{
  To split a polynomial $P(X)$ in factors of degree $k$, we use a
  algorithm due to Thion Ly \cite{Ly89}. 
  }
  */

#if 0
/*
  \TEX{
  \title{Equal degree polynomial factorization.}
  \input{A polynomial |P| to factorize, a polynomial |XqmodP| equal to
  $X^{2^n}\bmod P(X)$, the degree $k$ of the factors of $P(X)$.}
  \side{The factors are put in |*p_lfact|.}
  }
  */
int ZENPolyEDFactor(lfact, P, XqmodP, k, Rg)
     ZENPolyFact lfact;
     ZENElt P, XqmodP;
     int k;
     ZENRing Rg;{}
#endif
int _ZENPolyEDFactor(p_lfact, P, XqmodP, k, Rg)
     ZENPolyFact *p_lfact;
     ZENPoly P, XqmodP;
     int k;
     ZENRing Rg;
{
  ZENPoly Q, R, S;
  ZENPolyList list;
  int ret;

  list.m = 1; list.k = 0; list.Xq = NULL;

  if (ZENPolyDeg(P, Rg) == k) {
    ZENPolyAT(list.P, ZENPolyDeg(P, Rg), Rg, ;, return(ZERR));
    ZENPolyAssign(list.P, P, Rg);
    CheckReturn(ZENPolyFactAppend(*p_lfact, list, Rg), ZENPolyFree(list.P, Rg), return(ZERR));
    return(ZEN_HAS_INVERSE);
  }

  if (ZENPolyDeg(P, Rg) == 2) {
    ZENPoly Q, R;
    ZENElt elt;

    ZENPolyAT(Q, 1, Rg, ;, return(ZERR));
#define mbf_ED10 ZENPolyFree(Q, Rg)
    ZENPolyAT(R, 1, Rg, mbf_ED10, return(ZERR));
#define mbf_ED11 ZENPolyFree(R, Rg); mbf_ED10
    ret = ZENPolyRootsDegree2(Q, P, Rg);
    CheckReturn(ret != ZEN_HAS_INVERSE, mbf_ED11, return(ret));
    CheckReturn(ZENPolyDeg(Q, Rg) != 1, mbf_ED11, return(ret));
    
    ZENPolySetToXi(R, 1, Rg);

    elt = ZENPolyGetCoeffPtr(R, 0, Rg);
    ZENEltNegate(elt, ZENPolyGetCoeffPtr(Q, 0, Rg), Rg);
    ZENPolySetCoeff(R, 0, elt, Rg);
    ret = ZENPolyEDFactor(*p_lfact, R, NULL, k, Rg);
    CheckReturn(ret != ZEN_HAS_INVERSE, mbf_ED11, return(ret));

    elt = ZENPolyGetCoeffPtr(R, 0, Rg);
    ZENEltNegate(elt, ZENPolyGetCoeffPtr(Q, 1, Rg), Rg);
    ZENPolySetCoeff(R, 0, elt, Rg);
    ret = ZENPolyEDFactor(*p_lfact, R, NULL, k, Rg);
    mbf_ED11;
    return(ret);
  }
  
  if ((ZENRingPl(Rg) > 1) ||
      (ZBNDigit(ZENRingP(Rg), 0) != 2)) {
    Q = ZENPolyCamion(P, k, Rg);
  }
  else {
    Q = ZENPolyThiongLy(P, XqmodP, k, Rg);
  }
  CheckReturn(!Q, ;, return(ZERR));
#define mbf_ED0 ZENPolyFree(Q, Rg);

  ZENPolyAT(S, ZENPolyDeg(P, Rg), Rg, mbf_ED0, return(ZERR));
#define mbf_ED1 ZENPolyFree(S, Rg); mbf_ED0

  ret = ZENPolyDivide(S, S, XqmodP, Q, Rg);
  CheckReturn(ret != ZEN_HAS_INVERSE, mbf_ED1, return(ret));
  ret = ZENPolyEDFactor(*p_lfact, Q, S, k, Rg);
  CheckReturn(ret != ZEN_HAS_INVERSE, mbf_ED1, return(ret));
  
  ZENPolyAT(R, ZENPolyDeg(P, Rg)-ZENPolyDeg(Q, Rg), Rg, mbf_ED1, return(ZERR));
#define mbf_ED2 ZENPolyFree(R, Rg); mbf_ED1
  
  ret = ZENPolyDivide(R, S, P, Q, Rg);
  CheckReturn(ret != ZEN_HAS_INVERSE, mbf_ED2, return(ret));
  ret = ZENPolyDivide(S, S, XqmodP, R, Rg);
  CheckReturn(ret != ZEN_HAS_INVERSE, mbf_ED2, return(ret));
  ret = ZENPolyEDFactor(*p_lfact, R, S, k, Rg);
  CheckReturn(ret != ZEN_HAS_INVERSE, mbf_ED2, return(ret));
  
  mbf_ED2;
  return(ZEN_HAS_INVERSE);
}

/*
  \TEX{
  The Thiong Ly's  algorithm first find a
  $i$ such that $TX^i=X^i+\ldots+X^{i}$ is different than a constante
  modulo $P(X)$ and then a $j$ such that $u(X)=Tr(T^jTX^i)$ is
  different than a constant modulo $P(X)$. Finally $\gcd(u(X), P(X))$ is
  a non trivial factor of $P(X)$.
  }
  */

#if 0
/*
  \TEX{
  \title{.}
  \input{A polynomial |P| to factorize, $X^{2^n} \bmod P(X)$ and $k$,
  the degree of the factors of |P|.}
  \output{A factor of |P| is returned.}
  }
  */
ZENPoly ZENPolyThiongLy(P, XqmodP, k, Rg)
     ZENPoly P, XqmodP;
     int k;
     ZENRing Rg;{}
#endif
ZENPoly _ZENPolyThiongLy(P, XqmodP, k, Rg)
     ZENPoly P, XqmodP;
     int k;
     ZENRing Rg;
{
  ZENPoly Q, R;
  ZENElt Xj, X2, Xq, Xpq, T, t;
  int i, j, ret;
  ZENPrc prc;
  ZENRing KRg, rg;
  BigNum tmp, ptr;
  BigNumLength tmpl;
  int AbsDeg;
  
  /* Allocation of the extension */
  KRg = ZENExtRingCreate(P, Rg);
  CheckReturn(!KRg, ;, return(ZNULL));
#define mbf_LY0 ZENRingClose(KRg)
  if (ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_ELT_MULTIPLY)) {
    ZENPrcSetNone(prc); 
    ZENPrcSet(prc, ZENPRC_ELT_MULTIPLY);
    ZENRingAddPrc(KRg, prc);
  }

  ZENEltAT(Xpq, KRg, mbf_LY0, return(ZNULL));
#define mbf_LY1 ZENEltFree(Xpq, KRg); mbf_LY0

  ZENEltAT(Xq, KRg, mbf_LY1, return(ZNULL));
#define mbf_LY2 ZENEltFree(Xq, KRg); mbf_LY1

  ZENEltAT(X2, KRg, mbf_LY2, return(ZNULL));
#define mbf_LY2a ZENEltFree(X2, KRg); mbf_LY2

  ZENEltAT(Xj, KRg, mbf_LY2a, return(ZNULL));
#define mbf_LY3 ZENEltFree(Xj, KRg); mbf_LY2a

  ZENEltAT(T, KRg, mbf_LY3, return(ZNULL));
#define mbf_LY3a ZENEltFree(T, KRg); mbf_LY3

  ZBNAT(tmp, ZENPolyDeg(P, Rg)*ZENRingQl(Rg), mbf_LY3a, return(ZNULL));
#define mbf_LY4 ZBNF(tmp); mbf_LY3a

  ZENPolyAT(Q, ZENPolyDeg(P, Rg), Rg, mbf_LY4, return(ZNULL));
#define mbf_LY5 ZENPolyFree(Q, Rg); mbf_LY4

  /* The absolute degree of Rg */
  AbsDeg = 1;
  for(rg = Rg; ZENRingDef(rg); rg = ZENRingDef(rg)) {
    AbsDeg *= ZENRingDeg(rg);
  }

  /* X^q mod P(X) */
  ptr = ZENPolyToZ(&tmpl, XqmodP, Rg);
  ZENZToElt(Xpq, ptr, tmpl, KRg);
  ZBNF(ptr);
  
  for(j = 1; j < (k<<1); j += 2) {

    /* T(X^j) */
    ZENPolySetToXi(Q, j, Rg);
    ptr = ZENPolyToZ(&tmpl, Q, Rg);
    ZENZToElt(Xj, ptr, tmpl, KRg);
    ZBNF(ptr);
    
    ret = ZENShoopTrace(T, Xj, k, Xpq, KRg);
    CheckReturn(ret != ZEN_HAS_INVERSE, mbf_LY5, return(ZNULL));

    ZENEltToZ(tmp, &tmpl, T, KRg);
    R = ZENZToPoly(tmp, tmpl, Rg);
    CheckReturn(!R, mbf_LY5, return(ZNULL));
    if (ZENPolyDeg(R, Rg)) {break;}
    ZENPolyFree(R, Rg);
  }
      
  ret = ZENPolyGcd(Q, P, R, Rg);
  ZENPolyFree(R, Rg);
  CheckReturn(ret != ZEN_HAS_INVERSE, mbf_LY5, return(ZNULL));
  if (ZENPolyDeg(Q, Rg)) {
    mbf_LY4;
    return(Q);
  }

  /* T'T^iTX^j */
  for(;;) {

    ZENEltAT(t, Rg, mbf_LY5, return(ZNULL));
    ZENEltSetRandom(t, Rg);
    ZENEltToZ(tmp, &tmpl, t, Rg);
    ZENEltFree(t, Rg);
    ZENZToElt(Xq, tmp, tmpl, KRg);
    ZENEltMultiply(Xq, T, Xq, KRg);

    ZENEltAssign(X2, Xq, KRg);
    ZENEltAssign(Xj, Xq, KRg);
    for(i = 1; i < AbsDeg; i++) {
      ZENEltSquare(X2, X2, KRg);
      ZENEltAdd(Xj, X2, KRg);
    }
    ZENEltToZ(tmp, &tmpl, Xj, KRg);
    R = ZENZToPoly(tmp, tmpl, Rg);
    CheckReturn(!R, mbf_LY5, return(ZNULL));
    if (ZENPolyDeg(R, Rg)) {
      ret = ZENPolyGcd(Q, P, R, Rg);
      ZENPolyFree(R, Rg);
      CheckReturn(ret != ZEN_HAS_INVERSE, mbf_LY5, return(ZNULL));
      if (ZENPolyDeg(Q, Rg)) {break;}
    }
    else {
      ZENPolyFree(R, Rg);
    }
  }

  mbf_LY4;
  return(Q);
}



#if 0
/*
  \TEX{
  \title{.}
  \input{A polynomial |P| to factorize in a finite field of
  odd characteristic, $X^{2^n} \bmod P(X)$ and $k$,
  the degree of the factors of |P|.}
  \output{A factor of |P| is returned.}
  }
  */
ZENPoly ZENPolyCamion(P, k, Rg)
     ZENPoly P;
     int k;
     ZENRing Rg;{}
#endif
ZENPoly _ZENPolyCamion(P, k, Rg)
     ZENPoly P;
     int k;
     ZENRing Rg;
{
  ZENPoly  S, R;
  ZENElt Ku, u, elt;
  BigNum tmp, e, n;
  BigNumLength tmpl, el, nl;
  int i;
  ZENRing KRg;
  ZENPrc prc;

  /* The extension defined by P */
  KRg = ZENExtRingCreate(P, Rg);
  CheckReturn(!KRg, ;, return(ZNULL));
#define mbf_OC1 ZENRingClose(KRg)
  if (ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_ELT_MULTIPLY)) {
    ZENPrcSetNone(prc); 
    ZENPrcSet(prc, ZENPRC_ELT_MULTIPLY);
    ZENRingAddPrc(KRg, prc);
  }
  
  ZBNAT(tmp, ZENRingQl(KRg), mbf_OC1, return(ZNULL));
#define mbf_OC2 ZBNF(tmp); mbf_OC1

  ZENEltAT(Ku, KRg, mbf_OC2, return(ZNULL));
#define mbf_OC3 ZENEltFree(Ku, KRg); mbf_OC2

  ZENEltAT(u, Rg, mbf_OC3, return(ZNULL));
#define mbf_OC4 ZENEltFree(u, Rg); mbf_OC3

  ZBNAT(e, ZENRingQl(KRg), mbf_OC4, return(ZNULL));
#define mbf_OC5 ZENFree(e); mbf_OC4

  ZENPolyAT(R, ZENPolyDeg(P, Rg), Rg, mbf_OC5, return(ZNULL));
#define mbf_OC6 ZENPolyFree(R, Rg); mbf_OC5
 
  el = ZENRingQl(Rg);
  ZBNAssign(e, ZENRingQ(Rg), el);
  for(i = 1; i < k; i++) {
    nl = el+ZENRingQl(Rg)+1;
    CheckReturn(!(n = ZBNC(nl)), mbf_OC6, return(ZNULL));
    ZBNSetToZero(n, nl);
    ZBNMultiply(n, nl, e, el, ZENRingQ(Rg), ZENRingQl(Rg));
    ZBNF(e);
    e = n;
    el = ZBNNumDigits(n, nl);
  }
  ZBNShiftRight(e, el, 1);
  el = ZBNNumDigits(e, el);

  for(;;) {
    ZENEltSetRandom(Ku, KRg);
    ZENEltExp(Ku, e, el, Ku, KRg);

    ZENEltToZ(tmp, &tmpl, Ku, KRg);
    S = ZENZToPoly(tmp, tmpl, Rg);
    CheckReturn(!S, mbf_OC6, return(ZNULL));

    ZENEltSetToOne(u, Rg);
    elt = ZENPolyGetCoeffPtr(S, 0, Rg);
    ZENEltSubtract(elt, u, Rg);
    ZENPolySetCoeff(S, 0, elt, Rg);

    if (!ZENPolyIsZero(S, Rg)) {
      CheckReturn(ZENPolyGcd(R, P, S, Rg) != ZEN_HAS_INVERSE,
		  ZENPolyFree(S, Rg); mbf_OC6,
		  return(ZNULL));
      ZENPolyFree(S, Rg);
      if (ZENPolyDeg(R, Rg)) {break;}
    }
    else {
      ZENPolyFree(S, Rg);
    }
  }
  mbf_OC5;
  return(R);
}

/*
  \TEX{
  A recursive algorithm due to Shoop \cite{Sh92} computes traces of
  polynomial faster.
  }
  */
#if 0
/*
  \TEX{
  \title{Computing $Q(X)+Q(X)^{2^n}+...+Q(X)^(2^n(k-1)) \bmod P(X)$.}
  \input{The polynomial |Q|, the modulo |P|, |XqmodP|=$X^{2^n}\bmod
  P(X)$ and the integer |k|.}
  \side{The trace is put in |T|.}
  }
  */
int ZENShoopTrace(T, Q, k, XqmodP, KRg)
     ZENElt T, Q, XqmodP;
     int k;
     ZENRing KRg;{}
#endif
int _ZENShoopTrace(T, Q, k, XqmodP, KRg)
     ZENElt T, Q, XqmodP;
     int k;
     ZENRing KRg;
{
  ZENElt u, v, s, t;
  ZENElt *upowmodP;
  ZENPoly R;
  int kp, i;
  BigNum tmp;
  BigNumLength tmpl;
  ZENRing Rg;

  Rg = ZENRingDef(KRg);

  ZENEltAT(u, KRg, ;, return(ZERR));
#define mbf_SR0 ZENEltFree(u, KRg)

  ZENEltAT(v, KRg, mbf_SR0, return(ZERR));
#define mbf_SR1 ZENEltFree(v, KRg); mbf_SR0

  ZENEltAT(s, KRg, mbf_SR1, return(ZERR));
#define mbf_SR2 ZENEltFree(s, KRg); mbf_SR1

  ZENEltAT(t, KRg, mbf_SR2, return(ZERR));
#define mbf_SR2a ZENEltFree(t, KRg); mbf_SR2

  ZBNAT(tmp, ZENRingQl(KRg), mbf_SR2a, return(ZERR));
#define mbf_SR3 ZENFree(tmp); mbf_SR2a

  MallocAT(upowmodP, ZENRingDeg(KRg), ZENElt, mbf_SR3, return(ZERR));
  for(i = 1; i < ZENRingDeg(KRg); i++) {
    ZENEltAT(upowmodP[i], KRg, 
	     for(i--; i >=1; i--) {ZENEltFree(upowmodP[i], KRg);}; ZENFree(upowmodP); mbf_SR3, return(ZERR));
  }
#define mbf_SR4 for(i = 1; i < ZENRingDeg(KRg); i++) {ZENEltFree(upowmodP[i], KRg);}; ZENFree(upowmodP); mbf_SR3

  ZENEltAssign(u, XqmodP, KRg);
  ZENEltAssign(v, Q, KRg);
  ZENEltSetToZero(T, KRg);
  

  for (kp = k; kp > 1; kp >>= 1) {
    ZENEltAssign(upowmodP[1], u, KRg);
    for(i = 2; i < ZENRingDeg(KRg); i++) {
      if (i & 1) {
	ZENEltMultiply(upowmodP[i], upowmodP[i-1], upowmodP[1], KRg);
      }
      else {
	ZENEltSquare(upowmodP[i], upowmodP[i>>1], KRg);
      }
    }

    ZENEltToZ(tmp, &tmpl, u, KRg);
    R = ZENZToPoly(tmp, tmpl, Rg);    
    CheckReturn(!R, mbf_SR4, return(ZERR));   

    ZENEltToZ(tmp, &tmpl, ZENPolyGetCoeffPtr(R, 0, Rg), Rg);
    ZENZToElt(s, tmp, tmpl, KRg);

    for(i = 1; i <= ZENPolyDeg(R, Rg); i++) {
      ZENEltToZ(tmp, &tmpl, ZENPolyGetCoeffPtr(R, i, Rg), Rg);
      ZENZToElt(t, tmp, tmpl, KRg);
      ZENEltMultiply(t, upowmodP[i], t, KRg);
      ZENEltAdd(s, t, KRg);
    }
    ZENEltAssign(u, s, KRg);
    ZENPolyFree(R, Rg);

    if (kp&1) {
      ZENEltToZ(tmp, &tmpl, T, KRg);
      R = ZENZToPoly(tmp, tmpl, Rg);      
      CheckReturn(!R, mbf_SR4, return(ZERR)); 
      
      ZENEltToZ(tmp, &tmpl, ZENPolyGetCoeffPtr(R, 0, Rg), Rg);
      ZENZToElt(s, tmp, tmpl, KRg);
      
      for(i = 1; i <= ZENPolyDeg(R, Rg); i++) {
	ZENEltToZ(tmp, &tmpl, ZENPolyGetCoeffPtr(R, i, Rg), Rg);
	ZENZToElt(t, tmp, tmpl, KRg);
	ZENEltMultiply(t, upowmodP[i], t, KRg);
	ZENEltAdd(s, t, KRg);
      }
      ZENEltAssign(T, s, KRg);
      ZENEltAdd(T, v, KRg);
      ZENPolyFree(R, Rg);
    }

    ZENEltToZ(tmp, &tmpl, v, KRg);
    R = ZENZToPoly(tmp, tmpl, Rg);    
    CheckReturn(!R, mbf_SR4, return(ZERR));

    ZENEltToZ(tmp, &tmpl, ZENPolyGetCoeffPtr(R, 0, Rg), Rg);
    ZENZToElt(s, tmp, tmpl, KRg);

    for(i = 1; i <= ZENPolyDeg(R, Rg); i++) {
      ZENEltToZ(tmp, &tmpl, ZENPolyGetCoeffPtr(R, i, Rg), Rg);
      ZENZToElt(t, tmp, tmpl, KRg);
      ZENEltMultiply(t, upowmodP[i], t, KRg);
      ZENEltAdd(s, t, KRg);
    }
    ZENEltAdd(v, s, KRg);
    ZENPolyFree(R, Rg);
  }

  ZENEltAssign(upowmodP[1], u, KRg);
  for(i = 2; i < ZENRingDeg(KRg); i++) {
    if (i & 1) {
      ZENEltMultiply(upowmodP[i], upowmodP[i-1], upowmodP[1], KRg);
    }
    else {
      ZENEltSquare(upowmodP[i], upowmodP[i>>1], KRg);
    }
  }

  ZENEltToZ(tmp, &tmpl, T, KRg);
  R = ZENZToPoly(tmp, tmpl, Rg);    
  CheckReturn(!R, mbf_SR4, return(ZERR));

  ZENEltToZ(tmp, &tmpl, ZENPolyGetCoeffPtr(R, 0, Rg), Rg);
  ZENZToElt(s, tmp, tmpl, KRg);
  
  for(i = 1; i <= ZENPolyDeg(R, Rg); i++) {
    ZENEltToZ(tmp, &tmpl, ZENPolyGetCoeffPtr(R, i, Rg), Rg);
    ZENZToElt(t, tmp, tmpl, KRg);
    ZENEltMultiply(t, upowmodP[i], t, KRg);
    ZENEltAdd(s, t, KRg);
  }
  ZENEltAssign(T, s, KRg);
  ZENEltAdd(T, v, KRg);  
  ZENPolyFree(R, Rg);
  mbf_SR4;
  return(ZEN_HAS_INVERSE);

}

int _ZENPolyPRoot(P, Q, Rg)
     ZENPoly P, Q;
     ZENRing Rg;
{
  int i;
  BigNumDigit p;
  ZENElt elt;

  p = ZBNDigit(ZENRingP(Rg), 0);
  for(i = 0; i <= ZENPolyDeg(Q, Rg); i += p) {
    elt = ZENPolyGetCoeffPtr(P, i/p, Rg);
    CheckReturn(ZENEltPRoot(elt, ZENPolyGetCoeffPtr(Q, i, Rg), Rg), ;, return(ZERR));
    ZENPolySetCoeff(P, i/p, elt, Rg);
  }
  ZENPolyDeg(P, Rg) = ZENPolyDeg(Q, Rg)/p;
  return(ZEN_HAS_INVERSE);
}

