/*
 * zenfact/inputs/zparse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

				/* Concept index */
enum glo1 {			/* first level */
  bignum,
  optimization,
  polynom,
  rings,
  series,
  elements,
  curve,
  matrix,
  nbobjets
};

char *objets[nbobjets] = {	
  "big integers",
  "optimization",
  "polynomials",
  "rings",
  "series",
  "elements",
  "elliptic curves",
  "matrices",
};

enum glo2 {			/* second level */
  point,
  permutation,
  digit,
  modular,
  flag,
  clone,
  coeff,
  prc,
  prime,
  irreduc,
  monic,
  debug,
  nblev2
};

char *objets2[nblev2] = {
  "points",
  "permutation",
  "digit",
  "modular",
  "flags",
  "clones",
  "coefficients",
  "precomputations",
  "prime",
  "irreducible",
  "monic",
  "debugging"
};
  
enum glo3 {			/* third level */
  addition,
  negation,
  weight,
  mult,
  exp,
  inverse,
  sub,
  squ,
  create,
  shift,
  assign,
  compare,
  test,
  freeing,
  io,
  trace,
  convert,
  dup,
  gauss,
  det,
  kernel,
  permute,
  rank,
  mod,
  transpose,
  deg,
  derive,
  divide,
  dot,
  gcd,
  root,
  length,
  val,
  resultant,
  eval,
  flags,
  logarithm,
  doubling,
  factor,
  nblev3
};
  
char *objets3[nblev3] = {
  "addition",
  "negation",
  "weight",
  "multiplication",
  "exponentiation",
  "inverse",
  "subtraction",
  "square",
  "create",
  "shifting",
  "assignation",
  "comparison",
  "test",
  "freeing memory",
  "input/output",
  "trace",
  "conversion",
  "duplication",
  "gaussian elimination",
  "determinant",
  "kernel",
  "permuting",
  "rank",
  "modular reduction",
  "transpose",
  "degree",
  "derivation",
  "division",
  "scalar product",
  "greatest common divisor",
  "root",
  "length",
  "valuation",
  "resultant",
  "evaluation",
  "flags",
  "logarithm",
  "double",
  "factorization"
};


				/* FUNCTION INDEX */
enum ind2 {			/* second level */
  generic,
  tests,
  ze2m,
  ze2,
  defaults,
  zext,
  zeps,
  zep,
  zelog,
  zetab,
  zef,
  zec,
  nbfct
};

char *fctlev2[nbfct] = {
  "generic functions",
  "library tests",
  "extension over $\\GF{2}$",
  "$\\GF{2}$",
  "defaults functions",
  "extensions",
  "small modulus functions",
  "modular case",
  "clones with logarithms",
  "clones with tabulation",
  "rational functions",
  "chinese remainder construction"
};

#define check(level,str,ce1,ce2,ce3,le) \
    if((lvl<=level)&&(strncmp(pt,str,strlen(str))==0)) \
    { pt+=strlen(str); \
      if(ce1>=0) ce.l1=ce1; \
      if(ce2>=0) ce.l2=ce2; \
      if(ce3>=0) ce.l3=ce3; \
      if(ce1>=0) fe.l1=ce1; \
      if(le>=0) fe.l2=le; \
      lvl++; } else

#define IndexCheckAll()					\
do {							\
  check(0,"ZBN",bignum,-1,-1,-1);			\
  check(0,"ZEN",-1,-1,-1,-1);				\
  check(0,"\\_ZEN",-1,-1,-1,generic);			\
  check(0,"\\_\\_ZEN",-1,-1,-1,generic);		\
  check(0,"\\_Ze2m",-1,-1,-1,ze2m);			\
  check(0,"\\_Ze2",-1,-1,-1,ze2);			\
  check(0,"\\_\\_Ze2",-1,-1,-1,ze2);			\
  check(0,"Ze2",-1,-1,-1,ze2);				\
  check(0,"\\_Zeds",-1,-1,-1,defaults);			\
  check(0,"Zeds",-1,-1,-1,defaults);			\
  check(0,"\\_Zed",-1,-1,-1,defaults);			\
  check(0,"Zed",-1,-1,-1,defaults);			\
  check(0,"\\_Zext",-1,-1,-1,zext);			\
  check(0,"Zext",-1,-1,-1,zext);			\
  check(0,"\\_Zeps",-1,-1,-1,zeps);			\
  check(0,"Zeps",-1,-1,-1,zeps);			\
  check(0,"\\_Zep",-1,-1,-1,zep);			\
  check(0,"Zep",-1,-1,-1,zep);				\
  check(0,"\\_Zelog",-1,-1,-1,zelog);			\
  check(0,"Zelog",-1,-1,-1,zelog);			\
  check(0,"\\_Zetab",-1,-1,-1,zetab);			\
  check(0,"Zetab",-1,-1,-1,zetab);			\
  check(0,"\\_Zef",-1,-1,-1,zef);			\
  check(0,"Zef",-1,-1,-1,zef);				\
							\
  check(0,"div",bignum,digit,-1,-1);			\
  check(0,"mod",bignum,digit,-1,-1);			\
  check(0,"Bloc",bignum,digit,-1,-1);			\
  check(0,"ZBP",bignum,debug,io,-1);			\
  check(0,"ZBR",bignum,debug,io,-1);			\
  check(0,"ZEP",elements,debug,io,-1);			\
  check(0,"ZER",elements,debug,io,-1);			\
  check(0,"ZPP",polynom,debug,io,-1);			\
  check(0,"ZPR",polynom,debug,io,-1);			\
  check(0,"ZMP",matrix,debug,io,-1);			\
  check(0,"ZMR",matrix,debug,io,-1);			\
  check(0,"ZSP",series,debug,io,-1);			\
  check(0,"ZSR",series,debug,io,-1);			\
  check(0,"DDF",polynom,irreduc,test,-1);		\
  check(0,"Berlekamp",polynom,irreduc,test,-1);		\
							\
  check(1,"SetError",-1,debug,-1,-1);			\
  check(1,"Error",-1,debug,-1,-1);			\
  check(1,"ERR",-1,debug,-1,nbfct);			\
  check(1,"NULL",-1,debug,-1,nbfct);			\
  check(1,"Elt",elements,-1,-1,-1);			\
  check(1,"SetMemory",optimization,-1,-1,-1);		\
  check(1,"Clone",elements,-1,convert,-1);		\
  check(1,"Poly",polynom,-1,-1,-1);			\
  check(1,"Vect",matrix,-1,convert,-1);			\
  check(1,"Mat",matrix,-1,-1,-1);			\
  check(1,"Permutation",matrix,permutation,-1,-1);	\
  check(1,"Sr",series,-1,-1,-1);			\
  check(1,"MEc",curve,-1,-1,-1);			\
  check(1,"Ec",curve,-1,-1,-1);				\
  check(1,"Base",-1,-1,create,zep);			\
  check(1,"Shoop",polynom,-1,-1,-1);			\
  check(1,"Ext",-1,-1,create,zext);			\
  check(1,"Two",-1,-1,-1,-1);				\
  check(1,"GetFrom",-1,-1,io,-1);			\
  check(1,"Get",-1,-1,assign,-1);			\
  check(1,"Num",-1,digit,length,-1);			\
  check(1,"Set",-1,-1,assign,-1);			\
  check(1,"Long",-1,-1,-1,-1);				\
  check(1,"Small",-1,-1,-1,-1);				\
  check(1,"Any",-1,-1,-1,-1);				\
  check(1,"Mod",-1,modular,-1,-1);			\
							\
  check(2,"Pt",-1,point,-1,-1);				\
  check(2,"ZTo",bignum,-1,convert,-1);			\
  check(2,"To",-1,-1,convert,-1);			\
  check(2,"Absolute",-1,-1,-1,-1);			\
  check(2,"Long",-1,-1,-1,-1);				\
  check(2,"Small",-1,-1,-1,-1);				\
  check(2,"Col",-1,-1,-1,-1);				\
  check(2,"Row",-1,-1,-1,-1);				\
  check(2,"One",-1,-1,-1,-1);				\
  check(2,"SameType",-1,-1,compare,-1);			\
  check(2,"OppositeType",-1,-1,compare,-1);		\
  check(2,"Ring",rings,-1,-1,-1);			\
							\
  check(3,"Cln",optimization,clone,flags,-1);		\
  check(3,"Prc",optimization,prc,flags,-1);		\
  check(3,"Clone",optimization,clone,-1,-1);		\
  check(3,"ulo",-1,-1,mod,-1);				\
  check(3,"Log",-1,-1,logarithm,-1);			\
  check(3,"Sqrt",-1,-1,root,-1);			\
  check(3,"AddPrc",optimization,prc,-1,-1);		\
  check(3,"RmPrc",optimization,prc,-1,-1);		\
  check(3,"Add",-1,-1,addition,-1);			\
  check(3,"Negate",-1,-1,negation,-1);			\
  check(3,"Complement",-1,-1,negation,-1);		\
  check(3,"Multiply",-1,-1,mult,-1);			\
  check(3,"Mult",-1,-1,mult,-1);			\
  check(3,"Inverse",-1,-1,inverse,-1);			\
  check(3,"Subtract",-1,-1,sub,-1);			\
  check(3,"Double",-1,-1,doubling,-1);			\
  check(3,"SquareRoot",-1,-1,root,-1);			\
  check(3,"Squ",-1,-1,squ,-1);				\
  check(3,"Alloc",-1,-1,create,-1);			\
  check(3,"Create",-1,-1,create,-1);			\
  check(3,"Chinese",-1,-1,create,zec);			\
  check(3,"Set",-1,-1,assign,-1);			\
  check(3,"Next",-1,-1,assign,-1);			\
  check(3,"Weight",-1,-1,weight,-1);			\
  check(3,"Unset",-1,-1,assign,-1);			\
  check(3,"Are",-1,-1,compare,-1);			\
  check(3,"Assign",-1,-1,assign,-1);			\
  check(3,"Exp",-1,-1,exp,-1);				\
  check(3,"Free",-1,-1,freeing,-1);			\
  check(3,"GetFrom",-1,-1,io,-1);			\
  check(3,"Get",-1,-1,assign,-1);			\
  check(3,"Print",-1,-1,io,-1);				\
  check(3,"Put",-1,-1,io,-1);				\
  check(3,"Read",-1,-1,io,-1);				\
  check(3,"Trace",-1,-1,trace,-1);			\
  check(3,"Init",-1,-1,-1,-1);				\
  check(3,"Convert",-1,-1,convert,-1);			\
  check(3,"Copy",-1,-1,dup,-1);				\
  check(3,"Gauss",-1,-1,gauss,-1);			\
  check(3,"Det",-1,-1,det,-1);				\
  check(3,"Kernel",-1,-1,kernel,-1);			\
  check(3,"Permute",-1,-1,permute,-1);			\
  check(3,"Rank",-1,-1,rank,-1);			\
  check(3,"Transpose",-1,-1,transpose,-1);		\
  check(3,"Karatsuba",-1,-1,-1,-1);			\
  check(3,"Winograd",-1,-1,mult,-1);			\
  check(3,"Derive",-1,-1,derive,-1);			\
  check(3,"Normalize",-1,-1,-1,-1);			\
  check(3,"Divide",-1,-1,divide,-1);			\
  check(3,"Dot",-1,-1,dot,-1);				\
  check(3,"ExtGcd",-1,-1,gcd,-1);			\
  check(3,"Gcd",-1,-1,gcd,-1);				\
  check(3,"Eea",-1,-1,gcd,-1);				\
  check(3,"Has",-1,-1,-1,-1);				\
  check(3,"Lgt",-1,-1,length,-1);			\
  check(3,"Make",-1,-1,-1,-1);				\
  check(3,"Update",-1,-1,-1,-1);			\
  check(3,"Resultant",-1,-1,resultant,-1);		\
  check(3,"Eval",-1,-1,eval,-1);			\
  check(3,"Size",-1,-1,length,-1);			\
  check(3,"From",-1,-1,convert,-1);			\
  check(3,"To",-1,-1,convert,-1);			\
  check(3,"Full",-1,-1,-1,-1);				\
  check(3,"IsNot",-1,-1,test,-1);			\
  check(3,"IsA",-1,-1,test,-1);				\
  check(3,"Is",-1,-1,test,-1);				\
  check(3,"2",-1,-1,convert,-1);			\
							\
  check(4,"Camion",-1,-1,factor,-1);			\
  check(4,"ThiongLy",-1,-1,factor,-1);			\
  check(4,"MillerRabin",-1,prime,test,-1);		\
  check(4,"Prime",-1,prime,-1,-1);			\
  check(4,"Monic",-1,monic,-1,-1);			\
  check(4,"Deg",-1,-1,deg,-1);				\
  check(4,"Val",-1,-1,val,-1);				\
  check(4,"Close",-1,-1,freeing,-1);			\
  check(4,"Shift",-1,-1,shift,-1);			\
  check(4,"Coeff",-1,coeff,-1,-1);			\
  check(4,"DRoot",-1,-1,root,-1);			\
  check(4,"Root",-1,-1,root,-1);			\
  check(4,"Digit",-1,digit,-1,-1);			\
  check(4,"Good",-1,-1,-1,-1);				\
  check(4,"Random",-1,-1,-1,-1);			\
  check(4,"Smallest",-1,-1,-1,-1);			\
  check(4,"ED",-1,-1,-1,-1);				\
  check(4,"DD",-1,-1,-1,-1);				\
  check(4,"SF",-1,-1,-1,-1);				\
							\
  check(5,"Fact",-1,-1,factor,-1);			\
  check(5,"GCD",-1,-1,gcd,-1);				\
  check(5,"Irreducible",-1,irreduc,-1,-1);		\
							\
  check(1,"Berlekamp",polynom,irreduc,-1,-1);		\
  check(1,"C",-1,-1,create,-1);				\
  check(1,"F",-1,-1,freeing,-1);			\
} while(0)

#include "zparse.h"

int main(argc,argv)
     int argc;			
     char **argv;		
{
  exit(zparse(argc, argv));
}
