/*
 * zenfact/zmake.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "zmake.h"

/* Change this if the library zen is not in this directory */
#define ZENDIR "../zen"

/* Local files */
char *C_FILES = "psprime irreduc berlekamp roots factors random ring";

char *H_FILES = "zenfact";

char *DOC_ROOT = "doc";
char *DOC_FILES = "intro biblio title";

char *LIBLOCAL = "libzenfact";
char *LIBZEN = "libzen";

char *PRGM0 = "test";
char *PRGM1 = "bench";
char *PRGM2 = "example";

char *PARSER = "zparse";

char *INC = "include";
char *SRC = "c";
char *IPT = "inputs";

char *OBT = "obj";
char *BNT = "bin";
char *LBT = "lib";

char *OBJ_TOAR = "fact.dar";
char *AR_TOLIB = "fact.lar";

char *DOC = "doc";
char *TEX = "tex";
char *DVI = "dvi";

int main(argc, argv)
     int argc;
     char *argv[];
{
  /*
   * ZEN definitions
   ******************/
  char *ZEN_TOP = concat(TOPDIR, concat("/", ZENDIR));

  char *ZEN_INC = concat(ZEN_TOP, concat("/", INC));

  char *ZEN_LBT = concat(ZEN_TOP, concat("/", LBT));
  char *ZEN_LIB = concat(ZEN_LBT, concat("/", CompilingDir));
  char *ZEN_LIBRARY;
  char *ZEN_DEFINES;

  /*
   * Local definitions
   ********************/
  char *LOCAL_TOP = TOPDIR;

  char *LOCAL_SRC = concat(LOCAL_TOP, concat("/", SRC));
  char *LOCAL_INC = concat(LOCAL_TOP, concat("/", INC));

  char *LOCAL_OBT = concat(LOCAL_TOP, concat("/", OBT));
  char *LOCAL_OBJ = concat(LOCAL_OBT, concat("/", CompilingDir));

  char *LOCAL_LBT = concat(LOCAL_TOP, concat("/", LBT));
  char *LOCAL_LIB = concat(LOCAL_LBT, concat("/", CompilingDir));

  char *LOCAL_BNT = concat(LOCAL_TOP, concat("/", BNT));
  char *LOCAL_BIN = concat(LOCAL_BNT, concat("/", CompilingDir));
  
  char *LOCAL_DOC = concat(LOCAL_TOP, concat("/", DOC));
  char *LOCAL_TEX = concat(LOCAL_DOC, concat("/", TEX));
  char *LOCAL_DVI = concat(LOCAL_DOC, concat("/", DVI));

  char *LOCAL_IPT = concat(LOCAL_TOP, concat("/", IPT));
  
  char *LOCAL_INCLUDES, *LOCAL_INPUTS, *LOCAL_LIBRARIES;
  char *INCLUDED_DIRS;

 /* 
  * Internal targets 
  ********************/
  int cc = 0;
  int prgm0link = 0; int prgm1link = 0; int prgm2link = 0;
  int ludvi = 0;
  int clean_obj = 0; int clean_lib = 0; int clean_bin = 0;
  int clean_doc = 0; int clean_tex = 0; int clean_dvi = 0;
  int tidy_lib = 0; int tidy_obj = 0; int tidy_bin = 0;
  int tidy_doc = 0;
  
  /* 
   * External targets
   *******************/
  int all = Target("all", argc, argv);

  int ar = Target("ar", argc, argv);
  int prgm0 = Target("test", argc, argv);
  int prgm1 = Target("bench", argc, argv);
  int prgm2 = Target("example", argc, argv);

  int doc = Target("doc", argc, argv);
  int tex = Target("tex", argc, argv);
  int dvi = Target("dvi", argc, argv);

  int archtidy = Target("archtidy", argc, argv);
  int archclean = Target("archclean", argc, argv);
  int clean = Target("clean", argc, argv);
  int tidy = Target("tidy", argc, argv);

  int none = Target("none", argc, argv);

  /* 
   * Initialization
   *****************/
  LOCAL_INPUTS = concat(concat(CURDIR, ":"), LOCAL_IPT);
  LOCAL_INPUTS = concat(concat(LOCAL_INPUTS, ":"), LOCAL_TEX);

  LOCAL_INCLUDES = concat(" -I", ZEN_INC);
  LOCAL_INCLUDES = concat(concat(LOCAL_INCLUDES, " -I"), LOCAL_INC);

  INCLUDED_DIRS = concat("", ZEN_INC);
  append(&INCLUDED_DIRS, concat(" ", LOCAL_INC));

 /* Defines
  ***********/
#if OptimizingCode
  ZEN_DEFINES = " -DZEN_NO_TEST";
#else
  ZEN_DEFINES = "";
#endif
#if DebuggingMalloc
 ZEN_DEFINES = concat(ZEN_DEFINES," -DZEN_MALLOC_TEST");
#endif

  zmake_init(ZEN_DEFINES, LOCAL_INCLUDES, "", MathLibrary, argc, argv);

  LOCAL_LIBRARIES = concat(LOCAL_LIB, concat("/", concat(PREFIX, concat(LIBLOCAL, ASUF))));
  append(&LOCAL_LIBRARIES, " ");
  append(&LOCAL_LIBRARIES, concat(ZEN_LIB, concat("/", concat(PREFIX, concat(LIBZEN, ASUF)))));

  /* 
   * Targets
   ***********/

  /* default */
  TargetImply(defaults, all);

  /* all */
  TargetImply(all, prgm0=prgm1=prgm2);
  TargetImply(prgm0, ar=prgm0link);
  TargetImply(prgm1, ar=prgm1link);
  TargetImply(prgm2, ar=prgm2link);
  TargetImply(ar, cc);

  /* cc */
  SrcToObjects(cc,
	       LOCAL_OBJ, PREFIX,
	       LOCAL_SRC, "", C_FILES,
	       INCLUDED_DIRS, "");

  /* ar */
  ObjectsToArList(ar,
		  LOCAL_LIB, PREFIX, "", OBJ_TOAR,
		  LOCAL_OBJ, PREFIX, C_FILES);
  ArListToLibrary(ar,
		  LOCAL_LIB, PREFIX, LIBLOCAL, ASUF,
		  LOCAL_LIB, PREFIX, AR_TOLIB, "",
		  LOCAL_LIB, PREFIX, OBJ_TOAR, "");

  /* prgm0 */
  SrcToObjects(prgm0link,
	       LOCAL_OBJ, PREFIX,
	       LOCAL_SRC, "", PRGM0,
	       INCLUDED_DIRS, "");
  ObjectsToBinary(prgm0link,
		  LOCAL_BIN, PREFIX, PRGM0,
		  LOCAL_OBJ, PREFIX, PRGM0,
		  "", LOCAL_LIBRARIES);

  /* prgm1 */
  SrcToObjects(prgm1link,
	       LOCAL_OBJ, PREFIX,
	       LOCAL_SRC, "", PRGM1,
	       INCLUDED_DIRS, "");
  ObjectsToBinary(prgm1link,
		  LOCAL_BIN, PREFIX, PRGM1,
		  LOCAL_OBJ, PREFIX, PRGM1,
		  "", LOCAL_LIBRARIES);

  /* prgm2 */
  SrcToObjects(prgm2link,
	       LOCAL_OBJ, PREFIX,
	       LOCAL_SRC, "", PRGM2,
	       INCLUDED_DIRS, "");
  ObjectsToBinary(prgm2link,
		  LOCAL_BIN, PREFIX, PRGM2,
		  LOCAL_OBJ, PREFIX, PRGM2,
		  "", LOCAL_LIBRARIES);

  /* doc */
  TargetImply(doc, tex=ludvi);
  TargetImply(dvi, tex=ludvi);

  /* tex */
  SrcToTex(tex,
	   LOCAL_TEX, "",
	   LOCAL_SRC, "", C_FILES, CSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);
  SrcToTex(tex,
	   LOCAL_TEX, "",
	   LOCAL_SRC, "", PRGM0, CSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);
  SrcToTex(tex,
	   LOCAL_TEX, "",
	   LOCAL_SRC, "", PRGM1, CSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);
  SrcToTex(tex,
	   LOCAL_TEX, "",
	   LOCAL_SRC, "", PRGM2, CSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);
  SrcToTex(tex,
	   LOCAL_TEX, "",
	   LOCAL_INC, "", H_FILES, HSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);
  SrcToTex(tex,
	   LOCAL_TEX, "",
	   LOCAL_IPT, "", DOC_FILES, CSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);
  SrcToTex(tex,
	   LOCAL_TEX, "",
	   LOCAL_IPT, "", DOC_ROOT, CSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);

  /* dvi */
  TexToDvi(dvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", C_FILES,
	   LOCAL_INPUTS);
  TexToDvi(dvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", PRGM0,
	   LOCAL_INPUTS);
  TexToDvi(dvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", PRGM1,
	   LOCAL_INPUTS);
  TexToDvi(dvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", PRGM2,
	   LOCAL_INPUTS);
  TexToDvi(dvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", H_FILES,
	   LOCAL_INPUTS);
  TexToDvi(dvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", DOC_FILES,
	   LOCAL_INPUTS);
  TexToDvi(ludvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", DOC_ROOT,
	   LOCAL_INPUTS);

  /* Cleaning */
  TargetImply(tidy, tidy_obj=tidy_lib=tidy_bin=tidy_doc);
  TargetImply(archtidy, clean_obj=clean_lib=clean_doc=clean_bin);
  TargetImply(clean, tidy_obj=clean_tex);
  TargetImply(archclean, clean_obj=clean_tex);
  TargetImply(clean_doc, clean_dvi=clean_tex);
  
  CleaningRules(clean_obj, LOCAL_OBJ, PREFIX, C_FILES, OSUF);
  CleaningRules(clean_obj, LOCAL_OBJ, PREFIX, PRGM0, OSUF);
  CleaningRules(clean_obj, LOCAL_OBJ, PREFIX, PRGM1, OSUF);
  CleaningRules(clean_obj, LOCAL_OBJ, PREFIX, PRGM2, OSUF);

  CleaningRules(clean_obj, LOCAL_LIB, "", OBJ_TOAR, "");
  CleaningRules(clean_obj, LOCAL_LIB, "", AR_TOLIB, "");

  CleaningRules(clean_lib, LOCAL_LIB, PREFIX, LIBLOCAL, ASUF);

  CleaningRules(clean_bin, LOCAL_BIN, PREFIX, PRGM0, ESUF);
  CleaningRules(clean_bin, LOCAL_BIN, PREFIX, PRGM1, ESUF);
  CleaningRules(clean_bin, LOCAL_BIN, PREFIX, PRGM2, ESUF);

  CleaningRules(clean_tex, LOCAL_TEX, "", C_FILES, ".tex");
  CleaningRules(clean_tex, LOCAL_TEX, "", H_FILES, ".tex");
  CleaningRules(clean_tex, LOCAL_TEX, "", DOC_FILES, ".tex");
  CleaningRules(clean_tex, LOCAL_TEX, "", DOC_ROOT, ".tex");
  CleaningRules(clean_tex, LOCAL_TEX, "", PRGM0, ".tex");
  CleaningRules(clean_tex, LOCAL_TEX, "", PRGM1, ".tex");
  CleaningRules(clean_tex, LOCAL_TEX, "", PRGM2, ".tex");

  CleaningRules(clean_dvi, LOCAL_DVI, "", C_FILES, ".dvi");
  CleaningRules(clean_dvi, LOCAL_DVI, "", H_FILES, ".dvi");
  CleaningRules(clean_dvi, LOCAL_DVI, "", DOC_FILES, ".dvi");
  CleaningRules(clean_dvi, LOCAL_DVI, "", DOC_ROOT, ".dvi");
  CleaningRules(clean_dvi, LOCAL_DVI, "", PRGM0, ".dvi");
  CleaningRules(clean_dvi, LOCAL_DVI, "", PRGM1, ".dvi");
  CleaningRules(clean_dvi, LOCAL_DVI, "", PRGM2, ".dvi");

  CleaningRules(clean_tex|clean_dvi, LOCAL_DOC, "", "", "");

  TidyingRules(tidy_obj, LOCAL_OBT);
  TidyingRules(tidy_lib, LOCAL_LBT);
  TidyingRules(tidy_bin, LOCAL_BNT);
  TidyingRules(tidy_doc, LOCAL_DOC);

  /* 
   * Closing
   **********/
  zmake_close();

  exit(0);
}
