/*
 * zmake/defines.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef Verbose
# define Verbose 50
#endif

/* the following are needed before we can include zrules.h */
#ifndef SystemV
# define SystemV		NO	/* SYSV (R3) */
#endif
#ifndef SystemV4
# define SystemV4		NO	/* SVR4 */
#endif
#ifndef HasPurify
# define HasPurify		NO
#endif
#ifndef DoRanlibCmd
# if SystemV || SystemV4
#  define DoRanlibCmd NO
# else
#  define DoRanlibCmd YES
# endif
#endif

#ifndef OSMajorVersion
#define OSMajorVersion          0
#endif
#ifndef OSMinorVersion
#define OSMinorVersion          0
#endif
#ifndef HasLargeTmp
#define HasLargeTmp		NO	/* be paranoid */
#endif
#ifndef HasSockets
# define HasSockets		NO
#endif
#ifndef HasVFork
# if SystemV
#  define HasVFork		NO	/* not yet... */
# else
#  define HasVFork		YES
# endif
#endif
#ifndef HasPutenv
# define HasPutenv		YES	/* assume yes */
#endif
#ifndef HasBsearch
# define HasBsearch		YES	/* assume yes */
#endif
#ifndef HasLatex
# define HasLatex		YES
#endif
#ifndef StripInstalledPrograms
# define StripInstalledPrograms NO	/* leave symbol table just in case */
#endif
#ifndef HasVoidSignalReturn
# define HasVoidSignalReturn YES
#endif
#ifndef CIsAnsi
# define CIsAnsi YES
#endif

#ifndef DefaultUsrBin
# define DefaultUsrBin "/usr/bin"
#endif

#ifndef UsrLibDir
# define UsrLibDir "/usr/lib"
#endif

#ifndef ShLibDir
# define ShLibDir UsrLibDir
#endif

#ifndef IncRoot
# define IncRoot "/usr/include"
#endif

#ifndef ManDir
# if SystemV4
#  define ManDir "/usr/share/man"
# else
#  define ManDir "/usr/man"
# endif
#endif

#ifndef ProtoDefines
# if CIsAnsi
#  define ProtoDefines ""
# else
#  define ProtoDefines " -DNOCANSI"
# endif
#endif
#ifndef ExtraDefines
# if UseLongLong
#  if UseAssembler
#   define ExtraDefines " -DLONG_LONG"
#  else
#   define ExtraDefines " -DLONG_LONG -DNO_ASM"
#  endif
# else
#  if UseAssembler
#   define ExtraDefines ""
#  else
#   define ExtraDefines " -DNO_ASM"
#  endif
# endif
#endif
#ifndef ThreadsDefines
# define ThreadsDefines ""
#endif

#ifndef ThreadsCFlags
# define ThreadsCFlags ""
#endif
#if !HasOMP
# ifdef OMPCFlags
#  undef OMPCFlags
# endif
# ifdef OMPLoadOptions
#  undef OMPLoadOptions
# endif
#endif
#ifndef OMPCFlags
# define OMPCFlags ""
#endif
#ifndef OMPLoadOptions
# define OMPLoadOptions ""
#endif
#ifndef OptimizedCDebugFlags
# define OptimizedCDebugFlags " -O" /* this option is the only standard one for all cc */
# endif
#ifndef OptimizedCplusplusDebugFlags
# define OptimizedCplusplusDebugFlags OptimizedCDebugFlags
#endif
#ifndef DebuggableCDebugFlags
# define DebuggableCDebugFlags " -g"  /*-Wall -W -g -Wcast-align  -Wconversion  -Wstrict-prototypes -Wmissing-prototypes -Wmissing-declarations -Wnested-externs -Winline */
#endif
#ifndef DebuggableCplusplusDebugFlags
# define DebuggableCplusplusDebugFlags DebuggableCDebugFlags
#endif

#ifndef ProfiledCDebugFlags
# define ProfiledCDebugFlags " -pg"
#endif

#ifndef ProfiledOptimizedCDebugFlags
# define ProfiledOptimizedCDebugFlags OptimizedCDebugFlags
#endif

#ifndef MallocCDebugFlags
# define MallocCDebugFlags ""
#endif
/* FIXME: we must find a mean to see if bounds-checking is enabled or not
#ifndef MallocCDebugFlags
# if HasGcc
#  define MallocCDebugFlags " -fbounds-checking -static"
# else
#  define MallocCDebugFlags ""
# endif
#endif
*/
#ifndef ProfiledCplusplusDebugFlags
# define ProfiledCplusplusDebugFlags ProfiledCDebugFlags
#endif

#ifndef NoOpCDebugFlags
# define NoOpCDebugFlags ""
#endif

#ifndef DefaultCDebugFlags
# if  OptimizingCode
#  if  ProfilingCode
#   define DefaultCDebugFlags ProfiledOptimizedCDebugFlags
#  else
#   define DefaultCDebugFlags OptimizedCDebugFlags
#  endif
# else
#  define DefaultCDebugFlags DebuggableCDebugFlags
# endif
#endif

#ifndef DefaultCCOptions
# define DefaultCCOptions ""
#endif
#ifndef DefaultCplusplusOptions
# define DefaultCplusplusOptions ""
#endif

#ifndef InstPgmFlags
# if StripInstalledPrograms
#  define InstPgmFlags " -s"
# else
#  define InstPgmFlags ""
# endif
#endif
#ifndef InstBinFlags
# define InstBinFlags " -m 0755"
#endif
#ifndef InstUidFlags
# define InstUidFlags " -m 4755"
#endif
#ifndef InstLibFlags
# define InstLibFlags " -m 0644"
#endif
#ifndef InstIncFlags
# define InstIncFlags " -m 0444"
#endif
#ifndef InstManFlags
# define InstManFlags " -m 0444"
#endif
#ifndef InstDatFlags
# define InstDatFlags " -m 0444"
#endif

#ifndef ArCmdBase
# define ArCmdBase "ar"
#endif
#ifndef ArCreateOption
# if HasLargeTmp || SystemV4
#  define ArCreateOption " cq"
# else
#  define ArCreateOption " clq"
# endif
#endif
#ifndef ArAddOption
# if HasLargeTmp || SystemV4
#  define ArAddOption " ru"
# else
#  define ArAddOption " rul"
# endif
#endif
#ifndef ArRmOption
# define ArRmOption " d"
#endif
#ifndef ArExtractOption
# if HasLargeTmp || SystemV4
#  define ArExtractOption " x"
# else
#  define ArExtractOption " xl"
# endif
#endif

#ifndef HasGcc
# define HasGcc NO
#endif
#ifndef CcCmd
# if HasGcc
#  define CcCmd "gcc"
# else
#  define CcCmd "cc"
# endif
#endif

#ifndef HasCplusplus
# define HasCplusplus NO
#endif
#ifndef CplusplusCmd

# if HasGccForCplusplus
#  define CplusplusCmd "g++"
# else
#  define CplusplusCmd "CC"
# endif

# ifndef CplusplusStandardDefines
#  define CplusplusStandardDefines StandardDefines
# endif
# ifndef CplusplusThreadsDefines
#  define CplusplusThreadsDefines ""
# endif
# ifndef CplusplusExtraDefines
#  define CplusplusExtraDefines ""
# endif

# ifndef CplusplusExtraIncludes
#  define CplusplusExtraIncludes ""
# endif

# ifndef CplusplusOptions
#  define CplusplusOptions ""
# endif
# ifndef DefaultCplusplusDebugFlags
#  define DefaultCplusplusDebugFlags OptimizedCplusplusDebugFlags
# endif
# ifndef CplusplusThreadFlags
#  define CplusplusThreadFlags ""
# endif

#endif

#ifndef AsCmd
# define AsCmd "as"
#endif

#ifndef CppCmd
# define CppCmd "/lib/cpp"
#endif

#ifndef PreProcessCmd
# define PreProcessCmd concat(CcCmd, " -E")
#endif

#ifndef InstallCmd
# if SystemV || SystemV4
#  define InstallCmd "bsdinst"
# else
#  define InstallCmd "install"
# endif
#endif

#ifndef InstallFlags
# define InstallFlags " -c"
#endif

#ifndef LdCmd
# define LdCmd "ld"
#endif
#ifndef ThreadsLdFlags
# define ThreadsLdFlags ""
#endif
#ifndef ThreadsLibs
# define ThreadsLibs ""
#endif

#ifndef AwkCmd
# define AwkCmd "awk"
#endif

#ifndef LatexCmd
# define LatexCmd "latex"
#endif

#ifndef MakeIndexCmd
# define MakeIndexCmd "makeindex"
#endif

#ifndef CtangleCmd
# define CtangleCmd "ctangle"
#endif

#ifndef HycweaveCmd
# define HycweaveCmd "hycweave"
#endif

#ifndef LexCmd
# define LexCmd "lex"
#endif

#ifndef LexLib
# define LexLib " -ll"
#endif

#ifndef YaccCmd
# define YaccCmd "yacc"
#endif
#ifndef CplusplusYaccCmd
#define CplusplusYaccCmd YaccCmd
#endif

#ifndef ZMakeCmd
# define ZMakeCmd "zmake"
#endif

#ifndef LintCmd
#define LintCmd "lint"
#endif
#ifndef LintLibFlag
# if SystemV || SystemV4
#  define LintLibFlag " -o"
# else
#  define LintLibFlag " -C"
# endif
#endif
#ifndef LintOpts
# if SystemV || SystemV4
#  define LintOpts " -bh"
# else
#  define LintOpts " -axz"
# endif
#endif

#ifndef RanlibCmd
# define RanlibCmd "ranlib"
#endif
#ifndef RanlibInstFlags
# define RanlibInstFlags ""
#endif

#ifndef TroffCmd
#define TroffCmd "psroff"
#endif
#ifndef MsMacros
#define MsMacros " -ms"
#endif

#ifndef StandardIncludes
# define StandardIncludes ""
#endif
#ifndef ExtraIncludes
# define ExtraIncludes ""
#endif
#ifndef StandardDefines
# if SystemV
#  define StandardDefines "-DSYSV"
# else
#  if SystemV4
#   define StandardDefines "-DSVR4"
#  else
#   define StandardDefines ""
#  endif
# endif
#endif
#ifndef StandardCppDefines
# define StandardCppDefines StandardDefines
#endif

#ifndef LdPreLib
# define LdPreLib ""
#endif
#ifndef LdPostLib
#define LdPostLib ""
#endif
#ifndef MathLibrary
# define MathLibrary " -lm"
#endif
#ifndef ExtraLibraries
# if SystemV4
#  if HasSockets
#   define ExtraLibraries " -lsocket -lnsl -lw"
#  else
#   define ExtraLibraries " -lnsl -lw"
#  endif
# else
#  define ExtraLibraries ""
# endif
#endif
#ifndef SysLibraries
# define SysLibraries ""
#endif

#ifndef ExtraLoadOptions
# define ExtraLoadOptions ""
#endif
#ifndef ExtraLoadFlags
# define ExtraLoadFlags ""
#endif

#ifndef LdCombineFlags
# if SystemV4
#  define LdCombineFlags " -r"
# else
#  define LdCombineFlags " -X -r"
# endif
#endif
#ifndef LdStripFlags
# define LdStripFlags " -x"
#endif

#if HasPurify
# ifndef PurifyCmd
#  define PurifyCmd "purify -always-use-cache-dir"
# endif
# ifndef PurifyOptions
#  define PurifyOptions ""
# endif
#endif

#ifndef PathSeparator
# define PathSeparator "/"
#endif
#ifndef Osuf
# define Osuf ".o"
#endif
#ifndef Asuf
# define Asuf ".a"
#endif
#ifndef Esuf
# define Esuf ""
#endif
#ifndef Lsuf
# define Lsuf ".l"
#endif
#ifndef Ysuf
# define Ysuf ".y"
#endif
#ifndef Csuf
# define Csuf ".c"
#endif
#ifndef Hsuf
# define Hsuf ".h"
#endif
#ifndef TEXsuf
# define TEXsuf ".tex"
#endif
#ifndef DVIsuf
# define DVIsuf ".dvi"
#endif
#ifndef Nsuf
# define Bsuf ".bak"
#endif

#ifndef CCsuf
# define CCsuf "cxx"
#endif

#ifndef ManSuffix
# define ManSuffix	"n"
#endif
#ifndef LibManSuffix
# define LibManSuffix	"3"
#endif
#ifndef FileManSuffix
# define FileManSuffix	"4"
#endif

#ifndef ExpandManNames
# if SystemV
#  define ExpandManNames NO
# else
#  define ExpandManNames YES
# endif
#endif

#ifndef CURDIR
# define CURDIR zpwd()
#endif
#ifndef TOPDIR
# define TOPDIR CURDIR
#endif
#ifndef RELATIVE_ZCFDIR
# define RELATIVE_ZCFDIR  "/../zmake"
#endif
#ifndef ZCFDIR
# define ZCFDIR concat(TOPDIR, RELATIVE_ZCFDIR)
#endif

#ifndef CompilingDir
# define CompilingDir "unknown"
#endif

#ifndef LocalRulesFile
#define LocalRulesFile "noop.h"
#endif
