/*-*-LaTeX-*-
 * zmake/doc/intro.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
  This is a documentation file of the ZMAKE library.
  There is no C code in this file, only LaTeX commands.

\TEX{

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Introduction %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  
  In~\cite{imake}, it is stated that
  \begin{quote}
    ``In the UNIX world, software development typically involves
    |make|, using a |Makefile| describing how to build and install the
    programs in which we're interested. In essence, |make| functions
    as a command-generating engine, and the |Makefile| controls which
    command to generate for specific targets. This is a tremendous
    convenience for the programmer, since it eliminates the need to
    type out a bunch of commands each time you want to build
    something.
  \end{quote}

  \begin{quote}
    Unfortunately, |Makefiles| are not portable. The commands that
    make so conveniently generates at the drop of a hat are subject to
    variation from system to system, so a |Makefile| that works on my
    system may not work on yours. Ideally, we'd be able to take a
    software project, put it on another machine, compile and install
    it, and it would run. But it doesn't always work that way.''
  \end{quote}

  This point of view is developed in details in
  section~\ref{makefiles}.

  Two alternatives are known to avoid such troubles, the first one was
  developed for X-WINDOWS, it is called |imake|. The second one was
  developed by the GNU community, it is called |configure|.
  \begin{description}
  \item[Imake~:] With this solution, you don't write a |Makefile|,
    instead you write an |imakefile|, a machine independent
    description of the targets you want to build. Static machine
    dependencies are centralized into a set of configuration files
    isolated in their own directory, |config|, so they don't appear in
    the |imakefile|. The program |imake| merges information obtained
    from |config| and |imakefile| to generates a properly configured
    |Makefile|. Then you run |make| to build your programs.\\ 
                                %
    From a practical point of few, |imake| is just an interface for
    the preprocessor |cpp|. In a first approximation, the |Makefile|
    that |imake| generates, would be more or less the file obtained
    with |cpp -I. -Iconfig config/Imake.tmpl Makefile|. |cpp| usually
    defines specific variables for each machine and these variables
    are tested in |Imake.templ| to include the corresponding files of
    |config|. Then |Imake.templ| includes the current |imakefile|.
                                %
  \item[Configure~:] This solution generates |Makefiles| from a
    machine independent description of the targets you want to build
    and from machine dependencies dynamically collected at the
    compilation stage. \\ 
                                %
    From a practical point of few, |configure| is just a Bourne
    Script Shell which performs some tests in order to determine the
    way how the software must be compiled.
  \end{description}

  In |ZMAKE|, we tried to get the best of these two approaches. In
  fact, the strong assumptions we made in our design is that the only
  thing we can assume about the target system is that it has a
  C-compiler and that it supports POSIX system calls. Most of the OS
  we heard about verify this~! So, external tools like |imake|, |sh|,
  |make|, \ldots are no more needed.

  The way how a software must be compiled is no more specified in a a
  |Makefile|, an |Imakefile| or a |configure| script shells. It is
  simply specified in files written in C. These files are typically
  called |zmake.c|. Thus, the challenge that we had to face while
  developing |ZMAKE| was to give developers APIs which can be easily
  used. Like this, we expect to have |zmake.c| files easy to
  understand.  That's why these APIS are very similar to Makefile
  targets. These APIS are described in section~\ref{apis}.
  
  In order to have a documentation as close as possible to the C
  implementation, we used to interlace documentation and C code. This
  is done thanks to a specific parser called |zparse|, This parser is
  described in section~\ref{zparse}.

  Finally, a small example, the C file |zmake.c| which compiles this
  documentation is given in section~\ref{zmake}.

} */

