/*
 * zmake/doc/zmake.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \CiteC */ 
#include "zmake.h"
/* \EndCiteC */

/* \TEX{ At first, global variables containing main paths and files
   are defined.
   }
*/

/* \CiteC */ 
char *DOC_ROOT = "doc";
char *DOC_FILES = "intro title compile biblio zmake";
char *ZMAKE_DOC_FILES = "ztools zrules zmake";
char *ZMAKE_DOC_HFILES = "zmake";
char *PARSER = "zparse";
char *IPT = ".";
char *TEX = "tex";
char *DVI = "dvi";
/* \EndCiteC */

/* \TEX{
   The heart of a |zmake.c| is this procedure.
   }
*/
/* \CiteC */ 
int main(argc, argv)
     int argc;
     char *argv[];
{
/* \EndCiteC */

/* \TEX{
   We will need these variables.
   }
*/
/* \CiteC */ 
  char *LOCAL_TOP = TOPDIR;  
  char *LOCAL_BIN = concat(LOCAL_TOP, concat("/", CompilingDir));
  char *LOCAL_TEX = concat(LOCAL_TOP, concat("/", TEX));
  char *LOCAL_DVI = concat(LOCAL_TOP, concat("/", DVI));
  char *LOCAL_IPT = LOCAL_TOP;
  char *LOCAL_INPUTS;
/* \EndCiteC */

/* \TEX{
   Then, Makefiles target are simulated as follows.
   }
*/
/* \CiteC */ 
  int cc = 0;
  int prgm0link = 0; int prgm1link = 0; int prgm2link = 0;
  int ludvi = 0;
  int clean_obj = 0; int clean_lib = 0; int clean_bin = 0;
  int clean_doc = 0; int clean_tex = 0; int clean_dvi = 0;
  int tidy_lib = 0; int tidy_obj = 0; int tidy_bin = 0;
  int tidy_doc = 0;
/* \EndCiteC */

/* \TEX{ 
   Targets which can be called from the command line are defined here.
   }
*/
/* \CiteC */ 
  int all = Target("all", argc, argv);

  int ar = Target("ar", argc, argv);
  int prgm0 = Target("test", argc, argv);
  int prgm1 = Target("bench", argc, argv);
  int prgm2 = Target("example", argc, argv);

  int doc = Target("doc", argc, argv);
  int tex = Target("tex", argc, argv);
  int dvi = Target("dvi", argc, argv);

  int archtidy = Target("archtidy", argc, argv);
  int archclean = Target("archclean", argc, argv);
  int clean = Target("clean", argc, argv);
  int tidy = Target("tidy", argc, argv);

  int none = Target("none", argc, argv);

/* \EndCiteC
   \TEX{ 
   Afterwards, |ZMAKE| initialisation is done here.
   }
   \CiteC 
*/
  LOCAL_INPUTS = concat(concat(CURDIR, ":"), IPT);
  LOCAL_INPUTS = concat(concat(LOCAL_INPUTS, ":"), LOCAL_TEX);

  zmake_init("", "", "", "", argc, argv);

/* \EndCiteC
   \TEX{ 
   And pseudo makefiles targets can be handled as follows.
   }
   \CiteC 
*/
  TargetImply(defaults, all);

  TargetImply(all, doc);

  TargetImply(doc, tex=ludvi);
  TargetImply(dvi, tex=ludvi);

  SrcToTex(tex,
	   LOCAL_TEX, "",
	   LOCAL_IPT, "", DOC_FILES, CSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);
  SrcToTex(tex,
	   LOCAL_TEX, "",
	   ZMAKE_DIR, "", ZMAKE_DOC_FILES, CSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);
  SrcToTex(tex,
	   LOCAL_TEX, "h",
	   ZMAKE_DIR, "", ZMAKE_DOC_HFILES, HSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);
  SrcToTex(tex,
	   LOCAL_TEX, "",
	   ZMAKE_DIR, "", PARSER, HSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);
  SrcToTex(tex,
	   LOCAL_TEX, "",
	   LOCAL_IPT, "", DOC_ROOT, CSUF,
	   LOCAL_BIN, "", LOCAL_IPT, PARSER);

  TexToDvi(dvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", ZMAKE_DOC_FILES,
	   LOCAL_INPUTS);
  TexToDvi(dvi,
	   LOCAL_DVI, "h",
	   LOCAL_TEX, "h", ZMAKE_DOC_HFILES,
	   LOCAL_INPUTS);
  TexToDvi(dvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", DOC_FILES,
	   LOCAL_INPUTS);
  TexToDvi(dvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", PARSER,
	   LOCAL_INPUTS);
  TexToDvi(ludvi,
	   LOCAL_DVI, "",
	   LOCAL_TEX, "", DOC_ROOT,
	   LOCAL_INPUTS);

/* \EndCiteC
   \TEX{ 
   Cleaning rules are defined here.
   }
   \CiteC 
*/
  TargetImply(tidy, tidy_bin=tidy_doc);
  TargetImply(archtidy, clean_doc=clean_bin);
  TargetImply(clean, clean_tex);
  TargetImply(archclean, clean_tex);
  TargetImply(clean_doc, clean_dvi=clean_tex);
  
  CleaningRules(clean_tex, LOCAL_TEX, "", PARSER, ".tex");
  CleaningRules(clean_tex, LOCAL_TEX, "", ZMAKE_DOC_FILES, ".tex");
  CleaningRules(clean_tex, LOCAL_TEX, "h", ZMAKE_DOC_HFILES, ".tex");
  CleaningRules(clean_tex, LOCAL_TEX, "", DOC_FILES, ".tex");
  CleaningRules(clean_tex, LOCAL_TEX, "", DOC_ROOT, ".tex");

  CleaningRules(clean_dvi, LOCAL_DVI, "", ZMAKE_DOC_FILES, ".dvi");
  CleaningRules(clean_dvi, LOCAL_DVI, "h", ZMAKE_DOC_FILES, ".dvi");
  CleaningRules(clean_dvi, LOCAL_DVI, "", DOC_FILES, ".dvi");
  CleaningRules(clean_dvi, LOCAL_DVI, "", PARSER, ".dvi");
  CleaningRules(clean_dvi, LOCAL_DVI, "", DOC_ROOT, ".dvi");

  TidyingRules(tidy_bin, LOCAL_BIN);
  TidyingRules(tidy_doc, LOCAL_TEX);
  TidyingRules(tidy_doc, LOCAL_DVI);

/* \EndCiteC
   \TEX{ 
   Exit is finally done here.
   }
   \CiteC 
*/
  zmake_close();

  exit(0);
}
/* \EndCiteC */
