/*
 * zmake/zmake.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
 * Generic zmake template (a c imake-like template)
 */

/*
 * Modify zconfig.h when adding a new architecture, do not modify this file.
 * Platform-specific parameters may be set in the appropriate cf/vendor.h
 * configuration files. Site-specific parameters should be set in the file
 * site.h. Full rebuilds are recommended if any parameters are changed.
 */

#define ZMakeMajorRelease 1
#define ZMakeMinorRelease 0

#define YES     1
#define NO      0

#include "zconfig.h"

/*
 * Platform-specific configuration parameters - edit MacroFile to change
 */
#include MacroIncludeFile

/*
 * Site-specific configuration parameters.
 * Edit site.h to change.
 */
#include "site.h"

/*
 * Defaults for various generic parameters; set in site.h if needed
 */
#include "defines.h"

/*  
 * Makefile-like c-apis
 */
#include "ztools.c"

/*
 * Global variables.
 */
#undef _I
#define _I(x, y) char *x
#include "variable.h"

int VERBOSE;
int defaults;

/*
 * Local zmake rules.
 */
#include LocalRulesFile

/* 
 * Default zmake rules.
 */
#include "zrules.c"

/* \TEX{ 
   \Section{Initialization}
   } */

/* \TEX{ 
   
   \title{Initialization of a zmake.}

   \input{Standard defines |DEFINES|, standard includes |INCLUDES|,
   standard link flags |LOCAL_LDFLAGS| and standard libraries
   |LOCAL_LIBRARY| to use for any compilation and the well known
   |argc| and |argv| C arguments.}

   \side{If a statement |verbose=|100 is set on the command line,
   debugging output is done while procesing |ZMAKE|
   functions. Moreover, this function checks that any needed file for
   the compilation of |zmake| is older than |zmake|, otherwise
   recompilation of |zmake| is performed.}
   
   } */
void zmake_init(DEFINES, INCLUDES, LOCAL_LDFLAGS, LOCAL_LIBRARY,
                argc, argv)
     char *DEFINES;
     char *INCLUDES;
     char *LOCAL_LDFLAGS;
     char *LOCAL_LIBRARY;
     int argc;
     char *argv[];
{
  int i;


  /* Verbose */
  VERBOSE = Verbose;
  ReadIntOption(&VERBOSE, "verbose", argc, argv);
  if (VERBOSE >= 100) {
    for(i = 0; i < argc; i++) {
      zfprintf(stderr, " %s", argv[i]);
    }
    zfprintf(stderr, "\n");
  }

  /* Default target */
  defaults = 1;
  for(i = 1 ; i < argc; i++) {
    char *ptr;
    
    for(ptr = argv[i]; *ptr && (*ptr != '='); ptr++) {;}
    if (!(*ptr)) {
      defaults = 0;
    }
  }

  /* Initializations */
#undef _I
#if CIsAnsi
# define _I(x, y)                                       \
do {                                                    \
  (x) = (y);                                            \
  ReadStringOption(&(x), #x, argc, argv);               \
  if (VERBOSE >= 100) {                                 \
    zfprintf(stderr, "  %s = \"%s\"\n", #x, (x));       \
  }                                                     \
} while(0)
#else
# define _I(x, y)                               \
do {                                            \
  (x) = (y);                                    \
} while(0)
#endif
#include "variable.h"


  /* Recompile zmake ? */
  ZMakeZMake(CURRENT_DIR, argc, argv, 0);

  /* Print infos */
  if (VERBOSE >= 10) {
    int i;

    zfprintf(stderr, "\n--------------");
    for(i = 0; i < strlen(CURRENT_DIR); i++) {zfprintf(stderr, "-");}
    zfprintf(stderr, "\n");
    zfprintf(stderr, "Compiling in %s:\n", CURRENT_DIR);
  }
}

/* \TEX{ 
   \title{Ending a zmake.}
   \side{An |exit(0)| is performed.}
} */

void zmake_close()
{
  if (VERBOSE >= 10) {
    zfprintf(stderr, "-------------\n\n");
  }
  exit(0);
}
