/*
 * zmake/zparse.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/*************************************************************/
/*\TEX{ 

  This program builds |LaTeX| documentation source from the C files
  comments.  A few commands are used to perform clean output using
  |zen.sty| file.  These commands are only valid inside a C-comment:
  \begin{itemize}
  \item |\TEX{ZeNTeX}| allows to include |LaTeX| code in the resulting
    file as is. It is also intended to enumerate the procedures of the
    C files.  To include a procedure in the |LaTeX| file, there are
    fields that should be used to describe the function, all but the
    first of which are optional.
    \begin{enumerate}
    \item |\title| should shortly describe the function. This field is
      needed but can be set to ---.
    \item |\input| should describe the parameters of the function.
    \item |\output| should describe the output of the function.
    \item |\side| should describe the side effects of the procedure.
    \item |\note| allows to precise some programming features.
    \end{enumerate}
    Besides, as soon as a |\title{}| appears in a |\TEX| comment, the
    C-header of the following function is included and index entries
    for |MakeIndex| program are generated.  This feature is disabled
    by the -noindex option in the command line.

  \item |\TEX[ZeNTeX}| behaves in the same way, except that short
  descriptions of the functions are included.
  \item |\CiteC| makes the C source to be included in |verbatim| mode in
    the |LaTeX| code.
  \item |\EndCiteC| ends the citation of C source.
  \item Pipes symbols $\mid$ may delimit the programs names and
    declarations. Within pipes, the underscore symbols |_| are
    automatically backslashed ({\sf $\backslash$\_}) the stars |*|
    transformed in {\sf\$$\backslash$star\$} and the backslash |\|
    transformed in {\sf\$$\backslash$backslash\$}.
  \end{itemize}

\newpage
  \begin{note}
    By default, the output format of |LaTeX| is set to |a4paper|. It is
    possible to use default legal format by modifying the following
    compilation directive in file |zmake/zparse.h|.
    } \CiteC */ 
#define A4PAPER 1 
/* \EndCiteC */
/* \TEX{
    \end{note}   
    } 
*/


FILE *IN,*OUT;

typedef struct {
  enum glo1 l1;
  enum glo2 l2;
  enum glo3 l3;
} conceptentry;

typedef struct {
  enum glo1 l1;
  enum ind2 l2;
} fctentry;


char *header_proc[]={
  "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n",
  NULL};

char *header[]={
  "\\headerZeN{",
  "\\multicolumn{2}{c}{",
  NULL};

char *bottom_proc[]={
  "}}",
  "\\makeprocedure",
  NULL};
char *bottom_silent_proc[]={
  "}}",
  "\\makesilentprocedure",
  NULL};

char *header_tex[]= {
  "\\ifx\\undefined\\ZeNBloc",
  "\\gdef\\@FirstZeNBloc{}",
  "\\newcommand{\\ZeNBloc}[1]{\\if!\\@FirstZeNBloc!{",
  "    \\gdef\\@FirstZeNBloc{done}",
#if A4PAPER
  "    \\documentclass[a4paper]{book}",
#else
  "    \\documentclass{book}",
#endif
  "    \\usepackage{pstricks}",
  "    \\usepackage{ifthen}",
  "    \\usepackage{zen}",
  NULL};

char *header_tex_index[]= {
  "    \\usepackage{makeidx}",
  NULL};

char *header_tex_french[]= {
  "    \\usepackage[francais]{babel}",
  "    \\usepackage[latin1]{inputenc}",
  "    \\def\\procedurename{Proc\\'edure}",
  "    \\def\\inputname{Entr\\'ee~}",
  "    \\def\\outputname{Sortie~}",
  "    \\def\\sidename{Action~}",
  "    \\def\\notename{Note~}",
  NULL};

char *end_header_tex1[]= {
  "    \\setcounter{secnumdepth}{6}\\setcounter{tocdepth}{6}",
  "    \\makeindex",
  "    \\makeglossary",
  "    \\begin{document}#1",
  NULL};
char *end_header_tex2[]= {
  "  \\end{document}}",
  "  \\else #1 \\fi",
  "  }",
  "\\fi",
  "\\ZeNBloc{",
  NULL};

char *bottom_tex[]= {
  "}",
  NULL};

char *Msgerrparse[]= {
  "Unended \\TEX block",
  "Two \\title for the same function",
  "Unended pipe",
  "Unended C citation",
  NULL};

#define UNENDED_TEX_BLOCK 1
#define TWO_TEX_BLOCKS 2
#define UNENDED_PIPE 3
#define UNENDED_CITE_C 4

#ifdef DEBUG_PARSER
# define zengetc(stream) _fgetc(stream)
#else
# define zengetc(stream) fgetc(stream)
#endif

int _fgetc(stream)
     FILE *stream;
{int c;

 c=fgetc(stream);
 fputc(c,stdout);
 fflush(stdout);
 return(c);
}
/*************************************************************/
int FindNextTex ();
void Output ();
int FindToken ();
/*************************************************************/
int NO_INDEX,FRENCH;

int zparse(argc,argv)
     int argc;                  
     char **argv;               
{char **msg,file[1024];
 int i,err,ln,FLAG;
 
 IN=stdin;
 NO_INDEX=FRENCH=0;

 for(i=1;i<argc;i++)
   if (strcmp(argv[i],"-noindex")==0)
     NO_INDEX=1;
   else if (strcmp(argv[i],"-french")==0)
     FRENCH=1;
   else
     strcpy(file,argv[i]);

 
 IN=stdin;
 OUT=stdout;

 
 Output(header_tex);
 if(!NO_INDEX)
   Output(header_tex_index);
 if(FRENCH)
   Output(header_tex_french);
 Output(end_header_tex1);
 fprintf(OUT,"\\input{%s.ing}\n",file);
 fprintf(OUT,"\\input{%s.ind}\n",file);
 Output(end_header_tex2);

 ln=1;

 FLAG=0;
 while(err=FindNextTex(&ln))
   {
     fprintf(OUT,"\n");
     fprintf(stderr,"#### Something wrong in input around line %d\n",ln);
     fprintf(stderr,"%s\n",Msgerrparse[err-1]);
     FLAG=1;
   }

 fprintf(OUT,"\n");
 Output(bottom_tex);
 return(FLAG);
}

/*************************************************************/

#define SIZEINDEX 2048
void OutputIndex(idx)
     char idx[SIZEINDEX];
{
  char *pt;
  conceptentry ce;
  fctentry fe;
  int i,lvl=0;
  

  ce.l1=fe.l1=nbobjets;
  ce.l2=nblev2;
  ce.l3=nblev3;
  fe.l2=nbfct;

#ifndef ZEN_NO_TEST
  if(NO_INDEX)
    {
      fprintf(stderr,"No indexing for %s...\n",idx);
      return;
    }
  fprintf(stderr,"Indexing %s...\n",idx);
#endif
  pt=idx;
  

  IndexCheckAll();

#ifdef ZEN_NO_TEST
# define DEBUGINDEX 0
#else
# define DEBUGINDEX 1
#endif  
  fprintf(OUT,"\\glossary{");
  if(ce.l1<nbobjets)
    {
      fprintf(OUT,"%s@{\\Large\\bf %s}",objets[ce.l1],objets[ce.l1]);
      if(ce.l2<nblev2)
        {
          fprintf(OUT,"!%s@{\\large\\bf %s}",objets2[ce.l2],objets2[ce.l2]);
          if(ce.l3<nblev3)
            fprintf(OUT,"!%s",objets3[ce.l3]);
#if DEBUGINDEX
          else
            fprintf(OUT,"!{\\sf %s}",idx);
#endif
        }
      else
        {
          if(ce.l3<nblev3)
            fprintf(OUT,"!%s@{\\large\\bf %s}",objets3[ce.l3],objets3[ce.l3]);
#if DEBUGINDEX
          fprintf(OUT,"!{\\sf %s}",idx);
#endif
        }
    }
  else
    {
      if(ce.l2<nblev2)
        {
          fprintf(OUT,"%s@{\\Large\\bf %s}",objets2[ce.l2],objets2[ce.l2]);
          if(ce.l3<nblev3)
            fprintf(OUT,"!%s@{\\large\\bf %s}",objets3[ce.l3],objets3[ce.l3]);
#if DEBUGINDEX
          fprintf(OUT,"!{\\sf %s}",idx);
#endif
        }
      else
        {
          if(ce.l3<nblev3)
            {
              fprintf(OUT,"%s@{\\Large\\bf %s}",objets3[ce.l3],objets3[ce.l3]);
#if DEBUGINDEX
              fprintf(OUT,"!{\\sf %s}",idx);        
#endif
            }
#if DEBUGINDEX
          else
            fprintf(OUT,"{\\sf %s}",idx);
#endif    
        }
    }
  fprintf(OUT,"}\n");

  fprintf(OUT,"\\index{");
  if(fe.l1<nbobjets)
    {
      fprintf(OUT,"%s@{\\Large\\bf %s}",objets[fe.l1],objets[fe.l1]);
      if(fe.l2<nbfct)
        fprintf(OUT,"!%s@{\\large\\bf %s}",fctlev2[fe.l2],fctlev2[fe.l2]);
      fprintf(OUT,"!");
    }
  else
    {
      if(ce.l2<nbfct)
        {
          fprintf(OUT,"%s@{\\Large\\bf %s}",objets2[ce.l2],objets2[ce.l2]);
          if(fe.l2<nbfct)
            fprintf(OUT,"!%s@{\\large\\bf %s}",fctlev2[fe.l2],fctlev2[fe.l2]);
          fprintf(OUT,"!");
        }
      else if(fe.l2<nbfct)
        {
          fprintf(OUT,"%s@{\\Large\\bf %s}",fctlev2[fe.l2],fctlev2[fe.l2]);
          fprintf(OUT,"!");
        }      
    }  
  fprintf(OUT,"{\\sf %s}",idx);
  fprintf(OUT,"}\n");
}

int FindNextTex(ln)
     int *ln;
{int c;
 char FLAG,HEAD,INDEX,PIPE,HEAD_COMMENTS,CITE_C,SILENT_TEX;
 char memidx[SIZEINDEX],*idx;
 char unsidx[SIZEINDEX],*uns;

#define NOT_COMMENTS 0
#define COMMENTS 1
#define TEX_BLOCK 2

 
#define FORGET 0
#define SWALLOW 1
#define CONTINUE 2
#define FOUND 3
#define JUMP 4

 
 FLAG=NOT_COMMENTS;
 HEAD=INDEX=FORGET;
 PIPE=CITE_C=0;
 HEAD_COMMENTS=0;
 SILENT_TEX=0;
 idx=memidx;
 uns=unsidx;
 
 do
   {
     c=zengetc(IN);

     if(c == 012)
       (*ln) ++;
     
     switch(FLAG)
       {
/*************************/
        case NOT_COMMENTS :
          if(CITE_C)
            {
              switch(c)
                {
                case EOF:
                  fprintf(OUT,"\\E");
                  fprintf(OUT,"ndOfCite }\n");
                  return(0);

                case ' ':
                  fprintf(OUT,"\\hspace*{2mm}");
                  break;

                case '\t':
                  fprintf(OUT,"\\hspace*{16mm}");
                  break;

                case '/':       /* looking for beginning of comments */
                  c=zengetc(IN);
                  if(c == '*')
                    {
                      FLAG=COMMENTS;
                      fprintf(OUT,"}{\\small ");
                    }
                  else
                    fprintf(OUT,"/%c",c);
                break;

                case '~':
                  fprintf(OUT,"\\~{}");
                  break;

                case '^':
                  fprintf(OUT,"\\^{}");
                  break;

                case '\\':
                  fprintf(OUT,"$\\backslash$");
                  break;
                  /* 
                case '-':
                  c=zengetc(IN);
                  if(c == '>')
                    fprintf(OUT,"$\\rightarrow$");                  
                  else
                    fprintf(OUT,"-%c",c);
                  break;
                  */
                case '\n':
                   fputc('\\',OUT);
                   
                case '#':
                case '$':
                case '%':
                case '&':
                case '_':
                case '{':
                case '}':
                  fputc('\\',OUT);
                  
                default:
                  fputc(c,OUT);
                }
            }
          else
            {
              switch(HEAD)
                {
                case FOUND :            /* if we currently write a header */
                  switch(c)
                    {
                    case '{':   /* this is the end of the header */
                      HEAD=FORGET;
                      if(SILENT_TEX)
                        {
                          Output(bottom_silent_proc);
                          SILENT_TEX=0;
                        }
                      else
                        Output(bottom_proc);
                      if(INDEX) /* if we haven't yet output the index */
                        {
                          *idx=0;
                          idx=memidx;
                          uns=unsidx;                     
                          OutputIndex(memidx);
                          INDEX=FORGET;
                        }
                      break;
                    case 012 :
                      if(SILENT_TEX) /* first newline, we stop... */
                        {
                          Output(bottom_silent_proc);
                          SILENT_TEX=0;
                          HEAD=FORGET;
                          if(INDEX)     /* if we haven't yet output the index */
                            {
                              *idx=0;
                              idx=memidx;
                              uns=unsidx;                         
                              OutputIndex(memidx);
                              INDEX=FORGET;
                            }
                        }
                      else
                        {
                          fprintf(OUT,"}\\\\&{\n");
                          if(INDEX==FOUND)      /* if we have found the index */
                            INDEX=CONTINUE;
                        }
                      break;
                      
                    case '/' :
                      c=zengetc(IN);
                      if(c == '*')
                        {
                          HEAD_COMMENTS=1;
                          fprintf(OUT,"\\makebox[0in][l]{\\rm\\small /*");/**/
                            if(INDEX==FOUND)
                              INDEX=JUMP;
                        }
                      else
                        {
                          fputc('/',OUT);
                          fputc(c,OUT);
                          if(INDEX==FOUND)
                            {
                              *idx++ ='/';
                              *idx++ = c;
                            }
                        }
                      break;
                      
                    case '*' :
                      if(HEAD_COMMENTS)
                        {
                          fputc('*',OUT);
                          do
                            {
                              c=zengetc(IN);
                              fputc(c,OUT);
                            } while(c == '*');
                          if(c == '/')
                            {
                              HEAD_COMMENTS=0;
                              fputc('}',OUT);
                              if(INDEX==JUMP)
                                INDEX=FOUND;
                            }
                        }
                      else
                        fputc('*',OUT);
                      break;
                          
                    case '_':
                      fprintf(OUT,"\\_");
                      if(INDEX==FOUND)
                        {
                          *idx++ = '\\';
                          *idx++ = '_';
                        }
                      break;
                      
                    default:
                      fputc(c,OUT);     /* we write everything */
                      if(INDEX==FOUND)
                        *idx++ = c;
                      else if((INDEX==SWALLOW)
                              &&(HEAD_COMMENTS==0)
                              &&(c == ' '))
                        {
                          *uns=0;
                          if(strncmp(unsidx+1,"signed",strlen("signed"))) /* sordid hack !
                                                                             I have no time to make that
                                                                             better */
                            INDEX=FOUND;
                          uns=unsidx;
                        }
                      else
                        *uns++=c;
                    }
                    
                  break;
                  
                default :
                  if(c == '/')          /* looking for beginning of comments */
                    {
                      c=zengetc(IN);
                      if(c == '*')
                        {
                          FLAG=COMMENTS;
                          if(HEAD)      /* if this is a comment while
                                           waiting for a header, we memorize it */
                            HEAD=CONTINUE;
                        }
                    }
                  else if (HEAD)        /* we are looking for a header */
                    if((c != 012)
                       &&(c != '\t')
                       &&(c != ' ')) /* we swallow the white spaces */
                      {
                        HEAD = FOUND;
                        INDEX=SWALLOW;
                        if(c == '_')
                          fprintf(OUT,"\\_");
                        else
                          fputc(c,OUT);
                      }
                }
            }
          break;
/*************************/
       case COMMENTS :
         switch(c)
           {
           case '\\':           /* looking for a TEX block */
             c=zengetc(IN);
             switch(c)
               {
               case 'T':
                 if(FindToken("EX",NULL))
                   {
                     FLAG = TEX_BLOCK;
                     c=zengetc(IN);
                     switch(c)
                       {
                       case '[':
                         SILENT_TEX=1;
                       case '{':
                         Output(header_proc);
                         break;
                       default:
                         break;
                       }
                   }
                 break;

               case 'C':
                 if(FindToken("iteC",NULL))
                   {
                     if(CITE_C)
                       return(UNENDED_CITE_C);
                     
                     c='\n';
                     CITE_C= (char)-1;
                   }
                 break;

               case 'E':
                 if(CITE_C)
                   {
                     fprintf(OUT,"\\E");
                     if(FindToken("ndCiteC",OUT))
                       {
                         fprintf(OUT,"ndOfCite }\n");
                         CITE_C=0;
                       }
                   }
                 break;

               default:
                 if(CITE_C)
                   {
                     fputc('\\',OUT);
                     fputc(c,OUT);
                   }
               }
             break;
             
           case '*':            /* looking for the end of comments */
             do
               c=zengetc(IN);
             while(c == '*');
             
             if(c == '/')
               {
                 FLAG=NOT_COMMENTS;
                 if(HEAD == SWALLOW) /* the first time we are here we
                                      output \ZENHeader */
                   Output(header);
                 switch(CITE_C)
                   {
                   case (char)-1:
                     fprintf(OUT,"\n\\\n{\\tt ");
                     CITE_C=1;
                     break;
                   case 1:
                     fprintf(OUT," }{\\tt ");
                     break;
                   }
               }
             break;

           default:
             if(CITE_C)
               switch(c)
                 {
                 case '\n':
                   fputc('\\',OUT);
                   fputc(' ',OUT);
                   fputc('\\',OUT);
                   
                 case '#':
                 case '$':
                 case '%':
                 case '&':
                 case '~':
                 case '_':
                 case '^':
                 case '{':
                 case '}':
                   fputc('\\',OUT);
                   
                 default:
                   fputc(c,OUT);
                 }
           }
          
          break;

/*************************/
        default :
                                /* we are in a TEX Block */
         switch(c)
           {
           case '{' :           /* we count the brackets */
             if(PIPE)
               fprintf(OUT,"\\{");
             else
               {
                 FLAG++;
                 fputc(c,OUT);
               }
             break;
                 
           case '}' :
             if(PIPE)
               fprintf(OUT,"\\}");
             else
               {
                 FLAG--;
                 if(FLAG>=TEX_BLOCK)
                   fputc(c,OUT);
                 else if (CITE_C)
                   return(UNENDED_CITE_C);
               }
             break;
             
           case '|' :
             if (PIPE)
               fputc('}',OUT);
             else
               fprintf(OUT,"\\mbox{\\sf ");
             PIPE=!(PIPE);
             break;
             
           case '*' :           /* we check for end of comments and return
                                   error in this case */
             if(PIPE)
               fprintf(OUT,"\\mbox{$\\star$}");
             else
               {
                 fputc('*',OUT);
                 do
                   {
                     c=zengetc(IN);
                     fputc(c,OUT);
                   } while(c == '*');

                 switch(c)
                   {
                   case '{' :           /* we count the brackets */
                     FLAG++;
                     break;
                 
                   case '}' :
                     FLAG--;
                     if((FLAG < TEX_BLOCK)
                        &&(CITE_C))
                       return(UNENDED_CITE_C);
                     break;
                     
                   case '/':
                     return(UNENDED_TEX_BLOCK);
                   }
               }
             break;

           case '-':
             if(PIPE)
               {
                 c=zengetc(IN);
                 if(c == '>')
                   fprintf(OUT,"$\\rightarrow$");                   
                 else if(c == '|')
                   {
                     PIPE=!(PIPE);
                     fputc('}',OUT);
                   }
                 else
                   fprintf(OUT,"-%c",c);
               }
             else
               fputc('-',OUT);           
             break;

           case '\\' :          /* this is certainly a command */
             if(PIPE)
               fprintf(OUT,"\\mbox{$\\backslash$}");
             else
               {
                 fputc('\\',OUT);
                 if(FLAG == TEX_BLOCK) /* if we are at root for LaTeX */
                   {
                     c=zengetc(IN);
                     switch(c)
                       {
                       case 't':
                         fputc('t',OUT);
                         if(FindToken("itle{",OUT))
                           {
                             fprintf(OUT,"itleZeN{");
                             FLAG++;
                             if(HEAD)
                               return(TWO_TEX_BLOCKS);
                             HEAD=SWALLOW;
                           }
                         break;
                       case 'i':
                         fputc('i',OUT);
                         if(FindToken("nput{",OUT))
                           {
                             FLAG++;
                             if (HEAD == SWALLOW) {
                               fprintf(OUT,"nputZeN{");
                             }
                             else {
                               fprintf(OUT,"nput{");
                             }
                           }
                         break;
                       case 'o':
                         fputc('o',OUT);
                         if(FindToken("utput{",OUT))
                           {
                             FLAG++;
                             fprintf(OUT,"utputZeN{");
                           }
                         break;
                       case 's':
                         fputc('s',OUT);
                         if(FindToken("ide{",OUT))
                           {
                             FLAG++;
                             fprintf(OUT,"ideZeN{");
                           }
                         break;
                       case 'n':
                         fputc('n',OUT);
                         if(FindToken("ote{",OUT))
                           {
                             FLAG++;
                             fprintf(OUT,"oteZeN{");
                           }
                         break;

                       default:
                         fputc(c,OUT);
                       }
                   }
                 break;
               case '>':
                 if(PIPE)
                   fprintf(OUT,"\\mbox{$>$}",c);
                 else
                   fputc(c,OUT);
                 break;          
               case '_' :
               case '&' :
                 if(PIPE)
                   fprintf(OUT,"\\%c",c);
                 else
                   fputc(c,OUT);
                 break;
               default :
                 fputc(c,OUT);
               }
           }
       }
     
   } while(c != EOF);

 return(0);
}

/*************************************************************/
int FindToken(tok,out)
     char *tok;
     FILE *out;
{int i,c,l;

 l=strlen(tok);
 if(l==0)
   return(0);
 i=0;

 do
   {
     c=zengetc(IN);
     if(c != tok[i])
       {
         if(out != NULL)
           {
             for(l=0;l<i;l++)
               fputc(tok[l],out);
             fputc(c,out);
           }
         return(0);
     }
     i++;
   } while(i<l);
 return(1);
}

/*************************************************************/
void Output(msg)
     char *msg[];
{int i;

 if(msg!=NULL)
   for(i=0;
       msg[i]!=NULL;
       i++)
     fprintf(OUT,"%s\n",msg[i]);
}

/*************************************************************/
