/*
 * zmake/zrules.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */


/*
  \TEX{
  \Section{Processing files}
  } */


#ifndef IsADirectory
# define IsADirectory _IsADirectory
#endif
extern int _IsADirectory _H((char *name));
int _IsADirectory(name)
     char *name;
{
  char *path;
  int ret;
  zstatt st;

  ExpandedFileName(&path, name, PATHSEP);
  ret = zstat(path, &st);
  ZFree(path);
  if (ret) {return(-1);}
  if (ZS_ISDIR(st.st_mode)) {return(1);}
  return(0);
}

#ifndef MkdirHier
# define MkdirHier _MkdirHier
#endif
extern char *_znext_sep _H((char *path));
char *_znext_sep(path)
     char *path;
{
  char *ptr;

  for(ptr = path; *ptr; ptr++) {
    if ((*ptr == '/') &&
        ((ptr == path) || (*(ptr-1) != '\\'))) {
      return(ptr);
    }
  }
  return(NULL);
}
extern int _MkdirHier _H((char *path));
int _MkdirHier(path)
     char *path;
{
  int ret;
  char *next, *prev, *tmp, *dir;

  if (VERBOSE >= 50) {
    zfprintf(stderr, "mkdir %s\n", path);
  }
  for(prev = path; next = _znext_sep(prev); prev = next + 1) {
    char c;
    
    c = path[next-path];
    path[next-path] = '\0';
    ExpandedFileName(&dir, path, PATHSEP);
    zmkdir(dir);
    ZFree(dir);
    path[next-path] = c;
  }
  return(zmkdir(path));
}

#ifndef ChangeDirectory
# define ChangeDirectory _ChangeDirectory
#endif
extern void _ChangeDirectory _H((char *path));
void _ChangeDirectory(path)
     char *path;
{
  char *dir;

  ExpandedFileName(&dir, path, PATHSEP);
  if (VERBOSE >= 50) {
    zfprintf(stderr, "cd %s\n", dir);
  }
  if (zchdir(dir)) {
    zfprintf(stderr, "I can't go to \"%s\", I give up.\n", dir);
    zexit(-1);
  }
  ZFree(dir);
}

#ifndef RemoveFile
# define RemoveFile _RemoveFile
#endif
extern int _RemoveFile _H((char *path, char *prefix, char *file, char *suffix));
int _RemoveFile(path, prefix, file, suffix)
     char *path;
     char *prefix;
     char *file;
     char *suffix;
{
  int ret;
  char *name;

  name = PathPrefixFileSuffix(path, PATHSEP, prefix, file, suffix);
  ret = zunlink(name);
  if (!ret && (VERBOSE >= 50)) {
    zfprintf(stderr, "rm %s\n", name);
  }
  ZFree(name);
  return(ret);
}

#ifndef RemoveDir
# define RemoveDir _RemoveDir
#endif
extern int _RemoveDir _H((char *path));
int _RemoveDir(path)
     char *path;
{
  int ret;
  char *name;

  name = PathPrefixFileSuffix(path, PATHSEP, "", "", "");
  ret = zrmdir(name);
  if (!ret && (VERBOSE >= 50)) {
    zfprintf(stderr, "rmdir %s\n", name);
  }
  ZFree(name);
  return(ret);
}

#ifndef RmDirHier
# define RmDirHier _RmDirHier
#endif
extern char *_zprev_sep _H((char *begin, char *path));
char *_zprev_sep(begin, path)
     char *begin;
     char *path;
{
  char *ptr;

  for(ptr = path; ptr >= begin; ptr--) {
    if ((*ptr == '/') &&
        ((ptr == begin) || (*(ptr-1) != '\\'))) {
      return(ptr);
    }
  }
  return(NULL);
}
extern int _RmDirHier _H((char *path));
int _RmDirHier(path)
     char *path;
{
  int ret;
  char *name, *prev, *next;

  if ((ret = RemoveDir(path))) {
    return(ret);
  }
  for(prev = path+zstrlen(path)-1;
      next = _zprev_sep(path, prev);
      prev = next - 1) {
    char c;
    
    c = path[next-path]; path[next-path] = '\0';
    
    if ((ret = RemoveDir(path))) {
      path[next-path] = c;
      return(ret);
    }
    path[next-path] = c;
  }
  return(0);
}

#ifndef RemoveDirInAnyCase
# define RemoveDirInAnyCase _RemoveDirInAnyCase
#endif
extern int _RemoveDirInAnyCase _H((char *path));
int _RemoveDirInAnyCase(path)
     char *path;
{
  int ret;
  char *name;
  zDIR *dir;
  zdirent *entry;

  ExpandedFileName(&name, path, PATHSEP);
  dir = zopendir(name);
  ZFree(name);
  if (!dir) {return(-1);}
  while((entry = zreaddir(dir))) {
    if (zstrcmp(entry->d_name, ".") && zstrcmp(entry->d_name, "..")) {
      name = concat(path, PATHSEP); append(&name, entry->d_name);
      if (IsADirectory(name) == 1) {
        RemoveDirInAnyCase(name);
      }
      else {
        RemoveFile(path, "", entry->d_name, "");
      }
      ZFree(name);
    }
  }
  zclosedir(dir);

  return(RemoveDir(path));
}

#ifndef MoveFile
# define MoveFile _MoveFile
#endif
extern void _MoveFile _H((char *l_path, char *l_prefix, char *l_file, char *l_suffix, char *r_path, char *r_prefix, char *r_file, char *r_suffix));
void _MoveFile(l_path, l_prefix, l_file, l_suffix,
              r_path, r_prefix, r_file, r_suffix)
     char *l_path; char *l_prefix; char *l_file; char *l_suffix;
     char *r_path; char *r_prefix; char *r_file; char *r_suffix;
{
  char *lname, *rname;
  
  lname = PathPrefixFileSuffix(l_path, PATHSEP, l_prefix, l_file, l_suffix);
  rname = PathPrefixFileSuffix(r_path, PATHSEP, r_prefix, r_file, r_suffix);

  if (zstrcmp(lname, rname)) {
    if (VERBOSE >= 50) {
      zfprintf(stderr, "mv %s %s\n", lname, rname);
    }
    zrename(lname, rname);
  }

  ZFree(lname); ZFree(rname);
}

#ifndef CopyFile
# define CopyFile _CopyFile
#endif
extern int _CopyFile _H((char *l_path, char *l_prefix, char *l_file, char *l_suffix, char *r_path, char *r_prefix, char *r_file, char *r_suffix));
int _CopyFile(l_path, l_prefix, l_file, l_suffix,
               r_path, r_prefix, r_file, r_suffix)
     char *l_path; char *l_prefix; char *l_file; char *l_suffix;
     char *r_path; char *r_prefix; char *r_file; char *r_suffix;
{
  char buf[SIZE_BUF+1];
  int in, out, ret, size;
  char *lname, *rname;
  zstatt st;
  
  ret = 0;
  lname = PathPrefixFileSuffix(l_path, PATHSEP, l_prefix, l_file, l_suffix);
  rname = PathPrefixFileSuffix(r_path, PATHSEP, r_prefix, r_file, r_suffix);

  if (VERBOSE >= 50) {
    zfprintf(stderr, "cp %s %s\n", lname, rname);
  }
  /* Opening the file */
  if ((in = zopen(lname)) == -1) {ret = -1; goto end;}
  if ((out = zcreat(rname)) == -1) {ret = -1; close(in); goto end;}
  
  while((size = zread(in, buf, SIZE_BUF)) > 0) {
    if (zwrite(out, buf, size) < 0) {
      ret = -1; break;
    }
  }
  close(in); close(out);

  if (zstat(lname, &st) < 0) {ret = -1; goto end;}
  zchmod(rname, st.st_mode);

 end:
  ZFree(lname); ZFree(rname);
  return(ret);
}

#ifndef ReadFileInMemory
# define ReadFileInMemory _ReadFileInMemory
#endif
extern char *_ReadFileInMemory _H((char *path, char *prefix, char *file, char *suffix));
char *_ReadFileInMemory(path, prefix, file, suffix)
     char *path; char *prefix; char *file; char *suffix;
{
  int size, r, fd;
  char *text, *name, buf[SIZE_BUF];
  zstatt st;
  
  text = NULL;

  name = PathPrefixFileSuffix(path, PATHSEP, prefix, file, suffix);
  if (zstat(name, &st) < 0) {goto end;}
  if ((fd = zopen(name)) == -1) {goto end;}

  ZAlloc(text, st.st_size+1, char);
  for(size = 0; size < st.st_size; size += r) {
    r = zread(fd, buf, SIZE_BUF);
    if (size+r > st.st_size) {
      r = (st.st_size-size);
    }
    zstrncpy(text+size, buf, r);
  }
  text[st.st_size] = '\0';

 end:
  ZFree(name);
  return(text);
}

#ifndef ExecuteCommand
# define ExecuteCommand _ExecuteCommand
#endif
extern void _ExecuteCommand _H((char *command, char *indir, char *inprefix, char *infile, char *insuffix, char *outdir, char *outprefix, char *outfile, char *outsuffix));
void _ExecuteCommand(command,
                     indir, inprefix, infile, insuffix,
                     outdir, outprefix, outfile, outsuffix)
     char *command;
     char *indir; char *inprefix; char *infile; char *insuffix;
     char *outdir; char *outprefix; char *outfile; char *outsuffix;
{
  int ret;
  char *input, *output;

  if (infile) {
    input = PathPrefixFileSuffix(indir, PATHSEP, inprefix, infile, insuffix);
  }
  else {input = NULL;}
    
  if (outfile) {
    MkdirHier(outdir);
    output = PathPrefixFileSuffix(outdir, PATHSEP, outprefix, outfile, outsuffix);
  }
  else {output = NULL;}
    
  if (VERBOSE >= 10) {
    zfprintf(stderr, "%s", command);
    if (input) {zfprintf(stderr, " < %s", input);}
    if (output) {zfprintf(stderr, " > %s", output);}
    zfprintf(stderr, "\n");
  }

  if ((ret = Execute(command, PATHSEP, input, output))) {
    zfprintf(stderr, "Exit status is %d, I give up.\n", ret);
    zexit(-1);
  }
  
  if (input) {ZFree(input);}
  if (output) {ZFree(output);}
}

#ifndef ZMakeZMake
# define ZMakeZMake _ZMakeZMake
#endif
extern void _ZMakeZMake _H((char *current_dir, int argc, char *argv[], int remote));
void _ZMakeZMake(current_dir, argc, argv, remote)
     char *current_dir;
     int argc; char *argv[];
     int remote;
{
  int i;
  char *command, *pathes, *ptr, *zmake_dep;
  char **argvbis;

  if ((argc > 1) && !zstrcmp(argv[1], "internal_phase_1")) {
    RemoveFile(current_dir, "", ZMAKE, BSUF);
    MoveFile(current_dir, "", ZMAKE, ESUF,
             current_dir, "", ZMAKE, BSUF);
    ptr = concat(ZMAKE, "new");
    CopyFile(current_dir, "", ptr, ESUF,
             current_dir, "", ZMAKE, ESUF);
    ZFree(ptr);
    
    ZAlloc(argvbis, argc+1, char *);
    argvbis[0] = PathPrefixFileSuffix(current_dir, PATHSEP, "", ZMAKE, ESUF);
    argvbis[1] = "internal_phase_2";
    if (VERBOSE >= 50) {
      zfprintf(stderr, "%s internal_phase_2", argvbis[0]);
    }
    for(i = 2; i < argc; i++) {
      argvbis[i] = argv[i];
      if (VERBOSE >= 50) {
        zfprintf(stderr, " %s", argvbis[i]);
      }
    }
    argvbis[argc] = NULL;
    if (VERBOSE >= 50) {
      zfprintf(stderr, "\n");
    }
    
    zexecvp(argvbis[0], argvbis);
  }
  else if ((argc > 1) && !zstrcmp(argv[1], "internal_phase_2")) {
    ptr = concat(ZMAKE, "new");
    RemoveFile(current_dir, "", ptr, ESUF);
    ZFree(ptr);

    if ((argc > 2) && !zstrcmp(argv[2], "stop_after_internal_phases")) {
      exit(0);
    }
    
    ZAlloc(argvbis, argc+1, char *);
    argvbis[0] = argv[0];
    if (VERBOSE >= 50) {
      zfprintf(stderr, "%s", argvbis[0]);
    }
    for(i = 2; i < argc; i++) {
      argvbis[i-1] = argv[i];
      if (VERBOSE >= 50) {
        zfprintf(stderr, " %s", argvbis[i-1]);
      }
    }
    argvbis[i-1] = NULL;
    if (VERBOSE >= 50) {
      zfprintf(stderr, "\n");
    }
    if (VERBOSE >= 10) {
      zfprintf(stderr, "-------------\n");
    }

    zexecvp(argvbis[0], argvbis);
  }
  else {
    pathes = concat(ZMAKE_DIR, " ");
    append(&pathes, TOP_DIR); append(&pathes, " "); append(&pathes, current_dir);
    
    /* Do we have to rebuild zmake.c ? */
    zmake_dep = concat(ZMAKE, CSUF);
    append(&zmake_dep, " "); append(&zmake_dep, ZMAKE);
    
    if (NewerThan(current_dir, "", ZMAKE, ESUF,
                  pathes, "", zmake_dep, "",
                  PATHSEP, "# include \"", "\"") <= 0) {
      
      if (VERBOSE >= 10) {
        int i;
        
        zfprintf(stderr, "\n----------");
        for(i = 0;
            i < strlen(current_dir)+strlen(PATHSEP)+strlen(ZMAKE)+strlen(ESUF);
            i++) {
          zfprintf(stderr, "-");
        }
        zfprintf(stderr, "\n");
        
        zfprintf(stderr, "Updating %s%s%s%s:\n", current_dir, PATHSEP, ZMAKE, ESUF);
      }

      RemoveFile(current_dir, ZMAKE, "new", ESUF);

      command = concat(CC, ZCFCFLAGS);
      append(&command, " -DTOPDIR=\""); append(&command, TOP_DIR); append(&command, "\" ");      
      append(&command, " -DCURDIR=\""); append(&command, current_dir); append(&command, "\" ");
      append(&command, " -DZCFDIR=\""); append(&command, ZMAKE_DIR); append(&command, "\" ");
      append(&command, ZMAKE); append(&command, CSUF);
      append(&command, " -o ");
      append(&command, ZMAKE); append(&command, "new"); append(&command, ESUF);
      ExecuteCommand(command,
                     NULL, NULL, NULL, NULL,
                     NULL, NULL, NULL, NULL);

      ZFree(command);

      if (!remote) {
        ZAlloc(argvbis, argc+2, char *);
        argvbis[0] = PathPrefixFileSuffix(current_dir, PATHSEP, ZMAKE, "new", ESUF);
        argvbis[1] = "internal_phase_1";
        if (VERBOSE >= 50) {
          zfprintf(stderr, "%s internal_phase_1", argvbis[0]);
        }
        for(i = 1; i < argc; i++) {
          argvbis[i+1] = argv[i];
          if (VERBOSE >= 50) {
            zfprintf(stderr, " %s", argvbis[i+1]);
          }
        }
        argvbis[argc+1] = NULL;
        if (VERBOSE >= 50) {
          zfprintf(stderr, "\n");
        }

        zexecvp(argvbis[0], argvbis);
      }
      else {
        command = PathPrefixFileSuffix(current_dir, PATHSEP, ZMAKE, "new", ESUF);
        append(&command, " internal_phase_1 stop_after_internal_phases");
        ExecuteCommand(command,
                       NULL, NULL, NULL, NULL,
                       NULL, NULL, NULL, NULL);
      }
    }
    ZFree(zmake_dep);
    ZFree(pathes);
  }
}


#if 0
/* \TEX{
   
    \title{Recursively compiling and calling |zmake| command in
   directories.}
   
 \input{A target |target|, the list of directories |dirs| containing
   |zmake.c| files, targets |subtargets| which will be arguments for
   |zmake| called in these directories and the well known |argc| and
   |argv| C arguments.}
   
    \side{If target is non zero, in each directory |dirs|, |zmake| is
   compiled if this was not already done. The resulting |zmake| are
   called with target |subtarget|.}
   
    \note{If the calling |zmake| was just (re)compiled, the binaries
   |zmake| in the directories are automatically (re)compiled.}
   
   }  */
void NamedTargetSubdirs(target,dirs,subtarget,argc, argv)
     int target;
     char *dirs, *subtarget;   
     int argc; char *argv[];{}
#endif

#ifndef NamedTargetSubdirs
# define NamedTargetSubdirs _NamedTargetSubdirs
#endif
extern void _NamedTargetSubdirs _H((int target, char *dirs, char *subtarget, int argc, char *argv[]));
void _NamedTargetSubdirs(target,dirs,subtarget,argc, argv)
     int target;
     char *dirs;
     char *subtarget;   
     int argc; char *argv[];
{
  int i;
  char *cfile, *dir, *cdir, *command;

  if (!target) {return;}

  for(cfile = ExpandedFileName(&dir, dirs, PATHSEP);
      cfile;
      cfile = ExpandedFileName(&dir, cfile, PATHSEP)) {
    
    /* New zmake ? */
    cdir = concat(CURRENT_DIR, PATHSEP); append(&cdir, dir);
    ChangeDirectory(cdir);
    ZMakeZMake(cdir, argc, argv, 1);
    
    /* Subtarget */
    command = PathPrefixFileSuffix(cdir, PATHSEP, "", ZMAKE, ESUF);
    ZFree(cdir);
    for(i = 1 ; i < argc; i++) {
      char *ptr;

      for(ptr = argv[i]; *ptr && (*ptr != '='); ptr++) {;}
      if (*ptr == '=') {
        append(&command, " "); append(&command, argv[i]);
      }
    }
    append(&command, " "); append(&command, subtarget);
    ExecuteCommand(command,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);
    ZFree(command);
    ZFree(dir);
  }
  ZFree(dir);
}


#ifndef LexToSrc
# define LexToSrc _LexToSrc
#endif
extern int _LexToSrc _H((char *cprefix, char *cfile, char *ydir, char *yprefix, char *yfile, char *includes, char *cflags));
int _LexToSrc(cprefix, cfile,
                  ydir, yprefix, yfile,
                  includes, flags)
     char *cprefix; char *cfile;
     char *ydir; char *yprefix; char *yfile;
     char *includes; char *flags;
{
  int ret;
  char *command, *name, *dirs;

  ret = 0;
  dirs = concat(ydir, " "); append(&dirs, includes);

  if (NewerThan(".", cprefix, cfile, CSUF,
                dirs, yprefix, yfile, LSUF,
                PATHSEP, "# include \"", "\"") <= 0) {

    name = PathPrefixFileSuffix(ydir, PATHSEP, yprefix, yfile, LSUF);

    command = concat(LEX, " ");
    append(&command, flags);
    append(&command, " "); append(&command, name);

    ExecuteCommand(command,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);

    ZFree(command);

    MoveFile(".", "lex.", "yy", CSUF,
             ".", cprefix, cfile, CSUF);

    ZFree(name);
    ret = 1;
  }
  ZFree(dirs);

  return(ret);
}

#if 0
/*    \TEX{
   
     \title{Calling |lex| on source files to obtain C files.}
   
     \input{A target |target|, the directory |odir| which will receive
   the obtained C files, a prefix |cprefix| which will be added to
   these C files, the directory |ydir| which contains the |lex| files,
   the prefix |yprefix| of these |lex| files, the names |yfiles| of
   these |lex| files (without the usual suffix |.l|), directories
   |includes| which must be included and specific flags |flags| to add
   to |lex|.}
   
    \output{The integer 1 if an error occured, 0 otherwise.}
   
    \side{C files are created in |odir| if |target| is non zero.}
   
   }  */
int LexToSrcs(target, odir, cprefix, ydir, yprefix, yfiles,
              includes, flags)
     int target;
     char *odir; char *cprefix;
     char *ydir; char *yprefix; char *yfiles;
     char *includes; char *flags;{}
#endif

#ifndef LexToSrcs
# define LexToSrcs _LexToSrcs
#endif
extern int _LexToSrcs _H((int target, char *odir, char *cprefix, char *ydir, char *yprefix, char *yfiles, char *includes, char *flags));
int _LexToSrcs(target,
                   odir, cprefix,
                   ydir, yprefix, yfiles,
                   includes, flags)
     int target;
     char *odir; char *cprefix;
     char *ydir; char *yprefix; char *yfiles;
     char *includes; char *flags;
{
  int ret;
  char *yfile, *next;

  ret = 0;
  if (!target) {return(ret);}
  
  MkdirHier(odir);
  ChangeDirectory(odir);
  for(next = ExpandedFileName(&yfile, yfiles, PATHSEP);
      next;
      next = ExpandedFileName(&yfile, next, PATHSEP)) {
    if (LexToSrc(cprefix, yfile,
                 ydir, yprefix, yfile,
                 includes, flags)) {
      ret = 1;
    }
    ZFree(yfile);
  }
  ZFree(yfile);
  return(ret);
}

/*
 * YaccToSrc - Compiling one file
 */
#ifndef YaccToSrc
# define YaccToSrc _YaccToSrc
#endif
extern int _YaccToSrc _H((char *cprefix, char *cfile, char *ydir, char *yprefix, char *yfile, char *includes, char f*lags));
int _YaccToSrc(cprefix, cfile,
                  ydir, yprefix, yfile,
                  includes, flags)
     char *cprefix; char *cfile;
     char *ydir; char *yprefix; char *yfile;
     char *includes; char *flags;
{
  int ret;
  char *command, *name, *dirs;

  ret = 0;
  dirs = concat(ydir, " "); append(&dirs, includes);

  if (NewerThan(".", cprefix, cfile, CSUF,
                dirs, yprefix, yfile, YSUF,
                PATHSEP, "# include \"", "\"") <= 0) {

    name = PathPrefixFileSuffix(ydir, PATHSEP, yprefix, yfile, YSUF);

    command = concat(YACC, " ");
    append(&command, flags);
    append(&command, " "); append(&command, name);

    ExecuteCommand(command,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);

    ZFree(command);

    MoveFile(".", "y.", "tab", CSUF,
             ".", cprefix, cfile, CSUF);

    ZFree(name);
    ret = 1;
  }
  ZFree(dirs);

  return(ret);
}

#if 0
/*    \TEX{
   
      \title{Calling |yacc| on source files to obtain C files.}
   
      \input{A target |target|, the directory |odir| which will
   receive the obtained C files, a prefix |cprefix| which will be
   added to these C files, the directory |ydir| which contains the
   |yacc| files, the prefix |yprefix| of these |yacc| files, the names
   |yfiles| of these |yacc| files (without the usual suffix |.y|),
   directories |includes| which must be included and specific flags
   |flags| to add to |yacc|.}
   
      \output{The integer 1 if an error occured, 0 otherwise.}
   
      \side{C files are created in |odir| if |target| is non zero.}
   
   }  */
int YaccToSrcs(target, odir, cprefix, ydir, yprefix, yfiles,
               includes, flags)
     int target;
     char *odir; char *cprefix;
     char *ydir; char *yprefix; char *yfiles;
     char *includes; char *flags;{}
#endif

/*
 * YaccToSrcs - Compiling several files
 */
#ifndef YaccToSrcs
# define YaccToSrcs _YaccToSrcs
#endif
extern int _YaccToSrcs _H((int target, char *odir, char *cprefix, char *ydir, char *yprefix, char *yfiles, char *includes, char *flags));
int _YaccToSrcs(target,
                odir, cprefix,
                ydir, yprefix, yfiles,
                includes, flags)
     int target;
     char *odir; char *cprefix;
     char *ydir; char *yprefix; char *yfiles;
     char *includes; char *flags;
{
  int ret;
  char *yfile, *next;

  ret = 0;
  if (!target) {return(ret);}
  
  MkdirHier(odir);
  ChangeDirectory(odir);
  for(next = ExpandedFileName(&yfile, yfiles, PATHSEP);
      next;
      next = ExpandedFileName(&yfile, next, PATHSEP)) {
    if (YaccToSrc(cprefix, yfile,
                  ydir, yprefix, yfile,
                  includes, flags)) {
      ret = 1;
    }

    ZFree(yfile);
  }
  ZFree(yfile);
  return(ret);
}

/*
 * SrcToObject - Compiling one file
 */
#ifndef SrcToObject
# define SrcToObject _SrcToObject
#endif
extern int _SrcToObject _H((char *oprefix, char *ofile, char *cdir, char *cprefix, char *cfile, char *includes, char *cflags));
int _SrcToObject(oprefix, ofile,
                  cdir, cprefix, cfile,
                  includes, cflags)
     char *oprefix; char *ofile;
     char *cdir; char *cprefix; char *cfile;
     char *includes; char *cflags;
{
  int ret;
  char *command, *name, *dirs;

  ret = 0;
  dirs = concat(cdir, " "); append(&dirs, includes);

  if (NewerThan(".", oprefix, ofile, OSUF,
                dirs, cprefix, cfile, CSUF,
                PATHSEP, "# include \"", "\"") <= 0) {

    name = PathPrefixFileSuffix(cdir, PATHSEP, cprefix, cfile, CSUF);

    command = concat(CC, cflags); append(&command, CFLAGS);
    append(&command, " -c "); append(&command, name);

    ExecuteCommand(command,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);

    ZFree(command);

    MoveFile(".", cprefix, cfile, OSUF,
             ".", oprefix, ofile, OSUF);

    ZFree(name);
    ret = 1;
  }
  ZFree(dirs);

  return(ret);
}

#if 0
/*    \TEX{
   
      \title{Compiling C files to obtain object files.}
   
      \input{A target |target|, the directory |odir| which will
   receive the obtained object files, a prefix |oprefix| which will be
   added to these object files, the directory |cdir| which contains
   the C files, the prefix |cprefix| of these C files, the names
   |cfiles| of these C files (without the usual suffix |.c|),
   directories |includes| which must be included and specific flags
   |cflags| to add to |cc|.}
   
      \output{The integer 1 if an error occured, 0 otherwise.}
   
      \side{Object files are created in |cdirs| if |target| is non
   zero.}  }  */
int SrcToObjects(target, odir, oprefix, cdir, cprefix, cfiles,
                 includes, cflags)
     int target;
     char *odir; char *oprefix;
     char *cdir; char *cprefix; char *cfiles;
     char *includes; char *cflags;{}
#endif

#ifndef SrcToObjects
# define SrcToObjects _SrcToObjects
#endif
extern int _SrcToObjects _H((int target, char *odir, char *oprefix, char *cdir, char *cprefix, char *cfiles, char *includes, char *cflags));
int _SrcToObjects(target,
                   odir, oprefix,
                   cdir, cprefix, cfiles,
                   includes, cflags)
     int target;
     char *odir; char *oprefix;
     char *cdir; char *cprefix; char *cfiles;
     char *includes; char *cflags;
{
  int ret;
  char *cfile, *next;

  ret = 0;
  if (!target) {return(ret);}
  
  MkdirHier(odir);
  ChangeDirectory(odir);
  for(next = ExpandedFileName(&cfile, cfiles, PATHSEP);
      next;
      next = ExpandedFileName(&cfile, next, PATHSEP)) {
    if (SrcToObject(oprefix, cfile,
                    cdir, cprefix, cfile,
                    includes, cflags)) {
      ret = 1;
    }
    ZFree(cfile);
  }
  ZFree(cfile);
  return(ret);
}

#if 0
/*    \TEX{
   
      \title{Archiving object files in a list file.}
   
      \input{A target |target|, the directory |libdir| which will
   receive the list file, a prefix |lprefix| which will be added to
   list file, the name |lib| of the obtained list file, the suffix
   |objtoar| of this list, the directory |odir| which contains the
   object files, the prefix |rprefix| of these object files and the
   names |ofiles| of these object files (without the usual suffix
   |.o|).}
   
      \output{The integer 1 if an error occured, 0 otherwise.}
   
      \side{The list file is created in |libdir| if |target| is non
   zero.}

}  */
int ObjectsToArList(target, libdir, lprefix, lib, objtoar,
                    odir, rprefix, ofiles)
     int target;
     char *libdir; char *lprefix; char *lib; char *objtoar;
     char *odir; char *rprefix; char *ofiles;{}
#endif

#ifndef ObjectsToArList
# define ObjectsToArList _ObjectsToArList
#endif
extern int _ObjectsToArList _H((int target, char *libdir, char *lprefix, char *lib, char *objtoar, char *odir, char *rprefix, char *ofiles));
int _ObjectsToArList(target,
                      libdir, lprefix, lib, objtoar,
                      odir, rprefix, ofiles)
     int target;
     char *libdir; char *lprefix; char *lib; char *objtoar;
     char *odir; char *rprefix; char *ofiles;
{
  int ret;
  FILE *file;
  char *name, *next, *ofile;

  if (!target) {return(0);}
  
  ret = 0;
  if (NewerThan(libdir, lprefix, lib, objtoar,
                odir, rprefix, ofiles, OSUF,
                PATHSEP, NULL, NULL) <= 0) {
    ret = 1;
    MkdirHier(libdir);

    name = PathPrefixFileSuffix(libdir, PATHSEP, lprefix, lib, objtoar);
    if (VERBOSE >= 50) {
      zfprintf(stderr, "echo \"%s\" > %s\n", ofiles, name);
    }
    if (!(file = zfcreat(name))) {
      zfprintf(stderr, "I can't open \"%s\", I give up.\n", name);
      zexit(-1);
    }
    ZFree(name);

    for(next = ExpandedFileName(&ofile, ofiles, PATHSEP);
        next;
        next = ExpandedFileName(&ofile, next, PATHSEP)) {
      
      name = PathPrefixFileSuffix(odir, PATHSEP, rprefix, ofile, OSUF);
      ZFree(ofile);
      zfprintf(file, "%s ", name);
      ZFree(name);
    }
    ZFree(ofile);
    zfclose(file);
  }
  return(ret);
}

#if 0
/*    \TEX{
   
      \title{Creating an executable program from object files.}
   
      \input{A target |target|, the directory |bdir| which will
   receive the executable program |bfile|, a prefix |bprefix| which
   will be added to this binary, the name |bfile| of this binary, the
   directory |odir| which contains the object files, the names
   |ofiles| of these object files (without the usual suffix |.o|), the
   flags needed for |cc| and the library to link with.}
   
      \output{The integer 1 if an error occured, 0 otherwise.}
   
      \side{The binary is created in |bdir| if |target| is non zero.}
   
   }  */
int ObjectsToBinary(target, bdir, bprefix, bfile, odir, oprefix, ofiles,
                    ldflags, ldlibs)
     int target;
     char *bdir; char *bprefix; char *bfile;
     char *odir; char *oprefix; char *ofiles;
     char *ldflags; char *ldlibs;{}
#endif

#ifndef ObjectsToBinary
# define ObjectsToBinary _ObjectsToBinary
#endif
extern int _ObjectsToBinary _H((int target, char *bdir, char *bprefix, char *bfile, char *odir, char *oprefix, char *ofiles, char *ldflags, char *ldlibs));
int _ObjectsToBinary(target,
                     bdir, bprefix, bfile,
                     odir, oprefix, ofiles,
                     ldflags, ldlibs)
     int target;
     char *bdir; char *bprefix; char *bfile;
     char *odir; char *oprefix; char *ofiles;
     char *ldflags; char *ldlibs;
{
  char *ofile, *next;
  char *command;
  char *opath;

  if (!target) {return(0);}
  
  if (NewerThan(bdir, bprefix, bfile, ESUF,
                odir, oprefix, ofiles, OSUF,
                PATHSEP, NULL, NULL) > 0) {
    char *l_name, *r_name, *lib;
    zstatt ls, rs;

    l_name = PathPrefixFileSuffix(bdir, PATHSEP, bprefix, bfile, ESUF);
    if (VERBOSE >= 100) {
      zfprintf(stderr, "The LHS is \"%s\".\n", l_name);
    }
    if (!zstat(l_name, &ls)) {

      for(next = ExpandedFileName(&lib, ldlibs, PATHSEP);
          next;
          next = ExpandedFileName(&lib, next, PATHSEP)) {
        if (VERBOSE >= 100) {
          zfprintf(stderr, "Testing the LHS versus %s\n", lib);
        }
        if (!zstat(lib, &rs)) {
          ZFree(lib);
          if (ls.st_mtime < rs.st_mtime) {break;}
        }
        else {
          ZFree(lib);
        }
      }
      if (!next) {
        ZFree(lib);
        ZFree(l_name);
        return(0);
      }
    }
    ZFree(l_name);
  }

  MkdirHier(bdir);
  ChangeDirectory(bdir);
  
  command = concat(CC, LDOPTIONS);    

  ExpandedFileName(&opath, odir, PATHSEP);
  append(&opath, PATHSEP);
    
  for(next = ExpandedFileName(&ofile, ofiles, PATHSEP);
      next;
      next = ExpandedFileName(&ofile, next, PATHSEP)) {
    
    append(&command, " ");
    append(&command, opath);
    append(&command, oprefix);
    append(&command, ofile);
    append(&command, OSUF);
    
    ZFree(ofile);
  }
  ZFree(opath);
  ZFree(ofile);
  
  append(&command, " "); append(&command, ldlibs); append(&command, LDLIBS);
  append(&command, " "); append(&command, ldflags);
  append(&command, " -o ");
  append(&command, bprefix);
  append(&command, bfile);
  append(&command, ESUF);
  
  ExecuteCommand(command,
                 NULL, NULL, NULL, NULL,
                 NULL, NULL, NULL, NULL);
    
  ZFree(command);
  return(1);
}

/*
 * ObjectsToBinary2 - Compiling several files
 */
#ifndef ObjectsToBinary2
# define ObjectsToBinary2 _ObjectsToBinary2
#endif
extern int _ObjectsToBinary2 _H((int target, char *bdir, char *bprefix, char *bfile, char *odir, char *oprefix, char *ofiles, char *odir1, char *oprefix1, char *ofiles1, char *ldflags, char *ldlibs));
int _ObjectsToBinary2(target,
                      bdir, bprefix, bfile,
                      odir, oprefix, ofiles,
                      odir1, oprefix1, ofiles1,
                      ldflags, ldlibs)
     int target;
     char *bdir; char *bprefix; char *bfile;
     char *odir; char *oprefix; char *ofiles;
     char *odir1; char *oprefix1; char *ofiles1;
     char *ldflags; char *ldlibs;
{
  char *ofile, *next;
  char *command;
  char *opath;

  if (!target) {return(0);}
  
  if ((NewerThan(bdir, bprefix, bfile, ESUF,
                 odir, oprefix, ofiles, OSUF,
                 PATHSEP, NULL, NULL) > 0) &&
      (NewerThan(bdir, bprefix, bfile, ESUF,
                 odir1, oprefix1, ofiles1, OSUF,
                 PATHSEP, NULL, NULL) > 0)) {
    char *l_name, *r_name, *lib;
    zstatt ls, rs;

    l_name = PathPrefixFileSuffix(bdir, PATHSEP, bprefix, bfile, ESUF);
    if (VERBOSE >= 100) {
      zfprintf(stderr, "The LHS is \"%s\".\n", l_name);
    }
    if (!zstat(l_name, &ls)) {

      for(next = ExpandedFileName(&lib, ldlibs, PATHSEP);
          next;
          next = ExpandedFileName(&lib, next, PATHSEP)) {
        if (VERBOSE >= 100) {
          zfprintf(stderr, "Testing the LHS versus %s\n", lib);
        }
        if (!zstat(lib, &rs)) {
          ZFree(lib);
          if (ls.st_mtime < rs.st_mtime) {break;}
        }
        else {
          ZFree(lib);
        }
      }
      if (!next) {
        ZFree(lib);
        ZFree(l_name);
        return(0);
      }
    }
    ZFree(l_name);
  }

  MkdirHier(bdir);
  ChangeDirectory(bdir);
  
  command = concat(CC, LDOPTIONS);    
  append(&command, " "); append(&command, ldflags);

  ExpandedFileName(&opath, odir, PATHSEP);
  append(&opath, PATHSEP);
    
  for(next = ExpandedFileName(&ofile, ofiles, PATHSEP);
      next;
      next = ExpandedFileName(&ofile, next, PATHSEP)) {
    
    append(&command, " ");
    append(&command, opath);
    append(&command, oprefix);
    append(&command, ofile);
    append(&command, OSUF);
    
    ZFree(ofile);
  }
  ZFree(ofile);
  ZFree(opath);
  
  ExpandedFileName(&opath, odir1, PATHSEP);
  append(&opath, PATHSEP);
    
  for(next = ExpandedFileName(&ofile, ofiles1, PATHSEP);
      next;
      next = ExpandedFileName(&ofile, next, PATHSEP)) {
    
    append(&command, " ");
    append(&command, opath);
    append(&command, oprefix1);
    append(&command, ofile);
    append(&command, OSUF);
    
    ZFree(ofile);
  }
  ZFree(ofile);
  ZFree(opath);
  
  append(&command, " ");
  append(&command, ldlibs);
  append(&command, LDLIBS);
  append(&command, " -o ");
  append(&command, bprefix);
  append(&command, bfile);
  append(&command, ESUF);
  
  ExecuteCommand(command,
                 NULL, NULL, NULL, NULL,
                 NULL, NULL, NULL, NULL);
    
  ZFree(command);
  return(1);
}

/*
 * SrcToOneTex - Compiling one file
 */
#ifndef SrcToOneTex
# define SrcToOneTex _SrcToOneTex
#endif
extern int _SrcToOneTex _H((char *tprefix, char *tfile, char *cdir, char *cprefix, char *cfile, char *csuf,
                            char *fdir, char *fprefix, char *filter));
int _SrcToOneTex(tprefix, tfile,
                  cdir, cprefix, cfile, csuf,
                  fdir, fprefix, filter)
     char *tprefix; char *tfile;
     char *cdir; char *cprefix; char *cfile; char *csuf;
     char *fdir; char *fprefix; char *filter;
{
  int ret;
  
  ret = 0;
  if (NewerThan(".", tprefix, tfile, TEXSUF,
                cdir, cprefix, cfile, csuf,
                PATHSEP, NULL, NULL) <= 0) {
    char *command, *output;

    ret = 1;
    command = PathPrefixFileSuffix(fdir, PATHSEP, fprefix, filter, ESUF);
    append(&command, " ");
    output = PathPrefixFileSuffix(".", PATHSEP, tprefix, tfile, "");
    append(&command, output);
    ZFree(output);

    ExecuteCommand(command,
                   cdir, cprefix, cfile, csuf,
                   ".", tprefix, tfile, TEXSUF);

    ZFree(command);
  }
  return(ret);
}

#ifndef IsTheLibraryStillHere
# define IsTheLibraryStillHere _IsTheLibraryStillHere
#endif
extern int IsTheLibraryStillHere _H((int target, char *ldir, char *lprefix, char *lname, char *lsuf,
                                     char *mdir, char *mprefix, char *mnames, char *msuf));
int IsTheLibraryStillHere(target,
                           ldir, lprefix, lname, lsuf,
                           mdir, mprefix, mnames, msuf)
     int target;
     char *ldir; char *lprefix; char *lname; char *lsuf;
     char *mdir; char *mprefix; char *mnames; char *msuf;
{
  int ret;
  char *name;
  zstatt st;

  ret = 0;
  name = PathPrefixFileSuffix(ldir, PATHSEP, lprefix, lname, lsuf);
  if (zstat(name, &st) < 0) {
    char *next; char *mname;

    ret = 1;
    for(next = ExpandedFileName(&mname, mnames, PATHSEP);
        next;
        next = ExpandedFileName(&mname, next, PATHSEP)) {
      RemoveFile(mdir, mprefix, mname, msuf);
      ZFree(mname);
    }
    ZFree(mname);
  }
  return(ret);
}

#if 0
/*    \TEX{
   
      \title{Adding object files to a library.}
   
      \input{A target |target|, the directory |ldir| which contains
        the library, a prefix |lprefix|, the name |lname| and the
        suffix |lsuf| of the library, the directory |mdir|, prefix
        |mprefix|, name |mname| and suffix |msuf| of a temporary file
        and the directory |ddir|, prefix |dprefix|, name |dname| and
        suffix |dsuf| of a list file obtained with |ObjectsToArList|.}
   
      \output{The integer 1 if an error occured, 0 otherwise.}
   
      \side{The tex files are compiled if |target| is non zero.}
   
   }  */
int ArListToLibrary(target, ldir, lprefix, lname, lsuf,
                    mdir, mprefix, mname, msuf, ddir, dprefix, dname, dsuf)
     int target;
     char *ldir; char *lprefix; char *lname; char *lsuf;
     char *mdir; char *mprefix; char *mname; char *msuf;
     char *ddir; char *dprefix; char *dname; char *dsuf;{}
#endif

#ifndef ArListToLibrary
# define ArListToLibrary _ArListToLibrary
#endif
extern int _ArListToLibrary _H((int target, char *ldir, char *lprefix, char *lname, char *lsuf, char *mdir, char *mprefix, char *mname, char *msuf, char *ddir, char *dprefix, char *dname, char *dsuf));
int _ArListToLibrary(target,
                     ldir, lprefix, lname, lsuf,
                     mdir, mprefix, mname, msuf,
                     ddir, dprefix, dname, dsuf)
     int target;
     char *ldir; char *lprefix; char *lname; char *lsuf;
     char *mdir; char *mprefix; char *mname; char *msuf;
     char *ddir; char *dprefix; char *dname; char *dsuf;
{
  int ret;
  char *new_ofiles, *old_ofiles;
  char *command, *name;

  if (!target) {return(0);}
  ret = 0;

  if (IsTheLibraryStillHere(target,
                            ldir, lprefix, lname, lsuf,
                            mdir, mprefix, mname, msuf)) {
    ret = 1;
  }

  if (NewerThan(mdir, mprefix, mname, msuf,
                ddir, dprefix, dname, dsuf,
                PATHSEP, NULL, NULL) <= 0) {
    MkdirHier(ldir);
    ret = 1;

    new_ofiles = ReadFileInMemory(ddir, dprefix, dname, dsuf);
    if (!new_ofiles) {return(-1);}

    old_ofiles = ReadFileInMemory(mdir, mprefix, mname, msuf);
    if (old_ofiles && strcmp(old_ofiles, new_ofiles)) {
      command = concat(ARRM, " ");
      name = PathPrefixFileSuffix(ldir, PATHSEP, lprefix, lname, lsuf);
      append(&command, name);
      ZFree(name);
      append(&command, " ");
      append(&command, old_ofiles);
      ExecuteCommand(command,
                     NULL, NULL, NULL, NULL,
                     NULL, NULL, NULL, NULL);
      ZFree(command);
      ZFree(old_ofiles);
    }

    name = PathPrefixFileSuffix(ldir, PATHSEP, lprefix, lname, lsuf);

    command = concat(ARADD, " ");
    append(&command, name);
    append(&command, " ");
    append(&command, new_ofiles);
    ExecuteCommand(command,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);
    ZFree(new_ofiles);
    ZFree(command);

#if DoRanlibCmd
    command = concat(RANLIB, " ");
    append(&command, name);
    ExecuteCommand(command,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);
    ZFree(command);
#endif

    ZFree(name);

    CopyFile(ddir, dprefix, dname, dsuf,
             mdir, mprefix, mname, msuf);
  }
  return(ret);
}

#if 0
/*    \TEX{
   
      \title{Parsing C files to obtain |LaTeX| documentation.}
   
      \input{A target |target|, the directory |tdir| which will
   receive the |LaTeX| files, a prefix |tprefix| which will be added to
   this |LaTeX| files, the directory |cdir| which contains the C files,
   the prefix |cprefix| of these C files, the names |cfiles| of these
   C files (without the usual suffix ".c"), the suffix |csuf| of these
   C files (usually |.c| or |.h|), the directory |fdir| which will
   contain the compiled parser, the prefix |fprefix| of this parser,
   the directory |pdir| which contains the sources of the parser and
   the name |filter| of the parser.}
   
      \output{The integer 1 if an error occured, 0 otherwise.}
   
      \side{The parser is compiled and the |LaTeX| files are created
        if |target| is non zero.}
   
   }  */
int SrcToTex(target, tdir, tprefix, cdir, cprefix, cfiles, csuf,
             fdir, fprefix, pdir, filter)
     int target;
     char *tdir; char *tprefix;
     char *cdir; char *cprefix; char *cfiles; char *csuf;
     char *fdir; char *fprefix; char *pdir; char *filter;{}
#endif

/*
 * SrcToTex - Compiling several files
 */
#ifndef SrcToTex
# define SrcToTex _SrcToTex
#endif
extern int _SrcToTex _H((int target, char *tdir, char *tprefix, char *cdir, char *cprefix, char *cfiles, char *csuf, char *fdir, char *fprefix, char *filter));
int _SrcToTex(target,
               tdir, tprefix,
               cdir, cprefix, cfiles, csuf,
               fdir, fprefix, pdir, filter)
     int target;
     char *tdir; char *tprefix;
     char *cdir; char *cprefix; char *cfiles; char *csuf;
     char *fdir; char *fprefix; char *pdir; char *filter;
{
  int ret;
  char *cfile, *next;

  if (!target) {return(0);}
  ret = 0;

  if (NewerThan(fdir, fprefix, filter, ESUF,
                pdir, "", "zparse", CSUF,
                PATHSEP, NULL, NULL) <= 0) {
    char *command;

    ret = 1;
    MkdirHier(fdir);
    ChangeDirectory(fdir);

    command = concat(CC, ZCFCFLAGS);
    append(&command, " ");
    append(&command, pdir); append(&command, "/zparse"); append(&command, CSUF);
    append(&command, " -o ");
    append(&command, fprefix); append(&command, filter); append(&command, ESUF);
    ExecuteCommand(command,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);
    ZFree(command);
  }

  MkdirHier(tdir);
  ChangeDirectory(tdir);
  for(next = ExpandedFileName(&cfile, cfiles, PATHSEP);
      next;
      next = ExpandedFileName(&cfile, next, PATHSEP)) {
    if (SrcToOneTex(tprefix, cfile,
                cdir, cprefix, cfile, csuf,
                    fdir, fprefix, filter)) {ret = 1;}
    ZFree(cfile);
  }
  ZFree(cfile);
  return(ret);
}


/*
 * SrcToOneDvi - Compiling one file
 */
#ifndef TexToOneDvi
# define TexToOneDvi _TexToOneDvi
#endif
extern int _TexToOneDvi _H((char *dprefix, char *dfile, char *tdir, char *tprefix, char *tfile));
int _TexToOneDvi(dprefix, dfile,
                  tdir, tprefix, tfile)
     char *dprefix; char *dfile;
     char *tdir; char *tprefix; char *tfile;
{
  int ret;

  ret = 0;
  if (NewerThan(".", dprefix, dfile, DVISUF,
                tdir, tprefix, tfile, TEXSUF,
                PATHSEP, "\\ input {", "}") <= 0) {
    char *command, *latex, *name;
    
    ret = 1;

    /* echo > file.ind */
    name = PathPrefixFileSuffix(".", PATHSEP, tprefix, tfile, ".ind");
    if (VERBOSE >= 50) {
      zfprintf(stderr, "echo > %s\n", name);
    }
    zclose(zcreat(name));
    ZFree(name);

    /* echo > file.ing */
    name = PathPrefixFileSuffix(".", PATHSEP, tprefix, tfile, ".ing");
    if (VERBOSE >= 50) {
      zfprintf(stderr, "echo > %s\n", name);
    }
    zclose(zcreat(name));
    ZFree(name);

    /* $(LATEX) $(SRC_DIR)/file.tex; */
    /* $(LATEX) $(SRC_DIR)/file.tex; */
    latex = concat(LATEX, " ");
    name = PathPrefixFileSuffix(tdir, PATHSEP, tprefix, tfile, TEXSUF);
    append(&latex, name);
    ZFree(name);
    ExecuteCommand(latex,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);
    ExecuteCommand(latex,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);

    /* $(MV) file.idx file.glo $(SRC_DIR) */
    MoveFile(".", tprefix, tfile, ".idx",
             tdir, tprefix, tfile, ".idx");
    MoveFile(".", tprefix, tfile, ".glo",
             tdir, tprefix, tfile, ".glo");

    /* $(MAKEINDEX) -c $(SRC_DIR)/file -s index.ist ;  */
    command = concat(MAKEINDEX, " -c ");
    name = PathPrefixFileSuffix(tdir, PATHSEP, tprefix, tfile, "");
    append(&command, name);
    ZFree(name);
    append(&command, " -s index.ist");
    ExecuteCommand(command,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);
    ZFree(command);

    /* $(MAKEINDEX) -c $(SRC_DIR)/file.glo -s glossary.ist -o file.ing */
    command = concat(MAKEINDEX, " -c ");
    name = PathPrefixFileSuffix(tdir, PATHSEP, tprefix, tfile, ".glo");
    append(&command, name);
    ZFree(name);
    append(&command, " -s glossary.ist -o ");
    name = PathPrefixFileSuffix(".", PATHSEP, tprefix, tfile, ".ing");
    append(&command, name);
    ZFree(name);
    ExecuteCommand(command,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);
    ZFree(command);

    /*$(LATEX) $(SRC_DIR)/file.tex */
    ExecuteCommand(latex,
                   NULL, NULL, NULL, NULL,
                   NULL, NULL, NULL, NULL);
    MoveFile(".", tprefix, tfile, DVISUF,
             ".", dprefix, dfile, DVISUF);

    /* $(RM) file.log file.aux file.toc file.ilg file.idx file.glo file.ind  file.ing 
       $(SRC_DIR)/file.ilg $(SRC_DIR)/file.idx $(SRC_DIR)/file.glo 
       $(SRC_DIR)/file.ind $(SRC_DIR)/file.ing */
    RemoveFile(tdir, tprefix, tfile, ".glo");
    RemoveFile(tdir, tprefix, tfile, ".idx");
    RemoveFile(tdir, tprefix, tfile, ".ilg");
    RemoveFile(tdir, tprefix, tfile, ".ind");
    RemoveFile(tdir, tprefix, tfile, ".ing");

    RemoveFile(".", tprefix, tfile, ".toc");
    RemoveFile(".", tprefix, tfile, ".log");
    RemoveFile(".", tprefix, tfile, ".aux");
    RemoveFile(".", tprefix, tfile, ".glo");
    RemoveFile(".", tprefix, tfile, ".idx");
    RemoveFile(".", tprefix, tfile, ".ind");
    RemoveFile(".", tprefix, tfile, ".ing");   
  }

  return(ret);
}


#if 0
/*    \TEX{
   
      \title{Compiling |LaTeX| files to obtain |dvi| documentation.}
   
      \input{A target |target|, the directory |ddir| which will
        receive the |dvi| files, a prefix |dprefix| which will be
        added to this |dvi| files, the directory |tdir| which contains
        the |LaTeX| files, the prefix |tprefix| of these |LaTeX|
        files, the names |tfiles| of these |LaTeX| files (without the
        usual suffix |.tex|) and directories |inputs| where style
        |LaTeX| files can be found.}
   
      \output{The integer 1 if an error occured, 0 otherwise.}
   
      \side{The |LaTeX| files are compiled if |target| is non zero.}

}  */
int TexToDvi(target, ddir, dprefix, tdir, tprefix, tfiles, inputs)
     int target;
     char *ddir; char *dprefix;
     char *tdir; char *tprefix; char *tfiles;
     char *inputs;{}
#endif

#ifndef TexToDvi
# define TexToDvi _TexToDvi
#endif
extern int _TexToDvi _H((int target, char *ddir, char *dprefix, char *tdir, char *tprefix, char *tfiles, char *inputs));
int _TexToDvi(target,
              ddir, dprefix,
              tdir, tprefix, tfiles,
              inputs)
     int target;
     char *ddir; char *dprefix;
     char *tdir; char *tprefix; char *tfiles;
     char *inputs;
{
  int ret;
  char *tfile, *next, *texinputs, *indexstyle, *dir;

  if (!target) {return(0);}
  ret = 0;

  MkdirHier(ddir);
  ChangeDirectory(ddir);
  
  texinputs= concat("TEXINPUTS=", tdir);
  append(&texinputs, ":"); append(&texinputs, inputs);
  append(&texinputs, ":"); append(&texinputs, ZMAKE_DIR); append(&texinputs, "/tex");
  ExpandedFileName(&dir, texinputs, PATHSEP);
  ZFree(texinputs); texinputs=dir;
  /* From TeX manual:``An empty path component  will  be replaced with the paths
   * defined in the texmf.cnf file'', so if TEXINPUTS is not set we must be sure 
   * that TEXINPUTS end with a colon -- Bill Allombert.*/
  append(&texinputs, ":");
  if (zgetenv("TEXINPUTS")) {
    append(&texinputs, zgetenv("TEXINPUTS"));
  }
  if (VERBOSE >= 50) {
    zfprintf(stderr, "export %s\n", texinputs);
  }
  zputenv(texinputs);

  indexstyle = concat("INDEXSTYLE=", inputs);
  append(&indexstyle, ":"); append(&indexstyle, ZMAKE_DIR); append(&indexstyle, "/tex");
  ExpandedFileName(&dir, indexstyle, PATHSEP);
  ZFree(indexstyle); indexstyle=dir;
  if (VERBOSE >= 50) {
    zfprintf(stderr, "export %s\n", indexstyle);
  }
  zputenv(indexstyle);

  for(next = ExpandedFileName(&tfile, tfiles, PATHSEP);
      next;
      next = ExpandedFileName(&tfile, next, PATHSEP)) {
    if (TexToOneDvi(dprefix, tfile,
                    tdir, tprefix, tfile)) {ret = 1;}
    ZFree(tfile);
  }

  ZFree(indexstyle);
  ZFree(texinputs);
  ZFree(tfile);
  return(ret);
}

/*
  \TEX{
  \Section{Removing files}
  } */

#if 0
/* \TEX{
   
    \title{Removing files}
   
    \input{A target |target|, the directory |path| which contains
      files to remove, a prefix |prefix| of the files to remove, the
      names |files| of to remove, the suffix |suffix| of the files to
      remove.}
   
    \output{The integer 1 if an error occured, 0 otherwise.}
   
    \side{The files |prefix|-|names|-|suffix| are removed from
   |path|. If |path| is empty once these files are removed, |path| is
   removed too.}
   
      }  */
void CleaningRules(target, path, prefix, files, suffix)
     int target;
     char *path;
     char *prefix;
     char *files;
     char *suffix;{}
#endif

#ifndef CleaningRules
# define CleaningRules _CleaningRules
#endif
extern void _CleaningRules _H((int target, char *path, char *prefix, char *files, char *suffix));
void _CleaningRules(target, path, prefix, files, suffix)
     int target;
     char *path;
     char *prefix;
     char *files;
     char *suffix;
{
  char *next, *file, *name;
  
  if (!target) {return;}
  
  /* Removing files if possible */
  for(next = ExpandedFileName(&file, files, PATHSEP);
      next;
      next = ExpandedFileName(&file, next, PATHSEP)) {
    RemoveFile(path, prefix, file, suffix);
    ZFree(file);
  }
  ZFree(file);

  /* Removing path if possible*/
  RmDirHier(path);
}

#if 0
/*    \TEX{
   
      \title{Removing directories}
   
      \input{A target |target|, the directory |path| which contains
        files to remove.}
   
      \output{The integer 1 if an error occured, 0 otherwise.}
   
      \side{The directory and the files contained in this directory
   are removed.}
   
   }  */
void TidyingRules(target, path)
     int target;
     char *path;{}
#endif

#ifndef TidyingRules
# define TidyingRules _TidyingRules
#endif
extern void _TidyingRules _H((int target, char *path));
void _TidyingRules(target, path)
     int target;
     char *path;
{
  if (!target) {return;}

  RemoveDirInAnyCase(path);
}
