/*
 * zmake/ztools.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef stdio_apis
# include <stdio.h>
#endif

#ifndef stdlib_apis
# include <stdlib.h>
#endif

#ifndef types_apis
# include <sys/types.h>
#endif

#ifndef stat_apis
# include <sys/stat.h>
typedef struct stat zstatt;
#endif

#ifndef fcntl_apis
# include <fcntl.h>
#endif

#ifndef unistd_apis
# include <unistd.h>
#endif

#ifndef wait_apis
# include <sys/wait.h>
#endif

#ifndef string_apis
# include <string.h>
#endif

#ifndef dirent_apis
# include <dirent.h>
typedef DIR zDIR;
typedef struct dirent zdirent;
#endif

/*
 *  System calls used by zmake
 ******************************/

/* fopen(), fclose(), fprintf()
   Introduced by stdio_apis */
#ifndef zfcreat
# define zfcreat(path) fopen(path, "w")
#endif
#ifndef zfopen
# define zfopen(path)  fopen(path, "r")
#endif
#ifndef zfprintf
# define zfprintf fprintf
#endif
#ifndef zfclose
# define zfclose(file) fclose(file)
#endif
#ifndef zsscanf
# define zsscanf sscanf
#endif

/* malloc(), free(), exit()
   Introduced by stdlib_apis */
#ifndef zmalloc
# define zmalloc(a) malloc(a)
#endif
#ifndef zfree
# define zfree(a)   free(a)
#endif
#ifndef zexit
# define zexit(status) exit(status)
#endif
#ifndef zputenv
# define zputenv(var) putenv(var)
#endif
#ifndef zgetenv
# define zgetenv(var) getenv(var)
#endif

/* open(), close(), read(),  write()...
   Introduced by types_apis, stat_apis, fcntl_apis, unistd_apis */
#ifndef zcreat
# define zcreat(path) open((path), O_CREAT|O_WRONLY|O_TRUNC, 0666)
#endif
#ifndef zopen
# define zopen(path)  open((path), O_RDONLY)
#endif
#ifndef zdup
# define zdup(file)   dup(file)
#endif
#ifndef zclose
# define zclose(file) close(file)
#endif
#ifndef zread
# define zread(file, buf, size)  read(file, buf, size)
#endif
#ifndef zwrite
# define zwrite(file, buf, size) write(file, buf, size)
#endif

#ifndef zstat
# define zstat(path, st) stat(path, st)
#endif
#ifndef zchmod
# define zchmod(path, mode) chmod(path, mode)
#endif
#ifndef ZS_ISDIR
# define ZS_ISDIR(st) S_ISDIR(st)
#endif
#ifndef ZS_IXUSR
# define ZS_IXUSR(st) S_IXUSR(st)
#endif

#ifndef zchdir
# define zchdir(path)       chdir(path)
#endif
#ifndef zmkdir
# define zmkdir(path)       mkdir(path, 0777)
#endif
#ifndef zunlink
# define zunlink(path)      unlink(path)
#endif
#ifndef zrmdir
# define zrmdir(path)       rmdir(path)
#endif
#ifndef zrename
# define zrename(old, new)  rename(old, new)
#endif
#ifndef zpwd
# define zpwd() getcwd(NULL, 2048)
#endif

/* opendir(), readdir(), closedir()
   Introduced by dirent_apis */
#ifndef zopendir
# define zopendir(path) opendir(path)
#endif
#ifndef zreaddir
# define zreaddir(dir)  readdir(dir)
#endif
#ifndef zclosedir
# define zclosedir(dir) closedir(dir)
#endif

/* fork(), execvp()
   Introduced by unistd_apis */
#ifndef zfork
# define zfork() fork()
#endif
#ifndef zexecvp
# define zexecvp(path, argv) execvp(path, argv)
#endif

/* wait()
   Introduced by wait_apis */
#ifndef zwait
# define zwait(p_status) wait(p_status)
#endif
#ifndef ZWIFEXITED
# define ZWIFEXITED(status) WIFEXITED(status)
#endif
#ifndef ZWEXITSTATUS
# define ZWEXITSTATUS(status) WEXITSTATUS(status)
#endif

/* strcmp, strncmp
   Introduced by string_apis */
#ifndef zstrcmp
# define zstrcmp(s1, s2) strcmp(s1, s2)
#endif
#ifndef zstrncmp
# define zstrncmp(s1, s2, size) strncmp(s1, s2, size)
#endif
#ifndef zstrlen
# define zstrlen(s) strlen(s)
#endif
#ifndef zstrcat
# define zstrcat(s1, s2) strcat(s1, s2)
#endif
#ifndef zstrcpy
# define zstrcpy(s1, s2) strcpy(s1, s2)
#endif
#ifndef zstrncpy
# define zstrncpy(s1, s2, s) strncpy(s1, s2, s)
#endif

/*
 *  Types used by zmake
 ******************************/

typedef struct entry {
  dev_t device;
  ino_t inode;
  struct entry *next;
} Entry;
typedef Entry *Entries;

/*
 * Some Macros
 ***************/

#define ZAlloc(ptr, size, type)                                                 \
if (1) {                                                                                \
  if (!((ptr) = (type*)zmalloc(sizeof(type)*(size)))) {                                 \
    zfprintf(stderr, "I can't allocate %d bytes, I give up.\n", sizeof(type)*(size));   \
    zexit(-1);                                                                          \
  }                                                                                     \
}                                                                                       \
else 

#define ZFree(ptr) zfree(ptr)

extern int VERBOSE;

/* 
 * Headers
 **********/
#if AcceptPrototypes
# define _H(x) x
#else
# define _H(x) ()
#endif

extern void FreeList _H((Entries heap));
extern int AppendInList _H((Entries *heap, dev_t device, ino_t inode));
extern char *concat _H((char *s1, char *s2));
extern void append _H((char **s1, char *s2));
extern char *ExpandedFileName _H((char **path, char *names, char *separator));
extern int ParseFile _H((char **list, char *name, char *inc_lhs, char *inc_rhs));
extern char *PathPrefixFileSuffix _H((char *path, char *separator, char *prefix, char *file, char *suffix));
extern int NewerThanHeart _H((char *l_path, char *l_prefix, char *l_root, char *l_suffix, char *r_pathes, char *r_prefix, char *r_roots, char *r_suffix, char *separator, char *inc_lhs, char *inc_rhs, Entries *p_heap));
extern int NewerThan _H((char *l_path, char *l_prefix, char *l_root, char *l_suffix, char *r_pathes, char *r_prefix, char *r_roots, char *r_suffix, char *separator, char *inc_lhs, char *inc_rhs));
extern int Execute _H((char *command, char *separator, char *input, char *output));
extern int Target _H((char *target, int argc, char *argv[]));
extern int ReadIntOption _H((int *p_var, char *str, int argc, char *argv[]));
extern int ReadStringOption _H((char **p_var, char *str, int argc, char *argv[]));

/*
  \TEX{
  \Section{Handling targets}
  } */

/* \TEX{ 
   \title{Is a target specified on the command line.}
   \input{A target |target| and the well known C arguments |argc| and |argv|.}
   \output{1 if target was specified on the command line, 0 otherwise.}
} */
int Target(target, argc, argv)
     char *target;
     int argc;
     char *argv[];
{
  for(argc--; argc; argc--) {
    if (!zstrcmp(argv[argc], target)) {
      return(1);
    }
  }
  return(0);
}

/* Reading integer variable from command line */
int ReadIntOption(p_var, str, argc, argv)
     int *p_var;
     char *str;
     int argc;
     char *argv[];
{
  for(argc--; argc; argc--) {
    if (!zstrncmp(argv[argc], str, zstrlen(str))) {
      char *ptr;

      zsscanf(argv[argc]+zstrlen(str), "=%d", p_var);
      return(1);
    }
  }
  return(0);
}

/* Reading string variable from command line */
int ReadStringOption(p_var, str, argc, argv)
     char **p_var;
     char *str;
     int argc;
     char *argv[];
{
  for(argc--; argc; argc--) {
    if (!zstrncmp(argv[argc], str, zstrlen(str))) {
      char *ptr;

      ptr = argv[argc]+zstrlen(str);
      if (*ptr == '=') {*p_var = ptr+1;}
      return(1);
    }
  }
  return(0);
}

#if 0
/* \TEX{ 
   \title{Target implications.}
   \input{A target |target| and subtargets |subtargers|.}
   \side{Subtargets are set to one if target is non |NULL|.}
} */
void TargetImply(target, subtargets)
     int target;
     int subtargets;{}
#endif
#define TargetImply(target, subtargets) if (target) {subtargets = 1;}


/*
  \TEX{
  \Section{Handling paths}
  } */


/* Handling list of inodes
 **************************/

/* Freeing a list of inodes */
void FreeList(heap)
     Entries heap;
{
  Entries ptr;

  for(; heap; heap = ptr) {
    ptr = heap->next;
    ZFree(heap);
  }
}

/* Adding an inode to a list */
int AppendInList(heap, device, inode)
     Entries *heap;
     dev_t device;
     ino_t inode;
{
  Entries *ptr;
  
  for(ptr = heap; *ptr; ptr = &((*ptr)->next)) {
    if (((*ptr)->inode == inode) && 
        ((*ptr)->device == device)) {
      return(-1);
    }
  }
  ZAlloc(*ptr, 1, Entry);
  (*ptr)->next = NULL;
  (*ptr)->inode = inode;
  (*ptr)->device = device;
  return(0);
}

/* \TEX{ 
   \title{Concatening 2 strings.}
   \input{Two strings |s1| and |s2| to concatenate.}
   \output{An allocated strings which contains the concatenation of
   |s1| and |s2|.}
} */
char *concat(s1, s2)
     char *s1; char *s2;
{
  char *s;

  ZAlloc(s, zstrlen(s1)+zstrlen(s2)+1, char);
  zstrcpy(s, s1); zstrcat(s, s2);
  return(s);
}

/* \TEX{ 
   \title{Appending a string to another.}  
   \input{A string |s2| to append to |s1|.}
   \side{The string |s1| contains the concatenation of |s1| and |s2|.}
} */
void append(s1, s2)
     char **s1; char *s2;
{
  char *s;
  
  s = *s1; *s1 = concat(s, s2); ZFree(s);
}


/* Replacing `\ ' by `` '' and `\/` by ``separator'' in a zmake path
 *********************************************************************/
char *ExpandedFileName(path, names, separator)
     char **path;
     char *names;
     char *separator;
{
  char *end, *begin;
  int nb, i;

  /* Initialisation */
  ZAlloc(*path, 1, char); (*path)[0] = '\0';

  /* Skipping spaces */
  for(end = names; (*end) && ((*end) == ' '); end++) {;}
  if (!(*end)) {return(NULL);}
  
  /* Extracting the system path */
  for(; (*end) && ((*end) != ' '); end++) {
    char *current, *ptr;
    
    /* Getting the current name */
    nb = 0;
    for(begin = end; *end; end++) {
      if (((*end) == ' ') || ((*end) == '/')) {
        if ((end > begin) && (*(end-1) == '\\')) {nb++;} else {break;}
      }
    }
    
    /* Extracting the current name */
    ZAlloc(current, end-begin+1-nb, char);
    for(i = 0; begin < end; begin++) {
      if (((*begin) == ' ') || ((*begin) == '/')) {
        current[i-1] = *begin;
      }
      else {
        current[i++] = *begin;
      }
    }
    current[i] = '\0';

    /* Adding it to path */
    ptr = concat(*path, current);
    ZFree(current);
    ZFree(*path); *path = ptr;
    
    /* Do we have to loop again ? */
    if (!(*end) || ((*end) != '/')) {break;}
    ptr = concat(*path, separator);
    ZFree(*path); *path = ptr;
  }

  /* Output */
  return(end);
}

/* Looking for included files in another one
 ********************************************/
#define SIZE_BUF 1024
#define DEFAULT     0
#define HS_PARSE    1
#define HS_BLANK    2
#define NAME_PARSE  3

int ParseFile(list, name, inc_lhs, inc_rhs)
     char **list;
     char *name;
     char *inc_lhs; char *inc_rhs;
{
  char buf[SIZE_BUF+1];
  int fd, size, rhs, cut, state;
  char *ref, *tmp, *ptr;
  char *begin, *end;

  if (VERBOSE >= 100) {
    zfprintf(stderr, "Parsing %s\n", name);
  }

  /* Initialisation */
  ZAlloc(*list, 1, char); (*list)[0] = '\0';
  buf[SIZE_BUF] = '\0';
  
  /* Opening the file */
  if ((fd = zopen(name)) == -1) {return(-1);}
  
  /* Reading the file */
  state = DEFAULT; rhs = 0; cut = 0;
  ref = inc_lhs;
  while((size = zread(fd, buf, SIZE_BUF)) > 0) {
    
    for(ptr = buf; ptr < buf+size; ptr++) {
      switch(state) {

      case DEFAULT:                     /* No include directive */
        if (rhs) {
          if (*ptr != ' ') {
            state = HS_PARSE; ptr--;
          }
        }
        else {
          if (*ptr == *ref) {
            state = HS_PARSE; ref++;
          }
        }
        break;

      case HS_PARSE:                    /* Body of a hs include directive */
        if (!*ref) {
          if (rhs) {
            char c;

            c = *ptr; *ptr = '\0';
            if (!cut) {
              tmp = concat(*list, " "); ZFree(*list);
            }
            else {
              tmp = concat(*list, begin); ZFree(*list);
              ZFree(begin); begin = buf;
            }
            *list = concat(tmp, begin); ZFree(tmp);
            *ptr = c;
            cut = 0; rhs = 0;  ptr--; ref = inc_lhs;
            state = DEFAULT;
          }
          else {
            begin = ptr; ptr--;
            ref = inc_rhs; rhs = 1;
            state = NAME_PARSE;
          }
        }
        else if (*ref == ' ') {
          if (*ptr != ' ') {
            ptr--; ref++;
          }
          else {
            ref++;
            state = HS_BLANK;
          }
        }
        else {
          if (*ptr == *ref) {
            *ptr = '\0'; ref++;
          }
          else {
            ref = inc_lhs; rhs = 0; cut = 0;
            state = DEFAULT;
          }
        }
        break;

      case HS_BLANK:                    /* Spaces in a hs include directive */
        if (*ptr != ' ') {
          ptr--;
          state = HS_PARSE;
        }
        break;

      case NAME_PARSE:                  /* The included file */
        if (*ptr == *ref) {
          end = ptr; ref = inc_rhs; ptr--;
          state = DEFAULT;
        }
        break;

      }
    }
    
    if (rhs) {
      if (cut) {
        tmp = concat(begin, buf);       
        ZFree(begin); begin = tmp;
      }
      else {
        begin = concat(" ", begin);
        cut = 1;
      }
    }

  }
  if (cut) {ZFree(begin);}
  
  /* Closing the file */
  zclose(fd);

  return(0);
}

char *PathPrefixFileSuffix(path, separator, prefix, file, suffix)
     char *path;
     char *separator;
     char *prefix;
     char *file;
     char *suffix;
{
  char *name, *tmp, *ptr;
  
  ExpandedFileName(&tmp, path, separator);
  name = tmp;

  ptr = name; name = concat(ptr, separator); ZFree(ptr);

  ExpandedFileName(&tmp, prefix, separator);
  ptr = name; name = concat(ptr, tmp); ZFree(ptr);
  ZFree(tmp);

  ExpandedFileName(&tmp, file, separator);
  ptr = name; name = concat(ptr, tmp); ZFree(ptr);
  ZFree(tmp);

  ExpandedFileName(&tmp, suffix, separator);
  ptr = name; name = concat(ptr, tmp); ZFree(ptr);
  ZFree(tmp);
  
  return(name);
}


/* Emulating a Makefile's target
 * Return:
 *     -1 if l_name is older than the right hand side file,
 *     0 if l_name does not exist,
 *     1 otherwise
 ***************************************************/
int NewerThanHeart(l_path, l_prefix, l_root, l_suffix,
                   r_pathes, r_prefix, r_roots, r_suffix,
                   separator,
                   inc_lhs, inc_rhs,
                   p_heap)
     char *l_path; char *l_prefix; char *l_root; char *l_suffix;
     char *r_pathes; char *r_prefix; char *r_roots; char *r_suffix;
     char *separator;
     char *inc_lhs; char *inc_rhs;
     Entries *p_heap;
{
  char *tmp1, *tmp2;
  char *l_name, *r_name;
  char *cfile, *cpath, *r_path, *r_root;
  char *r_suffix_path, *r_prefix_path;
  zstatt ls, rs;
  int ret;

  /* Expanding the left hand side */
  l_name = PathPrefixFileSuffix(l_path, separator, l_prefix, l_root, l_suffix);
  if (VERBOSE >= 100) {
    zfprintf(stderr, "The LHS is \"%s\".\n", l_name);
  }

  /* Getting the modified time of l_name */
  ret = zstat(l_name, &ls);
  ZFree(l_name);
  if (ret) {return(0);}
  
  /* Expanding the right hand side */
  ExpandedFileName(&r_prefix_path, r_prefix, separator);
  ExpandedFileName(&r_suffix_path, r_suffix, separator);

  for(cfile = ExpandedFileName(&r_root, r_roots, separator);
      cfile;
      cfile = ExpandedFileName(&r_root, cfile, separator)) {
      
    for(cpath = ExpandedFileName(&r_path, r_pathes, separator);
        cpath;
        cpath = ExpandedFileName(&r_path, cpath, separator)) {

      r_name = PathPrefixFileSuffix(r_path, separator, r_prefix, r_root, r_suffix);
      ZFree(r_path);
  
      if (!zstat(r_name, &rs)) {

        if (VERBOSE >= 100) {
          zfprintf(stderr, "Testing the LHS versus %s\n", r_name);
        }

        if (ls.st_mtime < rs.st_mtime) {
          ZFree(r_name);
          ZFree(r_prefix_path);
          ZFree(r_root);
          ZFree(r_suffix_path);
          return(-1);
        }

        if (inc_lhs && !AppendInList(p_heap, rs.st_dev, rs.st_ino)) {
          char *list;
          
          ParseFile(&list, r_name, inc_lhs, inc_rhs);

          ret = NewerThanHeart(l_path, l_prefix, l_root, l_suffix,
                               r_pathes, "", list, "",
                               separator, inc_lhs, inc_rhs,
                               p_heap);

          ZFree(list);
          if (ret < 0) {
            ZFree(r_name);
            ZFree(r_prefix_path);
            ZFree(r_root);
            ZFree(r_suffix_path);
            return(ret);
          }
        }
      }
      ZFree(r_name);
    }
    ZFree(r_path);
    ZFree(r_root);
  }
  ZFree(r_root);
  ZFree(r_prefix_path);
  ZFree(r_suffix_path);
  return(1);
}

int NewerThan(l_path, l_prefix, l_root, l_suffix,
              r_pathes, r_prefix, r_roots, r_suffix,
              separator,
              inc_lhs, inc_rhs)
     char *l_path; char *l_prefix; char *l_root; char *l_suffix;
     char *r_pathes; char *r_prefix; char *r_roots; char *r_suffix;
     char *separator;
     char *inc_lhs; char *inc_rhs;
{
  int ret;
  static Entries heap;
  
  heap = NULL;
  ret = NewerThanHeart(l_path, l_prefix, l_root, l_suffix,
                       r_pathes, r_prefix, r_roots, r_suffix,
                       separator,
                       inc_lhs, inc_rhs,
                       &heap);
  FreeList(heap);
  return(ret);
}

int Execute(command, separator, input, output)
     char *command;
     char *separator;
     char *input; char *output;
{
  char *args, *arg, **argv;
  int argc;
  int status;
  pid_t pid;

  /* How many arguments ? */
  argc = 0;
  for(argc = 0, args = ExpandedFileName(&arg, command, separator);
      args;
      argc++, args = ExpandedFileName(&arg, args, separator)) {ZFree(arg);}
  ZFree(arg);

  /* Filling argv */
  ZAlloc(argv, argc+1, char *);
  for(argc = 0, args = ExpandedFileName(&(argv[argc]), command, separator);
      args;
      argc++, args = ExpandedFileName(&(argv[argc]), args, separator)) {;}
  ZFree(argv[argc]);
  argv[argc] = NULL;

  /* fork() call */
  pid = fork ();

  if (!pid) {    /* Son */
    int in, out;

    if (input && ((in = zopen(input)) == -1)) {
      zfprintf(stderr, "open error on \"%s\".\n", input);
      zexit(-1);
    }
    if (output && ((out = zcreat(output)) == -1)) {
      zfprintf(stderr, "open error on \"%s\".\n", output);
      zexit(-1);
    }
    if (input) {
      zclose(0); zdup(in); zclose(in);
    }
    if (output) {
      zclose(1); zdup(out); zclose(out);
    }

    zexecvp(argv[0], argv);
    zfprintf(stderr, "execvp() error.\n");
    zexit(-1);
  }
  else if (pid  < 0) { /* Error */
    zfprintf(stderr, "fork() error, I give up.\n");
    zexit(-1);
  }

  /* Freing argv structure */
  for(argc--; argc >= 0; argc--) {ZFree(argv[argc]);}
  ZFree(argv);

  /* Waiting for the end of the previous comannd */
  for(;;) {
    pid_t _pid;
    
    _pid = zwait(&status);
    if (_pid == pid) {
      break;
    }
    else if (_pid < 0) {
      zfprintf(stderr, "wait() error, I give up.\n");
      zexit(-1);
    }
  }
  
  /* Output status */
  return(ZWIFEXITED(status) ? ZWEXITSTATUS(status) : 2147483647);
}

